Package: nmfkc
Type: Package
Title: Non-Negative Matrix Factorization with Kernel Covariates
Version: 0.8.2
Date: 2026-06-09
Authors@R: 
    person("Kenichi", "Satoh",
           email = "kenichi-satoh@biwako.shiga-u.ac.jp",
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-4436-9347"))
Maintainer: Kenichi Satoh <kenichi-satoh@biwako.shiga-u.ac.jp>
URL: https://github.com/ksatohds/nmfkc,
        https://ksatohds.github.io/nmfkc/
BugReports: https://github.com/ksatohds/nmfkc/issues
Description: Performs Non-negative Matrix Factorization (NMF)
    with Kernel Covariates. Given an observation matrix and kernel
    covariates, it optimizes both a basis matrix and a parameter matrix.
    Notably, if the kernel matrix is an identity matrix, the method
    simplifies to standard NMF. Also provides NMF with Random Effects
    (NMF-RE) via nmfre(), which estimates a mixed-effects model combining
    covariate-driven scores with unit-specific random effects together
    with wild bootstrap inference, and NMF-based Structural Equation
    Modeling (NMF-SEM) via nmf.sem(), which fits a two-block input-output
    model for blind source separation and path analysis.
    References: Satoh (2025) <doi:10.48550/arXiv.2403.05359>;
    Satoh (2025) <doi:10.48550/arXiv.2510.10375>;
    Satoh (2025) <doi:10.48550/arXiv.2512.18250>;
    Satoh (2026) <doi:10.48550/arXiv.2603.01468>;
    Satoh (2026) <doi:10.1007/s42081-025-00314-0>.
License: MIT + file LICENSE
Imports: stats, graphics, utils, grDevices
Encoding: UTF-8
Language: en-US
ByteCompile: true
VignetteBuilder: knitr
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), mclust,
        palmerpenguins, quanteda, vars, DiagrammeR, MASS, nlme, lavaan
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.2.3
NeedsCompilation: no
Packaged: 2026-06-14 11:41:16 UTC; ksato
Author: Kenichi Satoh [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4436-9347>)
Repository: CRAN
Date/Publication: 2026-06-14 13:50:02 UTC
Built: R 4.5.2; ; 2026-06-14 13:58:21 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(coef,nmf)
S3method(coef,nmf.sem)
S3method(fitted,nmf)
S3method(fitted,nmf.sem)
S3method(fitted,nmfae)
S3method(plot,nmf.cluster.criteria)
S3method(plot,nmf.cluster.flow)
S3method(plot,nmf.rank)
S3method(plot,nmf.sem)
S3method(plot,nmfae)
S3method(plot,nmfae.cv)
S3method(plot,nmfae.ecv)
S3method(plot,nmfae.kernel.beta.cv)
S3method(plot,nmfae.signed)
S3method(plot,nmfkc)
S3method(plot,nmfkc.DOT)
S3method(plot,nmfkc.ard)
S3method(plot,nmfkc.consensus)
S3method(plot,nmfkc.signed)
S3method(plot,nmfre)
S3method(plot,predict.nmfae)
S3method(predict,nmfae)
S3method(predict,nmfae.signed)
S3method(predict,nmfkc)
S3method(predict,nmfkc.signed)
S3method(print,nmf.cluster.criteria)
S3method(print,nmf.cluster.flow)
S3method(print,nmf.inference)
S3method(print,nmf.rank)
S3method(print,nmfkc.ard)
S3method(print,nmfkc.consensus)
S3method(print,nmfre.dfU.scan)
S3method(print,summary.nmf.sem)
S3method(print,summary.nmfae)
S3method(print,summary.nmfae.inference)
S3method(print,summary.nmfae.signed)
S3method(print,summary.nmfae.signed.inference)
S3method(print,summary.nmfkc)
S3method(print,summary.nmfkc.inference)
S3method(print,summary.nmfkc.net)
S3method(print,summary.nmfkc.net.inference)
S3method(print,summary.nmfkc.net.signed)
S3method(print,summary.nmfkc.signed)
S3method(residuals,nmf)
S3method(residuals,nmf.sem)
S3method(residuals,nmfae)
S3method(residuals,nmfre)
S3method(summary,nmf.sem)
S3method(summary,nmfae)
S3method(summary,nmfae.inference)
S3method(summary,nmfae.signed)
S3method(summary,nmfae.signed.inference)
S3method(summary,nmfkc)
S3method(summary,nmfkc.inference)
S3method(summary,nmfkc.net)
S3method(summary,nmfkc.net.inference)
S3method(summary,nmfkc.net.signed)
S3method(summary,nmfkc.signed)
S3method(summary,nmfre)
export(nmf.cluster.criteria)
export(nmf.cluster.flow)
export(nmf.ffb)
export(nmf.ffb.DOT)
export(nmf.ffb.cv)
export(nmf.ffb.inference)
export(nmf.ffb.split)
export(nmf.sem)
export(nmf.sem.DOT)
export(nmf.sem.cv)
export(nmf.sem.inference)
export(nmf.sem.split)
export(nmfae)
export(nmfae.DOT)
export(nmfae.cv)
export(nmfae.ecv)
export(nmfae.heatmap)
export(nmfae.inference)
export(nmfae.kernel.beta.cv)
export(nmfae.rank)
export(nmfae.rename)
export(nmfae.signed)
export(nmfae.signed.ecv)
export(nmfae.signed.heatmap)
export(nmfae.signed.inference)
export(nmfae.signed.rank)
export(nmfae.signed.rename)
export(nmfkc)
export(nmfkc.DOT)
export(nmfkc.ar)
export(nmfkc.ar.DOT)
export(nmfkc.ar.degree.cv)
export(nmfkc.ar.predict)
export(nmfkc.ar.stationarity)
export(nmfkc.ard)
export(nmfkc.bicv)
export(nmfkc.class)
export(nmfkc.consensus)
export(nmfkc.criterion)
export(nmfkc.cv)
export(nmfkc.denormalize)
export(nmfkc.ecv)
export(nmfkc.inference)
export(nmfkc.kernel)
export(nmfkc.kernel.beta.cv)
export(nmfkc.kernel.beta.nearest.med)
export(nmfkc.kernel.gaussian)
export(nmfkc.net)
export(nmfkc.net.DOT)
export(nmfkc.net.ecv)
export(nmfkc.net.inference)
export(nmfkc.net.rank)
export(nmfkc.normalize)
export(nmfkc.rank)
export(nmfkc.residual.plot)
export(nmfkc.signed)
export(nmfkc.signed.cv)
export(nmfkc.signed.ecv)
export(nmfkc.signed.rank)
export(nmfkc.signed.rff)
export(nmfre)
export(nmfre.dfU.scan)
export(nmfre.inference)

# nmfkc: Non-negative Matrix Factorization with Kernel Covariates

[![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![GitHub version](https://img.shields.io/github/r-package/v/ksatohds/nmfkc)](https://github.com/ksatohds/nmfkc)

**nmfkc** is an R package that extends Non-negative Matrix Factorization (NMF) by incorporating **covariates** using kernel methods. It supports advanced features like rank selection via cross-validation, time-series modeling (NMF-VAR), supervised classification (NMF-LAB), **feed-forward + feedback structural modeling with equilibrium interpretation (NMF-FFB; formerly NMF-SEM)**, and **mixed-effects modeling with random effects (NMF-RE)**.

# Installation

```r
# Stable version (CRAN)
install.packages("nmfkc")

# Development version (GitHub, may be unstable)
# install.packages("remotes")
remotes::install_github("ksatohds/nmfkc@develop")

library(nmfkc)
```

# Help and Usage

```r
browseVignettes("nmfkc")
ls("package:nmfkc")
?nmfkc
```

# Citation

```r
citation("nmfkc")
```

# Quick Example

```r
library(nmfkc)

# Decompose a matrix Y into basis X and coefficient B with rank = 2
X_true <- cbind(c(1, 0, 1), c(0, 1, 0))
B_true <- cbind(c(1, 0), c(0, 1), c(1, 1))
Y <- X_true %*% B_true

res <- nmfkc(Y, rank = 2, epsilon = 1e-6)
plot(res)     # Convergence plot
summary(res)  # Summary statistics
```

See `browseVignettes("nmfkc")` for detailed examples covering rank selection, kernel NMF, time-series, classification, NMF-FFB, and NMF-RE.

# Comparison with Standard NMF

| Feature | Standard NMF | nmfkc |
| :--- | :--- | :--- |
| **Handles covariates** | No | **Yes** (Linear / Kernel) |
| **Feed-forward + feedback modeling** | No | **Yes** (NMF-FFB) |
| **Mixed-effects / Random effects** | No | **Yes** (NMF-RE) |
| **Classification** | No | **Yes** (NMF-LAB) |
| **Time series modeling** | No | **Yes** (NMF-VAR) |
| **Nonlinearity** | No | **Yes** (Kernel) |
| **Clustering support** | Limited | **Yes** (Hard/Soft) |
| **Rank selection / CV** | Limited (ad hoc) | **Yes** (Element-wise CV, Column-wise CV) |

# Statistical Model

The **nmfkc** package builds upon the standard NMF framework by incorporating external information (covariates):

$$Y(P,N) \approx X(P,Q) \times C(Q,R) \times A(R,N)$$

- **$Y$**: Observation matrix ($P$ features × $N$ samples)
- **$A$**: Covariate matrix ($R$ covariates × $N$ samples); defaults to identity (standard NMF)
- **$X$**: Basis matrix — learned latent patterns
- **$C$**: Parameter matrix — links covariates to latent structure

### Extensions

- **NMF-RE**: Adds unit-specific random effects $U$: $Y = X(\Theta A + U) + \mathcal{E}$, estimated via ridge-type BLUP with wild bootstrap inference.
- **NMF-FFB** (formerly NMF-SEM): Models feed-forward + feedback structure $Y_1 \approx X(\Theta_1 Y_1 + \Theta_2 Y_2)$, with equilibrium mapping $(I - X\Theta_1)^{-1} X\Theta_2$.

# Main Functions

| Function | Description |
|:---|:---|
| `nmfkc()` | Core NMF with covariates ($Y \approx XCA$); supports kernel matrices and formula interface |
| `nmfre()` / `nmfre.inference()` | NMF with Random Effects + wild bootstrap inference |
| `nmf.ffb()` / `nmf.ffb.inference()` | NMF Feed-Forward + Feedback model (formerly `nmf.sem*`, retained as alias) + inference for path coefficients |
| `nmfae()` / `nmfae.inference()` | NMF Autoencoder + inference |
| `nmfkc.rank()` | Rank selection via elbow, cross-validation, ECV, and CPCC |
| `nmfkc.inference()` | Sandwich SE and wild bootstrap p-values for `nmfkc` |
| `nmfkc.DOT()` / `nmfkc.ar.DOT()` / `nmf.ffb.DOT()` / `nmfae.DOT()` | Graphviz path diagrams; render with `plot()` |

S3 methods `coef()`, `fitted()`, `residuals()`, `plot()`, `summary()`, `predict()` are available for all model classes. See `?nmfkc` or `browseVignettes("nmfkc")` for the full function list.

# References

- Satoh, K. (2024). Applying Non-negative Matrix Factorization with Covariates to the Longitudinal Data as Growth Curve Model. arXiv:2403.05359. <https://arxiv.org/abs/2403.05359>
- Satoh, K. (2025). Applying non-negative Matrix Factorization with Covariates to Multivariate Time Series Data as a Vector Autoregression Model. *Japanese Journal of Statistics and Data Science*. <https://doi.org/10.1007/s42081-025-00314-0>
- Satoh, K. (2025). Applying non-negative matrix factorization with covariates to label matrix for classification. arXiv:2510.10375. <https://arxiv.org/abs/2510.10375>
- Satoh, K. (2025). Applying non-negative matrix factorization with covariates to structural equation modeling for blind input-output analysis. arXiv:2512.18250. <https://arxiv.org/abs/2512.18250>
- Satoh, K. (2026). Wild Bootstrap Inference for Non-Negative Matrix Factorization with Random Effects. arXiv:2603.01468. <https://arxiv.org/abs/2603.01468>

## ============================================================
## nmfkc.consensus(): consensus-clustering rank selection for NMF
## (Brunet et al. 2004).  A lightweight engine in the spirit of
## nmfkc.ecv / nmfkc.bicv: returns a stability score per rank.
## ============================================================

#' @title Consensus-clustering rank selection for NMF (Brunet 2004)
#' @description
#' The bioinformatics-standard stability approach to choosing the NMF rank.
#' A lightweight engine like \code{\link{nmfkc.ecv}} / \code{\link{nmfkc.bicv}}:
#' it returns one stability score per rank and nothing more.
#'
#' For each rank, NMF is run \code{nrun} times from different random
#' initializations (\code{X.init = "runif"}).  Each run gives a hard
#' clustering of the samples (the column \eqn{\arg\max} of the coefficient
#' matrix).  Averaging the \eqn{N \times N} connectivity matrices (1 if two
#' samples share a cluster) over the runs yields the \strong{consensus
#' matrix}; its crispness measures how reproducible the clustering is.  Two
#' summaries are reported per rank:
#' \itemize{
#'   \item \code{cophenetic}: the cophenetic correlation coefficient (CPCC)
#'     of the consensus matrix (Brunet et al. 2004).  Close to 1 = stable.
#'   \item \code{dispersion}: the Kim & Park (2007) dispersion
#'     \eqn{\frac{1}{N^2}\sum_{ij} 4 (C_{ij} - 1/2)^2 \in [0, 1]}; 1 when
#'     every consensus entry is exactly 0 or 1 (perfectly crisp).
#'   \item \code{pac}: the Proportion of Ambiguous Clustering
#'     (Senbabaoglu et al. 2014) -- the fraction of off-diagonal consensus
#'     entries falling in the ambiguous interval \code{pac.range}
#'     (default \eqn{(0.1, 0.9)}).  \strong{Lower is better} (less
#'     ambiguity); a more sensitive readout than \code{cophenetic}, which
#'     tends to saturate.
#' }
#' Unlike the cross-validation engines (where the rank \emph{minimizes} the
#' error), here a good rank \strong{maximizes} stability, or is the largest
#' rank before it drops.
#'
#' @param Y Observation matrix (\eqn{P \times N}), non-negative.
#' @param A Optional covariate matrix passed to \code{\link{nmfkc}}.
#' @param rank Integer vector of ranks to evaluate (\eqn{\ge 2}).
#' @param nrun Number of random-initialization runs per rank (default 30).
#' @param keep.consensus Logical; if \code{TRUE} also return the list of
#'   consensus matrices (one \eqn{N \times N} matrix per rank).
#' @param ... Advanced options, rarely needed (defaults in parentheses):
#'   \code{seed} (\code{123}, base seed; run \eqn{r} of rank index \eqn{i} uses
#'   \code{seed + 1000 * i + r}) and \code{pac.range} (\code{c(0.1, 0.9)}, the
#'   ambiguous interval \eqn{(u_1, u_2)} for the PAC measure).  Any other
#'   arguments are passed to \code{\link{nmfkc}} (e.g.\ \code{maxit});
#'   \code{X.init} is forced to \code{"runif"}.
#' @return An object of class \code{"nmfkc.consensus"} (a list) with:
#'   \item{cophenetic}{Cophenetic correlation coefficient for each rank.}
#'   \item{dispersion}{Dispersion coefficient (\eqn{[0, 1]}) for each rank.}
#'   \item{pac}{Proportion of Ambiguous Clustering (\eqn{[0, 1]}, lower is
#'     better) for each rank.}
#'   \item{rank}{The evaluated rank vector.}
#'   \item{nrun}{Number of runs per rank.}
#'   \item{consensus}{List of consensus matrices, or \code{NULL}.}
#'   It has \code{\link{print.nmfkc.consensus}} and
#'   \code{\link{plot.nmfkc.consensus}} (\code{type = "criteria"} /
#'   \code{"heatmap"}) methods.
#' @references
#' Brunet, J.-P., Tamayo, P., Golub, T. R., Mesirov, J. P. (2004).
#' Metagenes and molecular pattern discovery using matrix factorization.
#' \emph{PNAS} 101(12):4164--4169. \doi{10.1073/pnas.0308531101}.
#' Kim, H., Park, H. (2007).  Sparse non-negative matrix factorizations
#' \dots \emph{Bioinformatics} 23(12):1495--1502.
#' Senbabaoglu, Y., Michailidis, G., Li, J. Z. (2014).  Critical
#' limitations of consensus clustering in class discovery.
#' \emph{Sci. Rep.} 4:6207. \doi{10.1038/srep06207}.
#' @seealso \code{\link{nmfkc.rank}}, \code{\link{nmfkc.ecv}},
#'   \code{\link{nmfkc.bicv}}, \code{\link{nmfkc.ard}},
#'   \code{\link{nmf.cluster.criteria}}
#' @export
#' @examples
#' \donttest{
#' Y <- t(as.matrix(iris[, 1:4]))
#' cs <- nmfkc.consensus(Y, rank = 2:5, nrun = 20, keep.consensus = TRUE)
#' cs                       # stability table per rank
#' plot(cs)                 # type = "criteria": stability curves
#' plot(cs, type = "heatmap")            # all ranks, n2mfrow grid
#' plot(cs, type = "heatmap", rank = 3)  # one rank, with labels
#' }
nmfkc.consensus <- function(Y, A = NULL, rank = 2:4, nrun = 30,
                            keep.consensus = FALSE, ...) {
  Y <- base::as.matrix(Y)
  N <- base::ncol(Y)

  ## Advanced options live in `...` (the rest passes to nmfkc per run).
  dots      <- base::list(...)
  seed      <- if (base::is.null(dots$seed))      123         else dots$seed
  pac.range <- if (base::is.null(dots$pac.range)) c(0.1, 0.9) else dots$pac.range
  dots$seed <- NULL; dots$pac.range <- NULL

  ## nmfkc args: random init (required for stability), plain fast fit.
  ea <- dots; ea$Q <- NULL; ea$rank <- NULL; ea$seed <- NULL
  ea$X.init <- "runif"; ea$detail <- "fast"; ea$print.dims <- FALSE

  cophenetic <- stats::setNames(base::numeric(base::length(rank)),
                                base::paste0("rank=", rank))
  dispersion <- cophenetic
  pac        <- cophenetic
  consensus.list <- if (keep.consensus)
    base::vector("list", base::length(rank)) else NULL

  base::message(base::sprintf(
    "consensus: ranks %s, %d runs/rank (Brunet 2004); %d fits total...",
    base::paste(rank, collapse = ","), nrun, nrun * base::length(rank)))

  for (ki in base::seq_along(rank)) {
    k <- rank[ki]
    Cmat <- base::matrix(0, N, N)
    for (r in base::seq_len(nrun)) {
      fit <- base::suppressMessages(base::do.call("nmfkc",
        c(base::list(Y = Y, A = A, rank = k,
                     seed = seed + 1000L * ki + r), ea)))
      cl <- base::apply(fit$B, 2, base::which.max)   # hard cluster per sample
      Cmat <- Cmat + (base::outer(cl, cl, "==") + 0) # connectivity
    }
    Cmat <- Cmat / nrun                              # consensus in [0, 1]
    if (keep.consensus) {
      if (!base::is.null(base::colnames(Y)))
        base::dimnames(Cmat) <- base::list(base::colnames(Y), base::colnames(Y))
      consensus.list[[ki]] <- Cmat
    }

    ## Kim-Park dispersion: 1 = perfectly crisp (all 0/1).
    dispersion[ki] <- base::sum(4 * (Cmat - 0.5)^2) / (N * N)

    ## PAC (Senbabaoglu 2014): proportion of off-diagonal consensus entries
    ## in the ambiguous interval (pac.range); LOWER = crisper / better.
    lt <- Cmat[base::lower.tri(Cmat)]
    pac[ki] <- base::mean(lt > pac.range[1] & lt < pac.range[2])

    ## Cophenetic correlation of the consensus (Brunet).
    d <- stats::as.dist(1 - Cmat)
    if (k >= 2 && base::any(d > 0)) {
      hc <- stats::hclust(d, method = "average")
      cophenetic[ki] <- stats::cor(base::as.vector(d),
                                   base::as.vector(stats::cophenetic(hc)))
    } else {
      cophenetic[ki] <- NA_real_
    }
  }

  out <- base::list(cophenetic = cophenetic, dispersion = dispersion,
                    pac = pac, rank = rank, nrun = nrun,
                    consensus = consensus.list)
  base::class(out) <- "nmfkc.consensus"
  out
}


#' @title Print method for nmfkc.consensus objects
#' @param x An object of class \code{"nmfkc.consensus"}.
#' @param ... Unused.
#' @return \code{x}, invisibly.
#' @export
print.nmfkc.consensus <- function(x, ...) {
  base::cat(base::sprintf("Consensus rank selection (Brunet 2004), %d runs/rank\n",
                          x$nrun))
  base::cat(base::sprintf("  best: dispersion max = rank %d | PAC min = rank %d\n",
                          x$rank[base::which.max(x$dispersion)],
                          x$rank[base::which.min(x$pac)]))
  df <- base::data.frame(rank = x$rank,
                         cophenetic = base::round(x$cophenetic, 3),
                         dispersion = base::round(x$dispersion, 3))
  if (!base::is.null(x$pac)) df$pac <- base::round(x$pac, 3)
  base::print(df, row.names = FALSE)
  if (base::is.null(x$consensus))
    base::cat("  (consensus matrices not stored; use keep.consensus = TRUE for heatmaps)\n")
  base::invisible(x)
}


#' @title Plot a consensus rank-selection (nmfkc.consensus) object
#' @description
#' Two views of a \code{\link{nmfkc.consensus}} result:
#' \itemize{
#'   \item \code{type = "criteria"} (default): the stability curves
#'     \code{cophenetic} (blue) and \code{dispersion} (red) against rank.
#'     No "best" marker is drawn -- a stable rank tends to be a coarse one,
#'     so the curves are shown for inspection rather than as an optimum.
#'   \item \code{type = "heatmap"}: the consensus matrix of each requested
#'     \code{rank}, reordered by average-linkage hierarchical clustering of
#'     \eqn{1 - \bar C} (blue = 0 / different cluster, red = 1 / same
#'     cluster).  Requires \code{keep.consensus = TRUE} at compute time.
#' }
#' @param x An object of class \code{"nmfkc.consensus"}.
#' @param type \code{"criteria"} or \code{"heatmap"}.
#' @param rank For \code{type = "heatmap"}, the rank(s) to display.
#'   \code{NULL} (default) shows every rank stored in \code{x}.
#' @param mfrow Panel layout for multiple heatmaps, as
#'   \code{c(nrow, ncol)}.  \code{NULL} (default) uses
#'   \code{\link[grDevices]{n2mfrow}} for a near-square grid.
#' @param mar Per-panel margins \code{c(b, l, t, r)} for the heatmap
#'   view.  \code{NULL} (default) uses tight margins for a grid and wider
#'   margins (for sample labels) when a single rank is shown.
#' @param col Heatmap colour palette (length-50 blue-to-red by default).
#' @param main Plot title for the \code{"criteria"} view (heatmap panels
#'   are titled per rank).
#' @param ... Further arguments passed to the underlying
#'   \code{\link[graphics]{plot}} / \code{\link[graphics]{image}}.
#' @return \code{x}, invisibly.
#' @seealso \code{\link{nmfkc.consensus}}
#' @export
plot.nmfkc.consensus <- function(x, type = c("criteria", "heatmap"),
                                 rank = NULL, mfrow = NULL, mar = NULL,
                                 col = grDevices::hcl.colors(50, "Blue-Red"),
                                 main = NULL, ...) {
  type <- base::match.arg(type)

  if (type == "criteria") {
    graphics::plot(x$rank, x$cophenetic, type = "o", col = "blue", pch = 16,
                   lwd = 2, ylim = c(0, 1), xlab = "rank (Q)",
                   ylab = "stability (0-1)",
                   main = if (base::is.null(main)) "Consensus stability" else main,
                   ...)
    graphics::lines(x$rank, x$dispersion, type = "o", col = "red", pch = 17,
                    lwd = 2)
    leg <- c("cophenetic (Brunet, high=good)", "dispersion (Kim-Park, high=good)")
    lcol <- c("blue", "red"); lpch <- c(16, 17)
    if (!base::is.null(x$pac)) {
      graphics::lines(x$rank, x$pac, type = "o", col = "darkgreen", pch = 15,
                      lwd = 2)
      leg <- c(leg, "PAC (Senbabaoglu, low=good)")
      lcol <- c(lcol, "darkgreen"); lpch <- c(lpch, 15)
    }
    graphics::legend("left", leg, col = lcol, pch = lpch, lwd = 2,
                     bg = "white", cex = 0.78)
    return(base::invisible(x))
  }

  ## type == "heatmap"
  if (base::is.null(x$consensus))
    base::stop("No consensus matrices stored. Re-run ",
               "nmfkc.consensus(..., keep.consensus = TRUE).")
  ranks <- if (base::is.null(rank)) x$rank else base::intersect(rank, x$rank)
  if (!base::length(ranks))
    base::stop("Requested rank(s) are not among the evaluated ranks.")
  n <- base::length(ranks)
  if (base::is.null(mfrow)) mfrow <- grDevices::n2mfrow(n)
  show.labels <- (n == 1)
  if (base::is.null(mar))
    mar <- if (show.labels) c(6, 6, 3, 1) else c(1.5, 1.5, 3, 1)
  old <- graphics::par(mfrow = mfrow, mar = mar, oma = c(2, 0, 0, 0))
  base::on.exit(graphics::par(old))

  for (k in ranks) {
    i <- base::which(x$rank == k)
    C <- x$consensus[[i]]
    hc <- stats::hclust(stats::as.dist(1 - C), method = "average")
    Cr <- C[hc$order, hc$order]; N <- base::nrow(Cr)
    graphics::image(1:N, 1:N, base::t(Cr[N:1, ]), col = col, zlim = c(0, 1),
                    axes = FALSE, xlab = "", ylab = "",
                    main = base::sprintf("rank=%d  coph=%.3f  disp=%.3f",
                                         k, x$cophenetic[i], x$dispersion[i]),
                    ...)
    if (show.labels && !base::is.null(base::rownames(Cr))) {
      cax <- base::min(0.7, 28 / N)
      graphics::axis(1, 1:N, base::colnames(Cr), las = 2, cex.axis = cax)
      graphics::axis(2, 1:N, base::rev(base::rownames(Cr)), las = 2, cex.axis = cax)
    } else {
      graphics::box()
    }
  }
  graphics::mtext("blue = 0 (different cluster)  ->  red = 1 (same cluster)",
                  side = 1, outer = TRUE, cex = 0.7)
  base::invisible(x)
}

## Tests for nmfkc.consensus() (Brunet 2004 consensus-clustering rank selection)

make_blocks <- function(seed = 1) {
  set.seed(seed)
  X <- matrix(0, 30, 3); X[1:10, 1] <- 1; X[11:20, 2] <- 1; X[21:30, 3] <- 1
  B <- matrix(0, 3, 60); for (j in 1:60) B[((j - 1) %/% 20) + 1, j] <- 1
  X %*% B + matrix(abs(rnorm(30 * 60, 0, 0.1)), 30, 60)   # 3 clear clusters
}

test_that("nmfkc.consensus returns stability scores per rank", {
  Y <- make_blocks()
  cs <- suppressMessages(nmfkc.consensus(Y, rank = 2:5, nrun = 15, seed = 1))

  expect_type(cs, "list")
  expect_equal(cs$rank, 2:5)
  expect_equal(cs$nrun, 15)
  expect_length(cs$cophenetic, 4)
  expect_length(cs$dispersion, 4)
  expect_length(cs$pac, 4)
  ## valid ranges
  expect_true(all(cs$dispersion >= 0 & cs$dispersion <= 1, na.rm = TRUE))
  expect_true(all(abs(cs$cophenetic) <= 1, na.rm = TRUE))
  expect_true(all(cs$pac >= 0 & cs$pac <= 1, na.rm = TRUE))
  expect_null(cs$consensus)
})

test_that("nmfkc.consensus is most stable at the true rank (3 clusters)", {
  Y <- make_blocks()
  cs <- suppressMessages(nmfkc.consensus(Y, rank = 2:5, nrun = 15, seed = 1))
  ## true rank 3: dispersion ~1 and clearly above rank 2 (under-split)
  k3 <- which(cs$rank == 3)
  k2 <- which(cs$rank == 2)
  expect_gt(cs$dispersion[k3], 0.95)
  expect_gt(cs$dispersion[k3], cs$dispersion[k2])
  ## PAC is near 0 at the true rank (no ambiguous pairs) and the minimum
  expect_lt(cs$pac[k3], 0.05)
  expect_equal(cs$rank[which.min(cs$pac)], 3)
})

test_that("keep.consensus returns N x N consensus matrices", {
  Y <- make_blocks()
  cs <- suppressMessages(nmfkc.consensus(Y, rank = 2:3, nrun = 10,
                                         keep.consensus = TRUE))
  expect_s3_class(cs, "nmfkc.consensus")
  expect_length(cs$consensus, 2)
  expect_equal(dim(cs$consensus[[1]]), c(ncol(Y), ncol(Y)))
  ## consensus entries are in [0, 1]
  expect_true(all(cs$consensus[[2]] >= 0 & cs$consensus[[2]] <= 1))
})

test_that("print and both plot types work", {
  Y <- make_blocks()
  cs <- suppressMessages(nmfkc.consensus(Y, rank = 2:4, nrun = 10,
                                         keep.consensus = TRUE))
  expect_output(print(cs), "dispersion max")
  pdf(NULL)
  on.exit(dev.off(), add = TRUE)
  expect_no_error(plot(cs))                              # criteria
  expect_no_error(plot(cs, type = "heatmap"))            # all ranks (grid)
  expect_no_error(plot(cs, type = "heatmap", rank = 3))  # single rank
})

test_that("plot heatmap errors without stored consensus", {
  Y <- make_blocks()
  cs <- suppressMessages(nmfkc.consensus(Y, rank = 2:3, nrun = 8))  # no keep
  expect_error(plot(cs, type = "heatmap"), "keep.consensus")
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----lib----------------------------------------------------------------------
library(nmfkc)

## ----data, fig.width = 6, fig.height = 4--------------------------------------
set.seed(123)
r_true <- 3
P <- 24                       # features
n_each <- 20                  # samples per cluster
N <- r_true * n_each          # 60 samples

# basis: each latent factor is a distinct block of features
X <- matrix(0, P, r_true)
X[1:8, 1] <- 1; X[9:16, 2] <- 1; X[17:24, 3] <- 1
X <- X + matrix(runif(P * r_true, 0, 0.1), P, r_true)   # small background

# coefficients: each sample loads mainly on its own cluster's factor
cl_true <- rep(1:r_true, each = n_each)
B <- matrix(runif(r_true * N, 0, 0.2), r_true, N)
for (j in seq_len(N)) B[cl_true[j], j] <- runif(1, 1, 2)

Y <- X %*% B
Y <- Y + matrix(rnorm(P * N, 0, 0.15 * mean(X %*% B)), P, N)  # noise
Y[Y < 0] <- 0

dim(Y)
image(1:N, 1:P, t(Y), xlab = "sample", ylab = "feature",
      main = "Synthetic data (true rank = 3)")

## ----rank, fig.width = 7, fig.height = 6--------------------------------------
rk <- nmfkc.rank(Y, rank = 1:6)
rk$rank.best                       # recommended rank (ECV minimum)
round(rk$criteria, 3)

## ----cv-----------------------------------------------------------------------
ev <- nmfkc.ecv (Y, rank = 1:6)
bv <- nmfkc.bicv(Y, rank = 1:6)   # nfolds = 2 (Owen & Perry) by default
data.frame(rank = 1:6,
           sigma.ecv  = round(ev$sigma, 3),
           sigma.bicv = round(bv$sigma, 3))
cat("ECV  picks rank", which.min(ev$sigma),
    "| bi-CV picks rank", bv$rank[which.min(bv$sigma)], "\n")

## ----consensus, fig.width = 7, fig.height = 5---------------------------------
cs <- nmfkc.consensus(Y, rank = 2:6, nrun = 20, keep.consensus = TRUE)
cs
plot(cs)                               # stability curves

## ----consensus-heatmap, fig.width = 8, fig.height = 6-------------------------
plot(cs, type = "heatmap")   # all ranks

## ----ard, fig.width = 6.5, fig.height = 4.5-----------------------------------
ar <- nmfkc.ard(Y, rank = 8, nrun = 20)   # >=10-20 restarts for a stable mode
ar                                        # see "rank over runs" for confidence
plot(ar)

## ----summary------------------------------------------------------------------
data.frame(
  method = c("nmfkc.rank (ECV)", "nmfkc.ecv", "nmfkc.bicv",
             "nmfkc.consensus (dispersion)", "nmfkc.consensus (PAC)",
             "nmfkc.ard"),
  estimate = c(rk$rank.best,
               which.min(ev$sigma),
               bv$rank[which.min(bv$sigma)],
               cs$rank[which.max(cs$dispersion)],
               cs$rank[which.min(cs$pac)],
               ar$rank),
  true = r_true
)

## ----flow, fig.width = 8, fig.height = 6--------------------------------------
fits <- lapply(1:6, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
fl <- nmf.cluster.flow(fits, reference = 3)
head(fl$clusters)      # rows = individuals, columns = rank, entries = cluster id

# =====================================================
# nmfkc.signed.R — NMF-KC with signed covariate matrix
#
# Extends nmfkc() to allow the covariate matrix A to contain negative
# entries (and also the response Y).  The model is
#
#   Y ≈ X Θ A,    X ≥ 0,  Θ ∈ R^{Q×D},  A ∈ R^{D×N},
#
# where Θ and A may be signed.  Internally the function uses the
# sign-splitting trick (Ding, Li, & Jordan 2010) to solve
#
#   min_{X, C_+, C_-}  || Y − X (C_+ − C_-)(A_+ − A_-) ||_F^2
#   s.t.  X ≥ 0,  C_+ ≥ 0,  C_- ≥ 0, (optionally colSums(X) = 1)
#
# by a Direct Multiplicative Update (Direct MU) algorithm that avoids
# the N × N Gram matrix A A^T is D × D.  On each iteration, after the
# X update, columns of X are rescaled to satisfy the requested
# X.restriction (default colSums), with the scale absorbed into C_+,
# C_- so that X Θ A is unchanged.
#
# The primary application is kernel approximation via Random Fourier
# Features (RFF; Rahimi & Recht 2007), but A can be any real-valued
# matrix (PCA scores, wavelet coefficients, neural-net activations,
# RFF features, ...).
#
# References:
#   Ding, C. H. Q., Li, T., & Jordan, M. I. (2010).  Convex and
#     semi-nonnegative matrix factorizations.  IEEE TPAMI 32(1), 45–55.
#   Rahimi, A., & Recht, B. (2007).  Random features for large-scale
#     kernel machines.  NIPS.
# =====================================================


#' NMF-KC with signed covariate matrix
#'
#' @description
#' Solves
#' \deqn{Y \approx X\,\Theta\,A,\qquad X \ge 0,\;\Theta\in\R^{Q\times D},
#'   \;A\in\R^{D\times N},}
#' where the covariate matrix \eqn{A} and the coefficient matrix
#' \eqn{\Theta} may be \strong{signed}.  Internally \eqn{A = A_{+} - A_{-}}
#' and \eqn{\Theta = C_{+} - C_{-}} with \eqn{A_{\pm}, C_{\pm} \ge 0}
#' (sign-splitting trick, Ding et al. 2010), and the problem is solved
#' by a Direct Multiplicative Update algorithm whose iteration cost is
#' \eqn{O(Q D^2)}, independent of \eqn{N}.
#'
#' Only \eqn{X} is structurally constrained to be non-negative (Semi-NMF
#' sense of Ding, Li, & Jordan 2010).  In particular, \strong{\eqn{Y} may
#' contain negative entries}, in which case the response is fit in the
#' least-squares sense without any non-negativity requirement on \eqn{Y}.
#'
#' When \eqn{A \ge 0} (so \eqn{A_{-} = 0}), the result reduces to
#' \code{\link{nmfkc}}(Y, A, rank) with Euclidean loss, up to reordering.
#'
#' @param Y Real-valued \eqn{Q_{\mathrm{obs}} \times N} response matrix.
#'   Unlike \code{\link{nmfkc}}, negative entries are allowed.
#' @param A Real-valued \eqn{D \times N} covariate matrix (signed).
#'   A single matrix is passed; its positive and negative parts
#'   \eqn{A_{+} = \max(A, 0)} and \eqn{A_{-} = \max(-A, 0)} are computed
#'   internally.  When using Random Fourier Features (Rahimi & Recht
#'   2007) as \eqn{A}, supply the RFF parameters via the hidden
#'   \code{pars} argument so that \code{predict()} can regenerate
#'   features for new data (see \code{pars} entry in \code{\dots} below).
#' @param rank Integer.  Number of latent components \eqn{Q} in \eqn{X}.
#' @param epsilon Relative convergence tolerance on the objective
#'   (default \code{1e-4}).
#' @param maxit Maximum number of iterations (default \code{5000}).
#' @param verbose Logical.  Print dimensions at start (default \code{TRUE}).
#' @param ... Additional arguments:
#'   \itemize{
#'     \item \code{Q}: alias for \code{rank}.
#'     \item \code{X.restriction}: constraint applied to columns of
#'       \eqn{X} after every update, with the scale absorbed into
#'       \eqn{C_{+}, C_{-}}.  One of \code{"colSums"} (default,
#'       \eqn{\mathrm{colSums}(X) = 1}), \code{"colSqSums"},
#'       \code{"totalSum"}, \code{"none"}, \code{"fixed"}.
#'     \item \code{X.init}: initialization strategy for the basis matrix
#'       \eqn{X} (\eqn{Q_{\mathrm{obs}} \times Q}).  Accepts the same
#'       menu as \code{\link{nmfkc}}: \code{"kmeans"} (default),
#'       \code{"kmeansar"}, \code{"nndsvd"}, \code{"runif"}, or a
#'       user-supplied \eqn{Q_{\mathrm{obs}} \times Q} non-negative
#'       numeric matrix.  String methods delegate to the shared
#'       internal helper \code{.init_X_method()} (see \code{\link{nmfkc}}
#'       for the definitions of each method).  For signed \eqn{Y},
#'       \code{"kmeans"} cluster centers may contain negative entries;
#'       they are clipped to zero to satisfy \eqn{X \ge 0}, and any
#'       column that collapses to all-zeros is re-filled with small
#'       \eqn{\mathrm{Uniform}(0, 0.1)} noise.
#'     \item \code{C.init}: explicit initial \eqn{Q \times D} coefficient
#'       matrix \eqn{\Theta} (signed).  Split internally.
#'     \item \code{warm.start}: logical (default \code{TRUE}).  If
#'       \code{TRUE} \strong{and \eqn{Y \ge 0}}, runs
#'       \code{nmfkc(Y, A = rbind(A_+, A_-), rank = Q)} internally to
#'       seed \eqn{X, C_{+}, C_{-}}.  The user's \code{X.init},
#'       \code{seed}, \code{nstart}, and \code{X.restriction} are
#'       forwarded to the internal \code{\link{nmfkc}} call so that
#'       initialization choices propagate consistently between the
#'       warm-start and the signed MU loop.  Ignored when \eqn{Y} has
#'       negative entries (warm-start is disabled; \code{X.init} is
#'       used directly by the signed branch instead).
#'     \item \code{seed}: RNG seed for random initialization (default 123).
#'     \item \code{prefix}: name prefix for rows of \eqn{C} and columns
#'       of \eqn{X} (default \code{"Basis"}).
#'     \item \code{pars}: optional list \code{list(omega, b, D, beta)}
#'       of Random Fourier Feature parameters (Rahimi & Recht 2007;
#'       \code{omega}: frequency matrix, \code{b}: phase offset,
#'       \code{D}: feature dimension, \code{beta}: bandwidth).  When
#'       supplied, it is stored in the returned object so that
#'       \code{summary()} can report \eqn{\beta} and downstream
#'       \code{predict()} calls can regenerate RFF features for new
#'       data.  If \code{A} is not RFF features, leave this \code{NULL}.
#'     \item \code{Y.weights}: Optional non-negative weight matrix
#'       (\eqn{Q_{\mathrm{obs}} \times N}) or vector (length \eqn{N}),
#'       analogous to the \code{weights} argument of
#'       \code{\link[stats]{lm}}.  Loss becomes
#'       \eqn{\sum W_{ij} \, (Y_{ij} - (XCA)_{ij})^2}
#'       (\code{lm()}-style, \strong{linear} in \eqn{W}).  Logical
#'       matrices (\code{TRUE} / \code{FALSE}) are also accepted.
#'       Typical usage by \code{\link{nmfkc.signed.cv}} /
#'       \code{\link{nmfkc.signed.ecv}} passes a binary mask
#'       \eqn{W \in \{0,1\}} to hold out test elements; real-valued
#'       weights for observation-level importance weighting are also
#'       supported.  Default \code{NULL}: if \code{Y} has \code{NA},
#'       a binary mask is auto-constructed (0 for \code{NA}, 1
#'       elsewhere); otherwise no weighting.
#'     \item \code{nstart}: number of random restarts.  \strong{Signed
#'       models have more local minima than non-negative ones} because
#'       \eqn{\Theta = C_{+} - C_{-}} can take both positive and negative
#'       values.  Since \code{nmfkc.signed()} itself does not loop over
#'       restarts (callers control it), set the outer-loop size via e.g.
#'       running the function several times with different \code{seed}
#'       and keeping the fit with the smallest \code{$objfunc}.  A
#'       restart budget of 10-50 is recommended for publication-grade
#'       runs on signed data.
#'   }
#'
#' @return An object of class \code{c("nmfkc.signed", "nmfkc")} with
#' \itemize{
#'   \item \code{X}: \eqn{Q_{\mathrm{obs}} \times Q} basis matrix (non-negative,
#'     column-normalized according to \code{X.restriction}).
#'   \item \code{Cp}, \code{Cn}: \eqn{Q \times D} non-negative parts of
#'     \eqn{\Theta}, so that \eqn{\Theta = C_{+} - C_{-}}.
#'   \item \code{C}: \eqn{C_{+} - C_{-}} (= \eqn{\Theta}), signed.
#'   \item \code{B}: \eqn{C \, A}, \eqn{Q \times N} (signed).
#'   \item \code{objfunc.iter}: objective values per iteration.
#'   \item \code{objfunc}: final objective.
#'   \item \code{r.squared}: \eqn{\mathrm{cor}(Y, \widehat Y)^2} (Pearson; in \eqn{[0,1]}).
#'   \item \code{r.squared.uncentered}: uncentered \eqn{R^2 = 1 - \|Y - \widehat Y\|_F^2 / \|Y\|_F^2} (baseline = zero matrix).
#'   \item \code{r.squared.centered}: row-mean centered \eqn{1 - \|Y - \widehat Y\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2}.
#'   \item \code{mae}: mean absolute error.
#'   \item \code{iter}: number of iterations performed.
#'   \item \code{runtime}: elapsed seconds.
#'   \item \code{Y.signed}: logical; whether \eqn{Y} contained negative
#'     entries during fitting.
#'   \item \code{pars}: RFF generating parameters, if supplied.
#'   \item \code{call}: the matched call.
#' }
#'
#' @section Lifecycle:
#' This function is \strong{experimental}.  The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010).  Convex and
#' semi-nonnegative matrix factorizations.  \emph{IEEE TPAMI}, 32(1), 45–55.
#'
#' Rahimi, A., & Recht, B. (2007).  Random features for large-scale
#' kernel machines.  \emph{Advances in NIPS}, 20.
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{predict.nmfkc.signed}}
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' ## Example 1: signed A (e.g., hand-built RFF features), non-negative Y
#' ## Build simple signed features Z = sqrt(2/D) * cos(omega^T U + b):
#' U     <- matrix(stats::rnorm(5 * 40), 5, 40)           # raw input
#' D     <- 20                                            # feature dim
#' omega <- matrix(stats::rnorm(5 * D), 5, D)             # random freqs
#' b     <- stats::runif(D, 0, 2 * pi)                    # phase
#' Z     <- sqrt(2 / D) *
#'            cos(t(omega) %*% U + matrix(b, D, 40))      # D x 40, signed
#' Y     <- matrix(abs(stats::rnorm(8 * 40)), 8, 40)
#' res1  <- nmfkc.signed(Y, A = Z, rank = 3, maxit = 200)
#'
#' ## Example 2: signed Y (regression)
#' Y2    <- matrix(stats::rnorm(8 * 40), 8, 40)           # signed response
#' res2  <- nmfkc.signed(Y2, A = Z, rank = 3, maxit = 200,
#'                        warm.start = FALSE)
#' }
#'
#' @export
nmfkc.signed <- function(Y, A, rank = NULL,
                         epsilon = 1e-4, maxit = 5000,
                         verbose = TRUE, ...) {
  cl <- match.call()
  extra_args <- list(...)

  ## --- 1. Parameter extraction ---
  if (is.null(rank) && !is.null(extra_args$Q)) rank <- extra_args$Q
  if (is.null(rank)) stop("'rank' must be specified.")
  Q <- as.integer(rank)

  X.restriction <- if (!is.null(extra_args$X.restriction))
    extra_args$X.restriction else "colSums"
  X.restriction <- match.arg(X.restriction,
    c("colSums", "colSqSums", "totalSum", "none", "fixed"))

  X.init     <- if (!is.null(extra_args$X.init))     extra_args$X.init     else "kmeans"
  C.init     <- if (!is.null(extra_args$C.init))     extra_args$C.init     else NULL
  warm.start <- if (!is.null(extra_args$warm.start)) extra_args$warm.start else TRUE
  seed       <- if (!is.null(extra_args$seed))       extra_args$seed       else 123L
  prefix     <- if (!is.null(extra_args$prefix))     extra_args$prefix     else "Basis"
  pars_rff   <- extra_args$pars
  Y.weights  <- extra_args$Y.weights

  ## --- 2. Input preparation & validation ---
  if (is.vector(Y)) Y <- matrix(Y, nrow = 1)
  if (!is.matrix(Y)) Y <- as.matrix(Y)
  if (!is.matrix(A)) A <- as.matrix(A)
  Q_obs <- nrow(Y); N <- ncol(Y); D <- nrow(A)
  if (ncol(A) != N) stop("ncol(A) must equal ncol(Y).")
  if (any(is.na(A))) stop("A contains NA; please impute or remove.")

  ## --- Y.weights handling (supports vector, matrix, NA-autodetect) ---
  ## When supplied, the loss becomes sum(W * (Y - XCA)^2) and held-out
  ## elements (W = 0) are masked.  Y-NA auto-masking matches nmfkc().
  if (!is.null(Y.weights) && is.vector(Y.weights) && !is.matrix(Y.weights)) {
    if (length(Y.weights) == N) {
      Y.weights <- matrix(Y.weights, nrow = Q_obs, ncol = N, byrow = TRUE)
    } else if (length(Y.weights) == 1) {
      Y.weights <- matrix(Y.weights, nrow = Q_obs, ncol = N)
    } else {
      stop("Length of Y.weights vector must match ncol(Y) or be 1.")
    }
  }
  if (is.null(Y.weights) && any(is.na(Y))) {
    Y.weights <- matrix(1, nrow = Q_obs, ncol = N)
    Y.weights[is.na(Y)] <- 0
    Y[is.na(Y)] <- 0
  } else if (!is.null(Y.weights)) {
    Y.weights <- as.matrix(Y.weights)
    if (!all(dim(Y.weights) == dim(Y)))
      stop("Y.weights dimensions must match Y.")
    Y.weights[is.na(Y.weights)] <- 0
    Y[is.na(Y) | Y.weights == 0] <- 0
  } else if (any(is.na(Y))) {
    stop("Y contains NA; please impute, remove, or supply Y.weights.")
  }
  has.weights <- !is.null(Y.weights) && any(Y.weights != 1)
  if (!has.weights) Y.weights <- NULL
  Y_is_nonneg <- all(Y >= 0)

  if (isTRUE(verbose)) {
    msg <- sprintf("Y(%d,%d) ~ X(%d,%d) %%*%% C(%d,%d) %%*%% A(%d,%d)  [signed covariate]%s%s",
                   Q_obs, N, Q_obs, Q, Q, D, D, N,
                   if (Y_is_nonneg) "" else ", Y signed",
                   if (has.weights) ", weighted" else "")
    message(msg)
  }

  t0 <- proc.time()

  ## --- 3. Precomputation (avoids N x N Gram) ---
  Ap <- pmax(A,  0)
  An <- pmax(-A, 0)
  A_diff <- Ap - An                 # = A
  ## For weighted loss, S / G0 are NOT iteration-invariant, so we skip the
  ## precomputation and use a slower weighted path inside the main loop.
  ## For the unweighted loss, the precomputation gives the O(QD^2) speedup.
  if (!has.weights) {
    S   <- tcrossprod(A_diff)         # D x D, signed
    G0  <- Y %*% t(A_diff)            # Q_obs x D, signed (Y signed OK)
    S_p <- pmax(S,  0)
    S_n <- pmax(-S, 0)
  } else {
    S <- NULL; G0 <- NULL; S_p <- NULL; S_n <- NULL
  }
  Y_sqnorm <- sum(Y * Y)            # always >= 0

  ## --- 4. X.restriction helpers ---
  xscale <- switch(X.restriction,
    colSums   = function(X) colSums(X) + 1e-16,
    colSqSums = function(X) sqrt(colSums(X * X)) + 1e-16,
    totalSum  = function(X) rep(sum(X) + 1e-16, ncol(X)),
    none      = function(X) rep(1, ncol(X)),
    fixed     = function(X) rep(1, ncol(X)))

  ## --- 5. Initialization ---
  X <- NULL; Cp <- NULL; Cn <- NULL

  ## 5a. Warm-start via posneg nmfkc() (requires Y >= 0; disabled when weighted
  ## because the warm-start init can produce numerical explosion in the
  ## weighted MU loop — ECV CV callers get random init instead).
  explicit_X_mat <- is.matrix(X.init) || (is.numeric(X.init) && length(X.init) > 1)
  need_warm <- isTRUE(warm.start) && Y_is_nonneg && !has.weights &&
               is.null(C.init) && !explicit_X_mat
  if (need_warm) {
    warm_args <- list(Y, A = rbind(Ap, An), rank = Q,
                      epsilon = epsilon, maxit = maxit, verbose = FALSE,
                      seed = seed, X.restriction = X.restriction)
    if (has.weights) warm_args$Y.weights <- Y.weights
    ## Forward X.init (accepts the same menu as nmfkc()) so that the user's
    ## chosen initialization propagates into the posneg warm-start.
    warm_args$X.init <- X.init
    ## Forward nstart if the user supplied it (nmfkc() default is 1).
    if (!is.null(extra_args$nstart)) warm_args$nstart <- extra_args$nstart
    res0 <- do.call(nmfkc, warm_args)
    X  <- res0$X
    Cp <- res0$C[, 1:D, drop = FALSE]
    Cn <- res0$C[, (D + 1):(2 * D), drop = FALSE]
  } else {
    ## 5b. No warm-start: delegate to the shared .init_X_method() helper
    ## (same menu as nmfkc() / nmf.sem(): "kmeans", "kmeansar", "nndsvd",
    ## "runif", or a user-supplied Q_obs x Q matrix).
    if (explicit_X_mat) {
      X <- as.matrix(X.init)
      if (!identical(dim(X), c(Q_obs, Q)))
        stop("X.init must have dimensions (nrow(Y), rank).")
      X[X < 0] <- 0
    } else if (is.character(X.init)) {
      ## For signed Y, kmeans cluster centers may contain negative values;
      ## clip to non-negative since X must satisfy X >= 0.  .init_X_method's
      ## "nndsvd" (NNDSVDar) and "runif" paths are already non-negative.
      X <- .init_X_method(X.init, Y, Q, seed = seed)
      X[X < 0] <- 0
      ## Safety: if any column collapses to zero (can happen with kmeans
      ## on signed Y where a whole cluster mean is non-positive), fill with
      ## small positive noise so MU iterations can escape.
      bad <- colSums(X) < 1e-10
      if (any(bad)) {
        X[, bad] <- matrix(stats::runif(Q_obs * sum(bad), 0, 0.1),
                           Q_obs, sum(bad))
      }
    } else {
      stop("X.init must be one of \"kmeans\", \"kmeansar\", \"nndsvd\", ",
           "\"runif\", or a (nrow(Y) x rank) numeric matrix; got ",
           class(X.init)[1], ".")
    }
    ## C.init or random
    if (!is.null(C.init)) {
      if (!identical(dim(C.init), c(Q, D)))
        stop("C.init must have dimensions (rank, nrow(A)).")
      Cp <- pmax(C.init,  0)
      Cn <- pmax(-C.init, 0)
    } else {
      set.seed(seed + 1L)
      Cp <- matrix(abs(stats::rnorm(Q * D)) * 0.1,  Q, D)
      Cn <- matrix(abs(stats::rnorm(Q * D)) * 0.01, Q, D)
    }
  }

  ## Apply X.restriction initially (absorb into Cp, Cn so X %*% (Cp-Cn) %*% A is unchanged)
  if (X.restriction != "fixed" && X.restriction != "none") {
    d0 <- xscale(X)
    X  <- sweep(X,  2, d0, "/")
    Cp <- sweep(Cp, 1, d0, "*")
    Cn <- sweep(Cn, 1, d0, "*")
  }

  small <- 1e-16
  Wmat <- Y.weights  # short alias; NULL if no weights

  compute_obj <- function(X, Cp, Cn) {
    Yhat <- X %*% (Cp - Cn) %*% A_diff
    ## lm()-style weighted least squares: L = sum(W * (Y - Yhat)^2).
    ## W = W^2 for binary {0,1} masks so this is unchanged for ECV.
    if (has.weights) sum(Wmat * (Y - Yhat)^2) else sum((Y - Yhat)^2)
  }

  if (!has.weights) {
    P <- crossprod(X); G <- crossprod(X, G0)
    H <- Cp - Cn; PH <- P %*% H
    obj_prev <- Y_sqnorm - 2 * sum(G * H) + sum(H * (PH %*% S))
  } else {
    obj_prev <- compute_obj(X, Cp, Cn)
  }
  objfunc.iter <- numeric(maxit)

  ## --- 6. Main loop ---
  iter <- 0L
  for (iter in seq_len(maxit)) {

    if (!has.weights) {
      ## ---- Fast unweighted path (precomputed S, G0) ----
      ## 6a. Cp update
      G_p <- pmax(G, 0); G_n <- pmax(-G, 0)
      PCp <- P %*% Cp;   PCn <- P %*% Cn
      Cp  <- Cp * (G_p + PCp %*% S_n + PCn %*% S_p) /
                  (G_n + PCp %*% S_p + PCn %*% S_n + small)

      ## 6b. Cn update (Gauss-Seidel)
      PCp <- P %*% Cp
      Cn  <- Cn * (G_n + PCp %*% S_p + PCn %*% S_n) /
                  (G_p + PCp %*% S_n + PCn %*% S_p + small)

      ## 6c. X update
      if (X.restriction != "fixed") {
        H   <- Cp - Cn; Ht <- t(H)
        YMt <- G0 %*% Ht
        HS  <- H %*% S; MMt <- HS %*% Ht
        X <- X * (pmax(YMt, 0) + X %*% pmax(-MMt, 0)) /
                 (pmax(-YMt, 0) + X %*% pmax(MMt, 0) + small)
        if (X.restriction != "none") {
          d <- xscale(X)
          X  <- sweep(X,  2, d, "/")
          Cp <- sweep(Cp, 1, d, "*")
          Cn <- sweep(Cn, 1, d, "*")
        }
      }

      ## 6d. Refresh precomputed quantities & evaluate objective in closed form
      P  <- crossprod(X); G <- crossprod(X, G0)
      H  <- Cp - Cn; PH <- P %*% H
      obj_cur <- Y_sqnorm - 2 * sum(G * H) + sum(H * (PH %*% S))

    } else {
      ## ---- Weighted path (no S/G0 precompute) ----
      ## Let Yhat_+ = X Cp A, Yhat_- = X Cn A; residual under W.
      tX <- t(X)
      Yhat_p <- X %*% Cp %*% A_diff          # signed if A signed
      Yhat_n <- X %*% Cn %*% A_diff
      WY     <- Wmat * Y                      # weighted observations
      WYhp   <- Wmat * Yhat_p
      WYhn   <- Wmat * Yhat_n

      ## 6a'. Cp update via weighted split
      ## grad_{Cp} L = -2 X^T (W*(Y - X(Cp-Cn)A)) A^T
      ##             = -2 ( X^T (W*Y) A^T - X^T (W*Yhat_p) A^T + X^T (W*Yhat_n) A^T )
      G_w   <- tX %*% WY   %*% t(A_diff)      # Q x D, signed
      Hp_w  <- tX %*% WYhp %*% t(A_diff)      # Q x D, signed
      Hn_w  <- tX %*% WYhn %*% t(A_diff)      # Q x D, signed
      Gp <- pmax(G_w, 0); Gn <- pmax(-G_w, 0)
      Hpp <- pmax(Hp_w, 0); Hpn <- pmax(-Hp_w, 0)
      Hnp <- pmax(Hn_w, 0); Hnn <- pmax(-Hn_w, 0)
      Cp <- Cp * (Gp + Hpn + Hnp) / (Gn + Hpp + Hnn + small)
      ## Recompute Hp_w with updated Cp
      Yhat_p <- X %*% Cp %*% A_diff
      WYhp   <- Wmat * Yhat_p
      Hp_w   <- tX %*% WYhp %*% t(A_diff)
      Hpp <- pmax(Hp_w, 0); Hpn <- pmax(-Hp_w, 0)
      ## 6b'. Cn update
      Cn <- Cn * (Gn + Hpp + Hnn) / (Gp + Hpn + Hnp + small)

      ## 6c'. X update (weighted)
      ## -dL/(2) = (W*Y) M^T - (W*(XM)) M^T   where M = H A (signed).
      ## Ding split on the two signed terms A1, A2:
      ##   Pull = [A1]_+ + [A2]_-,  Push = [A1]_- + [A2]_+
      if (X.restriction != "fixed") {
        H   <- Cp - Cn
        M   <- H %*% A_diff        # Q x N, signed
        XM  <- X %*% M             # Q_obs x N, signed
        WY  <- Wmat * Y
        WXM <- Wmat * XM
        A1  <- WY  %*% t(M)        # Q_obs x Q, signed
        A2  <- WXM %*% t(M)        # Q_obs x Q, signed
        num <- pmax(A1, 0) + pmax(-A2, 0)
        den <- pmax(-A1, 0) + pmax(A2, 0)
        X <- X * num / (den + small)
        if (X.restriction != "none") {
          d <- xscale(X)
          X  <- sweep(X,  2, d, "/")
          Cp <- sweep(Cp, 1, d, "*")
          Cn <- sweep(Cn, 1, d, "*")
        }
      }

      obj_cur <- compute_obj(X, Cp, Cn)
    }

    objfunc.iter[iter] <- obj_cur
    ## Safety: bail out on NaN/Inf (numerical explosion in weighted MU)
    if (!is.finite(obj_cur)) {
      warning("nmfkc.signed: objective became non-finite at iter ", iter,
              "; stopping early.")
      break
    }
    if (is.finite(obj_prev) &&
        abs(obj_prev - obj_cur) / max(abs(obj_prev), 1e-12) < epsilon) break
    obj_prev <- obj_cur
  }
  ## Warn when the MU loop exhausts maxit without meeting the
  ## relative-tolerance criterion (matches nmfkc() / nmf.sem() convention).
  ## The non-finite-objective early break above already issued its own
  ## warning, so we additionally guard against double-warning here.
  if (iter == maxit && is.finite(obj_cur) && is.finite(obj_prev) &&
      abs(obj_prev - obj_cur) / max(abs(obj_prev), 1e-12) >= epsilon)
    warning(paste0("maximum iterations (", maxit, ") reached..."))
  objfunc.iter <- objfunc.iter[seq_len(iter)]

  ## --- 7. Post-processing: sort columns of X (nmfkc-style centroid order) ---
  if (ncol(X) > 1 && X.restriction != "fixed") {
    index <- order(as.vector(matrix(1:nrow(X) / nrow(X), nrow = 1) %*% X))
    X  <- X[, index, drop = FALSE]
    Cp <- Cp[index, , drop = FALSE]
    Cn <- Cn[index, , drop = FALSE]
  }

  ## --- 8. Reconstruction statistics ---
  C  <- Cp - Cn                     # Q x D, signed  (= Theta)
  B  <- C %*% A_diff                # Q x N, signed
  XB <- X %*% B                     # Q_obs x N, Yhat
  resid <- Y - XB
  if (has.weights) {
    ## lm()-style weighted least squares (matches compute_obj in the loop).
    objfunc <- sum(Wmat * resid^2)
    r2_all  <- .r.squared.all(Y, XB, Y.weights = Wmat)
    mae     <- sum(Wmat * abs(resid)) / max(sum(Wmat), small)
  } else {
    objfunc <- sum(resid * resid)
    r2_all  <- .r.squared.all(Y, XB)
    mae     <- mean(abs(resid))
  }
  r.squared          <- r2_all$r.squared
  r.squared.uncentered     <- r2_all$r.squared.uncentered
  r.squared.centered <- r2_all$r.squared.centered

  ## --- 9. Names ---
  rownames(X)  <- rownames(Y)
  colnames(X)  <- paste0(prefix, seq_len(Q))
  rownames(Cp) <- paste0(prefix, seq_len(Q))
  rownames(Cn) <- paste0(prefix, seq_len(Q))
  rownames(C)  <- paste0(prefix, seq_len(Q))
  if (!is.null(rownames(A))) {
    colnames(Cp) <- rownames(A)
    colnames(Cn) <- rownames(A)
    colnames(C)  <- rownames(A)
  }

  runtime <- as.numeric((proc.time() - t0)[3])

  ## Fields matching the main `nmfkc()` object for S3 method inheritance
  dims_str <- sprintf("Y(%d,%d)~X(%d,%d)C(%d,%d)A(%d,%d)",
                      Q_obs, N, Q_obs, Q, Q, D, D, N)
  n.valid   <- if (has.weights) sum(Wmat > 0) else Q_obs * N
  n.total   <- Q_obs * N
  n.missing <- n.total - n.valid
  sigma     <- if (n.valid > 0)
                 sqrt(sum(if (has.weights) Wmat * resid^2 else resid^2) / n.valid)
               else NA_real_

  result <- list(
    X             = X,
    Cp            = Cp,
    Cn            = Cn,
    C             = C,
    B             = B,
    XB            = XB,              # reconstruction (alias for fitted.nmf)
    objfunc.iter  = objfunc.iter,
    objfunc       = objfunc,
    r.squared          = r.squared,
    r.squared.uncentered     = r.squared.uncentered,
    r.squared.centered = r.squared.centered,
    sigma         = sigma,
    mae           = mae,
    iter          = iter,
    runtime       = runtime,
    rank          = Q,
    D             = D,
    dims          = dims_str,
    method        = "EU",
    n.missing     = n.missing,
    n.total       = n.total,
    X.restriction = X.restriction,
    Y.signed      = !Y_is_nonneg,
    pars          = pars_rff,
    call          = cl
  )
  class(result) <- c("nmfkc.signed", "nmfkc", "nmf")
  result
}


#' Predict method for nmfkc.signed
#'
#' @description
#' Computes \eqn{\widehat Y = X \, C \, A_{\mathrm{new}}}
#' (\eqn{= X (C_{+} - C_{-})(A_{+}^{\mathrm{new}} - A_{-}^{\mathrm{new}})}).
#' For \code{type = "response"} the raw prediction is returned
#' (possibly signed).  For \code{type = "prob"} and \code{"class"},
#' negative entries of \eqn{\widehat Y} are clipped to zero before
#' column normalization, since probabilities must be non-negative.
#'
#' @param object A fitted \code{"nmfkc.signed"} object.
#' @param newA Real-valued \eqn{D \times N_{\mathrm{new}}} covariate matrix.
#' @param type Output: \code{"response"} (raw signed), \code{"prob"},
#'   or \code{"class"}.
#' @param ... Unused.
#'
#' @return A numeric matrix (\code{"response"} or \code{"prob"}) or a
#'   character vector (\code{"class"}).
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
predict.nmfkc.signed <- function(object, newA = NULL,
                                  type = c("response", "prob", "class"),
                                  ...) {
  type <- match.arg(type)
  if (is.null(newA)) stop("'newA' must be supplied.")
  if (!is.matrix(newA)) newA <- as.matrix(newA)

  Yhat <- object$X %*% object$C %*% newA    # may be signed
  if (type == "response") return(Yhat)

  ## For prob / class, clip negatives and column-normalize
  Yhat <- pmax(Yhat, 0)
  col_sums <- colSums(Yhat) + 1e-12
  probs <- sweep(Yhat, 2, col_sums, "/")
  if (type == "prob") return(probs)

  classes <- rownames(Yhat)
  if (is.null(classes)) classes <- as.character(seq_len(nrow(Yhat)))
  classes[apply(probs, 2, which.max)]
}


#' Plot method for nmfkc.signed (convergence)
#'
#' @param x An \code{nmfkc.signed} object.
#' @param ... Passed to \code{plot()}.
#' @return Invisible \code{x}.
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
plot.nmfkc.signed <- function(x, ...) {
  extra_args <- list(...)
  args <- list(x = x$objfunc.iter, type = "l")
  if (is.null(extra_args$main))
    args$main <- sprintf("r.squared = %.3f", x$r.squared)
  if (is.null(extra_args$xlab)) args$xlab <- "iter"
  if (is.null(extra_args$ylab)) args$ylab <- "objfunc"
  do.call(graphics::plot, c(args, extra_args))
  invisible(x)
}


#' Summary method for nmfkc.signed
#'
#' @param object An \code{nmfkc.signed} object.
#' @param ... Unused.
#' @return An object of class \code{"summary.nmfkc.signed"}.
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
summary.nmfkc.signed <- function(object, ...) {
  .negfrac <- function(M) {
    if (is.null(M)) return(NA_real_)
    sum(pmax(-M, 0)) / (sum(abs(M)) + 1e-16)
  }
  .range   <- function(M) if (is.null(M)) c(NA_real_, NA_real_) else range(M)
  ans <- list(
    call          = object$call,
    Q_obs         = nrow(object$X),
    Q             = ncol(object$X),
    D             = object$D,
    X.restriction = object$X.restriction,
    Y.signed      = isTRUE(object$Y.signed),
    iter          = object$iter,
    runtime       = object$runtime,
    objfunc       = object$objfunc,
    r.squared          = object$r.squared,
    r.squared.uncentered     = object$r.squared.uncentered,
    r.squared.centered = object$r.squared.centered,
    mae           = object$mae,
    ## Sparsity
    X.sparsity    = if (!is.null(object$X))  mean(object$X  < 1e-4) else NA_real_,
    Cp.sparsity   = if (!is.null(object$Cp)) mean(object$Cp < 1e-4) else NA_real_,
    Cn.sparsity   = if (!is.null(object$Cn)) mean(object$Cn < 1e-4) else NA_real_,
    ## Range (min, max)
    X.range       = .range(object$X),
    Cp.range      = .range(object$Cp),
    Cn.range      = .range(object$Cn),
    C.range       = .range(object$C),
    B.range       = .range(object$B),
    ## Negative-mass ratio  sum(max(-M,0)) / sum(|M|)
    X.negfrac     = .negfrac(object$X),       # X >= 0, so always 0
    C.negfrac     = .negfrac(object$C),
    B.negfrac     = .negfrac(object$B),
    pars          = object$pars
  )
  class(ans) <- "summary.nmfkc.signed"
  ans
}


#' Print method for summary.nmfkc.signed
#'
#' @param x Object of class \code{"summary.nmfkc.signed"}.
#' @param digits Number of significant digits.
#' @param ... Unused.
#' @return Invisible \code{x}.
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
print.summary.nmfkc.signed <- function(x,
                                        digits = max(3L, getOption("digits") - 3L),
                                        ...) {
  cat("\nCall:\n",
      paste(deparse(x$call), sep = "\n", collapse = "\n"),
      "\n\n", sep = "")
  cat("Semi-NMF-KC with signed covariate (Direct MU)\n")
  cat(sprintf("  Y rows (Q_obs):    %d%s\n", x$Q_obs,
              if (x$Y.signed) "  (Y has negative entries)" else ""))
  cat(sprintf("  Rank (Q):          %d\n", x$Q))
  cat(sprintf("  Covariate dim (D): %d\n", x$D))
  cat(sprintf("  X.restriction:     %s\n", x$X.restriction))

  if (!is.null(x$pars)) {
    p_in  <- ncol(x$pars$omega)
    o_dim <- dim(x$pars$omega)
    cat("\nRFF parameters (for reference):\n")
    cat(sprintf("  Input dim (p):    %d\n", p_in))
    cat(sprintf("  beta (bandwidth): %s\n",
                format(x$pars$beta, digits = digits)))
    cat(sprintf("  omega:            %d x %d  (range [%s, %s])\n",
                o_dim[1], o_dim[2],
                format(min(x$pars$omega), digits = digits),
                format(max(x$pars$omega), digits = digits)))
  }

  cat("\nConvergence:\n")
  cat(sprintf("  Iterations:        %d\n", x$iter))
  cat(sprintf("  Runtime (secs):    %.2f\n", x$runtime))
  cat(sprintf("  Final objfunc:     %s\n", format(x$objfunc, digits = digits)))

  cat("\nGoodness of fit:\n")
  cat(sprintf("  R-squared (cor^2):    %s\n",
              format(x$r.squared, digits = digits)))
  if (!is.null(x$r.squared.uncentered))
    cat(sprintf("  R-squared (uncentered):     %s\n",
                format(x$r.squared.uncentered, digits = digits)))
  if (!is.null(x$r.squared.centered))
    cat(sprintf("  R-squared (centered): %s\n",
                format(x$r.squared.centered, digits = digits)))
  cat(sprintf("  MAE:                  %s\n",
              format(x$mae, digits = digits)))

  cat("\nStructure (range / sparsity / negative mass):\n")
  fmt_range <- function(r)
    sprintf("[%s, %s]",
            format(r[1], digits = digits, width = 1),
            format(r[2], digits = digits, width = 1))
  cat(sprintf("  X  (Q_obs x Q):  range %s   sparsity %5.1f%%   neg-mass %5.1f%%\n",
              fmt_range(x$X.range),  100 * x$X.sparsity,  100 * x$X.negfrac))
  cat(sprintf("  Cp (Q x D):      range %s   sparsity %5.1f%%   (non-negative)\n",
              fmt_range(x$Cp.range), 100 * x$Cp.sparsity))
  cat(sprintf("  Cn (Q x D):      range %s   sparsity %5.1f%%   (non-negative)\n",
              fmt_range(x$Cn.range), 100 * x$Cn.sparsity))
  cat(sprintf("  C=Cp-Cn (Q x D): range %s                   neg-mass %5.1f%%\n",
              fmt_range(x$C.range),                          100 * x$C.negfrac))
  cat(sprintf("  B=CA (Q x N):    range %s                   neg-mass %5.1f%%\n",
              fmt_range(x$B.range),                          100 * x$B.negfrac))
  cat("\n")
  invisible(x)
}


## ==============================================================
## Cross-validation helpers for nmfkc.signed
## ==============================================================

#' Column-wise k-fold cross-validation for nmfkc.signed
#'
#' @description Column-wise k-fold CV by held-out samples: for each
#' fold, the model is fit on the training columns and evaluated on the
#' held-out columns by solving for new-sample coefficients via a
#' weighted refit with \eqn{X} fixed.
#'
#' @param Y Real-valued \eqn{Q_{\mathrm{obs}} \times N} response matrix
#'   (signed entries allowed).
#' @param A Real-valued \eqn{D \times N} covariate matrix (signed).
#' @param rank Integer \eqn{Q}.
#' @param ... Passed to \code{\link{nmfkc.signed}}; also accepts
#'   \code{nfolds} (default 5; \code{div} alias), \code{seed}
#'   (default 123), \code{shuffle} (default \code{TRUE}).
#'
#' @section Lifecycle:
#' This function is \strong{experimental}.
#'
#' @return A list with \code{objfunc} (mean squared prediction error),
#'   \code{sigma} (RMSE), \code{objfunc.block} (per-fold MSE vector),
#'   \code{block} (integer fold assignment of length \eqn{N}).  Field
#'   names match \code{\link{nmfkc.cv}}.
#' @seealso \code{\link{nmfkc.signed}}, \code{\link{nmfkc.signed.ecv}}
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
nmfkc.signed.cv <- function(Y, A, rank = 2, ...) {
  extra <- list(...)
  if (!is.null(extra$Q)) rank <- extra$Q
  nfolds  <- if (!is.null(extra$nfolds)) extra$nfolds
             else if (!is.null(extra$div)) extra$div else 5
  seed    <- if (!is.null(extra$seed))    extra$seed    else 123
  shuffle <- if (!is.null(extra$shuffle)) extra$shuffle else TRUE

  Y <- as.matrix(Y); A <- as.matrix(A)
  N <- ncol(Y)
  if (ncol(A) != N) stop("ncol(A) must equal ncol(Y).")

  ## Strip CV-specific / user-overridable args
  fit_args <- extra
  fit_args$nfolds <- NULL; fit_args$div <- NULL
  fit_args$seed <- NULL; fit_args$shuffle <- NULL
  fit_args$Q <- NULL
  fit_args$verbose <- NULL; fit_args$Y.weights <- NULL

  ## Create folds over columns
  set.seed(seed)
  idx <- if (isTRUE(shuffle)) sample.int(N) else seq_len(N)
  chunk <- N %/% nfolds; rem <- N %% nfolds
  folds <- vector("list", nfolds); s <- 1L
  for (k in 1:nfolds) {
    sz <- chunk + ifelse(k <= rem, 1L, 0L)
    folds[[k]] <- idx[s:(s + sz - 1L)]; s <- s + sz
  }

  ## Per-fold: fit on train columns (via Y.weights = 0 on test columns),
  ## then score on held-out columns using Yhat = X C A
  objfunc.block <- numeric(nfolds)
  ## block: integer vector of length N assigning each column to a fold
  ## (matches main nmfkc.cv's output shape).
  block <- integer(N)
  for (k in 1:nfolds) block[folds[[k]]] <- k
  for (k in 1:nfolds) {
    test_cols <- folds[[k]]
    W <- matrix(1, nrow = nrow(Y), ncol = N)
    W[, test_cols] <- 0
    fit <- suppressMessages(do.call(
      nmfkc.signed,
      c(list(Y = Y, A = A, rank = rank, verbose = FALSE,
             Y.weights = W), fit_args)))
    Yhat_test <- fit$X %*% fit$C %*% A[, test_cols, drop = FALSE]
    objfunc.block[k] <- mean((Y[, test_cols, drop = FALSE] - Yhat_test)^2)
  }
  structure(list(
    objfunc = mean(objfunc.block),
    sigma = sqrt(mean(objfunc.block)),
    objfunc.block = objfunc.block,
    block = block
  ), class = c("nmfkc.signed.cv", "nmfkc.cv"))
}


#' Element-wise cross-validation for nmfkc.signed
#'
#' @description Element-wise k-fold CV (Wold's CV): held-out elements
#' are masked via \code{Y.weights = 0} during fitting, and the RMSE on
#' those elements is reported.  Loops over candidate \code{rank} values.
#'
#' @param Y Real-valued \eqn{Q_{\mathrm{obs}} \times N} response matrix
#'   (signed entries allowed).
#' @param A Real-valued \eqn{D \times N} covariate matrix (signed).
#' @param rank Integer vector of candidate ranks (default \code{1:3}).
#' @param ... Passed to \code{\link{nmfkc.signed}}; also accepts
#'   \code{nfolds} (default 5; \code{div} alias), \code{seed}
#'   (default 123).
#'
#' @section Lifecycle:
#' This function is \strong{experimental}.
#'
#' @return A list with \code{objfunc} (MSE per rank), \code{sigma}
#'   (RMSE), \code{objfunc.fold} (per-fold per-rank), \code{folds},
#'   \code{Q.grid}.
#' @seealso \code{\link{nmfkc.signed}}, \code{\link{nmfkc.signed.cv}}
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
nmfkc.signed.ecv <- function(Y, A, rank = 1:3, ...) {
  extra <- list(...)
  if (!is.null(extra$Q)) rank <- extra$Q
  nfolds <- if (!is.null(extra$nfolds)) extra$nfolds
            else if (!is.null(extra$div)) extra$div else 5
  seed   <- if (!is.null(extra$seed))   extra$seed   else 123

  Y <- as.matrix(Y); A <- as.matrix(A)
  P <- nrow(Y); N <- ncol(Y)
  if (ncol(A) != N) stop("ncol(A) must equal ncol(Y).")

  fit_args <- extra
  fit_args$nfolds <- NULL; fit_args$div <- NULL
  fit_args$seed <- NULL; fit_args$Q <- NULL
  fit_args$verbose <- NULL; fit_args$Y.weights <- NULL

  ## Create folds over valid elements (non-NA; shared helper)
  folds <- .ecv.make.folds(Y, nfolds, seed)

  run_one <- function(q, k) {
    test_idx <- folds[[k]]
    W <- matrix(1, nrow = P, ncol = N)
    if (any(is.na(Y))) W[is.na(Y)] <- 0
    W[test_idx] <- 0
    fit <- suppressMessages(do.call(
      nmfkc.signed,
      c(list(Y = Y, A = A, rank = q, verbose = FALSE,
             Y.weights = W), fit_args)))
    ## Yhat = X C A on held-out entries
    Yhat <- fit$X %*% fit$C %*% A
    mean((Y[test_idx] - Yhat[test_idx])^2)
  }

  message(sprintf("nmfkc.signed ECV: %d ranks, %d-fold.",
                  length(rank), nfolds))
  cv <- .ecv.run(sprintf("Q=%d", rank), nfolds,
                 run_one = function(i, k) run_one(rank[i], k),
                 progress = function(i, o, s)
                   message(sprintf("  Q=%d: MSE=%.6f, sigma=%.4f", rank[i], o, s)))
  structure(list(
    objfunc = cv$objfunc, sigma = cv$sigma,
    objfunc.fold = cv$objfunc.fold, folds = folds, Q.grid = rank
  ), class = c("nmfkc.signed.ecv", "nmfkc.ecv"))
}


#' @title Rank selection for nmfkc.signed (concise diagnostics)
#' @description
#' Fits \code{\link{nmfkc.signed}} across a range of ranks and reports
#' \code{r.squared}, the effective rank, and the element-wise CV error
#' \code{sigma.ecv}, with the same concise plot as
#' \code{\link{nmfkc.rank}}.
#' @param Y Observation matrix (may contain negative entries).
#' @param A Covariate matrix (may be signed).
#' @param rank Integer vector of ranks to evaluate.
#' @param detail \code{"full"} (default) also runs element-wise CV
#'   (\code{sigma.ecv}); \code{"fast"} skips it (plots r.squared and
#'   eff.rank only, and recommends the R-squared elbow).
#' @param plot Logical; draw the diagnostics plot (default \code{TRUE}).
#' @param ... Passed on to \code{\link{nmfkc.signed}} and
#'   \code{nmfkc.signed.ecv} (e.g.\ \code{maxit}, \code{nfolds},
#'   \code{seed}).
#' @return A list with \code{rank.best} and \code{criteria}
#'   (\code{rank}, \code{effective.rank}, \code{effective.rank.ratio},
#'   \code{r.squared}, \code{sigma.ecv}).
#' @seealso \code{\link{nmfkc.signed}}, \code{\link{nmfkc.rank}}
#' @references
#' Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
#' effective dimensionality.  \emph{Proc. EUSIPCO}, 606--610.
#' (\code{effective.rank})
#' Wold, S. (1978).  Cross-validatory estimation of the number of
#' components in factor and principal components models.
#' \emph{Technometrics}, 20(4), 397--405. (\code{sigma.ecv})
#' @export
nmfkc.signed.rank <- function(Y, A, rank = 1:5, detail = c("full", "fast"),
                              plot = TRUE, ...) {
  detail <- match.arg(detail)
  Y <- as.matrix(Y); A <- as.matrix(A)
  rs <- numeric(length(rank)); er <- numeric(length(rank))
  for (i in seq_along(rank)) {
    f <- suppressMessages(nmfkc.signed(Y, A, rank = rank[i],
                                       verbose = FALSE, ...))
    rs[i] <- f$r.squared
    er[i] <- .effective.rank(f$B)
  }
  ecv <- if (detail == "full")
    suppressMessages(nmfkc.signed.ecv(Y, A, rank = rank, ...))$sigma
    else rep(NA_real_, length(rank))
  criteria <- data.frame(rank = rank, effective.rank = er,
                         effective.rank.ratio = er / rank,
                         r.squared = rs, sigma.ecv = as.numeric(ecv))
  .rank.finish(criteria, plot = plot,
               main = "nmfkc.signed rank selection")
}

## ============================================================
## nmfkc.ard(): Automatic Relevance Determination for NMF rank
## (Tan & Fevotte 2013, beta = 2 / Euclidean).  PROTOTYPE.
##
## Fit NMF at an over-complete rank K and let a per-component relevance
## weight lambda_k (inverse-gamma prior) automatically prune unneeded
## components: the multiplicative updates gain a penalty term, and
## components whose energy collapses to ~0 are dropped.  The number of
## surviving components is the estimated rank -- a single fit, no scan.
## ============================================================

## Internal: one ARD-NMF fit from a single random initialization.
#' @keywords internal
#' @noRd
.ard.fit <- function(Y, K, prior, a, b, maxit, epsilon, tol, seed) {
  Fr <- base::nrow(Y); Nc <- base::ncol(Y); eps <- 1e-10; m <- base::mean(Y)
  if (!base::is.null(seed)) base::set.seed(seed)
  W <- base::matrix(stats::runif(Fr * K), Fr, K) * base::sqrt(m / K)
  H <- base::matrix(stats::runif(K * Nc), K, Nc) * base::sqrt(m / K)
  clam <- if (prior == "L2") (Fr + Nc) / 2 + a + 1 else (Fr + Nc) + a + 1

  objfunc <- base::numeric(maxit)
  for (it in base::seq_len(maxit)) {
    lam <- if (prior == "L2")
      (b + 0.5 * (base::colSums(W^2) + base::rowSums(H^2))) / clam
    else
      (b + base::colSums(W) + base::rowSums(H)) / clam
    num <- Y %*% base::t(H); den <- W %*% (H %*% base::t(H))
    pen <- if (prior == "L2") base::sweep(W, 2, 1 / lam, "*")
           else base::matrix(1 / lam, Fr, K, byrow = TRUE)
    W <- W * num / (den + pen + eps)
    num <- base::t(W) %*% Y; den <- (base::t(W) %*% W) %*% H
    pen <- if (prior == "L2") base::sweep(H, 1, 1 / lam, "*")
           else base::matrix(1 / lam, K, Nc)
    H <- H * num / (den + pen + eps)
    objfunc[it] <- 0.5 * base::sum((Y - W %*% H)^2)
    if (it > 1 && base::abs(objfunc[it - 1] - objfunc[it]) <
        epsilon * objfunc[it - 1]) { objfunc <- objfunc[1:it]; break }
  }

  energy <- base::sqrt(base::colSums(W^2) * base::rowSums(H^2))
  ord <- base::order(energy, decreasing = TRUE)
  energy <- energy[ord]; W <- W[, ord, drop = FALSE]; H <- H[ord, , drop = FALSE]
  lam <- lam[ord]
  relevance <- if (base::max(energy) > 0) energy / base::max(energy) else energy
  base::list(rank = base::sum(relevance > tol), relevance = relevance,
             lambda = lam, W = W, H = H, objfunc = objfunc)
}

#' @title Automatic relevance determination for NMF rank (experimental)
#' @description
#' \strong{Prototype} of Tan & Fevotte's (2013) ARD-NMF (Euclidean /
#' \eqn{\beta = 2}).  Unlike the cross-validation (\code{\link{nmfkc.ecv}},
#' \code{\link{nmfkc.bicv}}) and stability (\code{\link{nmfkc.consensus}})
#' engines that \emph{scan} a range of ranks, ARD fits NMF \strong{once} at
#' an over-complete rank and prunes automatically: each component \eqn{k}
#' carries a relevance weight \eqn{\lambda_k} with an inverse-gamma prior;
#' the multiplicative updates gain a penalty \eqn{+ w_{fk}/\lambda_k} (L2 /
#' half-normal prior) or \eqn{+ 1/\lambda_k} (L1 / exponential), which
#' drives unsupported components to zero.  The number of surviving
#' components is the estimated rank.  Covariates are ignored (plain NMF).
#'
#' This is a model-based point estimate: the result depends on the prior, the
#' starting rank and the random initialization, and can vary run to run.  Use
#' it as a complement to the CV / consensus engines, not a sole criterion.
#'
#' @details
#' \strong{Relation to Tan & Fevotte (2013).}  The update equations reproduce
#' the paper's \eqn{\ell_2} / \eqn{\ell_1} ARD-NMF for the \strong{Euclidean
#' (\eqn{\beta = 2})} case exactly: the multiplicative penalties, the
#' closed-form \eqn{\lambda_k} update \eqn{(f(w_k) + f(h_k) + b)/c} and the
#' constant \eqn{c} (their Eq. 33).  Two deliberate simplifications keep it
#' practical: only \eqn{\beta = 2} is implemented (the paper covers the general
#' \eqn{\beta}-divergence), and the default \code{b} is an \emph{empirical}
#' per-component energy scale \eqn{(F + N)/K \cdot \bar{Y}} that reliably avoids
#' the winner-take-all collapse, rather than the paper's method-of-moments value
#' (their Eq. 38).
#'
#' @param Y Observation matrix (\eqn{F \times N}), non-negative.
#' @param rank Over-complete starting rank \eqn{K} (must exceed the true
#'   rank).  \code{NULL} (default) uses \code{min(F, N, 20)}.
#' @param nrun Number of random-initialization restarts (default \code{10}).
#'   ARD is a sensitive point estimate, so several restarts are advisable; the
#'   reported rank is the \strong{mode} of the per-run estimates
#'   (\code{rank.runs}), with a representative modal fit kept for
#'   \code{plot}/\code{W}/\code{H}.
#' @param plot Logical; draw the relevance bar plot.
#' @param ... Advanced options, rarely needed (defaults in parentheses):
#'   \code{prior} (\code{"L2"}: half-normal / squared-energy group
#'   shrinkage, \code{"L1"}: exponential / sparser); \code{seed} (\code{123},
#'   random-initialization seed); \code{a} (\code{1}) and
#'   \code{b} (\code{(F + N)/K * mean(Y)}), the inverse-gamma prior (a smaller
#'   \code{b} over-prunes, a larger one prunes nothing); \code{maxit}
#'   (\code{3000}) and \code{epsilon} (\code{1e-6}) for optimisation control;
#'   \code{tol} (\code{1e-3}), the relevance threshold below which a component is
#'   counted as pruned.
#' @return An object of class \code{"nmfkc.ard"}: a list with
#'   \code{rank} (estimated = mode over restarts), \code{rank.runs} (the
#'   per-run estimates), \code{relevance} (representative run, descending),
#'   \code{lambda}, \code{W}, \code{H} (ordered by relevance),
#'   \code{rank.init}, \code{prior}, \code{nrun} and \code{objfunc}.
#' @references V. Y. F. Tan and C. Fevotte (2013).  Automatic relevance
#'   determination in nonnegative matrix factorization with the
#'   beta-divergence.  \emph{IEEE TPAMI} 35(7):1592--1605.
#'   \doi{10.1109/TPAMI.2012.240}.
#' @seealso \code{\link{nmfkc.ecv}}, \code{\link{nmfkc.bicv}},
#'   \code{\link{nmfkc.consensus}}, \code{\link{nmfkc.rank}}
#' @export
#' @examples
#' \donttest{
#' set.seed(1)
#' X <- matrix(abs(rnorm(40 * 3)), 40, 3)
#' B <- matrix(abs(rnorm(3 * 60)), 3, 60)
#' ar <- nmfkc.ard(X %*% B, rank = 10)   # over-complete start
#' ar$rank                                # ~ 3 surviving components
#' plot(ar)
#' }
nmfkc.ard <- function(Y, rank = NULL, nrun = 10, plot = FALSE, ...) {
  Y <- base::as.matrix(Y); Fr <- base::nrow(Y); Nc <- base::ncol(Y)
  K <- if (base::is.null(rank)) base::min(Fr, Nc, 20L) else rank

  ## Advanced options live in `...` with safe defaults (rarely changed).
  ## Default prior scale ties b to the initial per-component energy scale
  ## (F + N)/K * mean(Y); a fixed small fraction of mean(Y) over-prunes
  ## (winner-take-all collapse) when (F + N)/K is large.
  dots    <- base::list(...)
  prior   <- base::match.arg(if (base::is.null(dots$prior)) "L2" else dots$prior,
                             c("L2", "L1"))
  seed    <- if (base::is.null(dots$seed))    123     else dots$seed
  a       <- if (base::is.null(dots$a))       1       else dots$a
  b       <- if (base::is.null(dots$b))       (Fr + Nc) / K * base::mean(Y) else dots$b
  maxit   <- if (base::is.null(dots$maxit))   3000    else dots$maxit
  epsilon <- if (base::is.null(dots$epsilon)) 1e-6    else dots$epsilon
  tol     <- if (base::is.null(dots$tol))     1e-3    else dots$tol

  ## nrun random-initialization restarts (ARD is a sensitive point
  ## estimate); aggregate the integer rank by its MODE.
  fits <- base::lapply(base::seq_len(nrun), function(r)
    .ard.fit(Y, K, prior, a, b, maxit, epsilon, tol,
             seed = if (base::is.null(seed)) NULL else seed + r))
  rank.runs <- base::vapply(fits, function(f) f$rank, base::integer(1))
  tab <- base::sort(base::table(rank.runs), decreasing = TRUE)
  rank.mode <- base::as.integer(base::names(tab)[1])
  ## representative fit = a modal-rank run with the smallest residual
  modal <- base::which(rank.runs == rank.mode)
  rep.i <- modal[base::which.min(base::vapply(modal,
            function(i) fits[[i]]$objfunc[base::length(fits[[i]]$objfunc)],
            base::numeric(1)))]
  rep <- fits[[rep.i]]

  out <- base::structure(
    base::list(rank = rank.mode, rank.runs = rank.runs,
               relevance = rep$relevance, lambda = rep$lambda,
               W = rep$W, H = rep$H, rank.init = K, prior = prior,
               nrun = nrun, objfunc = rep$objfunc, tol = tol),
    class = "nmfkc.ard")
  if (plot) graphics::plot(out)
  out
}


#' @title Print method for nmfkc.ard objects
#' @param x An object of class \code{"nmfkc.ard"}.
#' @param ... Unused.
#' @return \code{x}, invisibly.
#' @export
print.nmfkc.ard <- function(x, ...) {
  base::cat(base::sprintf(
    "ARD-NMF rank determination (Tan-Fevotte 2013, %s prior, %d run%s)\n",
    x$prior, x$nrun, if (x$nrun > 1) "s" else ""))
  base::cat(base::sprintf("  starting rank: %d  ->  estimated rank (mode): %d\n",
                          x$rank.init, x$rank))
  if (!base::is.null(x$rank.runs) && x$nrun > 1) {
    tab <- base::table(x$rank.runs)
    base::cat("  rank over runs:",
              base::paste(base::sprintf("%s:%d", base::names(tab), tab),
                          collapse = "  "),
              base::sprintf("(median %g)\n", stats::median(x$rank.runs)))
  }
  base::cat("  relevance (representative run):",
            base::paste(base::round(x$relevance, 3), collapse = " "), "\n")
  base::invisible(x)
}


#' @title Plot method for nmfkc.ard objects
#' @description Bar plot of the per-component relevance (descending), with a
#'   line at the pruning threshold; bars above it (the estimated rank) are
#'   highlighted.
#' @param x An object of class \code{"nmfkc.ard"}.
#' @param main Plot title.
#' @param ... Passed to \code{\link[graphics]{barplot}}.
#' @return \code{x}, invisibly.
#' @export
plot.nmfkc.ard <- function(x, main = NULL, ...) {
  if (base::is.null(main))
    main <- base::sprintf("ARD-NMF: estimated rank = %d (of %d)",
                          x$rank, x$rank.init)
  keep <- x$relevance > x$tol
  cols <- base::ifelse(keep, "steelblue", "grey80")
  graphics::barplot(x$relevance, col = cols, border = NA,
                    names.arg = base::seq_along(x$relevance),
                    xlab = "component (by relevance)", ylab = "relevance",
                    main = main, ...)
  graphics::abline(h = x$tol, col = "red", lty = 2)
  base::invisible(x)
}

## Tests for nmfkc.ard() (ARD-NMF rank determination prototype)

test_that("nmfkc.ard prunes an over-complete fit toward the true rank", {
  set.seed(1)
  X <- matrix(abs(rnorm(40 * 3)), 40, 3)
  B <- matrix(abs(rnorm(3 * 60)), 3, 60)
  ar <- nmfkc.ard(X %*% B, rank = 10, seed = 1)   # clean rank-3 signal

  expect_s3_class(ar, "nmfkc.ard")
  expect_equal(ar$rank.init, 10)
  expect_length(ar$relevance, 10)
  ## relevance is in [0, 1], descending, first = 1
  expect_true(all(ar$relevance >= 0 & ar$relevance <= 1))
  expect_equal(ar$relevance[1], 1)
  expect_false(is.unsorted(rev(ar$relevance)))
  ## an over-complete fit should prune: estimated rank well below the start
  expect_lt(ar$rank, 10)
  expect_true(ar$rank >= 2 && ar$rank <= 5)   # near the true 3
  ## W, H are ordered/sized to the starting rank
  expect_equal(ncol(ar$W), 10)
  expect_equal(nrow(ar$H), 10)
})

test_that("nmfkc.ard print and plot run without error", {
  set.seed(2)
  Y <- matrix(abs(rnorm(30 * 40)), 30, 40)
  ar <- nmfkc.ard(Y, rank = 8, seed = 2)
  expect_output(print(ar), "estimated rank")
  pdf(NULL); on.exit(dev.off(), add = TRUE)
  expect_no_error(plot(ar))
})

test_that("nrun aggregates the rank by mode over restarts", {
  set.seed(4)
  X <- matrix(abs(rnorm(40 * 3)), 40, 3)
  B <- matrix(abs(rnorm(3 * 60)), 3, 60)
  ar <- nmfkc.ard(X %*% B, rank = 8, nrun = 10, seed = 4)
  expect_equal(ar$nrun, 10)
  expect_length(ar$rank.runs, 10)
  ## reported rank is the mode of the per-run estimates
  tab <- table(ar$rank.runs)
  expect_equal(ar$rank, as.integer(names(tab)[which.max(tab)]))
  expect_output(print(ar), "rank over runs")
})

test_that("L1 prior also returns a valid pruned rank", {
  set.seed(3)
  X <- matrix(abs(rnorm(40 * 3)), 40, 3)
  B <- matrix(abs(rnorm(3 * 60)), 3, 60)
  ar <- nmfkc.ard(X %*% B, rank = 10, prior = "L1", seed = 3)
  expect_equal(ar$prior, "L1")
  expect_true(ar$rank >= 1 && ar$rank <= 10)
})

# nmfae.R
# 3-layer NMF (NMF-AE): Y1 ≈ X1 C X2 Y2
# Author: Kenichi Satoh
# Date: 2026-03-07
#
# Depends: nmfkc
# Suggests: DiagrammeR, DiagrammeRsvg, rsvg (for DOT graph rendering)

#' @title Three-Layer Non-negative Matrix Factorization (NMF-AE)
#' @description
#' \code{nmfae} fits a three-layer nonnegative matrix factorization model
#' \eqn{Y_1 \approx X_1 \Theta X_2 Y_2}, where \eqn{X_1} is a decoder basis
#' (column sum 1), \eqn{\Theta} is a bottleneck parameter matrix,
#' \eqn{X_2} is an encoder basis (row sum 1), and \eqn{Y_2} is the input matrix.
#'
#' When \code{Y2 = Y1}, the model acts as a non-negative autoencoder.
#' When \code{Y1 != Y2}, it acts as a heteroencoder.
#'
#' Initialization uses a three-step NMF procedure via \code{\link{nmfkc}}:
#' (1) \code{nmfkc(Y1, rank=Q)} to obtain \eqn{X_1},
#' (2) \code{nmfkc(Y1, A=Y2, rank=Q)} with fixed \eqn{X_1} to obtain \eqn{C = \Theta X_2},
#' (3) \code{nmfkc(Y2, rank=R)} to factor \eqn{C} into \eqn{\Theta} and \eqn{X_2}.
#'
#' @param Y1 Output matrix \eqn{Y_1} (P1 x N). Non-negative. May contain \code{NA}s
#'   (handled via \code{Y1.weights}).
#' @param Y2 Input matrix \eqn{Y_2} (P2 x N). Non-negative. Default is \code{Y1} (autoencoder).
#' @param rank Integer. Rank of the decoder basis \eqn{X_1} (P1 x Q). Default is 2.
#'   For backward compatibility, \code{Q} is accepted via \code{...}.
#' @param rank.encoder Integer. Rank of the encoder basis \eqn{X_2} (R x P2).
#'   Default is \code{rank}. For backward compatibility, \code{R} is accepted via \code{...}.
#' @param epsilon Positive convergence tolerance. Default is \code{1e-4}.
#' @param maxit Maximum number of multiplicative update iterations. Default is 5000.
#' @param verbose Logical. If \code{TRUE}, prints progress messages during fitting. Default is \code{FALSE}.
#' @param ... Additional arguments:
#'   \describe{
#'     \item{\code{Y1.weights}}{Optional non-negative weight matrix
#'       (P1 x N) or vector for \eqn{Y_1}, analogous to the
#'       \code{weights} argument of \code{\link[stats]{lm}}.  Loss becomes
#'       \eqn{\sum W_{ij} \, (Y_{1,ij} - \hat Y_{1,ij})^2}
#'       (\code{lm()}-style, \strong{linear} in \eqn{W}).  Logical
#'       matrices (\code{TRUE} / \code{FALSE}) are also accepted.
#'       Typical ECV / CV usage passes a binary mask
#'       \eqn{W \in \{0,1\}} for held-out elements; real-valued weights
#'       for importance weighting are also supported.  Default: if
#'       \code{Y1} has \code{NA}, a binary mask is auto-generated
#'       (0 for \code{NA}, 1 elsewhere).}
#'     \item{\code{C.L1}}{L1 regularization parameter for \eqn{C}. Default is 0.}
#'     \item{\code{X1.L2.ortho}}{L2 orthogonality regularization for \eqn{X_1} columns. Default is 0.}
#'     \item{\code{X2.L2.ortho}}{L2 orthogonality regularization for \eqn{X_2} rows. Default is 0.}
#'     \item{\code{seed}}{Integer seed for reproducibility. Default is 123.}
#'     \item{\code{print.trace}}{Logical. If \code{TRUE}, prints progress. Default is \code{FALSE}.}
#'   }
#'
#' @return An object of class \code{"nmfae"}, a list with components:
#' \item{X1}{Decoder basis matrix (P1 x Q), column sum 1.}
#' \item{C}{Parameter matrix (Q x R).}
#' \item{X2}{Encoder basis matrix (R x P2), row sum 1.}
#' \item{Y1hat}{Fitted values \eqn{X_1 \Theta X_2 Y_2} (P1 x N).}
#' \item{rank}{Named integer vector \code{c(Q, R)}.}
#' \item{objfunc}{Final objective value.}
#' \item{objfunc.iter}{Objective values by iteration.}
#' \item{r.squared}{\eqn{\mathrm{cor}(Y, \widehat Y)^2} (Pearson; in \eqn{[0,1]}).}
#' \item{r.squared.uncentered}{Uncentered \eqn{R^2 = 1 - \|Y - \widehat Y\|_F^2 / \|Y\|_F^2} (baseline = zero matrix).}
#' \item{r.squared.centered}{Row-mean centered \eqn{1 - \|Y - \widehat Y\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2}.}
#' \item{niter}{Number of iterations performed.}
#' \item{runtime}{Elapsed time as a \code{difftime} object.}
#' \item{n.missing}{Number of missing (or zero-weighted) elements in \eqn{Y_1}.}
#' \item{n.total}{Total number of elements in \eqn{Y_1} (P1 x N).}
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in future versions.
#'
#' @seealso \code{\link{nmfae.inference}}, \code{\link{predict.nmfae}}, \code{\link{nmfae.ecv}}, \code{\link{nmfae.DOT}}, \code{\link{nmfkc}}
#' @export
#' @source Satoh, K. (2025). Applying Non-negative Matrix Factorization with Covariates
#'   to Multivariate Time Series. \emph{Japanese Journal of Statistics and Data Science}.
#' @references
#' Lee, D. D. and Seung, H. S. (2001). Algorithms for Non-negative Matrix Factorization.
#'   \emph{Advances in Neural Information Processing Systems}, 13.
#'
#' Saha, S. et al. (2022). Hierarchical Deep Learning Neural Network (HiDeNN):
#'   An Artificial Intelligence (AI) Framework for Computational Science and Engineering.
#'   \emph{Computer Methods in Applied Mechanics and Engineering}, 399.
#' @examples
#' # Autoencoder example
#' Y <- matrix(c(1,0,1,0, 0,1,0,1, 1,1,0,0), nrow=3, byrow=TRUE)
#' res <- nmfae(Y, rank=2, rank.encoder=2)
#' res$r.squared
#'
#' # Heteroencoder example
#' Y1 <- matrix(c(1,0,0,1), nrow=2)
#' Y2 <- matrix(runif(8), nrow=4)
#' res2 <- nmfae(Y1, Y2, rank=2, rank.encoder=2)
#'
nmfae <- function(Y1, Y2 = Y1, rank = 2, rank.encoder = rank,
                  epsilon = 1e-4, maxit = 5000, verbose = FALSE, ...) {

  cl <- match.call()

  extra_args <- list(...)
  # backward compatibility: Q -> rank, R -> rank.encoder
  if (!is.null(extra_args$Q)) rank <- extra_args$Q
  if (!is.null(extra_args$R)) rank.encoder <- extra_args$R
  Q <- rank
  R <- rank.encoder
  Y1.weights  <- if (!is.null(extra_args$Y1.weights))  extra_args$Y1.weights  else NULL
  C.L1        <- if (!is.null(extra_args$C.L1))        extra_args$C.L1        else 0
  X1.L2.ortho <- if (!is.null(extra_args$X1.L2.ortho)) extra_args$X1.L2.ortho else 0
  X2.L2.ortho <- if (!is.null(extra_args$X2.L2.ortho)) extra_args$X2.L2.ortho else 0
  seed        <- if (!is.null(extra_args$seed))        extra_args$seed        else 123
  print.trace <- verbose
  if (!is.null(extra_args$print.trace)) print.trace <- extra_args$print.trace  # backward compat

  # --- Input validation ---
  Y1 <- as.matrix(Y1); storage.mode(Y1) <- "double"
  Y2 <- as.matrix(Y2); storage.mode(Y2) <- "double"
  if (min(Y2, na.rm = TRUE) < 0) stop("Y2 must be non-negative.")
  if (ncol(Y1) != ncol(Y2)) stop("Y1 and Y2 must have the same number of columns (N).")

  P1 <- nrow(Y1); P2 <- nrow(Y2); N <- ncol(Y1)
  eps <- 1e-10
  start.time <- Sys.time()

  # --- Y1.weights handling (cf. nmfkc Y.weights) ---
  # Vector -> matrix expansion
  if (!is.null(Y1.weights) && is.vector(Y1.weights)) {
    if (length(Y1.weights) == N) {
      Y1.weights <- matrix(Y1.weights, nrow = P1, ncol = N, byrow = TRUE)
    } else if (length(Y1.weights) == 1) {
      Y1.weights <- matrix(Y1.weights, nrow = P1, ncol = N)
    } else {
      stop("Length of Y1.weights vector must match ncol(Y1) or be 1.")
    }
  }
  # NULL: auto-detect NAs
  if (is.null(Y1.weights)) {
    if (any(is.na(Y1))) {
      Y1.weights <- matrix(1, nrow = P1, ncol = N)
      Y1.weights[is.na(Y1)] <- 0
      Y1[is.na(Y1)] <- 0
      if (print.trace) message("  Notice: NAs in Y1 detected, Y1.weights set automatically.")
    }
  } else {
    if (!is.matrix(Y1.weights)) Y1.weights <- as.matrix(Y1.weights)
    if (!all(dim(Y1.weights) == dim(Y1))) stop("Dimension mismatch between Y1 and Y1.weights.")
    Y1.weights[is.na(Y1.weights)] <- 0
    Y1[is.na(Y1)] <- 0
    Y1[Y1.weights == 0] <- 0
  }
  # Determine code path
  has.weights <- !is.null(Y1.weights) && any(Y1.weights != 1)
  if (!has.weights) Y1.weights <- NULL

  if (min(Y1, na.rm = TRUE) < 0) stop("Y1 must be non-negative.")

  if (print.trace) {
    message(sprintf("Y1(%d,%d) ~ X1(%d,%d) C(%d,%d) X2(%d,%d) Y2(%d,%d)",
                    P1, N, P1, Q, Q, R, R, P2, P2, N))
  }

  # === Initialization using nmfkc ===
  # Step 1: X1 from nmfkc(Y1, rank=Q)
  if (print.trace) message("  Init step 1: nmfkc(Y1, rank=Q)...")
  res1 <- nmfkc(Y1, rank = Q, seed = seed, print.dims = FALSE,
                Y.weights = Y1.weights)
  X1 <- res1$X  # P1 x Q, column sum 1

  # Step 2: CX2 = C X2 with X1 fixed
  if (print.trace) message("  Init step 2: nmfkc(Y1, A=Y2, X.restriction='fixed')...")
  res2 <- nmfkc(Y1, A = Y2, rank = Q,
                 X.init = X1, X.restriction = "fixed",
                 seed = seed, print.dims = FALSE,
                 Y.weights = Y1.weights)
  CX2_init <- res2$C  # Q x P2

  # Step 3: X2 from nmfkc(Y2, rank=R), then C from CX2 ≈ C X2
  if (print.trace) message("  Init step 3: nmfkc(Y2, rank=R)...")
  res3 <- nmfkc(Y2, rank = R, seed = seed, print.dims = FALSE)
  X2 <- t(res3$X)             # R x P2, row sum 1
  C <- CX2_init %*% t(X2)     # Q x R, non-negative

  # Precompute (only needed for unweighted path)
  if (!has.weights) Y2Y2t <- tcrossprod(Y2)  # P2 x P2

  # === Multiplicative updates ===
  objfunc.iter <- numeric(maxit)

  W <- Y1.weights  # alias for readability (NULL when unweighted)

  for (iter in 1:maxit) {
    # 1. Update X1: Y1 ≈ X1 F, F = C X2 Y2
    F_mat <- C %*% X2 %*% Y2                           # Q x N
    if (has.weights) {
      num_X1 <- (W * Y1) %*% t(F_mat)
      den_X1 <- (W * (X1 %*% F_mat)) %*% t(F_mat) + eps
    } else {
      num_X1 <- tcrossprod(Y1, F_mat)
      den_X1 <- X1 %*% tcrossprod(F_mat) + eps
    }
    if (X1.L2.ortho > 0) {
      X1tX1 <- crossprod(X1); diag(X1tX1) <- 0
      den_X1 <- den_X1 + X1.L2.ortho * (X1 %*% X1tX1)
    }
    X1 <- X1 * (num_X1 / den_X1)

    # 2. Normalize X1 columns -> absorb scale into C (left)
    cs <- colSums(X1)
    X1 <- sweep(X1, 2, cs, "/")
    C <- sweep(C, 1, cs, "*")

    # 3. Update C: Y1 ≈ X1 C G, G = X2 Y2
    G <- X2 %*% Y2                                     # R x N
    if (has.weights) {
      num_C <- crossprod(X1, W * Y1) %*% t(G)
      den_C <- crossprod(X1, W * (X1 %*% C %*% G)) %*% t(G) + eps
    } else {
      num_C <- crossprod(X1, Y1) %*% t(G)
      den_C <- crossprod(X1) %*% C %*% tcrossprod(G) + eps
    }
    if (C.L1 > 0) den_C <- den_C + (C.L1 / 2)
    C <- C * (num_C / den_C)

    # 4. Update X2: Y1 ≈ H X2 Y2, H = X1 C
    H <- X1 %*% C                                      # P1 x R
    if (has.weights) {
      num_X2 <- crossprod(H, W * Y1) %*% t(Y2)
      den_X2 <- crossprod(H, W * (H %*% X2 %*% Y2)) %*% t(Y2) + eps
    } else {
      num_X2 <- crossprod(H, Y1) %*% t(Y2)
      den_X2 <- crossprod(H) %*% X2 %*% Y2Y2t + eps
    }
    if (X2.L2.ortho > 0) {
      X2X2t <- tcrossprod(X2); diag(X2X2t) <- 0
      den_X2 <- den_X2 + X2.L2.ortho * (X2X2t %*% X2)
    }
    X2 <- X2 * (num_X2 / den_X2)

    # 5. Normalize X2 rows -> absorb scale into C (right)
    rs <- rowSums(X2)
    X2 <- sweep(X2, 1, rs, "/")
    C <- sweep(C, 2, rs, "*")

    # Objective function (with regularization penalties)
    # lm()-style weighted least squares: L = sum(W * (Y1 - Y1hat)^2).
    # The MU (num_X1, num_C, etc.) carries W linearly, so reporting the
    # linear-W objective here keeps MU target and reported loss consistent.
    # For binary W in {0,1} (standard ECV / NA-mask case) this is identical
    # to sum((W*(Y1-Y1hat))^2) since W == W^2.
    Y1hat <- X1 %*% C %*% X2 %*% Y2
    if (has.weights) {
      obj <- sum(W * (Y1 - Y1hat)^2)
    } else {
      obj <- sum((Y1 - Y1hat)^2)
    }
    if (C.L1 > 0) obj <- obj + C.L1 * sum(C)
    if (X1.L2.ortho > 0) {
      X1tX1 <- crossprod(X1); diag(X1tX1) <- 0
      obj <- obj + (X1.L2.ortho / 2) * sum(X1tX1^2)
    }
    if (X2.L2.ortho > 0) {
      X2X2t <- tcrossprod(X2); diag(X2X2t) <- 0
      obj <- obj + (X2.L2.ortho / 2) * sum(X2X2t^2)
    }
    objfunc.iter[iter] <- obj

    if (print.trace && (iter %% 100 == 0 || iter == 1)) {
      message(sprintf("  iter %5d: objfunc = %.6f", iter, objfunc.iter[iter]))
    }

    # Convergence check
    if (iter > 1) {
      rel_change <- abs(objfunc.iter[iter] - objfunc.iter[iter-1]) /
                       (objfunc.iter[iter-1] + eps)
      if (rel_change < epsilon) {
        if (print.trace) message(sprintf("  Converged at iter %d", iter))
        break
      }
    }
  }
  ## Warn when the MU loop exhausts maxit without meeting the
  ## relative-tolerance criterion (matches nmfkc() / nmf.sem() convention).
  if (iter == maxit && exists("rel_change") && rel_change >= epsilon)
    warning(paste0("maximum iterations (", maxit, ") reached..."))

  niter <- iter
  objfunc.iter <- objfunc.iter[1:niter]
  diff.time <- as.numeric(difftime(Sys.time(), start.time, units = "secs"))

  # --- Reorder bases by centroid position (cf. nmfkc) ---
  # X1 (P1 x Q): sort columns by weighted centroid of row indices
  if (Q > 1) {
    idx1 <- order(matrix(seq_len(P1) / P1, nrow = 1) %*% X1)
    X1 <- X1[, idx1, drop = FALSE]
    C <- C[idx1, , drop = FALSE]
  }
  # X2 (R x P2): sort rows by weighted centroid of column indices
  if (R > 1) {
    idx2 <- order(X2 %*% matrix(seq_len(P2) / P2, ncol = 1))
    X2 <- X2[idx2, , drop = FALSE]
    C <- C[, idx2, drop = FALSE]
  }

  # --- Assign Dec/Enc names ---
  colnames(X1) <- paste0("Dec", 1:Q)
  rownames(C)  <- paste0("Dec", 1:Q)
  colnames(C)  <- paste0("Enc", 1:R)
  rownames(X2) <- paste0("Enc", 1:R)

  Y1hat <- X1 %*% C %*% X2 %*% Y2
  objfunc <- utils::tail(objfunc.iter, 1)

  # R-squared, sigma, mae, and missing count
  if (has.weights) {
    valid <- (W > 0)
    n.missing <- sum(!valid)
    n.valid <- sum(valid)
    r2_all <- .r.squared.all(Y1, Y1hat, Y.weights = W)
    sigma <- sqrt(objfunc / n.valid)
    mae <- mean(abs(Y1[valid] - Y1hat[valid]))
  } else {
    n.missing <- 0L
    n.valid <- P1 * N
    r2_all <- .r.squared.all(Y1, Y1hat)
    sigma <- sqrt(objfunc / n.valid)
    mae <- mean(abs(Y1 - Y1hat))
  }
  r.squared          <- r2_all$r.squared
  r.squared.uncentered     <- r2_all$r.squared.uncentered
  r.squared.centered <- r2_all$r.squared.centered

  if (print.trace) {
    msg <- sprintf("  Done: %d iterations, %.1f sec, R2 = %.4f", niter, diff.time, r.squared)
    if (n.missing > 0) msg <- paste0(msg, sprintf(", missing=%d", n.missing))
    message(msg)
  }

  # --- Soft/hard clustering (cf. nmfkc B.prob / B.cluster) ---
  H <- C %*% X2 %*% Y2   # Q x N encoding
  eps_bp <- 1e-16
  B.prob <- t( t(H) / (colSums(H) + eps_bp) )   # column-normalized
  B.cluster <- apply(B.prob, 2, which.max)
  B.cluster[colSums(H) == 0] <- NA

  result <- list(
    call = cl,
    X1 = X1,
    C = C,
    X2 = X2,
    Y1hat = Y1hat,
    H = H,
    B.prob = B.prob,
    B.cluster = B.cluster,
    rank = c(Q = Q, R = R),
    dims = c(P1 = P1, P2 = P2, N = N),
    objfunc = objfunc,
    objfunc.iter = objfunc.iter,
    r.squared          = r.squared,
    r.squared.uncentered     = r.squared.uncentered,
    r.squared.centered = r.squared.centered,
    sigma = sigma,
    mae = mae,
    niter = niter,
    runtime = diff.time,
    n.missing = n.missing,
    n.total = P1 * N
  )
  class(result) <- c("nmfae", "nmf")
  return(result)
}

#' @title Statistical Inference for NMF-AE Parameter Matrix
#' @description
#' Performs post-estimation inference for \eqn{\Theta} in the three-layer NMF model
#' \eqn{Y_1 \approx X_1 \Theta X_2 Y_2}, conditional on \eqn{(\hat{X}_1, \hat{X}_2)}.
#' Uses sandwich covariance estimation and one-step wild bootstrap with
#' non-negative projection.
#'
#' @param object An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.
#' @param Y1 Output matrix \eqn{Y_1} (P1 x N). Must match the data used in \code{nmfae()}.
#' @param Y2 Input matrix \eqn{Y_2} (P2 x N). Default is \code{Y1} (autoencoder).
#' @param wild.bootstrap Logical. If \code{TRUE} (default), performs wild bootstrap
#'   for bootstrap SE and confidence intervals. If \code{FALSE}, only sandwich SE
#'   and z-test p-values are computed (faster).
#' @param ... Additional arguments:
#'   \describe{
#'     \item{\code{wild.B}}{Number of bootstrap replicates. Default is 1000.}
#'     \item{\code{wild.seed}}{Seed for bootstrap. Default is 42.}
#'     \item{\code{wild.level}}{Confidence level for bootstrap CI. Default is 0.95.}
#'     \item{\code{sandwich}}{Logical. Use sandwich covariance. Default is \code{TRUE}.}
#'     \item{\code{C.p.side}}{P-value type: \code{"one.sided"} (default) or \code{"two.sided"}.}
#'     \item{\code{cov.ridge}}{Ridge stabilization for information matrix inversion. Default is 1e-8.}
#'     \item{\code{print.trace}}{Logical. If \code{TRUE}, prints progress. Default is \code{FALSE}.}
#'   }
#'
#' @return An object of class \code{c("nmfae.inference", "nmfae")}, inheriting all
#' components from the input \code{object}, with additional inference components:
#' \item{sigma2.used}{Estimated \eqn{\sigma^2} used for inference.}
#' \item{C.se}{Sandwich standard errors for \eqn{\Theta} (Q x R matrix).}
#' \item{C.se.boot}{Bootstrap standard errors for \eqn{\Theta} (Q x R matrix).}
#' \item{C.ci.lower}{Lower CI bounds for \eqn{\Theta} (Q x R matrix).}
#' \item{C.ci.upper}{Upper CI bounds for \eqn{\Theta} (Q x R matrix).}
#' \item{coefficients}{Data frame with Estimate, SE, BSE, z, p-value for each element of \eqn{\Theta}.}
#' \item{C.p.side}{P-value type used.}
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @seealso \code{\link{nmfae}}, \code{\link{summary.nmfae.inference}}
#' @export
#' @examples
#' Y <- matrix(c(1,0,1,0, 0,1,0,1, 1,1,0,0), nrow=3, byrow=TRUE)
#' res <- nmfae(Y, rank=2, rank.encoder=2)
#' res2 <- nmfae.inference(res, Y)
#' summary(res2)
#'
nmfae.inference <- function(object, Y1, Y2 = Y1,
                            wild.bootstrap = TRUE, ...) {
  if (!inherits(object, "nmfae")) stop("object must be of class 'nmfae'")

  extra_args <- list(...)
  wild.B     <- if (!is.null(extra_args$wild.B))     extra_args$wild.B     else 500
  wild.seed  <- if (!is.null(extra_args$wild.seed))  extra_args$wild.seed  else 123
  wild.level <- if (!is.null(extra_args$wild.level)) extra_args$wild.level else 0.95
  sandwich   <- if (!is.null(extra_args$sandwich))   extra_args$sandwich   else TRUE
  C.p.side   <- if (!is.null(extra_args$C.p.side))   extra_args$C.p.side   else "one.sided"
  cov.ridge  <- if (!is.null(extra_args$cov.ridge))  extra_args$cov.ridge  else 1e-8
  print.trace <- if (!is.null(extra_args$print.trace)) extra_args$print.trace else FALSE

  X1 <- object$X1
  C  <- object$C
  X2 <- object$X2
  Y1hat <- object$Y1hat
  Q  <- object$rank["Q"]
  R  <- object$rank["R"]
  P1 <- nrow(Y1)
  N  <- ncol(Y1)

  Z   <- X2 %*% Y2                        # R x N
  R_C <- Y1 - Y1hat                       # P1 x N  residuals

  # sigma2 estimate (denominator: PN - QR)
  denom <- max(P1 * N - Q * R, 1)
  sigma2.used <- sum(R_C^2) / denom

  # Information matrix: I = sigma^{-2} (ZZ' x X1'X1)
  X1tX1 <- crossprod(X1)                  # Q x Q
  ZZt   <- tcrossprod(Z)                  # R x R
  Info_core <- kronecker(ZZt, X1tX1)      # QR x QR
  Info <- Info_core / max(sigma2.used, 1e-12)
  Info <- Info + diag(cov.ridge, nrow(Info))

  Hinv <- tryCatch(solve(Info), error = function(e) {
    if (requireNamespace("MASS", quietly = TRUE)) MASS::ginv(Info)
    else stop("Information matrix singular; install MASS package.")
  })

  # Sandwich covariance: V = Hinv J Hinv
  V_sand <- NULL
  if (isTRUE(sandwich)) {
    X1t <- t(X1)
    J <- matrix(0, Q * R, Q * R)
    for (n in 1:N) {
      z_n <- Z[, n, drop = FALSE]
      r_n <- R_C[, n, drop = FALSE]
      g_n <- X1t %*% r_n
      S_n <- -(g_n %*% t(z_n)) / max(sigma2.used, 1e-12)
      s_n <- as.vector(S_n)
      J <- J + tcrossprod(s_n)
    }
    if (N > 1) J <- (N / (N - 1)) * J     # CR1 correction
    V_sand <- Hinv %*% J %*% Hinv
  }

  C.vec.cov <- if (!is.null(V_sand)) V_sand else Hinv

  # Sandwich SE
  se_vec <- sqrt(pmax(diag(C.vec.cov), 0))
  C.se <- matrix(se_vec, nrow = Q, ncol = R, byrow = FALSE)

  # ---- Wild bootstrap (one-step Newton) ----
  C.se.boot <- NULL
  C.ci.lower <- NULL
  C.ci.upper <- NULL

  if (isTRUE(wild.bootstrap)) {
    set.seed(wild.seed)
    X1t <- t(X1)
    score_mat <- matrix(0, Q * R, N)
    for (n in 1:N) {
      z_n <- Z[, n, drop = FALSE]
      r_n <- R_C[, n, drop = FALSE]
      g_n <- X1t %*% r_n
      G_n <- -(g_n %*% t(z_n)) / max(sigma2.used, 1e-12)
      score_mat[, n] <- as.vector(G_n)
    }

    C_hat_vec <- as.vector(C)
    C_boot <- matrix(NA_real_, nrow = Q * R, ncol = wild.B)
    for (b in 1:wild.B) {
      w <- stats::rexp(N, rate = 1) - 1     # Exp(1)-centered multiplier
      grad_b <- as.vector(score_mat %*% w)
      c_b <- C_hat_vec - as.vector(Hinv %*% grad_b)
      c_b <- pmax(c_b, 0)                   # project onto Theta >= 0
      C_boot[, b] <- c_b
    }

    # Bootstrap SE
    sd_vec <- apply(C_boot, 1, stats::sd, na.rm = TRUE)
    C.se.boot <- matrix(sd_vec, nrow = Q, ncol = R, byrow = FALSE)

    # Bootstrap CI
    alpha <- 1 - wild.level
    lo <- apply(C_boot, 1, stats::quantile, probs = alpha / 2, na.rm = TRUE, names = FALSE)
    hi <- apply(C_boot, 1, stats::quantile, probs = 1 - alpha / 2, na.rm = TRUE, names = FALSE)
    C.ci.lower <- matrix(lo, nrow = Q, ncol = R, byrow = FALSE)
    C.ci.upper <- matrix(hi, nrow = Q, ncol = R, byrow = FALSE)
  }

  # ---- Coefficients table ----
  Estimate <- as.vector(C)
  SE <- as.vector(C.se)
  BSE <- if (!is.null(C.se.boot)) as.vector(C.se.boot) else rep(NA_real_, length(Estimate))
  z_value <- ifelse(SE > 0, Estimate / SE, NA_real_)

  if (C.p.side == "one.sided") {
    p_value <- ifelse(is.finite(z_value), stats::pnorm(z_value, lower.tail = FALSE), NA_real_)
  } else {
    p_value <- ifelse(is.finite(z_value), 1 - stats::pchisq(z_value^2, df = 1), NA_real_)
  }

  # Row/column labels for C
  rlabs <- if (!is.null(rownames(C))) rownames(C) else paste0("Dec", 1:Q)
  clabs <- if (!is.null(colnames(C))) colnames(C) else paste0("Enc", 1:R)

  coefficients <- data.frame(
    Basis    = rep(rlabs, times = R),
    Covariate = rep(clabs, each = Q),
    Estimate = Estimate,
    SE       = SE,
    BSE      = BSE,
    z_value  = z_value,
    p_value  = p_value,
    CI_low   = if (!is.null(C.ci.lower)) as.vector(C.ci.lower) else NA_real_,
    CI_high  = if (!is.null(C.ci.upper)) as.vector(C.ci.upper) else NA_real_,
    row.names = NULL, stringsAsFactors = FALSE
  )

  if (print.trace) {
    if (isTRUE(wild.bootstrap)) {
      message("  Inference: sandwich SE + wild bootstrap done.")
    } else {
      message("  Inference: sandwich SE done (wild bootstrap skipped).")
    }
  }

  # Add inference fields to the object
  object$sigma2.used  <- sigma2.used
  object$C.se         <- C.se
  object$C.se.boot    <- C.se.boot
  object$C.ci.lower   <- C.ci.lower
  object$C.ci.upper   <- C.ci.upper
  object$coefficients <- coefficients
  object$C.p.side     <- C.p.side
  class(object) <- c("nmfae.inference", "nmf.inference", "nmfae", "nmf")
  return(object)
}

#' @title Rename decoder and encoder bases
#' @description
#' Assigns user-specified names to the decoder (X1 columns) and encoder
#' (X2 rows) bases of an \code{nmfae} object.  The names propagate to
#' \eqn{\Theta}, the coefficients table, and all downstream displays
#' such as \code{summary}, \code{nmfae.DOT}, and \code{nmfae.heatmap}.
#'
#' @param x An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.
#' @param X1.colnames Character vector of length \eqn{Q} for decoder bases
#'   (columns of \eqn{X_1} / rows of \eqn{\Theta}).  If \code{NULL}
#'   (default), the decoder names are left unchanged.
#' @param X2.rownames Character vector of length \eqn{R} for encoder bases
#'   (rows of \eqn{X_2} / columns of \eqn{\Theta}).  If \code{NULL}
#'   (default), the encoder names are left unchanged.
#'
#' @return A modified copy of \code{x} with updated names.
#' @examples
#' \donttest{
#' set.seed(1)
#' Y <- matrix(runif(15), nrow = 3)
#' res <- nmfae(Y, rank = 2, rank.encoder = 2)
#' res <- nmfae.rename(res,
#'   X1.colnames = c("Basis1", "Basis2"),
#'   X2.rownames = c("Enc1", "Enc2"))
#' summary(res)
#' }
#' @seealso \code{\link{nmfae}}
#' @export
nmfae.rename <- function(x, X1.colnames = NULL, X2.rownames = NULL) {
  if (!is.null(X1.colnames)) {
    Q <- ncol(x$X1)
    if (length(X1.colnames) != Q) stop("X1.colnames must have length ", Q)
    colnames(x$X1) <- X1.colnames
    rownames(x$C)  <- X1.colnames
    if (!is.null(x$coefficients)) {
      old <- paste0("Dec", 1:Q)
      for (k in seq_len(Q))
        x$coefficients$Basis[x$coefficients$Basis == old[k]] <- X1.colnames[k]
    }
  }
  if (!is.null(X2.rownames)) {
    R <- nrow(x$X2)
    if (length(X2.rownames) != R) stop("X2.rownames must have length ", R)
    rownames(x$X2) <- X2.rownames
    colnames(x$C)  <- X2.rownames
    if (!is.null(x$coefficients)) {
      old <- paste0("Enc", 1:R)
      for (k in seq_len(R))
        x$coefficients$Covariate[x$coefficients$Covariate == old[k]] <- X2.rownames[k]
    }
  }
  x
}

#' \code{plot.nmfae} displays the convergence trajectory of the objective function
#' across iterations. The title shows the achieved \eqn{R^2}.
#'
#' @param x An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.
#' @param ... Additional graphical parameters passed to \code{plot}.
#'
#' @return Invisible \code{NULL}. Called for its side effect (plot).
#' @seealso \code{\link{nmfae}}, \code{\link{nmfae.heatmap}}
#' @examples
#' \donttest{
#' set.seed(1)
#' Y <- matrix(runif(20), nrow = 4)
#' res <- nmfae(Y, rank = 2)
#' plot(res)
#' }
#' @export
plot.nmfae <- function(x, ...) {
  extra_args <- list(...)
  args <- list(x = x$objfunc.iter)
  if (is.null(extra_args$main))
    args$main <- paste0("R2 = ", round(x$r.squared, 3))
  if (is.null(extra_args$xlab)) args$xlab <- "iter"
  if (is.null(extra_args$ylab)) args$ylab <- "objfunc"
  if (is.null(extra_args$type)) args$type <- "l"
  all_args <- c(args, extra_args)
  do.call("plot", all_args)
  invisible(NULL)
}

#' @title Summary method for nmfae objects
#' @description
#' \code{summary.nmfae} produces a summary of a fitted NMF-AE model,
#' including dimensions, convergence status, goodness-of-fit statistics,
#' and structure diagnostics (sparsity of factor matrices).
#'
#' @param object An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.
#' @param ... Additional arguments (currently unused).
#'
#' @return An object of class \code{"summary.nmfae"}, a list with components:
#' \item{call}{The matched call.}
#' \item{dims}{Named vector \code{c(P1, P2, N)}.}
#' \item{Q}{Decoder rank.}
#' \item{R}{Encoder rank.}
#' \item{n.params}{Total number of parameters (P1*Q + Q*R + R*P2).}
#' \item{autoencoder}{Logical; TRUE if P1 == P2 and Y1 was used as Y2.}
#' \item{niter}{Number of iterations.}
#' \item{runtime}{Elapsed time.}
#' \item{objfunc}{Final objective value.}
#' \item{r.squared}{R-squared.}
#' \item{sigma}{Residual standard error (RMSE).}
#' \item{mae}{Mean absolute error.}
#' \item{n.missing}{Number of missing elements.}
#' \item{prop.missing}{Percentage of missing elements.}
#' \item{X1.sparsity}{Proportion of near-zero elements in X1.}
#' \item{C.sparsity}{Proportion of near-zero elements in C.}
#' \item{X2.sparsity}{Proportion of near-zero elements in X2.}
#'
#' @seealso \code{\link{nmfae}}, \code{\link{print.summary.nmfae}}
#' @export
summary.nmfae <- function(object, ...) {
  ans <- list()
  ans$call <- object$call
  ans$dims <- object$dims
  Q <- object$rank["Q"]
  R <- object$rank["R"]
  ans$Q <- Q
  ans$R <- R

  P1 <- object$dims["P1"]
  P2 <- object$dims["P2"]
  N  <- object$dims["N"]
  ans$n.params <- P1 * Q + Q * R + R * P2

  # Autoencoder detection
  ans$autoencoder <- (P1 == P2) &&
    !is.null(object$call$Y2) == FALSE
  # More robust: check if Y2 argument was not explicitly provided
  if (is.null(object$call$Y2)) {
    ans$autoencoder <- TRUE
  } else {
    ans$autoencoder <- FALSE
  }

  ans$niter <- object$niter
  ans$runtime <- object$runtime

  ans$objfunc <- object$objfunc
  ans$r.squared          <- object$r.squared
  ans$r.squared.uncentered     <- object$r.squared.uncentered
  ans$r.squared.centered <- object$r.squared.centered
  ans$sigma <- object$sigma
  ans$mae <- object$mae
  ## Effective rank of the latent encoding H (Q x N).
  ans$effective.rank <- if (!is.null(object$H)) .effective.rank(object$H) else NA_real_
  ans$rank <- if (!is.null(object$rank)) object$rank[1] else
              if (!is.null(object$H)) nrow(object$H) else NA

  # Missing values
  ans$n.missing <- object$n.missing
  if (!is.null(object$n.missing) && !is.null(object$n.total)) {
    ans$prop.missing <- object$n.missing / object$n.total * 100
  } else {
    ans$prop.missing <- 0
  }

  # Sparsity diagnostics
  if (!is.null(object$X1) && is.matrix(object$X1)) {
    ans$X1.sparsity <- mean(object$X1 < 1e-4)
  }
  if (!is.null(object$C) && is.matrix(object$C)) {
    ans$C.sparsity <- mean(object$C < 1e-4)
  }
  if (!is.null(object$X2) && is.matrix(object$X2)) {
    ans$X2.sparsity <- mean(object$X2 < 1e-4)
  }

  # Inference
  ans$coefficients <- object$coefficients
  ans$C.p.side <- object$C.p.side

  class(ans) <- "summary.nmfae"
  return(ans)
}

#' @title Print method for summary.nmfae objects
#' @description
#' Prints a formatted summary of an NMF-AE model fit.
#'
#' @param x An object of class \code{"summary.nmfae"}.
#' @param digits Minimum number of significant digits to be used.
#' @param max.coef Maximum number of coefficient rows to display. If the table
#'   has more rows, only significant rows (p < 0.05) are shown. Default is 20.
#' @param ... Additional arguments (currently unused).
#' @return Called for its side effect (printing). Returns \code{x} invisibly.
#' @seealso \code{\link{summary.nmfae}}
#' @export
print.summary.nmfae <- function(x, digits = max(3L, getOption("digits") - 3L),
                                max.coef = 20, ...) {
  cat("\nCall:\n",
      paste(deparse(x$call), sep = "\n", collapse = "\n"),
      "\n\n", sep = "")

  P1 <- x$dims["P1"]; P2 <- x$dims["P2"]; N <- x$dims["N"]
  Q <- x$Q; R <- x$R

  cat("Dimensions:\n")
  type_str <- if (x$autoencoder) "  (autoencoder)" else "  (heteroencoder)"
  cat("  Y1:             ", sprintf("%d x %d", P1, N), "\n")
  cat("  Y2:             ", sprintf("%d x %d", P2, N), type_str, "\n")
  cat("  Decoder rank Q: ", Q, "\n")
  cat("  Encoder rank R: ", R, "\n")
  cat("  Parameters:     ",
      sprintf("X1(%dx%d) + C(%dx%d) + X2(%dx%d) = %d",
              P1, Q, Q, R, R, P2, x$n.params), "\n")

  cat("\nConvergence:\n")
  cat("  Iterations:     ", x$niter, "\n")
  cat("  Runtime:        ", sprintf("%.1f secs", x$runtime), "\n")

  if (x$n.missing > 0) {
    cat("  Missing:        ", x$n.missing,
        sprintf("(%.1f%%)", x$prop.missing), "\n")
  }

  .print.fit.statistics(x, header = "Goodness of fit:", digits = digits)

  .print.structure.diagnostics(
    sparsity = c("Decoder (X1)" = x$X1.sparsity,
                 "Bottleneck (C)" = x$C.sparsity,
                 "Encoder (X2)" = x$X2.sparsity))

  # Coefficients table (inference)
  if (!is.null(x$coefficients) && is.data.frame(x$coefficients)) {
    cf <- x$coefficients
    n_total <- nrow(cf)
    rnames <- paste0(cf$Covariate, ":", cf$Basis)

    # Determine which rows to display
    if (n_total <= max.coef) {
      show_idx <- seq_len(n_total)
      truncated <- FALSE
    } else {
      sig_idx <- which(cf$p_value < 0.05)
      if (length(sig_idx) == 0) {
        show_idx <- seq_len(min(max.coef, n_total))
        truncated <- TRUE
      } else if (length(sig_idx) <= max.coef) {
        show_idx <- sig_idx
        truncated <- FALSE
      } else {
        show_idx <- sig_idx[seq_len(max.coef)]
        truncated <- TRUE
      }
    }

    n_sig <- sum(cf$p_value < 0.05, na.rm = TRUE)
    cat(sprintf("\nCoefficients (conditional on X1, X2): %d total, %d significant\n",
                n_total, n_sig))
    if (n_total > max.coef) {
      cat(sprintf("  (showing %d significant rows; use res$coefficients for full table)\n",
                  length(show_idx)))
    }

    p_side <- if (!is.null(x$C.p.side)) x$C.p.side else "one.sided"
    p_header <- if (p_side == "one.sided") "Pr(>z)" else "Pr(>|z|)"

    sig_stars <- function(p) {
      ifelse(!is.finite(p), " ",
        ifelse(p < 0.001, "***",
          ifelse(p < 0.01, "**",
            ifelse(p < 0.05, "*",
              ifelse(p < 0.1, ".", " ")))))
    }
    format_pval <- function(p) {
      ifelse(!is.finite(p), "      NA",
        ifelse(p < 2.2e-16, "  <2e-16",
          formatC(p, format = "g", digits = 4, width = 8)))
    }

    est <- formatC(cf$Estimate[show_idx], format = "f", digits = 3, width = 9)
    se  <- formatC(cf$SE[show_idx], format = "f", digits = 3, width = 10)
    bse <- formatC(cf$BSE[show_idx], format = "f", digits = 3, width = 6)
    zv  <- formatC(cf$z_value[show_idx], format = "f", digits = 2, width = 7)
    pv_str <- format_pval(cf$p_value[show_idx])
    stars <- sig_stars(cf$p_value[show_idx])
    show_names <- rnames[show_idx]

    max_lw <- max(nchar(show_names))
    hdr <- sprintf("%s %s %s %s %s %s",
                   formatC("Estimate", width = 9),
                   formatC("Std. Error", width = 10),
                   formatC("(Boot)", width = 6),
                   formatC("z value", width = 7),
                   formatC(p_header, width = 8), "")
    cat(sprintf("%s %s\n", formatC("Enc:Dec", width = max_lw), hdr))
    for (i in seq_along(show_names)) {
      cat(sprintf("%s %s %s %s %s %s %s\n",
                  formatC(show_names[i], width = max_lw),
                  est[i], se[i], bse[i], zv[i], pv_str[i], stars[i]))
    }
    cat("---\n")
    cat("Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1\n")
  }

  cat("\n")
  invisible(x)
}

#' @title Summary method for nmfae.inference objects
#' @description
#' Produces a summary of a fitted NMF-AE model with inference results,
#' including the coefficients table for \eqn{\Theta}.
#'
#' @param object An object of class \code{"nmfae.inference"}.
#' @param ... Additional arguments (currently unused).
#' @return An object of class \code{"summary.nmfae.inference"}.
#' @seealso \code{\link{nmfae.inference}}, \code{\link{summary.nmfae}}
#' @export
summary.nmfae.inference <- function(object, ...) {
  ans <- summary.nmfae(object, ...)
  class(ans) <- "summary.nmfae.inference"
  return(ans)
}

#' @title Print method for summary.nmfae.inference objects
#' @description
#' Prints a formatted summary including the coefficients table.
#' @param x An object of class \code{"summary.nmfae.inference"}.
#' @param digits Minimum number of significant digits.
#' @param max.coef Maximum coefficient rows to display. Default is 20.
#' @param ... Additional arguments (currently unused).
#' @return Called for its side effect (printing). Returns \code{x} invisibly.
#' @seealso \code{\link{summary.nmfae.inference}}
#' @export
print.summary.nmfae.inference <- function(x, digits = max(3L, getOption("digits") - 3L),
                                          max.coef = 20, ...) {
  print.summary.nmfae(x, digits = digits, max.coef = max.coef, ...)
}



#' @title Heatmap visualization of nmfae factor matrices
#' @description
#' \code{nmfae.heatmap} displays the three factor matrices \eqn{X_1}, \eqn{\Theta},
#' and \eqn{X_2} as side-by-side heatmaps. This provides an alternative to DOT graph
#' visualization, especially when \eqn{Y_2} has many variables (e.g., kernel matrix).
#'
#' @param x An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.
#' @param Y1.label Character vector of output variable names (rows of \eqn{X_1}).
#' @param X1.label Character vector of decoder basis labels (columns of \eqn{X_1}).
#' @param X2.label Character vector of encoder basis labels (rows of \eqn{X_2}).
#' @param Y2.label Character vector of input variable names (columns of \eqn{X_2}).
#' @param palette Color palette vector. Default is white-orange-red (64 colors).
#' @param ... Not used.
#'
#' @return Invisible \code{NULL}. Called for its side effect (plot).
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @seealso \code{\link{nmfae}}, \code{\link{plot.nmfae}}, \code{\link{nmfae.DOT}}
#' @examples
#' \donttest{
#' set.seed(1)
#' Y <- matrix(runif(20), nrow = 4)
#' res <- nmfae(Y, rank = 2)
#' nmfae.heatmap(res)
#' }
#' @export
nmfae.heatmap <- function(x,
                          Y1.label = NULL, X1.label = NULL,
                          X2.label = NULL, Y2.label = NULL,
                          palette = NULL, ...) {
  X1 <- x$X1
  C_mat <- x$C
  X2 <- x$X2
  P1 <- nrow(X1); Q <- ncol(X1)
  R <- nrow(X2); P2 <- ncol(X2)

  if (is.null(palette))
    palette <- grDevices::colorRampPalette(c("white", "orange", "red"))(64)

  # Default labels: use dimnames from the object, then fallback to prefix
  resolve_lab <- function(lab, dimnames_default, expected_len, prefix) {
    if (!is.null(lab) && length(lab) != expected_len) {
      warning(sprintf("Label length (%d) does not match dimension (%d). Labels ignored.",
                      length(lab), expected_len))
      lab <- NULL
    }
    if (is.null(lab)) lab <- dimnames_default
    if (is.null(lab)) lab <- paste0(prefix, 1:expected_len)
    lab
  }
  Y1.label <- resolve_lab(Y1.label, rownames(X1), P1, "Y1.")
  X1.label <- resolve_lab(X1.label, colnames(X1), Q,  "Q")
  X2.label <- resolve_lab(X2.label, rownames(X2), R,  "R")
  Y2.label <- resolve_lab(Y2.label, colnames(X2), P2, "Y2.")

  # Helper: plot a matrix as heatmap (row 1 at top)
  plot_mat <- function(mat, main, row.lab, col.lab, mar) {
    opar <- graphics::par(mar = mar)
    on.exit(graphics::par(opar))
    nr <- nrow(mat); nc <- ncol(mat)
    graphics::image(1:nc, 1:nr, t(mat[nr:1, , drop = FALSE]),
                    col = palette, axes = FALSE,
                    xlab = "", ylab = "", main = main)
    graphics::box()
    if (!is.null(col.lab)) {
      cex_col <- min(0.8, 20 / length(col.lab))
      graphics::axis(1, at = 1:nc, labels = col.lab, las = 2, cex.axis = cex_col)
    }
    if (!is.null(row.lab)) {
      cex_row <- min(0.8, 20 / length(row.lab))
      graphics::axis(2, at = 1:nr, labels = rev(row.lab), las = 1, cex.axis = cex_row)
    }
  }

  opar <- graphics::par(mfrow = c(1, 3))
  on.exit(graphics::par(opar))

  plot_mat(X1, sprintf("X1 (%d x %d)", P1, Q), Y1.label, X1.label,
           mar = c(5, 6, 3, 1))
  plot_mat(C_mat, sprintf("C (%d x %d)", Q, R), X1.label, X2.label,
           mar = c(5, 4, 3, 1))
  plot_mat(X2, sprintf("X2 (%d x %d)", R, P2), X2.label, Y2.label,
           mar = c(5, 4, 3, 1))

  invisible(NULL)
}

#' @title Predict method for nmfae objects
#' @description
#' \code{predict.nmfae} computes fitted or predicted values from a three-layer NMF model.
#' Without \code{newY2}, returns the in-sample fitted values \eqn{X_1 \Theta X_2 Y_2}.
#' With \code{newY2}, computes out-of-sample predictions \eqn{X_1 \Theta X_2 \cdot \mathrm{newY2}}.
#'
#' When \code{type = "class"}, each column is classified to the row with the
#' maximum predicted value (useful when \eqn{Y_1} is a one-hot class matrix
#' from \code{\link{nmfkc.class}}).
#'
#' If \code{Y1} (actual values) is provided, it is stored as an attribute so that
#' \code{plot.predict.nmfae} can produce an observed-vs-predicted scatter plot
#' (for \code{type = "response"}) or a confusion matrix heatmap
#' (for \code{type = "class"}).
#'
#' @param object An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.
#' @param newY2 Optional new input matrix (P2 x M) for prediction.
#'   If \code{NULL}, returns in-sample fitted values.
#' @param Y1 Optional actual output matrix for comparison plotting.
#' @param type Character. \code{"response"} (default) returns the predicted matrix.
#'   \code{"class"} returns a factor of predicted class labels (row with max value).
#' @param ... Not used.
#'
#' @return For \code{type = "response"}: a matrix of class \code{"predict.nmfae"}.
#'   For \code{type = "class"}: a factor of class \code{"predict.nmfae"} with
#'   predicted class labels. If \code{Y1} was provided, actual classes are stored
#'   in \code{attr(result, "actual")}.
#' @seealso \code{\link{nmfae}}, \code{\link{plot.predict.nmfae}},
#'   \code{\link{nmfkc.class}}
#' @examples
#' \donttest{
#' set.seed(1)
#' Y <- matrix(runif(20), nrow = 4)
#' res <- nmfae(Y, rank = 2)
#' pred <- predict(res)
#' }
#' @export
predict.nmfae <- function(object, newY2 = NULL, Y1 = NULL,
                          type = c("response", "class"), ...) {
  type <- match.arg(type)
  if (is.null(newY2)) pred_mat <- object$Y1hat
  else pred_mat <- object$X1 %*% object$C %*% object$X2 %*% as.matrix(newY2)

  if (type == "response") {
    result <- pred_mat
    if (!is.null(Y1)) attr(result, "Y1") <- Y1
    class(result) <- c("predict.nmfae", class(result))
  } else {
    # Class prediction: row with max value per column
    labels <- rownames(pred_mat)
    if (is.null(labels)) labels <- paste0("C", seq_len(nrow(pred_mat)))
    pred_class <- factor(labels[apply(pred_mat, 2, which.max)], levels = labels)
    result <- pred_class
    if (!is.null(Y1)) {
      Y1 <- as.matrix(Y1)
      act_labels <- rownames(Y1)
      if (is.null(act_labels)) act_labels <- labels
      actual <- factor(act_labels[apply(Y1, 2, which.max)], levels = act_labels)
      attr(result, "actual") <- actual
    }
    class(result) <- c("predict.nmfae", class(result))
    attr(result, "type") <- "class"
  }
  result
}

#' @title Plot method for predict.nmfae objects
#' @description
#' For \code{type = "response"}: if actual values \eqn{Y_1} were stored,
#' displays an observed-vs-predicted scatter plot with \eqn{R^2} in the title.
#' Otherwise, displays the predicted matrix as a heatmap.
#'
#' For \code{type = "class"}: if actual classes were stored, displays a
#' confusion matrix heatmap with accuracy (ACC) in the title.
#'
#' @param x An object of class \code{"predict.nmfae"} returned by \code{\link{predict.nmfae}}.
#' @param ... Additional graphical parameters passed to \code{plot} or \code{image}.
#'
#' @return Invisible \code{NULL}. Called for its side effect (plot).
#' @seealso \code{\link{predict.nmfae}}
#' @examples
#' \donttest{
#' set.seed(1)
#' Y <- matrix(runif(20), nrow = 4)
#' res <- nmfae(Y, rank = 2)
#' pred <- predict(res)
#' plot(pred)
#' }
#' @export
plot.predict.nmfae <- function(x, ...) {
  extra_args <- list(...)
  pred_type <- attr(x, "type")

  if (!is.null(pred_type) && pred_type == "class") {
    # --- Confusion matrix heatmap ---
    actual <- attr(x, "actual")
    if (is.null(actual)) {
      message("No actual classes stored. Use Y1 argument in predict().")
      return(invisible(NULL))
    }
    pred_class <- factor(x, levels = levels(actual))
    cm <- table(Actual = actual, Predicted = pred_class)
    acc <- sum(diag(cm)) / sum(cm)

    # Heatmap of confusion matrix
    mat <- as.matrix(cm)
    nr <- nrow(mat); nc <- ncol(mat)
    pal <- grDevices::colorRampPalette(c("white", "orange", "red"))(64)
    opar <- graphics::par(mar = c(5, 5, 3, 1))
    on.exit(graphics::par(opar))
    graphics::image(1:nc, 1:nr, t(mat[nr:1, , drop = FALSE]),
                    col = pal, axes = FALSE, xlab = "Predicted", ylab = "Actual")
    graphics::box()
    labs <- rownames(mat)
    graphics::axis(1, at = 1:nc, labels = colnames(mat), las = 2, cex.axis = 0.8)
    graphics::axis(2, at = 1:nr, labels = rev(labs), las = 1, cex.axis = 0.8)
    # Cell labels
    for (i in 1:nr) {
      for (j in 1:nc) {
        val <- mat[i, j]
        if (val > 0) {
          yi <- nr - i + 1
          graphics::text(j, yi, val, cex = 0.7,
                         col = ifelse(val > stats::median(mat), "white", "black"))
        }
      }
    }
    if (is.null(extra_args$main))
      graphics::title(main = sprintf("ACC = %.1f%%", acc * 100))
    else
      graphics::title(main = extra_args$main)

  } else {
    # --- Response type ---
    Y1 <- attr(x, "Y1")

    if (!is.null(Y1)) {
      # Scatter plot: observed vs predicted
      obs <- as.numeric(as.matrix(Y1))
      pred <- as.numeric(unclass(x))
      r2 <- stats::cor(obs, pred)^2
      args <- list(x = obs, y = pred)
      if (is.null(extra_args$main))
        args$main <- sprintf("R2 = %.3f", r2)
      if (is.null(extra_args$xlab)) args$xlab <- "Observed"
      if (is.null(extra_args$ylab)) args$ylab <- "Predicted"
      if (is.null(extra_args$pch))  args$pch <- 16
      if (is.null(extra_args$cex))  args$cex <- 0.8
      if (is.null(extra_args$col))  args$col <- grDevices::rgb(0, 0, 0, 0.4)
      all_args <- c(args, extra_args)
      do.call("plot", all_args)
      graphics::abline(0, 1, col = "red")
    } else {
      # Heatmap of predicted matrix
      mat <- unclass(x)
      nr <- nrow(mat); nc <- ncol(mat)
      pal <- grDevices::colorRampPalette(c("white", "orange", "red"))(64)
      args <- list(x = 1:nc, y = 1:nr, z = t(mat[nr:1, , drop = FALSE]),
                   col = pal, axes = FALSE, xlab = "Sample", ylab = "Variable")
      if (is.null(extra_args$main)) args$main <- "Predicted Y1"
      all_args <- c(args, extra_args)
      do.call("image", all_args)
      graphics::box()
    }
  }
  invisible(NULL)
}

#' @title Element-wise Cross-Validation for nmfae (Wold's CV)
#' @description
#' \code{nmfae.ecv} performs k-fold element-wise cross-validation by randomly
#' holding out individual elements of \eqn{Y_1}, assigning them a weight of 0
#' via \code{Y1.weights}, and evaluating the reconstruction error on those
#' held-out elements.
#'
#' This method (also known as Wold's CV) is suitable for determining the optimal
#' rank pair \eqn{(Q, R)} in three-layer NMF. Both \code{rank} and \code{rank.encoder} accept
#' vector inputs. When \code{rank.encoder = NULL} (default), \code{rank.encoder} is set equal to \code{rank}
#' and pairs are evaluated element-wise (i.e., \eqn{(Q_1, R_1), (Q_2, R_2), \dots}).
#' When \code{rank.encoder} is explicitly specified, all combinations of \code{rank} and \code{rank.encoder}
#' are evaluated via \code{expand.grid}.
#'
#' @param Y1 Output matrix \eqn{Y_1} (P1 x N).
#' @param Y2 Input matrix \eqn{Y_2} (P2 x N). Default is \code{Y1}.
#' @param rank Integer vector of decoder ranks to evaluate. Default is \code{1:2}.
#' @param rank.encoder Integer vector of encoder ranks to evaluate. Default is \code{NULL},
#'   which sets \code{rank.encoder = rank} and evaluates element-wise pairs.
#'   When explicitly specified, all combinations with \code{rank} are evaluated.
#' @param ... Additional arguments passed to \code{\link{nmfae}} (e.g., \code{epsilon}, \code{maxit}).
#'   Also accepts: \code{nfolds} (number of folds, default 5; \code{div} also accepted),
#'   \code{seed} (integer seed, default 123).
#'   For backward compatibility, \code{Q} and \code{R} are accepted as aliases for
#'   \code{rank} and \code{rank.encoder}.
#'
#' @return A list with components:
#' \item{objfunc}{Named numeric vector of mean MSE for each (Q, R) pair.}
#' \item{sigma}{Named numeric vector of RMSE (square root of MSE) for each pair.}
#' \item{objfunc.fold}{Named list of per-fold MSE vectors for each pair.}
#' \item{folds}{List of length \code{div} containing the held-out element indices for each fold.}
#' \item{QR}{Data frame with columns \code{Q} and \code{R} listing the evaluated pairs.}
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @seealso \code{\link{nmfae}}, \code{\link{nmfkc.ecv}}
#' @export
#' @examples
#' Y <- t(iris[1:30, 1:4])
#' # Default: rank.encoder=NULL -> paired rank=rank.encoder
#' res <- nmfae.ecv(Y, rank = 1:3, nfolds = 3, maxit = 500)
#' res$sigma
#' # Explicit rank.encoder: full grid
#' res2 <- nmfae.ecv(Y, rank = 1:3, rank.encoder = 1:3, nfolds = 3, maxit = 500)
#' res2$sigma
#'
nmfae.ecv <- function(Y1, Y2 = Y1, rank = 1:2, rank.encoder = NULL, ...) {
  extra_ecv <- list(...)
  if (!is.null(extra_ecv$Q)) rank <- extra_ecv$Q
  if (!is.null(extra_ecv$R)) rank.encoder <- extra_ecv$R
  nfolds <- if (!is.null(extra_ecv$nfolds)) extra_ecv$nfolds else if (!is.null(extra_ecv$div)) extra_ecv$div else 5
  seed   <- if (!is.null(extra_ecv$seed))   extra_ecv$seed   else 123
  Q <- rank; R <- rank.encoder
  div <- nfolds

  Y1 <- as.matrix(Y1); Y2 <- as.matrix(Y2)
  P1 <- nrow(Y1); N <- ncol(Y1)

  # R=NULL -> paired with Q; R specified -> full grid
  if (is.null(R)) {
    QR <- data.frame(Q = Q, R = Q)
  } else {
    QR <- expand.grid(Q = Q, R = R)
  }
  num_pairs <- nrow(QR)

  # Create folds (element-wise on Y1; shared helper)
  folds <- .ecv.make.folds(Y1, div, seed)

  # Prepare result storage
  pair_labels <- sprintf("Q=%d,R=%d", QR$Q, QR$R)
  has_na <- any(is.na(Y1))

  message(sprintf("Element-wise CV: %d (Q,R) pairs, %d-fold, %d tasks...",
                  num_pairs, div, num_pairs * div))

  extra_args <- list(...)

  # Model-specific worker: mask fold k, refit at pair i, held-out loss
  run_one <- function(i, k) {
    test_idx <- folds[[k]]
    weights_train <- matrix(1, nrow = P1, ncol = N)
    if (has_na) weights_train[is.na(Y1)] <- 0
    weights_train[test_idx] <- 0
    fit <- suppressMessages(
      do.call(nmfae, c(list(Y1 = Y1, Y2 = Y2, Q = QR$Q[i], R = QR$R[i],
                            Y1.weights = weights_train), extra_args))
    )
    mean((Y1[test_idx] - fit$Y1hat[test_idx])^2)
  }

  cv <- .ecv.run(pair_labels, div, run_one,
                 progress = function(i, o, s)
                   message(sprintf("  Q=%d, R=%d: MSE=%.6f, sigma=%.4f",
                                   QR$Q[i], QR$R[i], o, s)))

  result <- list(objfunc = cv$objfunc,
                 sigma = cv$sigma,
                 objfunc.fold = cv$objfunc.fold,
                 folds = folds,
                 QR = QR,
                 paired = is.null(R))
  class(result) <- "nmfae.ecv"
  return(result)
}


#' @title Rank selection for nmfae (paired rank, concise diagnostics)
#' @description
#' Fits \code{\link{nmfae}} with a \strong{paired} decoder/encoder rank
#' (\eqn{Q = R}) across a range of ranks and reports \code{r.squared},
#' the effective rank (of the latent encoding \eqn{H}), and the
#' element-wise CV error \code{sigma.ecv}, with the same concise plot as
#' \code{\link{nmfkc.rank}}.  For a full \eqn{(Q, R)} grid use
#' \code{\link{nmfae.ecv}} with \code{rank.encoder} and its heatmap.
#' @param Y1 Endogenous matrix (\eqn{P_1 \times N}).
#' @param Y2 Exogenous matrix; defaults to \code{Y1} (autoencoder).
#' @param rank Integer vector of (paired) ranks to evaluate.
#' @param detail \code{"full"} (default) also runs element-wise CV
#'   (\code{sigma.ecv}); \code{"fast"} skips it (plots r.squared and
#'   eff.rank only, and recommends the R-squared elbow).
#' @param plot Logical; draw the diagnostics plot (default \code{TRUE}).
#' @param ... Passed on to \code{\link{nmfae}} and \code{\link{nmfae.ecv}}
#'   (e.g.\ \code{maxit}, \code{nfolds}, \code{seed}).
#' @return A list with \code{rank.best} and \code{criteria}
#'   (\code{rank}, \code{effective.rank}, \code{effective.rank.ratio},
#'   \code{r.squared}, \code{sigma.ecv}).
#' @seealso \code{\link{nmfae}}, \code{\link{nmfae.ecv}},
#'   \code{\link{nmfkc.rank}}
#' @references
#' Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
#' effective dimensionality.  \emph{Proc. EUSIPCO}, 606--610.
#' (\code{effective.rank})
#' Wold, S. (1978).  Cross-validatory estimation of the number of
#' components in factor and principal components models.
#' \emph{Technometrics}, 20(4), 397--405. (\code{sigma.ecv})
#' @export
nmfae.rank <- function(Y1, Y2 = Y1, rank = 1:5, detail = c("full", "fast"),
                       plot = TRUE, ...) {
  detail <- match.arg(detail)
  Y1 <- as.matrix(Y1); Y2 <- as.matrix(Y2)
  rs <- numeric(length(rank)); er <- numeric(length(rank))
  for (i in seq_along(rank)) {
    f <- suppressMessages(nmfae(Y1, Y2, rank = rank[i],
                                rank.encoder = rank[i],
                                print.trace = FALSE, ...))
    rs[i] <- f$r.squared
    er[i] <- .effective.rank(f$H)
  }
  ecv <- if (detail == "full")
    suppressMessages(nmfae.ecv(Y1, Y2, rank = rank, ...))$sigma
    else rep(NA_real_, length(rank))
  criteria <- data.frame(rank = rank, effective.rank = er,
                         effective.rank.ratio = er / rank,
                         r.squared = rs, sigma.ecv = as.numeric(ecv))
  .rank.finish(criteria, plot = plot, main = "nmfae rank selection (paired Q=R)")
}

#' @title Plot method for nmfae.ecv objects
#' @description
#' Visualizes element-wise cross-validation results.
#' When \code{rank.encoder} was \code{NULL} (paired), a line plot of sigma vs rank is drawn.
#' When \code{rank.encoder} was explicitly specified (grid), a heatmap of sigma over the (rank, rank.encoder) grid is drawn.
#'
#' @param x An object of class \code{"nmfae.ecv"} returned by \code{\link{nmfae.ecv}}.
#' @param ... Additional graphical parameters (currently unused).
#'
#' @return Invisible \code{NULL}. Called for its side effect of producing a plot.
#' @seealso \code{\link{nmfae.ecv}}
#' @export
plot.nmfae.ecv <- function(x, ...) {
  Qs <- sort(unique(x$QR$Q))
  Rs <- sort(unique(x$QR$R))
  use_line <- x$paired || length(Qs) == 1 || length(Rs) == 1

  if (use_line) {
    # --- Line plot ---
    if (x$paired) {
      xvals <- x$QR$Q; xlab_str <- "Q = R"
    } else if (length(Qs) == 1) {
      xvals <- x$QR$R; xlab_str <- paste0("R (Q=", Qs, ")")
    } else {
      xvals <- x$QR$Q; xlab_str <- paste0("Q (R=", Rs, ")")
    }
    plot(xvals, x$sigma, type = "b", pch = 16,
         xlab = xlab_str, ylab = "sigma",
         main = "Element-wise CV",
         xaxt = "n")
    graphics::axis(1, at = xvals)
    best <- which.min(x$sigma)
    graphics::points(xvals[best], x$sigma[best], pch = 16, col = "red", cex = 1.5)
  } else {
    # --- Heatmap: sigma over (Q, R) grid ---
    Qs <- sort(unique(x$QR$Q))
    Rs <- sort(unique(x$QR$R))
    sigma_mat <- matrix(NA, length(Qs), length(Rs))
    for (i in 1:nrow(x$QR)) {
      qi <- which(Qs == x$QR$Q[i])
      ri <- which(Rs == x$QR$R[i])
      sigma_mat[qi, ri] <- x$sigma[i]
    }

    ncolors <- 64
    col_palette <- grDevices::hcl.colors(ncolors, "YlOrRd", rev = TRUE)

    opar <- graphics::par(mar = c(5, 5, 3, 8))
    on.exit(graphics::par(opar))

    graphics::image(Qs, Rs, sigma_mat, col = col_palette,
          xlab = "Q (decoder rank)", ylab = "R (encoder rank)",
          main = "Element-wise CV", axes = FALSE)
    graphics::axis(1, at = Qs)
    graphics::axis(2, at = Rs)
    graphics::box()

    # Cell values
    for (qi in seq_along(Qs)) {
      for (ri in seq_along(Rs)) {
        val <- sigma_mat[qi, ri]
        txt_col <- "black"
        graphics::text(Qs[qi], Rs[ri], sprintf("%.2f", val), cex = 0.65, col = txt_col)
      }
    }

    # Color bar — use cell step as unit for positioning
    old_par <- graphics::par(xpd = TRUE)
    on.exit(graphics::par(old_par), add = TRUE)
    q_step <- if (length(Qs) > 1) diff(Qs)[1] else 1
    r_step <- if (length(Rs) > 1) diff(Rs)[1] else 1
    bar_x <- max(Qs) + q_step * 0.7
    bar_w <- q_step * 0.3
    bar_y <- seq(min(Rs), max(Rs), length.out = ncolors + 1)
    for (i in 1:ncolors) {
      graphics::rect(bar_x, bar_y[i], bar_x + bar_w, bar_y[i + 1],
           col = col_palette[i], border = NA)
    }
    graphics::rect(bar_x, min(Rs), bar_x + bar_w, max(Rs))
    sigma_range <- range(sigma_mat)
    ticks <- pretty(sigma_range, 5)
    ticks <- ticks[ticks >= sigma_range[1] & ticks <= sigma_range[2]]
    tick_y <- min(Rs) + (ticks - sigma_range[1]) / diff(sigma_range) * diff(range(Rs))
    graphics::text(bar_x + bar_w + q_step * 0.15, tick_y,
         sprintf("%.1f", ticks), cex = 0.7, adj = 0)
    graphics::text(bar_x + bar_w / 2, max(Rs) + r_step * 0.5,
         "sigma", cex = 0.8)
  }
}

#' @title Sample-wise k-fold Cross-Validation for nmfae
#' @description
#' \code{nmfae.cv} performs k-fold cross-validation by splitting columns (samples)
#' of \eqn{Y_1} and \eqn{Y_2} into \code{div} folds. For each fold, the model
#' \eqn{Y_1 \approx X_1 \Theta X_2 Y_2} is fitted on the training samples and
#' predictive performance is evaluated on the held-out samples.
#'
#' When \code{Y2} is a kernel matrix created by \code{\link{nmfkc.kernel}}
#' (detected via attributes), the symmetric kernel splitting convention is used:
#' \code{Y2[train, train]} for training and \code{Y2[train, test]} for prediction.
#'
#' @param Y1 Output matrix \eqn{Y_1} (P1 x N). Non-negative.
#' @param Y2 Input matrix \eqn{Y_2} (P2 x N), or a kernel matrix (N x N).
#'   Default is \code{Y1} (autoencoder).
#' @param rank Integer. Rank of the decoder basis. Default is 2.
#' @param rank.encoder Integer. Rank of the encoder basis. Default is \code{rank}.
#' @param ... Additional arguments passed to \code{\link{nmfae}}
#'   (e.g., \code{epsilon}, \code{maxit}, \code{Y1.weights}).
#'   Also accepts: \code{nfolds} (number of folds, default 5; \code{div} also accepted),
#'   \code{seed} (integer seed, default 123), \code{shuffle} (logical, default \code{TRUE}).
#'   For backward compatibility, \code{Q}, \code{R} are accepted as aliases for
#'   \code{rank}, \code{rank.encoder}.
#'
#' @return A list with components:
#' \item{objfunc}{Mean squared error per valid element over all folds.}
#' \item{sigma}{Residual standard error (RMSE), same scale as \eqn{Y_1}.}
#' \item{objfunc.block}{Per-fold squared error totals.}
#' \item{block}{Integer vector of fold assignments (1, ..., \code{div}) for each column.}
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @seealso \code{\link{nmfae}}, \code{\link{nmfae.ecv}}, \code{\link{nmfae.kernel.beta.cv}},
#'   \code{\link{nmfkc.cv}}
#' @export
#' @examples
#' Y <- t(iris[1:30, 1:4])
#' res <- nmfae.cv(Y, rank = 2, rank.encoder = 2, nfolds = 5, maxit = 500)
#' res$sigma
#'
nmfae.cv <- function(Y1, Y2 = Y1, rank = 2, rank.encoder = rank, ...) {
  extra_cv <- list(...)
  if (!is.null(extra_cv$Q)) rank <- extra_cv$Q
  if (!is.null(extra_cv$R)) rank.encoder <- extra_cv$R
  nfolds  <- if (!is.null(extra_cv$nfolds))  extra_cv$nfolds  else if (!is.null(extra_cv$div)) extra_cv$div else 5
  seed    <- if (!is.null(extra_cv$seed))    extra_cv$seed    else 123
  shuffle <- if (!is.null(extra_cv$shuffle)) extra_cv$shuffle else TRUE
  Q <- rank; R <- rank.encoder
  div <- nfolds

  extra_args <- list(...)

  Y1 <- as.matrix(Y1); storage.mode(Y1) <- "double"
  Y2 <- as.matrix(Y2); storage.mode(Y2) <- "double"
  P1 <- nrow(Y1); N <- ncol(Y1)

  # --- Y1.weights preparation (cf. nmfkc.cv) ---
  Y1.weights <- extra_args$Y1.weights
  if (!is.null(Y1.weights) && is.vector(Y1.weights)) {
    if (length(Y1.weights) == N) {
      Y1.weights <- matrix(Y1.weights, nrow = P1, ncol = N, byrow = TRUE)
    } else if (length(Y1.weights) == 1) {
      Y1.weights <- matrix(Y1.weights, nrow = P1, ncol = N)
    } else {
      stop("Length of Y1.weights vector must match ncol(Y1) (or be 1).")
    }
  }
  if (is.null(Y1.weights)) {
    if (any(is.na(Y1))) {
      Y1.weights <- matrix(1, nrow = P1, ncol = N)
      Y1.weights[is.na(Y1)] <- 0
      Y1[is.na(Y1)] <- 0
    } else {
      Y1.weights <- matrix(1, nrow = P1, ncol = N)
    }
  } else {
    if (!is.matrix(Y1.weights)) Y1.weights <- as.matrix(Y1.weights)
    Y1.weights[is.na(Y1.weights)] <- 0
    Y1[is.na(Y1)] <- 0
  }

  # --- Detect kernel matrix (cf. nmfkc.cv) ---
  A.function <- attr(Y2, "function.name")
  is_kernel_matrix <- !is.null(A.function) && A.function == "nmfkc.kernel"
  is_symmetric <- isSymmetric(Y2, tol = .Machine$double.eps)

  # --- Create folds (sample-wise, same logic as nmfkc.cv) ---
  remainder <- N %% div
  division  <- N %/% div
  block <- integer(N)

  if (shuffle) {
    set.seed(seed)
    perm_index <- sample(1:N, N, replace = FALSE)
  } else {
    perm_index <- 1:N
  }

  processed <- 0
  for (i in 1:(div - 1)) {
    chunk <- division + ifelse(i <= remainder, 1, 0)
    idx <- perm_index[(processed + 1):(processed + chunk)]
    block[idx] <- i
    processed <- processed + chunk
  }
  block[perm_index[(processed + 1):N]] <- div

  # --- Prepare nmfae pass-through args ---
  nmfae_extra <- extra_args
  nmfae_extra$Y1.weights <- NULL
  nmfae_extra$print.trace <- NULL
  nmfae_extra$seed <- NULL

  objfunc.block <- numeric(div)
  total_valid <- 0

  for (j in 1:div) {
    train <- (block != j)
    test  <- (block == j)

    # Y1 split
    Y1_train <- Y1[, train, drop = FALSE]
    Y1_test  <- Y1[, test,  drop = FALSE]
    W_train  <- Y1.weights[, train, drop = FALSE]
    W_test   <- Y1.weights[, test,  drop = FALSE]

    # Y2 split
    if (is_symmetric && is_kernel_matrix) {
      Y2_train <- Y2[train, train, drop = FALSE]
      Y2_test  <- Y2[train, test,  drop = FALSE]
    } else {
      Y2_train <- Y2[, train, drop = FALSE]
      Y2_test  <- Y2[, test,  drop = FALSE]
    }

    # Fit on training set
    nmfae_args <- c(
      list(Y1 = Y1_train, Y2 = Y2_train, Q = Q, R = R,
           Y1.weights = W_train, print.trace = FALSE),
      nmfae_extra
    )
    res_j <- suppressMessages(do.call("nmfae", nmfae_args))

    # Predict on test set
    Y1hat_test <- res_j$X1 %*% res_j$C %*% res_j$X2 %*% Y2_test

    # Evaluate weighted error (lm-style: sum(W * resid^2))
    objfunc.block[j] <- sum(W_test * (Y1_test - Y1hat_test)^2)
    total_valid <- total_valid + sum(W_test > 0)
  }

  objfunc <- sum(objfunc.block) / max(total_valid, 1)
  sigma <- sqrt(objfunc)

  result <- list(objfunc = objfunc, sigma = sigma,
                 objfunc.block = objfunc.block, block = block)
  class(result) <- "nmfae.cv"
  result
}

#' @title Plot method for nmfae.cv objects
#' @description
#' Displays a bar chart of per-fold cross-validation errors from
#' \code{\link{nmfae.cv}}. The overall RMSE (sigma) is shown in the title.
#'
#' @param x An object of class \code{"nmfae.cv"} returned by \code{\link{nmfae.cv}}.
#' @param ... Additional graphical parameters passed to \code{barplot}.
#'
#' @return Invisible \code{NULL}. Called for its side effect (plot).
#' @seealso \code{\link{nmfae.cv}}
#' @export
plot.nmfae.cv <- function(x, ...) {
  extra_args <- list(...)
  args <- list(height = x$objfunc.block)
  if (is.null(extra_args$main))
    args$main <- sprintf("sigma = %.4f", x$sigma)
  if (is.null(extra_args$xlab)) args$xlab <- "Fold"
  if (is.null(extra_args$ylab)) args$ylab <- "Error"
  if (is.null(extra_args$col))  args$col <- "steelblue"
  if (is.null(extra_args$names.arg))
    args$names.arg <- seq_along(x$objfunc.block)
  all_args <- c(args, extra_args)
  do.call("barplot", all_args)
  invisible(NULL)
}

#' @title Optimize kernel beta for nmfae by cross-validation
#' @description
#' \code{nmfae.kernel.beta.cv} selects the optimal \code{beta} parameter of the
#' kernel function by evaluating \code{\link{nmfae.cv}} for each candidate value.
#' The kernel matrix \eqn{A = K(U, V; \beta)} replaces \eqn{Y_2} in the three-layer
#' NMF model.
#'
#' When \code{beta = NULL}, candidate values are automatically generated via
#' \code{\link{nmfkc.kernel.beta.nearest.med}}.
#'
#' @param Y1 Output matrix \eqn{Y_1} (P1 x N). Non-negative.
#' @param rank Integer. Rank of the decoder basis. Default is 2.
#' @param rank.encoder Integer. Rank of the encoder basis. Default is \code{rank}.
#' @param U Covariate matrix \eqn{U} (K x M). Rows are features, columns are samples
#'   (or knot points for non-symmetric kernels).
#' @param V Covariate matrix \eqn{V} (K x N). If \code{NULL} (default), \code{V = U}
#'   and a symmetric kernel is used.
#' @param beta Numeric vector of candidate beta values. If \code{NULL}, automatically
#'   determined via \code{\link{nmfkc.kernel.beta.nearest.med}}.
#' @param plot Logical. If \code{TRUE} (default), plots the objective function curve.
#' @param ... Additional arguments. Kernel-specific args (\code{kernel}, \code{degree})
#'   are passed to \code{\link{nmfkc.kernel}}; all others
#'   (\code{div}, \code{seed}, \code{shuffle}, \code{epsilon}, \code{maxit}, etc.)
#'   are passed to \code{\link{nmfae.cv}}.
#'   For backward compatibility, \code{Q} and \code{R} are accepted as aliases for
#'   \code{rank} and \code{rank.encoder}.
#'
#' @return A list with components:
#' \item{beta}{The beta value that minimizes the cross-validation objective.}
#' \item{objfunc}{Named numeric vector of objective function values for each candidate beta.}
#'
#' @seealso \code{\link{nmfae.cv}}, \code{\link{nmfkc.kernel}},
#'   \code{\link{nmfkc.kernel.beta.cv}}
#' @export
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' U <- matrix(cars$speed, nrow = 1)
#' res <- nmfae.kernel.beta.cv(Y, rank = 1, rank.encoder = 1, U = U,
#'                              beta = c(0.01, 0.02, 0.05), nfolds = 5)
#' res$beta
#'
nmfae.kernel.beta.cv <- function(Y1, rank = 2, rank.encoder = rank, U, V = NULL,
                                  beta = NULL, plot = TRUE, ...) {

  extra_args <- list(...)
  # backward compat: Q -> rank, R -> rank.encoder
  if (!is.null(extra_args$Q)) rank <- extra_args$Q
  if (!is.null(extra_args$R)) rank.encoder <- extra_args$R
  extra_args <- extra_args[!names(extra_args) %in% c("Q", "R")]

  # Separate kernel-specific args from cv/nmfae args
  kernel_only <- c("kernel", "degree")
  kernel_args <- extra_args[names(extra_args) %in% kernel_only]
  cv_args     <- extra_args[!names(extra_args) %in% kernel_only]

  # Auto-generate beta candidates if not provided
  if (is.null(beta)) {
    if (is.null(V)) V <- U
    result.beta <- nmfkc.kernel.beta.nearest.med(V)
    beta <- result.beta$beta_candidates
    if (is.null(beta) || length(beta) == 0)
      stop("Failed to determine beta candidates from nearest-neighbor median.")
  }

  objfuncs <- numeric(length(beta))
  for (i in seq_along(beta)) {
    start.time <- Sys.time()
    message(paste0("beta=", beta[i], "..."), appendLF = FALSE)

    kernel_call <- c(list(U = U, V = V, beta = beta[i]), kernel_args)
    A <- do.call("nmfkc.kernel", kernel_call)

    cv_call <- c(list(Y1 = Y1, Y2 = A, rank = rank, rank.encoder = rank.encoder), cv_args)
    result <- do.call("nmfae.cv", cv_call)

    objfuncs[i] <- result$objfunc

    end.time <- Sys.time()
    diff.time <- difftime(end.time, start.time, units = "sec")
    diff.time.st <- ifelse(diff.time <= 180,
                           paste0(round(diff.time, 1), "sec"),
                           paste0(round(diff.time / 60, 1), "min"))
    message(diff.time.st)
  }

  i0 <- which.min(objfuncs)
  beta.best <- beta[i0]

  names(objfuncs) <- beta
  result <- list(beta = beta.best, objfunc = objfuncs)
  class(result) <- "nmfae.kernel.beta.cv"
  if (plot) plot(result)
  result
}

#' @title Plot method for nmfae.kernel.beta.cv objects
#' @description
#' Displays the cross-validation objective function across candidate
#' \code{beta} values (log scale). The optimal beta is highlighted in red.
#'
#' @param x An object of class \code{"nmfae.kernel.beta.cv"} returned by
#'   \code{\link{nmfae.kernel.beta.cv}}.
#' @param ... Additional graphical parameters passed to \code{plot}.
#'
#' @return Invisible \code{NULL}. Called for its side effect (plot).
#' @seealso \code{\link{nmfae.kernel.beta.cv}}
#' @export
plot.nmfae.kernel.beta.cv <- function(x, ...) {
  beta <- as.numeric(names(x$objfunc))
  objfuncs <- x$objfunc
  i0 <- which.min(objfuncs)

  extra_args <- list(...)
  args <- list(x = beta, y = objfuncs, type = "l", col = 2, log = "x")
  if (is.null(extra_args$xlab)) args$xlab <- "beta"
  if (is.null(extra_args$ylab)) args$ylab <- "objfunc"
  if (is.null(extra_args$main))
    args$main <- sprintf("Best beta = %g", x$beta)
  all_args <- c(args, extra_args)
  do.call("plot", all_args)
  graphics::points(beta[i0], objfuncs[i0], cex = 3, col = 2)
  graphics::text(beta, objfuncs,
                 format(beta, scientific = TRUE, digits = 5))
  invisible(NULL)
}

#' @title DOT graph visualization for nmfae objects
#' @description
#' \code{nmfae.DOT} generates a DOT language string for visualizing the structure
#' of a three-layer NMF model. Two graph types are supported:
#' \code{"XCX"} shows encoder factors, \eqn{\Theta}, and decoder factors;
#' \code{"YXCXY"} shows the full structure from \eqn{Y_2} through \eqn{X_2}, \eqn{\Theta},
#' \eqn{X_1} to \eqn{Y_1}.
#'
#' Edge widths are proportional to matrix element values, and edges below
#' \code{threshold} are omitted for clarity.
#'
#' @param result An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.
#' @param type Character. Graph type: \code{"XCX"} (default) or \code{"YXCXY"}.
#' @param threshold Numeric. Edges with values below this are omitted. Default is 0.01.
#' @param sig.level Numeric or \code{NULL}. Significance level for filtering C edges
#'   when inference results are available. Only edges with p-value below \code{sig.level}
#'   are shown, with significance stars. Set to \code{NULL} to disable. Default is 0.1.
#' @param rankdir Character. Graph direction for DOT layout. Default is \code{"LR"} (left to right).
#' @param fill Logical. If \code{TRUE}, nodes are filled with color. Default is \code{TRUE}.
#' @param weight_scale Numeric. Base scale factor for edge widths. Default is 5.
#' @param weight_scale_x1 Numeric. Scale factor for \eqn{X_1} edges.
#' @param weight_scale_theta Numeric. Scale factor for \eqn{\Theta} edges.
#' @param weight_scale_x2 Numeric. Scale factor for \eqn{X_2} edges.
#' @param Y1.label Character vector of output variable labels.
#' @param X1.label Character vector of decoder basis labels.
#' @param X2.label Character vector of encoder basis labels.
#' @param Y2.label Character vector of input variable labels.
#' @param Y1.title Character. Title for output node group. Default is \code{"Output (Y1)"}.
#' @param X1.title Character. Title for decoder node group. Default is \code{"Decoder (X1)"}.
#' @param X2.title Character. Title for encoder node group. Default is \code{"Encoder (X2)"}.
#' @param Y2.title Character. Title for input node group. Default is \code{"Input (Y2)"}.
#' @param hide.isolated Logical. If \code{TRUE} (default), Y1 and Y2 nodes that have no
#'   edges at or above \code{threshold} are excluded from the graph. Only
#'   applies when \code{type = "YXCXY"}.
#'
#' @return A character string containing the DOT graph specification.
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @seealso \code{\link{nmfae}}
#' @examples
#' \donttest{
#' set.seed(1)
#' Y <- matrix(runif(20), nrow = 4)
#' res <- nmfae(Y, rank = 2)
#' dot <- nmfae.DOT(res)
#' }
#' @export
nmfae.DOT <- function(result,
                      type = c("XCX", "YXCXY"),
                      threshold = 0.01,
                      sig.level = 0.1,
                      rankdir = "LR",
                      fill = TRUE,
                      weight_scale = 5,
                      weight_scale_x1 = weight_scale,
                      weight_scale_theta = weight_scale,
                      weight_scale_x2 = weight_scale,
                      Y1.label = NULL, X1.label = NULL,
                      X2.label = NULL, Y2.label = NULL,
                      Y1.title = "Output (Y1)",
                      X1.title = "Decoder (X1)",
                      X2.title = "Encoder (X2)",
                      Y2.title = "Input (Y2)",
                      hide.isolated = TRUE) {

  type <- match.arg(type)

  X1 <- result$X1        # P1 x Q
  C_mat <- result$C   # Q x R
  X2 <- result$X2        # R x P2

  P1 <- nrow(X1); Q <- ncol(X1)
  R <- nrow(X2); P2 <- ncol(X2)

  # --- Labels ---
  Y1_labels <- if (!is.null(Y1.label)) Y1.label else rownames(X1)
  if (is.null(Y1_labels)) Y1_labels <- paste0("Y1_", seq_len(P1))
  X1_labels <- if (!is.null(X1.label)) X1.label else colnames(X1)
  if (is.null(X1_labels)) X1_labels <- paste0("D", seq_len(Q))
  X2_labels <- if (!is.null(X2.label)) X2.label else rownames(X2)
  if (is.null(X2_labels)) X2_labels <- paste0("E", seq_len(R))
  Y2_labels <- if (!is.null(Y2.label)) Y2.label else colnames(X2)
  if (is.null(Y2_labels)) Y2_labels <- paste0("Y2_", seq_len(P2))

  # --- Node IDs ---
  sanitize <- function(s) gsub("[^[:alnum:]_.]", "_", s, perl = TRUE)
  Y1_ids <- sanitize(paste0("Y1_", seq_len(P1)))
  X1_ids <- sanitize(paste0("X1_", seq_len(Q)))
  X2_ids <- sanitize(paste0("X2_", seq_len(R)))
  Y2_ids <- sanitize(paste0("Y2_", seq_len(P2)))

  # --- Filter isolated Y1/Y2 nodes (hide.isolated) ---
  idx_Y1 <- seq_len(P1)
  idx_Y2 <- seq_len(P2)
  if (isTRUE(hide.isolated) && type == "YXCXY") {
    # Y1 is connected if any X1[i, ] >= threshold
    used_Y1 <- apply(X1, 1L, function(row) any(row >= threshold, na.rm = TRUE))
    idx_Y1 <- which(used_Y1)
    # Y2 is connected if any X2[, j] >= threshold
    used_Y2 <- apply(X2, 2L, function(col) any(col >= threshold, na.rm = TRUE))
    idx_Y2 <- which(used_Y2)
  }

  # --- Helpers ---
  pw <- function(value, max_value, ws) {
    if (!is.finite(max_value) || max_value <= 0) return(0.5)
    if (!is.finite(value) || value <= 0) return(0.5)
    max(0.5, value * ws / max_value)
  }
  dg <- max(0L, nchar(sub("^0\\.", "", format(threshold, scientific = FALSE, trim = TRUE))))
  fmtc <- function(v) sprintf(paste0("%.", dg, "f"), v)

  # --- Cluster builder ---
  make_cluster <- function(id, title, ids, labels, shape, color) {
    style_str <- if (fill) {
      sprintf('    node [shape=%s, style="filled,rounded", fillcolor="%s", color=black, penwidth=1.5];\n', shape, color)
    } else {
      sprintf('    node [shape=%s, style="rounded", color=black, penwidth=1.5];\n', shape)
    }
    s <- paste0('  subgraph cluster_', id, ' {label="', title,
                '" style="rounded" color="black" penwidth=1.0;\n', style_str)
    for (i in seq_along(ids)) {
      s <- paste0(s, sprintf('    %s [label="%s"];\n', ids[i], labels[i]))
    }
    paste0(s, "  }\n")
  }

  # --- Edge builder ---
  make_edges <- function(from_ids, to_ids, mat, ws, comment,
                         dir = "row_to_col", stars_mat = NULL,
                         show_mat = NULL) {
    # dir = "row_to_col": mat[i,j] -> from_ids[j] -> to_ids[i]
    # dir = "col_to_row": mat[i,j] -> from_ids[i] -> to_ids[j]
    # stars_mat: optional matrix of "" or "*"/"**"/"***"
    # show_mat: optional logical matrix; if provided, overrides threshold filter
    s <- paste0('\n  // ', comment, '\n',
                '  edge [color="gray0", fontcolor="gray0", style=solid];\n')
    max_val <- suppressWarnings(max(mat, na.rm = TRUE))
    if (!is.finite(max_val) || max_val <= 0) return(s)
    nr <- nrow(mat); nc <- ncol(mat)
    for (i in seq_len(nr)) {
      for (j in seq_len(nc)) {
        val <- mat[i, j]
        show <- if (!is.null(show_mat)) show_mat[i, j]
                else is.finite(val) && val >= threshold
        if (show) {
          pen <- pw(val, max_val, ws)
          lab <- fmtc(val)
          if (!is.null(stars_mat)) lab <- paste0(lab, stars_mat[i, j])
          if (dir == "row_to_col") {
            s <- paste0(s, sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                                   from_ids[j], to_ids[i], lab, pen))
          } else {
            s <- paste0(s, sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                                   from_ids[i], to_ids[j], lab, pen))
          }
        }
      }
    }
    s
  }

  # --- Significance stars and p-value filter for Theta edges ---
  C_stars <- NULL
  C_show  <- NULL  # logical matrix for sig.level filter
  if (!is.null(result$coefficients)) {
    C_stars <- matrix("", nrow = Q, ncol = R)
    C_pval  <- matrix(NA_real_, nrow = Q, ncol = R)
    cf <- result$coefficients
    dec_names <- rownames(C_mat)  # current basis names (may be renamed)
    enc_names <- colnames(C_mat)  # current covariate names (may be renamed)
    for (k in seq_len(nrow(cf))) {
      q <- match(cf$Basis[k], dec_names)
      r <- match(cf$Covariate[k], enc_names)
      if (!is.na(q) && !is.na(r) && !is.na(cf$p_value[k])) {
        p <- cf$p_value[k]
        C_pval[q, r] <- p
        if (p < 0.001)      C_stars[q, r] <- "***"
        else if (p < 0.01)  C_stars[q, r] <- "**"
        else if (p < 0.05)  C_stars[q, r] <- "*"
      }
    }
    if (!is.null(sig.level)) {
      C_show <- !is.na(C_pval) & C_pval < sig.level
    }
  }

  # === Build DOT ===
  scr <- paste0(
    "digraph NMF_AE {\n",
    "  graph [rankdir=", rankdir, " compound=true];\n",
    "  splines=true; nodesep=0.4; ranksep=0.7; fontname=\"Arial\";\n",
    '  edge [fontname="Arial", fontsize=8, arrowhead=open];\n'
  )

  if (type == "YXCXY") {
    # Clusters: Y2 -> X2 -> X1 -> Y1
    scr <- paste0(scr,
      make_cluster("Y2", Y2.title, Y2_ids[idx_Y2], Y2_labels[idx_Y2], "box", "lightcoral"),
      make_cluster("X2", X2.title, X2_ids, X2_labels, "ellipse", "wheat"),
      make_cluster("X1", X1.title, X1_ids, X1_labels, "ellipse", "wheat"),
      make_cluster("Y1", Y1.title, Y1_ids[idx_Y1], Y1_labels[idx_Y1], "box", "lightblue")
    )
    # Edges: Y2 -> X2 (X2 matrix: R x P2, X2[r,j] = weight from Y2_j to X2_r)
    scr <- paste0(scr,
      make_edges(Y2_ids[idx_Y2], X2_ids, X2[, idx_Y2, drop = FALSE], weight_scale_x2, "Y2 -> X2 (encoder)", "row_to_col"))
    # Edges: X2 -> X1 (C: Q x R, C[q,r] = weight from X2_r to X1_q)
    scr <- paste0(scr,
      make_edges(X2_ids, X1_ids, C_mat, weight_scale_theta, "X2 -> X1 (C)",
                 "row_to_col", stars_mat = C_stars, show_mat = C_show))
    # Edges: X1 -> Y1 (X1: P1 x Q, X1[i,q] = weight from X1_q to Y1_i)
    scr <- paste0(scr,
      make_edges(X1_ids, Y1_ids[idx_Y1], X1[idx_Y1, , drop = FALSE], weight_scale_x1, "X1 -> Y1 (decoder)", "row_to_col"))

  } else {
    # XCX: X2 -> X1 only
    scr <- paste0(scr,
      make_cluster("X2", X2.title, X2_ids, X2_labels, "ellipse", "wheat"),
      make_cluster("X1", X1.title, X1_ids, X1_labels, "ellipse", "wheat")
    )
    scr <- paste0(scr,
      make_edges(X2_ids, X1_ids, C_mat, weight_scale_theta, "X2 -> X1 (C)",
                 "row_to_col", stars_mat = C_stars, show_mat = C_show))
  }

  result <- paste0(scr, "}\n")
  class(result) <- c("nmfae.DOT", "nmfkc.DOT")
  result
}

# nmfae.signed.R
# Signed-Bottleneck NMF-AE: Three-layer NMF-AE with signed bottleneck
# Model:  Y1 ~= X1 (Cp - Cn) X2 Y2,  X1, Cp, Cn, X2, Y2 >= 0,  Y1 signed allowed
# Author: Kenichi Satoh
# Date: 2026-04-18
#
# Depends: nmfkc (for warm.start via nmfae())
# References:
#   Ding, C. H. Q., Li, T., & Jordan, M. I. (2010).
#     Convex and Semi-Nonnegative Matrix Factorizations. IEEE TPAMI 32(1), 45-55.
#   Satoh, K. (2026). Signed-Bottleneck NMF-AE: Signed-Bottleneck 3-Layer NMF
#     (research memo 2026-04-18).
#
# This file provides:
#   nmfae.signed()              -- main Direct-MU fit
#   predict.nmfae.signed()      -- prediction on new Y2
#   plot.nmfae.signed()         -- convergence plot
#   summary.nmfae.signed()      -- summary with Cp/Cn diagnostics
#   print.summary.nmfae.signed()
#   nmfae.signed.rename()       -- rename Dec/Enc labels (propagates to Cp, Cn, C)
# ==============================================================

#' @title Signed-Bottleneck NMF-AE: Three-Layer NMF-AE with Signed Bottleneck
#' @description
#' \code{nmfae.signed} fits a three-layer non-negative matrix factorization
#' autoencoder with a \strong{signed} bottleneck, solving
#' \deqn{Y_1 \approx X_1 (C_{+} - C_{-}) X_2 Y_2,
#'   \quad X_1 \ge 0,\; C_{+} \ge 0,\; C_{-} \ge 0,\; X_2 \ge 0,}
#' where \eqn{\Theta = C_{+} - C_{-}} is the signed bottleneck.  The basis
#' matrices \eqn{X_1} (columns sum to 1) and \eqn{X_2} (rows sum to 1) retain
#' their non-negative "parts-based" interpretability, while \eqn{\Theta} can
#' express anti-correlations (e.g., refractive index up vs. Abbe number down).
#'
#' The algorithm uses \strong{Direct Multiplicative Updates} derived from
#' Ding et al. (2010) sign-splitting technique, applied block-wise to the
#' four non-negative blocks \eqn{(C_{+}, C_{-}, X_1, X_2)}.  Each block update
#' monotonically decreases the true objective
#' \eqn{\|Y_1 - X_1(C_{+} - C_{-})X_2 Y_2\|_F^2} (Lee-Seung auxiliary
#' function method).
#'
#' \strong{Relation to \code{\link{nmfae}}:} When \eqn{\Theta \ge 0} suffices
#' (the \code{nmfae} case), \code{nmfae.signed} reduces to \code{nmfae} up to
#' the \eqn{C_{+} - C_{-}} parameterization.  Use \code{nmfae.signed} when the
#' data exhibit negative cross-property correlations that tri-NMF-AE cannot
#' express (e.g., high refractive index <-> low Abbe number trade-off).
#'
#' @param Y1 Output matrix \eqn{Y_1} (P1 x N).  Negative entries allowed
#'   (\code{Y.signed = TRUE} is auto-detected).
#' @param Y2 Input matrix \eqn{Y_2} (P2 x N).  Must be non-negative.
#'   Default \code{Y1} (autoencoder).
#' @param rank Integer.  Decoder rank Q.  Default 2.
#'   Alias \code{Q} accepted via \code{...}.
#' @param rank.encoder Integer.  Encoder rank R.  Default \code{rank}.
#'   Alias \code{R} accepted via \code{...}.
#' @param epsilon Relative convergence tolerance on the objective.
#'   Default \code{1e-4}.
#' @param maxit Maximum iterations.  Default 5000.
#' @param verbose Logical.  Print progress.  Default \code{FALSE}.
#' @param ... Additional arguments:
#'   \describe{
#'     \item{\code{warm.start}}{One of \code{TRUE} (default, \strong{hybrid}:
#'       warm-start \eqn{X_1, X_2} from \code{\link{nmfae}} but initialize
#'       \eqn{C_{+}, C_{-}} randomly), \code{"full"} (warm-start everything
#'       including \eqn{C_{+} = C_{\mathrm{tri}}}, \eqn{C_{-} = \delta}),
#'       or \code{FALSE} (random for all blocks).  The hybrid default avoids
#'       the \eqn{C_{-} = 0} local-minimum trap inherited from tri-NMF-AE
#'       while still benefiting from good \eqn{X_1, X_2} initialization.
#'       Ignored when \eqn{Y_1} has negative entries.}
#'     \item{\code{nstart}}{Integer, default 1 (cf. \code{\link{kmeans}}).
#'       Number of random restarts for \eqn{C_{+}, C_{-}}.  Each restart
#'       uses seed \code{seed + 7919 * (s-1)}.  Returns the best run by
#'       final objective.  \strong{Signed models have more local minima
#'       than non-negative ones} because the bottleneck
#'       \eqn{\Theta = C_{+} - C_{-}} can take both positive and negative
#'       values; during exploration a larger \code{nstart} (e.g., 10-50)
#'       reduces the chance of being trapped at an inferior stationary
#'       point (particularly the \eqn{C_{-} = 0} trap from warm-start
#'       from non-negative tri-NMF-AE).  Use the default 1 for fast
#'       development and raise for publication-grade runs.}
#'     \item{\code{Y1.weights}}{Optional non-negative weight matrix
#'       (P1 x N) or vector (length N) for \eqn{Y_1}, analogous to the
#'       \code{weights} argument of \code{\link[stats]{lm}}.  Loss
#'       becomes \eqn{\sum W_{ij} \, (Y_{1,ij} - \hat Y_{1,ij})^2}
#'       (\code{lm()}-style, \strong{linear} in \eqn{W}).  Logical
#'       matrices (\code{TRUE} / \code{FALSE}) are also accepted.
#'       Used by \code{\link{nmfae.signed.ecv}} to hold out test
#'       elements via a binary mask \eqn{W \in \{0,1\}}; real-valued
#'       weights for importance weighting are also supported.  Default:
#'       if \code{Y1} has \code{NA}, a binary mask is auto-generated
#'       (0 for \code{NA}, 1 elsewhere).}
#'     \item{\code{Cp.init}, \code{Cn.init}}{Explicit \eqn{Q \times R}
#'       non-negative matrices for initialization.  Overrides warm.start.}
#'     \item{\code{C.init}}{Explicit signed \eqn{Q \times R} matrix,
#'       internally split into \eqn{(C_{+}, C_{-})}.}
#'     \item{\code{X1.init}, \code{X2.init}}{Explicit basis matrices.}
#'     \item{\code{seed}}{RNG seed.  Default 123.}
#'     \item{\code{print.trace}}{Logical.  Print iteration trace.
#'       Default \code{FALSE}.}
#'     \item{\code{prefix.dec}, \code{prefix.enc}}{Label prefixes for
#'       decoder/encoder factors.  Default \code{"Dec"}, \code{"Enc"}.}
#'   }
#'
#' @return An object of class \code{c("nmfae.signed", "nmfae", "nmf")} with:
#' \item{X1}{Decoder basis (P1 x Q), column sum 1.}
#' \item{Cp, Cn}{Non-negative parts of \eqn{\Theta} (each Q x R).}
#' \item{C}{Signed bottleneck \eqn{\Theta = C_{+} - C_{-}} (Q x R).}
#' \item{X2}{Encoder basis (R x P2), row sum 1.}
#' \item{Y1hat}{Fitted values \eqn{X_1 (C_{+} - C_{-}) X_2 Y_2}.}
#' \item{H}{Encoding \eqn{(C_{+} - C_{-}) X_2 Y_2} (Q x N, signed).}
#' \item{rank}{\code{c(Q = Q, R = R)}.}
#' \item{dims}{\code{c(P1, P2, N)}.}
#' \item{objfunc, objfunc.iter}{Final and per-iteration objective values.}
#' \item{r.squared}{\eqn{\mathrm{cor}(Y_1, \widehat Y_1)^2} (Pearson; in \eqn{[0,1]}).}
#' \item{r.squared.uncentered}{Uncentered \eqn{R^2 = 1 - \|Y_1 - \widehat Y_1\|_F^2 / \|Y_1\|_F^2} (baseline = zero matrix).}
#' \item{r.squared.centered}{Row-mean centered \eqn{1 - \|Y_1 - \widehat Y_1\|_F^2 / \|Y_1 - \bar Y_{p\cdot}\|_F^2}.}
#' \item{sigma, mae}{Residual SE and mean absolute error.}
#' \item{niter, runtime}{Iterations and elapsed seconds.}
#' \item{Y.signed}{Logical; whether \eqn{Y_1} contained negative entries.}
#' \item{call}{Matched call.}
#'
#' @section Lifecycle:
#' This function is \strong{experimental}; interface may change.
#'
#' @seealso \code{\link{nmfae}}, \code{\link{predict.nmfae.signed}},
#'   \code{\link{summary.nmfae.signed}}, \code{\link{nmfae.signed.rename}}
#'
#' @references
#' Ding, C.H.Q., Li, T., and Jordan, M.I. (2010).  Convex and
#' Semi-Nonnegative Matrix Factorizations.
#' \emph{IEEE TPAMI}, 32(1), 45-55.
#'
#' Satoh, K. (2026).  Signed-Bottleneck NMF-AE: Signed-Bottleneck 3-Layer NMF
#' (research memo, 2026-04-18).
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' Y1 <- matrix(abs(rnorm(12)), 3, 4)
#' Y2 <- matrix(abs(rnorm(20)), 5, 4)
#' res <- nmfae.signed(Y1, Y2, rank = 2, rank.encoder = 2, maxit = 500)
#' summary(res)
#' }
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
nmfae.signed <- function(Y1, Y2 = Y1, rank = 2, rank.encoder = rank,
                       epsilon = 1e-4, maxit = 5000,
                       verbose = FALSE, ...) {

  cl <- match.call()
  extra_args <- list(...)

  ## ---- 1. Parameter extraction (compatibility with nmfae) ----
  if (!is.null(extra_args$Q)) rank <- extra_args$Q
  if (!is.null(extra_args$R)) rank.encoder <- extra_args$R
  Q <- as.integer(rank)
  R <- as.integer(rank.encoder)

  warm.start <- if (!is.null(extra_args$warm.start)) extra_args$warm.start else TRUE
  nstart  <- if (!is.null(extra_args$nstart))  extra_args$nstart  else 1L
  Y1.weights <- if (!is.null(extra_args$Y1.weights)) extra_args$Y1.weights else NULL
  Cp.init    <- extra_args$Cp.init
  Cn.init    <- extra_args$Cn.init
  C.init     <- extra_args$C.init
  X1.init    <- extra_args$X1.init
  X2.init    <- extra_args$X2.init
  seed       <- if (!is.null(extra_args$seed))       extra_args$seed       else 123L
  print.trace <- verbose
  if (!is.null(extra_args$print.trace)) print.trace <- extra_args$print.trace
  prefix.dec <- if (!is.null(extra_args$prefix.dec)) extra_args$prefix.dec else "Dec"
  prefix.enc <- if (!is.null(extra_args$prefix.enc)) extra_args$prefix.enc else "Enc"

  ## Validate warm.start value
  if (is.logical(warm.start)) {
    warm_mode <- if (isTRUE(warm.start)) "hybrid" else "random"
  } else {
    warm_mode <- match.arg(as.character(warm.start),
                            c("hybrid", "full", "random"))
  }

  ## ---- 2. Input preparation & validation ----
  ## NA in Y1 is auto-masked via Y1.weights (matches nmfae() behavior);
  ## NA in Y2 is still an error since Y2 is not weighted.
  Y1 <- as.matrix(Y1); storage.mode(Y1) <- "double"
  Y2 <- as.matrix(Y2); storage.mode(Y2) <- "double"
  if (any(is.na(Y2))) stop("Y2 contains NA; please impute or remove.")
  if (min(Y2) < 0) stop("Y2 must be non-negative.")
  if (ncol(Y1) != ncol(Y2)) stop("Y1 and Y2 must have the same number of columns (N).")

  P1 <- nrow(Y1); P2 <- nrow(Y2); N <- ncol(Y1)
  Y1_signed <- any(Y1 < 0)
  small <- 1e-16
  eps_init <- 1e-6
  start.time <- Sys.time()

  if (print.trace) {
    message(sprintf(
      "Y1(%d,%d) ~ X1(%d,%d) [Cp(%d,%d)-Cn(%d,%d)] X2(%d,%d) Y2(%d,%d)%s",
      P1, N, P1, Q, Q, R, Q, R, R, P2, P2, N,
      if (Y1_signed) "  [Y1 signed]" else ""))
  }

  ## ---- 3a. Y1.weights handling (for ECV / missing-element masking) ----
  ## Vector of length N -> per-column; scalar -> all; NULL -> auto-detect NAs
  if (!is.null(Y1.weights) && is.vector(Y1.weights) && !is.matrix(Y1.weights)) {
    if (length(Y1.weights) == N) {
      Y1.weights <- matrix(Y1.weights, nrow = P1, ncol = N, byrow = TRUE)
    } else if (length(Y1.weights) == 1) {
      Y1.weights <- matrix(Y1.weights, nrow = P1, ncol = N)
    } else {
      stop("Length of Y1.weights vector must match ncol(Y1) or be 1.")
    }
  }
  if (is.null(Y1.weights) && any(is.na(Y1))) {
    Y1.weights <- matrix(1, nrow = P1, ncol = N)
    Y1.weights[is.na(Y1)] <- 0
    Y1[is.na(Y1)] <- 0
  } else if (!is.null(Y1.weights)) {
    Y1.weights <- as.matrix(Y1.weights)
    if (!all(dim(Y1.weights) == dim(Y1)))
      stop("Y1.weights dimensions must match Y1.")
    Y1.weights[is.na(Y1.weights)] <- 0
    Y1[is.na(Y1) | Y1.weights == 0] <- 0
  }
  has.weights <- !is.null(Y1.weights) && any(Y1.weights != 1)
  if (!has.weights) Y1.weights <- NULL
  Wmat <- Y1.weights  # short alias for weighted path

  ## ---- 3b. Pre-computation (unweighted path only; N-dependent, outer-loop) ----
  if (!has.weights) {
    S  <- tcrossprod(Y2)             # P2 x P2, >= 0 (since Y2 >= 0)
    G0 <- tcrossprod(Y1, Y2)         # P1 x P2 (= Y1 Y2^T), signed iff Y1 signed
  } else {
    S <- NULL; G0 <- NULL            # weighted path skips these precomputes
  }
  Y1_sqnorm <- sum(Y1 * Y1)

  ## ---- 4. Initialization ----
  ## Strategy:
  ##   hybrid: X1/X2 from tri-NMF-AE, Cp/Cn random (DEFAULT, best for Abbe-like data)
  ##   full:   X1/X2 from tri-NMF-AE, Cp = tri$C, Cn = eps (old tri-NMF solution)
  ##   random: all blocks random (slowest warm-up, widest exploration)
  X1 <- NULL; X2 <- NULL; Cp <- NULL; Cn <- NULL

  use_explicit_X <- !is.null(X1.init) || !is.null(X2.init)
  use_explicit_C <- !is.null(Cp.init) || !is.null(Cn.init) || !is.null(C.init)

  ## 4a. X1, X2: warm-start from tri-NMF-AE unless mode=random or explicit override
  tri_C_for_full <- NULL
  if (warm_mode %in% c("hybrid", "full") && !Y1_signed && !use_explicit_X) {
    if (print.trace) message("  Init: warm-start X1, X2 from nmfae() ...")
    res0 <- nmfae(Y1, Y2, rank = Q, rank.encoder = R,
                          epsilon = epsilon, maxit = maxit,
                          verbose = FALSE, seed = seed)
    X1 <- res0$X1
    X2 <- res0$X2
    tri_C_for_full <- res0$C          # for warm_mode == "full"
  } else {
    if (!is.null(X1.init)) {
      X1 <- as.matrix(X1.init)
      if (!identical(dim(X1), c(P1, Q)))
        stop("X1.init must have dimensions (nrow(Y1), rank).")
    } else {
      set.seed(seed)
      X1 <- matrix(abs(stats::rnorm(P1 * Q)) * 0.1, P1, Q)
    }
    if (!is.null(X2.init)) {
      X2 <- as.matrix(X2.init)
      if (!identical(dim(X2), c(R, P2)))
        stop("X2.init must have dimensions (rank.encoder, nrow(Y2)).")
    } else {
      set.seed(seed + 1L)
      X2 <- matrix(abs(stats::rnorm(R * P2)) * 0.1, R, P2)
    }
  }

  ## 4b. Cp, Cn: random (hybrid / random modes), tri (full mode), or explicit
  init_Cp_Cn <- function(s) {
    if (!is.null(Cp.init) || !is.null(Cn.init)) {
      if (is.null(Cp.init) || is.null(Cn.init))
        stop("Provide both Cp.init and Cn.init, or neither.")
      cp <- as.matrix(Cp.init); cn <- as.matrix(Cn.init)
      if (!identical(dim(cp), c(Q, R)) || !identical(dim(cn), c(Q, R)))
        stop("Cp.init and Cn.init must have dimensions (Q, R).")
      if (any(cp < 0) || any(cn < 0))
        stop("Cp.init, Cn.init must be non-negative.")
      return(list(Cp = cp, Cn = cn))
    }
    if (!is.null(C.init)) {
      C0 <- as.matrix(C.init)
      if (!identical(dim(C0), c(Q, R)))
        stop("C.init must have dimensions (Q, R).")
      return(list(Cp = pmax(C0, 0), Cn = pmax(-C0, 0)))
    }
    if (warm_mode == "full" && !is.null(tri_C_for_full)) {
      return(list(Cp = tri_C_for_full,
                  Cn = matrix(eps_init, Q, R)))
    }
    ## Sample signed C from U(-1, 1), split into Cp, Cn.  Scale is arbitrary
    ## because the multiplicative updates auto-rescale; non-zero + sign-
    ## symmetric diversity is what matters.
    set.seed(s)
    C0 <- matrix(stats::runif(Q * R, min = -1, max = 1), Q, R)
    list(Cp = pmax(C0, 0), Cn = pmax(-C0, 0))
  }

  ## ---- 5. Objective (weighted if Wmat present) ----
  compute_obj <- function(X1, Cp, Cn, X2) {
    Y1hat <- X1 %*% (Cp - Cn) %*% X2 %*% Y2
    ## lm()-style weighted least squares: L = sum(W * (Y1 - Y1hat)^2).
    ## W = W^2 for binary {0,1} masks so this is unchanged for ECV.
    if (has.weights) sum(Wmat * (Y1 - Y1hat)^2) else sum((Y1 - Y1hat)^2)
  }

  ## ---- 6. Main loop wrapped for multi-restart ----
  run_once <- function(X1, X2, Cp, Cn) {
    ## Normalize
    cs <- colSums(X1) + small
    X1 <- sweep(X1, 2, cs, "/"); Cp <- sweep(Cp, 1, cs, "*"); Cn <- sweep(Cn, 1, cs, "*")
    rs <- rowSums(X2) + small
    X2 <- sweep(X2, 1, rs, "/"); Cp <- sweep(Cp, 2, rs, "*"); Cn <- sweep(Cn, 2, rs, "*")
    obj_prev <- compute_obj(X1, Cp, Cn, X2)
    objfunc.iter <- numeric(maxit)
    iter <- 0L
    for (iter in seq_len(maxit)) {

    if (!has.weights) {
      ## ---- Fast unweighted path (uses precomputed S, G0) ----
      P1m <- crossprod(X1)
      SX  <- X2 %*% S %*% t(X2)
      GX  <- crossprod(X1, G0) %*% t(X2)
      GX_p <- pmax(GX,  0); GX_n <- pmax(-GX, 0)
      ## Cp update
      Cp <- Cp * (GX_p + P1m %*% Cn %*% SX) /
                 (GX_n + P1m %*% Cp %*% SX + small)
      ## Cn update (Gauss-Seidel)
      Cn <- Cn * (GX_n + P1m %*% Cp %*% SX) /
                 (GX_p + P1m %*% Cn %*% SX + small)
      ## X1 update
      G0X <- G0 %*% t(X2)
      G0X_p <- pmax(G0X, 0); G0X_n <- pmax(-G0X, 0)
      MMt_p <- Cp %*% SX %*% t(Cp) + Cn %*% SX %*% t(Cn)
      MMt_n <- Cp %*% SX %*% t(Cn) + Cn %*% SX %*% t(Cp)
      X1 <- X1 * (G0X_p %*% t(Cp) + G0X_n %*% t(Cn) + X1 %*% MMt_n) /
                 (G0X_n %*% t(Cp) + G0X_p %*% t(Cn) + X1 %*% MMt_p + small)
      ## Normalize X1 cols -> absorb into Cp, Cn rows
      cs <- colSums(X1) + small
      X1 <- sweep(X1, 2, cs, "/")
      Cp <- sweep(Cp, 1, cs, "*"); Cn <- sweep(Cn, 1, cs, "*")
      ## X2 update
      Amat <- crossprod(X1, G0)
      A_p <- pmax(Amat, 0); A_n <- pmax(-Amat, 0)
      HG_p <- crossprod(Cp, A_p) + crossprod(Cn, A_n)
      HG_n <- crossprod(Cp, A_n) + crossprod(Cn, A_p)
      P1m <- crossprod(X1)
      HtH_p <- crossprod(Cp, P1m) %*% Cp + crossprod(Cn, P1m) %*% Cn
      HtH_n <- crossprod(Cp, P1m) %*% Cn + crossprod(Cn, P1m) %*% Cp
      X2 <- X2 * (HG_p + HtH_n %*% X2 %*% S) /
                 (HG_n + HtH_p %*% X2 %*% S + small)
      ## Normalize X2 rows -> absorb into Cp, Cn cols
      rs <- rowSums(X2) + small
      X2 <- sweep(X2, 1, rs, "/")
      Cp <- sweep(Cp, 2, rs, "*"); Cn <- sweep(Cn, 2, rs, "*")

    } else {
      ## ---- Weighted path (no S/G0 precompute; ECV / missing data) ----
      F_mat <- X2 %*% Y2                                 # R x N
      ## Cp update
      XCpF <- X1 %*% Cp %*% F_mat                        # P1 x N, >= 0
      XCnF <- X1 %*% Cn %*% F_mat                        # P1 x N, >= 0
      WY1  <- Wmat * Y1
      WXCpF <- Wmat * XCpF
      WXCnF <- Wmat * XCnF
      G_w  <- crossprod(X1, WY1) %*% t(F_mat)            # Q x R, signed iff Y1 signed
      Hp_w <- crossprod(X1, WXCpF) %*% t(F_mat)          # Q x R, >= 0
      Hn_w <- crossprod(X1, WXCnF) %*% t(F_mat)          # Q x R, >= 0
      Gp <- pmax(G_w, 0); Gn <- pmax(-G_w, 0)
      Cp <- Cp * (Gp + Hn_w) / (Gn + Hp_w + small)
      ## Cn update (recompute Hp_w with new Cp)
      XCpF  <- X1 %*% Cp %*% F_mat
      WXCpF <- Wmat * XCpF
      Hp_w  <- crossprod(X1, WXCpF) %*% t(F_mat)
      Cn <- Cn * (Gn + Hp_w) / (Gp + Hn_w + small)
      ## X1 update
      Mp <- Cp %*% F_mat; Mn <- Cn %*% F_mat
      XMp <- X1 %*% Mp; XMn <- X1 %*% Mn
      WY1  <- Wmat * Y1
      WXMp <- Wmat * XMp; WXMn <- Wmat * XMn
      W1Mp <- WY1 %*% t(Mp); W1Mn <- WY1 %*% t(Mn)
      W1Mp_p <- pmax(W1Mp, 0); W1Mp_n <- pmax(-W1Mp, 0)
      W1Mn_p <- pmax(W1Mn, 0); W1Mn_n <- pmax(-W1Mn, 0)
      P_pp <- WXMp %*% t(Mp); P_pn <- WXMp %*% t(Mn)
      P_np <- WXMn %*% t(Mp); P_nn <- WXMn %*% t(Mn)
      X1 <- X1 * (W1Mp_p + W1Mn_n + P_pn + P_np) /
                 (W1Mp_n + W1Mn_p + P_pp + P_nn + small)
      cs <- colSums(X1) + small
      X1 <- sweep(X1, 2, cs, "/")
      Cp <- sweep(Cp, 1, cs, "*"); Cn <- sweep(Cn, 1, cs, "*")
      ## X2 update: Y1 ~= H X2 Y2, H = X1 Cp - X1 Cn
      Hp <- X1 %*% Cp; Hn <- X1 %*% Cn  # both >= 0
      HpF <- Hp %*% X2 %*% Y2; HnF <- Hn %*% X2 %*% Y2   # >= 0
      WHpF <- Wmat * HpF; WHnF <- Wmat * HnF
      WY1 <- Wmat * Y1
      Hp_tW1 <- crossprod(Hp, WY1) %*% t(Y2)              # R x P2, signed iff Y1 signed
      Hn_tW1 <- crossprod(Hn, WY1) %*% t(Y2)
      Hp_tW1_p <- pmax(Hp_tW1, 0); Hp_tW1_n <- pmax(-Hp_tW1, 0)
      Hn_tW1_p <- pmax(Hn_tW1, 0); Hn_tW1_n <- pmax(-Hn_tW1, 0)
      Q_pp <- crossprod(Hp, WHpF) %*% t(Y2)
      Q_pn <- crossprod(Hp, WHnF) %*% t(Y2)
      Q_np <- crossprod(Hn, WHpF) %*% t(Y2)
      Q_nn <- crossprod(Hn, WHnF) %*% t(Y2)
      X2 <- X2 * (Hp_tW1_p + Hn_tW1_n + Q_pn + Q_np) /
                 (Hp_tW1_n + Hn_tW1_p + Q_pp + Q_nn + small)
      rs <- rowSums(X2) + small
      X2 <- sweep(X2, 1, rs, "/")
      Cp <- sweep(Cp, 2, rs, "*"); Cn <- sweep(Cn, 2, rs, "*")
    }

    ## Objective and convergence
    obj_cur <- compute_obj(X1, Cp, Cn, X2)
    objfunc.iter[iter] <- obj_cur

    if (print.trace && (iter %% 100 == 0 || iter == 1)) {
      message(sprintf("  iter %5d: objfunc = %.6f", iter, obj_cur))
    }
    if (iter > 1) {
      rel <- abs(obj_prev - obj_cur) / (abs(obj_prev) + small)
      if (rel < epsilon) {
        if (print.trace) message(sprintf("  Converged at iter %d", iter))
        break
      }
    }
    obj_prev <- obj_cur
    }  # end for iter
    ## Warn when the MU loop exhausts maxit without convergence
    ## (matches nmfkc() / nmf.sem() convention).  Each restart inside
    ## the multi-start loop emits its own warning if it fails to
    ## converge, so users see how many restarts hit the cap.
    if (iter == maxit && exists("rel") && rel >= epsilon)
      warning(paste0("maximum iterations (", maxit, ") reached..."))
    list(X1 = X1, X2 = X2, Cp = Cp, Cn = Cn,
         objfunc.iter = objfunc.iter[seq_len(iter)], niter = iter,
         objfunc = obj_prev)
  }  # end run_once

  ## Execute run(s)
  best <- NULL
  for (s in seq_len(nstart)) {
    ## Wide seed spacing via a large prime (7919) to maximize init diversity
    s_seed <- seed + 7919L * (s - 1L)
    cc <- init_Cp_Cn(s_seed + 2L)
    if (print.trace && nstart > 1)
      message(sprintf("  Restart %d/%d (seed=%d) ...", s, nstart, s_seed))
    ## For random mode, also regenerate X1, X2 per restart
    Xs1 <- X1; Xs2 <- X2
    if (warm_mode == "random" && !use_explicit_X && nstart > 1) {
      set.seed(s_seed)
      Xs1 <- matrix(abs(stats::rnorm(P1 * Q)) * 0.1, P1, Q)
      set.seed(s_seed + 1L)
      Xs2 <- matrix(abs(stats::rnorm(R * P2)) * 0.1, R, P2)
    }
    out <- run_once(Xs1, Xs2, cc$Cp, cc$Cn)
    if (is.null(best) || out$objfunc < best$objfunc) best <- out
  }
  X1 <- best$X1; X2 <- best$X2; Cp <- best$Cp; Cn <- best$Cn
  objfunc.iter <- best$objfunc.iter
  iter <- best$niter
  niter <- iter
  diff.time <- as.numeric(difftime(Sys.time(), start.time, units = "secs"))

  ## ---- 7. Post-processing ----
  ## Reorder X1 columns by centroid of row indices
  if (Q > 1) {
    idx1 <- order(matrix(seq_len(P1) / P1, nrow = 1) %*% X1)
    X1 <- X1[, idx1, drop = FALSE]
    Cp <- Cp[idx1, , drop = FALSE]
    Cn <- Cn[idx1, , drop = FALSE]
  }
  ## Reorder X2 rows by centroid of column indices
  if (R > 1) {
    idx2 <- order(X2 %*% matrix(seq_len(P2) / P2, ncol = 1))
    X2 <- X2[idx2, , drop = FALSE]
    Cp <- Cp[, idx2, drop = FALSE]
    Cn <- Cn[, idx2, drop = FALSE]
  }

  ## Names
  rownames(X1) <- rownames(Y1)
  colnames(X2) <- rownames(Y2)
  colnames(X1) <- paste0(prefix.dec, seq_len(Q))
  rownames(Cp) <- paste0(prefix.dec, seq_len(Q))
  rownames(Cn) <- paste0(prefix.dec, seq_len(Q))
  colnames(Cp) <- paste0(prefix.enc, seq_len(R))
  colnames(Cn) <- paste0(prefix.enc, seq_len(R))
  rownames(X2) <- paste0(prefix.enc, seq_len(R))

  C <- Cp - Cn                                  # Q x R, signed
  H <- C %*% X2 %*% Y2                          # Q x N, signed encoding
  Y1hat <- X1 %*% H
  resid <- Y1 - Y1hat

  ## Goodness-of-fit statistics.  lm()-style weighted least squares:
  ## the reported objfunc / sigma / mae use sum(W * resid^2) to match the
  ## in-loop objective compute_obj().  For binary {0,1} masks (the
  ## standard ECV / CV / NA-mask case) this restricts statistics to valid
  ## elements; for real-valued weights it returns weighted averages.
  if (has.weights) {
    objfunc  <- sum(Wmat * resid^2)
    valid    <- (Wmat > 0)
    n.valid  <- sum(valid)
    r2_all   <- .r.squared.all(Y1, Y1hat, Y.weights = Wmat)
    sigma    <- if (n.valid > 0) sqrt(objfunc / n.valid) else NA_real_
    mae      <- if (sum(Wmat) > 0) sum(Wmat * abs(resid)) / sum(Wmat) else NA_real_
  } else {
    objfunc  <- sum(resid * resid)
    r2_all   <- .r.squared.all(Y1, Y1hat)
    sigma    <- sqrt(objfunc / (P1 * N))
    mae      <- mean(abs(resid))
  }
  r.squared          <- r2_all$r.squared
  r.squared.uncentered     <- r2_all$r.squared.uncentered
  r.squared.centered <- r2_all$r.squared.centered

  ## Soft/hard clustering of encoding (only meaningful when H has interpretable sign)
  eps_bp <- 1e-16
  ## Clip to non-negative for clustering probabilities
  Hp <- pmax(H, 0)
  col_mass <- colSums(Hp) + eps_bp
  B.prob <- sweep(Hp, 2, col_mass, "/")
  B.cluster <- apply(B.prob, 2, which.max)
  B.cluster[colSums(Hp) == 0] <- NA

  if (print.trace) {
    message(sprintf("  Done: %d iterations, %.1f sec, R2 = %.4f",
                    niter, diff.time, r.squared))
  }

  result <- list(
    call = cl,
    X1 = X1,
    Cp = Cp,
    Cn = Cn,
    C  = C,
    X2 = X2,
    Y1hat = Y1hat,
    H = H,
    B.prob = B.prob,
    B.cluster = B.cluster,
    rank = c(Q = Q, R = R),
    dims = c(P1 = P1, P2 = P2, N = N),
    objfunc = objfunc,
    objfunc.iter = objfunc.iter,
    r.squared          = r.squared,
    r.squared.uncentered     = r.squared.uncentered,
    r.squared.centered = r.squared.centered,
    sigma = sigma,
    mae = mae,
    niter = niter,
    runtime = diff.time,
    n.missing = if (has.weights) sum(Wmat == 0) else 0L,
    n.total = P1 * N,
    Y.signed = Y1_signed
  )
  class(result) <- c("nmfae.signed", "nmfae", "nmf")
  result
}


## ==============================================================
#' @title Statistical Inference for Signed-Bottleneck NMF-AE Signed Bottleneck
#' @description
#' Post-estimation inference for the \strong{signed} bottleneck
#' \eqn{\Theta = C_{+} - C_{-}} in the Signed-Bottleneck NMF-AE model
#' \eqn{Y_1 \approx X_1 \Theta X_2 Y_2}, conditional on
#' \eqn{(\hat X_1, \hat X_2)}.  Uses sandwich covariance and wild bootstrap
#' \strong{without} the non-negativity projection that \code{\link{nmfae.inference}}
#' applies (because \eqn{\Theta} is unconstrained in sign here).
#'
#' @param object A fitted \code{"nmfae.signed"} object.
#' @param Y1 Output matrix used during fitting.
#' @param Y2 Input matrix used during fitting.  Default \code{Y1}.
#' @param wild.bootstrap Logical.  Default \code{TRUE}.
#' @param ... Additional arguments:
#'   \describe{
#'     \item{\code{wild.B}}{Bootstrap replicates.  Default 500.}
#'     \item{\code{wild.seed}}{RNG seed.  Default 123.}
#'     \item{\code{wild.level}}{CI confidence level.  Default 0.95.}
#'     \item{\code{sandwich}}{Use sandwich covariance.  Default \code{TRUE}.}
#'     \item{\code{C.p.side}}{P-value type: \code{"two.sided"} (default for
#'       Signed-Bottleneck NMF-AE) or \code{"one.sided"}.}
#'     \item{\code{cov.ridge}}{Ridge stabilization.  Default \code{1e-8}.}
#'     \item{\code{print.trace}}{Logical.  Default \code{FALSE}.}
#'   }
#'
#' @return An object of class \code{c("nmfae.signed.inference",
#'   "nmfae.inference", "nmfae.signed", "nmfae", "nmf")} with added fields:
#' \item{sigma2.used}{Estimated \eqn{\sigma^2}.}
#' \item{C.se, C.se.boot}{Sandwich / bootstrap SEs for \eqn{\Theta} (Q x R).}
#' \item{C.ci.lower, C.ci.upper}{Bootstrap CIs.}
#' \item{coefficients}{Data frame with Estimate, SE, BSE, z, p-value, CI.}
#' \item{C.p.side}{P-value side used.}
#'
#' @seealso \code{\link{nmfae.signed}}, \code{\link{nmfae.inference}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
nmfae.signed.inference <- function(object, Y1, Y2 = Y1,
                                 wild.bootstrap = TRUE, ...) {
  if (!inherits(object, "nmfae.signed"))
    stop("object must be of class 'nmfae.signed'")

  extra_args <- list(...)
  wild.B     <- if (!is.null(extra_args$wild.B))     extra_args$wild.B     else 500
  wild.seed  <- if (!is.null(extra_args$wild.seed))  extra_args$wild.seed  else 123
  wild.level <- if (!is.null(extra_args$wild.level)) extra_args$wild.level else 0.95
  sandwich   <- if (!is.null(extra_args$sandwich))   extra_args$sandwich   else TRUE
  C.p.side   <- if (!is.null(extra_args$C.p.side))   extra_args$C.p.side   else "two.sided"
  cov.ridge  <- if (!is.null(extra_args$cov.ridge))  extra_args$cov.ridge  else 1e-8
  print.trace <- if (!is.null(extra_args$print.trace)) extra_args$print.trace else FALSE

  X1 <- object$X1
  C  <- object$C                      # signed Theta = Cp - Cn
  X2 <- object$X2
  Y1hat <- object$Y1hat
  Q  <- unname(object$rank["Q"])
  R  <- unname(object$rank["R"])
  P1 <- nrow(Y1)
  N  <- ncol(Y1)

  Z   <- X2 %*% Y2
  R_C <- Y1 - Y1hat

  denom <- max(P1 * N - Q * R, 1)
  sigma2.used <- sum(R_C^2) / denom

  X1tX1 <- crossprod(X1)
  ZZt   <- tcrossprod(Z)
  Info_core <- kronecker(ZZt, X1tX1)
  Info <- Info_core / max(sigma2.used, 1e-12)
  Info <- Info + diag(cov.ridge, nrow(Info))

  Hinv <- tryCatch(solve(Info), error = function(e) {
    if (requireNamespace("MASS", quietly = TRUE)) MASS::ginv(Info)
    else stop("Information matrix singular; install MASS package.")
  })

  V_sand <- NULL
  if (isTRUE(sandwich)) {
    X1t <- t(X1)
    J <- matrix(0, Q * R, Q * R)
    for (n in 1:N) {
      z_n <- Z[, n, drop = FALSE]
      r_n <- R_C[, n, drop = FALSE]
      g_n <- X1t %*% r_n
      S_n <- -(g_n %*% t(z_n)) / max(sigma2.used, 1e-12)
      s_n <- as.vector(S_n)
      J <- J + tcrossprod(s_n)
    }
    if (N > 1) J <- (N / (N - 1)) * J
    V_sand <- Hinv %*% J %*% Hinv
  }
  C.vec.cov <- if (!is.null(V_sand)) V_sand else Hinv
  se_vec <- sqrt(pmax(diag(C.vec.cov), 0))
  C.se <- matrix(se_vec, nrow = Q, ncol = R, byrow = FALSE)

  ## Wild bootstrap (NO non-negative projection; Theta is signed)
  C.se.boot <- NULL; C.ci.lower <- NULL; C.ci.upper <- NULL
  if (isTRUE(wild.bootstrap)) {
    set.seed(wild.seed)
    X1t <- t(X1)
    score_mat <- matrix(0, Q * R, N)
    for (n in 1:N) {
      z_n <- Z[, n, drop = FALSE]
      r_n <- R_C[, n, drop = FALSE]
      g_n <- X1t %*% r_n
      G_n <- -(g_n %*% t(z_n)) / max(sigma2.used, 1e-12)
      score_mat[, n] <- as.vector(G_n)
    }
    C_hat_vec <- as.vector(C)
    C_boot <- matrix(NA_real_, nrow = Q * R, ncol = wild.B)
    for (b in 1:wild.B) {
      w <- stats::rexp(N, rate = 1) - 1
      grad_b <- as.vector(score_mat %*% w)
      c_b <- C_hat_vec - as.vector(Hinv %*% grad_b)
      ## NOTE: no pmax(c_b, 0) -- Theta is signed
      C_boot[, b] <- c_b
    }
    sd_vec <- apply(C_boot, 1, stats::sd, na.rm = TRUE)
    C.se.boot <- matrix(sd_vec, nrow = Q, ncol = R, byrow = FALSE)
    alpha <- 1 - wild.level
    lo <- apply(C_boot, 1, stats::quantile, probs = alpha / 2, na.rm = TRUE, names = FALSE)
    hi <- apply(C_boot, 1, stats::quantile, probs = 1 - alpha / 2, na.rm = TRUE, names = FALSE)
    C.ci.lower <- matrix(lo, nrow = Q, ncol = R, byrow = FALSE)
    C.ci.upper <- matrix(hi, nrow = Q, ncol = R, byrow = FALSE)
  }

  Estimate <- as.vector(C)
  SE <- as.vector(C.se)
  BSE <- if (!is.null(C.se.boot)) as.vector(C.se.boot) else rep(NA_real_, length(Estimate))
  z_value <- ifelse(SE > 0, Estimate / SE, NA_real_)

  if (C.p.side == "one.sided") {
    p_value <- ifelse(is.finite(z_value), stats::pnorm(abs(z_value), lower.tail = FALSE), NA_real_)
  } else {
    p_value <- ifelse(is.finite(z_value), 2 * stats::pnorm(abs(z_value), lower.tail = FALSE), NA_real_)
  }

  rlabs <- if (!is.null(rownames(C))) rownames(C) else paste0("Dec", 1:Q)
  clabs <- if (!is.null(colnames(C))) colnames(C) else paste0("Enc", 1:R)

  coefficients <- data.frame(
    Basis     = rep(rlabs, times = R),
    Covariate = rep(clabs, each = Q),
    Estimate  = Estimate,
    SE        = SE,
    BSE       = BSE,
    z_value   = z_value,
    p_value   = p_value,
    CI_low    = if (!is.null(C.ci.lower)) as.vector(C.ci.lower) else NA_real_,
    CI_high   = if (!is.null(C.ci.upper)) as.vector(C.ci.upper) else NA_real_,
    row.names = NULL, stringsAsFactors = FALSE
  )

  if (print.trace) {
    message(sprintf("  Signed-Bottleneck NMF-AE inference: sandwich%s, p-value=%s",
                    if (isTRUE(wild.bootstrap)) " + wild bootstrap" else "",
                    C.p.side))
  }

  object$sigma2.used  <- sigma2.used
  object$C.se         <- C.se
  object$C.se.boot    <- C.se.boot
  object$C.ci.lower   <- C.ci.lower
  object$C.ci.upper   <- C.ci.upper
  object$coefficients <- coefficients
  object$C.p.side     <- C.p.side
  class(object) <- c("nmfae.signed.inference", "nmfae.inference",
                     "nmfae.signed", "nmfae", "nmf")
  object
}


## ==============================================================
#' @title Predict method for nmfae.signed
#' @description
#' Computes \eqn{\hat Y_1 = X_1 (C_{+} - C_{-}) X_2 Y_2^{\mathrm{new}}}.
#' Since \eqn{\Theta = C_{+} - C_{-}} is signed, predictions may contain
#' negative entries even when \eqn{Y_1 \ge 0} in training.
#'
#' @param object A fitted \code{"nmfae.signed"} object.
#' @param newY2 New input matrix (P2 x N_new).  If \code{NULL}, returns
#'   the training fitted values.
#' @param Y1 Optional reference Y1 for scatter / confusion plot.
#' @param type Output: \code{"response"} (raw signed) or \code{"class"}.
#' @param ... Unused.
#' @return A numeric matrix (\code{"response"}) or factor (\code{"class"}).
#' @seealso \code{\link{nmfae.signed}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
predict.nmfae.signed <- function(object, newY2 = NULL, Y1 = NULL,
                               type = c("response", "class"), ...) {
  type <- match.arg(type)
  if (is.null(newY2)) {
    pred_mat <- object$Y1hat
  } else {
    pred_mat <- object$X1 %*% object$C %*% object$X2 %*% as.matrix(newY2)
  }

  if (type == "response") {
    result <- pred_mat
    if (!is.null(Y1)) attr(result, "Y1") <- Y1
    class(result) <- c("predict.nmfae", class(result))
  } else {
    ## For class prediction, clip negatives then take argmax (probabilities)
    clipped <- pmax(pred_mat, 0)
    labels <- rownames(pred_mat)
    if (is.null(labels)) labels <- paste0("C", seq_len(nrow(clipped)))
    pred_class <- factor(labels[apply(clipped, 2, which.max)], levels = labels)
    result <- pred_class
    if (!is.null(Y1)) {
      Y1m <- as.matrix(Y1)
      act_labels <- rownames(Y1m); if (is.null(act_labels)) act_labels <- labels
      actual <- factor(act_labels[apply(Y1m, 2, which.max)], levels = act_labels)
      attr(result, "actual") <- actual
    }
    class(result) <- c("predict.nmfae", class(result))
    attr(result, "type") <- "class"
  }
  result
}


## ==============================================================
#' @title Plot method for nmfae.signed (convergence)
#' @description
#' Displays the convergence trajectory of the objective function.
#' @param x An \code{nmfae.signed} object.
#' @param ... Additional graphical parameters.
#' @return Invisible \code{NULL}.
#' @seealso \code{\link{nmfae.signed}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
plot.nmfae.signed <- function(x, ...) {
  extra_args <- list(...)
  args <- list(x = x$objfunc.iter)
  if (is.null(extra_args$main))
    args$main <- paste0("Signed-Bottleneck NMF-AE: R2 = ", round(x$r.squared, 3))
  if (is.null(extra_args$xlab)) args$xlab <- "iter"
  if (is.null(extra_args$ylab)) args$ylab <- "objfunc"
  if (is.null(extra_args$type)) args$type <- "l"
  do.call(graphics::plot, c(args, extra_args))
  invisible(NULL)
}


## ==============================================================
#' @title Summary method for nmfae.signed
#' @description
#' Produces a summary with dimensions, convergence, fit statistics,
#' and structure diagnostics (sparsity and negative-mass ratio).
#' @param object An \code{nmfae.signed} object.
#' @param ... Unused.
#' @return An object of class \code{"summary.nmfae.signed"}.
#' @seealso \code{\link{nmfae.signed}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
summary.nmfae.signed <- function(object, ...) {
  .sparsity <- function(M, thr = 1e-4) {
    if (is.null(M)) return(NA_real_); mean(abs(M) < thr)
  }
  .range <- function(M) if (is.null(M)) c(NA_real_, NA_real_) else range(M)
  .negfrac <- function(M) {
    if (is.null(M)) return(NA_real_)
    sum(pmax(-M, 0)) / (sum(abs(M)) + 1e-16)
  }
  n_params <- prod(dim(object$X1)) + prod(dim(object$Cp)) +
              prod(dim(object$Cn)) + prod(dim(object$X2))

  ans <- list(
    call        = object$call,
    dims        = object$dims,
    Q           = unname(object$rank["Q"]),
    R           = unname(object$rank["R"]),
    n.params    = n_params,
    Y.signed    = isTRUE(object$Y.signed),
    niter       = object$niter,
    runtime     = object$runtime,
    objfunc     = object$objfunc,
    r.squared          = object$r.squared,
    r.squared.uncentered     = object$r.squared.uncentered,
    r.squared.centered = object$r.squared.centered,
    sigma       = object$sigma,
    mae         = object$mae,
    X1.sparsity = .sparsity(object$X1),
    Cp.sparsity = .sparsity(object$Cp),
    Cn.sparsity = .sparsity(object$Cn),
    C.sparsity  = .sparsity(object$C),
    X2.sparsity = .sparsity(object$X2),
    X1.range    = .range(object$X1),
    Cp.range    = .range(object$Cp),
    Cn.range    = .range(object$Cn),
    C.range     = .range(object$C),
    X2.range    = .range(object$X2),
    C.negfrac   = .negfrac(object$C),
    H.negfrac   = .negfrac(object$H)
  )
  class(ans) <- "summary.nmfae.signed"
  ans
}

#' @title Print method for summary.nmfae.signed
#' @param x A \code{"summary.nmfae.signed"} object.
#' @param digits Number of significant digits.
#' @param ... Unused.
#' @return Invisible \code{x}.
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
print.summary.nmfae.signed <- function(x,
                                     digits = max(3L, getOption("digits") - 3L),
                                     ...) {
  cat("\nCall:\n",
      paste(deparse(x$call), sep = "\n", collapse = "\n"),
      "\n\n", sep = "")
  cat("Signed-Bottleneck NMF-AE (Direct MU; signed bottleneck Theta = Cp - Cn)\n")

  cat("\nDimensions:\n")
  cat(sprintf("  P1 (output rows): %d\n", x$dims["P1"]))
  cat(sprintf("  P2 (input rows):  %d\n", x$dims["P2"]))
  cat(sprintf("  N  (samples):     %d\n", x$dims["N"]))
  cat(sprintf("  Q (decoder rank): %d\n", x$Q))
  cat(sprintf("  R (encoder rank): %d\n", x$R))
  cat(sprintf("  Parameters:       %d\n", x$n.params))
  if (x$Y.signed) cat("  Y1 signed: TRUE\n")

  cat("\nConvergence:\n")
  cat(sprintf("  Iterations:       %d\n", x$niter))
  cat(sprintf("  Runtime (secs):   %.2f\n", x$runtime))
  cat(sprintf("  Final objfunc:    %s\n", format(x$objfunc, digits = digits)))

  cat("\nGoodness of fit:\n")
  cat(sprintf("  R-squared (cor^2):    %s\n", format(x$r.squared, digits = digits)))
  if (!is.null(x$r.squared.uncentered))
    cat(sprintf("  R-squared (uncentered):     %s\n", format(x$r.squared.uncentered, digits = digits)))
  if (!is.null(x$r.squared.centered))
    cat(sprintf("  R-squared (centered): %s\n", format(x$r.squared.centered, digits = digits)))
  cat(sprintf("  Sigma (RMSE):         %s\n", format(x$sigma,     digits = digits)))
  cat(sprintf("  MAE:                  %s\n", format(x$mae,       digits = digits)))

  cat("\nStructure (range / sparsity / negative mass):\n")
  fmt_range <- function(r)
    sprintf("[%s, %s]",
            format(r[1], digits = digits, width = 1),
            format(r[2], digits = digits, width = 1))
  cat(sprintf("  X1 (P1 x Q):     range %s   sparsity %5.1f%%\n",
              fmt_range(x$X1.range), 100 * x$X1.sparsity))
  cat(sprintf("  Cp (Q x R):      range %s   sparsity %5.1f%%\n",
              fmt_range(x$Cp.range), 100 * x$Cp.sparsity))
  cat(sprintf("  Cn (Q x R):      range %s   sparsity %5.1f%%\n",
              fmt_range(x$Cn.range), 100 * x$Cn.sparsity))
  cat(sprintf("  C=Cp-Cn (Q x R): range %s   sparsity %5.1f%%   neg-mass %5.1f%%\n",
              fmt_range(x$C.range), 100 * x$C.sparsity, 100 * x$C.negfrac))
  cat(sprintf("  X2 (R x P2):     range %s   sparsity %5.1f%%\n",
              fmt_range(x$X2.range), 100 * x$X2.sparsity))
  cat(sprintf("  Encoding H (Q x N):                                   neg-mass %5.1f%%\n",
              100 * x$H.negfrac))
  cat("\n")
  invisible(x)
}


## ==============================================================
#' @title Rename Dec/Enc labels on nmfae.signed objects
#' @description
#' Replaces the default \code{"Dec1", "Dec2", ...} (decoder / X1 columns
#' and Cp/Cn/C rows) and \code{"Enc1", "Enc2", ...} (encoder / X2 rows and
#' Cp/Cn/C columns) with user-supplied labels.  Propagates to coefficients
#' tables if present (e.g., from \code{nmfae.inference}).
#'
#' @param x An \code{"nmfae.signed"} object.
#' @param X1.colnames Character vector of length Q for decoder labels.
#' @param X2.rownames Character vector of length R for encoder labels.
#' @return The renamed object (same class).
#' @seealso \code{\link{nmfae.signed}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
nmfae.signed.rename <- function(x, X1.colnames = NULL, X2.rownames = NULL) {
  if (!is.null(X1.colnames)) {
    Q <- ncol(x$X1)
    if (length(X1.colnames) != Q) stop("X1.colnames must have length ", Q)
    colnames(x$X1) <- X1.colnames
    rownames(x$Cp) <- X1.colnames
    rownames(x$Cn) <- X1.colnames
    rownames(x$C)  <- X1.colnames
    if (!is.null(x$coefficients)) {
      old <- paste0("Dec", seq_len(Q))
      for (k in seq_len(Q))
        x$coefficients$Basis[x$coefficients$Basis == old[k]] <- X1.colnames[k]
    }
  }
  if (!is.null(X2.rownames)) {
    R <- nrow(x$X2)
    if (length(X2.rownames) != R) stop("X2.rownames must have length ", R)
    rownames(x$X2) <- X2.rownames
    colnames(x$Cp) <- X2.rownames
    colnames(x$Cn) <- X2.rownames
    colnames(x$C)  <- X2.rownames
    if (!is.null(x$coefficients)) {
      old <- paste0("Enc", seq_len(R))
      for (k in seq_len(R))
        x$coefficients$Covariate[x$coefficients$Covariate == old[k]] <- X2.rownames[k]
    }
  }
  x
}


## ==============================================================
#' @title Element-wise Cross-Validation for Signed-Bottleneck NMF-AE
#' @description
#' Element-wise k-fold cross-validation for \code{\link{nmfae.signed}} to
#' select the decoder / encoder ranks \eqn{(Q, R)}.  Mirrors
#' \code{\link{nmfae.ecv}} but uses the \strong{weighted} Signed-Bottleneck NMF-AE fit
#' path (\code{Y1.weights}): test-fold elements are zero-weighted during
#' fitting, and held-out MSE is computed on those elements.
#'
#' @param Y1 Output matrix (P1 x N).
#' @param Y2 Input matrix (P2 x N).  Default \code{Y1}.
#' @param rank Integer vector of candidate Q values.  Default \code{1:2}.
#' @param rank.encoder Integer vector of candidate R values, or \code{NULL}
#'   (default: pair R = Q, diagonal grid).
#' @param ... Additional arguments:
#'   \describe{
#'     \item{\code{nfolds} / \code{div}}{Number of folds.  Default 5.}
#'     \item{\code{seed}}{RNG seed for fold assignment.  Default 123.}
#'     \item{\code{nstart}}{Number of random restarts per fit.  Default 1.
#'       Signed models have more local minima (the bottleneck can carry
#'       both signs), so \code{nstart >= 10} is recommended for
#'       reproducible rank selection.}
#'     \item{Other args}{\code{epsilon}, \code{maxit}, \code{warm.start},
#'       etc.\ are passed to \code{\link{nmfae.signed}}.}
#'   }
#' @return An object of class \code{c("nmfae.signed.ecv", "nmfae.ecv")} with
#'   \code{objfunc} (MSE per pair), \code{sigma} (RMSE), \code{objfunc.fold}
#'   (per-fold MSE), \code{folds}, \code{QR}, \code{paired}.
#' @seealso \code{\link{nmfae.signed}}, \code{\link{nmfae.ecv}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
nmfae.signed.ecv <- function(Y1, Y2 = Y1, rank = 1:2, rank.encoder = NULL, ...) {
  extra_ecv <- list(...)
  if (!is.null(extra_ecv$Q)) rank <- extra_ecv$Q
  if (!is.null(extra_ecv$R)) rank.encoder <- extra_ecv$R
  nfolds <- if (!is.null(extra_ecv$nfolds)) extra_ecv$nfolds
            else if (!is.null(extra_ecv$div)) extra_ecv$div else 5
  seed   <- if (!is.null(extra_ecv$seed)) extra_ecv$seed else 123
  Q <- rank; R <- rank.encoder
  div <- nfolds

  Y1 <- as.matrix(Y1); Y2 <- as.matrix(Y2)
  P1 <- nrow(Y1); N <- ncol(Y1)

  if (is.null(R)) {
    QR <- data.frame(Q = Q, R = Q)
  } else {
    QR <- expand.grid(Q = Q, R = R)
  }
  num_pairs <- nrow(QR)

  ## Element-wise folds (shared helper)
  folds <- .ecv.make.folds(Y1, div, seed)

  pair_labels <- sprintf("Q=%d,R=%d", QR$Q, QR$R)
  has_na <- any(is.na(Y1))
  message(sprintf("Signed-Bottleneck NMF-AE ECV: %d (Q,R) pairs, %d-fold, %d tasks...",
                  num_pairs, div, num_pairs * div))

  ## Strip ecv-specific args before passing the rest to nmfae.signed
  fit_args <- extra_ecv
  fit_args$nfolds <- NULL; fit_args$div <- NULL
  fit_args$Q <- NULL; fit_args$R <- NULL

  run_one <- function(i, k) {
    test_idx <- folds[[k]]
    W_train <- matrix(1, nrow = P1, ncol = N)
    if (has_na) W_train[is.na(Y1)] <- 0
    W_train[test_idx] <- 0
    fit <- suppressMessages(
      do.call(nmfae.signed,
              c(list(Y1 = Y1, Y2 = Y2, rank = QR$Q[i], rank.encoder = QR$R[i],
                     Y1.weights = W_train), fit_args))
    )
    mean((Y1[test_idx] - fit$Y1hat[test_idx])^2)
  }

  cv <- .ecv.run(pair_labels, div, run_one,
                 progress = function(i, o, s)
                   message(sprintf("  Q=%d, R=%d: MSE=%.6f, sigma=%.4f",
                                   QR$Q[i], QR$R[i], o, s)))

  result <- list(objfunc = cv$objfunc,
                 sigma = cv$sigma,
                 objfunc.fold = cv$objfunc.fold,
                 folds = folds,
                 QR = QR,
                 paired = is.null(R))
  class(result) <- c("nmfae.signed.ecv", "nmfae.ecv")
  result
}


#' @title Rank selection for nmfae.signed (paired rank, concise diagnostics)
#' @description
#' Fits \code{\link{nmfae.signed}} with a \strong{paired} decoder/encoder
#' rank (\eqn{Q = R}) across a range of ranks and reports
#' \code{r.squared}, the effective rank (of the latent encoding \eqn{H}),
#' and the element-wise CV error \code{sigma.ecv}, with the same concise
#' plot as \code{\link{nmfkc.rank}}.  For a full \eqn{(Q, R)} grid use
#' \code{\link{nmfae.signed.ecv}}.
#' @param Y1 Endogenous matrix (\eqn{P_1 \times N}); may be signed.
#' @param Y2 Exogenous matrix; defaults to \code{Y1}.
#' @param rank Integer vector of (paired) ranks to evaluate.
#' @param plot Logical; draw the diagnostics plot (default \code{TRUE}).
#' @param detail \code{"full"} (default) also runs element-wise CV
#'   (\code{sigma.ecv}); \code{"fast"} skips it (plots r.squared and
#'   eff.rank only, and recommends the R-squared elbow).
#' @param ... Passed on to \code{\link{nmfae.signed}} and
#'   \code{\link{nmfae.signed.ecv}}.
#' @return A list with \code{rank.best} and \code{criteria}
#'   (\code{rank}, \code{effective.rank}, \code{effective.rank.ratio},
#'   \code{r.squared}, \code{sigma.ecv}).
#' @seealso \code{\link{nmfae.signed}}, \code{\link{nmfae.signed.ecv}},
#'   \code{\link{nmfkc.rank}}
#' @references
#' Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
#' effective dimensionality.  \emph{Proc. EUSIPCO}, 606--610.
#' (\code{effective.rank})
#' Wold, S. (1978).  Cross-validatory estimation of the number of
#' components in factor and principal components models.
#' \emph{Technometrics}, 20(4), 397--405. (\code{sigma.ecv})
#' @export
nmfae.signed.rank <- function(Y1, Y2 = Y1, rank = 1:5, detail = c("full", "fast"),
                              plot = TRUE, ...) {
  detail <- match.arg(detail)
  Y1 <- as.matrix(Y1); Y2 <- as.matrix(Y2)
  rs <- numeric(length(rank)); er <- numeric(length(rank))
  for (i in seq_along(rank)) {
    f <- suppressMessages(nmfae.signed(Y1, Y2, rank = rank[i],
                                       rank.encoder = rank[i],
                                       print.trace = FALSE, ...))
    rs[i] <- f$r.squared
    er[i] <- .effective.rank(f$H)
  }
  ecv <- if (detail == "full")
    suppressMessages(nmfae.signed.ecv(Y1, Y2, rank = rank, ...))$sigma
    else rep(NA_real_, length(rank))
  criteria <- data.frame(rank = rank, effective.rank = er,
                         effective.rank.ratio = er / rank,
                         r.squared = rs, sigma.ecv = as.numeric(ecv))
  .rank.finish(criteria, plot = plot,
               main = "nmfae.signed rank selection (paired Q=R)")
}


## ==============================================================
#' @title Summary method for nmfae.signed.inference objects
#' @description
#' Produces a summary of a fitted Signed-Bottleneck NMF-AE model with
#' inference results.  Extends \code{\link{summary.nmfae.signed}} by
#' attaching the \code{coefficients} table and p-value side from
#' \code{\link{nmfae.signed.inference}}.
#'
#' @param object An object of class \code{"nmfae.signed.inference"}.
#' @param ... Additional arguments (currently unused).
#' @return An object of class \code{"summary.nmfae.signed.inference"}.
#' @seealso \code{\link{nmfae.signed.inference}}, \code{\link{summary.nmfae.signed}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
summary.nmfae.signed.inference <- function(object, ...) {
  ans <- summary.nmfae.signed(object, ...)
  ans$coefficients <- object$coefficients
  ans$C.p.side     <- object$C.p.side
  ans$sigma2.used  <- object$sigma2.used
  class(ans) <- c("summary.nmfae.signed.inference",
                   "summary.nmfae.signed", "summary.nmfae.inference")
  ans
}

#' @title Print method for summary.nmfae.signed.inference objects
#' @description
#' Prints the Signed-Bottleneck NMF-AE summary followed by the
#' coefficients table of Theta.
#' @param x An object of class \code{"summary.nmfae.signed.inference"}.
#' @param digits Minimum number of significant digits.
#' @param ... Additional arguments (currently unused).
#' @return Called for its side effect (printing). Returns \code{x} invisibly.
#' @seealso \code{\link{summary.nmfae.signed.inference}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
print.summary.nmfae.signed.inference <- function(x,
    digits = max(3L, getOption("digits") - 3L), ...) {
  print.summary.nmfae.signed(x, digits = digits, ...)

  if (!is.null(x$coefficients) && is.data.frame(x$coefficients)) {
    cf <- x$coefficients
    p_side <- if (!is.null(x$C.p.side)) x$C.p.side else "two.sided"
    p_header <- if (p_side == "one.sided") "Pr(>z)" else "Pr(>|z|)"

    sig_stars <- function(p) {
      ifelse(!is.finite(p), " ",
        ifelse(p < 0.001, "***",
          ifelse(p < 0.01, "**",
            ifelse(p < 0.05, "*",
              ifelse(p < 0.1, ".", " ")))))
    }
    format_pval <- function(p) {
      ifelse(!is.finite(p), "      NA",
        ifelse(p < 2.2e-16, "  <2e-16",
          formatC(p, format = "g", digits = 4, width = 8)))
    }

    n_total <- nrow(cf)
    n_sig <- sum(cf$p_value < 0.05, na.rm = TRUE)
    cat(sprintf("\nTheta coefficients: %d total, %d significant (p < 0.05)\n",
                n_total, n_sig))
    rnames <- paste0(cf$Basis, ":", cf$Covariate)
    est <- formatC(cf$Estimate, format = "f", digits = 3, width = 9)
    se  <- formatC(cf$SE,       format = "f", digits = 3, width = 10)
    bse <- formatC(cf$BSE,      format = "f", digits = 3, width = 6)
    zv  <- formatC(cf$z_value,  format = "f", digits = 2, width = 7)
    pv_str <- format_pval(cf$p_value)
    stars <- sig_stars(cf$p_value)
    max_lw <- max(nchar(rnames))
    hdr <- sprintf("%s %s %s %s %s",
                   formatC("Estimate",   width = 9),
                   formatC("Std. Error", width = 10),
                   formatC("(Boot)",     width = 6),
                   formatC("z value",    width = 7),
                   formatC(p_header,     width = 8))
    cat(sprintf("%s %s\n", formatC("Dec:Enc", width = max_lw), hdr))
    for (i in seq_len(n_total)) {
      cat(sprintf("%s %s %s %s %s %s %s\n",
                  formatC(rnames[i], width = max_lw),
                  est[i], se[i], bse[i], zv[i], pv_str[i], stars[i]))
    }
    cat("---\n")
    cat("Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1\n")
  }
  cat("\n")
  invisible(x)
}


## ==============================================================
#' @title Heatmap visualization of nmfae.signed factor matrices
#' @description
#' Displays the factor blocks of a \code{\link{nmfae.signed}} fit as
#' side-by-side heatmaps.  Non-negative blocks (\eqn{X_1, C_{+}, C_{-},
#' X_2}) use the white-orange-red palette; the signed combined
#' bottleneck \eqn{C = C_{+} - C_{-}} is rendered with a diverging
#' blue-white-red palette so positive and negative weights are visually
#' distinguishable.
#'
#' @param x An object of class \code{"nmfae.signed"}.
#' @param Y1.label Character vector for rows of \eqn{X_1}.
#' @param X1.label Decoder basis labels.
#' @param X2.label Encoder basis labels.
#' @param Y2.label Input variable labels.
#' @param palette.pos Palette for non-negative blocks. Default white-orange-red.
#' @param palette.signed Palette for signed \eqn{C}. Default blue-white-red.
#' @param show.C Logical. If \code{TRUE} (default), shows the combined
#'   signed \eqn{C = C_{+} - C_{-}} as a separate panel.
#' @param ... Not used.
#'
#' @return Invisible \code{NULL}. Called for its side effect (plot).
#' @seealso \code{\link{nmfae.signed}}, \code{\link{nmfae.heatmap}}
#' @examples
#' \donttest{
#' set.seed(1)
#' Y1 <- matrix(abs(rnorm(12)), 3, 4)
#' Y2 <- matrix(abs(rnorm(20)), 5, 4)
#' res <- nmfae.signed(Y1, Y2, rank = 2, rank.encoder = 2, maxit = 200)
#' nmfae.signed.heatmap(res)
#' }
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions.
#'
#' @references
#' Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
#' semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
#' Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
#'
#' @export
nmfae.signed.heatmap <- function(x,
                                  Y1.label = NULL, X1.label = NULL,
                                  X2.label = NULL, Y2.label = NULL,
                                  palette.pos = NULL,
                                  palette.signed = NULL,
                                  show.C = TRUE, ...) {
  if (!inherits(x, "nmfae.signed"))
    stop("x must be of class 'nmfae.signed'.")
  if (is.null(palette.pos))
    palette.pos <- grDevices::colorRampPalette(c("white", "orange", "red"))(64)
  if (is.null(palette.signed))
    palette.signed <- grDevices::colorRampPalette(
      c("blue", "white", "red"))(64)

  X1 <- x$X1; Cp <- x$Cp; Cn <- x$Cn; X2 <- x$X2
  C  <- if (!is.null(x$C)) x$C else (Cp - Cn)
  Q  <- ncol(X1); R  <- nrow(X2)
  P1 <- nrow(X1); P2 <- ncol(X2)

  if (is.null(Y1.label)) Y1.label <- rownames(X1)
  if (is.null(X1.label)) X1.label <- colnames(X1)
  if (is.null(X2.label)) X2.label <- rownames(X2)
  if (is.null(Y2.label)) Y2.label <- colnames(X2)
  if (is.null(Y1.label)) Y1.label <- as.character(seq_len(P1))
  if (is.null(X1.label)) X1.label <- paste0("Dec", seq_len(Q))
  if (is.null(X2.label)) X2.label <- paste0("Enc", seq_len(R))
  if (is.null(Y2.label)) Y2.label <- as.character(seq_len(P2))

  .heat <- function(mat, title, pal, zlim = NULL) {
    m <- mat[nrow(mat):1, , drop = FALSE]
    if (is.null(zlim)) zlim <- range(m, finite = TRUE)
    graphics::image(x = seq_len(ncol(m)), y = seq_len(nrow(m)),
                    z = t(m), col = pal, zlim = zlim, axes = FALSE,
                    xlab = "", ylab = "", main = title)
    graphics::axis(1, at = seq_len(ncol(m)), labels = colnames(m),
                   las = 2, cex.axis = 0.7)
    graphics::axis(2, at = seq_len(nrow(m)), labels = rev(rownames(m)),
                   las = 1, cex.axis = 0.7)
    graphics::box()
  }

  n.panels <- if (isTRUE(show.C)) 5 else 4
  old.par <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old.par), add = TRUE)
  graphics::par(mfrow = c(1, n.panels), mar = c(4, 3, 2, 1))

  rownames(X1) <- Y1.label; colnames(X1) <- X1.label
  rownames(Cp) <- X1.label; colnames(Cp) <- X2.label
  rownames(Cn) <- X1.label; colnames(Cn) <- X2.label
  rownames(X2) <- X2.label; colnames(X2) <- Y2.label
  rownames(C)  <- X1.label; colnames(C)  <- X2.label

  .heat(X1, expression(X[1]), palette.pos)
  .heat(Cp, expression(C["+"]), palette.pos)
  .heat(Cn, expression(C["-"]), palette.pos)
  if (isTRUE(show.C)) {
    cmax <- max(abs(C), na.rm = TRUE)
    .heat(C, expression(C == C["+"] - C["-"]),
          palette.signed, zlim = c(-cmax, cmax))
  }
  .heat(X2, expression(X[2]), palette.pos)

  invisible(NULL)
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----load-data----------------------------------------------------------------
library(nmfkc)
library(nlme)

data(Orthodont)
head(Orthodont)

## ----prepare-Y----------------------------------------------------------------
Y <- matrix(Orthodont$distance, nrow = 4, ncol = 27)
colnames(Y) <- paste0("S", 1:27)
rownames(Y) <- paste("Age", c(8, 10, 12, 14))
Y[, 1:6]

## ----prepare-A----------------------------------------------------------------
male <- ifelse(Orthodont$Sex[seq(1, 108, 4)] == "Male", 1, 0)
A <- rbind(intercept = 1, male = male)
A[, 1:6]

## ----dfU-scan-----------------------------------------------------------------
scan_result <- nmfre.dfU.scan(1:10/10, Y, A, rank = 1)
scan_result

## ----fit-nmfre----------------------------------------------------------------
res <- nmfre(Y, A, rank = 1, df.rate = 0.2, prefix = "Trend")

## ----summary------------------------------------------------------------------
summary(res)

## ----plot-growth, fig.height=6------------------------------------------------
age <- c(8, 10, 12, 14)

plot(age, res$XB[, 1], type = "n", ylim = range(Y),
     xlab = "Age (years)", ylab = "Distance (mm)",
     main = "Orthodont: NMF-RE Growth Curves")

# Plot observed data points
for (j in 1:27) {
  pch_j <- ifelse(male[j] == 1, 4, 1)
  points(age, Y[, j], pch = pch_j, col = "gray60")
}

# Plot individual predictions (fixed + random effects)
for (j in 1:27) {
  lines(age, res$XB.blup[, j], col = "steelblue", lty = 3, lwd = 0.8)
}

# Plot population-level fixed effects (two lines: male and female)
for (j in 1:27) {
  lines(age, res$XB[, j], col = "red", lwd = 2)
}

legend("topleft",
  legend = c("Fixed effect (male/female)", "Fixed + Random (BLUP)",
             "Male (observed)", "Female (observed)"),
  lwd = c(2, 1, NA, NA), lty = c(1, 3, NA, NA),
  pch = c(NA, NA, 4, 1),
  col = c("red", "steelblue", "gray60", "gray60"),
  cex = 0.85)

## ----basis-X------------------------------------------------------------------
cat("Basis X (temporal pattern):\n")
print(round(res$X, 4))

## ----coef-C-------------------------------------------------------------------
cat("Coefficient matrix (Theta):\n")
print(round(res$C, 4))

## ----random-effects, fig.height=4---------------------------------------------
barplot(res$U[1, ], names.arg = colnames(Y),
        las = 2, cex.names = 0.7,
        col = ifelse(male == 1, "steelblue", "salmon"),
        main = "Random Effects (U) by Subject",
        ylab = "Random effect value")
legend("topright", fill = c("steelblue", "salmon"),
       legend = c("Male", "Female"), cex = 0.85)

## ----convergence--------------------------------------------------------------
cat("Converged:", res$converged, "\n")
cat("Iterations:", res$iter, "\n")
cat("Stop reason:", res$stop.reason, "\n")

## ----plot-convergence, fig.height=4-------------------------------------------
plot(res$objfunc.iter, type = "l",
     xlab = "Iteration", ylab = "Objective function",
     main = "Convergence Trace")

## ----residual-analysis, fig.height=4------------------------------------------
residuals <- Y - res$XB.blup
cat("Mean absolute residual (BLUP):", round(mean(abs(residuals)), 4), "\n")
cat("Mean absolute residual (fixed):", round(mean(abs(Y - res$XB)), 4), "\n")

# Fitted vs Observed
plot(as.vector(Y), as.vector(res$XB.blup),
     xlab = "Observed", ylab = "Fitted (BLUP)",
     main = "Observed vs Fitted", pch = 16, col = "steelblue")
abline(0, 1, col = "red", lwd = 2)

## ----compare-models-----------------------------------------------------------
# Standard NMF with covariates (no random effects)
res_fixed <- nmfkc(Y, A = A, rank = 1)

cat("=== Standard NMF (fixed effects only) ===\n")
cat("R-squared:", round(1 - sum((Y - res_fixed$XB)^2) / sum((Y - mean(Y))^2), 4), "\n\n")

cat("=== NMF-RE (fixed + random effects) ===\n")
cat("R-squared (XB):      ", round(res$r.squared.fixed, 4), "\n")
cat("R-squared (XB+blup): ", round(res$r.squared, 4), "\n")
cat("ICC:                 ", round(res$ICC, 4), "\n")

## ----separated-inference------------------------------------------------------
# Fit without bootstrap (fast)
res_fast <- nmfre(Y, A, rank = 1, df.rate = 0.2, prefix = "Trend",
                  wild.bootstrap = FALSE)

# Run inference separately
res_inf <- nmfre.inference(res_fast, Y, A, wild.B = 500)
res_inf$coefficients[, c("Basis", "Covariate", "Estimate", "SE", "p_value")]

## ----dot-visualization, eval=requireNamespace("DiagrammeR", quietly=TRUE)-----
dot <- nmfkc.DOT(res_inf, type = "YXA", sig.level = 0.05)
plot(dot)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----lib----------------------------------------------------------------------
library(nmfkc)

## ----data, fig.width = 6, fig.height = 5.2------------------------------------
set.seed(1)
K <- 3; n_each <- 20; N <- K * n_each       # 60 nodes, 3 communities
block <- rep(1:K, each = n_each)            # true community labels

p_in <- 0.6; p_out <- 0.05
Prob <- matrix(p_out, N, N)
for (k in 1:K) Prob[block == k, block == k] <- p_in
Y <- matrix(rbinom(N * N, 1, Prob), N, N)   # 0/1 adjacency
Y[lower.tri(Y)] <- t(Y)[lower.tri(Y)]        # make symmetric
diag(Y) <- 0

# add three bridge nodes, each also linked to a second community
bridge <- c(20, 40, 60); into <- c(2, 3, 1)
for (i in seq_along(bridge)) {
  b <- bridge[i]; tgt <- which(block == into[i])
  e <- rbinom(length(tgt), 1, 0.45)
  Y[b, tgt] <- pmax(Y[b, tgt], e); Y[tgt, b] <- Y[b, tgt]
}

isSymmetric(Y); sum(Y) / 2                   # symmetric, number of edges

# adjacency matrix in the original (random) node order -- structure is hidden
image(1:N, 1:N, Y[, N:1], col = c("white", "steelblue"),
      xlab = "node", ylab = "node", main = "Adjacency (original order)")

## ----fit----------------------------------------------------------------------
res <- nmfkc.net(Y, rank = 3, type = "bi", nstart = 10, maxit = 500)
res$r.squared                 # fit
head(round(res$X.prob, 2))    # soft membership: rows = nodes, columns = communities
head(res$X.cluster)           # hard label = argmax membership

## ----compare------------------------------------------------------------------
table(true = block, estimated = res$X.cluster)

# the three bridge nodes have genuinely split memberships
round(res$X.prob[bridge, ] * 100, 1)

## ----viz, fig.width = 8, fig.height = 4---------------------------------------
ord <- order(res$X.cluster)
op <- par(mfrow = c(1, 2), mar = c(4, 4, 3, 1))
image(1:N, 1:N, Y[ord, rev(ord)], col = c("white", "steelblue"),
      xlab = "node (reordered)", ylab = "", main = "Adjacency by community")
image(1:N, 1:K, res$X.prob[ord, , drop = FALSE],
      col = hcl.colors(20, "Blues", rev = TRUE),
      xlab = "node (reordered)", ylab = "community", axes = FALSE,
      main = "Soft membership")
axis(1); axis(2, at = 1:K)
par(op)

## ----dot-check, include = FALSE-----------------------------------------------
has_dg <- requireNamespace("DiagrammeR", quietly = TRUE)

## ----dot-soft, eval = has_dg, fig.width = 7, fig.height = 7-------------------
dot_soft <- nmfkc.net.DOT(res,
  Y.label = as.character(1:N), X.label = paste("Community", 1:K),
  Y.title = "Network nodes",   X.title = "Communities",
  layout  = "neato", threshold = 0.25, Y.cluster = "soft")
DiagrammeR::grViz(as.character(dot_soft))

## ----dot-hard, eval = has_dg, fig.width = 7, fig.height = 7-------------------
dot_hard <- nmfkc.net.DOT(res,
  Y.label = as.character(1:N), X.label = paste("Community", 1:K),
  Y.title = "Network nodes",   X.title = "Communities",
  layout  = "neato", threshold = 0.25, Y.cluster = "hard")
DiagrammeR::grViz(as.character(dot_hard))

## ----rank, fig.width = 7, fig.height = 6--------------------------------------
rk <- nmfkc.net.rank(Y, rank = 1:6, type = "bi", nstart = 5)
rk$rank.best
round(rk$criteria, 3)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----load-packages------------------------------------------------------------
library(nmfkc)
library(vars) # For Canada dataset

## ----air-data-prep------------------------------------------------------------
# Load and transform the ts object
d_air <- AirPassengers
d_air_log <- log10(d_air) # Still a ts object

## ----air-degree-cv------------------------------------------------------------
# Evaluate lag orders from 1 to 14
# Note: ts objects are automatically transposed to (Variables x Time) internally
cv_res <- nmfkc.ar.degree.cv(d_air_log, rank = 1, degree = 1:14, epsilon=1e-6, maxit=500000)

# Check the optimal degree
cv_res$degree

# For this example, we will proceed with D=12 (capturing monthly seasonality)
D <- 12

## ----air-model-fit------------------------------------------------------------
# Create matrices for the AR(12) model
a_air <- nmfkc.ar(d_air_log, degree = D, intercept = TRUE)

# Fit the NMF-AR model (Rank=1 for univariate)
res_air <- nmfkc(Y = a_air$Y, A = a_air$A, rank = 1, epsilon = 1e-6, maxit=500000)

# Check goodness of fit
res_air$r.squared

# Check for stationarity (spectral radius < 1)
nmfkc.ar.stationarity(res_air)

## ----air-forecast-------------------------------------------------------------
# Forecast next 2 years (24 months)
h <- 24
pred_res <- nmfkc.ar.predict(x = res_air, Y = a_air$Y, n.ahead = h)

# Convert predictions back to original scale
pred_val <- 10^as.vector(pred_res$pred)
pred_time <- pred_res$time # Future time points generated by the function

# --- Plotting ---
# Setup plot range
xlim_range <- range(c(time(d_air), pred_time))
ylim_range <- range(c(d_air, pred_val))

# 1. Observed data (Black)
plot(d_air, type = "l", col = "black", 
     xlim = xlim_range, ylim = ylim_range, lwd = 1,
     xlab = "Year", ylab = "Air Passengers", main = "NMF-VAR Forecast (h=24)")

# 2. Fitted values during training (Red)
# a_air$Y has column names as time strings; we parse them for plotting
fitted_time <- as.numeric(colnames(res_air$XB))
lines(fitted_time, 10^as.vector(res_air$XB), col = "red", lwd = 2)

# 3. Forecast (Blue)
# Connect the last observed point to the first forecast for a continuous line
last_t <- tail(as.numeric(time(d_air)), 1)
last_y <- tail(as.vector(d_air), 1)
lines(c(last_t, pred_time), c(last_y, pred_val), col = "blue", lwd = 2, lty = 2)

# Add legend
legend("topleft", legend = c("Observed", "Fitted", "Forecast"),
       col = c("black", "red", "blue"), lty = c(1, 1, 2), lwd = 2)

## ----canada-data-prep---------------------------------------------------------
# Load, difference, and normalize
d0_canada <- Canada
dd_canada <- apply(d0_canada, 2, diff) # Returns a matrix (Time x Vars)
dn_canada <- nmfkc.normalize(dd_canada)

# Transpose to (Variables x Time) for NMF
Y0_canada <- t(dn_canada)

# Create matrices for VAR(1)
a_canada <- nmfkc.ar(Y0_canada, degree = 1, intercept = TRUE)

## ----canada-model-fit---------------------------------------------------------
# Fit the NMF-VAR model
res_canada <- nmfkc(Y = a_canada$Y, A = a_canada$A, rank = 2, prefix = "Condition", epsilon = 1e-6)

# R-squared and Stationarity
res_canada$r.squared
nmfkc.ar.stationarity(res_canada)

## ----canada-soft-cluster------------------------------------------------------
# Visualize soft clustering of time trends
barplot(res_canada$B.prob, col = c(2, 3), border = NA,
        main = "Soft Clustering of Economic Conditions",
        xlab = "Time", ylab = "Probability",
        names.arg = colnames(a_canada$Y), las=2, cex.names = 0.5)
legend("topright", legend = colnames(res_canada$X), fill = c(2, 3), bg = "white")

## ----canada-dot---------------------------------------------------------------
# Generate DOT script for graph visualization
dot_script <- nmfkc.ar.DOT(res_canada, intercept = TRUE, threshold=0.01)

# plot(dot_script)  # requires DiagrammeR

## ============================================================
## nmfkc.bicv(): Bi-cross-validation for NMF rank selection
## (Owen & Perry 2009, AoAS 3(2):564-594).  PROTOTYPE / experimental.
##
## Idea: split rows into folds and columns into folds.  For a held-out
## row-block I and column-block J, the matrix is partitioned into
##   [ A B ]   A = Y[I, J]   (held-out block, predicted)
##   [ C D ]   B = Y[I, Jc]  C = Y[Ic, J]  D = Y[Ic, Jc]  (retained)
## Fit NMF only on D (D ~ L_D R_D), then "fold in" the held-out rows and
## columns by NON-NEGATIVE REGRESSION onto the fixed D-factors:
##   L_I = argmin_{>=0} ||B - L_I R_D||   (held rows' loadings)
##   R_J = argmin_{>=0} ||C - L_D R_J||   (held cols' scores)
## and predict A_hat = L_I R_J.  The held-out rows AND columns never
## enter the NMF fit, so there is no information leakage.
## ============================================================

#' @title Non-negative least squares by multiplicative updates (Internal)
#' @description Solve \eqn{\min_{L\ge 0}\|M - L F\|^2_F} with \eqn{F}
#'   fixed (\code{side = "left"}), or \eqn{\min_{R\ge 0}\|M - F R\|^2_F}
#'   with \eqn{F} fixed (\code{side = "right"}), via a few multiplicative
#'   updates.  Used by \code{\link{nmfkc.bicv}} to fold held-out rows /
#'   columns onto fixed factors.
#' @param M Target matrix.
#' @param F Fixed factor.
#' @param side \code{"left"} solves for \eqn{L} (\code{nrow(M) x k});
#'   \code{"right"} solves for \eqn{R} (\code{k x ncol(M)}).
#' @param maxit Number of multiplicative updates.
#' @return The non-negative factor.
#' @keywords internal
#' @noRd
.nnls.mu <- function(M, F, side = c("left", "right"), maxit = 100) {
  side <- base::match.arg(side)
  eps <- 1e-10
  if (side == "left") {
    k <- base::nrow(F)
    L <- base::matrix(base::mean(M) / (k * base::mean(F) + eps) + eps,
                      base::nrow(M), k)
    FFt <- F %*% base::t(F); MFt <- M %*% base::t(F)
    for (it in base::seq_len(maxit)) L <- L * MFt / (L %*% FFt + eps)
    L
  } else {
    k <- base::ncol(F)
    R <- base::matrix(base::mean(M) / (k * base::mean(F) + eps) + eps,
                      k, base::ncol(M))
    FtF <- base::t(F) %*% F; FtM <- base::t(F) %*% M
    for (it in base::seq_len(maxit)) R <- R * FtM / (FtF %*% R + eps)
    R
  }
}


#' @title Bi-cross-validation for NMF rank selection
#' @description
#' Owen & Perry's (2009) bi-cross-validation (BCV) for choosing the NMF
#' rank.  A lightweight CV engine in the spirit of \code{\link{nmfkc.ecv}}:
#' it returns the held-out error per rank and nothing more (no plot, no
#' table) -- pass the result to \code{which.min(sigma)}, or build your own
#' diagnostics.
#'
#' Unlike the element-wise CV of \code{\link{nmfkc.ecv}} (which holds out
#' scattered \emph{entries} and refits with weights), BCV holds out a whole
#' \strong{row-block and column-block} simultaneously: the model is fit only
#' on the retained block \eqn{D}, and the held-out block \eqn{A} is predicted
#' by folding the held-out rows/columns onto the fixed \eqn{D}-factors via
#' non-negative regression (\eqn{\hat A = L_I R_J}).  Because the held-out
#' rows and columns never enter the fit, there is no information leakage.
#' Covariates are ignored (plain NMF).  The recommended setting is to leave
#' out roughly half the rows and half the columns (\code{nfolds = 2}).
#'
#' @param Y Observation matrix (\eqn{P \times N}), non-negative.
#' @param rank Integer vector of ranks to evaluate.
#' @details Each fold keeps about \eqn{(1 - 1/\text{nfolds})} of the rows
#'   and columns, so the retained block \eqn{D} must have more than
#'   \code{rank} rows \strong{and} columns.  The largest testable rank is
#'   therefore about \eqn{(1 - 1/\text{nfolds})\min(P, N) - 1}; with
#'   \code{nfolds = 2} this is roughly \eqn{\min(P, N)/2 - 1}.  Ranks above
#'   this return \code{NA} and trigger a \code{\link[base]{warning}} that
#'   names the limit and the \code{nfolds} (or \code{\link{nmfkc.ecv}})
#'   that would reach the requested ranks.  Raising \code{nfolds} lifts the
#'   limit at the cost of a smaller hold-out and more compute
#'   (\eqn{(\text{nfolds} - 1)^2} full fits per rank).
#' @param ... Advanced options, rarely needed (defaults in parentheses):
#'   \code{nfolds} (\code{2}), the number of row and column folds (the grid is
#'   \code{nfolds x nfolds}; \code{2} leaves out half the rows / columns, Owen
#'   & Perry's recommendation); \code{seed} (\code{123}, fold-assignment seed);
#'   and \code{nnls.maxit} (\code{100}, multiplicative-update iterations for the
#'   fold-in non-negative regressions).  Any other arguments are passed to
#'   \code{\link{nmfkc}} for the per-block fits (e.g.\ \code{maxit}).
#' @return A list (cf.\ \code{\link{nmfkc.ecv}}) with:
#'   \item{objfunc}{Held-out mean squared error for each rank.}
#'   \item{sigma}{Its square root (RMSE) for each rank.}
#'   \item{rank}{The evaluated rank vector.}
#'   \item{nfolds}{The number of folds used.}
#' @references A. B. Owen and P. O. Perry (2009).  Bi-cross-validation of
#'   the SVD and the nonnegative matrix factorization.
#'   \emph{Ann. Appl. Stat.} 3(2):564--594. \doi{10.1214/08-AOAS227}.
#' @seealso \code{\link{nmfkc.rank}}, \code{\link{nmfkc.ecv}},
#'   \code{\link{nmfkc.consensus}}, \code{\link{nmfkc.ard}}
#' @export
#' @examples
#' \donttest{
#' ## rank-3 non-negative data; bi-CV needs enough kept rows/cols per
#' ## fold (> rank), so use a matrix with ample dimensions.
#' set.seed(1)
#' X <- matrix(abs(rnorm(30 * 3)), 30, 3)
#' B <- matrix(abs(rnorm(3 * 40)), 3, 40)
#' bv <- nmfkc.bicv(X %*% B, rank = 1:6)   # nfolds = 2 (Owen & Perry) by default
#' bv$sigma                  # held-out RMSE per rank
#' bv$rank[which.min(bv$sigma)]
#' }
nmfkc.bicv <- function(Y, rank = 1:3, ...) {
  Y <- base::as.matrix(Y)
  ## Advanced options live in `...` (the rest passes to nmfkc per block).
  ## nfolds = 2 is Owen & Perry's recommendation and rarely changed.
  dots       <- base::list(...)
  nfolds     <- if (base::is.null(dots$nfolds))     2   else dots$nfolds
  seed       <- if (base::is.null(dots$seed))       123 else dots$seed
  nnls.maxit <- if (base::is.null(dots$nnls.maxit)) 100 else dots$nnls.maxit
  dots$nfolds <- NULL; dots$seed <- NULL; dots$nnls.maxit <- NULL
  P <- base::nrow(Y); N <- base::ncol(Y)
  if (!base::is.null(seed)) base::set.seed(seed)
  row.fold <- base::sample(base::rep_len(1:nfolds, P))
  col.fold <- base::sample(base::rep_len(1:nfolds, N))

  ## Feasibility: the retained block D has min(P - largest row fold,
  ## N - largest col fold) rows/cols, and rank-k needs strictly more than
  ## k of each.  Warn (rather than silently return NA) for ranks the chosen
  ## nfolds cannot reach, and suggest the nfolds that would.
  kept.row <- P - base::max(base::table(row.fold))
  kept.col <- N - base::max(base::table(col.fold))
  rmax <- base::min(kept.row, kept.col) - 1L
  if (base::any(rank > rmax)) {
    g.need <- NA_integer_
    for (g in base::seq.int(nfolds + 1L, base::min(P, N))) {
      kr <- P - base::ceiling(P / g); kc <- N - base::ceiling(N / g)
      if (base::min(kr, kc) - 1L >= base::max(rank)) { g.need <- g; break }
    }
    hint <- if (base::is.na(g.need))
      "no nfolds can reach it (rank exceeds min(P, N) - 1); use a smaller rank"
      else base::sprintf("use nfolds = %d (smaller hold-out, more compute), or nmfkc.ecv", g.need)
    base::warning(base::sprintf(
      "nmfkc.bicv: ranks > %d are infeasible with nfolds = %d (retained block ~%d x %d; rank-k needs > k rows and cols) and return NA. To test up to rank %d, %s.",
      rmax, nfolds, kept.row, kept.col, base::max(rank), hint))
  }

  ## nmfkc args for the (plain-NMF) block fits: drop covariates, skip the
  ## O(N^2) clustering criteria, stay quiet about dimensions.
  ea <- dots; ea$A <- NULL; ea$Q <- NULL
  ea$detail <- "fast"; ea$print.dims <- FALSE

  objfunc <- stats::setNames(base::numeric(base::length(rank)),
                             base::paste0("rank=", rank))
  base::message(base::sprintf(
    "bi-CV: ranks %s, %dx%d fold grid (Owen-Perry 2009)...",
    base::paste(rank, collapse = ","), nfolds, nfolds))

  for (ki in base::seq_along(rank)) {
    k <- rank[ki]
    sse <- 0; cnt <- 0L
    for (fi in 1:nfolds) for (fj in 1:nfolds) {
      I  <- base::which(row.fold == fi); J  <- base::which(col.fold == fj)
      Ic <- base::which(row.fold != fi); Jc <- base::which(col.fold != fj)
      if (base::length(Ic) <= k || base::length(Jc) <= k) next
      D  <- Y[Ic, Jc, drop = FALSE]
      Bm <- Y[I,  Jc, drop = FALSE]   # held rows x kept cols
      Cm <- Y[Ic, J,  drop = FALSE]   # kept rows x held cols
      Am <- Y[I,  J,  drop = FALSE]   # held-out block
      fit <- base::suppressMessages(
        base::do.call("nmfkc", c(base::list(Y = D, rank = k), ea)))
      L_D <- fit$X; R_D <- fit$B
      L_I <- .nnls.mu(Bm, R_D, side = "left",  maxit = nnls.maxit)
      R_J <- .nnls.mu(Cm, L_D, side = "right", maxit = nnls.maxit)
      A_hat <- L_I %*% R_J
      sse <- sse + base::sum((Am - A_hat)^2)
      cnt <- cnt + base::length(Am)
    }
    objfunc[ki] <- if (cnt > 0) sse / cnt else NA_real_
  }

  base::list(objfunc = objfunc, sigma = base::sqrt(objfunc),
             rank = rank, nfolds = nfolds)
}

## Tests for nmfkc.bicv() (bi-cross-validation; lightweight, like nmfkc.ecv)

test_that("nmfkc.bicv returns sigma per rank and recovers a rank-3 signal", {
  set.seed(1)
  X <- matrix(abs(rnorm(30 * 3)), 30, 3)
  B <- matrix(abs(rnorm(3 * 40)), 3, 40)
  Y <- X %*% B                         # clean rank-3 non-negative matrix
  bv <- suppressMessages(nmfkc.bicv(Y, rank = 1:6, nfolds = 2, seed = 7))

  ## plain list, same spirit as nmfkc.ecv (objfunc / sigma)
  expect_type(bv, "list")
  expect_equal(bv$rank, 1:6)
  expect_length(bv$sigma, 6)
  expect_equal(bv$sigma, sqrt(bv$objfunc), tolerance = 1e-8)
  ## clean rank-3 data: error drops sharply from rank 2 to rank 3
  expect_lt(bv$sigma[3], bv$sigma[2])
  expect_true(bv$rank[which.min(bv$sigma)] >= 3)
})

test_that("nmfkc.bicv warns and returns NA for ranks too large for the block", {
  set.seed(2)
  Y <- matrix(abs(rnorm(8 * 50)), 8, 50)   # only 8 rows -> nfolds=2 keeps ~4
  ## rank 6 leaves <=4 kept rows per fold -> infeasible -> warning + NA
  expect_warning(
    bv <- suppressMessages(nmfkc.bicv(Y, rank = c(2, 6), nfolds = 2)),
    "infeasible")
  expect_false(is.na(bv$sigma[1]))
  expect_true(is.na(bv$sigma[2]))
})

# ============================================================
# S3 generic methods: coef, fitted, residuals, plot, summary
# ============================================================

# --- plot (convergence) ---

#' @title Plot convergence diagnostics for NMF models
#' @description
#' Plots the objective function value over iterations for \code{nmfre} and
#' \code{nmf.sem} objects. (For \code{nmfkc} and \code{nmfae}, plot methods
#' are defined in their respective source files.)
#'
#' @param x A fitted model object.
#' @param ... Additional graphical arguments passed to \code{\link{plot}}.
#' @return Invisible \code{NULL}.
#' @seealso \code{\link{nmfre}}, \code{\link{nmf.sem}}
#' @name plot.nmfre
#' @examples
#' \donttest{
#' set.seed(1)
#' Y <- matrix(runif(20), nrow = 4)
#' A <- diag(5)
#' res <- nmfre(Y, A, rank = 2, wild.bootstrap = FALSE)
#' plot(res)
#' }
#'
NULL

#' @rdname plot.nmfre
#' @export
plot.nmfre <- function(x, ...) {
  extra_args <- list(...)
  args <- list(x = x$objfunc.iter)
  if (is.null(extra_args$main)) {
    r2 <- if (!is.null(x$r.squared) && is.finite(x$r.squared)) round(x$r.squared, 3) else "NA"
    args$main <- paste0("R2 = ", r2)
  }
  if (is.null(extra_args$xlab)) args$xlab <- "iter"
  if (is.null(extra_args$ylab)) args$ylab <- "objfunc"
  if (is.null(extra_args$type)) args$type <- "l"
  all_args <- c(args, extra_args)
  do.call("plot", all_args)
  invisible(NULL)
}

#' @rdname plot.nmfre
#' @param which For \code{plot.nmf.sem}: which objective to plot.
#'   One of \code{"full"} (default; \code{loss + penalties}, the actual
#'   monotonically-decreasing quantity that the multiplicative updates
#'   minimize), \code{"reconstruction"} (Frobenius distance only,
#'   \eqn{\| Y_1 - X B \|_F^2}), or \code{"both"} (overlay both with a
#'   legend).  \code{"both"} is useful for diagnosing whether
#'   regularization is actively shaping the solution: if the two curves
#'   diverge, the penalties are pulling the optimizer away from the
#'   pure least-squares minimum.
#' @export
plot.nmf.sem <- function(x, ..., which = c("full", "reconstruction", "both")) {
  which <- match.arg(which)
  extra_args <- list(...)

  ## Pick the iteration trace(s) to plot.  Older nmf.sem objects may
  ## carry only x$objfunc (reconstruction loss); fall back gracefully.
  has_full <- !is.null(x$objfunc.full)
  if (which == "full" && !has_full) which <- "reconstruction"

  if (which == "both" && has_full) {
    y_full <- x$objfunc.full
    y_rec  <- x$objfunc
    iter_idx <- seq_along(y_full)
    main_default <- sprintf("MAE = %s, SC.cov = %s",
                            round(x$MAE, 3), round(x$SC.cov, 3))
    if (is.null(extra_args$main)) extra_args$main <- main_default
    if (is.null(extra_args$xlab)) extra_args$xlab <- "iter"
    if (is.null(extra_args$ylab)) extra_args$ylab <- "objective"
    do.call(plot, c(list(iter_idx, y_full, type = "l", lwd = 2,
                          col = "black"), extra_args))
    graphics::lines(iter_idx, y_rec, col = "tomato", lwd = 1.5, lty = 2)
    graphics::legend("topright",
                     legend = c("loss + penalties (full)", "reconstruction only"),
                     col    = c("black", "tomato"),
                     lty    = c(1, 2), lwd = c(2, 1.5),
                     bty    = "n", cex = 0.85)
  } else {
    y <- if (which == "full") x$objfunc.full else x$objfunc
    args <- list(x = y)
    if (is.null(extra_args$main)) {
      label <- if (which == "full") "loss + penalties"
               else "reconstruction only"
      args$main <- sprintf("%s | MAE = %s, SC.cov = %s",
                           label, round(x$MAE, 3), round(x$SC.cov, 3))
    }
    if (is.null(extra_args$xlab)) args$xlab <- "iter"
    if (is.null(extra_args$ylab))
      args$ylab <- if (which == "full") "objfunc.full" else "objfunc"
    if (is.null(extra_args$type)) args$type <- "l"
    all_args <- c(args, extra_args)
    do.call("plot", all_args)
  }
  invisible(NULL)
}


# --- summary.nmf.sem ---

#' @title Summary method for nmf.sem objects
#' @description
#' Produces a formatted summary of a fitted NMF-FFB model, including
#' matrix dimensions, convergence, stability diagnostics, fit statistics,
#' and inference results (if available).
#'
#' @param object An object of class \code{"nmf.ffb"} (or legacy
#'   \code{"nmf.sem"}) returned by \code{\link{nmf.ffb}} /
#'   \code{\link{nmf.sem}}.
#' @param ... Not used.
#' @return An object of class \code{"summary.nmf.sem"} (the fitted model
#'   tagged for printing); printed by \code{\link{print.summary.nmf.sem}}.
#' @seealso \code{\link{nmf.ffb}}, \code{\link{nmf.ffb.inference}}
#' @export
#' @examples
#' Y <- t(iris[, -5])
#' Y1 <- Y[1:2, ]; Y2 <- Y[3:4, ]
#' result <- nmf.ffb(Y1, Y2, rank = 2, maxit = 500)
#' summary(result)
#'
summary.nmf.sem <- function(object, ...) {
  class(object) <- "summary.nmf.sem"
  object
}

#' @title Print method for summary.nmf.sem objects
#' @description
#' Prints the NMF-FFB model summary (dimensions, convergence, stability
#' diagnostics, fit statistics, and inference results if available).
#' @param x An object of class \code{"summary.nmf.sem"} returned by
#'   \code{\link{summary.nmf.sem}}.
#' @param ... Not used.
#' @return Invisible \code{x}.
#' @seealso \code{\link{summary.nmf.sem}}
#' @export
print.summary.nmf.sem <- function(x, ...) {
  object <- x
  P1 <- nrow(object$X)
  Q  <- ncol(object$X)
  P2 <- ncol(object$C2)

  cat(sprintf("NMF-FFB: Y1(%d,N) = X(%d,%d) [C1(%d,%d) Y1 + C2(%d,%d) Y2]\n",
              P1, P1, Q, Q, P1, Q, P2))
  cat(sprintf("Iterations: %d\n", object$iter))

  cat("\nStability diagnostics:\n")
  cat(sprintf("  Spectral radius(XC1): %.4f %s\n",
              object$XC1.radius,
              if (object$XC1.radius < 1) "(stable)" else "(UNSTABLE)"))
  cat(sprintf("  ||XC1||_1:            %.4f\n", object$XC1.norm1))
  cat(sprintf("  Amplification:        %.4f (bound: %.4f)\n",
              object$amplification, object$amplification.bound))

  cat("\nFit statistics:\n")
  if (!is.null(object$SC.map) && is.finite(object$SC.map))
    cat(sprintf("  SC.map (mapping correlation):    %.4f\n", object$SC.map))
  if (!is.null(object$SC.cov) && is.finite(object$SC.cov))
    cat(sprintf("  SC.cov (covariance correlation): %.4f\n", object$SC.cov))
  if (!is.null(object$MAE)   && is.finite(object$MAE))
    cat(sprintf("  MAE (mean absolute error):       %.4f\n", object$MAE))
  if (!is.null(object$effective.rank) && is.finite(object$effective.rank))
    cat(sprintf("  Effective Rank:                  %.2f / %d  (%.1f%%)\n",
                object$effective.rank, Q, 100 * object$effective.rank / Q))

  # Coefficients from inference
  if (!is.null(object$coefficients) && is.data.frame(object$coefficients)) {
    cf <- object$coefficients

    ## Bootstrap meta-info (new full-pair-bootstrap inference; v0.6.8+)
    has_boot <- !is.null(object$bootstrap.B)
    if (has_boot) {
      cat(sprintf("\nBootstrap inference (X-fixed full pair bootstrap):\n"))
      cat(sprintf("  B = %d, valid = %d, threshold = %g, ci.level = %g\n",
                  object$bootstrap.B,
                  if (!is.null(object$bootstrap.n.valid)) object$bootstrap.n.valid else NA_integer_,
                  if (!is.null(object$bootstrap.threshold)) object$bootstrap.threshold else NA_real_,
                  if (!is.null(object$bootstrap.ci.level)) object$bootstrap.ci.level else NA_real_))
    }

    ## Print one block per Type ("C1" feedback / "C2" exogenous) when
    ## present.  Falls back to a single block for legacy inference output
    ## that lacks the Type column.
    print_block <- function(block, title) {
      if (nrow(block) == 0L) return(invisible(NULL))
      cat(sprintf("\n%s\n", title))
      rnames <- paste0(block$Covariate, " -> ", block$Basis)
      est <- formatC(block$Estimate, format = "g", digits = 4, width = 10)
      cl  <- if ("CI_low"  %in% names(block))
               formatC(block$CI_low,  format = "g", digits = 3, width = 10) else NULL
      cu  <- if ("CI_high" %in% names(block))
               formatC(block$CI_high, format = "g", digits = 3, width = 10) else NULL
      sup <- if ("support_rate" %in% names(block))
               formatC(block$support_rate, format = "f", digits = 3, width = 8) else NULL
      pv  <- block$p_value
      pv_str <- ifelse(!is.finite(pv), "      NA",
                 ifelse(pv < 2.2e-16, "  <2e-16",
                   formatC(pv, format = "g", digits = 4, width = 8)))
      stars <- if ("sig" %in% names(block)) {
                 block$sig
               } else {
                 ifelse(!is.finite(pv), " ",
                   ifelse(pv < 0.001, "***",
                     ifelse(pv < 0.01, "**",
                       ifelse(pv < 0.05, "*",
                         ifelse(pv < 0.1, ".", " ")))))
               }
      max_lw <- max(nchar(rnames))
      hdr_parts <- c(formatC("Estimate", width = 10))
      if (!is.null(cl)) hdr_parts <- c(hdr_parts, formatC("CI_low",  width = 10))
      if (!is.null(cu)) hdr_parts <- c(hdr_parts, formatC("CI_high", width = 10))
      if (!is.null(sup)) hdr_parts <- c(hdr_parts, formatC("support", width = 8))
      hdr_parts <- c(hdr_parts, formatC("Pr(>0)", width = 8), "")
      cat(sprintf("%s %s\n", formatC("", width = max_lw),
                  paste(hdr_parts, collapse = " ")))
      for (i in seq_along(rnames)) {
        row_parts <- c(est[i])
        if (!is.null(cl))  row_parts <- c(row_parts, cl[i])
        if (!is.null(cu))  row_parts <- c(row_parts, cu[i])
        if (!is.null(sup)) row_parts <- c(row_parts, sup[i])
        row_parts <- c(row_parts, pv_str[i], stars[i])
        cat(sprintf("%s %s\n",
                    formatC(rnames[i], width = max_lw),
                    paste(row_parts, collapse = " ")))
      }
    }

    if (!is.null(cf$Type)) {
      ## v0.6.8+ inference: separate C1 / C2 blocks
      cf_C1 <- cf[cf$Type == "C1", , drop = FALSE]
      cf_C2 <- cf[cf$Type == "C2", , drop = FALSE]
      print_block(cf_C1, "C1 Coefficients (Y1 -> Factor; feedback Theta_1):")
      print_block(cf_C2, "C2 Coefficients (Y2 -> Factor; exogenous Theta_2):")
    } else {
      ## Legacy inference output (only C2)
      print_block(cf, "C2 Coefficients (Covariate -> Basis):")
    }
    cat("---\n")
    cat("Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1\n")
  }

  invisible(object)
}


# --- coef ---

#' @title Extract coefficients from NMF models
#' @description
#' Returns the \code{coefficients} data frame from a fitted NMF model
#' that has been passed through an inference function
#' (\code{\link{nmfkc.inference}}, \code{\link{nmfae.inference}},
#' \code{\link{nmfre.inference}}).
#'
#' If inference has not been run, returns the parameter matrix \eqn{C}
#' (\eqn{\Theta}) directly.
#'
#' For \code{nmf.sem} objects, returns \eqn{C_2} (exogenous block) as fallback.
#'
#' @param object A fitted model object of class \code{"nmf"}, \code{"nmfkc"},
#'   \code{"nmfae"}, \code{"nmfre"}, or \code{"nmf.sem"}.
#' @param ... Not used.
#' @return A data frame of coefficients (if inference was performed),
#'   or the parameter matrix \eqn{C}.
#' @seealso \code{\link{nmfkc.inference}}, \code{\link{nmfae.inference}},
#'   \code{\link{nmfre.inference}}, \code{\link{nmf.sem.inference}}
#' @name coef.nmf
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(1, cars$speed)
#' result <- nmfkc(Y, A, rank = 1)
#' coef(result)  # returns C matrix
#'
#' result2 <- nmfkc.inference(result, Y, A)
#' coef(result2)  # returns coefficients data frame
#'
NULL

#' @rdname coef.nmf
#' @export
coef.nmf <- function(object, ...) {
  if (!is.null(object$coefficients)) object$coefficients else object$C
}

#' @rdname coef.nmf
#' @export
coef.nmf.sem <- function(object, ...) {
  ## With inference: return the unified coefficients data frame (rows for
  ## both C1 and C2, with CI / support_rate / sig columns).
  if (!is.null(object$coefficients)) return(object$coefficients)

  ## Without inference: return a long-format data frame with rows for
  ## every entry of C1 (feedback) and C2 (exogenous), so the column
  ## layout matches `coef(res_inf)`.  Users can post-hoc filter via
  ## `subset(coef(res), Type == "C1")` regardless of whether
  ## nmf.sem.inference() has been run.
  C1 <- object$C1
  C2 <- object$C2
  if (is.null(C1) || is.null(C2)) {
    ## Fallback (shouldn't normally happen for an nmf.sem result)
    return(if (!is.null(C2)) C2 else C1)
  }
  Q  <- nrow(C1)
  P1 <- ncol(C1)
  P2 <- ncol(C2)
  bas <- if (!is.null(rownames(C1))) rownames(C1) else paste0("Factor", seq_len(Q))
  y1  <- if (!is.null(colnames(C1))) colnames(C1) else paste0("Y1_", seq_len(P1))
  y2  <- if (!is.null(colnames(C2))) colnames(C2) else paste0("Y2_", seq_len(P2))
  C1_block <- data.frame(
    Type      = "C1",
    Basis     = rep(bas, times = P1),
    Covariate = rep(y1,  each  = Q),
    Estimate  = as.vector(C1),
    stringsAsFactors = FALSE
  )
  C2_block <- data.frame(
    Type      = "C2",
    Basis     = rep(bas, times = P2),
    Covariate = rep(y2,  each  = Q),
    Estimate  = as.vector(C2),
    stringsAsFactors = FALSE
  )
  out <- rbind(C1_block, C2_block)
  rownames(out) <- NULL
  out
}


# --- fitted ---

#' @title Extract fitted values from NMF models
#' @description
#' Returns the reconstructed matrix \eqn{\hat{Y} = X B} from a fitted
#' NMF model.
#'
#' For \code{nmf.sem} objects, returns the equilibrium prediction
#' \eqn{\hat{Y}_1 = M_{model} Y_2} if available. Supply \code{Y1} and
#' \code{Y2} to get the direct reconstruction
#' \eqn{X (C_1 Y_1 + C_2 Y_2)} instead.
#'
#' @param object A fitted model object of class \code{"nmf"}, \code{"nmfkc"},
#'   \code{"nmfae"}, \code{"nmfre"}, or \code{"nmf.sem"}.
#' @param ... For \code{nmf.sem}: optionally \code{Y1} and \code{Y2}.
#' @return The fitted matrix \eqn{X B}.
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfae}}, \code{\link{nmfre}},
#'   \code{\link{nmf.sem}}, \code{\link{residuals.nmf}}
#' @name fitted.nmf
#' @examples
#' result <- nmfkc(matrix(runif(50), 5, 10), rank = 2)
#' fitted(result)
#'
NULL

#' @rdname fitted.nmf
#' @export
fitted.nmf <- function(object, ...) {
  object$XB
}

#' @rdname fitted.nmf
#' @export
fitted.nmfae <- function(object, ...) {
  object$Y1hat
}

#' @rdname fitted.nmf
#' @export
fitted.nmf.sem <- function(object, ...) {
  extra <- list(...)
  Y1 <- extra$Y1
  Y2 <- extra$Y2
  if (!is.null(Y1) && !is.null(Y2)) {
    # Direct reconstruction: X(C1*Y1 + C2*Y2)
    object$X %*% (object$C1 %*% Y1 + object$C2 %*% Y2)
  } else {
    # Equilibrium prediction: M.model %*% Y2 (requires Y2)
    if (!is.null(Y2) && !anyNA(object$M.model)) {
      object$M.model %*% Y2
    } else {
      stop("Supply Y1 and Y2 for direct reconstruction, or Y2 for equilibrium prediction.")
    }
  }
}


# --- residuals ---

#' @title Extract residuals from NMF models
#' @description
#' Returns the residual matrix \eqn{Y - \hat{Y}} from a fitted NMF model.
#' Requires the original observation matrix \code{Y} to be supplied.
#'
#' For \code{nmfre} objects, residuals are computed from the BLUP
#' reconstruction (\eqn{Y - X(B_{blup})}) by default. Set
#' \code{type = "fixed"} to use fixed-effects only.
#'
#' @param object A fitted model object.
#' @param Y The original observation matrix used for fitting.
#' @param type For \code{nmfre} objects: \code{"blup"} (default) or
#'   \code{"fixed"}.
#' @param ... Not used.
#' @return The residual matrix.
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfae}}, \code{\link{nmfre}},
#'   \code{\link{nmf.sem}}, \code{\link{fitted.nmf}}
#' @name residuals.nmf
#' @examples
#' Y <- matrix(runif(50), 5, 10)
#' result <- nmfkc(Y, rank = 2)
#' residuals(result, Y)
#'
NULL

#' @rdname residuals.nmf
#' @export
residuals.nmf <- function(object, Y, ...) {
  Y - object$XB
}

#' @rdname residuals.nmf
#' @export
residuals.nmfae <- function(object, Y, ...) {
  Y - object$Y1hat
}

#' @rdname residuals.nmf
#' @export
residuals.nmfre <- function(object, Y, type = c("blup", "fixed"), ...) {
  type <- match.arg(type)
  if (type == "blup") {
    Y - object$XB.blup
  } else {
    Y - object$XB
  }
}

#' @rdname residuals.nmf
#' @export
residuals.nmf.sem <- function(object, Y, ...) {
  extra <- list(...)
  Y2 <- extra$Y2
  if (is.null(Y2)) stop("Supply Y2 to compute residuals.")
  if (!anyNA(object$M.model)) {
    Y - object$M.model %*% Y2
  } else {
    stop("Leontief inverse is NA; cannot compute residuals.")
  }
}


# --- print.nmf.inference ---

#' @title Print method for NMF inference objects
#' @description
#' Prints a summary of any NMF inference result object
#' (\code{"nmfkc.inference"} or \code{"nmfae.inference"}).
#' @param x An object of class \code{"nmf.inference"}.
#' @param ... Additional arguments passed to the corresponding
#'   \code{print.summary.*} method.
#' @return Called for its side effect (printing). Returns \code{x} invisibly.
#' @seealso \code{\link{nmfkc.inference}}, \code{\link{nmfae.inference}}
#' @export
print.nmf.inference <- function(x, ...) {
  print(summary(x), ...)
  invisible(x)
}

# ============================================================
# NMF-RE: Non-negative Matrix Factorization with Random Effects
# ============================================================

# ------------------------------------------------------------
# Internal helpers
# ------------------------------------------------------------

#' Safe element-wise division with fallback
#' @param num Numerator matrix.
#' @param den Denominator matrix.
#' @param eps Small constant to avoid division by zero.
#' @return Matrix of element-wise quotients.
#' @keywords internal
#' @noRd
.nmfre.safe.div <- function(num, den, eps = 1e-10) {
  res <- num / pmax(den, eps)
  res[is.na(res) | is.infinite(res)] <- 0
  res
}

#' Normalize columns of X to sum 1, rescaling C and U
#'
#' Applies X <- X D^{-1}, C <- D C, U <- D U where D = diag(colSums(X)).
#' @param X Basis matrix (P x Q).
#' @param C Coefficient matrix (Q x K).
#' @param U Random effects matrix (Q x N), or NULL.
#' @param eps Small constant to avoid division by zero.
#' @return A list with components \code{X}, \code{C}, \code{U} (if supplied), and \code{scale}.
#' @keywords internal
#' @noRd
.nmfre.normalize.X <- function(X, C, U = NULL, eps = 1e-12) {
  col_sums <- pmax(colSums(X), eps)
  Dinv <- diag(1 / col_sums, nrow = length(col_sums))
  D    <- diag(col_sums,     nrow = length(col_sums))

  Xn <- X %*% Dinv
  Cn <- D %*% C

  if (!is.null(U)) {
    Un <- D %*% U
    return(list(X = Xn, C = Cn, U = Un, scale = col_sums))
  }
  list(X = Xn, C = Cn, scale = col_sums)
}

#' Compute effective degrees of freedom for U from eigenvalues and lambda
#' @param d Eigenvalues of X'X.
#' @param N Number of observations.
#' @param lambda Ridge penalty parameter.
#' @return Scalar dfU value.
#' @keywords internal
#' @noRd
.nmfre.dfU.from.lambda <- function(d, N, lambda) {
  N * sum(d / (d + lambda))
}

#' Find lambda that achieves a target dfU cap via bisection
#' @param d Eigenvalues of X'X.
#' @param N Number of observations.
#' @param target Target dfU value.
#' @param lambda_min Lower bound for search.
#' @param lambda_max Upper bound for search.
#' @param tol Convergence tolerance.
#' @return Scalar lambda value.
#' @keywords internal
#' @noRd
.nmfre.lambda.for.dfU.cap <- function(d, N, target,
                                       lambda_min = 0,
                                       lambda_max = 1e12,
                                       tol = 1e-8) {
  if (.nmfre.dfU.from.lambda(d, N, lambda_min) <= target) return(lambda_min)

  hi <- lambda_min
  df_hi <- Inf
  while (hi < lambda_max) {
    hi <- max(1e-12, if (hi == 0) 1e-6 else hi * 10)
    df_hi <- .nmfre.dfU.from.lambda(d, N, hi)
    if (df_hi <= target) break
  }
  if (df_hi > target) return(lambda_max)

  f <- function(lam) .nmfre.dfU.from.lambda(d, N, lam) - target
  out <- stats::uniroot(f, lower = lambda_min, upper = hi, tol = tol)
  out$root
}


# ============================================================
# nmfre() - Main estimation function
# ============================================================

#' @title Non-negative Matrix Factorization with Random Effects
#' @description
#' Estimates the NMF-RE model
#' \deqn{Y = X(\Theta A + U) + \mathcal{E}}
#' where \eqn{Y} (\eqn{P \times N}) is a non-negative observation matrix,
#' \eqn{X} (\eqn{P \times Q}) is a non-negative basis matrix learned from the data,
#' \eqn{\Theta} (\eqn{Q \times K}) is a non-negative coefficient matrix capturing
#' systematic covariate effects on latent scores,
#' \eqn{A} (\eqn{K \times N}) is a covariate matrix, and
#' \eqn{U} (\eqn{Q \times N}) is a random effects matrix capturing
#' unit-specific deviations in the latent score space.
#'
#' NMF-RE can be viewed as a mixed-effects latent-variable model defined on a
#' reconstruction (mean) structure. The non-negativity constraint on \eqn{X}
#' induces sparse, parts-based loadings, achieving measurement-side variable
#' selection without an explicit sparsity penalty. Inference on \eqn{\Theta}
#' provides covariate-side variable selection by identifying which covariates
#' significantly affect which components.
#'
#' Estimation alternates ridge-type BLUP-like closed-form updates for \eqn{U}
#' with multiplicative non-negative updates for \eqn{X} and \eqn{\Theta}.
#' The effective degrees of freedom consumed by \eqn{U} are monitored and a
#' df-based cap can be enforced to prevent near-saturated fits.
#'
#' When \code{wild.bootstrap = TRUE}, inference on \eqn{\Theta} is performed
#' conditional on \eqn{(\hat{X}, \hat{U})} via asymptotic linearization,
#' a one-step Newton update, and a multiplier (wild) bootstrap, yielding
#' standard errors, z-values, p-values, and confidence intervals without
#' repeated constrained re-optimization.
#'
#' @param Y Observation matrix (P x N), non-negative.
#' @param A Covariate matrix (K x N). Default is a row of ones (intercept only).
#' @param rank Integer. Rank of the basis matrix \eqn{X}. Default is 2.
#'   For backward compatibility, \code{Q} is accepted via \code{...}.
#' @param df.rate Rate for computing the dfU cap (\code{cap = rate * N * Q}).
#'   For backward compatibility, \code{dfU.cap.rate} is accepted via \code{...}.
#'   If \code{NULL} (default), runs \code{\link{nmfre.dfU.scan}} internally and
#'   selects the minimum rate where the cap is not binding.
#'   Use \code{\link{nmfre.dfU.scan}} beforehand to examine dfU behavior across rates
#'   and choose an appropriate value.
#' @param wild.bootstrap Logical. If \code{TRUE} (default), perform wild bootstrap inference
#'   on \eqn{\Theta}.
#' @param epsilon Convergence tolerance for relative change in objective (default 1e-5).
#' @param maxit Maximum number of iterations.  Default \code{5000}
#'   (matches \code{\link{nmfkc}} and the other MU functions in the
#'   package).  When the cap is hit without meeting the relative-
#'   tolerance criterion, a \code{"maximum iterations (...) reached..."}
#'   warning is emitted so users notice unconverged fits.
#' @param ... Additional arguments for initialization, variance control, dfU control,
#'   optimization, and inference settings.
#'   \itemize{
#'     \item \code{X.init}: Initial basis matrix (P x Q), or \code{NULL}.
#'       When \code{NULL}, \code{\link{nmfkc}} is called internally to generate initial values.
#'     \item \code{C.init}: Initial coefficient matrix (Q x K), or \code{NULL}.
#'       When \code{NULL}, \code{\link{nmfkc}} is called internally to generate initial values.
#'     \item \code{U.init}: Initial random effects matrix (Q x N), or \code{NULL} (all zeros).
#'     \item \code{prefix}: Prefix for basis names (default \code{"Basis"}).
#'     \item \code{sigma2}: Initial residual variance (default 1).
#'     \item \code{sigma2.update}: Logical. Update \eqn{\sigma^2} during iterations (default \code{TRUE}).
#'     \item \code{tau2}: Initial random effect variance (default 1).
#'     \item \code{tau2.update}: Logical. Update \eqn{\tau^2} by moment matching (default \code{TRUE}).
#'       Disabled when dfU cap is active.
#'     \item \code{dfU.control}: Either \code{"cap"} (default) to enforce a cap on dfU,
#'       or \code{"off"} for no cap.
#'     \item \code{print.trace}: Logical. If \code{TRUE}, print progress every 100 iterations (default \code{FALSE}).
#'     \item \code{seed}: Integer seed for reproducibility (default 1).
#'     \item \code{C.p.side}: P-value sidedness: \code{"one.sided"} (default, for boundary null
#'       H0: C=0 vs H1: C>0) or \code{"two.sided"}.
#'     \item \code{wild.B}: Number of wild bootstrap replicates (default 500).
#'     \item \code{wild.seed}: Seed for wild bootstrap (default 123).
#'   }
#' @return A list of class \code{"nmfre"} with components.
#'   The model is \eqn{Y = X(\Theta A + U) + \mathcal{E}}.
#'
#'   \strong{Core matrices}
#'   \describe{
#'     \item{\code{X}}{Basis matrix \eqn{X} (\eqn{P \times Q}), columns normalized to sum to 1.}
#'     \item{\code{C}}{Coefficient matrix \eqn{\Theta} (\eqn{Q \times K}).}
#'     \item{\code{U}}{Random effects matrix \eqn{U} (\eqn{Q \times N}).}
#'   }
#'
#'   \strong{Variance components}
#'   \describe{
#'     \item{\code{sigma2}}{Residual variance \eqn{\hat{\sigma}^2}.}
#'     \item{\code{tau2}}{Random effect variance \eqn{\hat{\tau}^2}.}
#'     \item{\code{lambda}}{Ridge penalty \eqn{\lambda = \sigma^2 / \tau^2}.}
#'   }
#'
#'   \strong{Convergence diagnostics}
#'   \describe{
#'     \item{\code{converged}}{Logical. Whether the algorithm converged.}
#'     \item{\code{stop.reason}}{Character string describing why iteration stopped.}
#'     \item{\code{iter}}{Number of iterations performed.}
#'     \item{\code{maxit}}{Maximum iterations setting used.}
#'     \item{\code{epsilon}}{Convergence tolerance used.}
#'     \item{\code{objfunc}}{Final objective function value
#'       \eqn{\|Y - X(\Theta A + U)\|^2 + \lambda \|U\|^2}.}
#'     \item{\code{rel.change.final}}{Final relative change in objective.}
#'     \item{\code{objfunc.iter}}{Numeric vector of objective values per iteration.}
#'     \item{\code{rss.trace}}{Numeric vector of \eqn{\|Y - X(\Theta A + U)\|^2} per iteration.}
#'   }
#'
#'   \strong{Effective degrees of freedom (dfU) diagnostics}
#'   \describe{
#'     \item{\code{dfU}}{Final effective degrees of freedom
#'       \eqn{\mathrm{df}_U = N \sum_q d_q / (d_q + \lambda)},
#'       where \eqn{d_q} are eigenvalues of \eqn{X'X}.}
#'     \item{\code{dfU.cap}}{Upper bound imposed on \eqn{\mathrm{df}_U}.}
#'     \item{\code{dfU.cap.rate}}{Rate used to compute the cap.}
#'     \item{\code{dfU.cap.scan}}{Result of \code{\link{nmfre.dfU.scan}}, or \code{NULL}.}
#'     \item{\code{lambda.enforced}}{Final \eqn{\lambda} enforced to satisfy the cap.}
#'     \item{\code{dfU.hit.cap}}{Logical. Whether the cap was binding.}
#'     \item{\code{dfU.hit.iter}}{Iteration at which the cap first bound.}
#'     \item{\code{dfU.frac}}{\eqn{\mathrm{df}_U / (NQ)}, fraction of maximum df.}
#'     \item{\code{dfU.cap.frac}}{\eqn{\mathrm{df}_U^{\mathrm{cap}} / (NQ)}.}
#'   }
#'
#'   \strong{Fitted matrices}
#'   \describe{
#'     \item{\code{B}}{Fixed-effect scores \eqn{\Theta A} (\eqn{Q \times N}).}
#'     \item{\code{B.prob}}{Column-normalized probabilities from
#'       \eqn{\max(\Theta A, 0)}.}
#'     \item{\code{B.blup}}{BLUP scores \eqn{\Theta A + U} (\eqn{Q \times N}).}
#'     \item{\code{B.blup.pos}}{Non-negative BLUP scores
#'       \eqn{\max(\Theta A + U, 0)} (\eqn{Q \times N}).}
#'     \item{\code{B.blup.prob}}{Column-normalized probabilities from
#'       \eqn{\max(\Theta A + U, 0)}.}
#'     \item{\code{XB}}{Fitted values from fixed effects
#'       \eqn{X \Theta A} (\eqn{P \times N}).}
#'     \item{\code{XB.blup}}{Fitted values including random effects
#'       \eqn{X(\Theta A + U)} (\eqn{P \times N}).}
#'   }
#'
#'   \strong{Fit statistics}
#'   \describe{
#'     \item{\code{r.squared}}{Pearson \eqn{\mathrm{cor}(Y, X(\Theta A + U))^2}
#'       (BLUP prediction).}
#'     \item{\code{r.squared.uncentered}}{Uncentered
#'       \eqn{1 - \|Y - X(\Theta A + U)\|_F^2 / \|Y\|_F^2} (BLUP;
#'       baseline = zero matrix).}
#'     \item{\code{r.squared.centered}}{Row-mean centered
#'       \eqn{1 - \|Y - X(\Theta A + U)\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2}
#'       (BLUP; baseline = per-row mean).}
#'     \item{\code{r.squared.fixed}}{Pearson \eqn{\mathrm{cor}(Y, X\Theta A)^2}
#'       (fixed-only prediction).}
#'     \item{\code{r.squared.fixed.uncentered}, \code{r.squared.fixed.centered}}{Uncentered
#'       and centered \eqn{R^2} for the fixed-only prediction.}
#'     \item{\code{ICC}}{Trace-based Intraclass Correlation Coefficient.
#'       In the NMF-RE model, the conditional covariance of the \eqn{n}-th
#'       observation column is
#'       \eqn{\mathrm{Var}(Y_n) = \tau^2 X X^\top + \sigma^2 I_P},
#'       a \eqn{P \times P} matrix. Unlike a standard random intercept model
#'       where the design matrix \eqn{Z} is a simple indicator (so the ICC
#'       reduces to \eqn{\tau^2 / (\sigma^2 + \tau^2)}), the basis matrix
#'       \eqn{X} plays the role of \eqn{Z} in a random slopes model,
#'       making the variance contribution of \eqn{U} depend on \eqn{X}.
#'       To obtain a scalar summary, we take the trace of each component:
#'       \deqn{\mathrm{ICC} = \frac{\tau^2 \, \mathrm{tr}(X^\top X)}
#'       {\tau^2 \, \mathrm{tr}(X^\top X) + \sigma^2 P}.}
#'       This equals the average (over \eqn{P} dimensions) proportion of
#'       per-column variance attributable to the random effects.}
#'   }
#'
#'   \strong{Inference on \eqn{\Theta} (wild bootstrap)}
#'   \describe{
#'     \item{\code{sigma2.used}}{\eqn{\hat{\sigma}^2} used for inference.}
#'     \item{\code{C.vec.cov}}{Variance-covariance matrix for
#'       \eqn{\mathrm{vec}(\Theta)} (\eqn{QK \times QK}).}
#'     \item{\code{C.se}}{Standard error matrix for \eqn{\Theta} (\eqn{Q \times K}).}
#'     \item{\code{C.se.hess}}{Sandwich (Hessian-based) SE matrix for \eqn{\Theta}.}
#'     \item{\code{C.se.boot}}{Bootstrap SE matrix for \eqn{\Theta}.}
#'     \item{\code{coefficients}}{Data frame with columns Estimate, Std. Error, z value,
#'       Pr(>|z|), and confidence interval bounds for each element of \eqn{\Theta}.}
#'     \item{\code{C.ci.lower}}{Lower confidence interval matrix for \eqn{\Theta} (\eqn{Q \times K}).}
#'     \item{\code{C.ci.upper}}{Upper confidence interval matrix for \eqn{\Theta} (\eqn{Q \times K}).}
#'     \item{\code{C.boot.sd}}{Bootstrap standard deviation matrix for \eqn{\Theta} (\eqn{Q \times K}).}
#'     \item{\code{C.p.side}}{P-value sidedness used: \code{"one.sided"} or \code{"two.sided"}.}
#'   }
#' @seealso \code{\link{nmfre.inference}}, \code{\link{nmfre.dfU.scan}},
#'   \code{\link{nmfkc.DOT}}, \code{\link{summary.nmfre}}
#' @export
#' @references
#' Satoh, K. (2026). Wild Bootstrap Inference for Non-Negative Matrix
#'   Factorization with Random Effects. arXiv:2603.01468.
#'   \url{https://arxiv.org/abs/2603.01468}
#' @examples
#' # Example 1. cars data
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(intercept = 1, speed = cars$speed)
#' res <- nmfre(Y, A, rank = 1, maxit = 5000)
#' summary(res)
#'
#' \donttest{
#' # Example 2. Orthodont data (nlme)
#' if (requireNamespace("nlme", quietly = TRUE)) {
#'   Y <- matrix(nlme::Orthodont$distance, 4, 27)
#'   male <- ifelse(nlme::Orthodont$Sex[seq(1, 108, 4)] == "Male", 1, 0)
#'   A <- rbind(intercept = 1, male = male)
#'
#'   # Scan dfU cap rates to choose an appropriate value
#'   nmfre.dfU.scan(1:10/10, Y, A, rank = 1)
#'
#'   # Fit with chosen rate
#'   res <- nmfre(Y, A, rank = 1, df.rate = 0.2)
#'   summary(res)
#' }}
#'
nmfre <- function(Y, A = NULL, rank = 2, df.rate = NULL,
                  wild.bootstrap = TRUE, epsilon = 1e-5,
                  maxit = 5000, ...) {

  extra_args <- base::list(...)
  # backward compatibility
  if (!is.null(extra_args$Q)) rank <- extra_args$Q
  if (!is.null(extra_args$dfU.cap.rate)) df.rate <- extra_args$dfU.cap.rate
  Q <- rank
  dfU.cap.rate <- df.rate

  # --- Parameter Extraction from ... ---
  # initialization
  X.init    <- if (!is.null(extra_args$X.init)) extra_args$X.init else NULL
  C.init    <- if (!is.null(extra_args$C.init)) extra_args$C.init else NULL
  U.init    <- if (!is.null(extra_args$U.init)) extra_args$U.init else NULL
  prefix    <- if (!is.null(extra_args$prefix)) extra_args$prefix else "Basis"

  # variance handling
  sigma2        <- if (!is.null(extra_args$sigma2)) extra_args$sigma2 else 1
  sigma2.update <- if (!is.null(extra_args$sigma2.update)) extra_args$sigma2.update else TRUE
  tau2          <- if (!is.null(extra_args$tau2)) extra_args$tau2 else 1
  tau2.update   <- if (!is.null(extra_args$tau2.update)) extra_args$tau2.update else TRUE

  # variance update internals (fixed at defaults)
  sigma2.update.start <- 50
  sigma2.update.every <- 10
  sigma2.update.rate  <- 0.05
  sigma2.min <- 1e-12; sigma2.max <- 1e12
  tau2.update.start <- 1
  tau2.update.every <- 1
  tau2.update.rate  <- 0.2
  tau2.min <- 1e-12; tau2.max <- 1e12

  # dfU control
  dfU.control  <- if (!is.null(extra_args$dfU.control)) extra_args$dfU.control else "cap"

  # dfU internals (fixed at defaults)
  dfU.cap <- NULL
  dfU.enforce.every <- 1
  dfU.lambda.max <- 1e12

  # optimization extras
  print.trace <- if (!is.null(extra_args$print.trace)) extra_args$print.trace else FALSE
  seed        <- if (!is.null(extra_args$seed)) extra_args$seed else 1

  # inference settings
  C.p.side <- if (!is.null(extra_args$C.p.side)) extra_args$C.p.side else "one.sided"

  # inference internals (fixed at defaults)
  sigma2.hat  <- NULL
  df.sigma    <- "PN-df"
  C.info.mode <- "IminusH"
  sandwich    <- TRUE
  sandwich.cr1 <- TRUE
  se.rule     <- "sandwich"
  cov.ridge   <- 1e-8

  # wild bootstrap
  wild.B    <- if (!is.null(extra_args$wild.B)) extra_args$wild.B else 500
  wild.seed <- if (!is.null(extra_args$wild.seed)) extra_args$wild.seed else 123

  # wild bootstrap internals (fixed at defaults)
  wild.level <- 0.95

  set.seed(seed)
  .eps <- 1e-10

  dfU.control <- match.arg(dfU.control, choices = c("cap", "off"))
  C.p.side    <- match.arg(C.p.side, choices = c("one.sided", "two.sided"))

  # ---- dimensions ----
  P <- nrow(Y); N <- ncol(Y)
  if (is.null(A)) A <- matrix(1, 1, N)
  if (!is.matrix(A)) A <- as.matrix(A)
  stopifnot(ncol(A) == N)
  K <- nrow(A)

  # ---- init via nmfkc when X.init/C.init are NULL ----
  if (is.null(X.init) || is.null(C.init)) {
    res0 <- nmfkc(Y, A, Q = Q, epsilon = epsilon, seed = seed,
                          print.trace = print.trace, print.dims = FALSE)
    if (is.null(X.init)) X.init <- res0$X
    if (is.null(C.init)) C.init <- res0$C
  }

  X <- X.init
  stopifnot(nrow(X) == P, ncol(X) == Q)

  C_mat <- C.init
  stopifnot(nrow(C_mat) == Q, ncol(C_mat) == K)

  if (is.null(U.init)) {
    U <- matrix(0, Q, N)
  } else {
    U <- U.init
    stopifnot(nrow(U) == Q, ncol(U) == N)
  }

  # ---- dfU.cap.rate auto-selection via scan ----
  dfU.cap.scan <- NULL
  if (is.null(dfU.cap.rate) && dfU.control == "cap") {
    dfU.cap.scan <- nmfre.dfU.scan(
      Y = Y, A = A, Q = Q,
      X.init = X, C.init = C_mat, U.init = U,
      print.trace = FALSE, wild.bootstrap = FALSE,
      maxit = maxit, epsilon = epsilon, seed = seed
    )
    selected_rate <- dfU.cap.scan$cap.rate
    if (is.null(selected_rate) || is.na(selected_rate)) {
      message("No safeguard rate found in auto scan. Scan results:")
      print(dfU.cap.scan$table)
      stop("No safeguard rate found in auto scan. Try specifying df.rate manually.")
    }
    dfU.cap.rate <- selected_rate
  }
  if (is.null(dfU.cap.rate)) dfU.cap.rate <- 0.10

  if (!is.numeric(dfU.cap.rate) || length(dfU.cap.rate) != 1 ||
      !is.finite(dfU.cap.rate) || dfU.cap.rate <= 0) {
    stop("'dfU.cap.rate' must be a positive finite number (e.g., 0.05).")
  }

  # ---- normalize X columns to sum 1 (rescale C AND U) ----
  normed <- .nmfre.normalize.X(pmax(X, .eps), pmax(C_mat, .eps), U)
  X <- normed$X
  C_mat <- normed$C
  U <- normed$U

  clip_val <- function(x, xmin, xmax) min(max(x, xmin), xmax)

  # ---- convergence bookkeeping ----
  obj <- NA_real_
  obj_prev <- Inf
  rel_change <- NA_real_

  converged <- FALSE
  stop_reason <- "maxit"
  iter_done <- 0

  obj_trace <- rep(NA_real_, maxit)
  rss_trace <- rep(NA_real_, maxit)

  # ---- dfU diagnostics ----
  dfU_last <- NA_real_
  dfU_cap_last <- NA_real_
  lambda_last <- NA_real_

  dfU_hit_cap <- FALSE
  dfU_hit_iter <- integer(0)

  # =========================================================
  # main loop
  # =========================================================
  for (iter in 1:maxit) {
    iter_done <- iter

    # (0) current components
    CA <- C_mat %*% A  # Q x N

    # (1) U-step: ridge BLUP with lambda = sigma2/tau2
    tau2 <- clip_val(tau2, tau2.min, tau2.max)
    sigma2 <- clip_val(sigma2, sigma2.min, sigma2.max)
    lambda <- sigma2 / tau2

    XtX <- crossprod(X)  # Q x Q

    # ---- dfU cap enforcement ----
    if (dfU.control == "cap" && (iter %% dfU.enforce.every == 0)) {

      dfU_cap_now <- if (!is.null(dfU.cap)) dfU.cap else dfU.cap.rate * (N * Q)
      dfU_cap_now <- max(dfU_cap_now, 1)

      d <- eigen(XtX, symmetric = TRUE, only.values = TRUE)$values
      d <- pmax(d, 0)

      dfU_now <- .nmfre.dfU.from.lambda(d, N, lambda)

      if (dfU_now > dfU_cap_now) {
        lambda_new <- .nmfre.lambda.for.dfU.cap(
          d, N, target = dfU_cap_now,
          lambda_min = pmax(lambda, 0),
          lambda_max = dfU.lambda.max
        )
        lambda <- lambda_new
        tau2 <- clip_val(sigma2 / pmax(lambda, 1e-12), tau2.min, tau2.max)
        dfU_now <- .nmfre.dfU.from.lambda(d, N, lambda)

        dfU_hit_cap <- TRUE
        dfU_hit_iter <- c(dfU_hit_iter, iter)
      }

      dfU_last <- dfU_now
      dfU_cap_last <- dfU_cap_now
    }

    lambda_last <- lambda

    M <- XtX + diag(pmax(lambda, 1e-12), Q)
    cholM <- tryCatch(chol(M), error = function(e) NULL)
    if (is.null(cholM)) stop("Cholesky failed: XtX + lambda I not SPD. Increase lambda (or decrease tau2).")

    # residual without U
    fit_fixed <- X %*% CA
    R0 <- Y - fit_fixed

    # solve for each column u_n
    for (n in 1:N) {
      rhs <- crossprod(X, R0[, n, drop = FALSE])  # Q x 1
      u <- backsolve(cholM, forwardsolve(t(cholM), rhs))
      U[, n] <- as.numeric(u)
    }

    # center U (identifiability)
    U <- sweep(U, 1, rowMeans(U), "-")

    # (1.5) tau2 update by moment matching (disabled when dfU cap is on)
    if (isTRUE(tau2.update) && dfU.control != "cap" &&
        iter >= tau2.update.start && (iter %% tau2.update.every == 0)) {

      term1 <- sum(U^2) / (N * Q)

      Minv <- tryCatch(chol2inv(cholM), error = function(e) NULL)
      if (!is.null(Minv)) {
        term2 <- sigma2 * sum(diag(Minv)) / Q
        tau2_new <- clip_val(term1 + term2, tau2.min, tau2.max)
        tau2 <- (1 - tau2.update.rate) * tau2 + tau2.update.rate * tau2_new
      }
    }

    # (2) X-step (EU MU) with positive-part stabilization
    Y_tilde <- pmax(Y, 0)
    B_sem <- CA + U
    B_pos <- pmax(B_sem, 0)

    numX <- Y_tilde %*% t(B_pos)
    denX <- X %*% (B_pos %*% t(B_pos))
    X <- X * .nmfre.safe.div(numX, denX, eps = .eps)
    X <- pmax(X, .eps)

    # normalize columns sum 1; absorb scale into C AND U
    normed <- .nmfre.normalize.X(X, C_mat, U)
    X <- pmax(normed$X, .eps)
    C_mat <- pmax(normed$C, .eps)
    U <- normed$U

    # (3) C-step (EU MU) with positive-part stabilization
    Y_star <- Y_tilde - X %*% U
    Y_star_pos <- pmax(Y_star, 0)

    numC <- crossprod(X, Y_star_pos) %*% t(A)
    denC <- (crossprod(X, X) %*% C_mat) %*% (A %*% t(A))
    C_mat <- C_mat * .nmfre.safe.div(numC, denC, eps = .eps)
    C_mat <- pmax(C_mat, .eps)

    # (3.5) sigma2 update from residual
    if (isTRUE(sigma2.update) && iter >= sigma2.update.start &&
        (iter %% sigma2.update.every == 0)) {
      CA <- C_mat %*% A
      fit <- X %*% (CA + U)
      R <- Y - fit
      sigma2_new <- clip_val(mean(R^2), sigma2.min, sigma2.max)
      sigma2 <- (1 - sigma2.update.rate) * sigma2 + sigma2.update.rate * sigma2_new
    }

    # (4) objective & convergence
    CA <- C_mat %*% A
    fit <- X %*% (CA + U)
    R <- Y - fit

    lambda_obj <- sigma2 / clip_val(tau2, tau2.min, tau2.max)
    obj <- sum(R^2) + lambda_obj * sum(U^2)
    rss <- sum(R^2)

    obj_trace[iter] <- obj
    rss_trace[iter] <- rss

    if (is.finite(obj_prev)) {
      rel_change <- abs(obj_prev - obj) / (abs(obj_prev) + .eps)
    } else {
      rel_change <- NA_real_
    }

    if (!is.finite(obj)) {
      stop_reason <- "nonfinite_obj"
      converged <- FALSE
      break
    }

    if (is.finite(obj_prev) && rel_change < epsilon) {
      stop_reason <- "epsilon"
      converged <- TRUE
      break
    }

    obj_prev <- obj

    if (isTRUE(print.trace) && iter %% 100 == 0) {
      message(sprintf("iter=%d  obj=%.6g  rel=%.3g  sigma2=%.4g  tau2=%.4g  lambda=%.4g",
                      iter, obj, rel_change, sigma2, tau2, sigma2 / tau2))
    }
  }

  # ---- post loop: finalize ----
  obj_final <- obj
  rel_change_final <- rel_change

  if (iter_done < maxit && identical(stop_reason, "maxit")) {
    stop_reason <- "unknown_break"
    converged <- FALSE
  }

  ## Warn when the MU loop exhausted maxit without meeting the
  ## relative-tolerance criterion (matches nmfkc() / nmf.sem()
  ## convention).  The non-finite-objective break path issues its
  ## own diagnostic via stop_reason; here we only fire on maxit.
  if (identical(stop_reason, "maxit") && iter_done == maxit)
    warning(paste0("maximum iterations (", maxit, ") reached..."))

  # ---- reorder basis ----
  if (ncol(X) > 1) {
    w_ord <- matrix((1:P) / P, nrow = 1)
    score <- as.numeric(w_ord %*% X)
    index <- order(score)
    X <- X[, index, drop = FALSE]
    C_mat <- C_mat[index, , drop = FALSE]
    U <- U[index, , drop = FALSE]
  }

  # ---- final fitted matrices ----
  CA <- C_mat %*% A
  B_fixed <- CA
  B_blup_raw <- CA + U

  XB <- X %*% B_fixed
  XB.blup <- X %*% B_blup_raw

  # ---- final dfU recomputation ----
  lambda_final <- sigma2 / clip_val(tau2, tau2.min, tau2.max)
  XtX_final <- crossprod(X)
  d_final <- eigen(XtX_final, symmetric = TRUE, only.values = TRUE)$values
  d_final <- pmax(d_final, 0)
  dfU_final <- .nmfre.dfU.from.lambda(d_final, N, lambda_final)

  if (dfU.control == "cap") {
    dfU_cap_last <- if (!is.null(dfU.cap)) max(dfU.cap, 1) else max(dfU.cap.rate * (N * Q), 1)
  } else {
    dfU_cap_last <- NA_real_
  }
  dfU_last <- dfU_final
  lambda_last <- lambda_final

  # ---- names ----
  colnames(X) <- paste0(prefix, 1:ncol(X))
  rownames(C_mat) <- colnames(X)
  colnames(C_mat) <- rownames(A)
  rownames(U) <- rownames(C_mat)
  colnames(U) <- colnames(Y)

  # ---- Three-variant R^2 (cor^2, uncentered, row-mean
  # centered), separately for BLUP-prediction (XB.blup = XB + U) and
  # fixed-only prediction (XB).  Consistent with nmfkc / nmfae /
  # nmfae.signed / nmfkc.net / nmfkc.signed naming convention.
  r2_blup  <- .r.squared.all(Y, XB.blup)
  r2_fixed <- .r.squared.all(Y, XB)
  r.squared                <- r2_blup$r.squared
  r.squared.uncentered           <- r2_blup$r.squared.uncentered
  r.squared.centered       <- r2_blup$r.squared.centered
  r.squared.fixed          <- r2_fixed$r.squared
  r.squared.fixed.uncentered     <- r2_fixed$r.squared.uncentered
  r.squared.fixed.centered <- r2_fixed$r.squared.centered

  # ---- ICC (trace-based) ----
  trXtX <- sum(d_final)  # tr(X'X), using eigenvalues already computed
  ICC <- tau2 * trXtX / (tau2 * trXtX + sigma2 * P)

  # =========================================================
  # Inference on C
  # =========================================================
  C.vec.cov <- NULL
  C.se <- NULL
  C.se.hess <- NULL
  C.se.boot <- NULL
  coefficients <- NULL
  sigma2_used <- sigma2.hat

  C.ci.lower <- NULL
  C.ci.upper <- NULL
  C.boot.sd <- NULL

  if (isTRUE(wild.bootstrap)) {

    # Y_star for inference (conditioning on X,U)
    Y_star_inf <- Y - X %*% U
    R_C <- Y_star_inf - X %*% (C_mat %*% A)
    RSS_inf <- sum(R_C^2)

    lambda_inf <- sigma2 / pmax(tau2, 1e-12)

    XtX_now <- crossprod(X)
    AAt <- A %*% t(A)

    M_inf <- XtX_now + diag(pmax(lambda_inf, 1e-12), Q)
    cholM_inf <- tryCatch(chol(M_inf), error = function(e) NULL)
    if (!is.null(cholM_inf)) {
      Minv <- chol2inv(cholM_inf)
    } else {
      if (!requireNamespace("MASS", quietly = TRUE)) {
        stop("Package 'MASS' is required for generalized inverse when Cholesky fails.")
      }
      Minv <- tryCatch(solve(M_inf), error = function(e) MASS::ginv(M_inf))
    }

    trH <- sum(diag(Minv %*% XtX_now))
    dfU_inf <- N * trH

    if (is.null(sigma2_used)) {
      if (df.sigma == "PN") {
        denom <- P * N
      } else if (df.sigma == "PN-QR") {
        denom <- max(P * N - Q * K, 1)
      } else { # "PN-df"
        dfC <- Q * K
        denom <- max(P * N - dfU_inf - dfC, 1)
      }
      sigma2_used <- RSS_inf / denom
    }

    if (C.info.mode == "plain") {
      Info_core <- kronecker(AAt, XtX_now)
    } else {
      Xt_IH_X <- XtX_now - XtX_now %*% Minv %*% XtX_now
      Info_core <- kronecker(AAt, Xt_IH_X)
    }

    Info <- Info_core / max(sigma2_used, 1e-12)
    Info <- Info + diag(cov.ridge, nrow(Info))

    if (!requireNamespace("MASS", quietly = TRUE)) {
      Hinv <- tryCatch(solve(Info), error = function(e) {
        stop("Package 'MASS' is required for generalized inverse.")
      })
    } else {
      Hinv <- tryCatch(solve(Info), error = function(e) MASS::ginv(Info))
    }
    V_naive <- Hinv

    V_sand <- NULL
    if (isTRUE(sandwich)) {
      Xt <- t(X)
      J <- matrix(0, Q * K, Q * K)

      for (n in 1:N) {
        a_n <- A[, n, drop = FALSE]
        g_n <- Xt %*% R_C[, n, drop = FALSE]
        S_n <- -(g_n %*% t(a_n)) / max(sigma2_used, 1e-12)
        s_n <- as.vector(S_n)
        J <- J + tcrossprod(s_n)
      }

      if (isTRUE(sandwich.cr1) && N > 1) {
        J <- (N / (N - 1)) * J
      }
      V_sand <- Hinv %*% J %*% Hinv
    }

    if (se.rule == "naive") {
      C.vec.cov <- V_naive
    } else if (se.rule == "sandwich") {
      C.vec.cov <- if (is.null(V_sand)) V_naive else V_sand
    } else { # "max"
      if (is.null(V_sand)) {
        C.vec.cov <- V_naive
      } else {
        d1 <- pmax(diag(V_naive), 0)
        d2 <- pmax(diag(V_sand), 0)
        dmax <- pmax(d1, d2)
        C.vec.cov <- V_naive
        diag(C.vec.cov) <- dmax
      }
    }

    se_vec <- sqrt(pmax(diag(C.vec.cov), 0))
    C.se.hess <- matrix(se_vec, nrow = Q, ncol = K, byrow = FALSE)
    dimnames(C.se.hess) <- dimnames(C_mat)

    C.se <- C.se.hess

    # ---- Multiplier (wild) bootstrap ----
    set.seed(wild.seed)

    Xt <- t(X)
    score_mat <- matrix(0, Q * K, N)

    for (n in 1:N) {
      a_n <- A[, n, drop = FALSE]
      r_n <- R_C[, n, drop = FALSE]
      g_n <- Xt %*% r_n
      G_n <- -(g_n %*% t(a_n)) / max(sigma2_used, 1e-12)
      score_mat[, n] <- as.vector(G_n)
    }

    C_hat_vec <- as.vector(C_mat)
    C_boot <- matrix(NA_real_, nrow = Q * K, ncol = wild.B)

    for (b in 1:wild.B) {
      w <- stats::rexp(N, rate = 1) - 1
      grad_b <- as.vector(score_mat %*% w)
      c_b <- C_hat_vec - as.vector(Hinv %*% grad_b)
      c_b <- pmax(c_b, 0)
      C_boot[, b] <- c_b
    }

    alpha <- 1 - wild.level
    lo_q <- alpha / 2
    hi_q <- 1 - alpha / 2

    lo <- apply(C_boot, 1, stats::quantile, probs = lo_q, na.rm = TRUE, names = FALSE)
    hi <- apply(C_boot, 1, stats::quantile, probs = hi_q, na.rm = TRUE, names = FALSE)

    C.ci.lower <- matrix(lo, nrow = Q, ncol = K, byrow = FALSE)
    C.ci.upper <- matrix(hi, nrow = Q, ncol = K, byrow = FALSE)
    dimnames(C.ci.lower) <- dimnames(C_mat)
    dimnames(C.ci.upper) <- dimnames(C_mat)

    sd_vec <- apply(C_boot, 1, stats::sd, na.rm = TRUE)
    C.boot.sd <- matrix(sd_vec, nrow = Q, ncol = K, byrow = FALSE)
    dimnames(C.boot.sd) <- dimnames(C_mat)

    C.se.boot <- C.boot.sd

    # ---- coefficients table ----
    Estimate <- as.vector(C_mat)
    SE_hess_vec <- as.vector(C.se.hess)
    SE <- SE_hess_vec
    BSE_vec <- as.vector(C.boot.sd)

    z_value <- rep(NA_real_, length(Estimate))
    ok <- is.finite(Estimate) & is.finite(SE) & (SE > 0)
    z_value[ok] <- Estimate[ok] / SE[ok]

    Wald <- rep(NA_real_, length(Estimate))
    Wald[ok] <- z_value[ok]^2

    p_two <- rep(NA_real_, length(Estimate))
    p_two[ok] <- 1 - stats::pchisq(Wald[ok], df = 1)

    p_one <- rep(NA_real_, length(Estimate))
    p_one[ok] <- stats::pnorm(z_value[ok], lower.tail = FALSE)

    p_value <- if (C.p.side == "one.sided") p_one else p_two

    coefficients <- data.frame(
      Basis     = rep(rownames(C_mat), times = K),
      Covariate = rep(colnames(C_mat), each = Q),
      Estimate = Estimate,
      SE       = SE,
      BSE      = BSE_vec,
      z_value  = z_value,
      Wald     = Wald,
      p_value  = p_value,
      p_two_sided = p_two,
      p_one_sided = p_one,
      p_side_used = C.p.side,
      row.names = NULL,
      stringsAsFactors = FALSE
    )

    if (!is.null(C.ci.lower) && !is.null(C.ci.upper)) {
      coefficients$CI_low  <- as.vector(C.ci.lower)
      coefficients$CI_high <- as.vector(C.ci.upper)
    }
  }

  # ---- convenience probabilities ----
  colnorm_prob <- function(M, eps = 1e-12) {
    cs <- colSums(M)
    sweep(M, 2, pmax(cs, eps), "/")
  }

  out <- list(
    X = X,
    C = C_mat,
    U = U,

    sigma2 = sigma2,
    tau2 = tau2,
    lambda = sigma2 / tau2,

    # convergence diagnostics
    converged = converged,
    stop.reason = stop_reason,
    iter = iter_done,
    maxit = maxit,
    epsilon = epsilon,
    objfunc = obj_final,
    rel.change.final = rel_change_final,
    objfunc.iter = obj_trace[seq_len(iter_done)],
    rss.trace = rss_trace[seq_len(iter_done)],

    # dfU diagnostics
    dfU = dfU_last,
    dfU.cap = dfU_cap_last,
    dfU.cap.rate = dfU.cap.rate,
    dfU.cap.scan = dfU.cap.scan,
    lambda.enforced = lambda_last,
    dfU.hit.cap  = dfU_hit_cap,
    dfU.hit.iter = dfU_hit_iter,
    dfU.frac     = if (is.finite(dfU_last)) dfU_last / (N * Q) else NA_real_,
    dfU.cap.frac = if (is.finite(dfU_cap_last)) dfU_cap_last / (N * Q) else NA_real_,

    # fitted matrices
    B = B_fixed,
    B.prob = colnorm_prob(pmax(B_fixed, 0)),
    B.blup = B_blup_raw,
    B.blup.pos = pmax(B_blup_raw, 0),
    B.blup.prob = colnorm_prob(pmax(B_blup_raw, 0)),
    XB = XB,
    XB.blup = XB.blup,

    # fit statistics
    r.squared                = r.squared,
    r.squared.uncentered           = r.squared.uncentered,
    r.squared.centered       = r.squared.centered,
    r.squared.fixed          = r.squared.fixed,
    r.squared.fixed.uncentered     = r.squared.fixed.uncentered,
    r.squared.fixed.centered = r.squared.fixed.centered,
    ICC = ICC,

    # inference outputs
    sigma2.used = sigma2_used,
    C.vec.cov = C.vec.cov,
    C.se = C.se,
    C.se.hess = C.se.hess,
    C.se.boot = C.se.boot,
    coefficients = coefficients,
    C.ci.lower = C.ci.lower,
    C.ci.upper = C.ci.upper,
    C.boot.sd  = C.boot.sd,

    C.p.side = C.p.side
  )

  class(out) <- c("nmfre", "nmf")
  out
}


# ============================================================
# summary.nmfre() - S3 summary method
# ============================================================

#' @title Summary method for objects of class \code{nmfre}
#' @description
#' Displays a concise summary of an NMF-RE model fit, including dimensions,
#' convergence, variance components, and a coefficient table following
#' standard R regression output conventions.
#'
#' @param object An object of class \code{nmfre}, returned by \code{\link{nmfre}}.
#' @param show_ci Logical. If \code{TRUE}, show confidence interval columns (default \code{FALSE}).
#' @param ... Additional arguments (currently unused).
#' @return The input object, invisibly.
#' @seealso \code{\link{nmfre}}, \code{\link{nmfre.inference}}
#' @export
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(intercept = 1, speed = cars$speed)
#' res <- nmfre(Y, A, rank = 1, maxit = 5000)
#' summary(res)
#'
summary.nmfre <- function(object, show_ci = FALSE, ...) {

  x <- object
  P <- nrow(x$X); Q <- ncol(x$X)
  K <- ncol(x$C); N <- ncol(x$U)

  # ---- header ----
  cat(sprintf("NMF-RE: Y(%d,%d) = X(%d,%d) [C(%d,%d) A + U(%d,%d)]\n",
              P, N, P, Q, Q, K, Q, N))

  # ---- convergence ----
  conv_str <- if (isTRUE(x$converged)) "converged" else "NOT converged"
  cat(sprintf("Iterations: %d (%s, epsilon = %s)\n",
              x$iter, conv_str, format(x$epsilon, digits = 1)))

  # ---- R-squared ----
  if (!is.null(x$r.squared) && is.finite(x$r.squared)) {
    r2_fixed <- if (!is.null(x$r.squared.fixed) && is.finite(x$r.squared.fixed)) {
      sprintf(", %.4f (XB)", x$r.squared.fixed)
    } else ""
    cat(sprintf("R-squared (cor^2):    %.4f (XB+blup)%s\n", x$r.squared, r2_fixed))
    if (!is.null(x$r.squared.uncentered) && is.finite(x$r.squared.uncentered)) {
      r2f_fixed <- if (!is.null(x$r.squared.fixed.uncentered) && is.finite(x$r.squared.fixed.uncentered))
        sprintf(", %.4f (XB)", x$r.squared.fixed.uncentered) else ""
      cat(sprintf("R-squared (uncentered):     %.4f (XB+blup)%s\n", x$r.squared.uncentered, r2f_fixed))
    }
    if (!is.null(x$r.squared.centered) && is.finite(x$r.squared.centered)) {
      r2c_fixed <- if (!is.null(x$r.squared.fixed.centered) && is.finite(x$r.squared.fixed.centered))
        sprintf(", %.4f (XB)", x$r.squared.fixed.centered) else ""
      cat(sprintf("R-squared (centered): %.4f (XB+blup)%s\n", x$r.squared.centered, r2c_fixed))
    }
  }

  # ---- effective rank (BLUP scores B = Theta A + U, Q x N) ----
  eff <- if (!is.null(x$B.blup)) .effective.rank(x$B.blup) else NA_real_
  if (is.finite(eff)) cat(sprintf("Effective Rank:       %.2f / %d  (%.1f%%)\n",
                                  eff, Q, 100 * eff / Q))

  # ---- variance components ----
  cat(sprintf("\nVariance components:\n"))
  cat(sprintf("  sigma2 = %.4g  (residual)\n", x$sigma2))
  cat(sprintf("  tau2   = %.4g  (random effect)\n", x$tau2))
  cat(sprintf("  lambda = %.4g  (sigma2 / tau2)\n", x$lambda))
  cat(sprintf("  ICC    = %.4f  (tau2*tr(X'X) / (tau2*tr(X'X) + sigma2*P))\n", x$ICC))
  if (!is.null(x$dfU.cap) && is.finite(x$dfU.cap)) {
    cat(sprintf("  dfU    = %.2f <= %.2f (cap rate = %.2f)\n",
                x$dfU, x$dfU.cap, x$dfU.cap.rate))
  } else if (is.finite(x$dfU)) {
    cat(sprintf("  dfU    = %.2f\n", x$dfU))
  }

  # ---- coefficients ----
  if (!is.null(x$coefficients) && is.data.frame(x$coefficients)) {
    cat("\nCoefficients:\n")

    cf <- x$coefficients

    # row names: Covariate:Basis
    rnames <- paste0(cf$Covariate, ":", cf$Basis)

    # p-value formatting (lm-style)
    p_side <- if (!is.null(x$C.p.side)) x$C.p.side else "one.sided"
    pv <- cf$p_value

    format_pval <- function(p) {
      ifelse(!is.finite(p), "      NA",
        ifelse(p < 2.2e-16, "  <2e-16",
          formatC(p, format = "g", digits = 4, width = 8)))
    }

    # significance stars
    sig_stars <- function(p) {
      ifelse(!is.finite(p), " ",
        ifelse(p < 0.001, "***",
          ifelse(p < 0.01, "**",
            ifelse(p < 0.05, "*",
              ifelse(p < 0.1, ".", " ")))))
    }

    # build display matrix
    p_header <- if (p_side == "one.sided") "Pr(>z)" else "Pr(>|z|)"

    # columns
    est <- formatC(cf$Estimate, format = "f", digits = 3, width = 9)
    se  <- formatC(cf$SE, format = "f", digits = 3, width = 10)
    zv  <- formatC(cf$z_value, format = "f", digits = 2, width = 7)
    pv_str <- format_pval(pv)
    stars <- sig_stars(pv)

    has_bse <- "BSE" %in% names(cf) && any(is.finite(cf$BSE))
    if (has_bse) {
      bse <- formatC(cf$BSE, format = "f", digits = 3, width = 6)
    }

    # header line
    if (has_bse) {
      hdr <- sprintf("%s %s %s %s %s %s",
                     formatC("Estimate", width = 9),
                     formatC("Std. Error", width = 10),
                     formatC("(Boot)", width = 6),
                     formatC("z value", width = 7),
                     formatC(p_header, width = 8),
                     "")
    } else {
      hdr <- sprintf("%s %s %s %s %s",
                     formatC("Estimate", width = 9),
                     formatC("Std. Error", width = 10),
                     formatC("z value", width = 7),
                     formatC(p_header, width = 8),
                     "")
    }

    # max label width for alignment
    max_lw <- max(nchar(rnames))
    cat(sprintf("%s %s\n", formatC("", width = max_lw), hdr))

    for (i in seq_along(rnames)) {
      lab <- formatC(rnames[i], width = max_lw)
      if (has_bse) {
        cat(sprintf("%s %s %s %s %s %s %s\n",
                    lab, est[i], se[i], bse[i], zv[i], pv_str[i], stars[i]))
      } else {
        cat(sprintf("%s %s %s %s %s %s\n",
                    lab, est[i], se[i], zv[i], pv_str[i], stars[i]))
      }
    }

    cat("---\n")
    cat("Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1\n")

    # CI
    if (isTRUE(show_ci) && all(c("CI_low", "CI_high") %in% names(cf))) {
      level <- if (!is.null(x$wild.level)) x$wild.level else 0.95
      cat(sprintf("\n%.0f%% Bootstrap CI:\n", level * 100))
      ci_lo <- formatC(cf$CI_low, format = "f", digits = 3, width = 9)
      ci_hi <- formatC(cf$CI_high, format = "f", digits = 3, width = 9)
      for (i in seq_along(rnames)) {
        cat(sprintf("  %s: [%s, %s]\n", rnames[i], ci_lo[i], ci_hi[i]))
      }
    }
  }

  invisible(x)
}


# ============================================================
# nmfre.inference() - statistical inference on C (separated)
# ============================================================

#' @title Statistical inference for the coefficient matrix C from NMF-RE
#'
#' @description
#' \code{nmfre.inference} performs statistical inference on the coefficient
#' matrix \eqn{C} (\eqn{\Theta}) from a fitted \code{nmfre} model,
#' conditional on the estimated basis matrix \eqn{\hat{X}} and random
#' effects \eqn{\hat{U}}.
#'
#' Under the working model \eqn{Y^* = Y - X\hat{U} \approx X C A + \varepsilon},
#' inference is conducted via sandwich covariance estimation and
#' one-step wild bootstrap with non-negative projection.
#'
#' The result is compatible with \code{\link{nmfkc.DOT}} for visualization
#' (pass the result directly as \code{x} with \code{type = "YXA"}).
#'
#' @param object An object of class \code{"nmfre"} returned by
#'   \code{\link{nmfre}}.
#' @param Y Observation matrix (P x N). Must match the data used in
#'   \code{nmfre()}.
#' @param A Covariate matrix (K x N). Default is \code{NULL} (intercept only).
#' @param wild.bootstrap Logical. If \code{TRUE} (default), performs wild
#'   bootstrap for confidence intervals and bootstrap standard errors.
#' @param ... Additional arguments:
#'   \describe{
#'     \item{\code{wild.B}}{Number of bootstrap replicates. Default is 500.}
#'     \item{\code{wild.seed}}{Seed for bootstrap. Default is 123.}
#'     \item{\code{wild.level}}{Confidence level for bootstrap CI. Default is 0.95.}
#'     \item{\code{C.p.side}}{P-value type: \code{"one.sided"} (default) or \code{"two.sided"}.}
#'     \item{\code{cov.ridge}}{Ridge stabilization. Default is 1e-8.}
#'     \item{\code{print.trace}}{Logical. Default is \code{FALSE}.}
#'   }
#'
#' @return The input \code{object} with additional inference components:
#' \item{sigma2.used}{Estimated \eqn{\sigma^2} used for inference.}
#' \item{C.vec.cov}{Full covariance matrix for \eqn{vec(C)}.}
#' \item{C.se}{Sandwich standard errors for \eqn{C}.}
#' \item{C.se.boot}{Bootstrap standard errors for \eqn{C}.}
#' \item{C.ci.lower}{Lower CI bounds for \eqn{C}.}
#' \item{C.ci.upper}{Upper CI bounds for \eqn{C}.}
#' \item{coefficients}{Data frame with Basis, Covariate, Estimate, SE, BSE,
#'   z_value, p_value, CI_low, CI_high.}
#' \item{C.p.side}{P-value type used.}
#'
#' @seealso \code{\link{nmfre}}, \code{\link{nmfkc.DOT}},
#'   \code{\link{summary.nmfre}}
#' @references
#' Satoh, K. (2026). Wild Bootstrap Inference for Non-Negative Matrix
#'   Factorization with Random Effects. arXiv:2603.01468.
#'   \url{https://arxiv.org/abs/2603.01468}
#' @export
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(intercept = 1, speed = cars$speed)
#' res <- nmfre(Y, A, rank = 1, wild.bootstrap = FALSE)
#' res2 <- nmfre.inference(res, Y, A)
#' res2$coefficients
#'
nmfre.inference <- function(object, Y, A = NULL, wild.bootstrap = TRUE, ...) {
  if (!inherits(object, "nmfre"))
    stop("object must be of class 'nmfre' (returned by nmfre).")

  extra_args <- base::list(...)
  wild.B      <- if (!is.null(extra_args$wild.B))      extra_args$wild.B      else 500
  wild.seed   <- if (!is.null(extra_args$wild.seed))   extra_args$wild.seed   else 123
  wild.level  <- if (!is.null(extra_args$wild.level))  extra_args$wild.level  else 0.95
  C.p.side    <- if (!is.null(extra_args$C.p.side))    extra_args$C.p.side    else "one.sided"
  cov.ridge   <- if (!is.null(extra_args$cov.ridge))   extra_args$cov.ridge   else 1e-8
  print.trace <- if (!is.null(extra_args$print.trace)) extra_args$print.trace else FALSE

  Y <- base::as.matrix(Y)
  if (is.null(A)) A <- base::matrix(1, nrow = 1, ncol = base::ncol(Y))
  A <- base::as.matrix(A)

  X     <- object$X       # P x Q
  C_mat <- object$C       # Q x K
  U     <- object$U       # Q x N
  sigma2 <- object$sigma2
  tau2   <- object$tau2

  P <- base::nrow(Y)
  N <- base::ncol(Y)
  Q <- base::ncol(X)
  K <- base::nrow(A)

  lambda_inf <- sigma2 / base::pmax(tau2, 1e-12)

  # Y_star = Y - XU (remove random effects)
  Y_star_inf <- Y - X %*% U
  R_C <- Y_star_inf - X %*% (C_mat %*% A)
  RSS_inf <- base::sum(R_C^2)

  XtX_now <- base::crossprod(X)       # Q x Q
  AAt <- A %*% base::t(A)             # K x K

  M_inf <- XtX_now + base::diag(base::pmax(lambda_inf, 1e-12), Q)
  cholM_inf <- base::tryCatch(base::chol(M_inf), error = function(e) NULL)
  if (!is.null(cholM_inf)) {
    Minv <- base::chol2inv(cholM_inf)
  } else {
    Minv <- base::tryCatch(base::solve(M_inf), error = function(e) {
      if (base::requireNamespace("MASS", quietly = TRUE)) MASS::ginv(M_inf)
      else base::stop("Information matrix singular; install MASS package.")
    })
  }

  trH <- base::sum(base::diag(Minv %*% XtX_now))
  dfU_inf <- N * trH
  dfC <- Q * K
  denom <- base::max(P * N - dfU_inf - dfC, 1)
  sigma2_used <- RSS_inf / denom

  # Information matrix with (I-H) correction
  Xt_IH_X <- XtX_now - XtX_now %*% Minv %*% XtX_now
  Info_core <- base::kronecker(AAt, Xt_IH_X)
  Info <- Info_core / base::max(sigma2_used, 1e-12)
  Info <- Info + base::diag(cov.ridge, base::nrow(Info))

  Hinv <- base::tryCatch(base::solve(Info), error = function(e) {
    if (base::requireNamespace("MASS", quietly = TRUE)) MASS::ginv(Info)
    else base::stop("Information matrix singular; install MASS package.")
  })

  # Sandwich covariance
  V_sand <- NULL
  Xt <- base::t(X)
  J <- base::matrix(0, Q * K, Q * K)
  for (n in 1:N) {
    a_n <- A[, n, drop = FALSE]
    r_n <- R_C[, n, drop = FALSE]
    g_n <- Xt %*% r_n
    S_n <- -(g_n %*% base::t(a_n)) / base::max(sigma2_used, 1e-12)
    s_n <- base::as.vector(S_n)
    J <- J + base::tcrossprod(s_n)
  }
  if (N > 1) J <- (N / (N - 1)) * J    # CR1 correction
  V_sand <- Hinv %*% J %*% Hinv

  C.vec.cov <- V_sand
  se_vec <- base::sqrt(base::pmax(base::diag(C.vec.cov), 0))
  C.se <- base::matrix(se_vec, nrow = Q, ncol = K, byrow = FALSE)
  base::dimnames(C.se) <- base::dimnames(C_mat)

  # ---- Wild bootstrap (one-step Newton) ----
  C.se.boot  <- NULL
  C.ci.lower <- NULL
  C.ci.upper <- NULL

  if (base::isTRUE(wild.bootstrap)) {
    base::set.seed(wild.seed)
    score_mat <- base::matrix(0, Q * K, N)
    for (n in 1:N) {
      a_n <- A[, n, drop = FALSE]
      r_n <- R_C[, n, drop = FALSE]
      g_n <- Xt %*% r_n
      G_n <- -(g_n %*% base::t(a_n)) / base::max(sigma2_used, 1e-12)
      score_mat[, n] <- base::as.vector(G_n)
    }

    C_hat_vec <- base::as.vector(C_mat)
    C_boot <- base::matrix(NA_real_, nrow = Q * K, ncol = wild.B)
    for (b in 1:wild.B) {
      w <- stats::rexp(N, rate = 1) - 1
      grad_b <- base::as.vector(score_mat %*% w)
      c_b <- C_hat_vec - base::as.vector(Hinv %*% grad_b)
      c_b <- base::pmax(c_b, 0)
      C_boot[, b] <- c_b
    }

    alpha <- 1 - wild.level
    lo <- base::apply(C_boot, 1, stats::quantile, probs = alpha / 2, na.rm = TRUE, names = FALSE)
    hi <- base::apply(C_boot, 1, stats::quantile, probs = 1 - alpha / 2, na.rm = TRUE, names = FALSE)

    C.ci.lower <- base::matrix(lo, nrow = Q, ncol = K, byrow = FALSE)
    C.ci.upper <- base::matrix(hi, nrow = Q, ncol = K, byrow = FALSE)
    base::dimnames(C.ci.lower) <- base::dimnames(C_mat)
    base::dimnames(C.ci.upper) <- base::dimnames(C_mat)

    sd_vec <- base::apply(C_boot, 1, stats::sd, na.rm = TRUE)
    C.se.boot <- base::matrix(sd_vec, nrow = Q, ncol = K, byrow = FALSE)
    base::dimnames(C.se.boot) <- base::dimnames(C_mat)
  }

  # ---- Coefficients table (nmfkc.DOT compatible: Basis/Covariate) ----
  Estimate <- base::as.vector(C_mat)
  SE <- base::as.vector(C.se)
  BSE <- if (!is.null(C.se.boot)) base::as.vector(C.se.boot) else base::rep(NA_real_, base::length(Estimate))
  z_value <- base::ifelse(SE > 0, Estimate / SE, NA_real_)

  if (C.p.side == "one.sided") {
    p_value <- base::ifelse(base::is.finite(z_value), stats::pnorm(z_value, lower.tail = FALSE), NA_real_)
  } else {
    p_value <- base::ifelse(base::is.finite(z_value), 1 - stats::pchisq(z_value^2, df = 1), NA_real_)
  }

  rlabs <- if (!is.null(base::rownames(C_mat))) base::rownames(C_mat) else base::paste0("Basis", 1:Q)
  clabs <- if (!is.null(base::colnames(C_mat))) base::colnames(C_mat) else base::paste0("Cov", 1:K)

  coefficients <- base::data.frame(
    Basis     = base::rep(rlabs, times = K),
    Covariate = base::rep(clabs, each = Q),
    Estimate  = Estimate,
    SE        = SE,
    BSE       = BSE,
    z_value   = z_value,
    p_value   = p_value,
    CI_low    = if (!is.null(C.ci.lower)) base::as.vector(C.ci.lower) else NA_real_,
    CI_high   = if (!is.null(C.ci.upper)) base::as.vector(C.ci.upper) else NA_real_,
    row.names = NULL, stringsAsFactors = FALSE
  )

  if (print.trace) {
    base::message("  nmfre.inference: sandwich SE + wild bootstrap done.")
  }

  object$sigma2.used  <- sigma2_used
  object$C.vec.cov    <- C.vec.cov
  object$C.se         <- C.se
  object$C.se.hess    <- C.se
  object$C.se.boot    <- C.se.boot
  object$C.ci.lower   <- C.ci.lower
  object$C.ci.upper   <- C.ci.upper
  object$C.boot.sd    <- C.se.boot
  object$coefficients <- coefficients
  object$C.p.side     <- C.p.side
  return(object)
}


# ============================================================
# nmfre.dfU.scan() - dfU scan utility
# ============================================================

#' @title Scan dfU cap rates for NMF-RE
#' @description
#' Fits the NMF-RE model across a range of \code{dfU.cap.rate} values and
#' returns a diagnostic table showing the resulting effective degrees of freedom,
#' variance components, and convergence diagnostics for each rate.
#'
#' The dfU cap limits the effective degrees of freedom consumed by the random
#' effects \eqn{U}. The cap is computed as \code{rate * N * Q}, where \eqn{N}
#' is the number of observations and \eqn{Q} is the rank. A suitable rate is
#' one where the final \eqn{\mathrm{df}_U} is below the cap
#' (\code{safeguard = TRUE}) and the model has converged
#' (\code{converged = TRUE}).
#'
#' When called automatically by \code{\link{nmfre}} (i.e., \code{dfU.cap.rate = NULL}),
#' the minimum rate satisfying both \code{safeguard = TRUE} and
#' \code{converged = TRUE} is selected.
#'
#' @param rates Numeric vector of cap rates to scan (default \code{(1:10)/10}).
#' @param Y Observation matrix (P x N).
#' @param A Covariate matrix (K x N).
#' @param rank Integer. Rank of the basis matrix.
#'   For backward compatibility, \code{Q} is accepted via \code{...}.
#' @param X.init Initial basis matrix, or \code{NULL}.
#' @param C.init Initial coefficient matrix, or \code{NULL}.
#' @param U.init Initial random effects matrix, or \code{NULL}.
#' @param print.trace Logical. Print progress for each fit (default \code{FALSE}).
#' @param ... Additional arguments passed to \code{\link{nmfre}}.
#' @return An object of class \code{"nmfre.dfU.scan"} with two components:
#' \describe{
#'   \item{\code{table}}{A data frame with the following columns:
#'   \describe{
#'     \item{\code{rate}}{Cap rate used. The dfU cap is \code{rate * N * Q}.}
#'     \item{\code{dfU.cap}}{The dfU cap value (upper bound on effective degrees of freedom).}
#'     \item{\code{dfU}}{Final effective degrees of freedom for \eqn{U} at convergence.}
#'     \item{\code{safeguard}}{Logical. \code{TRUE} if the dfU cap is functioning
#'       as a safeguard (\code{dfU / dfU.cap < 0.99}): the cap prevents
#'       random-effects saturation without over-constraining \eqn{U}.
#'       \code{FALSE} if dfU is at or near the cap, indicating the cap is
#'       binding and the rate may be too small.}
#'     \item{\code{hit}}{Logical. \code{TRUE} if the cap was reached at least once
#'       during iteration, even if dfU later decreased below the cap.}
#'     \item{\code{converged}}{Logical. \code{TRUE} if the algorithm converged
#'       within the maximum number of iterations.}
#'     \item{\code{tau2}}{Final random effect variance \eqn{\hat{\tau}^2}.}
#'     \item{\code{sigma2}}{Final residual variance \eqn{\hat{\sigma}^2}.}
#'     \item{\code{ICC}}{Trace-based Intraclass Correlation Coefficient
#'       \eqn{\tau^2 \, \mathrm{tr}(X^\top X) /
#'       (\tau^2 \, \mathrm{tr}(X^\top X) + \sigma^2 P)}.
#'       See \code{\link{nmfre}} for details.}
#'   }}
#'   \item{\code{cap.rate}}{Optimal cap rate selected automatically.
#'     If rows with \code{safeguard = TRUE} and \code{hit = TRUE} exist,
#'     the maximum rate among them is chosen (safeguard activated but
#'     giving \eqn{U} the most freedom). Otherwise, the minimum rate
#'     with \code{safeguard = TRUE} and \code{hit = FALSE} is chosen.
#'     \code{NA} if no suitable rate is found.}
#' }
#'
#' When printed, only the \code{table} is displayed. Access \code{cap.rate}
#' directly from the returned object.
#' @seealso \code{\link{nmfre}}
#' @export
#' @examples
#' # Example 1. cars data (small maxit for speed)
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(intercept = 1, speed = cars$speed)
#' tab <- nmfre.dfU.scan(rates = c(0.1, 0.2), Y = Y, A = A, rank = 1, maxit = 1000)
#' print(tab)
#'
#' \donttest{
#' # Example 2. Orthodont data (nlme)
#' if (requireNamespace("nlme", quietly = TRUE)) {
#'   Y <- matrix(nlme::Orthodont$distance, 4, 27)
#'   male <- ifelse(nlme::Orthodont$Sex[seq(1, 108, 4)] == "Male", 1, 0)
#'   A <- rbind(intercept = 1, male = male)
#'   nmfre.dfU.scan(1:10/10, Y, A, rank = 1)
#' }}
#'
nmfre.dfU.scan <- function(
  rates = (1:10) / 10,
  Y, A, rank = NULL,
  X.init = NULL, C.init = NULL, U.init = NULL,
  print.trace = FALSE,
  ...
) {
  extra_scan <- base::list(...)
  if (!is.null(extra_scan$Q)) rank <- extra_scan$Q
  Q <- rank
  N <- ncol(Y)
  NQ <- N * Q

  one_run <- function(rate) {

    r <- nmfre(
      Y = Y, A = A, Q = Q,
      X.init = X.init, C.init = C.init, U.init = U.init,
      print.trace = print.trace,
      dfU.control = "cap",
      dfU.cap.rate = rate,
      ...
    )

    safeguard <- if (is.finite(r$dfU) && is.finite(r$dfU.cap) && r$dfU.cap > 0) {
      r$dfU / r$dfU.cap < 0.99
    } else {
      NA
    }
    data.frame(
      rate = rate,
      dfU.cap = r$dfU.cap,
      dfU = round(r$dfU, 2),
      safeguard = safeguard,
      hit = isTRUE(r$dfU.hit.cap),
      converged = isTRUE(r$converged),
      tau2 = round(r$tau2, 3),
      sigma2 = round(r$sigma2, 3),
      ICC = round(r$ICC, 4),
      stringsAsFactors = FALSE
    )
  }
  tbl <- do.call(rbind, lapply(rates, one_run))
  rownames(tbl) <- NULL

  # optimal rate selection
  sg_hit <- tbl[tbl$safeguard == TRUE & tbl$hit == TRUE & tbl$converged, ]
  sg_free <- tbl[tbl$safeguard == TRUE & tbl$hit == FALSE & tbl$converged, ]
  cap.rate <- if (nrow(sg_hit) > 0) {
    max(sg_hit$rate)
  } else if (nrow(sg_free) > 0) {
    min(sg_free$rate)
  } else {
    NA_real_
  }

  result <- list(table = tbl, cap.rate = cap.rate)
  class(result) <- "nmfre.dfU.scan"
  result
}

#' @seealso \code{\link{nmfre.dfU.scan}}
#' @export
print.nmfre.dfU.scan <- function(x, ...) {
  print(x$table, ...)
  invisible(x)
}

# =====================================================
# nmfkc.signed.rff.R — Random Fourier Features generator
#                      for use with nmfkc.signed()
#
# Given training covariates U (p x N), generates the signed RFF
# feature matrix
#     Z = sqrt(2 / D) * cos(omega U + b),       D x N,   Z^T Z ~ K,
# where omega_d ~ N(0, 2 beta I_p) and b_d ~ Uniform(0, 2 pi)
# (Rahimi & Recht, 2007).  The result is a list containing the
# sign-unrestricted feature matrix Z and the generating parameters
# pars = list(omega, b, D, beta), so that the same random map can be
# re-applied to new data and nmfkc.signed() can store pars for
# downstream summary.
#
# Design: kept intentionally minimal.  The kernel-faithful direct MU
# algorithm and all S3 methods (predict, plot, summary) live in
# nmfkc.signed.R; this file only provides the feature map.
#
# References:
#   Rahimi, A. & Recht, B. (2007). Random features for large-scale
#     kernel machines. NIPS 20.
# =====================================================


#' Random Fourier Features for nmfkc.signed()
#'
#' @description
#' Generates RFF random parameters
#' \eqn{\omega_d \sim \mathcal{N}(0, 2\beta I_p)},
#' \eqn{b_d \sim \mathrm{Uniform}(0, 2\pi)} (Rahimi & Recht, 2007) and
#' applies the RFF transform
#' \deqn{z_d(u) = \sqrt{2/D}\, \cos(\omega_d^\top u + b_d)}
#' to each column of \code{U}, yielding a sign-unrestricted
#' \eqn{D \times N} feature matrix \eqn{Z} such that
#' \eqn{Z^\top Z \approx K}, the Gaussian kernel matrix with bandwidth
#' \code{beta}.
#'
#' The return value is a list with the feature matrix \code{Z} and
#' the generating parameters \code{pars = list(omega, b, D, beta)} so
#' that the same random map can be re-applied to new data (by passing
#' \code{pars} back) and \code{\link{nmfkc.signed}} can record the
#' parameters for downstream summary.
#'
#' @param U A \eqn{p \times N} numeric matrix; columns are data points.
#' @param beta Positive scalar. Gaussian kernel bandwidth parameter.
#'   Can be obtained via \code{\link{nmfkc.kernel.beta.nearest.med}}.
#'   May be \code{NULL} only when \code{pars} is supplied through
#'   \code{...}.
#' @param D Integer. Number of random features. Defaults to
#'   \code{ceiling(ncol(U) / 2)}.  This default is intended for the
#'   \strong{training-time fresh generation} only; for test data, always
#'   supply \code{pars} via \code{...} to inherit the training \eqn{D}
#'   together with \eqn{\omega, b}.  For very large \eqn{N} the default
#'   may be excessive (direct MU cost is \eqn{O(QD^2)}); choose a
#'   smaller \eqn{D} manually.  For very small \eqn{N}, RFF is not
#'   recommended; use a full kernel matrix with \code{\link{nmfkc}}
#'   instead.
#' @param seed Optional integer passed to \code{set.seed()} before
#'   generating \eqn{\omega, b}, for reproducibility.  Ignored when
#'   \code{pars} is supplied.
#' @param ... Hidden option \code{pars}: a list
#'   \code{list(omega, b, D, beta)} obtained from a previous call
#'   (\code{Ztrain$pars}).  When supplied, \eqn{\omega, b} are reused
#'   and the \code{beta}, \code{D}, \code{seed} arguments are ignored.
#'   Use this to apply the same random map to test data.
#'
#' @return A list with two elements:
#'   \describe{
#'     \item{\code{Z}}{A \eqn{D \times N} sign-unrestricted numeric
#'       matrix.  Pass this to \code{\link{nmfkc.signed}} as its
#'       \code{A} argument.}
#'     \item{\code{pars}}{A list \code{list(omega, b, D, beta)}.  Pass
#'       this to \code{\link{nmfkc.signed}} via its \code{pars}
#'       argument (for summary display) and to subsequent
#'       \code{nmfkc.signed.rff()} calls (to reuse the same random map
#'       on new data).}
#'   }
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @seealso \code{\link{nmfkc.signed}},
#'   \code{\link{nmfkc.kernel.beta.nearest.med}}
#'
#' @examples
#' \donttest{
#' ## Iris 3-class classification with RFF + direct MU (Ding-Li-Jordan)
#' data(iris)
#' set.seed(1)
#' idx <- sample(nrow(iris), 100)    # 100 training, 50 test
#'
#' ## Scale features using TRAINING mean/sd; transpose to p x N layout
#' mn      <- colMeans(iris[idx, 1:4])
#' sc      <- apply(iris[idx, 1:4], 2, sd)
#' U.train <- t(scale(iris[idx,  1:4], center = mn, scale = sc))   # 4 x 100
#' U.test  <- t(scale(iris[-idx, 1:4], center = mn, scale = sc))   # 4 x  50
#'
#' ## One-hot encode training labels as a Q_obs x N target matrix
#' levs      <- levels(iris$Species)
#' Y.train   <- sapply(iris$Species[idx], function(s) as.integer(levs == s))
#' rownames(Y.train) <- levs                                       # 3 x 100
#' lab.train <- iris$Species[idx]
#' lab.test  <- iris$Species[-idx]
#'
#' ## Beta candidates from nearest-neighbour median heuristic
#' beta_info <- nmfkc.kernel.beta.nearest.med(U.train)
#' betas     <- beta_info$beta_candidates
#'
#' ## CV over beta candidates: for each beta, generate RFF, fit, and
#' ## evaluate column-wise CV-MSE on training data
#' cv_mse <- numeric(length(betas))
#' for (i in seq_along(betas)) {
#'   rff_i   <- nmfkc.signed.rff(U.train, beta = betas[i], D = 50, seed = 1)
#'   cv_i    <- nmfkc.signed.cv(Y.train, A = rff_i$Z, rank = 3, seed = 123)
#'   cv_mse[i] <- cv_i$objfunc
#' }
#' beta_best <- betas[which.min(cv_mse)]
#'
#' ## Generate signed RFF features with the best beta
#' rff.train <- nmfkc.signed.rff(U.train, beta = beta_best, D = 50, seed = 1)
#' rff.test  <- nmfkc.signed.rff(U.test, pars = rff.train$pars)
#'
#' ## Fit on training data only
#' res <- nmfkc.signed(Y.train, A = rff.train$Z, rank = 3,
#'                     pars = rff.train$pars, verbose = FALSE)
#'
#' ## Predict on training and test separately
#' pred.train <- predict(res, newA = rff.train$Z, type = "class")
#' pred.test  <- predict(res, newA = rff.test$Z,  type = "class")
#' mean(pred.train == as.character(lab.train))
#' mean(pred.test  == as.character(lab.test))
#' }
#'
#' @export
nmfkc.signed.rff <- function(U, beta = NULL,
                             D = ceiling(ncol(U) / 2),
                             seed = NULL, ...) {
  extra <- list(...)
  pars  <- extra$pars

  if (is.null(pars)) {
    if (is.null(beta))
      stop("'beta' must be specified (or supply 'pars' via ...).")
    if (!is.null(seed)) set.seed(seed)
    p <- nrow(U)
    pars <- list(
      omega = matrix(stats::rnorm(D * p, mean = 0, sd = sqrt(2 * beta)),
                     nrow = D, ncol = p),
      b     = stats::runif(D, 0, 2 * pi),
      D     = D,
      beta  = beta
    )
  }

  Z <- sqrt(2 / pars$D) * cos(pars$omega %*% U + pars$b)
  list(Z = Z, pars = pars)
}

# =====================================================
# nmfkc.net.R — Network analysis for symmetric NMF
# =====================================================


############################################################
## nmfkc.net.inference
############################################################

#' Statistical Inference for Symmetric NMF Parameters
#'
#' @description
#' Performs statistical inference on the parameter matrix \eqn{C} of a
#' symmetric NMF model (\eqn{Y \approx XCX^\top}).
#'
#' This is a wrapper around \code{\link{nmfkc.inference}} that automatically
#' sets the covariate matrix to \eqn{A = X^\top}, which is the defining
#' property of symmetric NMF.
#'
#' @param object A fitted model returned by \code{nmfkc()} with
#'   \code{Y.symmetric = "bi"} or \code{"tri"}, or the newer
#'   \code{\link{nmfkc.net}} with \code{type = "bi"} or \code{"tri"}.
#' @param Y The original symmetric matrix (P x P).
#' @param wild.bootstrap Logical; if TRUE (default), perform wild bootstrap
#'   inference for the C matrix.
#' @param ... Additional arguments passed to \code{nmfkc.inference()}
#'   (e.g., \code{wild.B}, \code{wild.seed}, \code{C.p.side}).
#'
#' @return The object augmented with inference fields
#'   (same as \code{nmfkc.inference}), with class
#'   \code{c("nmfkc.net.inference", "nmfkc.inference", "nmf.inference", "nmfkc", "nmf")}.
#'
#' @seealso \code{\link{nmfkc.inference}}, \code{\link{nmfkc.net.DOT}},
#'   \code{\link{summary.nmfkc.net.inference}}
#'
#' @examples
#' \donttest{
#' library(nmfkc)
#' Y <- matrix(c(0,1,1,0,0,0,
#'               1,0,1,0,0,0,
#'               1,1,0,1,0,0,
#'               0,0,1,0,1,1,
#'               0,0,0,1,0,1,
#'               0,0,0,1,1,0), 6, 6)
#' res <- nmfkc.net(Y, rank = 2, type = "tri", nstart = 20)
#' res_inf <- nmfkc.net.inference(res, Y)
#' summary(res_inf)
#' }
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @export
nmfkc.net.inference <- function(object, Y, wild.bootstrap = TRUE, ...) {
  if (!inherits(object, "nmfkc"))
    stop("object must be of class 'nmfkc'.")
  if (is.null(object$X))
    stop("object must contain X matrix.")

  ## Key: A = t(X) for symmetric NMF (Y = X C X')
  A <- t(object$X)
  result <- nmfkc.inference(object, Y, A = A, wild.bootstrap = wild.bootstrap, ...)

  ## Relabel coefficients for symmetric context: Covariate -> Basis.col
  if (!is.null(result$coefficients)) {
    cf <- result$coefficients
    names(cf)[names(cf) == "Basis"]     <- "Basis.row"
    names(cf)[names(cf) == "Covariate"] <- "Basis.col"

    ## Keep only strict upper triangle (off-diagonal) since C is symmetric
    ## Diagonal (within-group) is always large and trivially significant;
    ## only between-group interactions are meaningful to test.
    Q <- ncol(object$X)
    rlabs <- if (!is.null(rownames(object$C))) rownames(object$C)
             else paste0("Basis", 1:Q)
    clabs <- if (!is.null(colnames(object$C))) colnames(object$C)
             else paste0("Basis", 1:Q)
    row_idx <- match(cf$Basis.row, rlabs)
    col_idx <- match(cf$Basis.col, clabs)
    keep <- !is.na(row_idx) & !is.na(col_idx) & (row_idx < col_idx)
    cf <- cf[keep, , drop = FALSE]
    ## Unify Basis.col labels to match Basis.row naming
    cf$Basis.col <- rlabs[col_idx[keep]]
    rownames(cf) <- NULL
    result$coefficients <- cf
  }

  class(result) <- c("nmfkc.net.inference", "nmfkc.net",
                      "nmfkc.inference", "nmf.inference", "nmfkc", "nmf")
  result
}


############################################################
## summary / print for nmfkc.net.inference
############################################################

#' @title Summary method for nmfkc.net.inference objects
#' @description
#' Produces a summary of a symmetric NMF model with inference results,
#' including the coefficients table for \eqn{C}.
#'
#' @param object An object of class \code{"nmfkc.net.inference"}.
#' @param ... Additional arguments passed to \code{summary.nmfkc}.
#' @return An object of class \code{"summary.nmfkc.net.inference"}.
#' @seealso \code{\link{nmfkc.net.inference}}
#' @export
summary.nmfkc.net.inference <- function(object, ...) {
  ans <- summary.nmfkc(object, ...)
  ans$sigma2.used  <- object$sigma2.used
  ans$coefficients <- object$coefficients
  ans$C.p.side     <- object$C.p.side
  ans$C            <- object$C
  ans$Y.symmetric  <- tryCatch(eval(object$call$Y.symmetric),
                               error = function(e) NULL)
  class(ans) <- "summary.nmfkc.net.inference"
  ans
}


#' @title Print method for summary.nmfkc.net.inference objects
#' @description
#' Prints a formatted summary of a symmetric NMF model with inference results.
#' The coefficients table uses \code{Basis.row} and \code{Basis.col} labels
#' reflecting the symmetric structure \eqn{C_{qr}}.
#'
#' @param x An object of class \code{"summary.nmfkc.net.inference"}.
#' @param digits Minimum number of significant digits.
#' @param ... Additional arguments (currently unused).
#' @return Called for its side effect (printing). Returns \code{x} invisibly.
#' @seealso \code{\link{summary.nmfkc.net.inference}}
#' @export
print.summary.nmfkc.net.inference <- function(x,
    digits = max(3L, getOption("digits") - 3L), ...) {

  # Print base summary
  print.summary.nmfkc(x, digits = digits, ...)

  # Model type
  sym_type <- if (!is.null(x$Y.symmetric)) x$Y.symmetric else "unknown"
  cat(sprintf("Symmetric NMF type: %s\n", sym_type))

  # Inference section
  cat("Inference (conditional on X):\n")
  cat("  sigma^2:  ", format(x$sigma2.used, digits = digits), "\n")

  # C matrix
  if (!is.null(x$C)) {
    Q <- nrow(x$C)
    cat(sprintf("  C matrix:  %d x %d", Q, Q))
    if (sym_type == "tri") {
      sym_check <- max(abs(x$C - t(x$C)))
      cat(sprintf(" (max asymmetry: %.1e)", sym_check))
    }
    cat("\n")
  }

  # Coefficients table
  if (!is.null(x$coefficients) && is.data.frame(x$coefficients)) {
    cf <- x$coefficients

    p_side <- if (!is.null(x$C.p.side)) x$C.p.side else "one.sided"
    p_header <- if (p_side == "one.sided") "Pr(>z)" else "Pr(>|z|)"

    sig_stars <- function(p) {
      ifelse(!is.finite(p), " ",
        ifelse(p < 0.001, "***",
          ifelse(p < 0.01, "**",
            ifelse(p < 0.05, "*",
              ifelse(p < 0.1, ".", " ")))))
    }
    format_pval <- function(p) {
      ifelse(!is.finite(p), "      NA",
        ifelse(p < 2.2e-16, "  <2e-16",
          formatC(p, format = "g", digits = 4, width = 8)))
    }

    n_total <- nrow(cf)
    n_sig <- sum(cf$p_value < 0.05, na.rm = TRUE)
    cat(sprintf("\nC coefficients: %d total, %d significant\n",
                n_total, n_sig))

    rnames <- paste0(cf$Basis.row, ":", cf$Basis.col)

    est <- formatC(cf$Estimate, format = "f", digits = 3, width = 9)
    se  <- formatC(cf$SE, format = "f", digits = 3, width = 10)
    bse <- formatC(cf$BSE, format = "f", digits = 3, width = 6)
    zv  <- formatC(cf$z_value, format = "f", digits = 2, width = 7)
    pv_str <- format_pval(cf$p_value)
    stars <- sig_stars(cf$p_value)

    max_lw <- max(nchar(rnames))
    hdr <- sprintf("%s %s %s %s %s %s",
                   formatC("Estimate", width = 9),
                   formatC("Std. Error", width = 10),
                   formatC("(Boot)", width = 6),
                   formatC("z value", width = 7),
                   formatC(p_header, width = 8), "")
    cat(sprintf("%s %s\n", formatC("Row:Col", width = max_lw), hdr))
    for (i in seq_len(n_total)) {
      cat(sprintf("%s %s %s %s %s %s %s\n",
                  formatC(rnames[i], width = max_lw),
                  est[i], se[i], bse[i], zv[i], pv_str[i], stars[i]))
    }
    cat("---\n")
    cat("Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1\n")
  }

  cat("\n")
  invisible(x)
}


############################################################
## nmfkc.net.DOT
############################################################

#' Generate a Graphviz DOT Diagram for a Symmetric NMF Network
#'
#' @description
#' Creates a Graphviz DOT script that visualizes the two-layer structure
#' of a symmetric NMF model (\eqn{Y \approx X C X^\top}).
#'
#' The resulting diagram displays:
#' \itemize{
#'   \item outer nodes: the original network nodes (rows/columns of Y),
#'   \item inner nodes: the latent basis/group nodes (columns of X),
#'   \item directed edges from basis to node: membership weights from X,
#'   \item undirected edges between basis nodes: inter-group interactions
#'         from C matrix (for tri-symmetric models only).
#' }
#'
#' @param result A list returned by \code{nmfkc()} with
#'   \code{Y.symmetric = "bi"} or \code{"tri"}, or the newer
#'   \code{\link{nmfkc.net}} with \code{type = "bi"} or \code{"tri"}.
#'   If inference results are present (from \code{\link{nmfkc.net.inference}}),
#'   C edges are decorated with significance stars.
#' @param threshold Minimum coefficient value to display an edge.
#' @param sig.level Significance level for filtering C edges (if inference
#'   results are present). Set to \code{NULL} to show all edges above threshold.
#' @param weight_scale Base scaling factor for edge widths.
#' @param weight_scale_xy Scaling factor for X edges (basis -> node).
#' @param weight_scale_xx Scaling factor for C edges (basis <-> basis).
#' @param rankdir Graphviz rank direction (\code{"TB"} or \code{"LR"}).
#' @param fill Logical; whether nodes are filled with color.
#' @param hide.isolated Logical; if TRUE, omit outer nodes with no X edge
#'   above threshold.
#' @param Y.label Character vector of labels for outer nodes.
#' @param X.label Character vector of labels for basis nodes.
#' @param Y.title Cluster title for outer nodes.
#' @param X.title Cluster title for basis nodes.
#' @param show.theta Logical or NULL. Whether to draw C edges
#'   between basis nodes. NULL = auto-detect (TRUE for tri, FALSE for bi).
#' @param signed Logical. If \code{TRUE}, \eqn{C} is treated as a signed
#'   matrix: positive entries rendered as solid edges and negative as
#'   dashed, with edge visibility threshold on \eqn{|C|}.  Default is
#'   \code{inherits(result, "nmfkc.net.signed")} so that results from
#'   \code{\link{nmfkc.net}} are auto-detected.
#' @param cluster.box Style of cluster box: \code{"none"}, \code{"normal"},
#'   \code{"faint"}, \code{"invisible"}.
#' @param layout Graphviz layout engine, in recommended order:
#'   \code{"neato"} (default; spring model, clearest for small/medium community
#'   graphs), \code{"fdp"} (force-directed, scales to larger graphs),
#'   \code{"twopi"} (radial), \code{"circo"} (circular), \code{"dot"}
#'   (hierarchical).  For community networks, \code{"neato"} or \code{"fdp"} with
#'   a raised \code{threshold} (e.g.\ 0.2--0.3) separate the groups best.
#' @param X.color Color palette for basis nodes (length Q).
#' @param Y.cluster Coloring mode for outer nodes: \code{"soft"} (weighted mix)
#'   or \code{"hard"} (most probable basis color).
#'
#' @return A character string of class \code{c("nmfkc.net.DOT", "nmfkc.DOT")}
#'   representing a valid Graphviz DOT script.
#'   Use \code{plot()} to render (requires DiagrammeR).
#'
#' @seealso \code{\link{nmfkc.net.inference}}, \code{\link{nmfkc.DOT}},
#'   \code{\link{plot.nmfkc.DOT}}
#'
#' @examples
#' \donttest{
#' library(nmfkc)
#' Y <- matrix(c(0,1,1,0,0,0,
#'               1,0,1,0,0,0,
#'               1,1,0,1,0,0,
#'               0,0,1,0,1,1,
#'               0,0,0,1,0,1,
#'               0,0,0,1,1,0), 6, 6)
#' res <- nmfkc.net(Y, rank = 2, type = "tri", nstart = 20)
#' dot <- nmfkc.net.DOT(res)
#' plot(dot)
#' }
#'
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @export
nmfkc.net.DOT <- function(
    result,
    threshold       = 0.01,
    sig.level       = 0.1,
    weight_scale    = 5,
    weight_scale_xy = 1,
    weight_scale_xx = weight_scale,
    rankdir         = "TB",
    fill            = TRUE,
    hide.isolated   = TRUE,
    Y.label         = NULL,
    X.label         = NULL,
    Y.title         = "Nodes (Y)",
    X.title         = "Basis (X)",
    show.theta      = NULL,
    signed          = inherits(result, "nmfkc.net.signed"),
    cluster.box     = c("none", "normal", "faint", "invisible"),
    layout          = c("neato", "fdp", "twopi", "circo", "dot"),
    X.color         = NULL,
    Y.cluster       = c("soft", "hard")
) {

  ## ---------------------------------------------------------
  ## Extract matrices
  ## ---------------------------------------------------------
  X <- result$X          # P x Q
  C_mat <- result$C      # Q x Q

  if (is.null(X)) stop("result must contain X (basis matrix).")
  if (is.null(C_mat)) stop("result must contain C (parameter matrix).")

  P <- nrow(X)
  Q <- ncol(X)

  ## ---------------------------------------------------------
  ## Auto-detect bi vs tri model
  ## ---------------------------------------------------------
  if (is.null(show.theta)) {
    if (!is.null(result$type)) {
      ## New nmfkc.net API: $type is "bi" / "tri" / "signed".  Only "bi"
      ## has C = I (no inter-class interaction), so it draws no C edges.
      show.theta <- (result$type != "bi")
    } else {
      sym_arg <- tryCatch(eval(result$call$Y.symmetric), error = function(e) NULL)
      if (!is.null(sym_arg)) {
        show.theta <- (sym_arg == "tri")
      } else {
        ## Legacy fallback: detect C = I.  Drop dimnames first, otherwise
        ## all.equal() reports a names mismatch and mis-detects bi as tri.
        is_identity <- (all(dim(C_mat) == c(Q, Q)) &&
                        isTRUE(all.equal(unname(C_mat), diag(Q),
                                         tolerance = 1e-6)))
        show.theta <- !is_identity
      }
    }
  }

  ## Symmetrize C (numerical optimization may leave tiny asymmetry)
  if (show.theta) {
    C_mat <- (C_mat + t(C_mat)) / 2
  }

  ## ---------------------------------------------------------
  ## Argument matching
  ## ---------------------------------------------------------
  layout      <- match.arg(layout)
  Y.cluster   <- match.arg(Y.cluster)
  cluster.box <- match.arg(cluster.box)

  cluster_style <- switch(cluster.box,
                          normal    = "rounded",
                          faint     = "rounded,dashed",
                          invisible = "rounded",
                          none      = "none")
  cluster_color <- switch(cluster.box,
                          normal    = "black",
                          faint     = "gray80",
                          invisible = "none",
                          none      = "none")

  ## ---------------------------------------------------------
  ## Labels
  ## ---------------------------------------------------------
  Y_labels <- if (!is.null(Y.label)) Y.label else rownames(X)
  X_labels <- if (!is.null(X.label)) X.label else colnames(X)
  if (is.null(Y_labels)) Y_labels <- as.character(seq_len(P))
  if (is.null(X_labels)) X_labels <- paste0("Basis", seq_len(Q))

  Y_ids <- .nmfkc_dot_sanitize_id(paste0("Y_", seq_len(P)))
  X_ids <- .nmfkc_dot_sanitize_id(paste0("X_", seq_len(Q)))

  ## ---------------------------------------------------------
  ## Filter isolated outer nodes
  ## ---------------------------------------------------------
  idx_N <- seq_len(P)
  if (isTRUE(hide.isolated)) {
    used <- apply(X, 1L, function(row) any(row >= threshold, na.rm = TRUE))
    idx_N <- which(used)
  }

  ## ---------------------------------------------------------
  ## Membership probability: H (P x Q, each row sums to 1)
  ## New nmfkc.net objects carry $X.prob; legacy nmfkc(Y.symmetric=...)
  ## objects carry $B.prob (Q x P) which we transpose.
  ## ---------------------------------------------------------
  if (!is.null(result$X.prob)) {
    H <- result$X.prob
  } else if (!is.null(result$B.prob)) {
    H <- t(result$B.prob)
  } else if (!is.null(result$B)) {
    B_raw <- result$B
    cs <- colSums(B_raw); cs[cs == 0] <- 1
    H <- t(sweep(B_raw, 2, cs, "/"))
  } else {
    stop("result must contain X.prob, B.prob, or B.")
  }

  ## ---------------------------------------------------------
  ## Colors: basis nodes and outer nodes
  ## ---------------------------------------------------------
  if (is.null(X.color)) X.color <- grDevices::hcl.colors(Q, "Set 2")
  if (length(X.color) < Q) X.color <- rep_len(X.color, Q)

  basis_rgb <- t(grDevices::col2rgb(X.color)) / 255  # Q x 3

  node_hex <- character(P)
  if (Y.cluster == "soft") {
    for (i in seq_len(P)) {
      w <- H[i, ]
      mixed <- as.vector(w %*% basis_rgb)
      mixed <- pmin(pmax(mixed, 0), 1)
      node_hex[i] <- grDevices::rgb(mixed[1], mixed[2], mixed[3])
    }
  } else {
    for (i in seq_len(P)) {
      node_hex[i] <- X.color[which.max(H[i, ])]
    }
  }

  ## ---------------------------------------------------------
  ## DOT header
  ## ---------------------------------------------------------
  scr <- .nmfkc_dot_header(graph_name = "NMF_NET", rankdir = rankdir)
  scr <- paste0(scr, '  layout=', layout, ';\n',
                '  start=123;\n',
                '  node [fontsize=8];\n')

  ## ---------------------------------------------------------
  ## Basis node cluster
  ## ---------------------------------------------------------
  scr <- paste0(scr, '\n  // Basis nodes (latent groups)\n')
  if (!identical(cluster_style, "none")) {
    scr <- paste0(scr,
      sprintf('  subgraph cluster_Basis{label="%s" style="%s" color="%s" penwidth=1.0;\n',
              X.title, cluster_style, cluster_color))
  }
  for (q in seq_len(Q)) {
    scr <- paste0(scr,
      sprintf('  %s [label="%s", shape=ellipse, style="filled,rounded", fillcolor="%s"];\n',
              X_ids[q], X_labels[q], X.color[q]))
  }
  if (!identical(cluster_style, "none")) {
    scr <- paste0(scr, '  }\n')
  }

  ## ---------------------------------------------------------
  ## Outer node cluster
  ## ---------------------------------------------------------
  scr <- paste0(scr, '\n  // Original network nodes\n')
  if (!identical(cluster_style, "none")) {
    scr <- paste0(scr,
      sprintf('  subgraph cluster_Node{label="%s" style="%s" color="%s" penwidth=1.0;\n',
              Y.title, cluster_style, cluster_color))
  }
  for (i in idx_N) {
    if (fill) {
      scr <- paste0(scr,
        sprintf('  %s [label="%s", shape=circle, style="filled,rounded", fillcolor="%s", color=black, penwidth=0.5];\n',
                Y_ids[i], Y_labels[i], node_hex[i]))
    } else {
      scr <- paste0(scr,
        sprintf('  %s [label="%s", shape=circle, style="rounded", color=black, penwidth=0.5];\n',
                Y_ids[i], Y_labels[i]))
    }
  }
  if (!identical(cluster_style, "none")) {
    scr <- paste0(scr, '  }\n')
  }

  ## ---------------------------------------------------------
  ## Edge defaults
  ## ---------------------------------------------------------
  scr <- paste0(scr,
    '\n  edge [fontname="Arial", fontsize=8, arrowhead=open];\n')

  pw     <- .nmfkc_dot_penwidth
  digits <- .nmfkc_dot_digits_from_threshold(threshold)
  fmtc   <- function(x) .nmfkc_dot_format_coef(x, digits)

  ## ---------------------------------------------------------
  ## X edges: Basis -> Node (membership probability)
  ## ---------------------------------------------------------
  scr <- paste0(scr,
    '\n  // Membership edges: Basis -> Node (B.prob)\n',
    '  edge [color="gray0", fontcolor="gray0", style=solid];\n')

  if (max(H, na.rm = TRUE) > 0) {
    for (i in idx_N) {
      for (q in seq_len(Q)) {
        val <- H[i, q]
        if (is.finite(val) && val >= threshold) {
          pen <- pw(val, 1, weight_scale_xy)
          lab <- sprintf("%.2f", val)
          scr <- paste0(scr,
            sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                    X_ids[q], Y_ids[i], lab, pen))
        }
      }
    }
  }

  ## ---------------------------------------------------------
  ## Significance stars for C edges (from nmfkc.net.inference)
  ## ---------------------------------------------------------
  C_stars <- NULL
  C_show  <- NULL
  if (!is.null(result$coefficients) && show.theta) {
    C_stars <- matrix("", nrow = Q, ncol = Q)
    C_pval  <- matrix(NA_real_, nrow = Q, ncol = Q)
    cf <- result$coefficients

    ## Column names depend on whether inference was via nmfkc.net.inference
    row_col <- if ("Basis.row" %in% names(cf)) "Basis.row" else "Basis"
    col_col <- if ("Basis.col" %in% names(cf)) "Basis.col" else "Covariate"

    rlabs <- if (!is.null(rownames(result$C))) rownames(result$C)
             else paste0("Basis", 1:Q)
    clabs <- if (!is.null(colnames(result$C))) colnames(result$C)
             else paste0("Basis", 1:Q)

    for (k in seq_len(nrow(cf))) {
      q <- match(cf[[row_col]][k], rlabs)
      r <- match(cf[[col_col]][k], clabs)
      if (!is.na(q) && !is.na(r) && !is.na(cf$p_value[k])) {
        p <- cf$p_value[k]
        C_pval[q, r] <- p
        if      (p < 0.001) C_stars[q, r] <- "***"
        else if (p < 0.01)  C_stars[q, r] <- "**"
        else if (p < 0.05)  C_stars[q, r] <- "*"
      }
    }
    if (!is.null(sig.level)) {
      C_show <- !is.na(C_pval) & C_pval < sig.level
    }
  }

  ## ---------------------------------------------------------
  ## C edges: Basis <-> Basis (inter-group interaction, tri only)
  ## When `signed = TRUE`: positive C rendered solid, negative C dashed.
  ## When `signed = FALSE` (default for non-signed): all edges dashed.
  ## ---------------------------------------------------------
  if (isTRUE(show.theta) && Q >= 2) {
    default_style <- if (isTRUE(signed)) "solid" else "dashed"
    scr <- paste0(scr,
      '\n  // Inter-basis interaction edges (C, off-diagonal)\n',
      sprintf('  edge [color=black, fontcolor=black, style=%s, dir=none];\n',
              default_style))

    ## Visibility threshold: abs(val) >= threshold for signed; val >= threshold otherwise
    visible_val <- function(v) {
      if (isTRUE(signed)) abs(v) else v
    }
    max_C_off <- suppressWarnings(
      max(visible_val(C_mat[upper.tri(C_mat)]), na.rm = TRUE))
    if (is.finite(max_C_off) && max_C_off > 0) {
      for (q in seq_len(Q - 1)) {
        for (r in (q + 1):Q) {
          val <- C_mat[q, r]
          v_show <- visible_val(val)
          show_edge <- if (!is.null(C_show)) C_show[q, r]
                       else is.finite(v_show) && v_show >= threshold
          if (show_edge) {
            pen <- pw(v_show, max_C_off, weight_scale_xx)
            lab <- fmtc(val)
            if (!is.null(C_stars)) lab <- paste0(lab, C_stars[q, r])
            if (isTRUE(signed) && val < 0) {
              ## Negative C: dashed style, overriding the solid default
              scr <- paste0(scr,
                sprintf('  %s -> %s [label="%s", penwidth=%.2f, style=dashed, dir=none];\n',
                        X_ids[q], X_ids[r], lab, pen))
            } else {
              scr <- paste0(scr,
                sprintf('  %s -> %s [label="%s", penwidth=%.2f, dir=none];\n',
                        X_ids[q], X_ids[r], lab, pen))
            }
          }
        }
      }
    }
  }

  ## ---------------------------------------------------------
  ## Close and return
  ## ---------------------------------------------------------
  result <- paste0(scr, "}\n")
  class(result) <- c("nmfkc.net.DOT", "nmfkc.DOT")
  result
}


.nmfkc.net.run_once <- function(Y, X, C, maxit, epsilon,
                                X.restriction, C.L1, X.L2.ortho,
                                Wmat, has.weights) {
  N <- nrow(Y); Q <- ncol(X)
  small <- 1e-16
  norm_XC <- function(X, C) {
    if (X.restriction %in% c("fixed", "none")) return(list(X = X, C = C))
    cs <- colSums(X) + small
    if (X.restriction == "colSums") {
      d <- cs
    } else if (X.restriction == "colSqSums") {
      d <- sqrt(colSums(X * X)) + small
    } else d <- cs
    X <- sweep(X, 2, d, "/")
    C <- sweep(C, 1, d, "*"); C <- sweep(C, 2, d, "*")
    list(X = X, C = C)
  }
  cc <- norm_XC(X, C); X <- cc$X; C <- cc$C
  compute_obj <- function(X, C) {
    Yhat <- X %*% C %*% t(X)
    ## lm()-style weighted least squares: L = sum(W * (Y - Yhat)^2).
    ## W = W^2 for binary {0,1} masks, so this is unchanged for ECV.
    if (has.weights) sum(Wmat * (Y - Yhat)^2) else sum((Y - Yhat)^2)
  }
  obj_prev <- compute_obj(X, C)
  objfunc.iter <- numeric(maxit)
  iter <- 0L
  for (iter in seq_len(maxit)) {
    tX <- t(X)

    ## ---- C update (symmetric tri) ----
    if (has.weights) {
      Yhat <- X %*% C %*% tX
      num_C <- tX %*% (Wmat * Y) %*% X
      den_C <- tX %*% (Wmat * Yhat) %*% X
    } else {
      num_C <- tX %*% Y %*% X
      P <- tX %*% X
      den_C <- P %*% C %*% P
    }
    if (C.L1 > 0) den_C <- den_C + (C.L1 / 2)
    C <- C * (num_C / (den_C + small))

    ## ---- X update (bilateral gradient for Y = X C X^T, symmetric C) ----
    ##   num = (W*Y) X (C + C^T)
    ##   den = (W*Yhat) X (C + C^T)
    tX <- t(X)
    Yhat <- X %*% C %*% tX
    Ssym <- C + t(C)
    if (has.weights) {
      WY <- Wmat * Y; WYhat <- Wmat * Yhat
    } else {
      WY <- Y; WYhat <- Yhat
    }
    num_X <- WY    %*% X %*% Ssym
    den_X <- WYhat %*% X %*% Ssym
    if (X.L2.ortho > 0) {
      XtX <- crossprod(X); diag(XtX) <- 0
      den_X <- den_X + X.L2.ortho * (X %*% XtX)
    }
    X <- X * (num_X / (den_X + small))

    cc <- norm_XC(X, C); X <- cc$X; C <- cc$C

    obj_cur <- compute_obj(X, C)
    objfunc.iter[iter] <- obj_cur
    if (iter > 1) {
      rel <- abs(obj_prev - obj_cur) / (abs(obj_prev) + small)
      if (rel < epsilon) break
    }
    obj_prev <- obj_cur
  }
  if (iter == maxit && exists("rel") && rel >= epsilon)
    warning(paste0("maximum iterations (", maxit, ") reached..."))
  list(X = X, C = C, iter = iter,
       objfunc.iter = objfunc.iter[seq_len(iter)],
       objfunc = obj_prev)
}


## ==============================================================
## Internal runner for bi-symmetric: Y ~ X X^T (C = I_Q fixed)
## Cube-root damping (He et al. 2011) for monotone descent.
## ==============================================================

.nmfkc.net.run_once_bi <- function(Y, X, maxit, epsilon,
                                    X.restriction, X.L2.ortho,
                                    Wmat, has.weights) {
  N <- nrow(Y); Q <- ncol(X)
  small <- 1e-16
  compute_obj <- function(X) {
    Yhat <- tcrossprod(X)
    ## lm()-style weighted least squares (linear W).
    if (has.weights) sum(Wmat * (Y - Yhat)^2) else sum((Y - Yhat)^2)
  }
  obj_prev <- compute_obj(X)
  objfunc.iter <- numeric(maxit)
  iter <- 0L
  for (iter in seq_len(maxit)) {
    Yhat <- tcrossprod(X)
    if (has.weights) {
      num_X <- (Wmat * Y)    %*% X
      den_X <- (Wmat * Yhat) %*% X
    } else {
      num_X <- Y %*% X
      den_X <- X %*% crossprod(X)
    }
    if (X.L2.ortho > 0) {
      XtX <- crossprod(X); diag(XtX) <- 0
      den_X <- den_X + X.L2.ortho * (X %*% XtX)
    }
    update_ratio <- num_X / (den_X + small)
    X <- X * update_ratio^(1/3)

    if (X.restriction == "colSums") {
      cs <- colSums(X) + small
      X <- sweep(X, 2, cs, "/")
    } else if (X.restriction == "colSqSums") {
      d <- sqrt(colSums(X * X)) + small
      X <- sweep(X, 2, d, "/")
    }
    obj_cur <- compute_obj(X)
    objfunc.iter[iter] <- obj_cur
    if (iter > 1) {
      rel <- abs(obj_prev - obj_cur) / (abs(obj_prev) + small)
      if (rel < epsilon) break
    }
    obj_prev <- obj_cur
  }
  if (iter == maxit && exists("rel") && rel >= epsilon)
    warning(paste0("maximum iterations (", maxit, ") reached..."))
  list(X = X, C = diag(Q), iter = iter,
       objfunc.iter = objfunc.iter[seq_len(iter)],
       objfunc = obj_prev)
}


## ==============================================================
#' @title Symmetric NMF for networks (tri / bi / signed)
#' @description
#' Single entry point for symmetric NMF of network data with correct
#' multiplicative updates.  Three model types are supported via \code{type}:
#' \itemize{
#'   \item \strong{tri} (\code{type="tri"}, default):
#'     \eqn{Y \approx X C X^\top} with \eqn{X, C \ge 0} (both non-negative;
#'     \eqn{C} symmetric by design).  Uses Frobenius-full bilateral gradient.
#'   \item \strong{bi} (\code{type="bi"}):
#'     \eqn{Y \approx X X^\top} (\eqn{C} fixed to \eqn{I_Q}),
#'     cube-root damping (He et al. 2011).
#'   \item \strong{signed} (\code{type="signed"}):
#'     \eqn{Y \approx X (C_{+} - C_{-}) X^\top} with \eqn{X \ge 0} and signed
#'     \eqn{C = C_{+} - C_{-}}.  Preserves the soft-clustering interpretation
#'     of \eqn{X} while allowing negative off-diagonals of \eqn{C}
#'     (inter-cluster \emph{repulsion}).
#' }
#'
#' \strong{Non-negative adjacency matrix assumption.} All three types
#' assume \eqn{Y \ge 0} (a non-negative adjacency/affinity matrix).
#' The qualifier \dQuote{signed} in \code{type = "signed"} refers to the
#' middle coefficient \eqn{C}, not to \eqn{Y} itself.  The underlying
#' Ding, Li & Jordan (2010) sign-splitting updates require \eqn{Y \ge 0}
#' to guarantee monotone descent; supplying a signed \eqn{Y} triggers an
#' error.  For a signed data matrix, see \code{\link{nmfkc.signed}}.
#'
#' @param Y Symmetric (N x N) \strong{non-negative} adjacency matrix.
#'   \code{NA} entries are automatically treated as masked edges
#'   (equivalent to supplying \code{Y.weights} with 0 at those positions);
#'   see the note on \code{Y.weights} below.
#' @param rank Integer Q.
#' @param type \code{"tri"} (default), \code{"bi"}, or \code{"signed"}.
#' @param epsilon,maxit,verbose Standard.
#' @param ... Hidden options: \code{nstart} (default 1; see note below),
#'   \code{seed} (default 123), \code{X.restriction}, \code{X.init},
#'   \code{C.init} (tri only) or \code{Cp.init}/\code{Cn.init} (signed only),
#'   \code{Y.weights}, \code{C.L1} (tri only), \code{X.L2.ortho}, \code{prefix}.
#'
#' \strong{\code{Y.weights}} is an optional non-negative N x N weight
#' matrix (symmetric, same shape as \code{Y}).  When supplied, the loss
#' becomes \eqn{\sum W_{ij} \, (Y_{ij} - \hat Y_{ij})^2}
#' (\code{\link[stats]{lm}}()-style, \strong{linear} in \eqn{W}).
#' Logical matrices (\code{TRUE} / \code{FALSE}) are also accepted.
#' Typical usage by \code{\link{nmfkc.net.ecv}} is a binary mask
#' (\eqn{W \in \{0,1\}}) holding out test edges on the upper triangle;
#' real-valued weights for edge-level importance weighting are also
#' supported.  If \code{Y.weights} is \code{NULL} (default) and \code{Y}
#' contains \code{NA}, a binary mask is auto-generated (0 at \code{NA}
#' positions, 1 elsewhere), and the \code{NA} entries in \code{Y} are
#' replaced by 0 so the multiplicative updates can proceed.
#'
#' \strong{\code{X.init}} controls the initialization of the N x Q basis
#' matrix \eqn{X}.  Accepted values:
#' \itemize{
#'   \item \code{"kmeans"} (default): k-means on the rows of \eqn{Y}
#'     (equivalently columns, since \eqn{Y} is symmetric); the Q
#'     cluster centers become the columns of \eqn{X}.  Each node is
#'     treated as an N-dimensional connectivity profile, so clusters
#'     correspond to nodes with similar neighborhood structure --
#'     essentially a fast proxy for spectral clustering (Kuang, Yun &
#'     Park 2015, \emph{SymNMF}).  Scales well and is the recommended
#'     default for network data.
#'   \item \code{"kmeansar"}: \code{"kmeans"} followed by filling zero
#'     entries of \eqn{X} with \eqn{\mathrm{Uniform}(0, \bar Y / 100)}
#'     to escape trivial stationary points.
#'   \item \code{"nndsvd"}: Non-negative Double SVD with additive
#'     randomness (NNDSVDar).  Requires a full SVD of \eqn{Y}, so for
#'     very large networks (N > a few thousand) \code{"kmeans"} is
#'     preferable.
#'   \item \code{"runif"}: Uniform random entries in \eqn{[0, 1]}.
#'   \item \code{"random"}: Legacy default (pre-v0.6.8), equivalent to
#'     \code{abs(rnorm(N * Q)) * 0.1}.  Kept for backward compatibility.
#'   \item A numeric N x Q matrix supplied by the user (used as-is).
#' }
#' When \code{nstart > 1}, each restart uses a distinct seed so that
#' k-means / runif / NNDSVDar produce different candidate initial
#' values across the multi-start loop.
#'
#' \strong{Multi-start recommendation.} For \code{type = "signed"} the
#' \eqn{C = C_{+} - C_{-}} bottleneck can take both positive and negative
#' values, so the objective has more local minima than for \code{"tri"}
#' or \code{"bi"}.  A larger \code{nstart} (e.g., 10-50) is recommended
#' during exploration to reduce the chance of being trapped at a
#' suboptimal stationary point.  The default 1 is intended for fast
#' development; raise for publication-grade runs.
#' @return Object of class \code{c("nmfkc.net.<type>", "nmfkc.net", "nmfkc")}.
#'   For \code{type = "signed"} the return also carries \code{$Cp, $Cn}.
#' @seealso \code{\link{nmfkc.net.ecv}}, \code{\link{nmfkc.net.DOT}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @export
nmfkc.net <- function(Y, rank = 2, type = c("tri", "bi", "signed"),
                      epsilon = 1e-4, maxit = 5000,
                      verbose = FALSE, ...) {
  cl <- match.call()
  type <- match.arg(type)
  ex <- list(...)

  nstart        <- if (!is.null(ex$nstart))        ex$nstart        else 1L
  seed          <- if (!is.null(ex$seed))          ex$seed          else 123L
  X.restriction <- if (!is.null(ex$X.restriction)) ex$X.restriction
                   else if (type == "bi")          "none"
                   else                            "colSums"
  X.restriction <- match.arg(X.restriction,
                             c("colSums", "colSqSums", "none", "fixed"))
  X.init        <- if (!is.null(ex$X.init))        ex$X.init        else "kmeans"
  C.init        <- ex$C.init       # tri only
  Cp.init       <- ex$Cp.init      # signed only
  Cn.init       <- ex$Cn.init      # signed only
  Y.weights     <- ex$Y.weights
  C.L1          <- if (!is.null(ex$C.L1))          ex$C.L1          else 0
  X.L2.ortho    <- if (!is.null(ex$X.L2.ortho))    ex$X.L2.ortho    else 0
  prefix        <- if (!is.null(ex$prefix))        ex$prefix        else "Basis"

  Y <- as.matrix(Y); storage.mode(Y) <- "double"
  if (nrow(Y) != ncol(Y)) stop("Y must be square for nmfkc.net.")
  if (!isSymmetric(Y, tol = 1e-10)) {
    Y <- (Y + t(Y)) / 2
    if (verbose) message("nmfkc.net: symmetrized Y <- (Y + Y^T) / 2")
  }
  N <- nrow(Y); Q <- as.integer(rank)
  small <- 1e-16; t0 <- proc.time()

  ## ---- Y.weights handling (supports auto-NA mask, user mask, logical) ----
  ## Matches the convention of nmfkc() / nmfkc.signed(): if Y has NA and
  ## Y.weights is NULL, a binary mask is auto-generated with 0 at the
  ## NA positions and 1 elsewhere.  NA is then replaced by 0 in Y so the
  ## multiplicative updates can proceed (masked entries contribute
  ## nothing to the objective since W == 0 there).
  has.weights <- !is.null(Y.weights)
  if (is.null(Y.weights) && anyNA(Y)) {
    Y.weights <- matrix(1, nrow = N, ncol = N)
    Y.weights[is.na(Y)] <- 0
    ## Also zero the symmetric mirror so W stays symmetric
    Y.weights[is.na(t(Y))] <- 0
    Y[is.na(Y)] <- 0
    has.weights <- TRUE
    if (verbose) message("nmfkc.net: auto-generated Y.weights mask from ",
                         sum(Y.weights == 0), " NA entries in Y.")
  } else if (has.weights) {
    Y.weights <- as.matrix(Y.weights)
    if (!all(dim(Y.weights) == c(N, N))) stop("Y.weights must be N x N.")
    storage.mode(Y.weights) <- "double"  # coerce logical T/F to 1/0
    if (!isSymmetric(Y.weights, tol = 1e-10))
      Y.weights <- (Y.weights + t(Y.weights)) / 2
    Y.weights[is.na(Y.weights)] <- 0
    Y[is.na(Y) | Y.weights == 0] <- 0
  } else if (anyNA(Y)) {
    stop("Y contains NA; please impute, remove, or supply Y.weights.")
  }
  Wmat <- if (has.weights) Y.weights else NULL

  if (min(Y) < 0) {
    ## nmfkc.net is designed for non-negative adjacency matrices in network
    ## analysis.  `type = "signed"` means the middle coefficient matrix C is
    ## allowed to be signed (enabling inhibitory/excitatory latent factor
    ## interactions), NOT that Y itself can be signed.  The Ding-style
    ## multiplicative updates require Y >= 0 in all three types; allowing a
    ## signed Y here would silently violate the monotone descent guarantee
    ## (the C+/C- numerators could go negative).
    stop("nmfkc.net(type='", type, "') requires Y >= 0 ",
         "(Y is expected to be a non-negative adjacency matrix; ",
         "`type = \"signed\"` refers to the middle coefficient C, not Y).")
  }

  ## ---- run_once dispatches to the appropriate MU kernel ----
  run_once <- function(X, C_or_Cp, Cn = NULL) {
    if (type == "bi") {
      .nmfkc.net.run_once_bi(Y, X, maxit, epsilon,
                              X.restriction, X.L2.ortho, Wmat, has.weights)
    } else if (type == "signed") {
      .nmfkc.net.signed.run_once(Y, X, C_or_Cp, Cn, maxit, epsilon,
                                  X.restriction, Wmat, has.weights)
    } else {
      .nmfkc.net.run_once(Y, X, C_or_Cp, maxit, epsilon,
                          X.restriction, C.L1, X.L2.ortho, Wmat, has.weights)
    }
  }

  ## ---- multi-start ----
  best <- NULL
  explicit_X <- is.matrix(X.init) || (is.numeric(X.init) && length(X.init) > 1)
  ## "random" is the legacy default -- kept as a backward-compat alias for
  ## abs(rnorm(N*Q)) * 0.1, since pre-v0.6.8 nmfkc.net() used this form.
  ## Other strings ("kmeans", "kmeansar", "nndsvd", "runif") delegate to
  ## the shared .init_X_method() helper for consistency with nmfkc() /
  ## nmf.sem().
  for (s in seq_len(nstart)) {
    s_seed <- seed + 7919L * (s - 1L)
    if (explicit_X) {
      X0 <- as.matrix(X.init)
      if (!identical(dim(X0), c(N, Q)))
        stop("X.init must have dimensions (nrow(Y), rank).")
    } else if (identical(X.init, "random")) {
      set.seed(s_seed)
      X0 <- matrix(abs(stats::rnorm(N * Q)) * 0.1, N, Q)
    } else if (is.character(X.init)) {
      ## "kmeans" / "kmeansar" / "nndsvd" / "runif" via the shared helper.
      ## Each restart uses a distinct seed so k-means and runif produce
      ## different candidates across nstart; nndsvdar's additive randomness
      ## also varies.
      X0 <- .init_X_method(X.init, Y, Q, seed = s_seed)
    } else {
      stop("X.init must be one of \"kmeans\", \"kmeansar\", \"nndsvd\", ",
           "\"runif\", \"random\", or a numeric (N x Q) matrix; got ",
           class(X.init)[1], ".")
    }

    if (type == "bi") {
      out <- run_once(X0, diag(Q))
    } else if (type == "signed") {
      if (!is.null(Cp.init) && !is.null(Cn.init)) {
        Cp0 <- as.matrix(Cp.init); Cn0 <- as.matrix(Cn.init)
        if (!isSymmetric(Cp0, tol = 1e-10) || !isSymmetric(Cn0, tol = 1e-10)) {
          warning("nmfkc.net(type='signed'): Cp.init/Cn.init symmetrized via (M + M^T)/2.")
          Cp0 <- (Cp0 + t(Cp0)) / 2
          Cn0 <- (Cn0 + t(Cn0)) / 2
        }
      } else {
        set.seed(s_seed + 1L)
        C0 <- matrix(stats::runif(Q * Q, min = -1, max = 1), Q, Q)
        C0 <- (C0 + t(C0)) / 2       # symmetrize before sign split
        Cp0 <- pmax(C0, 0); Cn0 <- pmax(-C0, 0)
      }
      out <- run_once(X0, Cp0, Cn0)
    } else {  # tri
      if (!is.null(C.init)) {
        C0 <- as.matrix(C.init)
        if (!identical(dim(C0), c(Q, Q)))
          stop("C.init must have dimensions (rank, rank).")
        if (any(C0 < 0)) stop("C.init must be non-negative for type='tri'.")
        if (!isSymmetric(C0, tol = 1e-10)) {
          warning("nmfkc.net: C.init symmetrized via (C + C^T)/2.")
          C0 <- (C0 + t(C0)) / 2
        }
      } else {
        set.seed(s_seed + 1L)
        C0 <- matrix(abs(stats::rnorm(Q * Q)) * 0.1, Q, Q)
        C0 <- (C0 + t(C0)) / 2       # symmetrize for exact bilateral gradient
      }
      out <- run_once(X0, C0)
    }
    if (is.null(best) || out$objfunc < best$objfunc) best <- out
  }

  ## ---- extract best result ----
  X <- best$X
  objfunc.iter <- best$objfunc.iter
  iter <- best$iter
  if (type == "signed") {
    Cp <- best$Cp; Cn <- best$Cn; C <- Cp - Cn
  } else {
    C <- best$C; Cp <- NULL; Cn <- NULL
  }

  ## ---- centroid reorder ----
  if (Q > 1 && X.restriction != "fixed") {
    idx <- order(as.vector(matrix(seq_len(N) / N, nrow = 1) %*% X))
    X <- X[, idx, drop = FALSE]
    C <- C[idx, idx, drop = FALSE]
    if (!is.null(Cp)) Cp <- Cp[idx, idx, drop = FALSE]
    if (!is.null(Cn)) Cn <- Cn[idx, idx, drop = FALSE]
  }

  ## ---- names ----
  rownames(X) <- rownames(Y)
  colnames(X) <- paste0(prefix, seq_len(Q))
  rownames(C) <- colnames(C) <- paste0(prefix, seq_len(Q))
  if (!is.null(Cp)) { rownames(Cp) <- colnames(Cp) <- paste0(prefix, seq_len(Q)) }
  if (!is.null(Cn)) { rownames(Cn) <- colnames(Cn) <- paste0(prefix, seq_len(Q)) }

  ## ---- reconstruction and fit statistics ----
  Y1hat <- X %*% C %*% t(X)
  resid <- Y - Y1hat
  ## lm()-style weighted least squares (matches compute_obj inside the loop).
  objfunc <- if (has.weights) sum(Wmat * resid^2) else sum(resid^2)
  ## Unified three-variant R^2 (cor^2, uncentered,
  ## row-mean centered), all respecting Y.weights == 0 masking.
  r2_all <- .r.squared.all(Y, Y1hat,
                           Y.weights = if (has.weights) Wmat else NULL)
  r.squared          <- r2_all$r.squared
  r.squared.uncentered     <- r2_all$r.squared.uncentered
  r.squared.centered <- r2_all$r.squared.centered
  mae <- if (has.weights) sum(Wmat * abs(resid)) / max(sum(Wmat), small)
         else mean(abs(resid))

  X.prob <- X / (rowSums(X) + small)
  X.cluster <- apply(X.prob, 1, which.max)

  result <- list(
    call = cl,
    X = X, C = C, Cp = Cp, Cn = Cn,
    XB = Y1hat, Y1hat = Y1hat,          # XB alias for fitted.nmf compatibility
    X.prob = X.prob, X.cluster = X.cluster,
    rank = Q, dims = c(N = N),
    type = type,
    objfunc = objfunc, objfunc.iter = objfunc.iter,
    r.squared          = r.squared,
    r.squared.uncentered     = r.squared.uncentered,
    r.squared.centered = r.squared.centered,
    mae = mae,
    iter = iter,
    runtime = as.numeric((proc.time() - t0)[3]),
    X.restriction = X.restriction
  )
  class(result) <- c(paste0("nmfkc.net.", type), "nmfkc.net", "nmfkc", "nmf")
  result
}


## ==============================================================
## Signed: symmetric tri-NMF with signed C = Cp - Cn
## (X >= 0 for soft clustering; C signed for inter-cluster repulsion)
## ==============================================================

.nmfkc.net.signed.run_once <- function(Y, X, Cp, Cn, maxit, epsilon,
                                       X.restriction, Wmat, has.weights) {
  N <- nrow(Y); Q <- ncol(X)
  small <- 1e-16
  norm_XC <- function(X, Cp, Cn) {
    if (X.restriction %in% c("fixed", "none"))
      return(list(X = X, Cp = Cp, Cn = Cn))
    cs <- colSums(X) + small
    d <- switch(X.restriction,
                colSums   = cs,
                colSqSums = sqrt(colSums(X * X)) + small,
                cs)
    X <- sweep(X, 2, d, "/")
    Cp <- sweep(Cp, 1, d, "*"); Cp <- sweep(Cp, 2, d, "*")
    Cn <- sweep(Cn, 1, d, "*"); Cn <- sweep(Cn, 2, d, "*")
    list(X = X, Cp = Cp, Cn = Cn)
  }
  cc <- norm_XC(X, Cp, Cn); X <- cc$X; Cp <- cc$Cp; Cn <- cc$Cn
  compute_obj <- function(X, Cp, Cn) {
    Yhat <- X %*% (Cp - Cn) %*% t(X)
    ## lm()-style weighted least squares (linear W).
    if (has.weights) sum(Wmat * (Y - Yhat)^2) else sum((Y - Yhat)^2)
  }
  obj_prev <- compute_obj(X, Cp, Cn)
  objfunc.iter <- numeric(maxit)
  iter <- 0L

  for (iter in seq_len(maxit)) {
    tX <- t(X)
    P <- tX %*% X

    ## ---- Cp, Cn update (Ding sign split) ----
    if (has.weights) {
      Yh_p <- X %*% Cp %*% tX
      Yh_n <- X %*% Cn %*% tX
      G    <- tX %*% (Wmat * Y) %*% X
      Hp   <- tX %*% (Wmat * Yh_p) %*% X
      Hn   <- tX %*% (Wmat * Yh_n) %*% X
    } else {
      G  <- tX %*% Y %*% X
      Hp <- P %*% Cp %*% P
      Hn <- P %*% Cn %*% P
    }
    Cp <- Cp * (G + Hn) / (Hp + small)
    if (has.weights) {
      Yh_p <- X %*% Cp %*% tX
      Hp   <- tX %*% (Wmat * Yh_p) %*% X
    } else Hp <- P %*% Cp %*% P
    Cn <- Cn * (Hp) / (G + Hn + small)

    ## ---- X update (bilateral + Ding split, exact for symmetric Cp, Cn) ----
    tX <- t(X)
    Sp <- Cp + t(Cp); Sn <- Cn + t(Cn)
    Yh_p <- X %*% Cp %*% tX
    Yh_n <- X %*% Cn %*% tX
    if (has.weights) {
      WY   <- Wmat * Y;    WYhp <- Wmat * Yh_p;    WYhn <- Wmat * Yh_n
    } else {
      WY <- Y; WYhp <- Yh_p; WYhn <- Yh_n
    }
    num_X <- WY %*% X %*% Sp + WYhp %*% X %*% Sn + WYhn %*% X %*% Sp
    den_X <- WY %*% X %*% Sn + WYhp %*% X %*% Sp + WYhn %*% X %*% Sn
    X <- X * (num_X / (den_X + small))

    cc <- norm_XC(X, Cp, Cn); X <- cc$X; Cp <- cc$Cp; Cn <- cc$Cn

    obj_cur <- compute_obj(X, Cp, Cn)
    objfunc.iter[iter] <- obj_cur
    if (iter > 1) {
      rel <- abs(obj_prev - obj_cur) / (abs(obj_prev) + small)
      if (rel < epsilon) break
    }
    obj_prev <- obj_cur
  }
  if (iter == maxit && exists("rel") && rel >= epsilon)
    warning(paste0("maximum iterations (", maxit, ") reached..."))
  list(X = X, Cp = Cp, Cn = Cn, iter = iter,
       objfunc.iter = objfunc.iter[seq_len(iter)],
       objfunc = obj_prev)
}


## ECV: upper-triangle element-wise cross-validation
## ==============================================================

.nmfkc.net.make_uppertri_folds <- function(N, div = 5, seed = 123) {
  if (!is.null(seed)) set.seed(seed)
  ut_idx <- which(upper.tri(matrix(0, N, N), diag = TRUE))
  perm <- sample(ut_idx)
  n <- length(perm)
  chunk <- n %/% div; rem <- n %% div
  folds <- vector("list", div)
  s <- 1L
  for (k in 1:div) {
    sz <- chunk + ifelse(k <= rem, 1L, 0L)
    folds[[k]] <- perm[s:(s + sz - 1L)]
    s <- s + sz
  }
  folds
}

## Mirror a 0/1 mask to be symmetric via AND (upper & lower both retained).
## For general (non-binary) weights use (W + t(W))/2 instead.
.nmfkc.net.mirror_mask <- function(W) W * t(W)

#' Element-wise cross-validation for nmfkc.net (upper-triangle folds)
#'
#' @description k-fold CV with folds taken over the upper triangle of the
#' symmetric \eqn{Y} (mirrored to the lower triangle) to prevent information
#' leakage through symmetry.  A single entry point covers all three symmetric
#' NMF variants; \code{type} selects the fitting function:
#' \itemize{
#'   \item \code{"tri"} (default): \code{\link{nmfkc.net}} with \eqn{C \ge 0}
#'   \item \code{"bi"}: \code{\link{nmfkc.net}} with \eqn{C = I_Q}
#'   \item \code{"signed"}: \code{\link{nmfkc.net}} with signed \eqn{C = C_{+} - C_{-}}
#' }
#'
#' @param Y Symmetric N x N non-negative matrix.
#' @param rank Integer vector of ranks to evaluate. Default \code{1:3}.
#' @param type Model type: \code{"tri"} (default), \code{"bi"}, or \code{"signed"}.
#' @param ... Passed to the underlying fitter; also accepts \code{nfolds}
#'   (default 5; \code{div} alias), \code{seed} (default 123).
#'
#' @return A list with \code{objfunc}, \code{sigma}, \code{objfunc.fold},
#'   \code{folds}, \code{Q.grid}, \code{type}.
#' @seealso \code{\link{nmfkc.net}}
#' @section Lifecycle:
#' This function is \strong{experimental}. The interface may change in
#' future versions; details are to be described in an upcoming paper.
#'
#' @export
nmfkc.net.ecv <- function(Y, rank = 1:3,
                          type = c("tri", "bi", "signed"), ...) {
  type <- match.arg(type)
  ex <- list(...)
  nfolds <- if (!is.null(ex$nfolds)) ex$nfolds
            else if (!is.null(ex$div)) ex$div else 5
  seed <- if (!is.null(ex$seed)) ex$seed else 123
  Y <- as.matrix(Y); N <- nrow(Y)
  folds <- .nmfkc.net.make_uppertri_folds(N, div = nfolds, seed = seed)

  fit_args <- ex; fit_args$nfolds <- NULL; fit_args$div <- NULL
  fit_args$rank <- NULL; fit_args$type <- NULL

  run_one <- function(q, k) {
    test_ut <- folds[[k]]
    W <- matrix(1, N, N); W[test_ut] <- 0
    W <- .nmfkc.net.mirror_mask(W)
    fit <- suppressMessages(do.call(nmfkc.net,
             c(list(Y = Y, rank = q, type = type,
                    Y.weights = W, verbose = FALSE), fit_args)))
    Yhat <- fit$X %*% fit$C %*% t(fit$X)
    mean((Y[test_ut] - Yhat[test_ut])^2)
  }

  message(sprintf("nmfkc.net ECV (type=%s): %d ranks, %d-fold, upper-triangle.",
                  type, length(rank), nfolds))
  cv <- .ecv.run(sprintf("Q=%d", rank), nfolds,
                 run_one = function(i, k) run_one(rank[i], k),
                 progress = function(i, o, s)
                   message(sprintf("  Q=%d: MSE=%.6f, sigma=%.4f", rank[i], o, s)))
  cls <- if (type == "signed")
           c("nmfkc.net.signed.ecv", "nmfkc.net.ecv", "nmfkc.ecv")
         else
           c("nmfkc.net.ecv", "nmfkc.ecv")
  structure(list(objfunc = cv$objfunc, sigma = cv$sigma,
                 objfunc.fold = cv$objfunc.fold, folds = folds,
                 Q.grid = rank, type = type),
            class = cls)
}


#' @title Rank selection for nmfkc.net (concise diagnostics)
#' @description
#' Fits \code{\link{nmfkc.net}} across a range of ranks and reports the
#' three rank-selection criteria -- \code{r.squared}, the effective rank
#' (utilization), and the element-wise CV error \code{sigma.ecv} -- with
#' the same concise diagnostics plot as \code{\link{nmfkc.rank}}.
#' @param Y Symmetric (network) observation matrix.
#' @param rank Integer vector of ranks to evaluate.
#' @param type One of \code{"tri"}, \code{"bi"}, \code{"signed"}.
#' @param detail \code{"full"} (default) also runs element-wise CV
#'   (\code{sigma.ecv}); \code{"fast"} skips it (plots r.squared and
#'   eff.rank only, and recommends the R-squared elbow).
#' @param plot Logical; draw the diagnostics plot (default \code{TRUE}).
#' @param ... Passed on to \code{\link{nmfkc.net}} and
#'   \code{\link{nmfkc.net.ecv}} (e.g.\ \code{nstart}, \code{maxit},
#'   \code{nfolds}, \code{seed}).
#' @return A list with \code{rank.best} (ECV minimum, or the R-squared
#'   elbow under \code{detail = "fast"}) and \code{criteria} (data
#'   frame: \code{rank}, \code{effective.rank}, \code{effective.rank.ratio},
#'   \code{r.squared}, \code{sigma.ecv}).
#' @seealso \code{\link{nmfkc.net}}, \code{\link{nmfkc.net.ecv}},
#'   \code{\link{nmfkc.rank}}
#' @references
#' Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
#' effective dimensionality.  \emph{Proc. EUSIPCO}, 606--610.
#' (\code{effective.rank})
#' Wold, S. (1978).  Cross-validatory estimation of the number of
#' components in factor and principal components models.
#' \emph{Technometrics}, 20(4), 397--405. (\code{sigma.ecv})
#' @export
#' @examples
#' \donttest{
#' Y <- matrix(c(0,1,1,0,0,0, 1,0,1,0,0,0, 1,1,0,1,0,0,
#'               0,0,1,0,1,1, 0,0,0,1,0,1, 0,0,0,1,1,0), 6, 6)
#' nmfkc.net.rank(Y, rank = 1:3, type = "tri", nstart = 5, nfolds = 3)
#' }
nmfkc.net.rank <- function(Y, rank = 1:5, type = c("tri", "bi", "signed"),
                           detail = c("full", "fast"), plot = TRUE, ...) {
  type <- match.arg(type); detail <- match.arg(detail)
  Y <- as.matrix(Y)
  rs <- numeric(length(rank)); er <- numeric(length(rank))
  for (i in seq_along(rank)) {
    f <- suppressMessages(nmfkc.net(Y, rank = rank[i], type = type,
                                    verbose = FALSE, ...))
    rs[i] <- f$r.squared
    er[i] <- .effective.rank(t(f$X))
  }
  ecv <- if (detail == "full")
    suppressMessages(nmfkc.net.ecv(Y, rank = rank, type = type, ...))$sigma
    else rep(NA_real_, length(rank))
  criteria <- data.frame(rank = rank, effective.rank = er,
                         effective.rank.ratio = er / rank,
                         r.squared = rs, sigma.ecv = as.numeric(ecv))
  .rank.finish(criteria, plot = plot,
               main = sprintf("nmfkc.net rank selection (type=%s)", type))
}


## ==============================================================
## Summary methods for nmfkc.net and nmfkc.net.signed
## (Format unified with print.summary.nmfkc)
## ==============================================================

#' Summary method for nmfkc.net objects
#' @param object An \code{nmfkc.net} object.
#' @param ... Unused.
#' @return An object of class \code{"summary.nmfkc.net"}.
#' @seealso \code{\link{nmfkc.net}}
#' @export
summary.nmfkc.net <- function(object, ...) {
  N <- object$dims["N"]; Q <- object$rank
  small <- 1e-16
  dims_str <- sprintf("Y(%d,%d)~X(%d,%d)C(%d,%d)X'", N, N, N, Q, Q, Q)

  ans <- list(
    call = object$call,
    dims = dims_str,
    rank = Q,
    type = if (!is.null(object$type)) object$type else "tri",
    iter = object$iter,
    runtime = sprintf("%.2fsec", object$runtime),

    objfunc   = object$objfunc,
    r.squared          = object$r.squared,
    r.squared.uncentered     = object$r.squared.uncentered,
    r.squared.centered = object$r.squared.centered,
    mae       = object$mae,
    ## Effective rank over the Q factors' membership across the N nodes
    ## (X is N x Q, so transpose to factors x nodes).
    effective.rank = .effective.rank(t(object$X)),

    X.sparsity      = mean(object$X  < 1e-4),
    C.sparsity      = mean(abs(object$C) < 1e-4),

    X.range = range(object$X),
    C.range = range(object$C),
    X.cluster.counts = table(object$X.cluster)
  )
  class(ans) <- "summary.nmfkc.net"
  ans
}

#' Print method for summary.nmfkc.net objects
#' @param x An object of class \code{"summary.nmfkc.net"}.
#' @param digits Minimum number of significant digits.
#' @param ... Unused.
#' @return Invisible \code{x}.
#' @seealso \code{\link{summary.nmfkc.net}}
#' @export
print.summary.nmfkc.net <- function(x, digits = max(3L, getOption("digits") - 3L), ...) {
  cat("\nCall:\n", paste(deparse(x$call), sep = "\n", collapse = "\n"), "\n\n", sep = "")

  cat("Dimensions:", x$dims, "\n")
  cat("Rank (Q):   ", x$rank, "\n")
  cat("Type:       ", x$type, "\n")
  cat("Runtime:    ", x$runtime, "\n")
  cat("Iterations: ", x$iter, "\n")

  .print.fit.statistics(x, digits = digits)

  .print.structure.diagnostics(
    sparsity = c("Basis (X)" = x$X.sparsity, "C" = x$C.sparsity))
  cat(sprintf("  %-24s [%.3g, %.3g]\n", "X range:", x$X.range[1], x$X.range[2]))
  cat(sprintf("  %-24s [%.3g, %.3g]\n", "C range:", x$C.range[1], x$C.range[2]))

  cat("\nCluster sizes (hard assignment by argmax of X):\n  ")
  print(x$X.cluster.counts)
  cat("\n")
  invisible(x)
}

#' Summary method for nmfkc.net.signed objects
#' @param object An \code{nmfkc.net.signed} object.
#' @param ... Unused.
#' @return An object of class \code{"summary.nmfkc.net.signed"}.
#' @seealso \code{\link{nmfkc.net}}
#' @export
summary.nmfkc.net.signed <- function(object, ...) {
  ans <- summary.nmfkc.net(object, ...)
  ans$Cp.range <- range(object$Cp)
  ans$Cn.range <- range(object$Cn)
  ans$neg.frac <- sum(pmax(-object$C, 0)) / (sum(abs(object$C)) + 1e-16)
  class(ans) <- c("summary.nmfkc.net.signed", "summary.nmfkc.net")
  ans
}

#' Print method for summary.nmfkc.net.signed objects
#' @param x An object of class \code{"summary.nmfkc.net.signed"}.
#' @param digits Minimum number of significant digits.
#' @param ... Unused.
#' @return Invisible \code{x}.
#' @seealso \code{\link{summary.nmfkc.net.signed}}
#' @export
print.summary.nmfkc.net.signed <- function(x, digits = max(3L, getOption("digits") - 3L), ...) {
  cat("\nCall:\n", paste(deparse(x$call), sep = "\n", collapse = "\n"), "\n\n", sep = "")

  cat("Dimensions:", x$dims, "\n")
  cat("Rank (Q):   ", x$rank, "\n")
  cat("Type:       ", "signed (C = Cp - Cn)\n")
  cat("Runtime:    ", x$runtime, "\n")
  cat("Iterations: ", x$iter, "\n")

  .print.fit.statistics(x, digits = digits)

  .print.structure.diagnostics(
    sparsity = c("Basis (X)" = x$X.sparsity, "C" = x$C.sparsity))
  cat(sprintf("  %-24s [%.3g, %.3g]\n", "X range:",  x$X.range[1],  x$X.range[2]))
  cat(sprintf("  %-24s [%.3g, %.3g]\n", "Cp range:", x$Cp.range[1], x$Cp.range[2]))
  cat(sprintf("  %-24s [%.3g, %.3g]\n", "Cn range:", x$Cn.range[1], x$Cn.range[2]))
  cat(sprintf("  %-24s [%.3g, %.3g]\n", "C range:",  x$C.range[1],  x$C.range[2]))
  cat(sprintf("  %-24s %.1f%% (of sum |C|)\n", "Negative mass:", 100 * x$neg.frac))

  cat("\nCluster sizes (hard assignment by argmax of X):\n  ")
  print(x$X.cluster.counts)
  cat("\n")
  invisible(x)
}

## Tests for nmfkc.net family (symmetric NMF for networks)

## Small symmetric test network (two communities)
make_test_network <- function() {
  Y <- matrix(c(0, 1, 1, 0, 0, 0,
                1, 0, 1, 0, 0, 0,
                1, 1, 0, 1, 0, 0,
                0, 0, 1, 0, 1, 1,
                0, 0, 0, 1, 0, 1,
                0, 0, 0, 1, 1, 0), 6, 6)
  rownames(Y) <- colnames(Y) <- paste0("N", 1:6)
  Y
}


test_that("nmfkc.net (tri) fits a small symmetric network", {
  Y <- make_test_network()
  res <- nmfkc.net(Y, rank = 2, type = "tri", nstart = 5, maxit = 200)

  expect_s3_class(res, "nmfkc.net.tri")
  expect_s3_class(res, "nmfkc.net")
  expect_s3_class(res, "nmfkc")
  expect_equal(dim(res$X), c(6, 2))
  expect_equal(dim(res$C), c(2, 2))
  expect_true(all(res$X >= 0))
  expect_true(all(res$C >= 0))

  ## C is symmetric by design
  expect_true(isSymmetric(res$C, tol = 1e-8))

  ## X.prob rows sum to 1
  expect_true(all(abs(rowSums(res$X.prob) - 1) < 1e-8))
  expect_length(res$X.cluster, 6)

  expect_true(is.finite(res$r.squared))
})


test_that("nmfkc.net (bi) fits with C = I_Q fixed", {
  Y <- make_test_network()
  res <- nmfkc.net(Y, rank = 2, type = "bi", nstart = 5, maxit = 200)
  expect_s3_class(res, "nmfkc.net.bi")
  expect_equal(unname(res$C), diag(2))
  expect_true(all(res$X >= 0))
})


test_that("nmfkc.net(type='signed') produces signed C", {
  Y <- make_test_network()
  res <- nmfkc.net(Y, rank = 2, type = "signed", nstart = 5, maxit = 200)

  expect_s3_class(res, "nmfkc.net.signed")
  expect_s3_class(res, "nmfkc.net")
  expect_s3_class(res, "nmfkc")
  expect_true(all(res$X >= 0))
  expect_true(all(res$Cp >= 0))
  expect_true(all(res$Cn >= 0))

  ## C = Cp - Cn
  expect_equal(res$C, res$Cp - res$Cn, tolerance = 1e-10)
  ## C is symmetric by design
  expect_true(isSymmetric(res$C, tol = 1e-8))
})


test_that("nmfkc.net return structure is uniform (Cp/Cn NULL for tri/bi)", {
  Y <- make_test_network()
  res_tri    <- nmfkc.net(Y, rank = 2, type = "tri",    nstart = 3, maxit = 100)
  res_bi     <- nmfkc.net(Y, rank = 2, type = "bi",     nstart = 3, maxit = 100)
  res_signed <- nmfkc.net(Y, rank = 2, type = "signed", nstart = 3, maxit = 100)

  ## Uniform field presence
  for (res in list(res_tri, res_bi, res_signed)) {
    expect_true("X"  %in% names(res))
    expect_true("C"  %in% names(res))
    expect_true("Cp" %in% names(res))
    expect_true("Cn" %in% names(res))
  }
  ## Cp, Cn are NULL for tri/bi
  expect_null(res_tri$Cp); expect_null(res_tri$Cn)
  expect_null(res_bi$Cp);  expect_null(res_bi$Cn)
  ## Cp, Cn are matrices for signed
  expect_true(is.matrix(res_signed$Cp))
  expect_true(is.matrix(res_signed$Cn))
})


test_that("nmfkc.net.DOT works for tri, bi, and signed", {
  Y <- make_test_network()
  res_tri    <- nmfkc.net(Y, rank = 2, type = "tri", nstart = 3, maxit = 100)
  res_bi     <- nmfkc.net(Y, rank = 2, type = "bi",  nstart = 3, maxit = 100)
  res_signed <- nmfkc.net(Y, rank = 2, type = "signed", nstart = 3, maxit = 100)

  dot_tri <- nmfkc.net.DOT(res_tri)
  expect_s3_class(dot_tri, "nmfkc.DOT")
  expect_true(nchar(dot_tri) > 100)

  dot_bi <- nmfkc.net.DOT(res_bi)
  expect_s3_class(dot_bi, "nmfkc.DOT")

  ## Signed should auto-detect and produce a DOT string that references style="dashed"
  dot_signed <- nmfkc.net.DOT(res_signed)
  expect_s3_class(dot_signed, "nmfkc.DOT")
})


test_that("nmfkc.net.ecv supports tri/bi/signed via type argument", {
  Y <- make_test_network()

  ## maxit large enough to avoid the "maximum iterations reached"
  ## warning that nmfkc.net's MU loop now emits (commit 1d93c48 et al.).
  out_tri <- nmfkc.net.ecv(Y, rank = c(1, 2), type = "tri",
                            nfolds = 3, nstart = 3, maxit = 5000)
  expect_length(out_tri$objfunc, 2)
  expect_s3_class(out_tri, "nmfkc.net.ecv")
  expect_equal(out_tri$type, "tri")

  out_bi <- nmfkc.net.ecv(Y, rank = c(1, 2), type = "bi",
                           nfolds = 3, nstart = 3, maxit = 5000)
  expect_equal(out_bi$type, "bi")

  out_signed <- nmfkc.net.ecv(Y, rank = c(1, 2), type = "signed",
                               nfolds = 3, nstart = 3, maxit = 5000)
  expect_s3_class(out_signed, "nmfkc.net.signed.ecv")
  expect_s3_class(out_signed, "nmfkc.net.ecv")
  expect_equal(out_signed$type, "signed")
})


test_that("nmfkc(Y.symmetric = ...) stops and redirects to nmfkc.net()", {
  Y <- make_test_network()
  ## Symmetric NMF was removed from nmfkc(); passing Y.symmetric now
  ## errors with a message pointing to nmfkc.net().
  expect_error(
    nmfkc(Y, rank = 2, Y.symmetric = "tri",
          nstart = 3, maxit = 100, print.dims = FALSE),
    "nmfkc.net"
  )
  expect_error(
    nmfkc(Y, rank = 2, Y.symmetric = "bi",
          nstart = 3, maxit = 100, print.dims = FALSE),
    "no longer supported"
  )
})

test_that("nmfkc.ecv(Y.symmetric = ...) stops and redirects to nmfkc.net.ecv()", {
  Y <- make_test_network()
  expect_error(
    nmfkc.ecv(Y, rank = c(1, 2), Y.symmetric = "tri", nfolds = 3),
    "nmfkc.net.ecv"
  )
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7, 
  fig.height = 5
)

## ----load-packages------------------------------------------------------------
library(nmfkc)
library(quanteda)

## ----data-prep----------------------------------------------------------------
# Load the corpus from quanteda
corp <- corpus(data_corpus_inaugural)

# Preprocessing: tokenize, remove stopwords, and punctuation
tok <- tokens(corp, remove_punct = TRUE)
tok <- tokens_remove(tok, pattern = stopwords("en", source = "snowball"))

# Create DFM and filter
df <- dfm(tok)
df <- dfm_select(df, min_nchar = 3) # Remove short words (<= 2 chars)
df <- dfm_trim(df, min_termfreq = 100) # Remove rare words (appearing < 100 times)

# --- CRITICAL STEP ---
# quanteda's DFM is (Documents x Words).
# nmfkc expects (Words x Documents).
# We must transpose the matrix.
d <- as.matrix(df)
# Sort by frequency
index <- order(colSums(d), decreasing=TRUE) 
d <- d[,index]
Y <- t(d)

dim(Y) # Features (Words) x Samples (Documents)

## ----rank-selection, fig.width=7, fig.height=6--------------------------------
# Evaluate ranks from 2 to 6
nmfkc.rank(Y, rank = 2:6)

## ----model-fit----------------------------------------------------------------
rank <- 3
# Set seed for reproducibility
res_std <- nmfkc(Y, rank = rank, seed = 123, prefix = "Topic")

# Check Goodness of Fit (R-squared)
res_std$r.squared

## ----interpret-topics---------------------------------------------------------
# Extract top 10 words for each topic from X.prob (normalized X)
Xp <- res_std$X.prob
for(q in 1:rank){
  message(paste0("----- Featured words on Topic [", q, "] -----"))
  print(paste0(rownames(Xp), "(", rowSums(Y), ") ", round(100*Xp[,q], 1), "%")[Xp[,q]>=0.5])
}

## ----kernel-prep--------------------------------------------------------------
# Covariate: Year of the address
years <- as.numeric(substring(names(data_corpus_inaugural), 1, 4))
U <- t(as.matrix(years))

# Optimize beta (Gaussian Kernel width)
# We test a specific range of betas to locate the minimum CV error.
beta_candidates <- c(0.2, 0.5, 1, 2, 5) / 10000

# Run CV to find the best beta
# Note: We use the same rank (rank=3) as selected above.
cv_res <- nmfkc.kernel.beta.cv(Y, rank = rank, U = U, beta = beta_candidates, plot = FALSE)
best_beta <- cv_res$beta
print(best_beta)

## ----kernel-fit---------------------------------------------------------------
# Create Kernel Matrix
A <- nmfkc.kernel(U, beta = best_beta)

# Fit NMF with Kernel Covariates
res_ker <- nmfkc(Y, A = A, rank = rank, seed = 123, prefix = "Topic")

## ----plot-comparison, fig.width=8, fig.height=8-------------------------------
oldpar <- par(mfrow = c(2, 1), mar = c(4, 4, 2, 1))

# Prepare Axis Labels (Rounded to integers)
at_points <- seq(1, ncol(Y), length.out = 10)
labels_years <- round(seq(min(years), max(years), length.out = 10))

# 1. Standard NMF (Noisy)
barplot(res_std$B.prob, col = 2:(rank+1), border = NA, xaxt='n',
        main = "Standard NMF: Topic Proportions (Noisy)", ylab = "Probability")
axis(1, at = at_points, labels = labels_years)

# 2. Kernel NMF (Smooth trend)
barplot(res_ker$B.prob, col = 2:(rank+1), border = NA, xaxt='n',
        main = "Kernel NMF: Temporal Topic Evolution (Smooth)", ylab = "Probability")
axis(1, at = at_points, labels = labels_years)

# Legend
legend("topright", legend = paste("Topic", 1:rank), fill = 2:(rank+1), bg="white", cex=0.8)
par(oldpar)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----load-packages------------------------------------------------------------
library(nmfkc)
library(palmerpenguins)

## ----iris-data-prep-----------------------------------------------------------
# 1. Prepare Labels (Y)
label_iris <- iris$Species
Y_iris <- nmfkc.class(label_iris) # One-hot encoding (3 Classes x 150 Samples)
rank_iris <- length(unique(label_iris)) # Number of classes

# 2. Prepare Features (U)
# Normalize features to [0, 1] range and transpose to (Features x Samples)
U_iris <- t(nmfkc.normalize(iris[, -5]))

## ----iris-linear--------------------------------------------------------------
# Use Normalized Features directly as A
A_linear <- U_iris

# Fit Linear NMF-LAB
res_linear <- nmfkc(Y = Y_iris, A = A_linear, rank = rank_iris, seed = 123, prefix = "Class")

# --- Interpretability Check ---
# The matrix C (Q x R) shows the weight of each Feature (columns) for each Class (rows).
# Let's look at the estimated weights:
round(res_linear$C, 2)

## ----iris-linear-acc----------------------------------------------------------
pred_linear <- predict(res_linear, type = "class")
(f_linear <- table(fitted.label = pred_linear, label = label_iris))

# Calculate Accuracy (assuming diagonal correspondence)
acc_linear <- sum(diag(f_linear)) / sum(f_linear)
cat(paste0("Linear Model Accuracy: ", round(acc_linear * 100, 2), "%\n"))

## ----iris-kernel--------------------------------------------------------------
# 1. Optimize Kernel Width (beta)
# Heuristic estimation of beta
res_beta <- nmfkc.kernel.beta.nearest.med(U_iris)

# Cross-validation for fine-tuning (using generated candidates)
cv_res <- nmfkc.kernel.beta.cv(Y_iris, rank = rank_iris, U = U_iris, 
                               beta = res_beta$beta_candidates, plot = FALSE)
best_beta <- cv_res$beta

# 2. Fit Kernel NMF-LAB
A_kernel <- nmfkc.kernel(U_iris, beta = best_beta)
res_kernel <- nmfkc(Y = Y_iris, A = A_kernel, rank = rank_iris, seed = 123, prefix = "Class")

# 3. Prediction and Evaluation
fitted_label <- predict(res_kernel, type = "class")
(f_kernel <- table(fitted.label = fitted_label, label = label_iris))

# Calculate Accuracy
acc_kernel <- sum(diag(f_kernel)) / sum(f_kernel)
cat(paste0("Kernel Model Accuracy: ", round(acc_kernel * 100, 2), "%\n"))

## ----iris-plot-X, fig.height=5------------------------------------------------
image(t(res_kernel$X)[, nrow(res_kernel$X):1], 
      main = "Basis Matrix X (Kernel Model)\nMapping Factors to Species",
      axes = FALSE, col = hcl.colors(12, "YlOrRd", rev = TRUE))

axis(1, at = seq(0, 1, length.out = rank_iris), labels = colnames(res_kernel$X))
axis(2, at = seq(0, 1, length.out = rank_iris), labels = rev(rownames(res_kernel$X)), las = 2)
box()

## ----penguins-data-prep-------------------------------------------------------
# Load and clean data (remove rows with NAs)
d_penguins <- na.omit(palmerpenguins::penguins)

# Prepare Y (Labels)
label_penguins <- d_penguins$species
Y_penguins <- nmfkc.class(label_penguins)

# Prepare U (Features)
U_penguins <- t(nmfkc.normalize(d_penguins[, 3:6]))

## ----penguins-model-fit-------------------------------------------------------
rank_penguins <- length(unique(label_penguins))

# 1. Heuristic beta estimation
best_beta_penguins <- nmfkc.kernel.beta.nearest.med(U_penguins)$beta

# 2. Optimization
A_penguins <- nmfkc.kernel(U_penguins, beta = best_beta_penguins)
res_penguins <- nmfkc(Y = Y_penguins, A = A_penguins, rank = rank_penguins, 
                      seed = 123, prefix = "Class")

## ----penguins-soft-vis, fig.width=8, fig.height=4-----------------------------
# Get probabilistic predictions
probs <- predict(res_penguins, type = "prob")

# Visualize
barplot(probs, col = c("#FF8C00", "#9932CC", "#008B8B"), border = NA,
        main = "Soft Classification Probabilities (Penguins)",
        xlab = "Sample Index", ylab = "Probability")

legend("topright", legend = levels(label_penguins), 
       fill = c("#FF8C00", "#9932CC", "#008B8B"), bg = "white", cex = 0.8)

## ----penguins-acc-------------------------------------------------------------
fitted_label_p <- predict(res_penguins, type = "class")
(f_penguins <- table(Predicted = fitted_label_p, Actual = label_penguins))

acc_p <- sum(diag(f_penguins)) / sum(f_penguins)
cat(paste0("Penguins Accuracy: ", round(acc_p * 100, 2), "%\n"))

#' @title Create a Gaussian kernel matrix from covariates
#' @description
#' \code{nmfkc.kernel.gaussian} constructs a Gaussian (RBF) kernel matrix from covariate matrices.
#' The kernel is defined as \eqn{K(u,v) = \exp(-\beta \|u - v\|^2)}.
#' When \code{V} contains \code{NA} values, two methods are available via \code{na.method}:
#' \describe{
#'   \item{\code{"pds"}}{Partial Distance Strategy. Computes the kernel using only observed (non-NA) rows,
#'     with beta adjusted by \eqn{\beta_{adj} = \beta \times K / K_{obs}} where \eqn{K} is the total number of rows
#'     and \eqn{K_{obs}} is the number of observed rows.}
#'   \item{\code{"egk"}}{Expected Gaussian Kernel (Mesquita et al., 2019). Uses a Gaussian Mixture Model (GMM)
#'     to estimate the conditional distribution of missing values given observed values,
#'     then computes the expected kernel value via a Gamma approximation.
#'     Requires \code{gmm.means}, \code{gmm.sigmas}, and \code{gmm.weights} passed through \code{...}.}
#' }
#'
#' @param U Covariate matrix \eqn{U(K,N) = (u_1, \dots, u_N)}. Each row may be normalized in advance.
#' @param V Covariate matrix \eqn{V(K,M) = (v_1, \dots, v_M)}, typically used for prediction. If \code{NULL}, the default is \code{U}. May contain \code{NA} values.
#' @param beta Bandwidth parameter for the Gaussian kernel. Default is \code{0.5}.
#' @param na.method Method for handling \code{NA} values in \code{V}. Either \code{"pds"} or \code{"egk"}. Ignored if \code{V} has no \code{NA}.
#' @param ... Additional arguments for EGK method:
#'   \describe{
#'     \item{\code{gmm.G}}{Number of GMM components for EGK. Default is \code{3} (Mesquita et al., 2019).}
#'   }
#'
#' @return Kernel matrix \eqn{A(N,M)}.
#' @seealso \code{\link{nmfkc.kernel}}, \code{\link{nmfkc.kernel.beta.cv}}, \code{\link{nmfkc.kernel.beta.nearest.med}}
#' @export
#' @source Mesquita, D., Gomes, J. P., & Rodrigues, L. R. (2019).
#'   Gaussian kernels for incomplete data. \emph{Applied Soft Computing}, 77, 356--365.
#' @examples
#' U <- matrix(c(5,10,15,20,25),nrow=1)
#' V <- matrix(1:25,nrow=1)
#' A <- nmfkc.kernel.gaussian(U,V,beta=28/1000)
#' dim(A)
#'
#' # PDS example: V with NA in first row
#' U2 <- matrix(rnorm(20), nrow=2)
#' V2 <- matrix(rnorm(10), nrow=2)
#' V2[1, c(2,4)] <- NA
#' A2 <- nmfkc.kernel.gaussian(U2, V2, beta=0.5, na.method="pds")

nmfkc.kernel.gaussian <- function(U, V = NULL, beta = 0.5,
                                   na.method = c("pds", "egk"), ...){
  extra_args <- base::list(...)
  U <- as.matrix(U); storage.mode(U) <- "double"
  if (is.null(V)) V <- U else V <- as.matrix(V)
  storage.mode(V) <- "double"
  if (nrow(U) == 0) stop("'U' must have at least one row (feature).")
  if (nrow(U) != nrow(V)) stop("'U' and 'V' must have the same number of rows (features).")
  na.method <- match.arg(na.method)

  has_na <- anyNA(V)

  if (!has_na) {
    # --- No NA: fast vectorized computation ---
    G <- crossprod(U, V)
    u2 <- colSums(U * U)
    v2 <- colSums(V * V)
    D2 <- outer(u2, v2, "+") - 2 * G
    D2 <- pmax(D2, 0)
    K <- exp(-beta * D2)
  } else if (na.method == "pds") {
    # --- PDS: Partial Distance Strategy ---
    N <- ncol(U); M <- ncol(V); P <- nrow(U)
    K <- matrix(0, N, M)
    for (j in seq_len(M)) {
      obs <- !is.na(V[, j])
      P_obs <- sum(obs)
      if (P_obs == 0) next
      beta_adj <- beta * (P / P_obs)
      d2_j <- colSums((U[obs, , drop = FALSE] - V[obs, j])^2)
      K[, j] <- exp(-beta_adj * d2_j)
    }
  } else {
    # --- EGK: Expected Gaussian Kernel ---
    if (!requireNamespace("mclust", quietly = TRUE))
      stop("na.method='egk' requires the 'mclust' package. Install it with: install.packages('mclust')")
    gmm_G <- if (!is.null(extra_args$gmm.G)) extra_args$gmm.G else 3
    gmm <- mclust::Mclust(t(U), G = gmm_G, verbose = FALSE)
    gmm_means   <- gmm$parameters$mean
    gmm_sigmas  <- gmm$parameters$variance$sigma
    gmm_weights <- gmm$parameters$pro
    C_gmm <- length(gmm_weights)
    N <- ncol(U); M <- ncol(V); P <- nrow(U)
    K <- matrix(0, N, M)

    for (j in seq_len(M)) {
      obs_idx  <- which(!is.na(V[, j]))
      miss_idx <- which(is.na(V[, j]))
      P_obs  <- length(obs_idx)
      P_miss <- length(miss_idx)

      if (P_miss == 0) {
        # No missing: standard kernel
        d2_j <- colSums((U - V[, j])^2)
        K[, j] <- exp(-beta * d2_j)
        next
      }
      if (P_obs == 0) next

      x_obs <- V[obs_idx, j]

      # Conditional moments from GMM
      resp <- numeric(C_gmm)
      cond_mean_list <- vector("list", C_gmm)
      cond_var_list  <- vector("list", C_gmm)

      for (cc in seq_len(C_gmm)) {
        mu_O <- gmm_means[obs_idx, cc]
        mu_M <- gmm_means[miss_idx, cc]
        Sig_OO <- gmm_sigmas[obs_idx, obs_idx, cc]
        Sig_MO <- gmm_sigmas[miss_idx, obs_idx, cc]
        Sig_OM <- gmm_sigmas[obs_idx, miss_idx, cc]
        Sig_MM <- gmm_sigmas[miss_idx, miss_idx, cc]

        Sig_OO_inv <- solve(Sig_OO)
        diff_o <- x_obs - mu_O
        cond_mean_list[[cc]] <- as.vector(mu_M + Sig_MO %*% Sig_OO_inv %*% diff_o)
        cond_cov <- Sig_MM - Sig_MO %*% Sig_OO_inv %*% Sig_OM
        cond_var_list[[cc]] <- pmax(diag(cond_cov), 0)

        log_lik <- -0.5 * (P_obs * log(2 * pi) +
                            determinant(Sig_OO, logarithm = TRUE)$modulus +
                            sum(diff_o * (Sig_OO_inv %*% diff_o)))
        resp[cc] <- log(gmm_weights[cc]) + log_lik
      }

      # Normalize responsibilities (log-sum-exp trick)
      resp <- exp(resp - max(resp))
      resp <- resp / sum(resp)

      # Mixture moments for missing dimensions
      E_x_miss  <- rep(0, P_miss)
      E_x2_miss <- rep(0, P_miss)
      for (cc in seq_len(C_gmm)) {
        E_x_miss  <- E_x_miss  + resp[cc] * cond_mean_list[[cc]]
        E_x2_miss <- E_x2_miss + resp[cc] * (cond_mean_list[[cc]]^2 + cond_var_list[[cc]])
      }
      Var_x_miss <- pmax(E_x2_miss - E_x_miss^2, 0)

      # Compute EGK for each training sample
      for (i in seq_len(N)) {
        Ez_obs <- sum((U[obs_idx, i] - x_obs)^2)
        diff_miss <- U[miss_idx, i] - E_x_miss
        Ez_miss <- sum(diff_miss^2 + Var_x_miss)
        Ez <- Ez_obs + Ez_miss
        Varz <- sum(4 * diff_miss^2 * Var_x_miss + 2 * Var_x_miss^2)

        if (Varz < 1e-30) {
          K[i, j] <- exp(-beta * Ez)
        } else {
          alpha_g <- Ez^2 / Varz
          beta_g  <- Ez / Varz
          K[i, j] <- (beta_g / (beta_g + beta))^alpha_g
        }
      }
    }
  }

  K[K < 0] <- 0
  dimnames(K) <- list(colnames(U), colnames(V))
  attr(K, "params") <- beta
  attr(K, "kernel") <- "Gaussian"
  attr(K, "function.name") <- "nmfkc.kernel.gaussian"
  if (has_na) attr(K, "na.method") <- na.method
  return(K)
}


#' @title Create a kernel matrix from covariates
#' @description
#' \code{nmfkc.kernel} constructs a kernel matrix from covariate matrices.
#' It supports Gaussian, Exponential, Periodic, Linear, Normalized Linear, and Polynomial kernels.
#'
#' @param U Covariate matrix \eqn{U(K,N) = (u_1, \dots, u_N)}. Each row may be normalized in advance.
#' @param V Covariate matrix \eqn{V(K,M) = (v_1, \dots, v_M)}, typically used for prediction. If \code{NULL}, the default is \code{U}.
#' @param kernel Kernel function to use. Default is \code{"Gaussian"}. Options are \code{"Gaussian"}, \code{"Exponential"}, \code{"Periodic"}, \code{"Linear"}, \code{"NormalizedLinear"}, and \code{"Polynomial"}.
#' @param ... Additional arguments passed to the specific kernel function (e.g., \code{beta}, \code{degree}).
#'
#' @return Kernel matrix \eqn{A(N,M)}.
#' @seealso \code{\link{nmfkc.kernel.gaussian}}, \code{\link{nmfkc.cv}}
#' @export
#' @source Satoh, K. (2024). Applying Non-negative Matrix Factorization with Covariates to the Longitudinal Data as Growth Curve Model.
#'   arXiv preprint arXiv:2403.05359. \url{https://arxiv.org/abs/2403.05359}
#' @examples
#' # Example.
#' Y <- matrix(cars$dist,nrow=1)
#' U <- matrix(c(5,10,15,20,25),nrow=1)
#' V <- matrix(cars$speed,nrow=1)
#' A <- nmfkc.kernel(U,V,beta=28/1000)
#' dim(A)
#' result <- nmfkc(Y,A,rank=1)
#' plot(as.vector(V),as.vector(Y))
#' lines(as.vector(V),as.vector(result$XB),col=2,lwd=2)

nmfkc.kernel <- function(U, V = NULL,
                         kernel = c("Gaussian","Exponential","Periodic",
                                    "Linear","NormalizedLinear","Polynomial"),...){
  k_params <- list(...)
  U <- as.matrix(U); storage.mode(U) <- "double"
  if (is.null(V)) V <- U else V <- as.matrix(V)
  storage.mode(V) <- "double"
  if (nrow(U) == 0) stop("'U' must have at least one row (feature).")
  if (nrow(U) != nrow(V)) stop("'U' and 'V' must have the same number of rows (features).")
  kernel <- match.arg(kernel)
  G <- crossprod(U, V)

  # Determine the specific parameter used for the kernel (e.g., beta or degree)
  # This section extracts the effective parameter value to store in attributes.
  effective_param <- NULL
  if (kernel %in% c("Gaussian", "Exponential")) {
    effective_param <- if (!is.null(k_params$beta)) k_params$beta else 0.5
  } else if (kernel == "Periodic") {
    effective_param <- if (!is.null(k_params$beta)) k_params$beta else c(1,1)
  } else if (kernel == "Polynomial") {
    # For Polynomial, the effective param is (beta, degree) pair
    effective_param <- list(beta = if (!is.null(k_params$beta)) k_params$beta else 0,
                            degree = if (!is.null(k_params$degree)) k_params$degree else 2)
  }

  if (kernel %in% c("Gaussian","Exponential","Periodic")) {
    u2 <- colSums(U * U)
    v2 <- colSums(V * V)
    D2 <- outer(u2, v2, "+") - 2 * G
    D2 <- pmax(D2, 0)
  }

  K <- switch(kernel,
              Gaussian ={
                beta <- if (!is.null(k_params$beta)) k_params$beta else 0.5
                na_method <- if (!is.null(k_params$na.method)) k_params$na.method else "pds"
                KG <- nmfkc.kernel.gaussian(U, V, beta, na.method = na_method,
                                             gmm.G = k_params$gmm.G)
                attr(KG, "function.name") <- "nmfkc.kernel"
                KG
              },
              Exponential = {
                beta <- if (!is.null(k_params$beta)) k_params$beta else 0.5
                d <- sqrt(D2)
                exp(-beta * d)
              },
              Periodic = {
                beta <- if (!is.null(k_params$beta)) k_params$beta else c(1,1)
                if (length(beta) < 2L)
                  stop("For 'Periodic', set beta as c(beta1, beta2).")
                d <- sqrt(D2)
                exp(-beta[1] * (sin(beta[2] * d)^2))
              },
              Linear = G,
              NormalizedLinear = {
                nu <- sqrt(colSums(U * U))
                nv <- sqrt(colSums(V * V))
                nu[nu == 0] <- .Machine$double.eps
                nv[nv == 0] <- .Machine$double.eps
                G / outer(nu, nv)
              },
              Polynomial = {
                beta <- if (!is.null(k_params$beta)) k_params$beta else 0
                degree <- if (!is.null(k_params$degree)) k_params$degree else 2
                (G + beta)^degree
              }
  )
  K[K < 0] <- 0
  dimnames(K) <- list(colnames(U), colnames(V))

  # --- Store Kernel Metadata as Attributes (NEW BLOCK) ---
  # These attributes allow nmfkc to identify the kernel type and parameters used.
  attr(K, "params") <- effective_param
  attr(K, "kernel") <- kernel
  attr(K, "function.name") <- "nmfkc.kernel"
  # ------------------------------------------------------

  return(K)
}



#' @title Estimate Gaussian/RBF kernel parameter beta from covariates (supports landmarks)
#'
#' @description
#' Computes a data-driven reference scale for the Gaussian/RBF kernel from covariates
#' using a robust "median nearest-neighbor (or nearest-landmark) distance" heuristic,
#' and returns the corresponding kernel parameter \eqn{\beta}.
#'
#' The Gaussian/RBF kernel is assumed to be written in the form
#' \deqn{k(u,v) = \exp\{-\beta \|u-v\|^2\} = \exp\{-\|u-v\|^2/(2\sigma^2)\},}
#' hence \eqn{\beta = 1/(2\sigma^2)}. This function first estimates a typical distance
#' scale \eqn{\sigma_0} by the median of distances, then sets \eqn{\beta_0 = 1/(2\sigma_0^2)}.
#'
#' If \code{Uk} is \code{NULL}, \eqn{\sigma_0} is estimated as the median of
#' nearest-neighbor distances within \code{U} (excluding self-distance).
#' If \code{Uk} is provided, \eqn{\sigma_0} is estimated as the median of
#' nearest-landmark distances from each sample in \code{U} to its closest landmark in \code{Uk}.
#'
#' To control memory usage for large \code{N} (and \code{M}), distances are computed in blocks.
#' Optionally, columns of \code{U} can be randomly subsampled via \code{sample.size} to reduce cost.
#'
#' @details
#' \strong{Candidate grid:}
#' Along with \code{beta}, the function returns \code{beta_candidates}, a
#' logarithmic grid suitable for cross-validation.  The grid is symmetric on
#' the bandwidth scale \eqn{\sigma} around \eqn{\sigma_0}:
#' \deqn{\sigma = \sigma_0 \times 10^{t},}
#' and since \eqn{\beta = 1/(2\sigma^2)}, this corresponds to
#' \eqn{\beta = \beta_0 \times 10^{-2t}}.
#'
#' The grid of \eqn{t} values can be customized through the hidden argument
#' \code{candidates} (passed via \code{...}):
#' \itemize{
#'   \item \code{"7points"} (default): \eqn{t \in \{-1,-2/3,-1/3,0,1/3,2/3,1\}}
#'         (7 candidates spanning one decade, matches the grid used in the
#'         RFF-NMF research memo).
#'   \item \code{"4points"}: \eqn{t \in \{-1/2, 0, 1/2, 1\}} yielding
#'         \eqn{\beta_0 \times 10^{(1,0,-1,-2)}} (the legacy short grid).
#'   \item A numeric vector: user-specified \eqn{t} values.  The grid returned
#'         is \eqn{\beta_0 \times 10^{-2t}}.
#' }
#'
#' Prior to version 0.6.8, the grid depended on whether \code{Uk} was
#' supplied (4 candidates for \code{Uk = NULL}, 7 for supplied \code{Uk}).
#' The current implementation unifies both branches via \code{candidates}.
#'
#' \strong{Notes:}
#' \itemize{
#'   \item When \code{Uk} is identical to \code{U}, the function detects this case and excludes
#'         self-distances (distance 0) to avoid \eqn{\sigma_0=0}.
#'   \item \code{sample.size} performs random subsampling without setting a seed. For reproducible
#'         results, set \code{set.seed()} before calling this function.
#' }
#'
#' @param U A numeric matrix of covariates (\code{K x N}); columns are samples.
#' @param Uk An optional numeric matrix of landmarks (\code{K x M}); columns are landmark points.
#'   If provided, distances are computed from samples in \code{U} to landmarks in \code{Uk}.
#' @param block.size Integer. Number of columns of \code{U} processed per block when computing
#'   distances (controls memory usage). If \code{N <= 1000}, it is automatically set to \code{N}.
#' @param block.size.Uk Integer. Number of columns of \code{Uk} processed per block when \code{Uk}
#'   is not \code{NULL} (controls memory usage). If \code{M <= 2000}, it is automatically set to \code{M}.
#' @param sample.size Integer or \code{NULL}. If not \code{NULL}, randomly subsamples this many columns
#'   of \code{U} (without replacement) before computing distances, to reduce computational cost.
#' @param ... Additional arguments.  Hidden option \code{candidates} controls
#'   the candidate grid: one of \code{"7points"} (default), \code{"4points"},
#'   or a numeric vector of \eqn{t} values.  See Details.
#'
#' @return A list with elements:
#' \itemize{
#'   \item \code{beta}: Estimated kernel parameter \eqn{\beta_0 = 1/(2\sigma_0^2)}.
#'   \item \code{beta_candidates}: Numeric vector of candidate \eqn{\beta} values (logarithmic grid)
#'         intended for cross-validation.
#'   \item \code{dist_median}: The estimated distance scale \eqn{\sigma_0} (median of nearest-neighbor
#'         or nearest-landmark distances).
#'   \item \code{block.size.used}: The effective block size(s) used. Either a scalar (no \code{Uk}) or
#'         a named vector \code{c(U=..., Uk=...)} when \code{Uk} is provided.
#'   \item \code{sample.size.used}: The number of columns of \code{U} actually used (after subsampling).
#'   \item \code{uk_is_u}: Logical flag indicating whether \code{Uk} was detected as identical to \code{U}
#'         (only returned when \code{Uk} is provided).
#' }
#'
#' @examples
#' # Basic (nearest-neighbor within U)
#' U <- matrix(runif(20), nrow = 2)
#' beta_info <- nmfkc.kernel.beta.nearest.med(U)
#' beta0 <- beta_info$beta
#' betas <- beta_info$beta_candidates
#'
#' # With landmarks (nearest-landmark distances)
#' Uk <- matrix(runif(10), nrow = 2)
#' \donttest{
#' beta_info2 <- nmfkc.kernel.beta.nearest.med(U, Uk)
#' }
#'
#' @seealso \code{\link{nmfkc.kernel.gaussian}}, \code{\link{nmfkc.kernel.beta.cv}}
#' @export
nmfkc.kernel.beta.nearest.med <- function(
    U,
    Uk = NULL,
    block.size = 1000,
    block.size.Uk = 2000,
    sample.size = NULL,
    ...
){
  extra_bn <- list(...)
  if (!is.null(extra_bn$block.size)) block.size <- extra_bn$block.size
  if (!is.null(extra_bn$block.size.Uk)) block.size.Uk <- extra_bn$block.size.Uk
  if (!is.null(extra_bn$sample.size)) sample.size <- extra_bn$sample.size

  ## Hidden option `candidates` controls the bandwidth grid (see @details).
  ## Default "7points" unifies the legacy 4/7 branch inconsistency.
  candidates <- if (!is.null(extra_bn$candidates)) extra_bn$candidates else "7points"

  ## ---- resolve the t-grid from `candidates` ----
  t_grid <- if (is.character(candidates) && length(candidates) == 1L) {
    switch(candidates,
           "7points" = c(-1, -2/3, -1/3, 0, 1/3, 2/3, 1),
           "4points" = c(-1/2, 0, 1/2, 1),
           stop("'candidates' must be \"7points\", \"4points\", or a numeric vector."))
  } else if (is.numeric(candidates)) {
    as.numeric(candidates)
  } else {
    stop("'candidates' must be \"7points\", \"4points\", or a numeric vector.")
  }
  U <- as.matrix(U)
  storage.mode(U) <- "double"
  N <- ncol(U)
  if (N < 2) stop("U must have at least 2 columns.")

  # ---- optional subsampling over U columns (for speed) ----
  if (!is.null(sample.size)) {
    sample.size <- as.integer(sample.size)
    if (sample.size <= 1) stop("sample.size must be >= 2")
    if (sample.size < N) {
      idxU <- sample.int(N, sample.size)
      U <- U[, idxU, drop = FALSE]
      N <- ncol(U)
    }
  }
  sample.size.used <- N

  # If Uk is NULL, behave like the original function: NN within U (exclude self).
  if (is.null(Uk)) {
    X  <- t(U)                    # N x K
    if (N <= 1000) block.size <- N
    XX <- rowSums(X * X)          # length N
    min_d2 <- rep(Inf, N)

    for (i in seq(1, N, by = block.size)) {
      i2 <- min(i + block.size - 1, N)
      Xi <- X[i:i2, , drop = FALSE]
      Xi_norm <- rowSums(Xi * Xi)

      dist2 <- outer(Xi_norm, XX, "+") - 2 * Xi %*% t(X)
      idx <- i:i2
      dist2[cbind(seq_along(idx), idx)] <- Inf  # exclude self
      dist2[dist2 < 0] <- 0

      nn_local <- apply(dist2, 1, min)
      min_d2[idx] <- pmin(min_d2[idx], nn_local)

      rm(Xi, Xi_norm, dist2); gc(FALSE)
    }

    d_med <- stats::median(sqrt(min_d2))
    if (d_med <= 0) {
      warning("Median nearest-neighbor distance is 0 (identical points exist). Using machine epsilon.")
      d_med <- sqrt(.Machine$double.eps)
    }
    beta  <- 1 / (2 * d_med^2)
    return(list(
      beta = beta,
      beta_candidates = beta * 10^(-2 * t_grid),
      dist_median = d_med,
      block.size.used = block.size,
      sample.size.used = sample.size.used
    ))
  }

  # ---- Uk provided: nearest landmark distance (U vs Uk) ----
  Uk <- as.matrix(Uk)
  storage.mode(Uk) <- "double"
  if (nrow(Uk) != nrow(U)) stop("nrow(Uk) must equal nrow(U).")

  M <- ncol(Uk)
  if (M < 1) stop("Uk must have at least 1 column.")

  # We'll compute, for each column of U, min_j ||u_i - uk_j||^2.
  # Block over U (columns) and optionally over Uk to control memory.
  if (N <= 1000) block.size <- N
  if (M <= 2000) block.size.Uk <- M

  U2  <- colSums(U * U)           # length N
  Uk2 <- colSums(Uk * Uk)         # length M
  min_d2 <- rep(Inf, N)

  # Detect "Uk is U" case (same object / identical values) to exclude self-distance.
  uk_is_u <- (M == N) && isTRUE(all.equal(Uk, U, check.attributes = FALSE))

  for (i in seq(1, N, by = block.size)) {
    i2 <- min(i + block.size - 1, N)
    Ui <- U[, i:i2, drop = FALSE]
    Ui2 <- U2[i:i2]

    # we may need to accumulate min over Uk blocks
    cur_min <- rep(Inf, ncol(Ui))

    for (j in seq(1, M, by = block.size.Uk)) {
      j2 <- min(j + block.size.Uk - 1, M)
      Ukj <- Uk[, j:j2, drop = FALSE]
      Ukj2 <- Uk2[j:j2]

      # dist2: (block_U) x (block_Uk)
      # dist2 = ||u||^2 + ||uk||^2 - 2 uk^T u
      G <- crossprod(Ukj, Ui)  # (block_Uk) x (block_U)
      dist2 <- outer(Ui2, Ukj2, "+") - 2 * t(G)
      dist2[dist2 < 0] <- 0

      # exclude self-distances only when Uk == U
      if (uk_is_u) {
        # global indices in U: col(Ui) = (i..i2), col(Ukj) = (j..j2)
        # where they overlap, set that cell to Inf
        gi <- i:i2
        gj <- j:j2
        common <- intersect(gi, gj)
        if (length(common) > 0) {
          # map global index -> local positions
          li <- match(common, gi)
          lj <- match(common, gj)
          dist2[cbind(li, lj)] <- Inf
        }
      }

      cur_min <- pmin(cur_min, apply(dist2, 1, min))

      rm(Ukj, Ukj2, G, dist2); gc(FALSE)
    }

    min_d2[i:i2] <- pmin(min_d2[i:i2], cur_min)
    rm(Ui, Ui2, cur_min); gc(FALSE)
  }

  d_med <- stats::median(sqrt(min_d2))
  if (d_med <= 0) {
    warning("Median nearest-neighbor distance is 0 (identical points exist). Using machine epsilon.")
    d_med <- sqrt(.Machine$double.eps)
  }
  beta  <- 1 / (2 * d_med^2)

  list(
    beta = beta,
    beta_candidates = beta * 10^(-2 * t_grid),
    dist_median = d_med,
    block.size.used = c(U = block.size, Uk = block.size.Uk),
    sample.size.used = sample.size.used,
    uk_is_u = uk_is_u
  )
}





#' @title Optimize beta of the Gaussian kernel function by cross-validation
#' @description
#' \code{nmfkc.kernel.beta.cv} selects the optimal beta parameter of the kernel function by applying cross-validation over a set of candidate values.
#'
#' @param Y Observation matrix \eqn{Y(P,N)}.
#' @param rank Rank of the basis matrix.
#' @param U Covariate matrix \eqn{U(K,N) = (u_1, \dots, u_N)}. Each row may be normalized in advance.
#' @param V Covariate matrix \eqn{V(K,M) = (v_1, \dots, v_M)}, typically used for prediction. If \code{NULL}, the default is \code{U}.
#' @param beta A numeric vector of candidate kernel parameters to evaluate via cross-validation.
#' @param plot Logical. If TRUE (default), plots the objective function values for each candidate \code{beta}.
#' @param ... Additional arguments passed to \code{nmfkc.cv}.
#'
#' @return A list with components:
#' \item{beta}{The beta value that minimizes the cross-validation objective function.}
#' \item{objfunc}{Objective function values for each candidate \code{beta}.}
#' @seealso \code{\link{nmfkc.kernel.gaussian}}, \code{\link{nmfkc.kernel.beta.nearest.med}},
#'   \code{\link{nmfkc.kernel}}
#' @export
#' @examples
#' # Example.
#' Y <- matrix(cars$dist,nrow=1)
#' U <- matrix(c(5,10,15,20,25),nrow=1)
#' V <- matrix(cars$speed,nrow=1)
#' nmfkc.kernel.beta.cv(Y,rank=1,U,V,beta=25:30/1000)
#' A <- nmfkc.kernel(U,V,beta=28/1000)
#' result <- nmfkc(Y,A,rank=1)
#' plot(as.vector(V),as.vector(Y))
#' lines(as.vector(V),as.vector(result$XB),col=2,lwd=2)

nmfkc.kernel.beta.cv <- function(Y,rank=2,U,V=NULL,beta=NULL,plot=TRUE,...){
  extra_args <- list(...)
  if (!is.null(extra_args$Q)) rank <- extra_args$Q
  Q <- rank
  kernel_arg_names <- names(formals(nmfkc.kernel))
  cv_arg_names <- names(formals(nmfkc.cv))
  kernel_args_for_call <- extra_args[names(extra_args) %in% kernel_arg_names]
  cv_args_for_call <- extra_args[names(extra_args) %in% cv_arg_names]

  if(is.null(beta)){
    if(is.null(V)) V <- U
    med_args <- c(list(U = V), extra_args[names(extra_args) %in% names(formals(nmfkc.kernel.beta.nearest.med))])
    result.beta <- do.call("nmfkc.kernel.beta.nearest.med", med_args)
    beta <- result.beta$beta_candidates
    if (is.null(beta) || length(beta) == 0) stop("Failed to determine beta candidates from nearest-neighbor median.")
  }
  objfuncs <- numeric(length(beta))
  for(i in seq_along(beta)){
    start.time <- Sys.time()
    message(paste0("beta=",beta[i],"..."),appendLF=FALSE)

    kernel_args <- c(
      list(U = U, V = V, beta = beta[i]),
      kernel_args_for_call
    )
    A <- do.call("nmfkc.kernel", kernel_args)

    cv_args <- c(
      list(Y = Y, A = A, Q = Q),
      cv_args_for_call
    )
    result <- do.call("nmfkc.cv", cv_args)

    objfuncs[i] <- result$objfunc
    end.time <- Sys.time()
    diff.time <- difftime(end.time,start.time,units="sec")
    diff.time.st <- ifelse(diff.time<=180,paste0(round(diff.time,1),"sec"),
                           paste0(round(diff.time/60,1),"min"))
    message(diff.time.st)
  }
  i0 <- which.min(objfuncs)
  beta.best <- beta[i0]
  if(plot){
    plot(beta,objfuncs,type="l",col=2,xlab="beta",ylab="objfunc",log="x")
    graphics::points(beta[i0],objfuncs[i0],cex=3,col=2)
    graphics::text(beta,objfuncs,format(beta,scientific=TRUE,digits=5))
  }
  names(objfuncs) <- beta
  result <- list(beta=beta.best,objfunc=objfuncs)
  return(result)
}






#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------




#' @title Initialize W (X) matrix using NNDSVDar (Vectorized)
#' @description
#'   Internal function to compute the NNDSVDar (Nonnegative Double SVD and Random)
#'   initialization for the basis matrix X. This method fills the zero entries
#'   generated by basic NNDSVD with small random values scaled by the matrix average,
#'   improving stability and convergence.
#' @param Y Input matrix (P x N)
#' @param Q Rank (number of components)
#' @return X (P x Q) non-negative initial basis matrix
#' @keywords internal
#' @noRd
# Internal helper: initialize basis matrix X
.nmfkc_init_X <- function(Y, Q, X.init, Y.weights, seed, nstart, maxit, .eps) {
  # Impute NAs with row means for initialization
  Y_init <- Y
  if (is.matrix(Y.weights) && any(Y.weights == 0)) {
    row_means <- rowSums(Y) / (rowSums(Y.weights) + .eps)
    mask_binary <- (Y.weights > 0)
    idx_missing <- which(!mask_binary, arr.ind = TRUE)
    if (nrow(idx_missing) > 0) {
      Y_init[idx_missing] <- row_means[idx_missing[, 1]]
    }
  }

  if (is.matrix(X.init)) {
    X <- X.init
  } else if (is.character(X.init)) {
    if (!is.null(seed)) set.seed(seed)
    ## runif with nstart > 1 is an nmfkc-specific multi-start feature (10
    ## inner MU iterations per restart to rank the candidate seeds); all
    ## other string methods delegate to the shared .init_X_method().
    if (X.init == "runif" && nstart > 1) {
      best_obj <- Inf
      P <- nrow(Y_init); N <- ncol(Y_init)
      for (s in seq_len(nstart)) {
        Xs <- matrix(stats::runif(P * Q), nrow = P, ncol = Q)
        Bs <- matrix(stats::runif(Q * N), nrow = Q, ncol = N)
        for (iter in 1:10) {
          Bs <- Bs * (t(Xs) %*% Y_init) / (t(Xs) %*% Xs %*% Bs + .eps)
          Xs <- Xs * (Y_init %*% t(Bs)) / (Xs %*% Bs %*% t(Bs) + .eps)
        }
        obj <- sum((Y_init - Xs %*% Bs)^2)
        if (obj < best_obj) { best_obj <- obj; X <- Xs }
      }
    } else {
      ## Seed was already applied just above (`set.seed(seed)`), so pass
      ## seed = NULL to avoid double-setting; pass the outer MU maxit as
      ## kmeans iter.max to preserve pre-refactor behavior.
      X <- .init_X_method(X.init, Y_init, Q,
                          seed = NULL, nstart = nstart, kmeans.maxit = maxit)
    }
  } else if (ncol(Y) == Q) {
    X <- Y_init
  } else {
    X <- matrix(stats::runif(nrow(Y) * Q), nrow = nrow(Y), ncol = Q)
  }
  X
}


.nndsvdar <- function(Y, Q) {
  P <- nrow(Y)
  N <- ncol(Y)
  s <- svd(Y, nu = Q, nv = Q)
  W <- matrix(0, P, Q)
  W[, 1] <- sqrt(s$d[1]) * abs(s$u[, 1])
  if (Q > 1) {
    idx <- 2:Q
    U <- s$u[, idx, drop = FALSE]
    V <- s$v[, idx, drop = FALSE]
    D_sqrt <- sqrt(s$d[idx])
    U_p <- pmax(U, 0)
    U_n <- pmax(-U, 0)
    V_p <- pmax(V, 0)
    V_n <- pmax(-V, 0)
    norm_Up <- sqrt(colSums(U_p^2))
    norm_Un <- sqrt(colSums(U_n^2))
    norm_Vp <- sqrt(colSums(V_p^2))
    norm_Vn <- sqrt(colSums(V_n^2))
    Mp <- norm_Up * norm_Vp
    Mn <- norm_Un * norm_Vn
    eps <- .Machine$double.eps
    norm_Up_safe <- norm_Up; norm_Up_safe[norm_Up == 0] <- eps
    norm_Un_safe <- norm_Un; norm_Un_safe[norm_Un == 0] <- eps
    W_pos <- sweep(U_p, 2, (D_sqrt * Mp / norm_Up_safe), FUN = "*")
    W_neg <- sweep(U_n, 2, (D_sqrt * Mn / norm_Un_safe), FUN = "*")
    use_pos <- (Mp > Mn)
    W_combined <- W_pos
    W_combined[, !use_pos] <- W_neg[, !use_pos]
    W[, idx] <- W_combined
  }
  W[is.nan(W)] <- 0
  W[W < 0] <- 0
  avg_Y <- mean(Y)
  idx_zero <- which(W == 0)
  n_zero <- length(idx_zero)
  fill_values <- stats::runif(n_zero) * avg_Y / 100
  W[idx_zero] <- fill_values
  return(W)
}


## Internal: X initialization by named method.
## Shared across NMF variants (nmfkc, nmf.sem, ...) to avoid duplication
## of the "nndsvd" / "kmeans" / "kmeansar" / "runif" dispatch logic.
##
## @param method  One of "nndsvd", "kmeans", "kmeansar", "runif".
## @param Y       A P x N reference matrix used for SVD (nndsvd) or
##                column clustering (kmeans/kmeansar).  For "runif" only
##                its dimensions matter.  Must contain no NA.
## @param Q       Target rank.
## @param seed    Optional RNG seed (integer).  If not \code{NULL},
##                \code{set.seed(seed)} is called before any random step
##                so the result is reproducible.  Caller may also pre-seed
##                and pass \code{NULL} to leave the stream untouched.
## @param nstart  Number of restarts for \code{\link[stats]{kmeans}}
##                (used only by "kmeans" / "kmeansar").  Default 10.
## @param kmeans.maxit   \code{iter.max} for \code{\link[stats]{kmeans}}.
##                Default 100.
## @return A P x Q non-negative matrix.  Column normalization is the
##         caller's responsibility.
.init_X_method <- function(method, Y, Q,
                            seed = NULL,
                            nstart = 10,
                            kmeans.maxit = 100) {
  if (!is.null(seed)) set.seed(seed)
  P <- nrow(Y); N <- ncol(Y)
  if (identical(method, "nndsvd")) {
    if (Q <= min(P, N)) .nndsvdar(Y, Q)
    else matrix(stats::runif(P * Q), P, Q)
  } else if (identical(method, "kmeans") || identical(method, "kmeansar")) {
    res.kmeans <- if (N >= Q) {
      tryCatch(stats::kmeans(t(Y), centers = Q,
                             iter.max = kmeans.maxit,
                             nstart = nstart),
               error = function(e) NULL)
    } else NULL
    X <- if (!is.null(res.kmeans)) t(res.kmeans$centers)
         else matrix(stats::runif(P * Q), P, Q)
    if (identical(method, "kmeansar")) {
      avg_Y <- mean(Y)
      idx_zero <- which(X == 0)
      if (length(idx_zero) > 0)
        X[idx_zero] <- stats::runif(length(idx_zero)) * avg_Y / 100
    }
    X
  } else if (identical(method, "runif")) {
    matrix(stats::runif(P * Q), P, Q)
  } else {
    stop(".init_X_method: method must be one of \"nndsvd\", \"kmeans\", ",
         "\"kmeansar\", \"runif\"; got \"", method, "\".")
  }
}



#' @title Mean silhouette width from a distance matrix (Internal)
#' @description
#' Computes the standard mean silhouette width for a hard clustering,
#' given a pre-computed pairwise distance matrix and the cluster labels.
#' For each point \eqn{i}, \eqn{a(i)} is the mean distance to the other
#' members of its own cluster and \eqn{b(i)} is the minimum, over the
#' other clusters, of the mean distance to that cluster; the silhouette
#' is \eqn{(b(i) - a(i)) / \max(a(i), b(i))}.
#'
#' The distance matrix should describe distances in a \strong{fixed}
#' space that does not change with the NMF rank \eqn{Q} -- typically
#' the original data, \code{as.matrix(dist(t(Y)))} -- with the
#' per-sample hard labels \code{B.cluster}.  This is the k-means
#' convention.  (Computing silhouettes in the rank-\eqn{Q} \code{B.prob}
#' simplex instead makes the score monotone in \eqn{Q} and can hide
#' genuine cluster structure.)
#'
#' @param D An \eqn{N \times N} pairwise distance matrix (or object
#'   coercible to one) over the \eqn{N} samples.
#' @param labels Length-\eqn{N} integer cluster labels (\code{NA} allowed;
#'   such samples are dropped).
#'
#' @return The overall mean silhouette width (a single numeric), or
#'   \code{NA} when fewer than two clusters are present.
#' @keywords internal
#' @noRd
.silhouette.mean <- function(D, labels){
  D <- base::as.matrix(D)
  keep <- !is.na(labels)
  D <- D[keep, keep, drop = FALSE]
  labels <- labels[keep]
  clusters <- base::unique(labels)
  if (base::length(clusters) < 2) return(NA_real_)
  N <- base::length(labels)
  si <- base::numeric(N)
  for (i in base::seq_len(N)) {
    own  <- labels[i]
    same <- base::which(labels == own); same <- same[same != i]
    if (base::length(same) == 0) { si[i] <- 0; next }
    a <- base::mean(D[i, same])
    other <- base::setdiff(clusters, own)
    b <- base::min(base::vapply(other,
            function(cl) base::mean(D[i, labels == cl]), base::numeric(1)))
    si[i] <- if (base::max(a, b) > 0) (b - a) / base::max(a, b) else 0
  }
  base::mean(si)
}


#' @title Clustering-quality criteria from data and a coefficient matrix (Internal)
#' @description
#' Shared computation of the sample-clustering diagnostics used by
#' \code{\link{nmfkc.criterion}} (inside \code{nmfkc.rank}) and the
#' user-facing \code{\link{nmf.cluster.criteria}}.  From the original data
#' \eqn{Y} (\eqn{P \times N}) and a \eqn{Q \times N} coefficient/score
#' matrix \eqn{B} it returns:
#' \itemize{
#'   \item \code{dist.cor}: correlation between the sample distances in
#'     data space \code{dist(t(Y))} and in coefficient space
#'     \code{dist(t(B))} (embedding fidelity); always defined.
#'   \item \code{CPCC}: cophenetic correlation of a hierarchical
#'     clustering of \code{dist(t(B))} (Sokal-Rohlf); always defined for
#'     \eqn{Q \ge 2}.
#'   \item \code{silhouette}: mean silhouette width of the hard sample
#'     labels (argmax of the column-normalized \eqn{B}) in data space.
#'     This requires a \strong{non-negative} \eqn{B} (a valid membership
#'     simplex); for a signed \eqn{B} it is \code{NA}.
#' }
#' \code{dist.cor} and \code{CPCC} are distance-based and therefore work
#' for signed \eqn{B} too; only the hard-label \code{silhouette} needs
#' \eqn{B \ge 0}.
#' @param Y Data matrix (\eqn{P \times N}).
#' @param B Coefficient/score matrix (\eqn{Q \times N}).
#' @param labels Optional pre-computed hard labels (length \eqn{N}); when
#'   supplied they are used as-is for the silhouette (e.g.\ the
#'   \code{B.cluster} already held by a fitted \code{nmfkc} model),
#'   otherwise they are derived from \eqn{B} when \eqn{B \ge 0}.
#' @return A list: \code{silhouette}, \code{CPCC}, \code{dist.cor},
#'   \code{cluster} (the hard labels, or \code{NULL}), and \code{hard}
#'   (whether hard clustering was possible).
#' @keywords internal
#' @noRd
.cluster.criteria <- function(Y, B, labels = NULL) {
  Y <- base::as.matrix(Y); B <- base::as.matrix(B)
  Q <- base::nrow(B)
  dY <- stats::dist(base::t(Y))
  dB <- stats::dist(base::t(B))
  dist.cor <- stats::cor(base::as.vector(dY), base::as.vector(dB))
  CPCC <- if (Q >= 2) {
    coph <- stats::cophenetic(stats::hclust(dB))
    stats::cor(base::as.vector(dB), base::as.vector(coph))
  } else NA_real_
  nonneg <- base::all(base::is.finite(B)) && base::all(B >= 0)
  if (base::is.null(labels) && nonneg) {
    B.prob <- base::t(base::t(B) / (base::colSums(B) + .Machine$double.eps))
    labels <- base::apply(B.prob, 2, base::which.max)
    labels[base::colSums(B.prob) == 0] <- NA
  }
  silhouette <- if (!base::is.null(labels))
    .silhouette.mean(dY, labels) else NA_real_
  base::list(silhouette = silhouette, CPCC = CPCC, dist.cor = dist.cor,
             cluster = labels, hard = nonneg)
}





#' @title Unified R-squared computation (3 variants)
#' @description
#' Internal helper used by every NMF function in the package to report
#' three goodness-of-fit summaries on the same scale:
#' \itemize{
#'   \item \code{r.squared}: Pearson \eqn{\mathrm{cor}(Y, \widehat{Y})^2},
#'         scale-invariant, in \eqn{[0, 1]}.
#'   \item \code{r.squared.uncentered}: uncentered \eqn{R^2}
#'         \eqn{1 - \|Y - \widehat{Y}\|_F^2 / \|Y\|_F^2}.  Baseline is
#'         the zero matrix (natural for non-negative factorizations
#'         without an intercept).  Matches the "uncentered \eqn{R^2}"
#'         of intercept-free regression.  In \eqn{(-\infty, 1]}.
#'   \item \code{r.squared.centered}: row-mean centered
#'         \eqn{1 - \|Y - \widehat{Y}\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2},
#'         where \eqn{\bar Y_{p\cdot}} is the per-row mean over valid
#'         entries.  Equivalent to the multivariate regression
#'         \eqn{R^2} where each row is treated as one response.  In
#'         \eqn{(-\infty, 1]}.
#' }
#' All three respect the optional non-negative weight matrix
#' \code{Y.weights}: cells with \code{Y.weights == 0} (the standard
#' NA-mask / hold-out convention) are excluded from the sums.
#' @param Y observed matrix (\eqn{P \times N}).
#' @param Y_hat fitted matrix (\eqn{P \times N}).
#' @param Y.weights optional non-negative weight matrix of the same
#'   dimension as \code{Y}.  \code{NULL} means all weights are 1.
#' @return A list with components \code{r.squared}, \code{r.squared.uncentered},
#'   \code{r.squared.centered}, each a length-1 numeric (\code{NA_real_}
#'   when ill-defined, e.g.\ \code{||Y||_F = 0} or fewer than two valid
#'   entries).
#' @keywords internal
#' @noRd
.r.squared.all <- function(Y, Y_hat, Y.weights = NULL) {
  Y     <- base::as.matrix(Y)
  Y_hat <- base::as.matrix(Y_hat)
  P <- base::nrow(Y); N <- base::ncol(Y)

  if (base::is.null(Y.weights)) {
    Wm <- base::matrix(1, P, N)
  } else {
    Wm <- base::as.matrix(Y.weights)
    if (!base::identical(base::dim(Wm), base::c(P, N)))
      base::stop(".r.squared.all: Y.weights must match dim(Y).")
  }
  valid <- (Wm > 0)
  n_valid <- base::sum(valid)

  if (n_valid < 2) {
    return(base::list(r.squared          = NA_real_,
                      r.squared.uncentered     = NA_real_,
                      r.squared.centered = NA_real_))
  }

  y_v    <- Y[valid]
  yhat_v <- Y_hat[valid]
  sse    <- base::sum((y_v - yhat_v)^2)

  ## (A) Pearson cor^2
  r2_cor <- base::tryCatch(stats::cor(y_v, yhat_v)^2,
                           error = function(e) NA_real_)

  ## (C) Uncentered: 1 - SSE / ||Y||_F^2 (valid entries; baseline = 0)
  ss_y   <- base::sum(y_v^2)
  r2_unc <- if (is.finite(ss_y) && ss_y > 0) 1 - sse / ss_y else NA_real_

  ## (D) Row-mean centered: per-row mean over valid entries
  row_means <- base::numeric(P)
  for (p in base::seq_len(P)) {
    vp <- valid[p, ]
    if (base::any(vp)) row_means[p] <- base::mean(Y[p, vp])
    else                row_means[p] <- NA_real_
  }
  ## Build (Y - rowMean)^2 broadcasting row_means down columns, then
  ## sum over valid cells only.  rows with no valid entries contribute 0.
  Y_centered_sq <- (Y - row_means)^2
  Y_centered_sq[!valid] <- 0
  ## Drop rows with NA row_means (i.e., empty rows under masking)
  Y_centered_sq[base::is.na(Y_centered_sq)] <- 0
  ss_centered <- base::sum(Y_centered_sq)
  r2_centered <- if (is.finite(ss_centered) && ss_centered > 0) 1 - sse / ss_centered
                 else NA_real_

  base::list(r.squared          = r2_cor,
             r.squared.uncentered     = r2_unc,
             r.squared.centered = r2_centered)
}


#' @title Effective rank of a coefficient/score matrix (Internal)
#' @description
#' Returns the \strong{effective rank}: \eqn{\exp} of the Shannon
#' entropy of the explained-variance distribution
#' \eqn{p_k = \mathrm{var}(B_{k\cdot}) / \sum_j \mathrm{var}(B_{j\cdot})}
#' over the rows (latent factors) of \code{B}.  By the trace identity
#' \eqn{\sum_k \mathrm{var}(B_{k\cdot}) = \mathrm{tr}(\mathrm{Cov}(B))},
#' each \eqn{p_k} is the exact fraction of the total coefficient
#' variance carried by factor \eqn{k}, so the entropy measures how
#' evenly that variance is spread across factors.  Ranges in
#' \eqn{[1, Q]}.  Returns \code{NA} when there are fewer than two
#' factors (\eqn{Q < 2}) -- the entropy of a single point is trivially
#' 0 -- or fewer than two units (\eqn{N < 2}), or zero total variance.
#' @param B A \eqn{Q \times N} coefficient/score matrix (factors in
#'   rows, units in columns).
#' @return A single numeric (the effective rank), or \code{NA_real_}.
#' @keywords internal
#' @noRd
.effective.rank <- function(B) {
  B <- base::as.matrix(B)
  Q <- base::nrow(B); N <- base::ncol(B)
  if (Q < 2 || N < 2) return(NA_real_)
  v <- base::apply(B, 1, stats::var)
  s <- base::sum(v, na.rm = TRUE)
  if (!base::is.finite(s) || s <= 0) return(NA_real_)
  p <- v / s
  p <- p[p > 0]
  base::exp(-base::sum(p * base::log(p)))
}


#' @title Print a unified "Statistics" block for an NMF summary (Internal)
#' @description
#' Prints the common fit-statistics block shared by the NMF summary
#' methods (objective, the three R-squared variants, residual SE, MAE,
#' and effective rank), reading them from a list/object \code{x}.  Each
#' field is printed only when present and finite, so models that lack a
#' given statistic (e.g.\ \code{nmfkc.net} has no residual SE) simply
#' omit that line.  Labels are left-padded to a common width so values
#' align across all summaries.
#' @param x A list/object with any of \code{objfunc}, \code{r.squared},
#'   \code{r.squared.uncentered}, \code{r.squared.centered},
#'   \code{sigma}, \code{mae}, \code{effective.rank}, \code{rank}.
#' @param header Section header (default \code{"Statistics:"}).
#' @param digits Significant digits for numeric fields.
#' @return \code{NULL}, invisibly (called for its printing side effect).
#' @keywords internal
#' @noRd
.print.fit.statistics <- function(x, header = "Statistics:", digits = 4) {
  base::cat("\n", header, "\n", sep = "")
  w <- 24L
  num <- function(label, val) {
    if (!base::is.null(val) && base::length(val) == 1L && base::is.finite(val))
      base::cat(base::sprintf("  %-*s %s\n", w, label,
                              base::format(val, digits = digits)))
  }
  num("Objective function:",     x$objfunc)
  num("R-squared (cor^2):",      x$r.squared)
  num("R-squared (uncentered):", x$r.squared.uncentered)
  num("R-squared (centered):",   x$r.squared.centered)
  num("Residual Std Error:",     x$sigma)
  num("Mean Absolute Error:",    x$mae)
  if (!base::is.null(x$effective.rank) && base::is.finite(x$effective.rank) &&
      !base::is.null(x$rank) && base::is.finite(x$rank))
    base::cat(base::sprintf("  %-*s %.2f / %d  (%.1f%%)\n", w, "Effective Rank:",
                            x$effective.rank, x$rank,
                            100 * x$effective.rank / x$rank))
  base::invisible(NULL)
}


#' @title Print a unified "Structure Diagnostics" block (Internal)
#' @description
#' Prints the common structure-diagnostics block shared by the NMF
#' summary methods: one sparsity line per factor matrix plus an optional
#' clustering-crispness line.  Sparsities are supplied as a named
#' numeric vector (name = matrix label, value = fraction of near-zero
#' entries); labels are aligned to the same width as
#' \code{\link{.print.fit.statistics}}.  Model-specific extras (value
#' ranges, cluster sizes, \dots) are printed by the caller after this
#' block.
#' @param sparsity Named numeric vector of sparsity fractions in
#'   \eqn{[0, 1]}; names become the row labels (e.g.\ \code{"Basis (X)"}).
#' @param crispness Optional clustering crispness in \eqn{[1/Q, 1]}.
#' @param header Section header (default \code{"Structure Diagnostics:"}).
#' @return \code{NULL}, invisibly.
#' @keywords internal
#' @noRd
.print.structure.diagnostics <- function(sparsity = NULL, crispness = NULL,
                                         header = "Structure Diagnostics:") {
  base::cat("\n", header, "\n", sep = "")
  w <- 24L
  if (!base::is.null(sparsity)) {
    for (lab in base::names(sparsity)) {
      v <- sparsity[[lab]]
      if (!base::is.null(v) && base::is.finite(v))
        base::cat(base::sprintf("  %-*s %.1f%% (< 1e-4)\n", w,
                                base::paste0(lab, " Sparsity:"), 100 * v))
    }
  }
  if (!base::is.null(crispness) && base::is.finite(crispness))
    base::cat(base::sprintf("  %-*s %s (range: 1/Q-1, closer to 1 = more decisive assignment)\n",
                            w, "Clustering Crispness:",
                            base::format(crispness, digits = 4)))
  base::invisible(NULL)
}


#' @title Element-wise CV fold construction (Internal)
#' @description
#' Randomly partitions the non-\code{NA} elements of \code{Y} into
#' \code{nfolds} groups of (nearly) equal size for Wold-style
#' element-wise cross-validation.  Shared by the element-wise
#' \code{*.ecv} functions (\code{\link{nmfkc.ecv}}, \code{nmfae.ecv},
#' \code{nmfkc.signed.ecv}, \code{nmfae.signed.ecv}); the symmetric
#' \code{\link{nmfkc.net.ecv}} uses upper-triangle folds instead.
#' @param Y The observation matrix.
#' @param nfolds Number of folds.
#' @param seed Optional integer seed (set immediately before sampling so
#'   that the partition is reproducible).
#' @return A list of length \code{nfolds}; each element is an integer
#'   vector of linear indices into \code{Y}.
#' @keywords internal
#' @noRd
.ecv.make.folds <- function(Y, nfolds, seed = NULL) {
  if (!base::is.null(seed)) base::set.seed(seed)
  valid_indices <- base::which(!base::is.na(Y))
  n_valid <- base::length(valid_indices)
  perm_indices <- base::sample(valid_indices)
  folds <- base::vector("list", nfolds)
  chunk_size <- n_valid %/% nfolds
  remainder  <- n_valid %% nfolds
  start_idx <- 1L
  for (k in 1:nfolds) {
    current_size <- chunk_size + base::ifelse(k <= remainder, 1L, 0L)
    end_idx <- start_idx + current_size - 1L
    folds[[k]] <- perm_indices[start_idx:end_idx]
    start_idx <- end_idx + 1L
  }
  folds
}


#' @title Run an element-wise CV loop over a configuration list (Internal)
#' @description
#' Shared driver for every element-wise CV function -- the single-rank
#' ones (\code{\link{nmfkc.ecv}}, \code{\link{nmfkc.net.ecv}},
#' \code{nmfkc.signed.ecv}) and the \eqn{(Q, R)}-grid ones
#' (\code{nmfae.ecv}, \code{nmfae.signed.ecv}).  Configurations are
#' addressed by \strong{index}: for each config \eqn{i} (labelled by
#' \code{labels[i]}) and each of \code{nfolds} folds it calls the
#' model-specific \code{run_one(i, k)} closure -- which masks the
#' held-out fold, refits the model at config \eqn{i}, and returns that
#' fold's mean held-out loss -- then aggregates per config
#' (\code{objfunc} = mean over folds, \code{sigma} = its square root for
#' a non-negative objective, else \code{NA}).
#' @param labels Character vector of config labels (e.g.\ \code{"Q=2"} or
#'   \code{"Q=2,R=1"}); its length is the number of configs.
#' @param nfolds Number of folds.
#' @param run_one A function \code{run_one(i, k)} returning the held-out
#'   mean loss for config index \code{i} and fold \code{k}.
#' @param progress Optional \code{function(i, objfunc, sigma)} called
#'   after each config for progress reporting (default none).
#' @return A list with \code{objfunc} (named numeric), \code{sigma}
#'   (named numeric), and \code{objfunc.fold} (named list of per-fold
#'   loss vectors).  Callers using a loss that can be negative (e.g.\ KL)
#'   may additionally force \code{sigma} to \code{NA}.
#' @keywords internal
#' @noRd
.ecv.run <- function(labels, nfolds, run_one, progress = NULL) {
  n   <- base::length(labels)
  obj <- stats::setNames(base::numeric(n), labels)
  sig <- stats::setNames(base::numeric(n), labels)
  fld <- stats::setNames(base::vector("list", n), labels)
  for (i in base::seq_len(n)) {
    objs <- base::numeric(nfolds)
    for (k in 1:nfolds) objs[k] <- run_one(i, k)
    fld[[i]] <- objs
    obj[i]   <- base::mean(objs)
    sig[i]   <- if (base::is.finite(obj[i]) && obj[i] >= 0)
                  base::sqrt(obj[i]) else NA_real_
    if (!base::is.null(progress)) progress(i, obj[i], sig[i])
  }
  base::list(objfunc = obj, sigma = sig, objfunc.fold = fld)
}


#' @title Extract the Q x N coefficient/score matrix of a fitted MU model (Internal)
#' @description
#' Maps each multiplicative-update fit to its natural \eqn{Q \times N}
#' coefficient/score matrix for clustering diagnostics: \code{B} for
#' \code{nmfkc}/\code{nmfkc.signed}, \code{H} for \code{nmfae}/
#' \code{nmfae.signed}, \eqn{X^\top} for \code{nmfkc.net} (node
#' membership), the BLUP scores for \code{nmfre}, and
#' \eqn{C_1 Y_1 + C_2 Y_2} for \code{nmf.ffb}/\code{nmf.sem} (which needs
#' the exogenous block \code{Y2}).
#' @param object A fitted MU model.
#' @param Y The data matrix passed to \code{\link{nmf.cluster.criteria}} (used as
#'   \eqn{Y_1} for \code{nmf.ffb}).
#' @param Y2 Exogenous block, required only for \code{nmf.ffb}/\code{nmf.sem}.
#' @return A \eqn{Q \times N} numeric matrix.
#' @keywords internal
#' @noRd
.nmf.cluster.criteria.coef <- function(object, Y, Y2 = NULL) {
  if (base::inherits(object, "nmfkc.net")) return(base::t(object$X))
  if (base::inherits(object, "nmfae"))     return(object$H)
  if (base::inherits(object, c("nmf.ffb", "nmf.sem"))) {
    if (base::is.null(Y2))
      base::stop("For nmf.ffb / nmf.sem, also pass the exogenous block via Y2=.",
                 call. = FALSE)
    return(object$C1 %*% base::as.matrix(Y) + object$C2 %*% base::as.matrix(Y2))
  }
  if (base::inherits(object, "nmfre"))     return(object$B.blup)
  if (base::inherits(object, "nmfkc"))     return(object$B)
  base::stop("nmf.cluster.criteria(): unsupported object class '",
             base::paste(base::class(object), collapse = "/"), "'.", call. = FALSE)
}


#' @title Sample-clustering quality across ranks
#' @description
#' Computes the clustering-quality criteria \code{silhouette},
#' \code{CPCC}, and \code{dist.cor} for a list of models fitted at
#' different ranks (or a single fit), returning one row per rank.  These
#' are \strong{clustering-stability} diagnostics (how decisively and
#' faithfully the samples cluster), conceptually separate from the
#' rank-selection \code{*.rank} functions (which use r.squared, effective
#' rank, and ECV) and complementary to \code{\link{nmf.cluster.flow}}
#' (which shows how the hard clustering itself changes across ranks).
#'
#' Hard sample clustering requires a non-negative coefficient/score
#' matrix (so the columns form a membership simplex); when a model's
#' coefficient is signed (e.g.\ \code{nmfkc.signed}, \code{nmfae.signed},
#' \code{nmfre} fits whose coefficient has negative entries) the
#' hard-label \code{silhouette} is \code{NA} while the distance-based
#' \code{CPCC} and \code{dist.cor} are still computed.
#'
#' @param fits A list of fitted models, one per rank, all over the same
#'   \eqn{N} individuals (a single fitted model is also accepted and
#'   wrapped automatically).  Supported families: \code{\link{nmfkc}},
#'   \code{\link{nmfkc.signed}}, \code{\link{nmfae}}, \code{nmfae.signed},
#'   \code{\link{nmfkc.net}}, \code{\link{nmfre}}, and
#'   \code{\link{nmf.sem}} / \code{nmf.ffb}.
#' @param Y The original data matrix used to fit the models (\eqn{Y_1}
#'   for \code{nmf.ffb}); required for the data-space distances.
#' @param Y2 Exogenous block, required only for \code{nmf.ffb} /
#'   \code{nmf.sem}.
#' @param names Optional character vector (length \code{length(fits)}) of
#'   x-axis tick labels.  Defaults to each result's \code{$rank}.
#' @param plot Logical; draw the diagnostics plot immediately
#'   (default \code{TRUE}); see \code{\link{plot.nmf.cluster.criteria}}.
#' @param ... When \code{plot = TRUE}, graphical arguments forwarded to
#'   \code{\link{plot.nmf.cluster.criteria}}.
#' @return An object of class \code{"nmf.cluster.criteria"} (returned
#'   invisibly): a list with \code{criteria} (a data frame with one row
#'   per result and columns \code{rank}, \code{silhouette}, \code{CPCC},
#'   \code{dist.cor}, and \code{hard}) and \code{labels} (the x-axis
#'   labels).  Results are kept in the given order (not sorted).
#' @seealso \code{\link{nmf.cluster.flow}}, \code{\link{nmfkc.rank}}
#' @export
#' @examples
#' \donttest{
#' Y <- t(as.matrix(iris[, 1:4]))
#' fits <- lapply(2:6, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
#' cc <- nmf.cluster.criteria(fits, Y, plot = FALSE)
#' cc$criteria
#' plot(cc)
#' }
nmf.cluster.criteria <- function(fits, Y, Y2 = NULL, names = NULL,
                                 plot = TRUE, ...) {
  if (base::missing(Y))
    base::stop("nmf.cluster.criteria() requires the original data matrix Y.", call. = FALSE)
  ## A single fitted model (has a scalar $rank) is wrapped into a list.
  if (!base::is.null(fits$rank)) fits <- base::list(fits)
  if (!base::is.list(fits) || base::length(fits) < 1)
    base::stop("`fits` must be a fitted model or a non-empty list of them.",
               call. = FALSE)
  Y <- base::as.matrix(Y)

  ## Results are taken in the given order (NOT sorted).
  rows <- base::lapply(fits, function(f) {
    B  <- .nmf.cluster.criteria.coef(f, Y, Y2)
    cc <- .cluster.criteria(Y, B)
    rank <- if (!base::is.null(f$rank)) base::as.integer(f$rank)
            else base::nrow(base::as.matrix(B))
    base::data.frame(rank = rank, silhouette = cc$silhouette, CPCC = cc$CPCC,
                     dist.cor = cc$dist.cor, hard = cc$hard)
  })
  criteria <- base::do.call(base::rbind, rows)
  base::rownames(criteria) <- NULL

  R <- base::nrow(criteria)
  if (base::is.null(names)) labels <- base::as.character(criteria$rank)
  else {
    labels <- base::as.character(names)
    if (base::length(labels) != R)
      base::stop("`names` must have length(fits) = ", R, " entries.", call. = FALSE)
  }

  out <- base::list(criteria = criteria, labels = labels)
  base::class(out) <- "nmf.cluster.criteria"
  if (plot) graphics::plot(out, ...)
  base::invisible(out)
}


#' @title Plot clustering-quality criteria across a sequence of fits
#' @description
#' Line plot of \code{silhouette}, \code{CPCC}, and \code{dist.cor}
#' against the result index, for an object from
#' \code{\link{nmf.cluster.criteria}}.  X-axis ticks default to each
#' result's \code{$rank} (overridable via the \code{names} argument of
#' \code{\link{nmf.cluster.criteria}}).
#' @param x An object of class \code{"nmf.cluster.criteria"}.
#' @param main Plot title.
#' @param xlab,ylab Axis labels.
#' @param lwd Line width.
#' @param ... Further arguments passed to the initial
#'   \code{\link[graphics]{plot}}.
#' @return \code{x}, invisibly.
#' @seealso \code{\link{nmf.cluster.criteria}}
#' @export
plot.nmf.cluster.criteria <- function(x, main = "Clustering quality across rank",
                                      xlab = "rank (Q)", ylab = "criterion",
                                      lwd = 2, ...) {
  cr <- x$criteria
  labels <- if (!base::is.null(x$labels)) x$labels else base::as.character(cr$rank)
  xi <- base::seq_len(base::nrow(cr))       # result index on the x-axis
  metrics <- c(silhouette = 7, CPCC = 6, dist.cor = 5)  # name = colour
  yr <- base::range(base::unlist(cr[base::names(metrics)]), na.rm = TRUE)
  if (!base::all(base::is.finite(yr))) yr <- c(0, 1)
  graphics::plot(NA, xlim = base::range(xi), ylim = yr, xaxt = "n",
                 xlab = xlab, ylab = ylab, main = main, ...)
  graphics::axis(1, at = xi, labels = labels)
  for (q in xi) graphics::abline(v = q, col = "gray90", lwd = 0.5)
  for (m in base::names(metrics)) {
    v <- cr[[m]]
    if (base::all(base::is.na(v))) next
    graphics::lines(xi, v, col = metrics[[m]], lwd = lwd)
    graphics::points(xi, v, pch = 16, col = metrics[[m]], cex = 0.8)
  }
  graphics::legend("bottomright", legend = base::names(metrics),
                   col = base::unlist(metrics), lty = 1, lwd = 2,
                   bg = "white", cex = 0.8)
  base::invisible(x)
}


#' @title Print method for nmf.cluster.criteria objects
#' @param x An object of class \code{"nmf.cluster.criteria"}.
#' @param ... Passed to the criteria table's \code{print}.
#' @return \code{x}, invisibly.
#' @export
print.nmf.cluster.criteria <- function(x, ...) {
  labels <- if (!base::is.null(x$labels)) x$labels else base::as.character(x$criteria$rank)
  base::cat("Sample-clustering quality across ", base::nrow(x$criteria),
            " results [", base::paste(labels, collapse = ", "), "]\n", sep = "")
  base::cat("(silhouette / CPCC / dist.cor; silhouette = NA when the coefficient is signed)\n\n")
  base::print(x$criteria, row.names = FALSE, ...)
  base::invisible(x)
}


#' @title Hard cluster labels of a fitted MU model (Internal)
#' @description
#' Returns the per-individual hard cluster labels (argmax of the
#' \eqn{Q \times N} coefficient/score matrix) for a fitted model, used by
#' \code{\link{nmf.cluster.flow}}.  For \code{nmfkc} this equals the
#' stored \code{B.cluster}.
#' @param object A fitted MU model.
#' @return An integer vector of length \eqn{N} (named by the columns of
#'   the coefficient matrix when available).
#' @keywords internal
#' @noRd
.nmf.hard.labels <- function(object) {
  B <- if (base::inherits(object, "nmfkc.net")) base::t(object$X)
       else if (base::inherits(object, "nmfae")) object$H
       else if (base::inherits(object, "nmfre")) object$B.blup
       else if (base::inherits(object, "nmfkc")) object$B
       else base::stop("nmf.cluster.flow(): unsupported model class '",
                       base::paste(base::class(object), collapse = "/"),
                       "'.", call. = FALSE)
  B <- base::as.matrix(B)
  lab <- base::apply(B, 2, base::which.max)
  nm <- base::colnames(B)
  if (!base::is.null(nm)) base::names(lab) <- nm
  lab
}


#' @title Place individuals within a rank, with inter-cluster gaps (Internal)
#' @description
#' Layout for one rank of \code{\link{nmf.cluster.flow}}.  Clusters are
#' stacked in increasing order of the mean key (\code{key}) of their
#' members -- a barycenter ordering that reduces alluvial crossings --
#' and within each cluster the members are sorted by \code{key}.  Members
#' are placed one unit apart, and a \strong{gap} of \eqn{N / k} units
#' (one average cluster's worth, \eqn{k} = number of clusters) is left
#' between consecutive clusters so the cluster boxes are clearly
#' separated.  Positions are normalized per rank in the plot, so the
#' varying total span across ranks does not matter.
#' @param labels Integer cluster labels at this rank (length \eqn{N}).
#' @param key Numeric ordering key (length \eqn{N}); the positions of the
#'   adjacent already-placed rank, or, for the reference rank, the
#'   reference labels themselves (so clusters stack in id order).
#' @return A numeric vector of positions for this rank.
#' @keywords internal
#' @noRd
.flow.place <- function(labels, key) {
  N <- base::length(labels)
  cls <- base::unique(labels)
  bary <- base::vapply(cls, function(c) base::mean(key[labels == c], na.rm = TRUE),
                       base::numeric(1))
  cls <- cls[base::order(bary)]
  gap <- N / base::length(cls)          # inter-cluster gap (one avg cluster)
  y <- base::numeric(N)
  pos <- 0
  for (c in cls) {
    idx <- base::which(labels == c)
    idx <- idx[base::order(key[idx])]
    nc <- base::length(idx)
    y[idx] <- pos + (base::seq_len(nc) - 1)
    pos <- pos + nc + gap               # advance past this cluster + gap
  }
  y
}


#' @title Adjusted Rand Index between two hard clusterings (Internal)
#' @description
#' Standard adjusted Rand index (Hubert & Arabie 1985) measuring the
#' agreement between two label vectors over the same items, used by
#' \code{\link{nmf.cluster.flow}} for adjacent ranks.  \code{NA} items
#' (in either labelling) are dropped pairwise.
#' @param a,b Integer label vectors of equal length.
#' @return The adjusted Rand index in \eqn{[-1, 1]} (\code{NA} when it is
#'   undefined, e.g.\ fewer than two valid items or a degenerate pair).
#' @keywords internal
#' @noRd
.ari <- function(a, b) {
  keep <- !(base::is.na(a) | base::is.na(b))
  a <- a[keep]; b <- b[keep]
  n <- base::length(a)
  if (n < 2) return(NA_real_)
  tab <- base::table(a, b)
  cc <- function(x) base::sum(x * (x - 1) / 2)
  sij <- cc(base::as.vector(tab))
  si  <- cc(base::rowSums(tab))
  sj  <- cc(base::colSums(tab))
  expct <- si * sj / (n * (n - 1) / 2)
  maxi  <- (si + sj) / 2
  if (maxi - expct == 0) return(NA_real_)
  (sij - expct) / (maxi - expct)
}


#' @title Cluster-flow (alluvial) diagram across a sequence of fits
#' @description
#' Visualizes how the hard sample clustering changes across a sequence of
#' fitted models -- typically the same model at increasing ranks, but
#' also different models at the same rank.  Each individual is a line
#' flowing left-to-right across the results (x-axis); its vertical
#' position at each result is determined by its cluster, and clusters are
#' reordered (barycenter method) to reduce crossings.  Lines are coloured
#' by the individual's cluster in the \code{reference} result, so one can
#' see how the reference clusters split or merge.  The adjusted Rand
#' index (ARI) between each pair of adjacent results is printed along the
#' top of the figure.  X-axis ticks default to each result's \code{$rank}
#' and can be overridden with \code{names}.
#'
#' Works for any non-negative multiplicative-update family
#' (\code{nmfkc}, \code{nmfae}, \code{nmfkc.net}, \code{nmfre},
#' and the signed variants); the hard label is the argmax of the
#' coefficient/score matrix.
#'
#' @param fits A list (length \eqn{\ge 2}) of fitted models, all over the
#'   \strong{same} \eqn{N} individuals.  The results are taken in the
#'   given order (\strong{not} sorted), so they may be different ranks or
#'   different models at the same rank.
#' @param reference The \strong{index} (1-based position in \code{fits})
#'   of the result whose clustering defines the line colours.  Defaults
#'   to the central result, \code{floor(length(fits) / 2) + 1} (e.g.\ the
#'   2nd of 2 or 3 results).
#' @param names Optional character vector (length \code{length(fits)}) of
#'   x-axis tick labels.  Defaults to each result's \code{$rank}.
#' @param plot Logical; draw the diagram immediately by calling
#'   \code{\link{plot.nmf.cluster.flow}} (default \code{TRUE}).  Set
#'   \code{FALSE} to only build the object and plot it later.
#' @param ... When \code{plot = TRUE}, graphical arguments forwarded to
#'   \code{\link{plot.nmf.cluster.flow}} (e.g.\ \code{col}, \code{lwd},
#'   \code{xlab}, \code{ylab}, \code{main}).
#' @return An object of class \code{"nmf.cluster.flow"} (returned
#'   invisibly): a list with \code{clusters} (the \eqn{N \times R} table:
#'   rows = individuals, columns = results, entries = cluster number = the
#'   dominant-factor index of each fit, so it matches the factor/basis
#'   numbering of \code{fits}; a factor that never dominates leaves an
#'   empty, unused cluster number),
#'   \code{ypos} (the layout positions), \code{ranks} (each result's
#'   rank), \code{labels} (the x-axis labels), \code{reference} (the
#'   reference index), \code{ref.cluster} (the reference hard labels),
#'   \code{ARI} (adjusted Rand index between each pair of adjacent
#'   results, length \eqn{R - 1}), and \code{colors}
#'   (the default per-individual reference colour).  Call
#'   \code{\link{plot}} on it to (re)draw the diagram.
#' @seealso \code{\link{plot.nmf.cluster.flow}},
#'   \code{\link{nmf.cluster.criteria}}, \code{\link{nmfkc.rank}}
#' @export
#' @examples
#' \donttest{
#' Y <- t(as.matrix(iris[, 1:4]))
#' fits <- lapply(2:6, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
#' fl <- nmf.cluster.flow(fits, reference = 2, plot = FALSE)  # 2nd result
#' head(fl$clusters)
#' plot(fl, lwd = 2, main = "iris cluster flow")
#' }
nmf.cluster.flow <- function(fits, reference = NULL, names = NULL,
                             plot = TRUE, ...) {
  if (!base::is.list(fits) || base::length(fits) < 2)
    base::stop("`fits` must be a list of at least two fitted models.", call. = FALSE)

  lab_list <- base::lapply(fits, .nmf.hard.labels)
  N <- base::length(lab_list[[1]])
  if (base::any(base::vapply(lab_list, base::length, 1L) != N))
    base::stop("all fits must share the same number of individuals (N).", call. = FALSE)
  R <- base::length(fits)

  ## Each result's rank (for the default labels); the given order is kept.
  ranks <- base::vapply(fits, function(f)
    if (!base::is.null(f$rank)) base::as.integer(f$rank) else NA_integer_, 1L)
  if (base::is.null(names)) labels <- base::as.character(ranks)
  else {
    labels <- base::as.character(names)
    if (base::length(labels) != R)
      base::stop("`names` must have length(fits) = ", R, " entries.", call. = FALSE)
  }

  ## The cluster number is the dominant-factor index (argmax of the
  ## coefficient), kept as-is so it matches the factor/basis numbering of
  ## the supplied fits; a factor that never dominates leaves a gap.
  clusters <- base::matrix(base::unlist(lab_list), nrow = N, ncol = R)
  ind_names <- base::names(lab_list[[1]])
  if (base::is.null(ind_names)) ind_names <- base::paste0("i", base::seq_len(N))
  base::rownames(clusters) <- ind_names
  base::colnames(clusters) <- labels

  ## reference = index (1..R) of the result that defines the colours;
  ## default to the central result floor(R/2)+1 (e.g. 2 for 2 or 3 results)
  if (base::is.null(reference)) reference <- base::floor(R / 2) + 1L
  ref_col <- base::as.integer(reference)
  if (base::length(ref_col) != 1L || base::is.na(ref_col) ||
      ref_col < 1L || ref_col > R)
    base::stop("`reference` must be a single result index in 1..", R, ".",
               call. = FALSE)
  ref_lab <- clusters[, ref_col]

  ## --- layout: positions per result, with inter-cluster gaps ---
  ypos <- base::matrix(NA_real_, N, R)
  ypos[, ref_col] <- .flow.place(ref_lab, ref_lab)
  if (ref_col < R) for (q in (ref_col + 1):R)
    ypos[, q] <- .flow.place(clusters[, q], ypos[, q - 1])
  if (ref_col > 1) for (q in (ref_col - 1):1)
    ypos[, q] <- .flow.place(clusters[, q], ypos[, q + 1])

  ## Adjacent-result agreement: ARI between results q and q+1.
  ari <- base::vapply(base::seq_len(R - 1),
                      function(q) .ari(clusters[, q], clusters[, q + 1]),
                      base::numeric(1))
  base::names(ari) <- base::sprintf("%s-%s", labels[-R], labels[-1L])

  out <- base::list(clusters = clusters, ypos = ypos, ranks = ranks,
                    labels = labels, reference = ref_col, ref.cluster = ref_lab,
                    ARI = ari,
                    colors = stats::setNames(.flow.colors(ref_lab)[ref_lab], ind_names))
  base::class(out) <- "nmf.cluster.flow"
  if (plot) graphics::plot(out, ...)
  base::invisible(out)
}


#' @title Reference-cluster colours for a cluster-flow diagram (Internal)
#' @description
#' Maps reference hard labels to colours.  With \code{col = NULL} a
#' strong, well-separated qualitative palette is used (ColorBrewer
#' \dQuote{Dark 2} for up to 8 clusters, HCL \dQuote{Dark 3} beyond, both
#' from the standard \code{grDevices} package); a user \code{col} vector
#' is indexed by cluster id (recycled if too short).
#' @param ref_lab Integer reference cluster labels (length \eqn{N}).
#' @param col Optional colour vector indexed by cluster id.
#' @return The resolved \strong{palette}: a character vector of length
#'   \eqn{K} (one colour per reference cluster).  Index it by cluster id,
#'   e.g.\ \code{pal[ref_lab]} for per-individual colours.
#' @keywords internal
#' @noRd
.flow.colors <- function(ref_lab, col = NULL) {
  K <- base::max(ref_lab, na.rm = TRUE)
  if (base::is.null(col)) {
    n <- base::max(K, 2L)
    col <- if (n <= 8L) grDevices::palette.colors(8L, "Dark 2")[base::seq_len(n)]
           else grDevices::hcl.colors(n, "Dark 3")
  } else if (base::length(col) < K) {
    col <- base::rep_len(col, K)
  }
  col[base::seq_len(K)]
}


#' @title Plot a cluster-flow (alluvial) diagram
#' @description
#' Draws the alluvial / Sankey-style cluster-flow diagram for an object
#' created by \code{\link{nmf.cluster.flow}}.
#' @param x An object of class \code{"nmf.cluster.flow"}.
#' @param col Optional colour vector indexed by \strong{reference}
#'   cluster id (\code{col[k]} colours every individual whose reference
#'   cluster is \code{k}); recycled if shorter than the number of
#'   reference clusters.  Defaults to the object's palette.
#' @param lwd Line width of the flow segments.
#' @param xlab,ylab Axis labels.
#' @param main Plot title.
#' @param ... Further arguments passed to the initial
#'   \code{\link[graphics]{plot}} call.
#' @return \code{x}, invisibly.
#' @seealso \code{\link{nmf.cluster.flow}}
#' @export
plot.nmf.cluster.flow <- function(x, col = NULL, lwd = 1,
                                  xlab = "rank (Q)", ylab = "individuals",
                                  main = "Cluster flow across rank", ...) {
  clusters <- x$clusters; ypos <- x$ypos
  labels <- if (!base::is.null(x$labels)) x$labels else x$ranks
  ref_lab <- x$ref.cluster
  N <- base::nrow(clusters); R <- base::ncol(clusters)
  pal <- .flow.colors(ref_lab, col)   # one colour per reference cluster
  ind_col <- pal[ref_lab]             # per-individual line colour

  ## Normalize each result (column) independently to [0, 1]: the layout
  ## inserts inter-cluster gaps, so the total span differs per result.
  yn <- base::apply(ypos, 2, function(p) {
    rg <- base::range(p)
    if (base::diff(rg) == 0) base::rep(0.5, base::length(p))
    else (p - rg[1]) / base::diff(rg)
  })
  xs <- base::seq_len(R)
  old <- graphics::par(mar = c(4, 4, 4, 2) + 0.1); base::on.exit(graphics::par(old))
  ## extra head-room at the top for the adjacent-result ARI labels
  graphics::plot(NA, xlim = c(1, R), ylim = c(-0.03, 1.12),
                 xaxt = "n", yaxt = "n", xlab = xlab, ylab = ylab, main = main, ...)
  graphics::axis(1, at = xs, labels = labels)

  ## Adjacent-rank ARI, printed between the two columns it compares.
  if (!base::is.null(x$ARI)) for (q in base::seq_len(R - 1)) {
    if (!base::is.finite(x$ARI[q])) next
    lab <- if (q == 1) base::sprintf("ARI=%.2f", x$ARI[q])
           else base::sprintf("%.2f", x$ARI[q])
    graphics::text((xs[q] + xs[q + 1]) / 2, 1.08, lab,
                   cex = 0.85, font = 2, col = "gray25")
  }

  hw  <- 0.10        # half-width of the cluster box (x units)

  ## 1. flow lines + points (drawn first, so the boxes sit in front)
  for (q in 1:(R - 1))
    graphics::segments(xs[q], yn[, q], xs[q + 1], yn[, q + 1],
                       col = ind_col, lwd = lwd)
  for (q in 1:R)
    graphics::points(base::rep(xs[q], N), yn[, q], pch = 16,
                     col = ind_col, cex = 0.5)

  ## 2. foreground box per cluster, sized to the min/max of its members
  ##    (tiny pad so a singleton still shows).  The box is tinted by the
  ##    MAJORITY reference colour among its members; ties are broken in
  ##    favour of the earliest palette entry (smallest reference id).
  pad <- 0.004
  for (q in 1:R) for (c in base::unique(clusters[, q])) {
    sel <- clusters[, q] == c
    ys  <- yn[sel, q]
    tab <- base::table(ref_lab[sel])
    cand <- base::as.integer(base::names(tab)[tab == base::max(tab)])
    majref <- base::min(cand)                 # earliest palette colour on ties
    box_fill <- grDevices::adjustcolor(pal[majref], alpha.f = 0.45)
    graphics::rect(xs[q] - hw, base::min(ys) - pad,
                   xs[q] + hw, base::max(ys) + pad,
                   col = box_fill, border = pal[majref], lwd = 1)
    graphics::text(xs[q], base::mean(ys), c, cex = 1.2, font = 2, col = "gray10")
  }
  base::invisible(x)
}


#' @title Print method for nmf.cluster.flow objects
#' @description
#' Prints a one-line header, the adjacent-result ARI, and the
#' \eqn{N \times R} cluster table (rows = individuals, columns =
#' results, entries = cluster number).  Use \code{\link{plot}} for the
#' diagram.
#' @param x An object of class \code{"nmf.cluster.flow"}.
#' @param ... Passed to the table's \code{print}.
#' @return \code{x}, invisibly.
#' @export
print.nmf.cluster.flow <- function(x, ...) {
  labels <- if (!base::is.null(x$labels)) x$labels else base::as.character(x$ranks)
  base::cat(base::sprintf("Cluster flow: %d individuals across %d results [%s] (reference = result %d: %s)\n",
                          base::nrow(x$clusters), base::length(labels),
                          base::paste(labels, collapse = ", "),
                          x$reference, labels[x$reference]))
  if (!base::is.null(x$ARI)) {
    base::cat("\nAdjacent-result ARI:\n")
    base::print(base::round(x$ARI, 3))
  }
  base::cat("\nClusters (rows = individuals, columns = rank, entries = cluster number):\n")
  base::print(x$clusters, ...)
  base::invisible(x)
}


#' @title Best-rank selection and concise diagnostics plot (Internal)
#' @description
#' Shared back-end for the rank-selection functions (\code{nmfkc.rank},
#' \code{nmfkc.net.rank}, \code{nmfkc.signed.rank}, \code{nmfae.rank},
#' \code{nmfae.signed.rank}).  From a pre-built \code{criteria} data
#' frame it determines the recommended rank and draws a concise
#' three-criterion plot, using only \code{r.squared}, the effective-rank
#' utilization \code{effective.rank.ratio}, and the cross-validation
#' error \code{sigma.ecv}.  Any other columns present (e.g.\ \code{ARI},
#' \code{silhouette}, \code{CPCC}, \code{dist.cor} from \code{nmfkc.rank})
#' are kept in the returned table but \strong{not plotted}.  All three
#' criteria are drawn as lines with points and rank-number labels.  Two
#' carry a highlighted "Best" marker that drives the recommended rank:
#' \itemize{
#'   \item \code{r.squared}: elbow (kneedle) -- \dQuote{Best (Elbow)}.
#'   \item \code{sigma.ecv}: minimum CV error -- \dQuote{Best (Min)}.
#' }
#' The green line is the \code{effective.rank.index} (not the raw
#' \code{eff.rank / Q}): \code{(eff.rank - E) / (Q - E)} clamped to
#' \eqn{[0, 1]}, where \code{E = exp(H_Q - 1)} is the broken-stick
#' expected effective rank (\code{H_Q} = the \code{Q}-th harmonic
#' number).  This anchors 0 at the random null and 1 at perfect
#' evenness, removing the small-rank inflation of \code{eff.rank / Q}.
#' It is shown for context only and carries \strong{no} best marker: it
#' is a factor-utilization diagnostic (most even relative to the null),
#' not a predictive rank optimum -- the recommended rank is driven by
#' the cross-validation minimum and the R-squared elbow.
#' @param criteria Data frame with at least \code{rank}, \code{r.squared},
#'   \code{effective.rank.ratio}, and (optionally) \code{sigma.ecv}.
#' @param plot Logical; draw the diagnostics plot.
#' @param main Plot title.
#' @return A list with \code{rank.best} (ECV minimum if available, else
#'   the R-squared elbow) and \code{criteria} (the input data frame).
#' @keywords internal
#' @noRd
.rank.finish <- function(criteria, plot = TRUE,
                         main = "Rank Selection Diagnostics") {
  base::message("Note: sample-clustering quality (silhouette / CPCC / ",
                "dist.cor) is not part of rank selection; compute it from a ",
                "list of fits with nmf.cluster.criteria().  See ?nmf.cluster.criteria")
  rk <- criteria$rank
  nq <- base::length(rk)

  has_ecv <- !base::is.null(criteria$sigma.ecv) &&
             base::any(base::is.finite(criteria$sigma.ecv))
  rank.best.ecv <- if (has_ecv)
    rk[base::which.min(criteria$sigma.ecv)] else NA

  rank.best.r2 <- NA
  if (nq > 2 && !base::is.null(criteria$r.squared)) {
    x <- base::seq_len(nq); y <- criteria$r.squared
    yr <- base::max(y) - base::min(y); xr <- base::max(x) - base::min(x)
    yn <- if (yr > 0) (y - base::min(y)) / yr else base::rep(0.5, nq)
    xn <- if (xr > 0) (x - base::min(x)) / xr else base::rep(0.5, nq)
    x1 <- xn[1]; y1 <- yn[1]; x2 <- xn[nq]; y2 <- yn[nq]
    den <- base::sqrt((y2 - y1)^2 + (x2 - x1)^2)
    d <- if (den > 0)
      base::abs((y2 - y1) * xn - (x2 - x1) * yn + x2 * y1 - y2 * x1) / den
      else base::numeric(nq)
    rank.best.r2 <- rk[base::which.max(d)]
  }

  ## Broken-stick correction of the effective-rank utilization.
  ## Expected effective rank under a random (uniform-Dirichlet) split of
  ## the variance among Q factors is exp(H_Q - 1), H_Q = sum_{j<=Q} 1/j.
  ## The [0,1] index anchors 0 at this null and 1 at perfect evenness
  ## (eff.rank = Q): (eff.rank - expected) / (Q - expected), clamped.
  ## This removes the small-Q inflation of eff.rank / Q, so its maximum
  ## is a meaningful rank.
  if (!base::is.null(criteria$effective.rank)) {
    Hq <- base::vapply(rk, function(Q) base::sum(1 / base::seq_len(Q)),
                       base::numeric(1))
    e_null <- base::exp(Hq - 1)
    idx <- (criteria$effective.rank - e_null) / (rk - e_null)
    idx[!base::is.finite(idx)] <- NA_real_
    criteria$effective.rank.expected <- e_null
    criteria$effective.rank.index <- base::pmin(base::pmax(idx, 0), 1)
  }

  has_idx <- !base::is.null(criteria$effective.rank.index) &&
             base::any(base::is.finite(criteria$effective.rank.index))
  rank.best.idx <- if (has_idx)
    rk[base::which.max(criteria$effective.rank.index)] else NA

  rank.final <- if (!base::is.na(rank.best.ecv)) rank.best.ecv else rank.best.r2

  out <- base::list(rank.best = rank.final, criteria = criteria,
                    best = base::list(ecv = rank.best.ecv, r2 = rank.best.r2,
                                      eff.idx = rank.best.idx),
                    main = main)
  base::class(out) <- "nmf.rank"
  if (plot) graphics::plot(out)
  base::invisible(out)
}


#' @title Plot a rank-selection (nmf.rank) object
#' @description
#' Draws the concise three-criterion rank-selection plot for an object
#' returned by \code{\link{nmfkc.rank}} (or \code{nmfkc.net.rank},
#' \code{nmfkc.signed.rank}, \code{nmfae.rank}, \code{nmfae.signed.rank}):
#' \code{r.squared} (red) and the effective-rank utilization
#' \code{eff.rank} (green) on the left \eqn{[0, 1]} axis, and the
#' cross-validation error \code{sigma.ecv} (blue) on the right axis, each
#' with points, rank-number labels and a highlighted best marker.
#' @param x An object of class \code{"nmf.rank"}.
#' @param main Plot title (defaults to the title stored in \code{x}).
#' @param xlab,ylab Axis labels.
#' @param lwd Line width of the criterion curves.
#' @param ... Further arguments passed to the initial
#'   \code{\link[graphics]{plot}}.
#' @return \code{x}, invisibly.
#' @seealso \code{\link{nmfkc.rank}}
#' @export
plot.nmf.rank <- function(x, main = NULL, xlab = "Rank (Q)",
                          ylab = "R-squared / eff.rank (0-1)", lwd = 3, ...) {
  if (base::is.null(main)) main <- x$main
  criteria <- x$criteria
  rk <- criteria$rank
  rank.best.ecv <- x$best$ecv; rank.best.r2 <- x$best$r2
  rank.best.idx <- x$best$eff.idx
  has_ecv <- !base::is.null(criteria$sigma.ecv) &&
             base::any(base::is.finite(criteria$sigma.ecv))
  has_idx <- !base::is.null(criteria$effective.rank.index) &&
             base::any(base::is.finite(criteria$effective.rank.index))

  old_par <- graphics::par(mar = c(5, 4, 4, 5) + 0.1)
  base::on.exit(graphics::par(old_par))
  decorate <- function(yv, col, best_rank, best_lab, numpos) {
    graphics::points(rk, yv, pch = 16, col = col, cex = 0.8)
    graphics::text(rk, yv, rk, pos = numpos, col = col, cex = 0.8)
    if (!base::is.na(best_rank)) {
      j <- base::which(rk == best_rank)
      graphics::points(best_rank, yv[j], pch = 16, col = col, cex = 1.6)
      graphics::text(best_rank, yv[j], best_lab, pos = 4, col = col, cex = 0.8)
    }
  }
  leg_txt <- "r.squared"; leg_col <- 2

  graphics::plot(rk, criteria$r.squared, type = "l", col = 2, lwd = lwd,
                 xlab = xlab, ylab = ylab, ylim = c(0, 1), main = main, ...)
  for (q in rk) graphics::abline(v = q, col = "gray90", lwd = 0.5)
  decorate(criteria$r.squared, 2, rank.best.r2, "Best (Elbow)", 3)

  if (has_idx) {
    ## Broken-stick-corrected effective-rank index in [0, 1], shown for
    ## context only: line + points + numbers, NO best marker (it is a
    ## utilization diagnostic, not a predictive rank optimum).
    graphics::lines(rk, criteria$effective.rank.index, col = "forestgreen", lwd = lwd)
    decorate(criteria$effective.rank.index, "forestgreen", NA, "", 1)
    leg_txt <- c(leg_txt, "eff.rank"); leg_col <- c(leg_col, "forestgreen")
  }

  if (has_ecv) {
    graphics::par(new = TRUE)
    graphics::plot(rk, criteria$sigma.ecv, type = "l", col = "blue", lwd = lwd,
                   axes = FALSE, xlab = "", ylab = "")
    decorate(criteria$sigma.ecv, "blue", rank.best.ecv, "Best (Min)", 3)
    graphics::axis(side = 4, col = "blue", col.axis = "blue")
    graphics::mtext("ECV Sigma (RMSE)", side = 4, line = 3, col = "blue")
    leg_txt <- c(leg_txt, "sigma.ecv"); leg_col <- c(leg_col, "blue")
  }

  graphics::legend("right", legend = leg_txt, col = leg_col,
                   lty = 1, lwd = 2, bg = "white", cex = 0.7)
  base::invisible(x)
}


#' @title Print method for rank-selection (nmf.rank) objects
#' @param x An object of class \code{"nmf.rank"}.
#' @param ... Passed to the criteria table's \code{print}.
#' @return \code{x}, invisibly.
#' @export
print.nmf.rank <- function(x, ...) {
  fmt <- function(v) if (base::is.null(v) || base::is.na(v)) "NA" else base::as.character(v)
  base::cat(base::sprintf("Rank selection over ranks %s\n",
                          base::paste(x$criteria$rank, collapse = ", ")))
  base::cat(base::sprintf("  recommended rank (rank.best): %s\n", fmt(x$rank.best)))
  base::cat(base::sprintf("  best by:  ECV min = %s | R-squared elbow = %s\n",
                          fmt(x$best$ecv), fmt(x$best$r2)))
  base::cat("\nCriteria:\n")
  base::print(x$criteria, row.names = FALSE, ...)
  base::invisible(x)
}


#' @title Parse formula and prepare Y and A matrices
#' @description
#' Internal function to handle formula input, parse variables, and generate
#' the transposed Y and A matrices (features/covariates in rows, samples in columns)
#' as required by the core \code{nmfkc} function. Supports 'data' mode
#' (variable names and dot notation) and direct expression evaluation mode.
#' @param formula A formula object.
#' @param data Optional data frame or environment.
#' @return A list containing the prepared matrices: \code{Y} (P x N) and \code{A} (R x N or NULL).
#' @keywords internal
#' @noRd
.nmfkc_parse_formula <- function(formula, data) {

  # Helper function to abbreviate column messages (shows max_show columns)
  .abbreviate_msg <- function(cols, prefix, max_show = 5) {
    total_len <- base::length(cols)
    if (total_len > max_show) {
      msg <- base::paste0(base::paste(cols[1:max_show], collapse = ", "),
                          base::sprintf("... (Total: %d)", total_len))
    } else {
      msg <- base::paste(cols, collapse = ", ")
    }
    base::message(base::paste(prefix, "created using columns:", msg))
  }

  # Capture the formula object and set the environment
  f <- formula
  f_env <- if (base::missing(data)) base::environment(f) else base::environment()

  # Extract the expressions for Y and A
  Y_expr <- f[[2]]
  A_expr <- f[[3]]

  # --- Input Mode Branching ---

  if (!base::missing(data)) {
    # MODE 1: data provided (Variable names and dot notation)
    data <- base::as.data.frame(data)
    all_cols <- base::names(data)

    # Check if A_expr is structurally present in the formula object.
    A_is_structurally_present <- base::length(f) >= 3

    # Extract Y and A expressions as clean strings using deparse
    Y_expr_str <- base::paste(base::deparse(Y_expr), collapse = " ")

    A_expr_str <- if (A_is_structurally_present) {
      base::paste(base::deparse(A_expr), collapse = " ")
    } else {
      NA # Use simple NA for missing structural part
    }

    # Determine dot notation and missing status based on the safe string A_expr_str
    Y_is_dot <- (Y_expr_str == ".")
    A_is_missing <- base::is.na(A_expr_str)
    A_is_dot <- (!A_is_missing && A_expr_str == ".")

    # --- Check for explicit A omission symbols (0 or -1) ---
    A_is_explicitly_omitted <- (!A_is_missing &&
                                  (A_expr_str == "0" || A_expr_str == "-1" ||
                                     A_expr_str == "0 + ." || A_expr_str == "-1 + ." ||
                                     A_expr_str == ". + 0" || A_expr_str == ". + -1"))

    # Update A_is_missing status if explicitly omitted
    A_is_missing <- A_is_missing || A_is_explicitly_omitted

    Y_cols <- NULL
    A_cols <- NULL

    # Helper to clean and split variable names from expression string
    .extract_cols <- function(expr_str) {
      # Split by '+', '~', ' ' and remove empty elements. This handles "A1 + A2" structure.
      cols <- base::unlist(base::strsplit(expr_str, "[ +~]", fixed = FALSE))
      cols <- cols[cols != ""]
      # Remove explicit omission symbols if they were part of the split result
      cols <- cols[!(cols %in% base::c("0", "-1"))]
      return(cols)
    }

    if (!Y_is_dot) { Y_cols <- .extract_cols(Y_expr_str) }

    # A_cols is only extracted if A is NOT missing and NOT explicitly omitted (and not dot)
    if (!A_is_missing && !A_is_dot) {
      A_cols <- .extract_cols(A_expr_str)
    }

    # Error Check: Both sides cannot be '.'
    if (Y_is_dot && A_is_dot) {
      base::stop("Formula error: '.' cannot be used on both the left (Y) and right (A) sides simultaneously when 'data' is provided.")
    }

    used_cols <- base::unique(base::c(Y_cols, A_cols))
    remaining_cols <- base::setdiff(all_cols, used_cols)

    # Assign dot notation variables
    if (Y_is_dot) {
      Y_cols <- remaining_cols
      if (base::length(Y_cols) == 0) { base::stop("Formula error: '.' for Y resulted in no remaining variables.") }
      .abbreviate_msg(Y_cols, "Y")
    } else if (A_is_dot) {
      A_cols <- remaining_cols
      if (base::length(A_cols) == 0) { base::stop("Formula error: '.' for A resulted in no remaining variables.") }
      # NOTE: Message output for A is suppressed here to avoid duplication
      #       and is handled in the A_mat creation block below.
    }

    # Validate column existence
    missing_Y <- base::setdiff(Y_cols, all_cols)
    if (base::length(missing_Y) > 0) {
      base::stop(base::paste0("Formula error: Y columns not found in data: ",
                               base::paste(missing_Y, collapse = ", ")))
    }
    if (!A_is_missing) {
      missing_A <- base::setdiff(A_cols, all_cols)
      if (base::length(missing_A) > 0) {
        base::stop(base::paste0("Formula error: A columns not found in data: ",
                                 base::paste(missing_A, collapse = ", ")))
      }
    }

    # Matrix Creation
    Y_mat <- data[, Y_cols, drop = FALSE]

    # --- A Matrix Finalization and Message Output ---
    if (A_is_missing) { # Catches structural omission, NA, or explicit 0/-1
      A_mat <- NULL
      base::message("A (covariate matrix) is omitted. Performing standard NMF (Y ~ X B).")
    } else {
      # This block now runs if A is explicitly defined with variables or is dot notation
      A_mat <- data[, A_cols, drop = FALSE]
      .abbreviate_msg(A_cols, "A") # Message output for A is performed here ONCE
    }

  } else {
    # MODE 2: data omitted (Direct matrix expression evaluation)
    Y_cols <- NULL
    A_cols <- NULL

    if (base::is.symbol(Y_expr) && base::as.character(Y_expr) == ".") {
      base::stop("Formula error: '.' is not supported for direct matrix evaluation mode (without 'data' argument).")
    }

    Y_mat <- base::tryCatch({ base::as.matrix(base::eval(Y_expr, envir = f_env)) },
                            error = function(e) { base::stop(base::paste("Error evaluating Y expression:", base::conditionMessage(e))) })

    A_mat <- base::tryCatch({
      # Check if A_expr exists (length >= 3) or is explicitly 0/-1
      A_is_structurally_present <- base::length(f) >= 3
      if (!A_is_structurally_present || (base::is.numeric(A_expr) && A_expr == 0) || (base::is.numeric(A_expr) && A_expr == -1)) { NULL
      } else { base::as.matrix(base::eval(A_expr, envir = f_env)) }
    }, error = function(e) { base::stop(base::paste("Error evaluating A expression:", base::conditionMessage(e))) })

    if (base::is.null(A_mat)) {
      base::message("A (covariate matrix) is omitted. Performing standard NMF (Y ~ X B).")
    }
    # NOTE: No column abbreviation message needed for Mode 2 as matrices are evaluated directly.
  }

  # Transpose: R rows=samples -> nmfkc columns=samples (P x N or R x N)
  # NOTE: R data frames/matrices are typically samples x variables. nmfkc requires variables x samples.
  Y <- base::t(Y_mat)

  if (!base::is.null(A_mat)) {
    A <- base::t(A_mat)
    if (base::ncol(Y) != base::ncol(A)) {
      base::stop(base::paste0("Dimension error: Number of columns (samples) in Y (", base::ncol(Y), ") must match number of columns (samples) in A (", base::ncol(A), ")."))
    }
  } else {
    A <- NULL
  }

  return(base::list(Y = Y, A = A, Y_cols = Y_cols, A_cols = A_cols))
}


#' Resolve formula input to Y/A matrices with metadata
#'
#' Internal helper that detects formula input, parses it, and returns
#' the Y/A matrices along with formula metadata for downstream use.
#' The \code{data_missing} flag solves the \code{missing(data)} propagation
#' problem in R (cannot check \code{missing()} through nested calls).
#'
#' @param Y A formula or matrix.
#' @param A A matrix or NULL.
#' @param data_missing Logical; TRUE if \code{data} was not supplied by the caller.
#' @param data_value The value of \code{data} (or NULL if missing).
#' @return A list with \code{Y}, \code{A}, and \code{formula.meta} (NULL if not formula mode).
#' @keywords internal
#' @noRd
.nmfkc_resolve_formula <- function(Y, A, data_missing, data_value) {
  if (!base::inherits(Y, "formula")) {
    return(base::list(Y = Y, A = A, formula.meta = NULL))
  }
  if (data_missing) {
    data_list <- .nmfkc_parse_formula(formula = Y)
  } else {
    data_list <- .nmfkc_parse_formula(formula = Y, data = data_value)
  }
  formula.meta <- base::list(
    formula = Y,
    Y_cols  = data_list$Y_cols,
    A_cols  = data_list$A_cols
  )
  base::list(Y = data_list$Y, A = data_list$A, formula.meta = formula.meta)
}










#' @title Optimize NMF with kernel covariates (Full Support for Missing Values)
#' @description
#' \code{nmfkc} fits a nonnegative matrix factorization with kernel covariates
#' under the tri-factorization model \eqn{Y \approx X C A = X B}.
#'
#' This function supports two major input modes:
#' 1. **Matrix Mode (Existing)**: \code{nmfkc(Y=matrix, A=matrix, ...)}
#' 2. **Formula Mode (New)**: \code{nmfkc(formula=Y_vars ~ A_vars, data=df, rank=Q, ...)}
#'
#' The rank of the basis matrix can be specified using either the \code{rank} argument
#' (preferred for formula mode) or the hidden \code{Q} argument (for backward compatibility).
#'
#' @param Y Observation matrix (P x N), OR a formula object for Formula Mode.
#'   In Formula Mode, use \code{Y1 + Y2 ~ A1 + A2} with \code{data}, or
#'   \code{Y_matrix ~ A_matrix} for direct matrix evaluation.
#'   Supports dot notation (\code{. ~ A1 + A2}) when \code{data} is supplied.
#' @param A Covariate matrix. Default is \code{NULL} (no covariates).
#'   Ignored when \code{Y} is a formula.
#' @param rank Integer. The rank of the basis matrix \eqn{X} (Q). Preferred over \code{Q}.
#' @param data Optional. A data frame from which variables in the formula should be taken.
#' @param epsilon Positive convergence tolerance.
#' @param maxit Maximum number of iterations.
#' @param verbose Logical. If \code{TRUE} (default), prints matrix dimensions and elapsed time.
#' @param ... Additional arguments passed for fine-tuning regularization, initialization, constraints,
#'   and output control. This includes the backward-compatible arguments \code{Q} and \code{method}.
#'   \itemize{
#'     \item \code{Y.weights}: Optional weight matrix (P x N) or vector
#'       (length N) with non-negative entries, analogous to the \code{weights}
#'       argument of \code{\link[stats]{lm}}.  When supplied, the objective
#'       becomes \eqn{\sum W_{ij} \, (Y_{ij} - (XB)_{ij})^2}
#'       (i.e.\ \strong{linear} in \eqn{W}; \code{lm()}-style weighted
#'       least squares).  Logical matrices (\code{TRUE} / \code{FALSE})
#'       are also accepted and coerced to 1 / 0.  The primary use case is
#'       missing-value masking for ECV / CV, where \eqn{W_{ij} \in \{0, 1\}}
#'       (\code{FALSE} / \code{TRUE}) indicates held-out vs.\ used elements;
#'       real-valued weights for observation-level importance weighting are
#'       also supported.  Default \code{NULL}: if \code{Y} contains \code{NA}
#'       a binary mask is auto-constructed (0 for \code{NA}, 1 elsewhere);
#'       otherwise no weighting.
#'     \item \code{X.L2.ortho}: Nonnegative penalty parameter for the orthogonality of \eqn{X} (default: 0).
#'       It minimizes the off-diagonal elements of the Gram matrix \eqn{X^\top X}, reducing the correlation
#'       between basis vectors (conceptually minimizing \eqn{\| X^\top X - \mathrm{diag}(X^\top X) \|_F^2}).
#'       (Formerly \code{lambda.ortho}).
#'     \item \code{B.L1}: Nonnegative penalty parameter for L1 regularization on \eqn{B = C A} (default: 0).
#'       Promotes **sparsity in the coefficients**. (Formerly \code{gamma}).
#'     \item \code{C.L1}: Nonnegative penalty parameter for L1 regularization on \eqn{C} (default: 0).
#'       Promotes **sparsity in the parameter matrix**. (Formerly \code{lambda}).
#'     \item \code{Q}: Backward-compatible name for the rank of the basis matrix (Q).
#'     \item \code{method}: Objective function: Euclidean distance \code{"EU"} (default) or Kullback–Leibler divergence \code{"KL"}.
#'     \item \code{X.restriction}: Constraint for columns of \eqn{X}. Options: \code{"colSums"} (default), \code{"colSqSums"}, \code{"totalSum"}, \code{"none"}, or \code{"fixed"}.
#'       \code{"none"} applies no normalization to \eqn{X} after each update, allowing it to absorb the scale freely.
#'     \item \code{X.init}: Method for initializing the basis matrix \eqn{X}. Options: \code{"kmeans"} (default), \code{"kmeansar"}, \code{"runif"}, \code{"nndsvd"}, or a user-specified matrix. \code{"kmeansar"} applies \eqn{k}-means initialization and then fills zero entries with \code{Uniform(0, mean(Y)/100)}, analogous to NNDSVDar.
#'     \item \code{nstart}: Number of random starts for initialization of \eqn{X} (default: 1).
#'       Used by \code{kmeans} (when \code{X.init = "kmeans"} or \code{"kmeansar"}) and by the
#'       multi-start evaluation (when \code{X.init = "runif"}).
#'     \item \code{seed}: Integer seed for reproducibility (default: 123).
#'     \item \code{C.init}: Optional numeric matrix giving the initial value of the parameter matrix \eqn{C}
#'       (i.e., \eqn{\Theta}). If \code{A} is \code{NULL}, \code{C} has dimension \eqn{Q \times N} (equivalently \eqn{B});
#'       otherwise, \code{C} has dimension \eqn{Q \times K} where \eqn{K = nrow(A)}. Default initializes all entries to 1.
#'     \item \code{Y.symmetric}: \strong{Removed.} Symmetric NMF
#'       (\eqn{Y \approx X X^\top} or \eqn{X C X^\top}) has moved to the
#'       dedicated \code{\link{nmfkc.net}} function (types \code{"tri"},
#'       \code{"bi"}, \code{"signed"}), which uses the correct
#'       Frobenius bilateral-gradient updates.  Passing \code{Y.symmetric}
#'       to \code{nmfkc()} now stops with a message pointing to
#'       \code{nmfkc.net()}.
#'     \item \code{prefix}: Prefix for column names of \eqn{X} and row names of \eqn{B} (default: "Basis").
#'     \item \code{print.trace}: Logical. If \code{TRUE}, prints progress every 10 iterations (default: \code{FALSE}).
#'     \item \code{print.dims}: Deprecated. Use \code{verbose} instead.
#'     \item \code{detail}: Level of post-fit criterion computation.
#'       \code{"full"} computes all criteria including silhouette, CPCC, dist.cor;
#'       \code{"fast"} skips expensive distance-based criteria;
#'       \code{"minimal"} returns only information criteria.
#'       Default is \code{"full"}. For backward compatibility,
#'       \code{save.time = TRUE} maps to \code{"fast"} and
#'       \code{save.memory = TRUE} maps to \code{"minimal"}.
#'   }
#' @return A list with components:
#' \item{call}{The matched call, as captured by `match.call()`.}
#' \item{dims}{A character string summarizing the matrix dimensions of the model.}
#' \item{runtime}{A character string summarizing the computation time.}
#' \item{X}{Basis matrix. Column normalization depends on \code{X.restriction}.}
#' \item{B}{Coefficient matrix \eqn{B = C A}.}
#' \item{XB}{Fitted values for \eqn{Y}.}
#' \item{C}{Parameter matrix.}
#' \item{B.prob}{Soft-clustering probabilities derived from columns of \eqn{B}.}
#' \item{B.cluster}{Hard-clustering labels (argmax over \eqn{B.prob} for each column).}
#' \item{X.prob}{Row-wise soft-clustering probabilities derived from \eqn{X}.}
#' \item{X.cluster}{Hard-clustering labels (argmax over \eqn{X.prob} for each row).}
#' \item{A.attr}{List of attributes of the input covariate matrix \code{A}, containing metadata like lag order and intercept status if created by \code{nmfkc.ar} or \code{nmfkc.kernel}.}
#' \item{formula.meta}{If fitted via Formula Mode, a list with \code{formula}, \code{Y_cols}, and \code{A_cols}; otherwise \code{NULL}.}
#' \item{objfunc}{Final objective value.}
#' \item{objfunc.iter}{Objective values by iteration.}
#' \item{r.squared}{\eqn{R^2 = \mathrm{cor}(Y, XB)^2} (Pearson; scale-invariant; \eqn{[0,1]}).}
#' \item{r.squared.uncentered}{Uncentered \eqn{R^2 = 1 - \|Y - XB\|_F^2 / \|Y\|_F^2} (baseline = zero matrix; natural for non-negative factorizations without an intercept).}
#' \item{r.squared.centered}{Row-mean centered \eqn{R^2 = 1 - \|Y - XB\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2}, the multivariate regression \eqn{R^2}.}
#' \item{method}{Character string indicating the optimization method used (\code{"EU"} or \code{"KL"}).}
#' \item{n.missing}{Number of missing (or zero-weighted) elements in \eqn{Y}.}
#' \item{n.total}{Total number of elements in \eqn{Y}.}
#' \item{rank}{The rank \eqn{Q} used in the factorization.}
#' \item{sigma}{The residual standard error, representing the typical deviation of the observed values \eqn{Y} from the fitted values \eqn{X B}.}
#' \item{mae}{Mean Absolute Error between \eqn{Y} and \eqn{X B}.}
#' \item{criterion}{A list of selection criteria: \code{silhouette} (mean silhouette width of the hard clustering, computed in the original data space \code{dist(t(Y))} with the per-sample labels), \code{CPCC} (cophenetic correlation of a hierarchical clustering of the coefficient distances \code{dist(t(B))}), \code{dist.cor} (correlation between original-data and coefficient distances), \code{B.prob.max.mean} (clustering crispness: mean dominant-cluster membership, in \eqn{[1/Q, 1]}; meaningful at fixed \eqn{Q} as a confidence check before using \code{B.cluster} as hard labels), and \code{effective.rank}.  The last is the \strong{effective rank}: \eqn{\exp} of the Shannon entropy of the explained-variance distribution \eqn{p_k = \mathrm{var}(B_{k\cdot}) / \sum_j \mathrm{var}(B_{j\cdot})}.  By the trace identity \eqn{\sum_k \mathrm{var}(B_{k\cdot}) = \mathrm{tr}(\mathrm{Cov}(B))}, \eqn{p_k} is the exact fraction of the total coefficient variance carried by factor \eqn{k}, so the entropy measures how that variance is spread across factors.  It ranges in \eqn{[1, Q]} (1 when one factor carries all the variance, \eqn{Q} when all contribute equally) and counts the number of latent factors that actively shape across-sample variation.  This is the PCA-style explained-variance / effective-dimensionality measure and reuses the \eqn{\exp(\mathrm{entropy})} functional form of Roy & Vetterli (2007).}
#' @seealso \code{\link{nmfkc.cv}}, \code{\link{nmfkc.rank}}, \code{\link{nmfkc.kernel}}, \code{\link{nmfkc.ar}}, \code{\link{predict.nmfkc}}
#' @export
#' @references
#' Satoh, K. (2024). Applying Non-negative Matrix Factorization with Covariates
#'   to the Longitudinal Data as Growth Curve Model. arXiv:2403.05359.
#'   \url{https://arxiv.org/abs/2403.05359}
#'
#' Satoh, K. (2025). Applying non-negative matrix factorization with covariates
#'   to multivariate time series data as a vector autoregression model.
#'   \emph{Japanese Journal of Statistics and Data Science}. arXiv:2501.17446.
#'   \doi{10.1007/s42081-025-00314-0}
#'
#' Satoh, K. (2025). Applying non-negative matrix factorization with covariates
#'   to label matrix for classification. arXiv:2510.10375.
#'   \url{https://arxiv.org/abs/2510.10375}
#'
#' Ding, C., Li, T., Peng, W., & Park, H. (2006). Orthogonal Nonnegative Matrix
#'   Tri-Factorizations for Clustering.
#'   In \emph{Proc. 12th ACM SIGKDD} (pp. 126--135).
#'   \doi{10.1145/1150402.1150420}
#'
#' Roy, O., & Vetterli, M. (2007). The effective rank: A measure of
#'   effective dimensionality. In \emph{15th European Signal Processing
#'   Conference (EUSIPCO)} (pp. 606--610).
#' @examples
#' # Example 1. Matrix Mode (Existing)
#' X <- cbind(c(1,0,1),c(0,1,0))
#' B <- cbind(c(1,0),c(0,1),c(1,1))
#' Y <- X %*% B
#' rownames(Y) <- paste0("P",1:nrow(Y))
#' colnames(Y) <- paste0("N",1:ncol(Y))
#' print(X); print(B); print(Y)
#' res <- nmfkc(Y,rank=2,epsilon=1e-6)
#' res$X
#' res$B
#'
#' # Example 2. Formula Mode
#' set.seed(1)
#' dummy_data <- data.frame(Y1=rpois(10,5), Y2=rpois(10,10),
#'                          A1=abs(rnorm(10,5)), A2=abs(rnorm(10,3)))
#' res_f <- nmfkc(Y1 + Y2 ~ A1 + A2, data=dummy_data, rank=2)
#'
#' # For symmetric NMF (Y approximated by X X^T or X C X^T),
#' # use \code{\link{nmfkc.net}()} instead.
#'
nmfkc <- function(Y, A=NULL, rank=NULL, data, epsilon=1e-4, maxit=5000, verbose=TRUE, ...){
  # A small constant for numerical stability to prevent division by zero and log(0).
  .eps <- 1e-10

  extra_args <- base::list(...)

  # --- 1. Parameter Extraction ---
  Q_hidden <- if (!base::is.null(extra_args$Q)) extra_args$Q else NULL
  Q_val <- if (!base::is.null(rank)) rank else if (!base::is.null(Q_hidden)) Q_hidden else 2
  Q <- Q_val

  C.L1 <- if (!base::is.null(extra_args$C.L1)) extra_args$C.L1 else 0
  B.L1 <- if (!base::is.null(extra_args$B.L1)) extra_args$B.L1 else 0
  X.L2.ortho <- if (!base::is.null(extra_args$X.L2.ortho)) extra_args$X.L2.ortho else 0

  if (C.L1 == 0 && !base::is.null(extra_args$lambda)) C.L1 <- extra_args$lambda
  if (B.L1 == 0 && !base::is.null(extra_args$gamma)) B.L1 <- extra_args$gamma
  if (X.L2.ortho == 0 && !base::is.null(extra_args$lambda.ortho)) X.L2.ortho <- extra_args$lambda.ortho

  method <- if (!base::is.null(extra_args$method)) extra_args$method else "EU"
  X.restriction <- if (!base::is.null(extra_args$X.restriction)) extra_args$X.restriction else "colSums"
  X.init <- if (!base::is.null(extra_args$X.init)) extra_args$X.init else "kmeans"
  nstart <- if (!base::is.null(extra_args$nstart)) extra_args$nstart else 1
  seed <- if (!base::is.null(extra_args$seed)) extra_args$seed else 123
  C.init <- if (!is.null(extra_args$C.init)) extra_args$C.init else NULL
  ## Symmetric NMF (Y ~ X X^T or X C X^T) has moved out of nmfkc() into
  ## the dedicated nmfkc.net() function (Frobenius bilateral-gradient
  ## updates; types "tri", "bi", "signed").  The old Y.symmetric option
  ## was deprecated in v0.7.x and is removed here: passing it now stops
  ## with a redirect message rather than silently running plain NMF.
  if (!base::is.null(extra_args$Y.symmetric) &&
      !base::identical(base::as.character(extra_args$Y.symmetric), "none")) {
    ys <- base::as.character(extra_args$Y.symmetric)[1]
    ys_type <- if (ys %in% c("bi", "tri")) ys else "tri"
    base::stop(
      "`Y.symmetric` is no longer supported in nmfkc().\n",
      "Symmetric NMF has moved to nmfkc.net(). Use:\n",
      "    nmfkc.net(Y, rank, type = \"", ys_type, "\")\n",
      "which implements the correct Frobenius bilateral-gradient updates.\n",
      "See help(nmfkc.net) for the types \"tri\", \"bi\", and \"signed\".",
      call. = FALSE
    )
  }

  prefix <- if (!base::is.null(extra_args$prefix)) extra_args$prefix else "Basis"
  print.trace <- if (!base::is.null(extra_args$print.trace)) extra_args$print.trace else FALSE
  print.dims <- verbose
  if (!base::is.null(extra_args$print.dims)) print.dims <- extra_args$print.dims  # backward compat
  save.time <- if (!base::is.null(extra_args$save.time)) extra_args$save.time else FALSE
  save.memory <- if (!base::is.null(extra_args$save.memory)) extra_args$save.memory else FALSE
  detail <- if (!base::is.null(extra_args$detail)) extra_args$detail else NULL
  if (base::is.null(detail)) {
    # backward compatibility: derive detail from save.time / save.memory
    if (save.memory) detail <- "minimal"
    else if (save.time) detail <- "fast"
    else detail <- "full"
  }
  detail <- base::match.arg(detail, base::c("full", "fast", "minimal"))

  Y.weights <- if (!base::is.null(extra_args$Y.weights)) extra_args$Y.weights else NULL

  # --- 2. Input Data Preparation ---
  formula.meta <- NULL
  if (base::inherits(Y, "formula")) {
    data_missing <- base::missing(data)
    resolved <- .nmfkc_resolve_formula(Y, A, data_missing, if (!data_missing) data else NULL)
    Y <- resolved$Y
    A <- resolved$A
    formula.meta <- resolved$formula.meta
  } else {
    if(base::is.vector(Y)) Y <- base::matrix(Y,nrow=1)
    if(!base::is.matrix(Y)) Y <- base::as.matrix(Y)
  }

  # --- Input Validation (after formula dispatch) ---
  if(!base::is.null(A)) {
    if(any(is.na(A))) base::stop("Covariate matrix A contains NAs. Please impute or remove them.")
    if(base::min(A, na.rm=TRUE)<0) base::stop("The matrix A should be non-negative.")
  }
  if(base::min(Y, na.rm=TRUE)<0) base::stop("The matrix Y should be non-negative.")

  # === Weights Handling ===
  # 1. Vector Expansion (Column-wise weights)
  if (!is.null(Y.weights) && is.vector(Y.weights)) {
    if (length(Y.weights) == ncol(Y)) {
      # Expand vector to matrix: each row gets the same weight vector
      Y.weights <- matrix(Y.weights, nrow = nrow(Y), ncol = ncol(Y), byrow = TRUE)
    } else if (length(Y.weights) == 1) {
      Y.weights <- matrix(Y.weights, nrow = nrow(Y), ncol = ncol(Y))
    } else {
      stop("Length of Y.weights vector must match ncol(Y) (or be 1).")
    }
  }

  # 2. Check Dimensions & Handle NAs
  if (is.null(Y.weights)) {
    if (any(is.na(Y))) {
      Y.weights <- matrix(1, nrow=nrow(Y), ncol=ncol(Y))
      Y.weights[is.na(Y)] <- 0
      Y[is.na(Y)] <- 0
      if(print.dims) message("Notice: Missing values (NA) in Y were treated as weights=0.")
    } else {
      Y.weights <- matrix(1, nrow = nrow(Y), ncol = ncol(Y))
    }
  } else {
    if (!is.matrix(Y.weights)) Y.weights <- as.matrix(Y.weights)
    if (!all(dim(Y.weights) == dim(Y))) stop("Dimension mismatch between Y and Y.weights.")
    Y.weights[is.na(Y.weights)] <- 0
    Y[is.na(Y)] <- 0
    Y[Y.weights == 0] <- 0
  }

  # --- 3. Algorithm Setup ---
  X.restriction <- base::match.arg(X.restriction, base::c("colSums", "colSqSums", "totalSum", "none", "fixed"))
  xnorm <- base::switch(X.restriction,
                        colSums   = function(X) base::sweep(X, 2, base::colSums(X), "/"),
                        colSqSums = function(X) base::sweep(X, 2, base::sqrt(base::colSums(X^2)), "/"),
                        totalSum  = function(X) X / base::sum(X),
                        none  = function(X) X,
                        fixed = function(X) X
  )

  if(base::is.null(A)){
    dims <- base::sprintf("Y(%d,%d)~X(%d,%d)B(%d,%d)",
                          base::nrow(Y),base::ncol(Y),base::nrow(Y),Q,Q,base::ncol(Y))
  }else{
    dims <- base::sprintf("Y(%d,%d)~X(%d,%d)C(%d,%d)A(%d,%d)=XB(%d,%d)",
                          base::nrow(Y),base::ncol(Y),base::nrow(Y),Q,Q,base::nrow(A),base::nrow(A),base::ncol(Y),Q,base::ncol(Y))
  }
  if(print.dims) base::message(base::paste0(dims,"..."),appendLF=FALSE)
  start.time <- base::Sys.time()

  # Initialize X
  is.X.scalar <- FALSE
  if(nrow(Y)>=2){
    X <- .nmfkc_init_X(Y, Q, X.init, Y.weights, seed, nstart, maxit, .eps)
  }else{
    X <- matrix(data=1,nrow=1,ncol=1)
    is.X.scalar <- TRUE
  }
  X <- xnorm(X)

  # [FIX: Initialization of tX]
  # Initialize tX here so it exists even if the X update loop is skipped (e.g., scalar X)
  tX <- t(X)

  if(is.null(A)){
    if(is.null(C.init)) C <- matrix(1, nrow=Q, ncol=ncol(Y)) else C <- C.init
  }else{
    if(is.null(C.init)) C <- matrix(1, nrow=Q, ncol=nrow(A)) else C <- C.init
  }
  hasA <- !is.null(A)

  ones_QN <- matrix(1, nrow=Q, ncol=ncol(Y))
  if(hasA) {
    At <- t(A)
    ones_QN_At <- ones_QN %*% At
  }

  epsilon.iter <- Inf
  objfunc.iter <- 0*(1:maxit)
  i_end <- NULL

  # --- 4. Main Loop (Weighted) ---
  for(i in 1:maxit){
    if(is.null(A)) B <- C else B <- C %*% A
    XB <- X %*% B
    if(print.trace && i %% 10==0) message(paste0(format(Sys.time(), "%X")," ",i,"..."))

    if(method=="EU"){
      if(!is.X.scalar && X.restriction!="fixed"){
        num_X <- (Y.weights * Y) %*% t(B)
        den_X <- (Y.weights * XB) %*% t(B)
        if (X.L2.ortho > 0) {
          XtX <- crossprod(X); diag(XtX) <- 0
          den_X <- den_X + X.L2.ortho * (X %*% XtX)
        }
        update_ratio <- num_X / (den_X + .eps)
        X <- X * update_ratio
        X <- xnorm(X)
        tX <- t(X)
      }
      if(is.null(A)) {
        num_C <- tX %*% (Y.weights * Y)
        den_C <- tX %*% (Y.weights * XB)
        if (C.L1 != 0) den_C <- den_C + (C.L1/2) * ones_QN
        if (B.L1 != 0) den_C <- den_C + (B.L1/2) * ones_QN
        C <- C * (num_C / (den_C + .eps))
      } else {
        num_C <- tX %*% (Y.weights * Y) %*% At
        den_C <- tX %*% (Y.weights * XB) %*% At
        if (C.L1 != 0) den_C <- den_C + (C.L1/2) * matrix(1, nrow=Q, ncol=nrow(A))
        if (B.L1 != 0) den_C <- den_C + (B.L1/2) * ones_QN_At
        C <- C * (num_C / (den_C + .eps))
      }
      ## lm()-style weighted least squares: L = sum(W * (Y - XB)^2).
      ## The multiplicative updates (num/den above) already carry W linearly,
      ## so reporting the linear-W objective here makes MU target and reported
      ## loss consistent for any non-negative W.  For binary W in {0,1}
      ## (the standard ECV / CV / NA-mask case) this is identical to
      ## sum((W*(Y-XB))^2) since W = W^2.
      obj <- sum(Y.weights * (Y - XB)^2)

    }else{ # KL
      if(!is.X.scalar && X.restriction!="fixed"){
        ratio <- Y.weights * (Y / (XB + .eps))
        num_X <- ratio %*% t(B)
        den_X <- Y.weights %*% t(B)
        if (X.L2.ortho > 0) {
          XtX <- crossprod(X); diag(XtX) <- 0
          den_X <- den_X + X.L2.ortho * (X %*% XtX)
        }
        update_ratio <- num_X / (den_X + .eps)
        X <- X * update_ratio
        X <- xnorm(X)
        tX <- t(X)
      }
      if(is.null(A)) {
        ratio <- Y.weights * (Y / (XB + .eps))
        num_C <- tX %*% ratio
        den_C <- tX %*% Y.weights
        if (C.L1 != 0) den_C <- den_C + C.L1 * ones_QN
        if (B.L1 != 0) den_C <- den_C + B.L1 * ones_QN
        C <- C * (num_C / (den_C + .eps))
      } else {
        ratio <- Y.weights * (Y / (XB + .eps))
        num_C <- tX %*% ratio %*% At
        den_C <- tX %*% Y.weights %*% At
        if (C.L1 != 0) den_C <- den_C + C.L1 * matrix(1, nrow=Q, ncol=nrow(A))
        if (B.L1 != 0) den_C <- den_C + B.L1 * ones_QN_At
        C <- C * (num_C / (den_C + .eps))
      }
      term1 <- - (Y.weights * Y) * log(XB + .eps)
      term2 <- Y.weights * XB
      obj <- sum(term1 + term2)
    }

    if (C.L1 != 0) obj <- obj + C.L1 * sum(C)
    if (B.L1 != 0) obj <- obj + if (hasA) B.L1 * sum(C %*% A) else B.L1 * sum(C)
    if (X.L2.ortho != 0) {
      XtX <- crossprod(X); diag(XtX) <- 0
      obj <- obj + (X.L2.ortho / 2) * sum(XtX^2)
    }
    objfunc.iter[i] <- obj

    if(i>=10){
      #epsilon.iter <- abs(objfunc.iter[i]-objfunc.iter[i-1])/(abs(objfunc.iter[i])+0.1)
      epsilon.iter <- abs(objfunc.iter[i]-objfunc.iter[i-1]) / pmax(abs(objfunc.iter[i]), 1)
      if(epsilon.iter <= abs(epsilon)){ i_end <- i; break }
    }
  }

  if(is.null(A)) B <- C else B <- C %*% A
  XB <- X %*% B

  if(method=="EU"){
    ## lm()-style weighted least squares; see note above near line 1446.
    objfunc <- sum(Y.weights * (Y - XB)^2)
  } else {
    term1 <- - (Y.weights * Y) * log(XB + .eps)
    term2 <- Y.weights * XB
    objfunc <- sum(term1 + term2)
  }

  if(!is.null(i_end)){ objfunc.iter <- objfunc.iter[10:i_end]
  } else if (i >= 10){ objfunc.iter <- objfunc.iter[10:i]
  } else { objfunc.iter <- objfunc.iter[1:i] }

  if(ncol(X) > 1 && X.restriction != "fixed"){
    index <- order(matrix(1:nrow(X)/nrow(X),nrow=1) %*% X)
    X <- X[,index,drop=FALSE]; B <- B[index,,drop=FALSE]
    C <- C[index,,drop=FALSE]
  }
  rownames(C) <- paste0(prefix,1:nrow(C))
  if (!is.null(A)) {
    colnames(C) <- if (!is.null(rownames(A))) rownames(A) else paste0("Cov", 1:ncol(C))
  }
  rownames(X) <- rownames(Y); colnames(X) <- paste0(prefix,1:ncol(X))
  rownames(B) <- paste0(prefix,1:nrow(B)); colnames(B) <- colnames(Y)

  # --- Compute criteria via nmfkc.criterion ---
  crit_result <- nmfkc.criterion(
    base::list(X = X, B = B, C = C, XB = XB, method = method, A.attr = if(!base::is.null(A)) base::attributes(A) else NULL),
    Y, detail = detail, Y.weights = Y.weights, X.restriction = X.restriction
  )
  r2          <- crit_result$r.squared
  r2.uncentered     <- crit_result$r.squared.uncentered
  r2.centered <- crit_result$r.squared.centered
  sigma       <- crit_result$sigma
  mae         <- crit_result$mae
  B.prob    <- crit_result$B.prob
  B.cluster <- crit_result$B.cluster
  X.prob    <- crit_result$X.prob
  X.cluster <- crit_result$X.cluster
  if (detail == "minimal") XB <- NA

  if(epsilon.iter > abs(epsilon)) warning(paste0("maximum iterations (",maxit,") reached..."))
  end.time <- Sys.time()
  diff.time.st <- paste0(round(difftime(end.time,start.time,units="sec"),1),"sec")
  if(print.dims) message(diff.time.st)

  n.missing <- sum(Y.weights == 0)
  n.total <- prod(dim(Y))
  A.attr <- NULL
  if (!is.null(A)) A.attr <- attributes(A)

  result <- list(
    call      = match.call(),
    dims      = dims,
    runtime   = diff.time.st,
    method    = method,
    X         = X,
    B         = B,
    XB        = XB,
    C         = C,
    B.prob    = B.prob,
    B.cluster = B.cluster,
    X.prob    = X.prob,
    X.cluster = X.cluster,
    A.attr    = A.attr,
    formula.meta = formula.meta,
    n.missing = n.missing,
    n.total   = n.total,
    rank      = Q,
    objfunc   = objfunc,
    objfunc.iter = objfunc.iter,
    r.squared          = r2,
    r.squared.uncentered     = r2.uncentered,
    r.squared.centered = r2.centered,
    sigma     = sigma,
    mae = mae,
    criterion = crit_result$criterion
  )
  class(result) <- c("nmfkc", "nmf")
  return(result)
}















#' @title Plot method for objects of class \code{nmfkc}
#' @description
#' \code{plot.nmfkc} produces a diagnostic plot for the return value of
#' \code{nmfkc}, showing the objective function across iterations.
#'
#' @param x An object of class \code{nmfkc}, i.e., the return value of \code{nmfkc}.
#' @param ... Additional arguments passed to the base \code{\link{plot}} function.
#' @return Called for its side effect (a plot). Returns \code{NULL} invisibly.
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(1, cars$speed)
#' result <- nmfkc(Y, A, rank = 1)
#' plot(result)
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{summary.nmfkc}}
#' @export
plot.nmfkc <- function(x,...){
  extra_args <- list(...)
  args <- list(x = x$objfunc.iter)
  if(is.null(extra_args$main)) args$main <- paste0("r.squared=",round(x$r.squared,3))
  if(is.null(extra_args$xlab)) args$xlab <- "iter"
  if(is.null(extra_args$ylab)) args$ylab <- "objfunc"
  all_args <- c(args, extra_args)
  do.call("plot",all_args)
}




#' @title Summary method for objects of class \code{nmfkc}
#' @description
#' Produces a summary of an \code{nmfkc} object, including matrix dimensions,
#' runtime, fit statistics, and diagnostics.
#'
#' @param object An object of class \code{nmfkc}, i.e., the return value of \code{nmfkc}.
#' @param ... Additional arguments (currently unused).
#' @return An object of class \code{summary.nmfkc}, containing summary statistics.
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(1, cars$speed)
#' result <- nmfkc(Y, A, rank = 1)
#' summary(result)
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfkc.inference}}, \code{\link{plot.nmfkc}}
#' @export
summary.nmfkc <- function(object, ...) {
  ans <- list()
  ans$call <- object$call
  ans$dims <- object$dims
  ans$rank <- object$rank
  ans$runtime <- object$runtime

  # Missing values
  if(!is.null(object$n.missing) && !is.null(object$n.total)){
    ans$n.missing <- object$n.missing
    ans$prop.missing <- object$n.missing / object$n.total * 100
  } else {
    ans$n.missing <- NULL
  }

  ans$formula.meta <- object$formula.meta
  ans$method <- object$method
  ans$iter <- length(object$objfunc.iter)
  ans$objfunc <- object$objfunc
  ans$r.squared          <- object$r.squared
  ans$r.squared.uncentered     <- object$r.squared.uncentered
  ans$r.squared.centered <- object$r.squared.centered
  ans$sigma <- object$sigma
  ans$mae <- object$mae
  if(!is.null(object$criterion)){
    ans$effective.rank <- object$criterion$effective.rank
  } else {
    ans$effective.rank <- NULL
  }
  ans$rank <- object$rank

  # --- Diagnostics (Sparsity & Clustering Quality) ---

  # 1. Basis (X)
  if (!is.null(object$X) && is.matrix(object$X)) {
    # Sparsity: Proportion of elements close to zero (< 1e-4)
    ans$X.sparsity <- mean(object$X < 1e-4)
  }

  # 2. Probabilities (B.prob)
  if (!is.null(object$B.prob)){
    # Sparsity
    ans$B.prob.sparsity <- mean(object$B.prob < 1e-4)
    ans$B.prob.max.mean <- object$criterion$B.prob.max.mean
  }

  class(ans) <- "summary.nmfkc"
  return(ans)
}

#' @title Print method for \code{summary.nmfkc} objects
#' @description
#' Prints a formatted summary of an \code{nmfkc} model fit.
#'
#' @param x An object of class \code{summary.nmfkc}.
#' @param digits Minimum number of significant digits to be used.
#' @param ... Additional arguments (currently unused).
#' @return Called for its side effect (printing). Returns \code{x} invisibly.
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(1, cars$speed)
#' result <- nmfkc(Y, A, rank = 1)
#' print(summary(result))
#'
#' @export
print.summary.nmfkc <- function(x, digits = max(3L, getOption("digits") - 3L), ...) {
  cat("\nCall:\n", paste(deparse(x$call), sep = "\n", collapse = "\n"), "\n\n", sep = "")

  if (!is.null(x$formula.meta)) {
    cat("Formula:    ", deparse(x$formula.meta$formula), "\n")
  }
  cat("Dimensions:", x$dims, "\n")
  if(!is.null(x$rank)) cat("Rank (Q):   ", x$rank, "\n")
  cat("Runtime:    ", x$runtime, "\n")
  if (!is.null(x$method)) cat("Method:     ", x$method, "\n")
  cat("Iterations: ", x$iter, "\n")

  if (!is.null(x$n.missing)) {
    cat("Missing:    ", x$n.missing,
        sprintf("(%.1f%%)", x$prop.missing), "\n")
  }

  .print.fit.statistics(x, digits = digits)

  .print.structure.diagnostics(
    sparsity  = c("Basis (X)" = x$X.sparsity, "Coef (B)" = x$B.prob.sparsity),
    crispness = x$B.prob.max.mean)
  cat("\n")
  invisible(x)
}














#' @title Normalize a matrix to the range \eqn{[0,1]}
#' @description
#' \code{nmfkc.normalize} rescales the values of a matrix to lie between 0 and 1
#' using the column-wise minimum and maximum values of a reference matrix.
#'
#' @param x A numeric matrix (or vector) to be normalized.
#' @param ref A reference matrix from which the column-wise minima and maxima are taken.
#'   Default is \code{x}.
#'
#' @return A matrix of the same dimensions as \code{x}, with each column rescaled to the \eqn{[0,1]} range.
#' @seealso \code{\link{nmfkc.denormalize}}
#' @export
#' @examples
#' # Example.
#' x <- nmfkc.normalize(iris[,-5])
#' apply(x,2,range)
nmfkc.normalize <- function(x,ref=x){
  if(is.vector(x)){
    x <- matrix(x,ncol=1)
    ref <- matrix(ref,ncol=1)
  }
  r <- apply(ref,2,range)
  denom <- r[2, ] - r[1, ]
  denom[denom == 0] <- 1   # leave zero-width columns unchanged
  y <- sweep(x, 2, r[1, ], FUN = "-")
  y <- sweep(y, 2, denom,   FUN = "/")
  return(y)
}




#' @title Denormalize a matrix from \eqn{[0,1]} back to its original scale
#' @description
#' \code{nmfkc.denormalize} rescales a matrix with values in \eqn{[0,1]} back to its
#' original scale using the column-wise minima and maxima of a reference matrix.
#'
#' @param x A numeric matrix (or vector) with values in \eqn{[0,1]} to be denormalized.
#' @param ref A reference matrix used to obtain the original column-wise minima
#'   and maxima. Must have the same number of columns as \code{x}.
#'
#' @return A numeric matrix with values transformed back to the original scale.
#' @seealso \code{\link{nmfkc.normalize}}
#' @export
#' @examples
#' x <- nmfkc.normalize(iris[, -5])
#' x_recovered <- nmfkc.denormalize(x, iris[, -5])
#' apply(x_recovered - iris[, -5], 2, max)
nmfkc.denormalize <- function(x, ref=x) {
  if (is.vector(x)) {
    x <- matrix(x, ncol = 1)
    ref <- matrix(ref, ncol = 1)
  }
  r <- apply(ref, 2, range)
  y <- sweep(x, 2, r[2, ] - r[1, ], FUN = "*")
  y <- sweep(y, 2, r[1, ], FUN = "+")
  return(y)
}




#' @title Create a class (one-hot) matrix from a categorical vector
#' @description
#' \code{nmfkc.class} converts a categorical or factor vector into a class matrix
#' (one-hot encoded representation), where each row corresponds to a category
#' and each column corresponds to an observation.
#'
#' @param x A categorical vector or a factor.
#'
#' @return A binary matrix with one row per unique category and one column per observation. Each column has exactly one entry equal to 1, indicating the category of the observation.
#' @seealso \code{\link{nmfkc}}
#' @export
#' @examples
#' # Example.
#' Y <- nmfkc.class(iris$Species)
#' Y[,1:6]
nmfkc.class <- function(x){
  if(!is.factor(x)) x <- as.factor(x)
  lev <- levels(x)
  X <- outer(lev, x, "==")
  mode(X) <- "numeric"
  rownames(X) <- lev
  if(!is.null(names(x))) colnames(X) <- names(x) else colnames(X) <- seq_along(x)
  X
}





#' @title Prediction method for objects of class \code{nmfkc}
#' @description
#' \code{predict.nmfkc} generates predictions from an object of class \code{nmfkc},
#' either using the fitted covariates or a new covariate matrix.
#'
#' When the model was fitted using a formula (Formula Mode), a \code{newdata}
#' data frame can be supplied instead of \code{newA}; the covariate matrix is
#' then constructed automatically from the stored formula metadata.
#'
#' @param object An object of class \code{nmfkc}, i.e., the return value of \code{nmfkc}.
#' @param newA Optional. A new covariate matrix to be used for prediction.
#' @param newdata Optional data frame. Only available when the model was fitted
#'   using a formula. Covariate columns are extracted automatically using the
#'   stored formula metadata. If both \code{newdata} and \code{newA} are
#'   supplied, \code{newdata} takes precedence (with a warning).
#' @param type Type of prediction to return. Options are "response" (fitted values matrix),
#'   "prob" (soft-clustering probabilities), or "class" (hard-clustering labels based on row names of X).
#' @param ... Further arguments passed to or from other methods.
#' @return Depending on \code{type}: a numeric matrix (\code{"response"} or \code{"prob"})
#'   or a character vector of class labels (\code{"class"}).
#' @examples
#' # Prediction with newA
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(1, cars$speed)
#' result <- nmfkc(Y, A, rank = 1)
#' newA <- rbind(1, c(10, 20, 30))
#' predict(result, newA = newA)
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfkc.cv}}
#' @export
predict.nmfkc <- function(object, newA = NULL, newdata = NULL, type = "response", ...) {
  x <- object
  .eps <- 1e-10

  # --- newdata handling (Formula Mode) ---
  if (!is.null(newdata)) {
    if (is.null(x$formula.meta)) {
      stop("'newdata' can only be used when the model was fitted with a formula.")
    }
    if (!is.null(newA)) {
      warning("Both 'newdata' and 'newA' supplied; 'newdata' takes precedence.")
    }
    A_cols <- x$formula.meta$A_cols
    if (is.null(A_cols)) {
      stop("Formula metadata has no A columns. Model was fitted without covariates (standard NMF).")
    }
    missing_cols <- setdiff(A_cols, names(newdata))
    if (length(missing_cols) > 0) {
      stop(paste0("'newdata' is missing required columns: ", paste(missing_cols, collapse = ", ")))
    }
    newdata <- as.data.frame(newdata)
    newA <- t(as.matrix(newdata[, A_cols, drop = FALSE]))
  }

  if(is.null(newA)){
    if(type=="response"){
      result <- x$X %*% x$B
    }else{
      XB.prob <- x$X %*% x$B.prob
      if(type=="prob"){
        result <- XB.prob
      }else{
        result <- rownames(x$X)[apply(XB.prob,2,which.max)]
      }
    }
  }else{
    B <- x$C %*% newA
    if(type=="response"){
      result <- x$X %*% B
    }else{
      B.prob <- sweep(B, 2, colSums(B) + .eps, "/")
      XB.prob <- x$X %*% B.prob
      if(type=="prob"){
        result <- XB.prob
      }else{
        result <- rownames(x$X)[apply(XB.prob,2,which.max)]
      }
    }
  }
  return(result)
}








#' @title Perform k-fold cross-validation for NMF with kernel covariates
#'
#' @description
#' \code{nmfkc.cv} performs k-fold cross-validation for the tri-factorization model
#' \eqn{Y \approx X C A = X B}, where
#' \itemize{
#'   \item \eqn{Y(P,N)} is the observation matrix,
#'   \item \eqn{A(R,N)} is the covariate (or kernel) matrix,
#'   \item \eqn{X(P,Q)} is the basis matrix,
#'   \item \eqn{C(Q,R)} is the parameter matrix, and
#'   \item \eqn{B(Q,N)} is the coefficient matrix (\eqn{B = C A}).
#' }
#' Given \eqn{Y} (and optionally \eqn{A}), \eqn{X} and \eqn{C} are fitted on each
#' training split and predictive performance is evaluated on the held-out split.
#'
#' @param Y Observation matrix, or a formula (see \code{\link{nmfkc}} for Formula Mode).
#' @param A Covariate matrix. If \code{NULL}, the identity matrix is used.
#'   Ignored when \code{Y} is a formula.
#' @param rank Rank of the basis matrix \eqn{X}. Default is 2.
#' @param data A data frame (required when \code{Y} is a formula with column names).
#' @param ... Additional arguments controlling CV and the internal \code{\link{nmfkc}} call:
#'   \describe{
#'     \item{\code{Y.weights}}{Non-negative weight matrix or vector
#'       (\code{lm()}-style: loss \eqn{\sum W \, r^2}).  Binary
#'       \code{{0,1}} masks (\code{TRUE} / \code{FALSE} also accepted)
#'       are the typical ECV usage -- 0/\code{FALSE} excludes an element.
#'       See \code{\link{nmfkc}} for full details.}
#'     \item{\code{div}}{Number of folds (\eqn{k}); default: \code{5}.}
#'     \item{\code{seed}}{Integer seed for reproducible partitioning; default: \code{123}.}
#'     \item{\code{shuffle}}{Logical. If \code{TRUE} (default), randomly shuffles samples (standard CV);
#'       if \code{FALSE}, splits sequentially (block CV; recommended for time series).}
#'     \item{\code{Q}}{(Deprecated) Alias for \code{rank}.}
#'     \item{\emph{Arguments passed to} \code{\link{nmfkc}}}{e.g., \code{gamma} (\code{B.L1}), \code{epsilon},
#'       \code{maxit}, \code{method} (\code{"EU"} or \code{"KL"}), \code{X.restriction}, \code{X.init}, etc.}
#'   }
#'
#' @return A list with components:
#'   \describe{
#'     \item{\code{objfunc}}{Mean loss per valid entry over all folds (MSE for \code{method="EU"}).}
#'     \item{\code{sigma}}{Residual standard error (RMSE). Available only if \code{method="EU"}; on the same scale as \code{Y}.}
#'     \item{\code{objfunc.block}}{Loss for each fold.}
#'     \item{\code{block}}{Vector of fold indices (1, …, \code{div}) assigned to each column of \eqn{Y}.}
#'   }
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfkc.kernel.beta.cv}}, \code{\link{nmfkc.ar.degree.cv}}
#'
#' @examples
#' # Example 1 (with explicit covariates):
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(1, cars$speed)
#' res <- nmfkc.cv(Y, A, rank = 1)
#' res$objfunc
#'
#' # Example 2 (kernel A and beta sweep):
#' Y <- matrix(cars$dist, nrow = 1)
#' U <- matrix(c(5, 10, 15, 20, 25), nrow = 1)
#' V <- matrix(cars$speed, nrow = 1)
#' betas <- 25:35/1000
#' obj <- numeric(length(betas))
#' for (i in seq_along(betas)) {
#'   A <- nmfkc.kernel(U, V, beta = betas[i])
#'   obj[i] <- nmfkc.cv(Y, A, rank = 1, nfolds = 10)$objfunc
#' }
#' betas[which.min(obj)]
#'
#' @export

nmfkc.cv <- function(Y, A=NULL, rank=2, data, ...){
  # --- Formula Mode ---
  if (base::inherits(Y, "formula")) {
    resolved <- .nmfkc_resolve_formula(Y, A, base::missing(data), if (!base::missing(data)) data else NULL)
    Y <- resolved$Y
    A <- resolved$A
  }

  # A small constant for numerical stability to prevent division by zero and log(0).
  .eps <- 1e-10

  extra_args <- list(...)
  if (!is.null(extra_args$Q)) rank <- extra_args$Q
  Q <- rank

  div <- if (!is.null(extra_args$div)) extra_args$div else 5
  seed <- if (!is.null(extra_args$seed)) extra_args$seed else 123
  shuffle <- if (!is.null(extra_args$shuffle)) extra_args$shuffle else TRUE

  gamma <- if (!is.null(extra_args$B.L1)) extra_args$B.L1 else if (!is.null(extra_args$gamma)) extra_args$gamma else 0
  epsilon <- if (!is.null(extra_args$epsilon)) extra_args$epsilon else 1e-4
  maxit   <- if (!is.null(extra_args$maxit))   extra_args$maxit   else 5000
  method  <- if (!is.null(extra_args$method))  extra_args$method  else "EU"
  Y.weights <- if (!is.null(extra_args$Y.weights)) extra_args$Y.weights else NULL

  if(!is.matrix(Y)) Y <- as.matrix(Y)
  P <- nrow(Y)
  N <- ncol(Y)

  # === Weights Preparation (Same as nmfkc) ===
  # Expand vector to matrix for easier slicing
  if (!is.null(Y.weights) && is.vector(Y.weights)) {
    if (length(Y.weights) == N) {
      Y.weights <- matrix(Y.weights, nrow = P, ncol = N, byrow = TRUE)
    } else if (length(Y.weights) == 1) {
      Y.weights <- matrix(Y.weights, nrow = P, ncol = N)
    } else {
      stop("Length of Y.weights vector must match ncol(Y) (or be 1).")
    }
  }
  # Handle NAs
  if (is.null(Y.weights)) {
    if (any(is.na(Y))) {
      Y.weights <- matrix(1, nrow=P, ncol=N)
      Y.weights[is.na(Y)] <- 0
      Y[is.na(Y)] <- 0
    } else {
      Y.weights <- matrix(1, nrow=P, ncol=N) # Full 1s for easier slicing logic
    }
  } else {
    if (!is.matrix(Y.weights)) Y.weights <- as.matrix(Y.weights)
    Y.weights[is.na(Y.weights)] <- 0
    Y[is.na(Y)] <- 0
    Y <- Y * Y.weights
  }

  # --- Helper: Weighted Optimization of B given fixed X and Y_test ---
  optimize.B.from.Y <- function(result, Y_test, W_test, gamma, epsilon, maxit, method){
    X <- result$X
    # Initialize C (which acts as B here)
    C <- matrix(1, nrow=ncol(X), ncol=ncol(Y_test))

    # Precompute ones for penalty (must match test dimension)
    ones_QN <- matrix(1, nrow=ncol(X), ncol=ncol(Y_test))

    oldSum <- 0
    epsilon.iter <- Inf

    for(l in 1:maxit){
      B <- C
      XB <- X %*% B

      if(method=="EU"){
        # Weighted Update for B (EU)
        # Num: X^T (W * Y)
        num <- t(X) %*% (W_test * Y_test)
        # Den: X^T (W * XB)
        den <- t(X) %*% (W_test * XB)

        if(gamma != 0) den <- den + (gamma/2) * ones_QN
        C <- C * ( num / (den + .eps) )

      }else{ # KL
        # Weighted Update for B (KL)
        # Num: X^T (W * (Y/XB))
        ratio <- W_test * (Y_test / (XB + .eps))
        num <- t(X) %*% ratio
        # Den: X^T W
        den <- t(X) %*% W_test

        if(gamma != 0) den <- den + gamma * ones_QN
        C <- C * ( num / (den + .eps) )
      }

      newSum <- sum(C)
      if(l>=10){
        #epsilon.iter <- abs(newSum-oldSum)/(abs(newSum)+0.1)
        epsilon.iter <- abs(newSum-oldSum) / pmax(abs(newSum), 1)
        if(epsilon.iter <= abs(epsilon)) break
      }
      oldSum <- sum(C)
    }

    B <- C
    XB <- X %*% B
    # Note: colnames handling omitted for speed in CV
    return(list(B=B, XB=XB))
  }

  is.identity.matrix <- function(A, tol = .Machine$double.eps) {
    if (nrow(A) != ncol(A)) return(FALSE)
    isTRUE(all.equal(A, diag(nrow(A)), tolerance = tol))
  }

  if(is.null(A)){
    is_identity <- TRUE
    is_symmetric.matrix <- FALSE
  }else{
    is_identity <- is.identity.matrix(A)
    is_symmetric.matrix <- isSymmetric(A, tol=.Machine$double.eps)
    A.function <- attr(A, "function.name")
    is_kernel_matrix <- !is.null(A.function) && A.function == "nmfkc.kernel"
  }

  # Create Folds
  remainder <- N %% div
  division <- N %/% div
  block <- 0*(1:N)

  if(shuffle){
    set.seed(seed)
    perm_index <- sample(1:N, N, replace=FALSE)
  } else {
    perm_index <- 1:N
  }

  processed_count <- 0
  for(i in 1:(div-1)){
    plus <- ifelse(i <= remainder, 1, 0)
    chunk_size <- division + plus
    target_indices <- perm_index[(processed_count + 1):(processed_count + chunk_size)]
    block[target_indices] <- i
    processed_count <- processed_count + chunk_size
  }
  target_indices <- perm_index[(processed_count + 1):N]
  block[target_indices] <- div

  objfunc.block <- numeric(div)
  total_valid_obs <- 0 # Denominator for weighted mean error

  for(j in 1:div){
    # Slice Y and Weights
    # Train
    Y_train <- Y[,block!=j, drop=FALSE]
    W_train <- Y.weights[,block!=j, drop=FALSE]

    # Test
    Y_test <- Y[,block==j, drop=FALSE]
    W_test <- Y.weights[,block==j, drop=FALSE]

    # Slice A
    if(is_identity){
      A_train <- NULL
    }else{
      if(is_symmetric.matrix && is_kernel_matrix){
        A_train <- A[block!=j,block!=j, drop=FALSE]
        A_test  <- A[block!=j,block==j, drop=FALSE]
      }else{
        A_train <- A[,block!=j, drop=FALSE]
        A_test  <- A[,block==j, drop=FALSE]
      }
    }

    # Run NMF on Training set (passing W_train)
    nmfkc_args <- c(
      list(...),
      list(Y = Y_train, A = A_train, Q = Q, Y.weights = W_train, # Pass weights!
           seed=NULL, print.trace = FALSE, print.dims = FALSE,
           save.time = TRUE, save.memory = TRUE)
    )
    nmfkc_args$shuffle <- NULL

    # Suppress messages from inner nmfkc calls
    res_j <- suppressMessages(do.call("nmfkc", nmfkc_args))

    # Predict on Test set
    if(is_identity){
      # Standard NMF: Optimize B for test set using weights
      resj <- optimize.B.from.Y(res_j, Y_test, W_test, gamma, epsilon, maxit, method)
      XB_test <- resj$XB
    }else{
      # Covariate NMF: Predict using A_test
      # XB = X * C * A_test
      XB_test <- res_j$X %*% res_j$C %*% A_test
    }

    # Evaluate Error (Weighted)
    if(method=="EU"){
      resid <- W_test * (Y_test - XB_test)
      objfunc.block[j] <- sum(resid^2)
    }else{
      term1 <- - (W_test * Y_test) * log(XB_test + .eps)
      term2 <- W_test * XB_test
      objfunc.block[j] <- sum(term1 + term2)
    }

    total_valid_obs <- total_valid_obs + sum(W_test > 0)
  }

  # Mean error per valid observation
  # (Avoid division by zero if all weights are 0, though unlikely)
  objfunc <- sum(objfunc.block) / max(total_valid_obs, 1)

  # Calculate RMSE for EU (This corresponds to 'sigma')
  sigma <- if(method == "EU") sqrt(objfunc) else NA

  return(list(objfunc=objfunc, sigma=sigma, objfunc.block=objfunc.block, block=block))
}





#' @title Perform Element-wise Cross-Validation (Wold's CV)
#' @description
#' \code{nmfkc.ecv} performs k-fold cross-validation by randomly holding out
#' individual elements of the data matrix (element-wise), assigning them a
#' weight of 0 via `Y.weights`, and evaluating the reconstruction error on
#' those held-out elements.
#'
#' This method (also known as Wold's CV) is theoretically robust for determining
#' the optimal rank (Q) in NMF. This function supports vector input for `Q`,
#' allowing simultaneous evaluation of multiple ranks on the same folds.
#'
#' For symmetric (network) data use \code{\link{nmfkc.net.ecv}}, which
#' creates upper-triangle folds to prevent information leakage through
#' the symmetric entries \eqn{Y_{ij} = Y_{ji}}.  Passing the old
#' \code{Y.symmetric} argument here is no longer supported and stops
#' with a redirect message.
#'
#' @param Y Observation matrix, or a formula (see \code{\link{nmfkc}} for Formula Mode).
#' @param A Covariate matrix. Ignored when \code{Y} is a formula.
#' @param rank Vector of ranks to evaluate (e.g., 1:5). For backward compatibility,
#'   \code{Q} is accepted via \code{...}.
#' @param data A data frame (required when \code{Y} is a formula with column names).
#' @param ... Additional arguments passed to \code{\link{nmfkc}} (e.g., \code{method="EU"}).
#'   Also accepts: \code{nfolds} (number of folds, default 5; \code{div} also accepted),
#'   \code{seed} (integer seed, default 123).
#'
#' @return A list with components:
#' \item{objfunc}{Numeric vector containing the Mean Squared Error (MSE) for each Q.}
#' \item{sigma}{Numeric vector containing the Residual Standard Error (RMSE) for each Q. Only available if method="EU".}
#' \item{objfunc.fold}{List of length equal to Q vector. Each element contains the MSE values for the k folds.}
#' \item{folds}{A list of length \code{div}, containing the linear indices of held-out elements for each fold (shared across all Q).}
#' @references
#' Wold, S. (1978).  Cross-validatory estimation of the number of
#' components in factor and principal components models.
#' \emph{Technometrics}, 20(4), 397--405.
#' \doi{10.1080/00401706.1978.10489693}
#' Owen, A. B., & Perry, P. O. (2009).  Bi-cross-validation of the SVD
#' and the nonnegative matrix factorization.  \emph{Ann. Appl. Stat.}
#' 3(2), 564--594. \doi{10.1214/08-AOAS227}  (cross-validation of the
#' NMF rank; see also \code{\link{nmfkc.bicv}}).
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfkc.cv}}; other rank-selection
#'   criteria: \code{\link{nmfkc.rank}}, \code{\link{nmfkc.bicv}},
#'   \code{\link{nmfkc.consensus}}, \code{\link{nmfkc.ard}}.
#' @examples
#' # Element-wise CV to select rank
#' Y <- t(iris[1:30, 1:4])
#' res <- nmfkc.ecv(Y, rank = 1:2, nfolds = 3)
#' res$objfunc
#'
#' @export
nmfkc.ecv <- function(Y, A=NULL, rank=1:3, data, ...){
  extra_ecv <- list(...)
  if (!is.null(extra_ecv$Q)) rank <- extra_ecv$Q
  nfolds <- if (!is.null(extra_ecv$nfolds)) extra_ecv$nfolds else if (!is.null(extra_ecv$div)) extra_ecv$div else 5
  seed   <- if (!is.null(extra_ecv$seed))   extra_ecv$seed   else 123
  Q <- rank
  div <- nfolds
  # --- Formula Mode ---
  if (base::inherits(Y, "formula")) {
    resolved <- .nmfkc_resolve_formula(Y, A, base::missing(data), if (!base::missing(data)) data else NULL)
    Y <- resolved$Y
    A <- resolved$A
  }

  if(!is.matrix(Y)) Y <- as.matrix(Y)
  P <- nrow(Y)
  N <- ncol(Y)

  # --- Argument Handling ---
  extra_args <- list(...)

  # If user mistakenly passed 'rank' in ..., treat it as Q
  if (!is.null(extra_args$rank)) {
    Q <- extra_args$rank
  }

  # Symmetric element-wise CV (upper-triangle folds) has moved to
  # nmfkc.net.ecv().  Passing Y.symmetric here stops with a redirect.
  if (!is.null(extra_args$Y.symmetric) &&
      !identical(as.character(extra_args$Y.symmetric), "none")) {
    ys <- as.character(extra_args$Y.symmetric)[1]
    ys_type <- if (ys %in% c("bi", "tri")) ys else "tri"
    stop("`Y.symmetric` is no longer supported in nmfkc.ecv().\n",
         "Symmetric element-wise CV has moved to nmfkc.net.ecv(). Use:\n",
         "    nmfkc.net.ecv(Y, rank, type = \"", ys_type, "\")\n",
         "See help(nmfkc.net.ecv).", call. = FALSE)
  }

  # 1. Create Folds (shared element-wise helper)
  folds <- .ecv.make.folds(Y, div, seed)

  method <- if(!is.null(extra_args$method)) extra_args$method else "EU"

  # Clean extra args for the inner nmfkc() calls (once, outside the loop)
  nmfkc_clean_args <- extra_args
  nmfkc_clean_args$Q <- NULL
  nmfkc_clean_args$rank <- NULL
  nmfkc_clean_args$print.trace <- NULL
  nmfkc_clean_args$print.dims <- NULL
  nmfkc_clean_args$save.time <- NULL
  nmfkc_clean_args$save.memory <- NULL

  # Model-specific worker: mask fold k, refit at rank Q[i], held-out loss
  run_one <- function(i, k) {
    q_curr <- Q[i]
    test_idx <- folds[[k]]
    weights_train <- matrix(1, nrow = P, ncol = N)
    if (any(is.na(Y))) weights_train[is.na(Y)] <- 0
    weights_train[test_idx] <- 0
    nmfkc_args <- c(list(Y = Y, A = A, Q = q_curr, Y.weights = weights_train,
                         print.trace = FALSE, print.dims = FALSE,
                         save.time = TRUE), nmfkc_clean_args)
    fit <- suppressMessages(do.call("nmfkc", nmfkc_args))
    pred <- fit$XB
    if (method == "KL") {
      .eps <- 1e-10
      mean(-Y[test_idx] * log(pred[test_idx] + .eps) + pred[test_idx])
    } else {
      mean((Y[test_idx] - pred[test_idx])^2)
    }
  }

  # 2. Loop over Q via shared driver
  message(paste0("Performing Element-wise CV for Q = ", paste(Q, collapse=","), " (", div, "-fold)..."))
  cv <- .ecv.run(paste0("Q=", Q), div, run_one)
  if (method != "EU") cv$sigma[] <- NA   # sigma = RMSE only for EU loss

  return(list(objfunc = cv$objfunc,
              sigma = cv$sigma,
              objfunc.fold = cv$objfunc.fold,
              folds = folds))
}


#' @title Compute model selection criteria for a fitted nmfkc model
#' @description
#' \code{nmfkc.criterion} computes the effective rank, clustering-quality
#' measures (silhouette, CPCC, dist.cor), and the clustering-crispness
#' statistic (\code{B.prob.max.mean}) from a fitted \code{nmfkc} model.
#'
#' This function can be called on a model that was fitted with
#' \code{detail = "fast"} or \code{detail = "minimal"} to compute the
#' full set of criteria afterwards.
#'
#' @param object An object of class \code{"nmfkc"} returned by
#'   \code{\link{nmfkc}}.
#' @param Y The original observation matrix (P x N) used for fitting.
#' @param detail Character string controlling the level of computation:
#'   \code{"full"} (default) computes all criteria including silhouette,
#'   CPCC and dist.cor;
#'   \code{"fast"} skips the expensive distance-based criteria;
#'   \code{"minimal"} skips the fit and clustering statistics.
#' @param ... Additional arguments: \code{Y.weights} (non-negative
#'   weight matrix; \code{lm()}-style loss \eqn{\sum W \, r^2}; default:
#'   all ones).  See \code{\link{nmfkc}} for full details.
#'
#' @return A list with components:
#' \describe{
#'   \item{r.squared}{R-squared between Y and XB.}
#'   \item{sigma}{Residual standard deviation.}
#'   \item{mae}{Mean absolute error.}
#'   \item{B.prob}{Column-normalized coefficient matrix (soft-clustering probabilities).}
#'   \item{B.cluster}{Hard clustering labels (argmax of B.prob per column).}
#'   \item{X.prob}{Row-normalized basis matrix.}
#'   \item{X.cluster}{Hard clustering labels per row of X.}
#'   \item{criterion}{Named list: B.prob.max.mean, effective.rank,
#'     silhouette, CPCC, dist.cor.}
#' }
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfkc.rank}}
#' @export
#' @examples
#' Y <- t(iris[, -5])
#' res <- nmfkc(Y, rank = 3, detail = "fast")
#' crit <- nmfkc.criterion(res, Y)
#' crit$criterion$silhouette
#'
nmfkc.criterion <- function(object, Y, detail = c("full", "fast", "minimal"), ...) {
  detail <- match.arg(detail)
  extra_args <- base::list(...)

  X <- object$X
  B <- object$B
  C <- object$C
  XB <- object$XB
  method <- if (!is.null(object$method)) object$method else "EU"
  Q <- base::ncol(X)
  .eps <- 1e-10

  Y <- base::as.matrix(Y)
  P <- base::nrow(Y)
  N <- base::ncol(Y)

  # Y.weights
  Y.weights <- if (!is.null(extra_args$Y.weights)) extra_args$Y.weights
               else base::matrix(1, nrow = P, ncol = N)

  # Recompute XB if missing (detail="minimal" case)
  if (base::is.null(XB) || (base::length(XB) == 1 && base::is.na(XB))) {
    hasA <- !base::is.null(object$A.attr)
    if (hasA) {
      # Cannot recompute without A; use object$B
      XB <- X %*% B
    } else {
      XB <- X %*% B
    }
  }

  # --- Fit statistics and clustering (detail != "minimal") ---
  if (detail != "minimal") {
    valid_idx <- (Y.weights > 0)
    if (base::any(valid_idx)) {
      r2_all <- .r.squared.all(Y, XB, Y.weights = Y.weights)
      r2 <- r2_all$r.squared
      r2.uncentered <- r2_all$r.squared.uncentered
      r2.centered <- r2_all$r.squared.centered
      sigma <- stats::sd(Y[valid_idx] - XB[valid_idx])
      mae <- base::mean(base::abs(Y[valid_idx] - XB[valid_idx]))
    } else {
      r2 <- NA; r2.uncentered <- NA; r2.centered <- NA
      sigma <- NA; mae <- NA
    }

    B.prob <- base::t(base::t(B) / (base::colSums(B) + .eps))
    ## Clustering crispness: mean over samples of the dominant-cluster
    ## membership.  Range [1/Q, 1]; higher = more decisive (hard-like)
    ## soft assignment.  Meaningful at fixed Q (e.g. as a confidence
    ## check before treating B.cluster as hard labels); not used for
    ## rank selection (it is monotone in Q).
    if (Q > 1) {
      B.prob.max.mean <- base::mean(base::apply(B.prob, 2, base::max))
    } else {
      B.prob.max.mean <- 1
    }
    B.cluster <- base::apply(B.prob, 2, base::which.max)
    B.cluster[base::colSums(B.prob) == 0] <- NA
    X.prob <- X / (base::rowSums(X) + .eps)
    X.cluster <- base::apply(X.prob, 1, base::which.max)

    ## Effective rank: exp(Shannon entropy) of the explained-variance
    ## distribution p_k = var(B[k, ]) / sum_j var(B[j, ]).  We use the
    ## across-sample VARIANCE (not the standard deviation) on purpose:
    ## by the trace identity sum_k var(B[k, ]) = tr(Cov(B)) = total
    ## variance, so p_k is the exact fraction of total variance carried
    ## by factor k -- a genuine additive decomposition, which is what
    ## makes the entropy "effective number" interpretation well-posed
    ## (standard deviations are not additive, so sum_k sd_k has no such
    ## meaning).  This is the PCA-style "explained-variance entropy" /
    ## effective-dimensionality measure; it reuses the exp(entropy)
    ## functional form of Roy & Vetterli (2007).  See .effective.rank().
    ## NA at Q = 1 (single-point variance distribution -> trivially 1),
    ## consistent with silhouette / CPCC which are also NA at Q = 1.
    effective.rank <- .effective.rank(B)

    # Distance-based criteria (detail == "full" only); computed by the
    # shared .cluster.criteria() helper using the hard labels B.cluster
    # already derived above (so values are unchanged).  The same helper
    # backs the user-facing nmf.cluster.criteria().
    if (detail == "full" && !base::any(Y.weights == 0)) {
      cc <- .cluster.criteria(Y, B, labels = B.cluster)
      silhouette <- cc$silhouette
      CPCC       <- cc$CPCC
      dist.cor   <- cc$dist.cor
    } else {
      silhouette <- NA; CPCC <- NA; dist.cor <- NA
    }
  } else {
    r2 <- NA; r2.uncentered <- NA; r2.centered <- NA
    sigma <- NA; mae <- NA
    B.prob <- NA; B.cluster <- NA
    B.prob.max.mean <- NA
    X.prob <- NA; X.cluster <- NA
    silhouette <- NA; CPCC <- NA; dist.cor <- NA
    effective.rank <- NA_real_
  }

  base::list(
    r.squared          = r2,
    r.squared.uncentered     = r2.uncentered,
    r.squared.centered = r2.centered,
    sigma     = sigma,
    mae       = mae,
    B.prob    = B.prob,
    B.cluster = B.cluster,
    X.prob    = X.prob,
    X.cluster = X.cluster,
    criterion = base::list(
      B.prob.max.mean     = B.prob.max.mean,
      effective.rank      = effective.rank,
      silhouette = silhouette,
      CPCC       = CPCC,
      dist.cor   = dist.cor
    )
  )
}


#' @title Rank selection diagnostics with graphical output
#' @description
#' \code{nmfkc.rank} provides diagnostic criteria for selecting the rank (\eqn{Q})
#' in NMF with kernel covariates. Three rank-selection measures are computed
#' (R-squared, the effective rank, and the element-wise CV error), and results
#' can be visualized in a plot. Sample-clustering quality (silhouette / CPCC /
#' dist.cor) is no longer part of rank selection; use \code{\link{nmf.cluster.criteria}}
#' on a fitted model for those.
#'
#' By default (\code{save.time = FALSE}), this function also computes the
#' Element-wise Cross-Validation error (Wold's CV Sigma) using \code{\link{nmfkc.ecv}}.
#'
#' The plot explicitly marks the "BEST" rank based on two criteria:
#' \enumerate{
#'   \item **Elbow Method (Red)**: Based on the curvature of the R-squared values (always computed if Q > 2).
#'   \item **Min RMSE (Blue)**: Based on the minimum Element-wise CV Sigma (only if \code{detail="full"}).
#' }
#'
#' @param Y Observation matrix, or a formula (see \code{\link{nmfkc}} for Formula Mode).
#' @param A Covariate matrix. If \code{NULL}, the identity matrix is used.
#'   Ignored when \code{Y} is a formula.
#' @param rank A vector of candidate ranks to be evaluated.
#' @param detail \code{"full"} (default) also runs the element-wise CV
#'   (\code{sigma.ecv}); \code{"fast"} skips it (the plot then shows only
#'   r.squared and eff.rank, and the recommended rank falls back to the
#'   R-squared elbow).
#' @param plot Logical. If \code{TRUE} (default), draws a plot of the diagnostic criteria.
#' @param data A data frame (required when \code{Y} is a formula with column names).
#' @param ... Additional arguments passed to \code{\link{nmfkc}} and \code{\link{nmfkc.ecv}}.
#'   \itemize{
#'     \item \code{Q}: (Deprecated) Alias for \code{rank}.
#'     \item \code{save.time}: (Deprecated) \code{TRUE} maps to \code{detail = "fast"}.
#'   }
#'
#' @return A list containing:
#' \item{rank.best}{The estimated optimal rank. Prioritizes ECV minimum if available, otherwise R-squared Elbow.}
#' \item{criteria}{A data frame containing diagnostic metrics for each rank.
#'   The \code{effective.rank} column gives the effective rank
#'   (\eqn{\exp} of the Shannon entropy of the explained-variance
#'   distribution \eqn{p_k = \mathrm{var}(B_{k\cdot}) / \sum_j \mathrm{var}(B_{j\cdot})},
#'   in \eqn{[1, Q]}); when it plateaus well below the nominal
#'   \code{rank}, the extra factors are not carrying additional
#'   coefficient variance, which suggests an over-specified rank.
#'   The \code{effective.rank.ratio} column is \code{effective.rank /
#'   rank} in \eqn{[0, 1]} (the utilization fraction plotted as
#'   \code{eff.rank} when \code{plot = TRUE}); a peak marks the rank
#'   at which the latent factors carry the most evenly distributed
#'   variance.}
#' @export
#' @references
#' Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
#' effective dimensionality.  \emph{Proc. 15th European Signal Processing
#' Conf. (EUSIPCO)}, 606--610.  (\code{effective.rank})
#' Wold, S. (1978).  Cross-validatory estimation of the number of
#' components in factor and principal components models.
#' \emph{Technometrics}, 20(4), 397--405.
#' \doi{10.1080/00401706.1978.10489693}  (\code{sigma.ecv})
#' @seealso \code{\link{nmfkc}},
#'   \code{\link{nmfkc.ecv}} (element-wise CV, used internally),
#'   \code{\link{nmfkc.bicv}} (block bi-cross-validation),
#'   \code{\link{nmfkc.consensus}} (stability) and
#'   \code{\link{nmfkc.ard}} (Bayesian ARD) for alternative rank criteria.
#' @examples
#' # Example.
#' Y <- t(iris[,-5])
#' # Full run (default)
#' nmfkc.rank(Y, rank=1:4)
#' # Fast run (skip ECV)
#' nmfkc.rank(Y, rank=1:4, detail="fast")

nmfkc.rank <- function(Y, A=NULL, rank=1:2, detail="full", plot=TRUE, data, ...){
  # --- Formula Mode ---
  if (base::inherits(Y, "formula")) {
    resolved <- .nmfkc_resolve_formula(Y, A, base::missing(data), if (!base::missing(data)) data else NULL)
    Y <- resolved$Y
    A <- resolved$A
  }

  extra_args <- list(...)

  # Backward compatibility: Q -> rank, save.time -> detail
  if (!is.null(extra_args$Q)) rank <- extra_args$Q
  if (!is.null(extra_args$save.time) && extra_args$save.time && detail == "full") detail <- "fast"
  Q <- rank
  # ---------------------------------------------
  num_q <- length(Q)
  results_df <- data.frame(
    rank = Q,
    effective.rank = numeric(num_q),
    effective.rank.ratio = numeric(num_q),
    r.squared = numeric(num_q),
    sigma.ecv = numeric(num_q)
  )

  # --- Main loop: per-rank fit -> effective rank + r.squared ---
  # Fitted with detail = "fast": the (O(N^2)) sample-clustering criteria
  # silhouette / CPCC / dist.cor are no longer part of rank selection.
  # For per-fit clustering quality use nmf.cluster.criteria() instead.
  for(q_idx in 1:num_q){
    current_Q <- Q[q_idx]

    extra_args_nmfkc <- extra_args
    extra_args_nmfkc$save.memory <- NULL
    extra_args_nmfkc$save.time <- NULL
    extra_args_nmfkc$detail <- "fast"
    extra_args_nmfkc$Q <- NULL

    nmfkc_args <- c(list(Y = Y, A = A, rank = current_Q), extra_args_nmfkc)
    result <- do.call("nmfkc", nmfkc_args)

    results_df$effective.rank[q_idx] <- result$criterion$effective.rank
    results_df$effective.rank.ratio[q_idx] <-
      result$criterion$effective.rank / current_Q
    results_df$r.squared[q_idx] <- result$r.squared
  }

  # --- Element-wise CV (Wold's CV) ---
  if(detail == "full"){
    ecv_args <- list(Y = Y, A = A, Q = Q)
    extra_args_ecv <- extra_args
    extra_args_ecv$save.time <- NULL
    extra_args_ecv$save.memory <- NULL
    extra_args_ecv$Q <- NULL

    ecv_full_args <- c(ecv_args, extra_args_ecv)

    message("Running Element-wise CV (this may take time)...")
    ecv_res <- do.call("nmfkc.ecv", ecv_full_args)
    results_df$sigma.ecv <- ecv_res$sigma
  } else {
    results_df$sigma.ecv <- NA
  }

  # Best-rank selection + concise 3-criterion plot (shared back-end).
  # criteria now holds only the rank-selection columns; clustering
  # quality is provided separately by nmf.cluster.criteria().
  return(.rank.finish(results_df, plot = plot,
                      main = "Rank Selection Diagnostics"))
}







#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------



#' @title Plot Diagnostics: Original, Fitted, and Residual Matrices as Heatmaps
#' @description
#' This function generates a side-by-side plot of three heatmaps: the original
#' observation matrix Y, the fitted matrix XB (from NMF), and the residual matrix E (Y - XB).
#' This visualization aids in diagnosing whether the chosen rank Q is adequate
#' by assessing if the residual matrix E appears to be random noise.
#'
#' The axis labels (X-axis: Samples, Y-axis: Features) are integrated into the main title of each plot
#' to maximize the plot area, reflecting the compact layout settings.
#'
#' @param Y The original observation matrix (P x N).
#' @param result The result object returned by the nmfkc function.
#' @param fitted.palette A vector of colors for Y and XB heatmaps. Defaults to white-orange-red.
#'   For backward compatibility, \code{Y_XB_palette} is accepted via \code{...}.
#' @param residual.palette A vector of colors for the residuals heatmap. Defaults to blue-white-red.
#'   For backward compatibility, \code{E_palette} is accepted via \code{...}.
#' @param ... Additional graphical parameters passed to the internal image calls.
#'
#' @return NULL. The function generates a plot.
#' @examples
#' Y <- t(iris[1:30, 1:4])
#' result <- nmfkc(Y, rank = 2)
#' nmfkc.residual.plot(Y, result)
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{residuals.nmf}}
#' @export
nmfkc.residual.plot <- function(Y, result,
                                fitted.palette = grDevices::colorRampPalette(c("white", "orange", "red"))(256),
                                residual.palette = grDevices::colorRampPalette(c("blue", "white", "red"))(256), ...){
  extra_rp <- list(...)
  if (!is.null(extra_rp$Y_XB_palette)) fitted.palette <- extra_rp$Y_XB_palette
  if (!is.null(extra_rp$E_palette)) residual.palette <- extra_rp$E_palette
  if (!inherits(result, "nmfkc")) {
    stop("The 'result' argument must be an object of class 'nmfkc'.")
  }
  XB <- result$XB
  if (is.null(XB) || (length(XB) == 1 && is.na(XB))) {
    stop("'result$XB' is not available. The model may have been fitted with save.memory=TRUE.")
  }
  E <- Y - XB
  if(nrow(Y) != nrow(E) || ncol(Y) != ncol(E)){
    stop("Dimension mismatch between Y and result$XB. Check input matrices.")
  }
  old_par <- graphics::par(mfrow = c(1, 3), mar = c(1,0.5,5,0.5) + 0.1)
  on.exit(graphics::par(old_par))
  min_YX <- min(Y, XB, na.rm = TRUE)
  max_YX <- max(Y, XB, na.rm = TRUE)
  max_abs_E <- max(abs(E), na.rm = TRUE)
  min_E <- -max_abs_E
  max_E <- max_abs_E
  graphics::image(t(Y)[, nrow(Y):1],
                  col = fitted.palette,
                  zlim = c(min_YX, max_YX),
                  main = "1. Original Matrix Y\n X-axis: Samples (N), Y-axis: Features (P)",
                  xlab = "",
                  ylab = "",
                  axes = FALSE, ...)
  graphics::box()
  graphics::image(t(XB)[, nrow(XB):1],
                  col = fitted.palette,
                  zlim = c(min_YX, max_YX),
                  main = paste0("2. Fitted Matrix XB (Q=",ncol(result$X),")\n X-axis: Samples (N), Y-axis: Features (P)"),
                  xlab = "",
                  ylab = "",
                  axes = FALSE, ...)
  graphics::box()
  graphics::image(t(E)[, nrow(E):1],
                  col = residual.palette,
                  zlim = c(min_E, max_E),
                  main = "3. Residual Matrix E (Y - XB)\n X-axis: Samples (N), Y-axis: Features (P)",
                  xlab = "",
                  ylab = "",
                  axes = FALSE, ...)
  graphics::box()
}



#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
.nmfkc_dot_digits_from_threshold <- function(threshold) {
  if (!is.finite(threshold) || threshold <= 0) {
    return(2L)
  }
  if (threshold >= 1) {
    return(0L)
  }
  s <- format(threshold, scientific = FALSE, trim = TRUE)
  # Examples: "0.01" -> "01", "0.005" -> "005"
  s_dec <- sub("^0\\.", "", s)
  nchar(s_dec)
}


#' Format coefficient values for DOT edge labels
#'
#' If \code{digits} is \code{NULL}, the function uses the traditional
#' magnitude-based formatting rules.
#' If \code{digits} is provided, the coefficient is formatted with the
#' specified number of decimal places (typically derived from the threshold).
#'
#' @param x Numeric scalar.
#' @param digits Integer or \code{NULL}; number of decimal places.
#' @return Character string suitable for use inside DOT \code{label="..."}.
#' @keywords internal
#' @noRd
.nmfkc_dot_format_coef <- function(x, digits = NULL) {
  if (!is.finite(x)) return("")

  # Default behavior: magnitude-dependent auto formatting
  if (is.null(digits)) {
    ax <- abs(x)
    if (ax == 0) return("0")

    if (ax >= 1) {
      sprintf("%.2f", x)
    } else if (ax >= 1e-1) {
      sprintf("%.3f", x)
    } else if (ax >= 1e-2) {
      sprintf("%.4f", x)
    } else {
      sprintf("%.3e", x)
    }

    # Threshold-based fixed-digit formatting
  } else {
    if (digits <= 0) {
      sprintf("%.0f", x)
    } else {
      fmt <- paste0("%.", digits, "f")
      sprintf(fmt, x)
    }
  }
}


#' Sanitize character vectors for DOT node / cluster IDs
#'
#' Replace characters that are not alphanumeric, underscore, or dot
#' with underscores to generate safe DOT identifiers.
#'
#' @param x Character vector.
#' @return Character vector with sanitized identifiers.
#' @keywords internal
#' @noRd
.nmfkc_dot_sanitize_id <- function(x) {
  gsub("[^[:alnum:]_.]", "_", x, perl = TRUE)
}


#' Generate a standard DOT graph header for NMF-related diagrams
#'
#' Provides a consistent DOT header used by NMF visualization functions.
#'
#' @param graph_name Character; name of the DOT graph.
#' @param rankdir Character; Graphviz rank direction (e.g., "LR", "RL", "TB", "BT").
#' @param fontname Character; default font name used in the graph.
#'
#' @return Character scalar containing DOT header lines.
#' @keywords internal
#' @noRd
.nmfkc_dot_header <- function(graph_name = "NMF_GRAPH",
                              rankdir    = "LR",
                              fontname   = "Arial") {
  paste0(
    "digraph ", graph_name, " {\n",
    "  graph [rankdir=", rankdir, " compound=true];\n",
    "  splines=true; nodesep=0.4; ranksep=0.7; fontname=\"", fontname, "\";\n"
  )
}


#' Generate a DOT subgraph cluster with nodes
#'
#' Internal helper to define a cluster (subgraph) of nodes with shared
#' visual properties.
#' If \code{cluster_style = "none"}, nodes are listed individually without
#' creating a DOT subgraph.
#'
#' @param cluster_id Character; cluster identifier suffix (e.g. "Y", "X", "F").
#' @param title Character; cluster label (ignored if \code{cluster_style = "none"}).
#' @param node_ids Character vector; internal DOT node IDs.
#' @param node_labels Character vector; display labels for nodes.
#' @param shape Character; Graphviz node shape.
#' @param fill Logical; whether to use filled node shapes.
#' @param fillcolor Character; fill color (if \code{fill = TRUE}).
#' @param line_width Numeric; node border width.
#' @param indent Character; indentation prefix for formatting.
#' @param cluster_style Character; boundary style ("rounded", "dashed", "invis", "none").
#' @param cluster_color Character; cluster boundary color.
#' @param cluster_penwidth Numeric; cluster boundary width.
#'
#' @return Character scalar containing DOT code for the cluster.
#' @keywords internal
#' @noRd
.nmfkc_dot_cluster_nodes <- function(cluster_id,
                                     title,
                                     node_ids,
                                     node_labels,
                                     shape      = "box",
                                     fill       = TRUE,
                                     fillcolor  = "white",
                                     line_width = 1.5,
                                     indent     = "  ",
                                     cluster_style    = "rounded",
                                     cluster_color    = "black",
                                     cluster_penwidth = 1.0) {

  if (length(node_ids) == 0L) return("")
  if (length(node_ids) != length(node_labels)) {
    stop("node_ids and node_labels must have the same length.")
  }

  # Node style definition
  if (fill) {
    node_style <- sprintf(
      '%snode [shape=%s, style="filled,rounded", fillcolor="%s", color=black, penwidth=%.1f];\n',
      indent, shape, fillcolor, line_width
    )
  } else {
    node_style <- sprintf(
      '%snode [shape=%s, style="rounded", color=black, penwidth=%.1f];\n',
      indent, shape, line_width
    )
  }

  # Case 1: no cluster → output nodes directly
  if (identical(cluster_style, "none")) {
    scr <- node_style
    for (i in seq_along(node_ids)) {
      scr <- paste0(
        scr,
        sprintf('%s%s [label="%s"];\n', indent, node_ids[i], node_labels[i])
      )
    }
    return(scr)
  }

  # Case 2: cluster with boundary
  scr <- paste0(
    indent, "subgraph cluster_", cluster_id,
    '{label="', title, '"',
    ' style="', cluster_style, '"',
    ' color="', cluster_color, '"',
    ' penwidth=', sprintf("%.1f", cluster_penwidth), ";\n",
    node_style
  )

  for (i in seq_along(node_ids)) {
    scr <- paste0(
      scr,
      sprintf('%s  %s [label="%s"];\n', indent, node_ids[i], node_labels[i])
    )
  }
  paste0(scr, indent, "}\n")
}


#' Scale DOT edge width based on coefficient magnitude
#'
#' Computes an appropriate \code{penwidth} value for Graphviz edges
#' by scaling the coefficient relative to the maximum coefficient in
#' its category.
#'
#' @param value Numeric; coefficient value.
#' @param max_value Numeric; maximum coefficient within its group.
#' @param weight_scale Numeric; global scale factor for edge width.
#' @param min_pw Numeric; minimum penwidth.
#'
#' @return Numeric penwidth value.
#' @keywords internal
#' @noRd
.nmfkc_dot_penwidth <- function(value,
                                max_value,
                                weight_scale = 5,
                                min_pw       = 0.5) {
  if (!is.finite(max_value) || max_value <= 0) return(min_pw)
  if (!is.finite(value)     || value <= 0)     return(min_pw)
  max(min_pw, value * weight_scale / max_value)
}


#' @title Statistical inference for the parameter matrix C (Theta)
#' @description
#' \code{nmfkc.inference} performs statistical inference on the parameter matrix
#' \eqn{C} (\eqn{\Theta}) from a fitted \code{nmfkc} model, conditional on
#' the estimated basis matrix \eqn{\hat{X}}.
#'
#' Under the working model \eqn{Y = X C A + \varepsilon} where
#' \eqn{\varepsilon_{pn} \stackrel{iid}{\sim} N(0, \sigma^2)},
#' inference is conducted via sandwich covariance estimation and
#' one-step wild bootstrap with non-negative projection.
#'
#' @param object An object of class \code{"nmfkc"} returned by \code{\link{nmfkc}}.
#' @param Y Observation matrix (P x N). Must match the data used in \code{nmfkc()}.
#' @param A Covariate matrix (K x N). Default is \code{NULL} (same as identity;
#'   in this case \eqn{B = C} and inference is on \eqn{B} directly).
#' @param wild.bootstrap Logical. If \code{TRUE} (default), performs wild bootstrap
#'   for confidence intervals and bootstrap standard errors. Set to \code{FALSE}
#'   to skip bootstrap (faster, only sandwich SE is computed).
#' @param ... Additional arguments:
#'   \describe{
#'     \item{\code{wild.B}}{Number of bootstrap replicates. Default is 1000.}
#'     \item{\code{wild.seed}}{Seed for bootstrap. Default is 42.}
#'     \item{\code{wild.level}}{Confidence level for bootstrap CI. Default is 0.95.}
#'     \item{\code{sandwich}}{Logical. Use sandwich covariance. Default is \code{TRUE}.}
#'     \item{\code{C.p.side}}{P-value type: \code{"one.sided"} (default) or \code{"two.sided"}.}
#'     \item{\code{cov.ridge}}{Ridge stabilization for information matrix inversion. Default is 1e-8.}
#'     \item{\code{print.trace}}{Logical. If \code{TRUE}, prints progress. Default is \code{FALSE}.}
#'   }
#'
#' @return An object of class \code{c("nmfkc.inference", "nmfkc")}, inheriting all
#' components from the input \code{object}, with additional inference components:
#' \item{sigma2.used}{Estimated \eqn{\sigma^2} used for inference.}
#' \item{C.se}{Sandwich standard errors for \eqn{C} (Q x K matrix).}
#' \item{C.se.boot}{Bootstrap standard errors for \eqn{C} (Q x K matrix).}
#' \item{C.ci.lower}{Lower CI bounds for \eqn{C} (Q x K matrix).}
#' \item{C.ci.upper}{Upper CI bounds for \eqn{C} (Q x K matrix).}
#' \item{coefficients}{Data frame with Estimate, SE, BSE, z, p-value for each element of \eqn{C}.}
#' \item{C.p.side}{P-value type used.}
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{summary.nmfkc.inference}}
#' @references
#' Satoh, K. (2026). Wild Bootstrap Inference for Non-Negative Matrix
#'   Factorization with Random Effects. arXiv:2603.01468.
#'   \url{https://arxiv.org/abs/2603.01468}
#' @export
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(intercept = 1, speed = cars$speed)
#' result <- nmfkc(Y, A, rank = 1)
#' result2 <- nmfkc.inference(result, Y, A)
#' summary(result2)
#'
nmfkc.inference <- function(object, Y, A = NULL,
                            wild.bootstrap = TRUE, ...) {
  if (!inherits(object, "nmfkc")) stop("object must be of class 'nmfkc'")

  extra_args <- base::list(...)
  wild.B      <- if (!is.null(extra_args$wild.B))      extra_args$wild.B      else 500
  wild.seed   <- if (!is.null(extra_args$wild.seed))   extra_args$wild.seed   else 123
  wild.level  <- if (!is.null(extra_args$wild.level))  extra_args$wild.level  else 0.95
  sandwich    <- if (!is.null(extra_args$sandwich))     extra_args$sandwich    else TRUE
  C.p.side    <- if (!is.null(extra_args$C.p.side))    extra_args$C.p.side    else "one.sided"
  cov.ridge   <- if (!is.null(extra_args$cov.ridge))   extra_args$cov.ridge   else 1e-8
  print.trace <- if (!is.null(extra_args$print.trace)) extra_args$print.trace else FALSE

  X <- object$X   # P x Q
  C_mat <- object$C   # Q x K (or Q x N if A is NULL)

  # If A is NULL, use identity (B = C)
  if (is.null(A)) A <- diag(ncol(Y))

  XB <- X %*% C_mat %*% A   # P x N  fitted values
  Q <- ncol(X)
  K <- nrow(A)
  P <- nrow(Y)
  N <- ncol(Y)

  R_C <- Y - XB   # P x N  residuals

  # sigma2 estimate (denominator: PN - QK)
  denom <- max(P * N - Q * K, 1)
  sigma2.used <- sum(R_C^2) / denom

  # Information matrix: I = sigma^{-2} (AA' x X'X)
  XtX <- crossprod(X)       # Q x Q
  AAt <- tcrossprod(A)      # K x K
  Info_core <- kronecker(AAt, XtX)   # QK x QK
  Info <- Info_core / max(sigma2.used, 1e-12)
  Info <- Info + diag(cov.ridge, nrow(Info))

  Hinv <- tryCatch(solve(Info), error = function(e) {
    if (requireNamespace("MASS", quietly = TRUE)) MASS::ginv(Info)
    else stop("Information matrix singular; install MASS package.")
  })

  # Sandwich covariance: V = Hinv J Hinv
  V_sand <- NULL
  if (isTRUE(sandwich)) {
    Xt <- t(X)
    J <- matrix(0, Q * K, Q * K)
    for (n in 1:N) {
      a_n <- A[, n, drop = FALSE]
      r_n <- R_C[, n, drop = FALSE]
      g_n <- Xt %*% r_n
      S_n <- -(g_n %*% t(a_n)) / max(sigma2.used, 1e-12)
      s_n <- as.vector(S_n)
      J <- J + tcrossprod(s_n)
    }
    if (N > 1) J <- (N / (N - 1)) * J   # CR1 correction
    V_sand <- Hinv %*% J %*% Hinv
  }

  C.vec.cov <- if (!is.null(V_sand)) V_sand else Hinv

  # Sandwich SE
  se_vec <- sqrt(pmax(diag(C.vec.cov), 0))
  C.se <- matrix(se_vec, nrow = Q, ncol = K, byrow = FALSE)

  # ---- Wild bootstrap (one-step Newton) ----
  C.se.boot <- NULL
  C.ci.lower <- NULL
  C.ci.upper <- NULL

  if (isTRUE(wild.bootstrap)) {
    set.seed(wild.seed)
    Xt <- t(X)
    score_mat <- matrix(0, Q * K, N)
    for (n in 1:N) {
      a_n <- A[, n, drop = FALSE]
      r_n <- R_C[, n, drop = FALSE]
      g_n <- Xt %*% r_n
      G_n <- -(g_n %*% t(a_n)) / max(sigma2.used, 1e-12)
      score_mat[, n] <- as.vector(G_n)
    }

    C_hat_vec <- as.vector(C_mat)
    C_boot <- matrix(NA_real_, nrow = Q * K, ncol = wild.B)
    for (b in 1:wild.B) {
      w <- stats::rexp(N, rate = 1) - 1   # Exp(1)-centered multiplier
      grad_b <- as.vector(score_mat %*% w)
      c_b <- C_hat_vec - as.vector(Hinv %*% grad_b)
      c_b <- pmax(c_b, 0)   # project onto C >= 0
      C_boot[, b] <- c_b
    }

    # Bootstrap SE
    sd_vec <- apply(C_boot, 1, stats::sd, na.rm = TRUE)
    C.se.boot <- matrix(sd_vec, nrow = Q, ncol = K, byrow = FALSE)

    # Bootstrap CI
    alpha <- 1 - wild.level
    lo <- apply(C_boot, 1, stats::quantile, probs = alpha / 2, na.rm = TRUE, names = FALSE)
    hi <- apply(C_boot, 1, stats::quantile, probs = 1 - alpha / 2, na.rm = TRUE, names = FALSE)
    C.ci.lower <- matrix(lo, nrow = Q, ncol = K, byrow = FALSE)
    C.ci.upper <- matrix(hi, nrow = Q, ncol = K, byrow = FALSE)
  }

  # ---- Coefficients table ----
  Estimate <- as.vector(C_mat)
  SE <- as.vector(C.se)
  BSE <- if (!is.null(C.se.boot)) as.vector(C.se.boot) else rep(NA_real_, length(Estimate))
  z_value <- ifelse(SE > 0, Estimate / SE, NA_real_)

  if (C.p.side == "one.sided") {
    p_value <- ifelse(is.finite(z_value), stats::pnorm(z_value, lower.tail = FALSE), NA_real_)
  } else {
    p_value <- ifelse(is.finite(z_value), 1 - stats::pchisq(z_value^2, df = 1), NA_real_)
  }

  # Row/column labels for C (rows = Basis, cols = Covariate)
  rlabs <- if (!is.null(rownames(C_mat))) rownames(C_mat) else paste0("Basis", 1:Q)
  clabs <- if (!is.null(colnames(C_mat))) colnames(C_mat) else paste0("Cov", 1:K)

  coefficients <- data.frame(
    Basis    = rep(rlabs, times = K),
    Covariate = rep(clabs, each = Q),
    Estimate = Estimate,
    SE       = SE,
    BSE      = BSE,
    z_value  = z_value,
    p_value  = p_value,
    CI_low   = if (!is.null(C.ci.lower)) as.vector(C.ci.lower) else NA_real_,
    CI_high  = if (!is.null(C.ci.upper)) as.vector(C.ci.upper) else NA_real_,
    row.names = NULL, stringsAsFactors = FALSE
  )

  if (print.trace) {
    if (isTRUE(wild.bootstrap)) {
      message("  Inference: sandwich SE + wild bootstrap done.")
    } else {
      message("  Inference: sandwich SE done (wild bootstrap skipped).")
    }
  }

  # Add inference fields to the object
  object$sigma2.used  <- sigma2.used
  object$C.se         <- C.se
  object$C.se.boot    <- C.se.boot
  object$C.ci.lower   <- C.ci.lower
  object$C.ci.upper   <- C.ci.upper
  object$coefficients <- coefficients
  object$C.p.side     <- C.p.side
  class(object) <- c("nmfkc.inference", "nmf.inference", "nmfkc", "nmf")
  return(object)
}


#' @title Summary method for nmfkc.inference objects
#' @description
#' Produces a summary of a fitted NMF model with inference results,
#' including the coefficients table for \eqn{C} (\eqn{\Theta}).
#'
#' @param object An object of class \code{"nmfkc.inference"}.
#' @param ... Additional arguments (currently unused).
#' @return An object of class \code{"summary.nmfkc.inference"}.
#' @seealso \code{\link{nmfkc.inference}}, \code{\link{summary.nmfkc}}
#' @export
summary.nmfkc.inference <- function(object, ...) {
  ans <- summary.nmfkc(object, ...)
  ans$sigma2.used  <- object$sigma2.used
  ans$coefficients <- object$coefficients
  ans$C.p.side     <- object$C.p.side
  class(ans) <- "summary.nmfkc.inference"
  return(ans)
}


#' @title Print method for summary.nmfkc.inference objects
#' @description
#' Prints a formatted summary including the coefficients table.
#' @param x An object of class \code{"summary.nmfkc.inference"}.
#' @param digits Minimum number of significant digits.
#' @param max.coef Maximum coefficient rows to display. Default is 20.
#' @param ... Additional arguments (currently unused).
#' @return Called for its side effect (printing). Returns \code{x} invisibly.
#' @seealso \code{\link{summary.nmfkc.inference}}
#' @export
print.summary.nmfkc.inference <- function(x, digits = max(3L, getOption("digits") - 3L),
                                           max.coef = 20, ...) {
  # Print base summary
  print.summary.nmfkc(x, digits = digits, ...)

  # Print inference section
  cat("Inference (conditional on X):\n")
  cat("  sigma^2:  ", format(x$sigma2.used, digits = digits), "\n")

  # Coefficients table (formatted like lm summary)
  if (!is.null(x$coefficients) && is.data.frame(x$coefficients)) {
    cf <- x$coefficients
    n_total <- nrow(cf)
    rnames <- paste0(cf$Covariate, ":", cf$Basis)

    # Determine which rows to display
    if (n_total <= max.coef) {
      show_idx <- seq_len(n_total)
      truncated <- FALSE
    } else {
      sig_idx <- which(cf$p_value < 0.05)
      if (length(sig_idx) == 0) {
        show_idx <- seq_len(min(max.coef, n_total))
        truncated <- TRUE
      } else if (length(sig_idx) <= max.coef) {
        show_idx <- sig_idx
        truncated <- FALSE
      } else {
        show_idx <- sig_idx[seq_len(max.coef)]
        truncated <- TRUE
      }
    }

    n_sig <- sum(cf$p_value < 0.05, na.rm = TRUE)
    cat(sprintf("\nCoefficients (conditional on X): %d total, %d significant\n",
                n_total, n_sig))
    if (n_total > max.coef) {
      cat(sprintf("  (showing %d significant rows; use res$coefficients for full table)\n",
                  length(show_idx)))
    }

    p_side <- if (!is.null(x$C.p.side)) x$C.p.side else "one.sided"
    p_header <- if (p_side == "one.sided") "Pr(>z)" else "Pr(>|z|)"

    sig_stars <- function(p) {
      ifelse(!is.finite(p), " ",
        ifelse(p < 0.001, "***",
          ifelse(p < 0.01, "**",
            ifelse(p < 0.05, "*",
              ifelse(p < 0.1, ".", " ")))))
    }
    format_pval <- function(p) {
      ifelse(!is.finite(p), "      NA",
        ifelse(p < 2.2e-16, "  <2e-16",
          formatC(p, format = "g", digits = 4, width = 8)))
    }

    est <- formatC(cf$Estimate[show_idx], format = "f", digits = 3, width = 9)
    se  <- formatC(cf$SE[show_idx], format = "f", digits = 3, width = 10)
    bse <- formatC(cf$BSE[show_idx], format = "f", digits = 3, width = 6)
    zv  <- formatC(cf$z_value[show_idx], format = "f", digits = 2, width = 7)
    pv_str <- format_pval(cf$p_value[show_idx])
    stars <- sig_stars(cf$p_value[show_idx])
    show_names <- rnames[show_idx]

    max_lw <- max(nchar(show_names))
    hdr <- sprintf("%s %s %s %s %s %s",
                   formatC("Estimate", width = 9),
                   formatC("Std. Error", width = 10),
                   formatC("(Boot)", width = 6),
                   formatC("z value", width = 7),
                   formatC(p_header, width = 8), "")
    cat(sprintf("%s %s\n", formatC("Cov:Basis", width = max_lw), hdr))
    for (i in seq_along(show_names)) {
      cat(sprintf("%s %s %s %s %s %s %s\n",
                  formatC(show_names[i], width = max_lw),
                  est[i], se[i], bse[i], zv[i], pv_str[i], stars[i]))
    }
    cat("---\n")
    cat("Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1\n")
  }

  cat("\n")
  invisible(x)
}

test_that("workwell", {
  X <- cbind(c(1,0,1),c(0,1,0))
  B <- cbind(c(1,0),c(0,1),c(1,1))
  Y <- X %*% B
  res <- nmfkc::nmfkc(Y,rank=2,epsilon=1e-12,maxit=50000,print.dims = FALSE)
  tolerance <- 1e-3
  expect_equal(Y,res$XB, tolerance = tolerance)
})

## Tests for nmf.cluster.criteria() sample-clustering quality across ranks

test_that("nmf.cluster.criteria builds a per-rank criteria table from a fit list", {
  Y <- t(as.matrix(iris[, 1:4]))
  fits <- lapply(2:4, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
  cc <- nmf.cluster.criteria(fits, Y, plot = FALSE)

  expect_s3_class(cc, "nmf.cluster.criteria")
  expect_equal(cc$criteria$rank, 2:4)
  expect_setequal(names(cc$criteria),
                  c("rank", "silhouette", "CPCC", "dist.cor", "hard"))
  expect_true(all(cc$criteria$hard))                 # non-negative B
  expect_true(all(is.finite(cc$criteria$silhouette)))
  expect_true(all(is.finite(cc$criteria$CPCC)))
  expect_true(all(is.finite(cc$criteria$dist.cor)))
})

test_that("nmf.cluster.criteria accepts a single fit (wrapped)", {
  Y <- t(as.matrix(iris[, 1:4]))
  fit <- nmfkc(Y, Q = 3, print.dims = FALSE)
  cc <- nmf.cluster.criteria(fit, Y, plot = FALSE)
  expect_equal(nrow(cc$criteria), 1L)
  expect_equal(cc$criteria$rank, 3L)
})

test_that("nmf.cluster.criteria sets silhouette NA when the coefficient is signed", {
  set.seed(7)
  Y <- matrix(rnorm(8 * 20), 8, 20)
  A <- rbind(intercept = 1, x = rnorm(20))
  fit <- suppressWarnings(nmfkc.signed(Y, A, rank = 3, maxit = 500))
  skip_if_not(any(fit$B < 0), "coefficient happened to be non-negative")

  cc <- nmf.cluster.criteria(fit, Y, plot = FALSE)
  expect_false(cc$criteria$hard[1])
  expect_true(is.na(cc$criteria$silhouette[1]))
  ## distance-based criteria are still defined for a signed B
  expect_true(is.finite(cc$criteria$CPCC[1]))
  expect_true(is.finite(cc$criteria$dist.cor[1]))
})

test_that("nmf.cluster.criteria matches nmfkc.criterion's values (shared helper)", {
  Y <- t(as.matrix(iris[, 1:4]))
  fit <- nmfkc(Y, Q = 3, print.dims = FALSE)
  cc  <- nmf.cluster.criteria(fit, Y, plot = FALSE)
  crit <- fit$criterion
  expect_equal(cc$criteria$CPCC,       crit$CPCC,       tolerance = 1e-8)
  expect_equal(cc$criteria$dist.cor,   crit$dist.cor,   tolerance = 1e-8)
  expect_equal(cc$criteria$silhouette, crit$silhouette, tolerance = 1e-8)
})

test_that("nmf.cluster.criteria requires Y", {
  Y <- t(as.matrix(iris[, 1:4]))
  fit <- nmfkc(Y, Q = 2, print.dims = FALSE)
  expect_error(nmf.cluster.criteria(fit), "requires the original data")
})

## Tests for nmfkc.kernel.beta.nearest.med() candidates option

test_that("candidates argument controls the beta grid", {
  set.seed(1)
  U <- matrix(stats::rnorm(5 * 50), 5, 50)

  ## default = "7points"
  info7 <- nmfkc.kernel.beta.nearest.med(U)
  expect_length(info7$beta_candidates, 7)
  expect_equal(info7$beta_candidates[4], info7$beta, tolerance = 1e-12)

  ## "4points"
  info4 <- nmfkc.kernel.beta.nearest.med(U, candidates = "4points")
  expect_length(info4$beta_candidates, 4)

  ## custom numeric t-grid
  t_custom <- c(-0.5, 0, 0.5)
  info_c <- nmfkc.kernel.beta.nearest.med(U, candidates = t_custom)
  expect_length(info_c$beta_candidates, 3)
  expect_equal(info_c$beta_candidates,
               info_c$beta * 10^(-2 * t_custom), tolerance = 1e-12)

  ## invalid character -> error
  expect_error(nmfkc.kernel.beta.nearest.med(U, candidates = "bogus"),
               "7points|4points|numeric")

  ## works with Uk supplied as well (7 candidates by default)
  Uk <- U[, 1:5, drop = FALSE]
  infoUk <- nmfkc.kernel.beta.nearest.med(U, Uk = Uk)
  expect_length(infoUk$beta_candidates, 7)
})

## Tests for nmf.cluster.flow() cluster-flow diagram across a sequence of fits

test_that("nmf.cluster.flow returns an N x R cluster table", {
  Y <- t(as.matrix(iris[, 1:4]))
  fits <- lapply(2:5, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
  fl <- nmf.cluster.flow(fits, reference = 2, plot = FALSE)

  expect_equal(dim(fl$clusters), c(ncol(Y), 4L))
  expect_equal(fl$ranks, 2:5)
  expect_equal(fl$labels, as.character(2:5))
  expect_equal(fl$reference, 2L)                 # reference is an index
  expect_equal(colnames(fl$clusters), as.character(2:5))
  ## entries are valid cluster numbers (1..rank at each result)
  for (j in seq_along(fl$ranks)) {
    expect_true(all(fl$clusters[, j] >= 1))
    expect_true(all(fl$clusters[, j] <= fl$ranks[j]))
  }
  ## one colour per individual
  expect_length(fl$colors, ncol(Y))
})

test_that("nmf.cluster.flow keeps the given order (no rank sorting)", {
  Y <- t(as.matrix(iris[, 1:4]))
  fits <- lapply(c(4, 2, 3), function(q) nmfkc(Y, Q = q, print.dims = FALSE))
  fl <- nmf.cluster.flow(fits, plot = FALSE)
  expect_equal(fl$ranks, c(4L, 2L, 3L))          # order preserved
  expect_equal(fl$labels, c("4", "2", "3"))
})

test_that("nmf.cluster.flow honours a custom `names` for the labels", {
  Y <- t(as.matrix(iris[, 1:4]))
  fits <- lapply(c(3, 3, 3), function(q) nmfkc(Y, Q = q, print.dims = FALSE))
  fl <- nmf.cluster.flow(fits, names = c("modelA", "modelB", "modelC"),
                         plot = FALSE)
  expect_equal(fl$labels, c("modelA", "modelB", "modelC"))
  expect_equal(colnames(fl$clusters), c("modelA", "modelB", "modelC"))
  expect_error(nmf.cluster.flow(fits, names = c("a", "b"), plot = FALSE),
               "length\\(fits\\)")
})

test_that("nmf.cluster.flow rejects a single fit and an out-of-range reference", {
  Y <- t(as.matrix(iris[, 1:4]))
  one <- list(nmfkc(Y, Q = 2, print.dims = FALSE))
  expect_error(nmf.cluster.flow(one, plot = FALSE), "at least two")

  fits <- lapply(2:4, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
  expect_error(nmf.cluster.flow(fits, reference = 9, plot = FALSE),
               "result index in 1")
})

# test-formula-mode.R
# Tests for Formula Mode support across nmfkc functions

# --- Helper data ---
# Create a small non-negative dataset for testing
make_test_data <- function() {
  set.seed(42)
  data.frame(
    Y1 = rpois(20, 5),
    Y2 = rpois(20, 10),
    Y3 = rpois(20, 8),
    A1 = abs(rnorm(20, 5, 1)),
    A2 = abs(rnorm(20, 3, 1))
  )
}

# ============================================================
# 1. .nmfkc_parse_formula (internal function)
# ============================================================

test_that(".nmfkc_parse_formula: MODE 1 (data provided) basic", {
  df <- make_test_data()
  res <- nmfkc:::.nmfkc_parse_formula(Y1 + Y2 + Y3 ~ A1 + A2, data = df)

  expect_true(is.matrix(res$Y))
  expect_true(is.matrix(res$A))
  # Y is P x N (3 variables x 20 samples)
  expect_equal(nrow(res$Y), 3)
  expect_equal(ncol(res$Y), 20)
  # A is R x N (2 covariates x 20 samples)
  expect_equal(nrow(res$A), 2)
  expect_equal(ncol(res$A), 20)
  # Column names returned
  expect_equal(res$Y_cols, c("Y1", "Y2", "Y3"))
  expect_equal(res$A_cols, c("A1", "A2"))
})

test_that(".nmfkc_parse_formula: MODE 1 dot notation on RHS", {
  df <- make_test_data()
  res <- nmfkc:::.nmfkc_parse_formula(Y1 + Y2 + Y3 ~ ., data = df)

  expect_equal(nrow(res$A), 2)  # A1 and A2 remain
  expect_equal(res$A_cols, c("A1", "A2"))
})

test_that(".nmfkc_parse_formula: MODE 1 dot notation on LHS", {
  df <- make_test_data()
  res <- nmfkc:::.nmfkc_parse_formula(. ~ A1 + A2, data = df)

  expect_equal(nrow(res$Y), 3)  # Y1, Y2, Y3 remain
})

test_that(".nmfkc_parse_formula: MODE 1 A omitted (standard NMF)", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc:::.nmfkc_parse_formula(Y1 + Y2 + Y3 ~ 0, data = df))

  expect_null(res$A)
  expect_equal(nrow(res$Y), 3)
})

test_that(".nmfkc_parse_formula: MODE 2 (direct matrix eval)", {
  # Both matrices: rows=samples, cols=variables. After t(): variables x samples.
  Y_mat <- matrix(1:6, nrow = 3, ncol = 2)   # 3 samples, 2 variables
  A_mat <- matrix(1:6, nrow = 3, ncol = 2)   # 3 samples, 2 covariates
  res <- nmfkc:::.nmfkc_parse_formula(Y_mat ~ A_mat)

  expect_true(is.matrix(res$Y))
  expect_true(is.matrix(res$A))
  # Mode 2: Y_cols and A_cols are NULL
  expect_null(res$Y_cols)
  expect_null(res$A_cols)
})

# ============================================================
# 2. nmfkc() with formula
# ============================================================

test_that("nmfkc: formula mode matches matrix mode", {
  df <- make_test_data()
  Y_mat <- t(as.matrix(df[, c("Y1", "Y2", "Y3")]))
  A_mat <- t(as.matrix(df[, c("A1", "A2")]))

  set.seed(1)
  res_mat <- suppressMessages(nmfkc::nmfkc(Y_mat, A_mat, rank = 2,
                                           epsilon = 1e-4, maxit = 100,
                                           print.dims = FALSE, seed = 1))
  set.seed(1)
  res_frm <- suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                                           rank = 2, epsilon = 1e-4, maxit = 100,
                                           print.dims = FALSE, seed = 1))

  # Same optimization result
  expect_equal(res_mat$r.squared, res_frm$r.squared, tolerance = 1e-3)
  expect_equal(res_mat$rank, res_frm$rank)
})

test_that("nmfkc: formula mode stores formula.meta", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                                       rank = 2, epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  expect_false(is.null(res$formula.meta))
  expect_true(inherits(res$formula.meta$formula, "formula"))
  expect_equal(res$formula.meta$Y_cols, c("Y1", "Y2", "Y3"))
  expect_equal(res$formula.meta$A_cols, c("A1", "A2"))
})

test_that("nmfkc: formula mode without A (standard NMF)", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ 0, data = df,
                                       rank = 2, epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  expect_false(is.null(res$formula.meta))
  expect_null(res$formula.meta$A_cols)
})

test_that("nmfkc: formula mode rejects negative values", {
  df <- make_test_data()
  df$Y1[1] <- -1
  expect_error(
    suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                                  rank = 2, print.dims = FALSE)),
    "non-negative"
  )
})

test_that("nmfkc: matrix mode formula.meta is NULL", {
  Y <- matrix(rpois(30, 5), nrow = 3)
  res <- suppressMessages(nmfkc::nmfkc(Y, rank = 2, epsilon = 1e-4,
                                       maxit = 100, print.dims = FALSE))
  expect_null(res$formula.meta)
})

# ============================================================
# 3. predict.nmfkc() with newdata
# ============================================================

test_that("predict: newdata works for formula mode models", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                                       rank = 2, epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  new_df <- data.frame(A1 = c(5, 6), A2 = c(3, 4))
  pred <- predict(res, newdata = new_df)

  expect_true(is.matrix(pred))
  expect_equal(nrow(pred), 3)   # P = 3 variables
  expect_equal(ncol(pred), 2)   # 2 new samples
})

test_that("predict: newdata errors for matrix-mode models", {
  Y <- matrix(rpois(30, 5), nrow = 3)
  res <- suppressMessages(nmfkc::nmfkc(Y, rank = 2, epsilon = 1e-4,
                                       maxit = 100, print.dims = FALSE))
  expect_error(
    predict(res, newdata = data.frame(A1 = 1)),
    "formula"
  )
})

test_that("predict: newdata errors when columns are missing", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                                       rank = 2, epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  expect_error(
    predict(res, newdata = data.frame(A1 = c(5, 6))),
    "missing required columns"
  )
})

test_that("predict: newdata + newA gives warning and uses newdata", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                                       rank = 2, epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  new_df <- data.frame(A1 = c(5, 6), A2 = c(3, 4))
  dummy_A <- matrix(1, nrow = 2, ncol = 2)

  expect_warning(
    pred <- predict(res, newA = dummy_A, newdata = new_df),
    "newdata"
  )
  expect_true(is.matrix(pred))
})

test_that("predict: type='class' works with newdata", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                                       rank = 2, epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  new_df <- data.frame(A1 = c(5, 6), A2 = c(3, 4))
  pred <- predict(res, newdata = new_df, type = "class")

  expect_true(is.character(pred))
  expect_equal(length(pred), 2)
})

test_that("predict: backward compatible (newA still works)", {
  df <- make_test_data()
  Y_mat <- t(as.matrix(df[, c("Y1", "Y2", "Y3")]))
  A_mat <- t(as.matrix(df[, c("A1", "A2")]))

  res <- suppressMessages(nmfkc::nmfkc(Y_mat, A_mat, rank = 2,
                                       epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  newA_mat <- matrix(abs(rnorm(4)), nrow = 2, ncol = 2)
  pred <- predict(res, newA = newA_mat)

  expect_true(is.matrix(pred))
  expect_equal(ncol(pred), 2)
})

# ============================================================
# 4. CV functions with formula
# ============================================================

test_that("nmfkc.cv: formula mode works", {
  df <- make_test_data()
  res <- suppressWarnings(suppressMessages(
    nmfkc::nmfkc.cv(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                     rank = 2, nfolds = 3, maxit = 100,
                     print.dims = FALSE, print.trace = FALSE)
  ))
  expect_true(is.list(res))
  expect_true(!is.null(res$objfunc))
  expect_true(is.finite(res$objfunc))
})

test_that("nmfkc.ecv: formula mode works", {
  df <- make_test_data()
  res <- suppressWarnings(suppressMessages(
    nmfkc::nmfkc.ecv(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                      rank = 1:2, nfolds = 3, maxit = 100,
                      print.dims = FALSE, print.trace = FALSE)
  ))
  expect_true(is.list(res))
  expect_equal(length(res$objfunc), 2)
})

test_that("nmfkc.rank: formula mode works", {
  df <- make_test_data()
  res <- suppressWarnings(suppressMessages(
    nmfkc::nmfkc.rank(Y1 + Y2 + Y3 ~ A1 + A2, data = df,
                       rank = 1:2, detail = "fast",
                       plot = FALSE, maxit = 100,
                       print.dims = FALSE, print.trace = FALSE)
  ))
  expect_true(is.list(res))
  expect_true(!is.null(res$rank.best))
})

# ============================================================
# 5. Edge cases
# ============================================================

test_that("nmfkc: single Y variable formula", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc::nmfkc(Y1 ~ A1 + A2, data = df,
                                       rank = 1, epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  expect_equal(nrow(res$X), 1)  # P = 1
})

test_that("predict: newdata with formula mode (no A = standard NMF) errors", {
  df <- make_test_data()
  res <- suppressMessages(nmfkc::nmfkc(Y1 + Y2 + Y3 ~ 0, data = df,
                                       rank = 2, epsilon = 1e-4, maxit = 100,
                                       print.dims = FALSE))
  expect_error(
    predict(res, newdata = data.frame(A1 = 5)),
    "no A columns"
  )
})

# nmfkc.ar.R — AR/VAR related functions
# nmfkc.ar, nmfkc.ar.predict, nmfkc.ar.degree.cv, nmfkc.ar.stationarity, nmfkc.ar.DOT

#' @title Construct observation and covariate matrices for a vector autoregressive model
#' @description
#' \code{nmfkc.ar} generates the observation matrix and covariate matrix
#' corresponding to a specified autoregressive lag order.
#'
#' If the input \code{Y} is a \code{ts} object, its time properties are preserved
#' in the \code{"tsp_info"} attribute, adjusted for the lag.
#' Additionally, the column names of \code{Y} and \code{A} are set to the corresponding time points.
#'
#' @param Y An observation matrix (P x N) or a \code{ts} object.
#'   If \code{Y} is a \code{ts} object (typically N x P), it is automatically transposed to match the (P x N) format.
#' @param degree The lag order of the autoregressive model. The default is 1.
#' @param intercept Logical. If TRUE (default), an intercept term is added to the covariate matrix.
#'
#' @return A list containing:
#' \item{Y}{Observation matrix (P x N_A) used for NMF. Includes adjusted \code{"tsp_info"} attribute and time-based column names.}
#' \item{A}{Covariate matrix (R x N_A) constructed according to the specified lag order. Includes adjusted \code{"tsp_info"} attribute and time-based column names.}
#' \item{A.columns}{Index matrix used to generate \code{A}.}
#' \item{degree.max}{Maximum lag order.}
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfkc.ar.degree.cv}}, \code{\link{nmfkc.ar.stationarity}}, \code{\link{nmfkc.ar.DOT}}
#' @references
#' Satoh, K. (2025). Applying non-negative matrix factorization with covariates
#'   to multivariate time series data as a vector autoregression model.
#'   \emph{Japanese Journal of Statistics and Data Science}. arXiv:2501.17446.
#'   \doi{10.1007/s42081-025-00314-0}
#' @examples
#' # Example using AirPassengers (ts object)
#' d <- AirPassengers
#' ar_data <- nmfkc.ar(d, degree = 2)
#' dim(ar_data$Y)
#' dim(ar_data$A)
#'
#' # Example using matrix input
#' Y <- matrix(1:20, nrow = 2)
#' ar_data <- nmfkc.ar(Y, degree = 1)
#' ar_data$degree.max
#'
#' @export
nmfkc.ar <- function(Y, degree=1, intercept=TRUE){

  # --- 1. Time Series Object Handling & Standardization ---
  # Initialize basic tsp info
  tsp_start <- 1
  tsp_freq <- 1

  # [FIX 1] Record if input was originally a ts object
  is_ts_object <- stats::is.ts(Y)

  if(is_ts_object){
    # Capture original time properties BEFORE transforming Y
    y_tsp_orig <- stats::tsp(Y)
    tsp_start <- y_tsp_orig[1]
    tsp_freq <- y_tsp_orig[3]

    # Standard 'ts' objects are Time(rows) x Variables(cols).
    # 'nmfkc' requires Variables(rows) x Time(cols).
    Y <- t(as.matrix(Y))

  } else {
    # Handle vector input (not ts) -> 1 x N matrix
    if(is.vector(Y)) Y <- matrix(Y, nrow=1)
    if(!is.matrix(Y)) Y <- as.matrix(Y)

    # If Y is a regular matrix, we assume it is already Variables x Time.
    # Default tsp (Start=1, Freq=1) is used.
  }

  # --- 2. NA Check ---
  if(any(is.na(Y))){
    stop("Y contains missing values (NA). In NMF-VAR, lagged Y is used to construct the covariate matrix A, which cannot contain missing values. Please impute Y before calling nmfkc.ar().")
  }

  P <- nrow(Y) # Features/Variables
  N <- ncol(Y) # Time points

  # --- 3. Degree Validity Check ---
  if (degree < 1) {
    stop("The 'degree' (lag order) must be 1 or greater.")
  }
  if (degree >= N) {
    stop(paste0("The 'degree' (", degree, ") must be strictly less than the number of time points in Y (", N, ")."))
  }

  # Length of the time series after lagging (N_A)
  N_A <- N - degree
  t_start_idx <- degree + 1

  # --- 4. Calculate Adjusted tsp for Y and A ---
  tsp_start_new <- tsp_start + (degree / tsp_freq)
  tsp_end_new <- tsp_start_new + (N_A - 1) / tsp_freq
  y_tsp_new <- c(tsp_start_new, tsp_end_new, tsp_freq)

  # --- 5. Generate Time Sequence for Column Names [FIXED] ---
  # Determine correct column names
  if (!is_ts_object && !is.null(colnames(Y))) {
    # Case A: Not a ts object AND has existing column names.
    # Preserve existing names by slicing them.
    time_names <- colnames(Y)[t_start_idx : N]
  } else {
    # Case B: Is a ts object OR no column names exist.
    # Generate numeric time sequence based on tsp info.
    time_seq <- seq(from = tsp_start_new, by = 1/tsp_freq, length.out = N_A)
    time_names <- as.character(round(time_seq, 5))
  }

  # --- 6. A.columns.index ---
  A.columns.index <- matrix(0, nrow = degree, ncol = N_A)
  for(i in 1:degree) {
    A.columns.index[i, ] <- (t_start_idx - i) : (N - i)
  }

  # --- 7. A: Construct covariate matrix A ---
  A_data <- Y[, as.vector(A.columns.index)]
  A <- matrix(A_data, nrow = P * degree, ncol = N_A, byrow = FALSE)

  # Set row labels
  if(is.null(rownames(Y))) rownames(Y) <- 1:P
  label <- unlist(lapply(1:degree, function(i) paste0(rownames(Y), "_", i)))
  rownames(A) <- label

  # Add Intercept
  if(intercept){
    A <- rbind(A, matrix(1, nrow=1, ncol=ncol(A)))
    rownames(A) <- c(label,"(Intercept)")
  }

  # --- 8. Ya (Non-lagged Observation Matrix) ---
  Ya <- Y[, t_start_idx : N, drop=FALSE]

  # --- 9. Set Column Names ---
  colnames(Ya) <- time_names
  colnames(A) <- time_names

  # --- 10. Metadata Calculation and Attribute Storage ---
  degree.max <- min(ncol(Ya), floor(10*log10(ncol(Ya))))

  # Store AR parameters
  attr(A, "degree") <- degree
  attr(A, "intercept") <- intercept
  attr(A, "function.name") <- "nmfkc.ar"

  # Store Adjusted Time Properties
  attr(A, "tsp_info") <- y_tsp_new
  attr(Ya, "tsp_info") <- y_tsp_new

  list(Y = Ya,
       A = A,
       A.columns = A.columns.index,
       degree.max = degree.max)
}




#' @title Forecast future values for NMF-VAR model
#' @description
#' \code{nmfkc.ar.predict} computes multi-step-ahead forecasts for a fitted NMF-VAR model
#' using recursive forecasting.
#'
#' If the fitted model contains time series property information (from \code{nmfkc.ar}),
#' the forecasted values will have appropriate time-based column names.
#'
#' @param x An object of class \code{nmfkc} (the fitted model).
#' @param Y The historical observation matrix used for fitting (or at least the last \code{degree} columns).
#' @param degree Optional integer. Lag order (D). If \code{NULL} (default), it is inferred
#'   from \code{x$A.attr} (when available) or from the dimensions of \code{x$C}.
#' @param n.ahead Integer (>=1). Number of steps ahead to forecast.
#'
#' @return A list with components:
#' \item{pred}{A \eqn{P \times n.ahead} matrix of predicted values. Column names are future time points if time information is available.}
#' \item{time}{A numeric vector of future time points corresponding to the columns of \code{pred}.}
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfkc.ar}}
#' @examples
#' # Forecast AirPassengers
#' d <- AirPassengers
#' ar_data <- nmfkc.ar(d, degree = 2)
#' result <- nmfkc(ar_data$Y, ar_data$A, rank = 1)
#' pred <- nmfkc.ar.predict(result, Y = matrix(d, nrow = 1), degree = 2, n.ahead = 3)
#' pred$pred
#'
#' @export
nmfkc.ar.predict <- function(x, Y, degree = NULL, n.ahead = 1){
  # --- Basic checks ---
  if (!inherits(x, "nmfkc")) stop("Argument 'x' must be an object of class 'nmfkc'.")
  if (is.vector(Y)) Y <- matrix(Y, nrow = 1)
  if (!is.matrix(Y)) Y <- as.matrix(Y)
  if (!is.numeric(Y)) stop("'Y' must be numeric.")
  if (length(n.ahead) != 1 || !is.finite(n.ahead) || n.ahead < 1) {
    stop("'n.ahead' must be a positive integer.")
  }
  n.ahead <- as.integer(n.ahead)

  # Fit objects & dims
  Xfit <- x$X
  Cfit <- x$C
  if (is.null(Xfit) || is.null(Cfit)) stop("The fitted object 'x' must contain 'X' and 'C'.")
  P_fit <- nrow(Xfit)
  Q_fit <- ncol(Xfit)
  R_fit <- ncol(Cfit)

  P <- nrow(Y)
  if (P != P_fit) {
    stop(sprintf("Row dimension mismatch: nrow(Y)=%d but nrow(x$X)=%d.", P, P_fit))
  }

  # --- Degree & intercept detection ---
  A.attributes <- x$A.attr
  has_meta <- !is.null(A.attributes) &&
    !is.null(A.attributes$`function.name`) &&
    A.attributes$`function.name` == "nmfkc.ar"

  # 1) user-specified degree has the highest priority
  if (!is.null(degree)) {
    if (length(degree) != 1 || !is.finite(degree) || degree < 1) {
      stop("'degree' must be a positive integer.")
    }
    degree_final <- as.integer(degree)
    # intercept: if metadata available, trust it; otherwise infer by dimension
    if (has_meta && !is.null(A.attributes$intercept)) {
      has_intercept_final <- isTRUE(A.attributes$intercept)
    } else {
      # Infer from C dimensions: R_fit == P*D (+1 if intercept)
      if ((R_fit - 1) %% P_fit == 0 && (R_fit - 1) / P_fit >= 1) {
        # ambiguous if user degree doesn't match dims; check and warn/stop
        if ((R_fit - 1) / P_fit == degree_final) {
          has_intercept_final <- TRUE
        } else if (R_fit %% P_fit == 0 && R_fit / P_fit == degree_final) {
          has_intercept_final <- FALSE
        } else {
          stop(sprintf("Dimension mismatch for supplied degree=%d with C (ncol=%d).", degree_final, R_fit))
        }
      } else if (R_fit %% P_fit == 0 && R_fit / P_fit >= 1) {
        if (R_fit / P_fit == degree_final) {
          has_intercept_final <- FALSE
        } else {
          stop(sprintf("Dimension mismatch for supplied degree=%d with C (ncol=%d).", degree_final, R_fit))
        }
      } else {
        stop("Could not reconcile 'degree' with the dimensions of 'C'.")
      }
    }
  } else if (has_meta) {
    degree_final <- as.integer(A.attributes$degree)
    has_intercept_final <- isTRUE(A.attributes$intercept)
  } else {
    # 3) fall back to dimension inference from C
    if ((R_fit - 1) %% P_fit == 0 && (R_fit - 1) / P_fit >= 1) {
      has_intercept_final <- TRUE
      degree_final <- as.integer((R_fit - 1) / P_fit)
    } else if (R_fit %% P_fit == 0 && R_fit / P_fit >= 1) {
      has_intercept_final <- FALSE
      degree_final <- as.integer(R_fit / P_fit)
    } else {
      stop("Could not infer lag order and intercept status. Please supply 'degree'.")
    }
  }

  # --- Data sufficiency check ---
  if (ncol(Y) < degree_final) {
    stop(sprintf("Not enough historical data in Y (%d columns) for degree=%d.", ncol(Y), degree_final))
  }

  # --- Forecast loop ---
  preds <- matrix(0, nrow = P, ncol = n.ahead)
  rownames(preds) <- rownames(Y)
  colnames(preds) <- paste0("t+", seq_len(n.ahead)) # Default names

  current_Y <- Y
  # Expected length of newA
  expected_R <- P_fit * degree_final + if (has_intercept_final) 1L else 0L
  if (expected_R != R_fit) {
    stop(sprintf("Internal check failed: expected R=%d (from degree/intercept) but ncol(C)=%d.", expected_R, R_fit))
  }

  for (step in seq_len(n.ahead)) {
    # Build lag vector: [Y_t-1; Y_t-2; ...; Y_t-degree] stacked by rows (P each)
    a_vec <- numeric(P * degree_final + if (has_intercept_final) 1L else 0L)
    pos <- 0L
    for (k in 1:degree_final) {
      col_idx <- ncol(current_Y) - k + 1L
      yk <- current_Y[, col_idx, drop = FALSE]
      a_vec[(pos + 1L):(pos + P)] <- as.numeric(yk)
      pos <- pos + P
    }
    if (has_intercept_final) a_vec[length(a_vec)] <- 1

    # Safety check for newA length
    if (length(a_vec) != R_fit) {
      stop(sprintf("Length of constructed lag vector (%d) != ncol(C) (%d).", length(a_vec), R_fit))
    }

    newA <- matrix(a_vec, ncol = 1)
    pred_step <- Xfit %*% (Cfit %*% newA)  # (P x Q) %*% (Q x R) %*% (R x 1) = (P x 1)
    preds[, step] <- as.numeric(pred_step)
    current_Y <- cbind(current_Y, pred_step)
  }

  # --- Time inference logic ---
  future_times <- NULL

  # Strategy 1: Use "tsp_info" from attributes (Robust method)
  if (has_meta && !is.null(A.attributes$tsp_info)) {
    # tsp_info = c(start, end, freq)
    t_info <- A.attributes$tsp_info
    last_time <- t_info[2]
    freq <- t_info[3]

    # If user passed a subset Y (not the full training Y), last_time in attribute might be old.
    # We should check if colnames(Y) can be parsed to sync with Y's actual end time.
    # However, if Y is just the tail of training data, we can try to infer dt from freq.

    if (freq <= 0) stop("'freq' in tsp_info must be positive.")
    dt <- 1 / freq

    # Try to find the last time point of current Y to start prediction from
    last_time_Y <- NULL
    if (!is.null(colnames(Y))) {
      num_cols <- suppressWarnings(as.numeric(colnames(Y)))
      if (!any(is.na(num_cols))) {
        last_time_Y <- utils::tail(num_cols, 1)
      }
    }

    # If Y colnames are not numeric, fallback to the model's last_time (assuming prediction continues from training)
    # OR if Y seems to be the full training set.
    if (is.null(last_time_Y)) {
      start_time_pred <- last_time + dt
    } else {
      start_time_pred <- last_time_Y + dt
    }

    future_times <- seq(from = start_time_pred, by = dt, length.out = n.ahead)

  }
  # Strategy 2: Infer from numeric colnames of Y (Fallback method)
  else if (!is.null(colnames(Y))) {
    times <- suppressWarnings(as.numeric(colnames(Y)))
    if (!any(is.na(times)) && length(times) >= 2) {
      n_check <- min(length(times), 10L)
      recent_times <- utils::tail(times, n_check)
      diffs <- diff(recent_times)
      if (length(diffs) > 0) {
        tol <- 0.01 * abs(mean(diffs)) + 1e-8  # ~1% tolerance
        if (max(diffs) - min(diffs) < tol) {
          dt <- mean(diffs)
          last_time <- utils::tail(times, 1)
          future_times <- last_time + seq_len(n.ahead) * dt
        }
      }
    }
  }

  if (!is.null(future_times)) {
    # Rounding to keep colnames clean (consistent with nmfkc.ar)
    time_names <- as.character(round(future_times, 5))
    colnames(preds) <- time_names
  }

  return(list(pred = preds, time = future_times))
}










#' @title Optimize lag order for the autoregressive model
#' @description
#' \code{nmfkc.ar.degree.cv} selects the optimal lag order for an autoregressive model
#' by applying cross-validation over candidate degrees.
#'
#' This function accepts both standard matrices (Variables x Time) and \code{ts} objects
#' (Time x Variables). \code{ts} objects are automatically transposed internally.
#'
#' @param Y Observation matrix \eqn{Y(P,N)} or a \code{ts} object.
#' @param rank Rank of the basis matrix. For backward compatibility,
#'   \code{Q} is accepted via \code{...}.
#' @param degree A vector of candidate lag orders to be evaluated.
#' @param intercept Logical. If TRUE (default), an intercept is added to the covariate matrix.
#' @param plot Logical. If TRUE (default), a plot of the objective function values is drawn.
#' @param ... Additional arguments passed to \code{nmfkc.cv}.
#'
#' @return A list with components:
#' \item{degree}{The lag order that minimizes the cross-validation objective function.}
#' \item{degree.max}{Maximum recommended lag order, computed as \eqn{10 \log_{10}(N)}
#'   following the \code{ar} function in the \pkg{stats} package.}
#' \item{objfunc}{Objective function values for each candidate lag order.}
#' @seealso \code{\link{nmfkc.ar}}, \code{\link{nmfkc.cv}}
#' @export
#' @examples
#' # Example using ts object directly
#' d <- AirPassengers
#'
#' # Selection of degree (using ts object)
#' # Note: Y is automatically transposed if it is a ts object
#' nmfkc.ar.degree.cv(Y=d, rank=1, degree=11:14)

nmfkc.ar.degree.cv <- function(Y, rank=1, degree=1:2, intercept=TRUE, plot=TRUE, ...){
  extra_cv <- list(...)
  if (!is.null(extra_cv$Q)) rank <- extra_cv$Q
  Q <- rank

  # --- 1. Time Series Object Handling & Standardization ---
  # Ensure Y is (P x N) matrix before processing
  if(stats::is.ts(Y)){
    # ts objects are Time x Var (N x P) -> Transpose to Var x Time (P x N)
    Y <- t(as.matrix(Y))
  } else {
    if(is.vector(Y)) Y <- matrix(Y, nrow=1)
    if(!is.matrix(Y)) Y <- as.matrix(Y)
  }

  # --- 2. Main Processing ---
  # Remove backward-compat aliases from extra_args to avoid duplicate argument errors
  extra_args <- extra_cv[!names(extra_cv) %in% c("Q")]
  objfuncs <- numeric(length(degree))
  success_status <- logical(length(degree))

  for(i in seq_along(degree)){
    start.time <- Sys.time()
    current_degree <- degree[i]
    message(paste0("degree=", current_degree, "..."), appendLF = FALSE)

    res <- tryCatch({
      # 1) build lagged Y & A
      # Y is already standardized to (P x N), so nmfkc.ar handles it as a matrix
      a <- nmfkc.ar(Y = Y, degree = current_degree, intercept = intercept)

      # 2) prepare args for CV (block CV for time series)
      main_args <- list(Y = a$Y, A = a$A, Q = Q)
      all_args  <- c(extra_args, main_args, list(shuffle = FALSE))

      # 3) run CV
      result.cv <- suppressMessages(do.call("nmfkc.cv", all_args))

      # 4) success payload
      list(ok = TRUE, obj = result.cv$objfunc, err = NULL)
    }, error = function(e){
      warning(paste0("Skipping degree=", current_degree, " due to error: ", e$message), call. = FALSE)
      list(ok = FALSE, obj = NA_real_, err = e)
    })

    objfuncs[i] <- res$obj
    success_status[i] <- res$ok

    end.time <- Sys.time()
    diff.time <- difftime(end.time, start.time, units = "sec")
    diff.time.st <- ifelse(diff.time <= 180, paste0(round(diff.time, 1), "sec"),
                           paste0(round(diff.time/60, 1), "min"))
    message(if (res$ok) diff.time.st else "Skipped (Error)")
  }

  valid_indices <- which(success_status)
  if(length(valid_indices) == 0) stop("Cross-validation failed for all candidate degrees.")

  i0 <- valid_indices[which.min(objfuncs[valid_indices])]
  best.degree <- degree[i0]

  # Calculate degree.max based on the standardized Y (ncol = Time points N)
  degree.max <- min(ncol(Y), floor(10 * log10(ncol(Y))))

  if(plot){
    plot(degree, objfuncs, type = "l", col = 2,
         xlab = paste0("degree (max=", degree.max, ")"),
         ylab = "objfunc")
    graphics::points(degree[valid_indices], objfuncs[valid_indices], cex = 1, col = 2)
    graphics::points(degree[i0], objfuncs[i0], cex = 3, col = 2)
    graphics::text(degree, objfuncs, degree, pos = 3)
  }

  names(objfuncs) <- degree
  list(degree = best.degree, degree.max = degree.max, objfunc = objfuncs)
}









#' @title Check stationarity of an NMF-VAR model
#' @description
#' \code{nmfkc.ar.stationarity} assesses the dynamic stability of a VAR model
#' by computing the spectral radius of its companion matrix.
#' It returns both the spectral radius and a logical indicator of stationarity.
#'
#' @param x The return value of \code{nmfkc} for a VAR model.
#'
#' @return A list with components:
#' \item{spectral.radius}{Numeric. The spectral radius of the companion matrix. A value less than 1 indicates stationarity.}
#' \item{stationary}{Logical. \code{TRUE} if the spectral radius is less than 1 (i.e., the system is stationary), \code{FALSE} otherwise.}
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfkc.ar}}
#' @examples
#' # Check stationarity of fitted AR model
#' d <- AirPassengers
#' ar_data <- nmfkc.ar(d, degree = 2)
#' result <- nmfkc(ar_data$Y, ar_data$A, rank = 1)
#' nmfkc.ar.stationarity(result)
#'
#' @export

nmfkc.ar.stationarity <- function(x){
  if (!inherits(x, "nmfkc")) stop("'x' must be an object of class 'nmfkc'.")
  if (is.null(x$X) || is.null(x$C)) stop("'x' must contain 'X' and 'C' components (from nmfkc.ar).")
  X <- x$X  # P × Q
  Theta <- x$C  # Q × (P * D [+1] )
  P <- nrow(X)
  Q <- ncol(X)
  if (P == 0 || Q == 0) stop("'X' must have positive dimensions.")
  total_cols <- ncol(Theta)
  has_intercept <- (total_cols - 1) %% P == 0
  D <- if (has_intercept) (total_cols - 1) %/% P else total_cols %/% P
  if (D <= 0) stop("Cannot determine lag order D from dimensions of 'C'.")
  Theta_lags <- if (has_intercept) Theta[, 1:(total_cols - 1), drop = FALSE] else Theta
  Xi_list <- lapply(1:D, function(d){
    cols <- ((d - 1) * P + 1):(d * P)
    X %*% Theta_lags[, cols]})
  companion_matrix <- matrix(0, nrow = P * D, ncol = P * D)
  for (d in 1:D)companion_matrix[1:P, ((d - 1) * P + 1):(d * P)] <- Xi_list[[d]]
  if (D > 1)companion_matrix[(P + 1):(P * D), 1:(P * (D - 1))] <- diag(P * (D - 1))
  rho <- max(Mod(eigen(companion_matrix)$values))
  return(list(spectral.radius = rho, stationary = (rho < 1)))
}


#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------




############################################################
## nmfkc.ar.DOT  (Graphviz visualization for NMF-AR / VAR models)
############################################################

#' Generate a Graphviz DOT Diagram for NMF-AR / NMF-VAR Models
#'
#' @description
#' Produces a Graphviz DOT script for visualizing autoregressive
#' NMF-with-covariates models constructed via \code{nmfkc.ar} + \code{nmfkc}.
#'
#' The diagram displays three types of directed relationships:
#' \itemize{
#'   \item Lagged predictors: \eqn{T_{t-k} \rightarrow X},
#'   \item Current latent factors: \eqn{X \rightarrow T_t},
#'   \item Optional intercept effects: \code{Const -> X}.
#' }
#'
#' Importantly, *no direct edges from lagged variables to current outputs*
#' (\eqn{T_{t-k} \rightarrow T_t}) are drawn, in accordance with the NMF-AR
#' formulation.
#'
#' Each block of lagged variables is displayed in its own DOT subgraph
#' (e.g., “T-1”, “T-2”, ...), while latent factor nodes and current-time
#' outputs are arranged in separate clusters.
#'
#' @param result A fitted \code{nmfkc} object representing the AR model.
#'   Must contain matrices \code{X} and \code{C}.
#' @param degree Maximum AR lag to visualize.
#' @param intercept Logical; if \code{TRUE}, draws intercept nodes for
#'   columns named "(Intercept)" in matrix \code{C}.
#'   Default is \code{TRUE} when an intercept column is detected in \code{C},
#'   \code{FALSE} otherwise (auto-detected).
#' @param threshold Minimum coefficient magnitude required to draw an edge.
#' @param rankdir Graphviz rank direction (e.g., \code{"RL"}, \code{"LR"}, \code{"TB"}).
#' @param fill Logical; whether nodes are filled with color.
#' @param weight_scale_xy Scaling factor for edges \eqn{X \rightarrow T}.
#' @param weight_scale_lag Scaling factor for lagged edges \eqn{T-k \rightarrow X}.
#' @param weight_scale_int Scaling factor for intercept edges.
#' @param hide.isolated Logical. If \code{TRUE} (default), Y nodes that have no
#'   edges at or above \code{threshold} are excluded from the graph.
#'
#' @return A character string representing a Graphviz DOT file.
#' @seealso \code{\link{nmfkc.ar}}, \code{\link{nmfkc}}, \code{\link{plot.nmfkc.DOT}}
#' @examples
#' d <- AirPassengers
#' ar_data <- nmfkc.ar(d, degree = 2)
#' result <- nmfkc(ar_data$Y, ar_data$A, rank = 1)
#' dot <- nmfkc.ar.DOT(result, degree = 2)
#' cat(dot)
#'
#' @export
nmfkc.ar.DOT <- function(result,
                         degree    = 1,
                         intercept = any(colnames(result$C) == "(Intercept)"),
                         threshold = 0.1,
                         rankdir   = "RL",
                         fill      = TRUE,
                         weight_scale_xy  = 5,
                         weight_scale_lag = 5,
                         weight_scale_int = 3,
                         hide.isolated   = TRUE) {

  ## -------------------------------------------------------------
  ## Extract required AR components
  ## -------------------------------------------------------------
  X     <- result$X
  C_raw <- result$C

  if (is.null(X) || is.null(C_raw)) {
    stop("result must contain matrices X and C.")
  }

  C <- as.matrix(C_raw)

  ## Ensure node labels exist
  if (is.null(rownames(X))) rownames(X) <- paste0("Y", seq_len(nrow(X)))
  if (is.null(colnames(X))) colnames(X) <- paste0("X", seq_len(ncol(X)))

  if (is.null(colnames(C))) {
    colnames(C)     <- paste0("A_", seq_len(ncol(C)))
    colnames(C_raw) <- colnames(C)
  }

  X_lab <- colnames(X)
  Y_lab <- rownames(X)
  C_lab <- colnames(C_raw)

  X_ids <- .nmfkc_dot_sanitize_id(X_lab)
  Y_ids <- .nmfkc_dot_sanitize_id(Y_lab)
  C_ids <- .nmfkc_dot_sanitize_id(colnames(C))

  ## Helper: strip numeric suffix _k for lag grouping
  base_from_name <- function(nm) sub("_([0-9]+)$", "", nm)

  has_intercept_col <- any(C_lab == "(Intercept)")

  ## -------------------------------------------------------------
  ## Determine lag structure
  ## -------------------------------------------------------------
  A_labels_raw <- unique(base_from_name(C_lab))
  A_labels_raw <- setdiff(A_labels_raw, "(Intercept)")

  A_per_lag <- length(A_labels_raw)
  if (A_per_lag == 0L) {
    stop("Unable to determine covariate base names for lag separation.")
  }

  total_cols        <- ncol(C)
  total_cols_no_int <- if (has_intercept_col) total_cols - 1L else total_cols
  Dmax              <- floor(total_cols_no_int / A_per_lag)

  if (Dmax < 1L) {
    stop("Insufficient columns in C to construct at least one lag block.")
  }

  D <- min(Dmax, as.integer(degree))

  ## -------------------------------------------------------------
  ## Compute edge width scaling maxima
  ## -------------------------------------------------------------
  max_X   <- suppressWarnings(max(X,   na.rm = TRUE))
  max_C   <- suppressWarnings(max(C[, C_lab != "(Intercept)"], na.rm = TRUE))
  max_int <- if (has_intercept_col) {
    suppressWarnings(max(C[, C_lab == "(Intercept)"], na.rm = TRUE))
  } else {
    NA_real_
  }

  pw     <- .nmfkc_dot_penwidth
  digits <- .nmfkc_dot_digits_from_threshold(threshold)
  fmtc   <- function(v) .nmfkc_dot_format_coef(v, digits)

  ## -------------------------------------------------------------
  ## Filter isolated Y nodes (hide.isolated)
  ## -------------------------------------------------------------
  idx_Y <- seq_along(Y_lab)
  if (isTRUE(hide.isolated)) {
    used_Y <- apply(X, 1L, function(row) any(row >= threshold, na.rm = TRUE))
    idx_Y <- which(used_Y)
  }

  ## -------------------------------------------------------------
  ## DOT header
  ## -------------------------------------------------------------
  scr <- .nmfkc_dot_header(graph_name = "NMF_AR", rankdir = rankdir)

  ## -------------------------------------------------------------
  ## Current-time output cluster (T)
  ## -------------------------------------------------------------
  scr <- paste0(
    scr,
    '\n  // Current-time outputs (T)\n',
    .nmfkc_dot_cluster_nodes(
      cluster_id  = "Y",
      title       = "T",
      node_ids    = Y_ids[idx_Y],
      node_labels = Y_lab[idx_Y],
      shape       = "box",
      fill        = fill,
      fillcolor   = "lightblue",
      line_width  = 1.5
    )
  )

  ## -------------------------------------------------------------
  ## Latent factors cluster (X)
  ## -------------------------------------------------------------
  scr <- paste0(
    scr,
    '\n  // Latent variables (X)\n',
    .nmfkc_dot_cluster_nodes(
      cluster_id  = "X",
      title       = "Latent Variables",
      node_ids    = X_ids,
      node_labels = X_lab,
      shape       = "ellipse",
      fill        = fill,
      fillcolor   = "wheat",
      line_width  = 1.0
    )
  )

  ## -------------------------------------------------------------
  ## Edge defaults
  ## -------------------------------------------------------------
  scr <- paste0(
    scr,
    '\n  edge [fontname="Arial", fontsize=8, arrowhead=open];\n'
  )

  ## -------------------------------------------------------------
  ## 1. X → T edges (factor loadings)
  ## -------------------------------------------------------------
  scr <- paste0(
    scr,
    '\n  // X -> T edges (factor loadings)\n',
    '  edge [color="gray0", fontcolor="gray0", style=solid];\n'
  )

  if (is.finite(max_X) && max_X > 0) {
    for (i in idx_Y) {
      for (j in seq_len(ncol(X))) {
        val <- X[i, j]
        if (is.finite(val) && val >= threshold) {
          pen <- pw(val, max_X, weight_scale_xy)
          lab <- fmtc(val)
          scr <- paste0(
            scr,
            sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                    X_ids[j], Y_ids[i], lab, pen)
          )
        }
      }
    }
  }

  ## -------------------------------------------------------------
  ## 2. Optional intercept nodes
  ## -------------------------------------------------------------
  if (isTRUE(intercept) && has_intercept_col && is.finite(max_int) && max_int > 0) {

    scr <- paste0(scr, '\n  // Intercept nodes\n')
    int_col <- which(C_lab == "(Intercept)")

    for (j in seq_len(ncol(X))) {
      val <- C[j, int_col]
      if (is.finite(val) && val >= threshold) {
        pen <- pw(val, max_int, weight_scale_int)
        lab <- fmtc(val)
        node_id <- paste0("Const", j)

        scr <- paste0(
          scr,
          sprintf('  %s [shape=circle, label="%s"];\n', node_id, lab),
          sprintf('  %s -> %s [penwidth=%.2f];\n', node_id, X_ids[j], pen)
        )
      }
    }
  }

  ## -------------------------------------------------------------
  ## 3. Lag blocks and T-k → X edges
  ## -------------------------------------------------------------
  for (k in seq_len(D)) {

    start <- (k - 1L) * A_per_lag
    cols  <- (start + 1L):(start + A_per_lag)

    Ck     <- C[, cols, drop = FALSE]
    Ck_ids <- C_ids[cols]
    Ck_lab <- base_from_name(C_lab[cols])

    ## Skip lag block if no coefficient exceeds threshold
    if (suppressWarnings(max(Ck, na.rm = TRUE)) < threshold) {
      next
    }

    ## ---- Cluster for lag k ----
    st <- sprintf('  subgraph cluster_C%d {label="T-%d" style="rounded";\n', k, k)

    if (fill) {
      st <- paste0(
        st,
        '    node [shape=box, style="filled,rounded", fillcolor="lightcoral", color=black, penwidth=1.5];\n'
      )
    } else {
      st <- paste0(
        st,
        '    node [shape=box, style="rounded", color=black, penwidth=1.5];\n'
      )
    }

    for (j in seq_len(ncol(Ck))) {
      if (suppressWarnings(max(Ck[, j], na.rm = TRUE)) >= threshold) {
        st <- paste0(
          st,
          sprintf('    %s [label="%s"];\n', Ck_ids[j], Ck_lab[j])
        )
      }
    }
    st  <- paste0(st, "  }\n")

    scr <- paste0(scr, "\n  // Lag block T-", k, "\n", st)

    ## ---- Lag → X edges ----
    scr <- paste0(
      scr,
      '  // T-', k, ' -> X edges\n',
      '  edge [color=black, fontcolor=black, style=solid];\n'
    )

    if (is.finite(max_C) && max_C > 0) {
      for (q in seq_len(nrow(Ck))) {
        for (j in seq_len(ncol(Ck))) {
          val <- Ck[q, j]
          if (is.finite(val) && val >= threshold) {
            pen <- pw(val, max_C, weight_scale_lag)
            lab <- fmtc(val)
            scr <- paste0(
              scr,
              sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                      Ck_ids[j], X_ids[q], lab, pen)
            )
          }
        }
      }
    }
  }

  result <- paste0(scr, "}\n")
  class(result) <- c("nmfkc.ar.DOT", "nmfkc.DOT")
  result
}

# nmf.sem.R — NMF-FFB (formerly NMF-SEM) related functions
# nmf.sem, nmf.sem.cv, nmf.sem.split, nmf.sem.DOT

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#' @title NMF-FFB Main Estimation Algorithm (formerly NMF-SEM)
#'
#' @description
#' Fits the NMF-FFB model
#' \deqn{
#'   Y_1 \approx X \bigl( \Theta_1 Y_1 + \Theta_2 Y_2 \bigr)
#' }
#' under non-negativity constraints with orthogonality and sparsity regularization.
#' The function returns the estimated latent factors, structural coefficient matrices,
#' and the implied equilibrium (input–output) mapping.
#'
#' At equilibrium, the model can be written as
#' \deqn{
#'   Y_1 \approx (I - X \Theta_1)^{-1} X \Theta_2 Y_2
#'   \equiv M_{\mathrm{model}} Y_2,
#' }
#' where \eqn{M_{\mathrm{model}} = (I - X \Theta_1)^{-1} X \Theta_2} is a
#' Leontief-type cumulative-effect operator in latent space.
#'
#' Internally, the latent feedback and exogenous loading matrices are stored as
#' \code{C1} and \code{C2}, corresponding to \eqn{\Theta_1} and \eqn{\Theta_2},
#' respectively.
#'
#' @param Y1 A non-negative numeric matrix of endogenous variables with
#'   \strong{rows = variables (P1), columns = samples (N)}.
#' @param Y2 A non-negative numeric matrix of exogenous variables with
#'   \strong{rows = variables (P2), columns = samples (N)}.
#'   Must satisfy \code{ncol(Y1) == ncol(Y2)}.
#' @param rank Integer; number of latent factors \eqn{Q}. If \code{NULL},
#'   \eqn{Q} is taken from a hidden argument in \code{...} or defaults to
#'   \code{nrow(Y2)}.
#' @param X.init Initialization strategy for the basis matrix
#'   \code{X} (\eqn{P_1 \times Q}).  One of:
#'   \itemize{
#'     \item \code{"nndsvd"} (default): Non-negative Double SVD with
#'       additive randomness (NNDSVDar; Boutsidis & Gallopoulos 2008),
#'       computed internally via \code{.nndsvdar(Y1, Q)}.  Requires
#'       \eqn{Q \le \min(P_1, N)} (over-rank case falls back to
#'       \code{"runif"}).  Uses a full SVD of \eqn{Y_1}, so for very
#'       large \eqn{Y_1} consider switching to \code{"kmeans"} to
#'       avoid SVD memory / compute cost.
#'     \item \code{"kmeans"}: k-means on the columns of \eqn{Y_1}
#'       (samples clustered into \eqn{Q} groups); the transposed
#'       cluster centers become \eqn{X}.  Scales well for large
#'       \eqn{Y_1}; this is the default of \code{\link{nmfkc}}.
#'     \item \code{"kmeansar"}: \code{"kmeans"} followed by filling
#'       zero entries of \eqn{X} with \eqn{\mathrm{Uniform}(0,
#'       \bar Y_1 / 100)} (NNDSVDar-style additive randomness to
#'       escape trivial stationary points).
#'     \item \code{"runif"}: Uniform random entries in \eqn{[0, 1]}.
#'     \item A numeric \eqn{P_1 \times Q} matrix supplied by the user;
#'       negative entries are projected to 0.
#'     \item \code{NULL}: backward-compatible alias for \code{"nndsvd"}.
#'   }
#'   In all cases the result is column-normalized to \code{colSums(X) = 1}
#'   before iteration.  The menu mirrors \code{\link{nmfkc}}'s
#'   \code{X.init} option for consistency across the package.
#' @param X.L2.ortho L2 orthogonality penalty for \code{X}. This controls
#'   the penalty term \eqn{\lambda_X \lVert X^\top X - \mathrm{diag}(X^\top X)
#'   \rVert_F^2}. Default: \code{100}.
#' @param C1.L1 L1 sparsity penalty for \code{C1} (i.e., \eqn{\Theta_1}).
#'   Default: \code{1.0}.
#' @param C2.L1 L1 sparsity penalty for \code{C2} (i.e., \eqn{\Theta_2}).
#'   Default: \code{0.1}.
#' @param epsilon Relative convergence threshold for the objective function.
#'   Iterations stop when the relative change in reconstruction loss falls
#'   below this value. Default: \code{1e-6}.
#' @param maxit Maximum number of iterations for the multiplicative updates.
#'   Default: \code{5000} (matches \code{\link{nmfkc}} and other MU
#'   functions in the package).
#' @param seed Random seed used to initialize \code{X}, \code{C1}, and \code{C2}.
#'   Default: \code{123}.
#' @param ... Additional hidden arguments controlling the optional
#'   feedforward baseline (used both as an \eqn{X} warm-start and as
#'   the reference for \code{SC.map}, the input-output structural
#'   fidelity defined in Satoh (2025) §4.SC.map):
#'   \describe{
#'     \item{\code{nmfkc.baseline}}{Controls whether a feedforward
#'       \code{\link{nmfkc}}(Y1, A = Y2) fit is used as baseline.
#'       Possible values:
#'       \itemize{
#'         \item Default (not given) — \code{nmf.sem} runs
#'           \code{\link{nmfkc}} \strong{internally} when \code{X.init}
#'           is a string method (\code{"nndsvd"}, \code{"kmeans"},
#'           \dots) or \code{NULL}, forwarding \code{X.init},
#'           \code{X.L2.ortho}, \code{epsilon}, \code{maxit},
#'           \code{seed}.  The fitted \eqn{X} of the baseline is then
#'           used as warm-start for the nmf.sem MU iterations, and
#'           \code{SC.map} is computed.  This means
#'           \code{nmf.sem(Y1, Y2, rank = Q)} runs end-to-end without
#'           a prior \code{nmfkc} call.
#'         \item \code{TRUE} — same as above, but force the internal
#'           \code{\link{nmfkc}} call even when \code{X.init} is a
#'           user-supplied matrix (the matrix is overridden).
#'         \item \code{FALSE} — opt out; no internal call,
#'           \code{SC.map = NA} (pre-v0.6.8 behavior).
#'         \item An \code{\link{nmfkc}} result (list with \code{$X}
#'           and \code{$C}) — use as the baseline directly (no
#'           internal call); also adopted as \code{X.init} when the
#'           latter is a string / NULL.
#'       }}
#'     \item{\code{M.simple}}{Optional \eqn{P_1 \times P_2} pre-computed
#'       baseline mapping.  Takes precedence over \code{nmfkc.baseline}
#'       for the SC.map calculation but does not affect warm-start.}
#'     \item{\code{Q}}{Backward-compat alias for \code{rank}.}
#'   }
#'
#' @return A list with components:
#'   \item{X}{Estimated basis matrix (\eqn{P_1 \times Q}).}
#'   \item{C1}{Estimated latent feedback matrix (\eqn{\Theta_1}, \eqn{Q \times P_1}).}
#'   \item{C2}{Estimated exogenous loading matrix (\eqn{\Theta_2}, \eqn{Q \times P_2}).}
#'   \item{XC1}{Feedback matrix \eqn{X \Theta_1}.}
#'   \item{XC2}{Direct-effect matrix \eqn{X \Theta_2}.}
#'   \item{XC1.radius}{Spectral radius \eqn{\rho(X \Theta_1)}.}
#'   \item{XC1.norm1}{Induced 1-norm \eqn{\lVert X \Theta_1 \rVert_{1,\mathrm{op}}}.}
#'   \item{Leontief.inv}{Leontief-type inverse \eqn{(I - X \Theta_1)^{-1}.}}
#'   \item{M.model}{Equilibrium mapping
#'     \eqn{M_{\mathrm{model}} = (I - X \Theta_1)^{-1} X \Theta_2}.}
#'   \item{amplification}{Latent amplification factor
#'     \eqn{\lVert M_{\mathrm{model}} \rVert_{1,\mathrm{op}} /
#'          \bigl\lVert X \Theta_2 \bigr\rVert_{1,\mathrm{op}}}.}
#'   \item{amplification.bound}{Geometric-series upper bound
#'     \eqn{1 / (1 - \lVert X \Theta_1 \rVert_{1,\mathrm{op}})} if
#'     \eqn{\lVert X \Theta_1 \rVert_{1,\mathrm{op}} < 1}, otherwise \code{Inf}.}
#'   \item{Q}{Effective latent dimension used in the fit.}
#'   \item{SC.cov}{Correlation between sample and model-implied covariance
#'     (flattened) of \eqn{Y_1}.  See \emph{second-moment fidelity} in
#'     Satoh (2025).}
#'   \item{SC.map}{Correlation between the equilibrium operator
#'     \eqn{M_{\mathrm{model}}} and a feedforward baseline mapping
#'     \eqn{M_{\mathrm{simple}} = X_0 \Theta_0}, computed only when the
#'     baseline is supplied via \code{M.simple} or \code{nmfkc.baseline}
#'     in \code{...}; otherwise \code{NA}.  See \emph{input-output
#'     structural fidelity} in Satoh (2025).}
#'   \item{MAE}{Mean absolute error between \eqn{Y_1} and its equilibrium
#'     prediction \eqn{\hat Y_1 = M_{\mathrm{model}} Y_2}.}
#'   \item{objfunc}{Vector of reconstruction losses per iteration.}
#'   \item{objfunc.full}{Vector of penalized objective values per iteration.}
#'   \item{iter}{Number of iterations actually performed.}
#'
#' @examples
#' # Simple NMF-FFB with iris data (non-negative)
#' Y <- t(iris[, -5])
#' Y1 <- Y[1:2, ]  # Sepal
#' Y2 <- Y[3:4, ]  # Petal
#' result <- nmf.sem(Y1, Y2, rank = 2, maxit = 500)
#' result$MAE
#'
#' @seealso \code{\link{nmf.sem.inference}}, \code{\link{nmf.sem.cv}},
#'   \code{\link{nmf.sem.split}}, \code{\link{nmf.sem.DOT}},
#'   \code{\link{summary.nmf.sem}}
#' @references
#' Satoh, K. (2025). Applying non-negative matrix factorization with covariates
#'   to structural equation modeling for blind input-output analysis.
#'   arXiv:2512.18250. \url{https://arxiv.org/abs/2512.18250}
#' @export
nmf.sem <- function(
    Y1, Y2,
    rank = NULL,
    X.init = "nndsvd",
    X.L2.ortho = 100.0,
    C1.L1 = 1.0,
    C2.L1 = 0.1,
    epsilon = 1e-6,
    maxit = 5000,
    seed  = 123,
    ...
) {
  # ------------------------------ checks ------------------------------
  if (!is.matrix(Y1)) Y1 <- as.matrix(Y1)
  if (!is.matrix(Y2)) Y2 <- as.matrix(Y2)

  if (any(!is.finite(Y1)) || any(!is.finite(Y2)))
    stop("Y1 and Y2 must not contain NA/NaN/Inf.")
  if (min(Y1) < 0 || min(Y2) < 0)
    stop("Y1 and Y2 must be non-negative.")
  if (ncol(Y1) != ncol(Y2))
    stop("ncol(Y1) must be equal to ncol(Y2).")

  extra_args <- list(...)
  Q_hidden <- if (!is.null(extra_args$Q)) extra_args$Q else NULL
  Q0 <- if (!is.null(rank)) rank else if (!is.null(Q_hidden)) Q_hidden else nrow(Y2)

  P1 <- nrow(Y1); P2 <- nrow(Y2); N <- ncol(Y1)
  Q  <- Q0
  if (Q < 1) stop("Rank Q must be >= 1.")

  # -------------------------- labels (output) -------------------------
  Y1_labels    <- if (!is.null(rownames(Y1))) rownames(Y1) else paste0("Y1_", 1:P1)
  Y2_labels    <- if (!is.null(rownames(Y2))) rownames(Y2) else paste0("Y2_", 1:P2)
  Basis_labels <- paste0("Factor", 1:Q)

  set.seed(seed)
  .eps <- 1e-10
  .xnorm  <- function(X) sweep(X, 2, pmax(colSums(X), .eps), "/")
  mat1norm <- function(A) max(colSums(abs(A)))

  # ---------- (optional) internal nmfkc warm-start + SC.map baseline --
  ## Resolution rules for `nmfkc.baseline`:
  ##   * not given (default): if X.init is a string method (or NULL),
  ##     run nmfkc(Y1, A = Y2) INTERNALLY using the same X.init,
  ##     X.L2.ortho, epsilon, maxit, seed; use its X as warm-start
  ##     and X * C as M.simple for SC.map.  This is the typical
  ##     workflow described in Satoh (2025) and means
  ##       res <- nmf.sem(Y1, Y2, rank = Q)
  ##     can run end-to-end without first calling nmfkc().
  ##   * TRUE: same as above (explicit opt-in even when X.init is a
  ##     user-supplied matrix; the matrix is overridden by nmfkc's X).
  ##   * FALSE: opt-out — no internal nmfkc, no warm-start, SC.map = NA
  ##     (equivalent to pre-v0.6.8 behavior).
  ##   * an nmfkc result (list with $X and $C): use as the baseline
  ##     for SC.map and as warm-start for X.init (no internal call).
  baseline_for_scmap <- NULL

  user_baseline    <- extra_args$nmfkc.baseline
  user_M.simple    <- extra_args$M.simple

  baseline_is_obj  <- is.list(user_baseline) &&
                      !is.null(user_baseline$X) && !is.null(user_baseline$C)
  baseline_is_TRUE <- isTRUE(user_baseline)
  baseline_is_FALSE <- isFALSE(user_baseline)
  baseline_is_default <- is.null(user_baseline)

  ## Default auto-run when baseline is unspecified AND X.init is a string
  ## method (NULL is treated as "nndsvd" string).
  auto_nmfkc <- if (baseline_is_FALSE) {
    FALSE
  } else if (baseline_is_TRUE) {
    TRUE
  } else if (baseline_is_obj) {
    FALSE
  } else if (baseline_is_default) {
    is.null(X.init) || is.character(X.init)
  } else {
    FALSE
  }

  if (auto_nmfkc) {
    ## Internal nmfkc call.  Forward only the genuinely shared options
    ## (X.init, X.L2.ortho, epsilon, maxit, seed); the nmf.sem-specific
    ## C1.L1 / C2.L1 do not apply to the feedforward baseline model.
    nmfkc_xinit <- if (is.null(X.init)) "nndsvd" else X.init
    baseline_for_scmap <- nmfkc(
      Y = Y1, A = Y2, Q = Q,
      X.init = nmfkc_xinit,
      X.L2.ortho = X.L2.ortho,
      epsilon = epsilon,
      maxit = maxit,
      seed = seed,
      verbose = FALSE,
      print.dims = FALSE
    )
    ## Override X.init with the nmfkc-fitted X for nmf.sem warm-start
    X.init <- baseline_for_scmap$X
  } else if (baseline_is_obj) {
    baseline_for_scmap <- user_baseline
    ## When the user supplies an nmfkc result AND X.init is still a
    ## string / NULL, also use baseline$X as warm-start (rmd workflow).
    if (is.null(X.init) || is.character(X.init)) {
      X.init <- baseline_for_scmap$X
    }
  }

  # ---------------------------- init X,C1,C2 --------------------------
  ## X.init dispatch: delegate string methods to the shared internal
  ## helper .init_X_method() (defined in R/nmfkc.R).  Accepts:
  ##   "nndsvd" (default), "kmeans", "kmeansar", "runif",
  ##   a numeric P1 x Q matrix, or NULL (alias for "nndsvd").
  if (is.null(X.init)) X.init <- "nndsvd"
  if (is.character(X.init)) {
    X <- .init_X_method(X.init, Y1, Q)
  } else {
    X <- as.matrix(X.init)
    if (!all(dim(X) == c(P1, Q))) {
      stop("X.init must have dimension (nrow(Y1) x rank).")
    }
    X[X < 0] <- 0
  }
  X <- .xnorm(X)

  C1 <- matrix(stats::runif(Q * P1, 0.01, 0.1), nrow = Q, ncol = P1)
  C2 <- matrix(stats::runif(Q * P2, 0.001, 0.01), nrow = Q, ncol = P2)

  min_dim <- min(Q, P2)
  for (i in 1:min_dim) {
    C2[i, i] <- stats::runif(1, 0.1, 0.2)
  }

  objfunc      <- numeric(maxit)
  objfunc.full <- numeric(maxit)

  # ----------------------------- main loop ----------------------------
  for (it in 1:maxit) {
    M  <- C1 %*% Y1 + C2 %*% Y2
    Mt <- t(M)

    # 2.1 update X
    Numerator_X       <- Y1 %*% Mt
    Denominator_X_rec <- X %*% M %*% Mt

    if (X.L2.ortho > 0) {
      XtX <- t(X) %*% X
      XtX_offdiag <- XtX
      diag(XtX_offdiag) <- 0
      Denominator_X_ortho <- X.L2.ortho * X %*% XtX_offdiag
    } else {
      Denominator_X_ortho <- 0
    }

    X <- X * (Numerator_X / (Denominator_X_rec + Denominator_X_ortho + .eps))
    X <- .xnorm(X)

    Xt  <- t(X)
    XtX <- Xt %*% X

    # 2.2 update C1
    Numerator_C1   <- Xt %*% Y1 %*% t(Y1)
    Denominator_C1 <- XtX %*% (C1 %*% Y1 + C2 %*% Y2) %*% t(Y1) + C1.L1 + .eps
    C1 <- C1 * (Numerator_C1 / Denominator_C1)

    # 2.3 update C2
    Numerator_C2   <- Xt %*% Y1 %*% t(Y2)
    Denominator_C2 <- XtX %*% (C1 %*% Y1 + C2 %*% Y2) %*% t(Y2) + C2.L1 + .eps
    C2 <- C2 * (Numerator_C2 / Denominator_C2)

    # loss + penalties
    XB <- X %*% (C1 %*% Y1 + C2 %*% Y2)
    loss_rec <- sum((Y1 - XB)^2)
    objfunc[it] <- loss_rec

    if (X.L2.ortho > 0) {
      XtX_off <- XtX
      diag(XtX_off) <- 0
      pen_X_ortho <- 0.5 * X.L2.ortho * sum(XtX_off^2)
    } else {
      pen_X_ortho <- 0
    }
    pen_C1_L1 <- C1.L1 * sum(C1)
    pen_C2_L1 <- C2.L1 * sum(C2)
    objfunc.full[it] <- loss_rec + pen_X_ortho + pen_C1_L1 + pen_C2_L1

    if (it >= 10) {
      epsilon_iter <- abs(objfunc[it] - objfunc[it - 1]) / pmax(abs(objfunc[it]), 1)
      if (epsilon_iter <= epsilon) break
    }
  }
  ## Warn when the MU loop exhausts maxit without meeting the
  ## relative-tolerance criterion (matches the nmfkc() convention).
  if (it == maxit && exists("epsilon_iter") && epsilon_iter > abs(epsilon))
    warning(paste0("maximum iterations (", maxit, ") reached..."))

  # ------------------ reorder factors (nmfkc centroid order) ----------
  centroid <- as.numeric((1:nrow(X)) / nrow(X)) %*% X
  index <- order(centroid)
  X  <- X[, index, drop = FALSE]
  C1 <- C1[index, , drop = FALSE]
  C2 <- C2[index, , drop = FALSE]

  # ------------------------------ names -------------------------------
  colnames(X)  <- Basis_labels
  rownames(C1) <- Basis_labels
  rownames(C2) <- Basis_labels
  rownames(X)  <- Y1_labels
  colnames(C1) <- Y1_labels
  colnames(C2) <- Y2_labels

  # -------------------- feedback + stability diagnostics --------------
  XC1  <- X %*% C1
  eigs <- eigen(XC1, only.values = TRUE)$values
  rho  <- max(abs(eigs))
  if (rho >= 1)
    warning("Leontief.inv may be unstable; spectral radius >= 1.")

  XC1_norm1 <- mat1norm(XC1)

  # -------------------- Leontief inverse + equilibrium mapping --------
  I_mat <- diag(nrow(XC1))
  XC2   <- X %*% C2

  Leontief.inv <- tryCatch(
    base::solve(I_mat - XC1),
    error = function(e) {
      warning("Failed to compute Leontief.inv via solve(I - XC1). Returning NA matrices.")
      matrix(NA_real_, nrow = nrow(XC1), ncol = ncol(XC1))
    }
  )
  M.model <- Leontief.inv %*% XC2

  amplification <- mat1norm(M.model) / (mat1norm(XC2) + .eps)
  amplification.bound <- if (XC1_norm1 < 1) 1 / (1 - XC1_norm1) else Inf

  # -------------------- fit indices (equilibrium prediction) ----------
  if (anyNA(Leontief.inv)) {
    Y1_hat <- matrix(NA_real_, nrow = nrow(Y1), ncol = ncol(Y1))
    SC.cov <- NA_real_
    MAE    <- NA_real_
  } else {
    Y1_hat   <- M.model %*% Y2
    S.sample <- Y1 %*% t(Y1)
    S.model  <- Y1_hat %*% t(Y1_hat)
    SC.cov   <- stats::cor(as.numeric(S.sample), as.numeric(S.model))
    MAE      <- mean(abs(Y1 - Y1_hat))
  }

  ## -------------------- input-output structural fidelity (SC.map) -----
  ## SC.map = cor(vec(M.model), vec(M.simple)), where M.simple = X0 * Theta0
  ## is the feedforward baseline mapping (Satoh 2025 §4.SC.map).
  ## `baseline_for_scmap` was set above to either:
  ##   - the result of an internal nmfkc call (default auto path), or
  ##   - the user-supplied `nmfkc.baseline` list, or
  ##   - NULL if the user opted out (nmfkc.baseline = FALSE) or
  ##     supplied X.init as a numeric matrix without nmfkc.baseline.
  ## A user-supplied `M.simple` matrix takes precedence over both.
  SC.map <- NA_real_
  M.simple <- if (!is.null(user_M.simple)) user_M.simple
              else if (!is.null(baseline_for_scmap))
                baseline_for_scmap$X %*% baseline_for_scmap$C
              else NULL
  if (!is.null(M.simple) && !anyNA(M.model)) {
    M.simple <- as.matrix(M.simple)
    if (all(dim(M.simple) == dim(M.model))) {
      SC.map <- tryCatch(
        stats::cor(as.numeric(M.simple), as.numeric(M.model)),
        error = function(e) NA_real_
      )
    } else {
      warning("M.simple has dimension ", paste(dim(M.simple), collapse = "x"),
              " but M.model is ", paste(dim(M.model), collapse = "x"),
              "; SC.map not computed.")
    }
  }

  out <- list(
    X                   = X,
    C1                  = C1,
    C2                  = C2,
    XC1                 = XC1,
    XC2                 = XC2,
    XC1.radius          = rho,
    XC1.norm1           = XC1_norm1,
    Leontief.inv        = Leontief.inv,
    M.model             = M.model,
    amplification       = amplification,
    amplification.bound = amplification.bound,
    Q                   = Q,
    SC.cov              = SC.cov,
    SC.map              = SC.map,
    MAE                 = MAE,
    ## Effective rank of the latent scores B = C1 Y1 + C2 Y2 (Q x N).
    effective.rank      = .effective.rank(C1 %*% Y1 + C2 %*% Y2),
    objfunc             = objfunc[1:it],
    objfunc.full        = objfunc.full[1:it],
    iter                = it
  )
  ## Carry both the canonical NMF-FFB class (paper-aligned, primary) and
  ## the legacy "nmf.sem" class (back-compat).  S3 methods registered on
  ## either class are dispatched correctly via inheritance.
  class(out) <- c("nmf.ffb", "nmf.sem")
  out
}


#' @title Statistical inference for NMF-FFB via X-fixed full pair bootstrap
#' @description
#' \code{nmf.sem.inference} performs statistical inference on the structural
#' coefficient matrices \eqn{C_1} (latent feedback, \eqn{\Theta_1}) and
#' \eqn{C_2} (exogenous loading, \eqn{\Theta_2}) from a fitted
#' \code{\link{nmf.sem}} model.
#'
#' The procedure is a \strong{full pair bootstrap} that holds the basis
#' matrix \eqn{\hat X} from the original fit fixed across all replicates
#' (which avoids label switching and gives a clean conditional
#' interpretation: ``uncertainty of the structural coefficients given the
#' measurement model''):
#' \enumerate{
#'   \item For each replicate \eqn{b = 1, \dots, B}, resample column indices
#'     \eqn{(i_1, \dots, i_N)} with replacement from \eqn{\{1, \dots, N\}}
#'     and form \eqn{Y_1^{(b)} = Y_1[, i]}, \eqn{Y_2^{(b)} = Y_2[, i]}.
#'   \item Re-estimate \eqn{(C_1^{(b)}, C_2^{(b)})} by running the
#'     \code{\link{nmf.sem}} multiplicative updates \emph{with \eqn{X = \hat X}
#'     held fixed} (no \eqn{X} update; no centroid sort), using the same
#'     \code{C1.L1}, \code{C2.L1} as the original fit.
#'   \item Discard replicates that violate stationarity
#'     (\eqn{\rho(X C_1^{(b)}) \ge 1}) or have an amplification ratio
#'     exceeding the geometric-series bound by more than 1\%.
#' }
#' Because \eqn{C_1, C_2 \ge 0} are non-negative by construction, exact
#' zeros are essentially never observed in the bootstrap distribution.
#' Significance is assessed via a \strong{support rate} at a small display
#' threshold \eqn{\delta} (default \code{0.01}):
#' \deqn{
#'   \mathrm{sup}(c) \;=\; \frac{1}{|\mathrm{valid}|}
#'     \sum_{b \in \mathrm{valid}} \mathbf{1}\!\left( \hat c^{(b)} > \delta \right).
#' }
#' This is a one-sided counterpart of the classical \eqn{p}-value:
#' large \code{support_rate} indicates strong evidence that the entry is
#' meaningfully positive.  Significance markers follow the lavaan
#' convention with the natural correspondence \eqn{p = 1 - \mathrm{sup}}:
#' \code{*} (sup > 0.95), \code{**} (sup > 0.99), \code{***}
#' (sup > 0.999).  Cutoffs use strict greater-than so the rule
#' mirrors the standard R convention for p-values (p < 0.05 / 0.01 /
#' 0.001 → */**/***), translated to support_rate via
#' \eqn{\mathrm{sup} = 1 - p}.
#'
#' @param object A fitted object returned by \code{\link{nmf.sem}}.  Must
#'   contain \code{X}, \code{C1}, \code{C2}.
#' @param Y1 Endogenous variable matrix (P1 x N).  Must match the data
#'   used in \code{nmf.sem()}.
#' @param Y2 Exogenous variable matrix (P2 x N).  Same.
#' @param B Number of bootstrap replicates.  Default \code{1000}; required
#'   for the \code{***} threshold (sup > 0.999).  Reduce to 500 for
#'   exploratory speed (only \code{*} / \code{**} stay reliable).
#' @param threshold Display threshold \eqn{\delta} for the support rate
#'   \eqn{\Pr_{\mathrm{boot}}(\hat c^{(b)} > \delta)}.  Default
#'   \code{0.01}; entries below this magnitude are treated as effectively
#'   zero in the path diagram.
#' @param ci.level Confidence level for the percentile bootstrap CI.
#'   Default \code{0.95}.
#' @param C1.L1,C2.L1 L1 sparsity penalties used by the original
#'   \code{\link{nmf.sem}} fit.  These must match the fit's hyperparameters
#'   for the bootstrap to estimate the correct model.  Defaults
#'   (\code{1.0}, \code{0.1}) match \code{nmf.sem}'s defaults but you
#'   should pass the actual values used.
#' @param seed Base RNG seed for the bootstrap.  Each replicate uses
#'   \code{seed + b} (resampling) and \code{seed + 1000 + b}
#'   (\eqn{C_1, C_2} initialization).  Default \code{123}.
#' @param ... Hidden options:
#'   \describe{
#'     \item{\code{epsilon}}{Convergence tolerance for the inner fixed-X MU
#'       loop.  Default \code{1e-6}.}
#'     \item{\code{maxit}}{Maximum iterations for the inner MU loop.
#'       Default \code{5000}.}
#'     \item{\code{ncores}}{Number of parallel workers.  Default \code{1}
#'       (serial).  Cross-platform: uses \code{parallel::mclapply} on
#'       Linux/macOS and \code{parallel::parLapply} (PSOCK cluster) on
#'       Windows.}
#'     \item{\code{print.trace}}{Logical, print progress.  Default
#'       \code{FALSE}.}
#'   }
#'
#' @return The input \code{object} with additional bootstrap inference
#'   components:
#' \item{coefficients}{Data frame with rows for every entry of \eqn{C_1}
#'   and \eqn{C_2} and columns \code{Type} ("C1" / "C2"), \code{Basis},
#'   \code{Covariate}, \code{Estimate}, \code{CI_low}, \code{CI_high},
#'   \code{support_rate}, \code{p_value} (\eqn{= 1 - \mathrm{support\_rate}},
#'   for compatibility with downstream consumers such as
#'   \code{\link{nmf.sem.DOT}}), and \code{sig}.}
#' \item{C1.support.rate, C2.support.rate}{Per-element support rates
#'   (Q x P1 and Q x P2 matrices).}
#' \item{C1.ci.lower, C1.ci.upper, C2.ci.lower, C2.ci.upper}{Per-element
#'   percentile CI bounds.}
#' \item{C1.array, C2.array}{Bootstrap distributions: 3D arrays of shape
#'   B x Q x P1 (and B x Q x P2).  Invalid replicates contain \code{NA}.}
#' \item{rho.boot, AR.boot, iter.boot}{Per-replicate spectral radius,
#'   amplification ratio, and inner-loop iteration count.}
#' \item{bootstrap.B, bootstrap.threshold, bootstrap.ci.level}{Inputs
#'   recorded for reproducibility.}
#' \item{bootstrap.n.valid, bootstrap.n.invalid}{Validity counts.}
#'
#' @section Lifecycle:
#' This function's interface changed at v0.6.8: the legacy 1-step Newton
#' wild bootstrap (with sandwich SE) has been replaced by the full pair
#' bootstrap described above, following the paper revision.  The fields
#' \code{sigma2.used}, \code{C2.se}, \code{C2.se.boot}, \code{C2.p.side}
#' that the previous implementation produced are no longer present.
#'
#' @seealso \code{\link{nmf.sem}}, \code{\link{nmf.sem.DOT}}
#' @references
#' Satoh, K. (2025). Applying non-negative matrix factorization with covariates
#'   to structural equation modeling for blind input-output analysis.
#'   arXiv:2512.18250. \url{https://arxiv.org/abs/2512.18250}
#' @export
#' @examples
#' \donttest{
#' Y <- t(iris[, -5])
#' Y1 <- Y[1:2, ]; Y2 <- Y[3:4, ]
#' res  <- nmf.sem(Y1, Y2, rank = 2)
#' res2 <- nmf.sem.inference(res, Y1, Y2, B = 200)  # quick demo
#' head(res2$coefficients)
#' }
nmf.sem.inference <- function(object, Y1, Y2,
                               B = 1000L,
                               threshold = 0.01,
                               ci.level = 0.95,
                               C1.L1 = 1.0,
                               C2.L1 = 0.1,
                               seed = 123L,
                               ...) {
  if (is.null(object$X) || is.null(object$C1) || is.null(object$C2))
    stop("object must contain X, C1, and C2 (returned by nmf.sem).")

  extra_args  <- base::list(...)
  epsilon     <- if (!is.null(extra_args$epsilon))     extra_args$epsilon     else 1e-6
  maxit       <- if (!is.null(extra_args$maxit))       extra_args$maxit       else 5000L
  ncores      <- if (!is.null(extra_args$ncores))      extra_args$ncores      else 1L
  print.trace <- if (!is.null(extra_args$print.trace)) extra_args$print.trace else FALSE

  Y1 <- base::as.matrix(Y1)
  Y2 <- base::as.matrix(Y2)
  if (ncol(Y1) != ncol(Y2)) stop("Y1 and Y2 must have the same number of columns.")

  X.hat  <- object$X
  C1.hat <- object$C1
  C2.hat <- object$C2

  Q  <- ncol(X.hat)
  P1 <- nrow(Y1)
  P2 <- nrow(Y2)
  N  <- ncol(Y1)
  if (nrow(X.hat) != P1)
    stop("nrow(X) must equal nrow(Y1); did Y1 change shape?")
  if (Q < 1L) stop("Inferred rank Q < 1.")

  ## ----------------------------------------------------------------
  ## one_boot: a single bootstrap replicate.
  ## Closure captures Y1, Y2, X.hat, C1.L1, C2.L1, Q, P1, P2, N,
  ## epsilon, maxit, seed.  Returns a list with valid flag plus
  ## (C1, C2, rho, AR, iter) when valid, or NA placeholders otherwise.
  ## ----------------------------------------------------------------
  one_boot <- function(b) {
    ## Sample column indices with replacement (pair bootstrap)
    set.seed(seed + b)
    idx  <- sample.int(N, size = N, replace = TRUE)
    Y1_b <- Y1[, idx, drop = FALSE]
    Y2_b <- Y2[, idx, drop = FALSE]

    ## Independent random init for C1, C2 (X is held at X.hat)
    set.seed(seed + 1000L + b)
    C1 <- matrix(stats::runif(Q * P1, 0.01, 0.1),  nrow = Q)
    C2 <- matrix(stats::runif(Q * P2, 0.001, 0.01), nrow = Q, ncol = P2)
    min_dim <- min(Q, P2)
    if (min_dim >= 1L) {
      for (i in 1:min_dim) C2[i, i] <- stats::runif(1, 0.1, 0.2)
    }

    .eps_local <- 1e-10
    Xt  <- t(X.hat)
    XtX <- Xt %*% X.hat
    Y1tY1_b <- tcrossprod(Y1_b)        # P1 x P1
    XtY1_b  <- Xt %*% Y1_b              # Q x N (re-used via crossprods below)
    XtY1Y1t <- XtY1_b %*% t(Y1_b)       # Q x P1   = Xt %*% Y1_b %*% t(Y1_b)
    XtY1Y2t <- XtY1_b %*% t(Y2_b)       # Q x P2
    prev_loss <- Inf
    iter_used <- 0L

    ## Fixed-X MU loop on (C1, C2)
    for (it in seq_len(maxit)) {
      M_b <- C1 %*% Y1_b + C2 %*% Y2_b   # Q x N latent representation

      ## C1 update
      Den_C1 <- XtX %*% M_b %*% t(Y1_b) + C1.L1 + .eps_local
      C1 <- C1 * (XtY1Y1t / Den_C1)

      ## C2 update (uses updated C1 via fresh M_b)
      M_b <- C1 %*% Y1_b + C2 %*% Y2_b
      Den_C2 <- XtX %*% M_b %*% t(Y2_b) + C2.L1 + .eps_local
      C2 <- C2 * (XtY1Y2t / Den_C2)

      ## Convergence check (relative change in reconstruction loss)
      XB <- X.hat %*% (C1 %*% Y1_b + C2 %*% Y2_b)
      loss <- sum((Y1_b - XB)^2)
      if (it >= 10L) {
        rel <- abs(loss - prev_loss) / max(abs(loss), 1)
        if (rel <= epsilon) { iter_used <- it; break }
      }
      prev_loss <- loss
      iter_used <- it
    }

    ## ---- Validity checks (mirrors the paper's bootstrap helper) ----
    XC1_b <- X.hat %*% C1
    rho <- tryCatch(
      max(abs(eigen(XC1_b, only.values = TRUE, symmetric = FALSE)$values)),
      error = function(e) NA_real_
    )
    norm1_XC1 <- max(colSums(abs(XC1_b)))
    AR_bound <- if (is.finite(norm1_XC1) && norm1_XC1 < 1)
                  1 / (1 - norm1_XC1)
                else Inf

    if (!is.finite(rho) || rho >= 1) {
      return(list(valid = FALSE, C1 = NULL, C2 = NULL,
                  rho = rho, AR = NA_real_, iter = iter_used))
    }
    ## XC1_b = X (P1 x Q) %*% C1 (Q x P1) is P1 x P1, so the identity
    ## here must be P1 x P1 too.  (Earlier draft used diag(Q) which is
    ## conformable only when Q == P1, otherwise solve() throws an error
    ## that the tryCatch swallowed -- causing every replicate to be
    ## marked invalid with AR = NA.)
    I_mat   <- diag(nrow(XC1_b))
    Linv    <- tryCatch(solve(I_mat - XC1_b), error = function(e) NULL)
    if (is.null(Linv)) {
      return(list(valid = FALSE, C1 = NULL, C2 = NULL,
                  rho = rho, AR = NA_real_, iter = iter_used))
    }
    XC2_b   <- X.hat %*% C2
    M.model <- Linv %*% XC2_b
    AR <- max(colSums(abs(M.model))) / (max(colSums(abs(XC2_b))) + .eps_local)
    if (!is.finite(AR) || AR > AR_bound * 1.01) {
      return(list(valid = FALSE, C1 = NULL, C2 = NULL,
                  rho = rho, AR = AR, iter = iter_used))
    }

    list(valid = TRUE, C1 = C1, C2 = C2,
         rho = rho, AR = AR, iter = iter_used)
  }

  ## ----------------------------------------------------------------
  ## Run the B replicates (parallel on Linux/macOS via mclapply,
  ## PSOCK cluster on Windows; serial when ncores == 1).
  ## ----------------------------------------------------------------
  if (print.trace)
    base::message(sprintf("  Bootstrap: B=%d, ncores=%d, threshold=%.3g, ci.level=%.2f",
                          B, ncores, threshold, ci.level))

  if (ncores > 1L) {
    if (.Platform$OS.type == "windows") {
      cl <- parallel::makeCluster(ncores)
      on.exit(parallel::stopCluster(cl), add = TRUE)
      parallel::clusterExport(cl,
        varlist = c("Y1", "Y2", "X.hat", "C1.L1", "C2.L1",
                    "Q", "P1", "P2", "N", "epsilon", "maxit", "seed"),
        envir = environment())
      res_list <- parallel::parLapply(cl, seq_len(B), one_boot)
    } else {
      res_list <- parallel::mclapply(seq_len(B), one_boot, mc.cores = ncores)
    }
  } else {
    res_list <- lapply(seq_len(B), one_boot)
  }

  ## ----------------------------------------------------------------
  ## Aggregate replicates into 3D arrays (B x Q x P1) and (B x Q x P2)
  ## ----------------------------------------------------------------
  C1.array <- array(NA_real_, dim = c(B, Q, P1))
  C2.array <- array(NA_real_, dim = c(B, Q, P2))
  rho.vec  <- rep(NA_real_, B)
  AR.vec   <- rep(NA_real_, B)
  iter.vec <- rep(NA_integer_, B)
  valid.vec <- logical(B)

  for (b in seq_len(B)) {
    r <- res_list[[b]]
    if (is.null(r) || inherits(r, "try-error")) next
    valid.vec[b] <- isTRUE(r$valid)
    if (!is.null(r$rho))  rho.vec[b]  <- r$rho
    if (!is.null(r$AR))   AR.vec[b]   <- r$AR
    if (!is.null(r$iter)) iter.vec[b] <- as.integer(r$iter)
    if (valid.vec[b]) {
      C1.array[b, , ] <- r$C1
      C2.array[b, , ] <- r$C2
    }
  }
  n.valid <- sum(valid.vec)

  if (n.valid < 10L) {
    warning(sprintf("Only %d / %d bootstrap replicates were valid; CIs / support rates may be unreliable. Consider checking convergence (raise maxit, lower epsilon) or the stationarity of the original fit (rho < 1).",
                    n.valid, B))
  }

  ## ----------------------------------------------------------------
  ## Per-element summary: support rate at threshold, percentile CI
  ## ----------------------------------------------------------------
  alpha <- 1 - ci.level

  apply_finite <- function(arr, FUN) {
    apply(arr, c(2, 3), function(v) {
      v <- v[is.finite(v)]
      if (!length(v)) NA_real_ else FUN(v)
    })
  }

  C1.support  <- apply_finite(C1.array, function(v) mean(v > threshold))
  C2.support  <- apply_finite(C2.array, function(v) mean(v > threshold))
  C1.ci.lower <- apply_finite(C1.array, function(v) stats::quantile(v, alpha / 2,    names = FALSE))
  C1.ci.upper <- apply_finite(C1.array, function(v) stats::quantile(v, 1 - alpha / 2, names = FALSE))
  C2.ci.lower <- apply_finite(C2.array, function(v) stats::quantile(v, alpha / 2,    names = FALSE))
  C2.ci.upper <- apply_finite(C2.array, function(v) stats::quantile(v, 1 - alpha / 2, names = FALSE))

  ## Significance markers from support rate (one-sided; lavaan / stats
  ## convention).  Cutoffs use strict greater-than so the rule mirrors
  ## the R convention for p-values (p < 0.05 / 0.01 / 0.001 → */**/***),
  ## translated to support_rate via support_rate = 1 - p.  This matches
  ## the verbal description "support rates exceeding 0.95, 0.99, 0.999"
  ## in Satoh (2025), arXiv:2512.18250.
  sig.from.support <- function(s) {
    ifelse(!is.finite(s), " ",
      ifelse(s > 0.999, "***",
        ifelse(s > 0.99,  "**",
          ifelse(s > 0.95, "*", " "))))
  }

  ## ----------------------------------------------------------------
  ## Build coefficients table (rows for both C1 and C2)
  ## p_value = 1 - support_rate is provided so that downstream
  ## consumers (e.g., nmf.sem.DOT) that filter / star by p_value
  ## continue to work without modification.
  ## ----------------------------------------------------------------
  Q_lab  <- if (!is.null(rownames(C1.hat))) rownames(C1.hat) else paste0("Factor", 1:Q)
  Y1_lab <- if (!is.null(colnames(C1.hat))) colnames(C1.hat) else paste0("Y1_", 1:P1)
  Y2_lab <- if (!is.null(colnames(C2.hat))) colnames(C2.hat) else paste0("Y2_", 1:P2)

  build_block <- function(type_label, Mhat, sup, lo, hi, basislabs, varlabs) {
    Q_local <- nrow(Mhat); P_local <- ncol(Mhat)
    s_vec <- as.vector(sup)
    data.frame(
      Type         = type_label,
      Basis        = rep(basislabs, times = P_local),
      Covariate    = rep(varlabs,   each  = Q_local),
      Estimate     = as.vector(Mhat),
      CI_low       = as.vector(lo),
      CI_high      = as.vector(hi),
      support_rate = s_vec,
      p_value      = ifelse(is.finite(s_vec), 1 - s_vec, NA_real_),
      sig          = sig.from.support(s_vec),
      stringsAsFactors = FALSE
    )
  }

  C1_block <- build_block("C1", C1.hat, C1.support, C1.ci.lower, C1.ci.upper, Q_lab, Y1_lab)
  C2_block <- build_block("C2", C2.hat, C2.support, C2.ci.lower, C2.ci.upper, Q_lab, Y2_lab)
  coefficients <- rbind(C1_block, C2_block)
  rownames(coefficients) <- NULL

  if (print.trace)
    base::message(sprintf("  Bootstrap done: %d / %d valid replicates.", n.valid, B))

  ## ----------------------------------------------------------------
  ## Append to object and return
  ## ----------------------------------------------------------------
  object$bootstrap.B          <- B
  object$bootstrap.threshold  <- threshold
  object$bootstrap.ci.level   <- ci.level
  object$bootstrap.n.valid    <- n.valid
  object$bootstrap.n.invalid  <- B - n.valid
  object$rho.boot             <- rho.vec
  object$AR.boot              <- AR.vec
  object$iter.boot            <- iter.vec
  object$C1.array             <- C1.array
  object$C2.array             <- C2.array
  object$C1.support.rate      <- C1.support
  object$C2.support.rate      <- C2.support
  object$C1.ci.lower          <- C1.ci.lower
  object$C1.ci.upper          <- C1.ci.upper
  object$C2.ci.lower          <- C2.ci.lower
  object$C2.ci.upper          <- C2.ci.upper
  object$coefficients         <- coefficients

  ## Add NMF-FFB inference class on top of the legacy SEM class.
  ## Final class vector for a typical input:
  ##   c("nmf.ffb.inference", "nmf.sem.inference", "nmf.ffb", "nmf.sem")
  ## Existing S3 methods (e.g. summary.nmf.sem) still dispatch via
  ## inheritance.
  if (!inherits(object, "nmf.sem.inference"))
    class(object) <- c("nmf.sem.inference", class(object))
  if (!inherits(object, "nmf.ffb.inference"))
    class(object) <- c("nmf.ffb.inference", class(object))
  object
}


#' @title Cross-Validation for NMF-FFB
#' @description
#' Performs K-fold cross-validation to evaluate the equilibrium mapping of
#' the NMF-FFB model.
#'
#' For each fold, \code{nmf.sem} is fitted on the training samples,
#' yielding an equilibrium mapping \eqn{\hat Y_1 = M_{\mathrm{model}} Y_2}.
#' The held-out endogenous variables \eqn{Y_1} are then predicted from \eqn{Y_2}
#' using this mapping, and the mean absolute error (MAE) over all entries in the
#' test block is computed. The returned value is the average MAE across folds.
#'
#' This implements the hyperparameter selection strategy described in the paper:
#' hyperparameters are chosen by predictive cross-validation rather than direct
#' inspection of the internal structural matrices.
#'
#' @param Y1 A non-negative numeric matrix of endogenous variables with
#'   \strong{rows = variables (P1), columns = samples (N)}.
#' @param Y2 A non-negative numeric matrix of exogenous variables with
#'   \strong{rows = variables (P2), columns = samples (N)}.
#'   Must satisfy \code{ncol(Y1) == ncol(Y2)}.
#' @param rank Integer; rank (number of latent factors) passed to \code{nmf.sem}.
#'   If \code{NULL}, \code{nmf.sem} decides the effective rank (via \code{...} or \code{nrow(Y2)}).
#' @param X.init Initialization strategy for \code{X}, forwarded to
#'   \code{\link{nmf.sem}}.  One of \code{"nndsvd"} (default),
#'   \code{"kmeans"}, \code{"kmeansar"}, \code{"runif"}, a numeric
#'   \eqn{P_1 \times Q} matrix, or \code{NULL} (alias for
#'   \code{"nndsvd"}).  See \code{\link{nmf.sem}} for details.
#' @param X.L2.ortho L2 orthogonality penalty for \code{X}.
#' @param C1.L1 L1 sparsity penalty for \code{C1} (\eqn{\Theta_1}).
#' @param C2.L1 L1 sparsity penalty for \code{C2} (\eqn{\Theta_2}).
#' @param epsilon Convergence threshold for \code{nmf.sem}.
#' @param maxit Maximum number of iterations for \code{nmf.sem}.
#' @param ... Additional arguments passed to \code{nmf.sem} (except for
#'   \code{rank}, \code{seed}, \code{div}, \code{shuffle}, which are handled here).
#'   Also accepts: \code{nfolds} (number of folds, default 5; \code{div} also accepted),
#'   \code{seed} (master random seed, default \code{NULL}),
#'   \code{shuffle} (logical, default \code{TRUE}).
#'
#' @return A numeric scalar: mean MAE across CV folds.
#'
#' @examples
#' Y <- t(iris[, -5])
#' Y1 <- Y[1:2, ]
#' Y2 <- Y[3:4, ]
#' mae <- nmf.sem.cv(Y1, Y2, rank = 2, maxit = 500, nfolds = 3)
#' mae
#'
#' @seealso \code{\link{nmf.sem}}
#' @export
nmf.sem.cv <- function(
    Y1, Y2,
    rank = NULL,
    X.init = "nndsvd",
    X.L2.ortho = 100.0,
    C1.L1 = 1.0,        # L1 sparsity for C1 (Theta1)
    C2.L1 = 0.1,        # L1 sparsity for C2 (Theta2)
    epsilon = 1e-6,     # Convergence tolerance passed to nmf.sem
    maxit = 5000,
    ...
){
  extra_cv <- base::list(...)
  nfolds  <- if (!is.null(extra_cv$nfolds))  extra_cv$nfolds  else if (!is.null(extra_cv$div)) extra_cv$div else 5
  seed    <- if (!is.null(extra_cv$seed))    extra_cv$seed    else NULL
  shuffle <- if (!is.null(extra_cv$shuffle)) extra_cv$shuffle else TRUE
  div <- nfolds
  # ------------------------------------------------------------------
  # 1. Basic input checks
  #
  # NMF-FFB requires non-negative matrices. We also require that Y1 and Y2
  # share the same number of samples (columns) to allow paired CV splits.
  # ------------------------------------------------------------------
  if (!is.matrix(Y1)) Y1 <- as.matrix(Y1)
  if (!is.matrix(Y2)) Y2 <- as.matrix(Y2)
  if (any(!is.finite(Y1)) || any(!is.finite(Y2))) stop("Y1 and Y2 must not contain NA/NaN/Inf.")
  div <- as.integer(div)
  if (min(Y1) < 0 || min(Y2) < 0) stop("Y1 and Y2 must be non-negative.")
  if (ncol(Y1) != ncol(Y2)) {
    stop("Y1 and Y2 must have the same number of columns (samples).")
  }

  P1 <- nrow(Y1)
  P2 <- nrow(Y2)
  N  <- ncol(Y1)

  if (div < 2L) {
    stop("div (number of CV folds) must be >= 2.")
  }
  if (div > N) {
    stop("div (number of CV folds) must be <= number of samples.")
  }

  # ------------------------------------------------------------------
  # 2. Handle extra arguments for nmf.sem
  #
  # We collect additional arguments in 'extra_args' but explicitly remove
  # those that are managed at the CV level:
  #   - div, shuffle : used only here, not passed to nmf.sem.
  #   - rank        : passed explicitly from nmf.sem.cv.
  #   - seed        : fold-specific seeds are generated here.
  # ------------------------------------------------------------------
  extra_args <- list(...)

  extra_args$div     <- NULL
  extra_args$nfolds  <- NULL
  extra_args$shuffle <- NULL
  extra_args$rank    <- NULL
  extra_args$seed    <- NULL

  # ------------------------------------------------------------------
  # 3. Set RNG for CV partition and per-fold seeds
  #
  # If a master 'seed' is given:
  #   - it is used to define the CV partition (sample permutation),
  #   - independent seeds for each nmf.sem run are drawn.
  # If 'seed' is NULL:
  #   - CV partition uses the current RNG state,
  #   - nmf.sem runs use whatever the global RNG state is at call time.
  # ------------------------------------------------------------------
  if (!is.null(seed)) {
    set.seed(seed)
  }

  # Sample indices for fold division
  if (shuffle) {
    perm_index <- sample.int(N)
  } else {
    perm_index <- seq_len(N)
  }

  # Per-fold seeds for nmf.sem (optional, only if master seed specified)
  if (!is.null(seed)) {
    seeds_fold <- sample.int(.Machine$integer.max, div)
  } else {
    seeds_fold <- rep(NA_integer_, div)
  }

  # ------------------------------------------------------------------
  # 4. Create CV folds
  #
  # We assign approximately N/div samples to each fold, distributing
  # any remainder one-by-one to the earliest folds.
  # 'block[i] = k' means sample i belongs to fold k (as test set).
  # ------------------------------------------------------------------
  remainder <- N %% div
  division  <- N %/% div
  block     <- integer(N)

  processed_count <- 0L
  for (i in 1:(div - 1L)) {
    plus       <- ifelse(i <= remainder, 1L, 0L)
    chunk_size <- division + plus
    idx_range  <- (processed_count + 1L):(processed_count + chunk_size)
    target_idx <- perm_index[idx_range]
    block[target_idx] <- i
    processed_count   <- processed_count + chunk_size
  }
  # Last fold gets all remaining samples
  target_idx <- perm_index[(processed_count + 1L):N]
  block[target_idx] <- div

  # Per-fold CV loss (MAE) will be stored here
  objfunc.block <- numeric(div)

  # ------------------------------------------------------------------
  # 5. Cross-validation loop
  #
  # For each fold j:
  #   - train on all samples not in fold j,
  #   - test on samples in fold j,
  #   - fit nmf.sem on training data,
  #   - compute MAE on test block from equilibrium mapping M.model.
  # ------------------------------------------------------------------
  for (j in 1:div) {
    # Train / Test split
    train_idx <- block != j
    test_idx  <- block == j

    Y1_train <- Y1[, train_idx, drop = FALSE]
    Y1_test  <- Y1[, test_idx,  drop = FALSE]
    Y2_train <- Y2[, train_idx, drop = FALSE]
    Y2_test  <- Y2[, test_idx,  drop = FALSE]

    # Fold-specific seed for nmf.sem
    seed_j <- if (!is.null(seed)) seeds_fold[j] else NULL

    # Assemble arguments for nmf.sem
    nmf.sem.args <- c(
      extra_args,   # User-specified additional arguments (e.g., Q)
      list(
        Y1         = Y1_train,
        Y2         = Y2_train,
        rank       = rank,
        X.init     = X.init,
        X.L2.ortho = X.L2.ortho,
        C1.L1      = C1.L1,
        C2.L1      = C2.L1,
        epsilon    = epsilon,
        maxit      = maxit
      )
    )
    # Attach seed only when it is defined
    if (!is.null(seed_j)) {
      nmf.sem.args$seed <- seed_j
    }

    # Call nmf.sem on the training data (suppress messages for cleaner CV output)
    res_j <- suppressMessages(do.call("nmf.sem", nmf.sem.args))

    # If mapping is not usable, penalize this fold (do not crash CV)
    if (is.null(res_j$M.model) || any(!is.finite(res_j$M.model))) {
      objfunc.block[j] <- Inf
      next
    }
    Pre_test <- res_j$M.model %*% Y2_test
    if (any(!is.finite(Pre_test))) {
      objfunc.block[j] <- Inf
      next
    }
    objfunc.block[j] <- mean(abs(Y1_test - Pre_test))
  }

  # ------------------------------------------------------------------
  # 6. Aggregate CV score
  #
  # The overall CV criterion is the average MAE across all folds.
  # This is typically minimized over hyperparameter grids
  # (e.g., rank, X.L2.ortho, C1.L1, C2.L1) when tuning NMF-FFB.
  # ------------------------------------------------------------------
  objfunc <- mean(objfunc.block)
  return(objfunc)
}



#' @title Heuristic Variable Splitting for NMF-FFB
#'
#' @description
#' Infers a heuristic partition of observed variables into exogenous (\eqn{Y_2})
#' and endogenous (\eqn{Y_1}) blocks for use in NMF-FFB.
#' The method is based on positive-SEM logic, causal ordering, and optional
#' sign alignment using the first principal component (PC1).
#'
#' The procedure:
#' \itemize{
#'   \item internally standardizes variables (mean 0, sd 1),
#'   \item optionally flips signs so that most variables align positively with PC1,
#'   \item infers a causal ordering by repeatedly regressing each variable on the
#'         remaining ones and selecting the variable with the largest minimum
#'         standardized coefficient,
#'   \item determines an exogenous block by scanning the ordering from upstream
#'         and stopping at the first variable whose strongest parent coefficient
#'         exceeds \code{threshold}.
#' }
#'
#' If \code{n.exogenous} is supplied, it overrides the automatic threshold rule.
#'
#' @param x A numeric matrix or data frame with
#'   \strong{rows = samples} and \strong{columns = observed variables}.
#' @param n.exogenous Optional integer specifying the number of exogenous variables
#'   (\eqn{Y_2}). If \code{NULL}, the number is inferred automatically by the
#'   coefficient cut-off rule.
#' @param threshold Standardized regression-coefficient threshold used in the
#'   automatic exogenous–endogenous split. A variable is treated as endogenous
#'   once its maximum standardized parent coefficient exceeds this value.
#'   (Default: \code{0.1})
#' @param auto.flipped Logical; if \code{TRUE}, applies PC1-based automatic
#'   sign flipping after standardization to ensure consistent orientation.
#'   (Default: \code{TRUE})
#' @param verbose Logical; if \code{TRUE}, prints progress messages and the
#'   resulting variable split. (Default: \code{FALSE})
#' @param ... Reserved for future use; currently unused (also accepted
#'   by the \code{\link{nmf.ffb.split}} alias for argument forwarding).
#'
#' @return A list with:
#'   \item{endogenous.variables}{
#'     Character vector of variables selected as endogenous (\eqn{Y_1}).}
#'   \item{exogenous.variables}{
#'     Character vector of variables selected as exogenous (\eqn{Y_2}).}
#'   \item{ordered.variables}{
#'     Variables in inferred causal order (from exogenous to endogenous).}
#'   \item{is.flipped}{
#'     Logical vector indicating which variables were sign-flipped during processing.}
#'   \item{n.exogenous}{
#'     Integer giving the number of exogenous variables.}
#'
#' @examples
#' # Infer exogenous/endogenous split from iris
#' sp <- nmf.sem.split(iris[, -5], n.exogenous = 2)
#' sp$endogenous.variables
#' sp$exogenous.variables
#'
#' @seealso \code{\link{nmf.sem}}
#' @export
nmf.sem.split <- function(x, n.exogenous = NULL, threshold = 0.1,
                          auto.flipped = TRUE, verbose = FALSE) {

  if (!is.matrix(x) && !is.data.frame(x))
    stop("x must be a numeric matrix or data frame.")

  X_raw <- as.matrix(x)
  P <- ncol(X_raw)
  col_names <- colnames(X_raw)
  if (is.null(col_names)) {
    col_names <- paste0("V", 1:P)
    colnames(X_raw) <- col_names
  }

  # --------------------------------------------------------------------
  # Preprocessing Step 1: Standardize all variables
  #
  # Variables are centered and scaled (mean 0, sd 1). NMF-FFB requires
  # non-negative matrices, but the purpose of this function is only to
  # infer variable roles (Y1/Y2), so standardized values are allowed.
  #
  # Missing or NaN values resulting from constant columns are set to 0.
  # --------------------------------------------------------------------
  X_calc <- scale(X_raw, center = TRUE, scale = TRUE)
  X_calc[is.na(X_calc)] <- 0
  X_calc[is.nan(X_calc)] <- 0

  all_indices <- 1:P

  # --------------------------------------------------------------------
  # Preprocessing Step 2: Optional sign flipping based on PC1 alignment
  #
  # In positive SEM (and NMF-FFB), variables should ideally have
  # consistent sign orientation. To enforce this heuristic, variables
  # negatively correlated with the first principal component are flipped.
  #
  # This stabilizes the causal-ordering heuristic by avoiding mixtures
  # of arbitrary sign conventions in the raw data.
  # --------------------------------------------------------------------
  is.flipped <- rep(FALSE, P)
  names(is.flipped) <- col_names

  if (auto.flipped) {
    if (verbose) message("Step 0: Checking correlations with PC1 (on standardized data)...")

    svd_res <- svd(X_calc)
    pc1 <- svd_res$u[, 1]

    cors <- stats::cor(X_calc, pc1)
    # Ensure majority alignment with PC1
    if (stats::median(cors, na.rm = TRUE) < 0) cors <- -cors

    flip_idx <- which(cors < 0)

    if (length(flip_idx) > 0) {
      is.flipped[flip_idx] <- TRUE
      X_calc[, flip_idx] <- -X_calc[, flip_idx]

      if (verbose) {
        message(sprintf("   -> Detected %d flipped variables: %s",
                    length(flip_idx), paste(col_names[flip_idx], collapse=", ")))
      }
    }
  }

  # --------------------------------------------------------------------
  # Step 1: Causal ordering heuristic
  #
  # We infer an ordering of variables consistent with positive-SEM logic:
  # repeatedly select the variable that has the *largest minimum* coefficient
  # when regressed on the remaining variables. This favors variables that
  # are least explained by others → likely exogenous.
  #
  # The resulting order approximates a causal topological order in which
  # exogenous variables appear early and endogenous variables later.
  # --------------------------------------------------------------------
  if (verbose) message("Step 1: Inferring Causal Ordering...")

  active_set <- all_indices
  ordering_reversed <- integer(P)

  for (t in 1:(P - 1)) {
    scores <- numeric(length(active_set))

    for (i in seq_along(active_set)) {
      target_col <- active_set[i]
      pred_cols <- active_set[-i]

      y_vec <- X_calc[, target_col]
      X_mat <- X_calc[, pred_cols, drop = FALSE]

      coefs <- tryCatch({
        stats::coef(stats::lm(y_vec ~ X_mat - 1))
      }, error = function(e) rep(NA_real_, length(pred_cols)))

      # Positive SEM → keep the smallest coefficient (weakest positive predictor)
      if (all(is.na(coefs))) {
        scores[i] <- -Inf
      } else {
        scores[i] <- min(coefs, na.rm = TRUE)
      }
    }

    best_idx <- which.max(scores)
    ordering_reversed[t] <- active_set[best_idx]
    active_set <- active_set[-best_idx]
  }
  ordering_reversed[P] <- active_set[1]

  # Causal order: exogenous → endogenous
  ordering_indices <- rev(ordering_reversed)

  # --------------------------------------------------------------------
  # Step 2: Automatic identification of exogenous variables
  #
  # Sweep through the causal ordering. For each variable, regress it on all
  # earlier variables. If its strongest parent coefficient exceeds the
  # threshold, the variable is considered endogenous.
  #
  # Variables before this point → exogenous (Y2)
  # Variables after this point → endogenous (Y1)
  #
  # If n.exogenous is given, it overrides this automatic rule.
  # --------------------------------------------------------------------
  if (is.null(n.exogenous)) {
    if (verbose) message("Step 2: Detecting optimal cut-off for exogenous variables...")
    cutoff <- 1

    for (k in seq_len(max(P - 2, 0)) + 1) {
      curr_idx <- ordering_indices[k]
      parent_indices <- ordering_indices[seq_len(k - 1)]

      y_vec <- X_calc[, curr_idx]
      if (length(parent_indices) == 0) next
      X_parents <- X_calc[, parent_indices, drop = FALSE]

      coefs <- stats::coef(stats::lm(y_vec ~ X_parents - 1))
      coefs <- coefs[is.finite(coefs)]
      max_influence <- if (length(coefs) > 0) max(coefs) else 0

      if (max_influence > threshold) {
        if (verbose)
          message(sprintf("   [%d] %s : Max std.coef=%.3f -> Endogenous (Stop)",
                      k, col_names[curr_idx], max_influence))
        break
      } else {
        cutoff <- k
        if (verbose)
          message(sprintf("   [%d] %s : Max std.coef=%.3f -> Exogenous (Continue)",
                      k, col_names[curr_idx], max_influence))
      }
    }
    n.exogenous <- cutoff
  }

  # --------------------------------------------------------------------
  # Step 3: Final classification into Y1 and Y2
  #
  # Variables appearing early in the ordering (determined by cut-off) are
  # treated as exogenous (Y2). The remainder are endogenous (Y1).
  #
  # Ordered list shows the full inferred causal sequence.
  # --------------------------------------------------------------------
  if(n.exogenous < 1 || n.exogenous >= P)
    stop("n.exogenous must be between 1 and P-1.")
  idx_exo <- ordering_indices[1:n.exogenous]
  idx_endo <- ordering_indices[(n.exogenous + 1):P]

  exogenous.variables <- col_names[idx_exo]
  endogenous.variables <- col_names[idx_endo]
  ordered.variables <- col_names[ordering_indices]

  if (verbose) {
    message("\n--- Auto Split Result ---")
    message(sprintf("Exogenous (Y2, n=%d): %s",
                n.exogenous, paste(exogenous.variables, collapse=", ")))
    message(sprintf("Endogenous (Y1, n=%d): %s ...",
                length(endogenous.variables),
                paste(utils::head(endogenous.variables, 3), collapse=", ")))
  }

  return(list(
    endogenous.variables = endogenous.variables,
    exogenous.variables = exogenous.variables,
    ordered.variables = ordered.variables,
    is.flipped = is.flipped,
    n.exogenous = as.integer(n.exogenous)
  ))
}




############################################################
## Common DOT Helpers
############################################################

############################################################
## 1. nmf.sem.DOT  (for NMF-FFB visualization)
############################################################

#' Generate a Graphviz DOT Diagram for an NMF-FFB Model
#'
#' @description
#' Creates a Graphviz DOT script that visualizes the structural network
#' estimated by \code{nmf.sem}.
#' The resulting diagram displays:
#' \itemize{
#'   \item endogenous observed variables (\eqn{Y_1}),
#'   \item exogenous observed variables (\eqn{Y_2}),
#'   \item latent factors (\eqn{F_1}, \dots, \eqn{F_Q}),
#' }
#' together with the non-negative path coefficients whose magnitudes
#' exceed a user-specified threshold.
#'
#' Directed edges represent estimated relationships:
#' \itemize{
#'   \item \eqn{Y_2 \rightarrow F_q}: entries of \code{C2} (exogenous loadings),
#'   \item \eqn{F_q \rightarrow Y_1}: rows of \code{X} (factor-to-endogenous mappings),
#'   \item \eqn{Y_1 \rightarrow F_q}: entries of \code{C1} (feedback paths).
#' }
#'
#' Edge widths are scaled by coefficient magnitude, and nodes are placed
#' in optional visual clusters. Only variables participating in
#' edges above the threshold are displayed, while latent factors are always shown.
#'
#' @param result A list returned by \code{nmf.sem}, containing matrices
#'   \code{X}, \code{C1}, and \code{C2}.
#' @param weight_scale Base scaling factor for edge widths.
#' @param weight_scale_c2 Scaling factor for edges
#'   \eqn{Y_2 \rightarrow F_q} (C2 matrix). Defaults to \code{weight_scale}.
#' @param weight_scale_x1 Scaling factor for edges
#'   \eqn{F_q \rightarrow Y_1} (X matrix). Defaults to \code{weight_scale}.
#' @param weight_scale_feedback Scaling factor for feedback edges
#'   \eqn{Y_1 \rightarrow F_q} (C1 matrix). Defaults to \code{weight_scale}.
#' @param threshold Minimum coefficient value needed for an edge to be drawn.
#' @param sig.level Significance level for filtering edges by p-value
#'   (requires inference results). Edges with p-value above this level are omitted.
#' @param rankdir Graphviz rank direction (e.g., \code{"LR"}, \code{"TB"}).
#' @param fill Logical; whether to use filled node shapes.
#' @param ... For backward compatibility: accepts deprecated names
#'   \code{weight_scale_y2f} (use \code{weight_scale_c2}) and
#'   \code{weight_scale_fy1} (use \code{weight_scale_x1}).
#' @param cluster.box Character string controlling the visibility and style
#'   of cluster frames around Y2, factors, and Y1 blocks.
#'   One of \code{"normal"}, \code{"faint"}, \code{"invisible"}, \code{"none"}.
#' @param cluster.labels Optional character vector of length 3 giving custom
#'   labels for the Y2, factor, and Y1 clusters.
#' @param hide.isolated Logical. If \code{TRUE} (default), Y1 and Y2 nodes
#'   that have no edges at or above \code{threshold} are excluded from the graph.
#' @param sig.level Significance level for filtering structural edges
#'   (\eqn{C_1} feedback and \eqn{C_2} exogenous loadings) when
#'   inference results are present.  If \code{result} contains a
#'   \code{coefficients} data frame from \code{\link{nmf.sem.inference}},
#'   only edges with \code{p_value < sig.level} are drawn, with
#'   significance stars (\code{*} \code{**} \code{***}) appended to
#'   the edge label.  The \eqn{X} (factor-to-\eqn{Y_1}) edges are
#'   never starred since the basis is not the inference target.
#'   Set to \code{NULL} to disable significance filtering and fall
#'   back to the \code{threshold} magnitude filter for both \eqn{C_1}
#'   and \eqn{C_2}.  Default is \code{0.1}.
#'
#' @return A character string representing a valid Graphviz DOT script.
#'
#' @examples
#' Y <- t(iris[, -5])
#' Y1 <- Y[1:2, ]
#' Y2 <- Y[3:4, ]
#' result <- nmf.sem(Y1, Y2, rank = 2, maxit = 500)
#' dot <- nmf.sem.DOT(result)
#' cat(dot)
#'
#' @seealso \code{\link{nmf.sem}}, \code{\link{nmf.sem.inference}},
#'   \code{\link{plot.nmfkc.DOT}}
#' @export
nmf.sem.DOT <- function(result,
                        weight_scale          = 5,
                        weight_scale_c2       = weight_scale,
                        weight_scale_x1       = weight_scale,
                        weight_scale_feedback = weight_scale,
                        threshold             = 0.01,
                        sig.level             = 0.1,
                        rankdir               = "LR",
                        fill                  = TRUE,
                        cluster.box           = c("normal", "faint", "invisible", "none"),
                        cluster.labels        = NULL,
                        hide.isolated         = TRUE,
                        ...) {

  ## Backward compatibility: accept deprecated names via ...
  extra_args <- base::list(...)
  if (!base::is.null(extra_args$weight_scale_y2f)) weight_scale_c2 <- extra_args$weight_scale_y2f
  if (!base::is.null(extra_args$weight_scale_fy1)) weight_scale_x1 <- extra_args$weight_scale_fy1

  ## ---------------------------------------------------------------
  ## Cluster style selection
  ## ---------------------------------------------------------------
  cluster.box <- match.arg(cluster.box)

  cluster_style <- switch(cluster.box,
                          normal    = "rounded",
                          faint     = "rounded,dashed",
                          invisible = "rounded",
                          none      = "none")

  cluster_color <- switch(cluster.box,
                          normal    = "black",
                          faint     = "gray80",
                          invisible = "none",
                          none      = "none")

  cluster_penwidth <- switch(cluster.box,
                             normal    = 1.0,
                             faint     = 0.7,
                             invisible = 0.0,
                             none      = 0.0)

  ## ---------------------------------------------------------------
  ## Cluster titles
  ## ---------------------------------------------------------------
  default.labels <- c("Exogenous (Y2)", "Latent Factors", "Endogenous (Y1)")
  if (is.null(cluster.labels)) {
    titles <- default.labels
  } else {
    if (!is.character(cluster.labels) || length(cluster.labels) != 3L) {
      stop("cluster.labels must be a character vector of length 3: c(label_Y2, label_F, label_Y1).")
    }
    titles <- cluster.labels
  }

  ## ---------------------------------------------------------------
  ## Extract matrices and labels
  ## ---------------------------------------------------------------
  X  <- result$X
  C1 <- result$C1
  C2 <- result$C2

  if (is.null(X) || is.null(C1) || is.null(C2)) {
    stop("result must contain elements X, C1, and C2.")
  }

  Y1_labels <- rownames(X)
  Y2_labels <- colnames(C2)

  if (is.null(Y1_labels)) Y1_labels <- paste0("Y1_", seq_len(nrow(X)))
  if (is.null(Y2_labels)) Y2_labels <- paste0("Y2_", seq_len(ncol(C2)))

  P1 <- length(Y1_labels)
  P2 <- length(Y2_labels)
  Q  <- ncol(X)

  ## ---------------------------------------------------------------
  ## Identify nodes involved in edges >= threshold
  ## ---------------------------------------------------------------
  if (isTRUE(hide.isolated)) {
    used_y2 <- apply(C2, 2L, function(col) any(col >= threshold, na.rm = TRUE))
    used_y1_from_X  <- apply(X,  1L, function(row) any(row >= threshold, na.rm = TRUE))
    used_y1_from_C1 <- apply(C1, 2L, function(col) any(col >= threshold, na.rm = TRUE))
    used_y1 <- used_y1_from_X | used_y1_from_C1
    idx_y1 <- which(used_y1)
    idx_y2 <- which(used_y2)
  } else {
    idx_y1 <- seq_len(P1)
    idx_y2 <- seq_len(P2)
  }

  ## ---------------------------------------------------------------
  ## Assign internal DOT node IDs
  ## ---------------------------------------------------------------
  Y1_ids <- .nmfkc_dot_sanitize_id(paste0("Y1_", seq_len(P1)))
  Y2_ids <- .nmfkc_dot_sanitize_id(paste0("Y2_", seq_len(P2)))
  F_ids  <- .nmfkc_dot_sanitize_id(paste0("F_",  seq_len(Q)))

  ## Node colors
  COLOR_Y2_NODE     <- "lightcoral"
  COLOR_Y1_NODE     <- "lightblue"
  COLOR_FACTOR_NODE <- "wheat"

  ## ---------------------------------------------------------------
  ## Header
  ## ---------------------------------------------------------------
  dot_script <- .nmfkc_dot_header(
    graph_name = "NMF_SEM_Full_Mechanism",
    rankdir    = rankdir
  )

  ## ---------------------------------------------------------------
  ## Y2 cluster
  ## ---------------------------------------------------------------
  dot_script <- paste0(
    dot_script,
    "\n  // Exogenous variables (Y2)\n",
    .nmfkc_dot_cluster_nodes(
      cluster_id        = "Y2",
      title             = titles[1],
      node_ids          = Y2_ids[idx_y2],
      node_labels       = Y2_labels[idx_y2],
      shape             = "box",
      fill              = fill,
      fillcolor         = COLOR_Y2_NODE,
      line_width        = 1.5,
      cluster_style     = cluster_style,
      cluster_color     = cluster_color,
      cluster_penwidth  = cluster_penwidth
    )
  )

  ## ---------------------------------------------------------------
  ## Y1 cluster
  ## ---------------------------------------------------------------
  dot_script <- paste0(
    dot_script,
    "\n  // Endogenous variables (Y1)\n",
    .nmfkc_dot_cluster_nodes(
      cluster_id        = "Y1",
      title             = titles[3],
      node_ids          = Y1_ids[idx_y1],
      node_labels       = Y1_labels[idx_y1],
      shape             = "box",
      fill              = fill,
      fillcolor         = COLOR_Y1_NODE,
      line_width        = 1.5,
      cluster_style     = cluster_style,
      cluster_color     = cluster_color,
      cluster_penwidth  = cluster_penwidth
    )
  )

  ## ---------------------------------------------------------------
  ## Latent factor cluster
  ## ---------------------------------------------------------------
  dot_script <- paste0(
    dot_script,
    "\n  // Latent Factors (F)\n",
    .nmfkc_dot_cluster_nodes(
      cluster_id        = "F",
      title             = titles[2],
      node_ids          = F_ids,
      node_labels       = paste0("Factor ", seq_len(Q)),
      shape             = "ellipse",
      fill              = fill,
      fillcolor         = COLOR_FACTOR_NODE,
      line_width        = 1.0,
      cluster_style     = cluster_style,
      cluster_color     = cluster_color,
      cluster_penwidth  = cluster_penwidth
    )
  )

  ## ---------------------------------------------------------------
  ## Edge defaults
  ## ---------------------------------------------------------------
  dot_script <- paste0(
    dot_script,
    '\n  edge [fontname="Arial", fontsize=8, arrowhead=open];\n'
  )

  pw     <- .nmfkc_dot_penwidth
  digits <- .nmfkc_dot_digits_from_threshold(threshold)
  fmtc   <- function(x) .nmfkc_dot_format_coef(x, digits)

  ## ---------------------------------------------------------------
  ## Significance stars for C1 (feedback) and C2 (exogenous) edges,
  ## both from nmf.sem.inference().  X (F -> Y1) edges are NOT
  ## starred even when inference results are present, since the
  ## basis is not the inference target.
  ##
  ## Filters by the optional `Type` column ("C1" / "C2") if present
  ## (newer inference output); falls back to row-name matching for
  ## back-compatibility with pre-v0.6.8 inference output that only
  ## carried C2.
  ## ---------------------------------------------------------------
  pval_to_stars <- function(p) {
    if (!is.finite(p)) ""
    else if (p < 0.001) "***"
    else if (p < 0.01)  "**"
    else if (p < 0.05)  "*"
    else                ""
  }

  C1_stars <- NULL; C1_show <- NULL
  C2_stars <- NULL; C2_show <- NULL
  if (!is.null(result$coefficients)) {
    cf        <- result$coefficients
    fac_names <- rownames(C2)
    y1_names  <- colnames(C1)
    y2_names  <- colnames(C2)

    has_type <- !is.null(cf$Type)
    cf_C1 <- if (has_type) cf[cf$Type == "C1", , drop = FALSE] else NULL
    cf_C2 <- if (has_type) cf[cf$Type == "C2", , drop = FALSE] else cf

    ## ---- C2 stars (Y2 -> F) ----
    C2_stars <- matrix("", nrow = Q, ncol = P2)
    C2_pval  <- matrix(NA_real_, nrow = Q, ncol = P2)
    for (k in seq_len(nrow(cf_C2))) {
      q  <- match(cf_C2$Basis[k], fac_names)
      p2 <- match(cf_C2$Covariate[k], y2_names)
      if (!is.na(q) && !is.na(p2) && !is.na(cf_C2$p_value[k])) {
        C2_pval[q, p2]  <- cf_C2$p_value[k]
        C2_stars[q, p2] <- pval_to_stars(cf_C2$p_value[k])
      }
    }
    if (!is.null(sig.level)) {
      C2_show <- !is.na(C2_pval) & C2_pval < sig.level
    }

    ## ---- C1 stars (Y1 -> F, feedback) ----
    if (!is.null(cf_C1) && nrow(cf_C1) > 0L) {
      C1_stars <- matrix("", nrow = Q, ncol = P1)
      C1_pval  <- matrix(NA_real_, nrow = Q, ncol = P1)
      for (k in seq_len(nrow(cf_C1))) {
        q  <- match(cf_C1$Basis[k], fac_names)
        p1 <- match(cf_C1$Covariate[k], y1_names)
        if (!is.na(q) && !is.na(p1) && !is.na(cf_C1$p_value[k])) {
          C1_pval[q, p1]  <- cf_C1$p_value[k]
          C1_stars[q, p1] <- pval_to_stars(cf_C1$p_value[k])
        }
      }
      if (!is.null(sig.level)) {
        C1_show <- !is.na(C1_pval) & C1_pval < sig.level
      }
    }
  }

  ## ---------------------------------------------------------------
  ## 1. Y2 → F edges (C2)
  ## ---------------------------------------------------------------
  dot_script <- paste0(
    dot_script,
    '\n  // 1. External Driving (Y2 -> Factor) [C2]\n',
    '  edge [color=black, fontcolor=black, style=solid];\n'
  )

  max_C2 <- suppressWarnings(max(C2, na.rm = TRUE))
  if (is.finite(max_C2) && max_C2 > 0) {
    for (q in seq_len(Q)) {
      for (p2 in idx_y2) {
        weight <- C2[q, p2]
        show <- if (!is.null(C2_show)) C2_show[q, p2]
                else is.finite(weight) && weight >= threshold
        if (show) {
          pen <- pw(weight, max_C2, weight_scale_c2)
          lab <- fmtc(weight)
          if (!is.null(C2_stars)) lab <- paste0(lab, C2_stars[q, p2])
          path <- sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                          Y2_ids[p2], F_ids[q], lab, pen)
          dot_script <- paste0(dot_script, path)
        }
      }
    }
  }

  ## ---------------------------------------------------------------
  ## 2. F → Y1 edges (X)
  ## ---------------------------------------------------------------
  dot_script <- paste0(
    dot_script,
    '\n  // 2. Generation (Factor -> Y1) [X]\n',
    '  edge [color="gray0", fontcolor="gray0", style=solid];\n'
  )

  max_X <- suppressWarnings(max(X, na.rm = TRUE))
  if (is.finite(max_X) && max_X > 0) {
    for (q in seq_len(Q)) {
      for (p1 in idx_y1) {
        weight <- X[p1, q]
        if (is.finite(weight) && weight >= threshold) {
          pen <- pw(weight, max_X, weight_scale_x1)
          lab <- fmtc(weight)
          path <- sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                          F_ids[q], Y1_ids[p1], lab, pen)
          dot_script <- paste0(dot_script, path)
        }
      }
    }
  }

  ## ---------------------------------------------------------------
  ## 3. Y1 → F edges (C1 feedback)
  ## ---------------------------------------------------------------
  dot_script <- paste0(
    dot_script,
    '\n  // 3. Internal Feedback (Y1 -> Factor) [C1]\n',
    '  edge [style=dashed, color="gray0", fontcolor="gray0"];\n'
  )

  max_C1 <- suppressWarnings(max(C1, na.rm = TRUE))
  if (is.finite(max_C1) && max_C1 > 0) {
    for (q in seq_len(Q)) {
      for (p1 in idx_y1) {
        weight <- C1[q, p1]
        ## When inference results provide C1 p-values, prefer the
        ## significance-based filter (parallels the C2 branch above);
        ## otherwise fall back to the magnitude threshold.
        show <- if (!is.null(C1_show)) C1_show[q, p1]
                else is.finite(weight) && weight >= threshold
        if (show) {
          pen <- pw(weight, max_C1, weight_scale_feedback)
          lab <- fmtc(weight)
          if (!is.null(C1_stars)) lab <- paste0(lab, C1_stars[q, p1])
          path <- sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                          Y1_ids[p1], F_ids[q], lab, pen)
          dot_script <- paste0(dot_script, path)
        }
      }
    }
  }

  result <- paste0(dot_script, "}\n")
  class(result) <- c("nmf.sem.DOT", "nmfkc.DOT")
  result
}





############################################################
## 2. nmfkc.DOT  (Static NMF / NMF-with-covariates visualization)
############################################################

#' Generate Graphviz DOT Scripts for NMF or NMF-with-Covariates Models
#'
#' @description
#' Produces a Graphviz DOT script visualizing the structure of an NMF model
#' (\eqn{Y \approx X C A}) or its simplified forms.
#'
#' Supported visualization types:
#' \itemize{
#'   \item \code{"YX"} — Standard NMF view: latent factors \eqn{X} map to observations \eqn{Y}.
#'   \item \code{"YA"} — Direct regression view: covariates \eqn{A} map directly to \eqn{Y}
#'         using the combined coefficient matrix \eqn{X C}.
#'   \item \code{"YXA"} — Full tri-factorization: \eqn{A \rightarrow C \rightarrow X \rightarrow Y}.
#' }
#'
#' Edge widths are scaled by coefficient magnitude, and nodes with no edges
#' above the threshold are omitted from the visualization.
#'
#' @param result The return value from \code{nmfkc}, containing matrices
#'   \code{X}, \code{B}, and optionally \code{C}.
#' @param type Character string specifying the visualization style:
#'   one of \code{"YX"}, \code{"YA"}, \code{"YXA"}.
#' @param threshold Minimum coefficient magnitude to display an edge.
#' @param sig.level Significance level for filtering C edges when inference
#'   results are available (i.e., \code{x} is of class \code{"nmfkc.inference"}).
#'   Only edges with p-value below \code{sig.level} are shown, decorated with
#'   significance stars (\code{*}, \code{**}, \code{***}). Set to \code{NULL}
#'   to disable filtering and show all edges above \code{threshold}. Default is 0.1.
#' @param rankdir Graphviz rank direction (e.g., \code{"LR"}, \code{"TB"}).
#' @param fill Logical; whether nodes should be drawn with filled shapes.
#' @param weight_scale Base scaling factor for edge widths.
#' @param weight_scale_ax Scaling factor for edges \eqn{A \rightarrow X} (type \code{"YXA"}).
#' @param weight_scale_xy Scaling factor for edges \eqn{X \rightarrow Y}.
#' @param weight_scale_ay Scaling factor for edges \eqn{A \rightarrow Y} (type \code{"YA"}).
#' @param Y.label Optional character vector for labels of Y nodes.
#' @param X.label Optional character vector for labels of X (latent factor) nodes.
#' @param A.label Optional character vector for labels of A (covariate) nodes.
#' @param Y.title Cluster title for Y nodes.
#' @param X.title Cluster title for X nodes.
#' @param A.title Cluster title for A nodes.
#' @param hide.isolated Logical. If \code{TRUE} (default), Y and A nodes that have no
#'   edges at or above \code{threshold} are excluded from the graph.
#'
#' @return A character string representing a Graphviz DOT script.
#'
#' @seealso \code{\link{nmfkc}}, \code{\link{nmfae.DOT}}, \code{\link{nmf.sem.DOT}},
#'   \code{\link{nmfkc.ar.DOT}}, \code{\link{plot.nmfkc.DOT}}
#' @examples
#' Y <- matrix(cars$dist, nrow = 1)
#' A <- rbind(1, cars$speed)
#' result <- nmfkc(Y, A, rank = 1)
#' dot <- nmfkc.DOT(result)
#' cat(dot)
#'
#' @export
nmfkc.DOT <- function(
    result,
    type = c("YX","YA","YXA"),
    threshold = 0.01,
    sig.level = 0.1,
    rankdir   = "LR",
    fill      = TRUE,
    weight_scale    = 5,
    weight_scale_ax = weight_scale,
    weight_scale_xy = weight_scale,
    weight_scale_ay = weight_scale,
    Y.label = NULL, X.label = NULL, A.label = NULL,
    Y.title = "Observation (Y)",
    X.title = "Basis (X)",
    A.title = "Covariates (A)",
    hide.isolated = TRUE
) {

  type <- match.arg(type)

  ## ---------------------------------------------------------
  ## Required matrices
  ## ---------------------------------------------------------
  X <- result$X
  B <- result$B
  if (is.null(X) || is.null(B)) {
    stop("result must contain X and B.")
  }

  ## If C exists and is a proper NMF-with-covariates factor:
  hasA <- !is.null(result$C) && ncol(result$C) != ncol(B)

  P <- nrow(X)
  Q <- ncol(X)

  ## ---------------------------------------------------------
  ## Labels
  ## ---------------------------------------------------------
  Y_labels <- if (is.null(Y.label)) rownames(X) else Y.label
  X_labels <- if (is.null(X.label)) colnames(X) else X.label

  if (is.null(Y_labels)) Y_labels <- paste0("Y", seq_len(P))
  if (is.null(X_labels)) X_labels <- paste0("Factor", seq_len(Q))

  Y_ids <- .nmfkc_dot_sanitize_id(paste0("Y_", seq_len(P)))
  X_ids <- .nmfkc_dot_sanitize_id(paste0("X_", seq_len(Q)))

  ## ---------------------------------------------------------
  ## Covariates and tri-factorization
  ## ---------------------------------------------------------
  if (hasA) {
    C <- as.matrix(result$C)          # Q x R
    A_cols <- ncol(C)
    A_labels <- if (is.null(A.label)) colnames(C) else A.label
    if (is.null(A_labels)) A_labels <- paste0("A", seq_len(A_cols))
    A_ids <- .nmfkc_dot_sanitize_id(paste0("A_", seq_len(A_cols)))

    ## Combined mapping for type = "YA"
    XC_mat <- X %*% C   # P x R

  } else if (type == "YX") {
    ## No A block needed
    C <- NULL
    A_cols <- 0L
    A_labels <- NULL
    A_ids <- NULL
    XC_mat <- NULL

  } else {
    stop("The model structure (matrix C) is incompatible with type 'YXA' or 'YA'.")
  }

  ## ---------------------------------------------------------
  ## Filter isolated nodes (hide.isolated)
  ## ---------------------------------------------------------
  idx_Y <- seq_len(P)
  idx_A <- if (hasA) seq_len(A_cols) else integer(0)
  if (isTRUE(hide.isolated)) {
    if (type == "YX" || type == "YXA") {
      # Y is connected via X: X[i, ] >= threshold
      used_Y <- apply(X, 1L, function(row) any(row >= threshold, na.rm = TRUE))
      idx_Y <- which(used_Y)
    } else if (type == "YA" && !is.null(XC_mat)) {
      # Y is connected via XC: XC_mat[i, ] >= threshold
      used_Y <- apply(XC_mat, 1L, function(row) any(row >= threshold, na.rm = TRUE))
      idx_Y <- which(used_Y)
    }
    if (hasA && type %in% c("YXA", "YA")) {
      if (type == "YXA" && !is.null(C)) {
        # A is connected via C: C[, k] >= threshold
        used_A <- apply(C, 2L, function(col) any(col >= threshold, na.rm = TRUE))
        idx_A <- which(used_A)
      } else if (type == "YA" && !is.null(XC_mat)) {
        # A is connected via XC: XC_mat[, k] >= threshold
        used_A <- apply(XC_mat, 2L, function(col) any(col >= threshold, na.rm = TRUE))
        idx_A <- which(used_A)
      }
    }
  }

  ## ---------------------------------------------------------
  ## DOT header
  ## ---------------------------------------------------------
  scr <- .nmfkc_dot_header(graph_name = "NMF", rankdir = rankdir)

  ## ---------------------------------------------------------
  ## Y node cluster
  ## ---------------------------------------------------------
  scr <- paste0(
    scr,
    '\n  // Output variables (Y)\n',
    .nmfkc_dot_cluster_nodes(
      cluster_id  = "Y",
      title       = Y.title,
      node_ids    = Y_ids[idx_Y],
      node_labels = Y_labels[idx_Y],
      shape       = "box",
      fill        = fill,
      fillcolor   = "lightblue",
      line_width  = 1.5
    )
  )

  ## ---------------------------------------------------------
  ## X node cluster
  ## (Hidden when type = "YA")
  ## ---------------------------------------------------------
  if (type != "YA") {
    scr <- paste0(
      scr,
      '\n  // Latent factors (X)\n',
      .nmfkc_dot_cluster_nodes(
        cluster_id  = "X",
        title       = X.title,
        node_ids    = X_ids,
        node_labels = X_labels,
        shape       = "ellipse",
        fill        = fill,
        fillcolor   = "wheat",
        line_width  = 1.0
      )
    )
  }

  ## ---------------------------------------------------------
  ## A node cluster
  ## (Only for "YA" and "YXA")
  ## ---------------------------------------------------------
  if (type != "YX" && hasA) {
    scr <- paste0(
      scr,
      '\n  // Covariates (A)\n',
      .nmfkc_dot_cluster_nodes(
        cluster_id  = "A",
        title       = A.title,
        node_ids    = A_ids[idx_A],
        node_labels = A_labels[idx_A],
        shape       = "box",
        fill        = fill,
        fillcolor   = "lightcoral",
        line_width  = 1.5
      )
    )
  }

  ## ---------------------------------------------------------
  ## Significance stars for C edges (from nmfkc.inference)
  ## ---------------------------------------------------------
  C_stars <- NULL
  C_show  <- NULL
  if (!is.null(result$coefficients) && hasA) {
    C_stars <- matrix("", nrow = Q, ncol = A_cols)
    C_pval  <- matrix(NA_real_, nrow = Q, ncol = A_cols)
    cf <- result$coefficients
    basis_names <- rownames(C)
    cov_names   <- colnames(C)
    for (k in seq_len(nrow(cf))) {
      q <- match(cf$Basis[k], basis_names)
      r <- match(cf$Covariate[k], cov_names)
      if (!is.na(q) && !is.na(r) && !is.na(cf$p_value[k])) {
        p <- cf$p_value[k]
        C_pval[q, r] <- p
        if (p < 0.001)      C_stars[q, r] <- "***"
        else if (p < 0.01)  C_stars[q, r] <- "**"
        else if (p < 0.05)  C_stars[q, r] <- "*"
      }
    }
    if (!is.null(sig.level)) {
      C_show <- !is.na(C_pval) & C_pval < sig.level
    }
  }

  ## ---------------------------------------------------------
  ## Edge defaults
  ## ---------------------------------------------------------
  scr <- paste0(
    scr,
    '\n  edge [fontname="Arial", fontsize=8, arrowhead=open];\n'
  )

  pw     <- .nmfkc_dot_penwidth
  digits <- .nmfkc_dot_digits_from_threshold(threshold)
  fmtc   <- function(x) .nmfkc_dot_format_coef(x, digits)


  ## =========================================================
  ## Case 1: Full tri-factorization (A → X → Y)
  ## =========================================================
  if (type == "YXA") {

    ## ---- A → X (C) edges ----
    scr <- paste0(
      scr,
      '\n  // A -> X edges (C)\n',
      '  edge [color=black, fontcolor=black, style=solid];\n'
    )

    max_C <- suppressWarnings(max(C, na.rm = TRUE))
    if (is.finite(max_C) && max_C > 0) {
      for (q in seq_len(Q)) {
        for (k in idx_A) {
          val <- C[q, k]
          show <- if (!is.null(C_show)) C_show[q, k]
                  else is.finite(val) && val >= threshold
          if (show) {
            pen <- pw(val, max_C, weight_scale_ax)
            lab <- fmtc(val)
            if (!is.null(C_stars)) lab <- paste0(lab, C_stars[q, k])
            scr <- paste0(
              scr,
              sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                      A_ids[k], X_ids[q], lab, pen)
            )
          }
        }
      }
    }

    ## ---- X → Y edges ----
    scr <- paste0(
      scr,
      '\n  // X -> Y edges (X)\n',
      '  edge [color="gray0", fontcolor="gray0", style=solid];\n'
    )

    max_X <- suppressWarnings(max(X, na.rm = TRUE))
    if (is.finite(max_X) && max_X > 0) {
      for (i in idx_Y) {
        for (j in seq_len(Q)) {
          val <- X[i, j]
          if (is.finite(val) && val >= threshold) {
            pen <- pw(val, max_X, weight_scale_xy)
            lab <- fmtc(val)
            scr <- paste0(
              scr,
              sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                      X_ids[j], Y_ids[i], lab, pen)
            )
          }
        }
      }
    }


    ## =========================================================
    ## Case 2: Direct regression view (A → Y)
    ## =========================================================
  } else if (type == "YA") {

    scr <- paste0(
      scr,
      '\n  // A -> Y edges (X %*% C)\n',
      '  edge [color=black, fontcolor=black, style=solid];\n'
    )

    max_XC <- suppressWarnings(max(XC_mat, na.rm = TRUE))
    if (is.finite(max_XC) && max_XC > 0) {
      for (i in idx_Y) {
        for (k in idx_A) {
          val <- XC_mat[i, k]
          if (is.finite(val) && val >= threshold) {
            pen <- pw(val, max_XC, weight_scale_ay)
            lab <- fmtc(val)
            scr <- paste0(
              scr,
              sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                      A_ids[k], Y_ids[i], lab, pen)
            )
          }
        }
      }
    }


    ## =========================================================
    ## Case 3: Standard NMF (X → Y)
    ## =========================================================
  } else if (type == "YX") {

    scr <- paste0(
      scr,
      '\n  // X -> Y edges (X)\n',
      '  edge [color="gray0", fontcolor="gray0", style=solid];\n'
    )

    max_X <- suppressWarnings(max(X, na.rm = TRUE))
    if (is.finite(max_X) && max_X > 0) {
      for (i in idx_Y) {
        for (j in seq_len(Q)) {
          val <- X[i, j]
          if (is.finite(val) && val >= threshold) {
            pen <- pw(val, max_X, weight_scale_xy)
            lab <- fmtc(val)
            scr <- paste0(
              scr,
              sprintf('  %s -> %s [label="%s", penwidth=%.2f];\n',
                      X_ids[j], Y_ids[i], lab, pen)
            )
          }
        }
      }
    }
  }

  result <- paste0(scr, "}\n")
  class(result) <- "nmfkc.DOT"
  result
}


#' @title Plot method for nmfkc.DOT objects
#' @description
#' Renders a DOT graph string using \code{DiagrammeR::grViz}.
#' If the \pkg{DiagrammeR} package is not installed, prints the DOT source
#' to the console instead.
#'
#' This method handles all DOT objects produced by the nmfkc package:
#' \code{\link{nmfkc.DOT}}, \code{\link{nmfae.DOT}}, \code{\link{nmf.sem.DOT}},
#' and \code{\link{nmfkc.ar.DOT}}.
#'
#' @param x An object of class \code{"nmfkc.DOT"} (or a subclass thereof).
#' @param ... Not used.
#' @return Called for its side effect (rendering). Returns \code{x} invisibly.
#' @seealso \code{\link{nmfkc.DOT}}, \code{\link{nmfae.DOT}},
#'   \code{\link{nmf.sem.DOT}}, \code{\link{nmfkc.ar.DOT}}
#' @export
plot.nmfkc.DOT <- function(x, ...) {
  if (requireNamespace("DiagrammeR", quietly = TRUE)) {
    print(DiagrammeR::grViz(as.character(x)))
  } else {
    message("DiagrammeR package not installed. Printing DOT source:")
    cat(as.character(x))
  }
  invisible(x)
}

# ============================================================
#  R/nmf.ffb.R — NMF-FFB aliases for the nmf.sem* family
#
#  "NMF-FFB" (Non-negative Matrix Factorization with Feed-Forward +
#  Feedback) is the canonical name adopted in Satoh (2025), arXiv:
#  2512.18250.  It is identical to "NMF-SEM" -- only the name differs.
#
#  This file provides thin wrappers
#      nmf.ffb()           = nmf.sem()
#      nmf.ffb.inference() = nmf.sem.inference()
#      nmf.ffb.cv()        = nmf.sem.cv()
#      nmf.ffb.DOT()       = nmf.sem.DOT()
#      nmf.ffb.split()     = nmf.sem.split()
#  so that the package's user-facing API matches the paper.  The
#  legacy nmf.sem* functions remain fully functional (no deprecation
#  is planned at this time).
#
#  Class hierarchy: result objects carry both new and legacy classes,
#      class(res) = c("nmf.ffb", "nmf.sem")
#  so existing S3 methods (summary.nmf.sem, plot.nmf.sem, ...) are
#  automatically reused via S3 inheritance.
# ============================================================


#' @rdname nmf.sem
#' @export
nmf.ffb <- function(Y1, Y2, rank = NULL, X.init = "nndsvd",
                    X.L2.ortho = 100.0, C1.L1 = 1.0, C2.L1 = 0.1,
                    epsilon = 1e-6, maxit = 5000, seed = 123, ...) {
  nmf.sem(Y1, Y2, rank = rank, X.init = X.init,
          X.L2.ortho = X.L2.ortho, C1.L1 = C1.L1, C2.L1 = C2.L1,
          epsilon = epsilon, maxit = maxit, seed = seed, ...)
}


#' @rdname nmf.sem.inference
#' @export
nmf.ffb.inference <- function(object, Y1, Y2,
                              B = 1000L, threshold = 0.01,
                              ci.level = 0.95,
                              C1.L1 = 1.0, C2.L1 = 0.1,
                              seed = 123L, ...) {
  nmf.sem.inference(object, Y1, Y2,
                    B = B, threshold = threshold,
                    ci.level = ci.level,
                    C1.L1 = C1.L1, C2.L1 = C2.L1,
                    seed = seed, ...)
}


#' @rdname nmf.sem.cv
#' @export
nmf.ffb.cv <- function(...) nmf.sem.cv(...)


#' @rdname nmf.sem.DOT
#' @export
nmf.ffb.DOT <- function(result, ...) nmf.sem.DOT(result, ...)


#' @rdname nmf.sem.split
#' @export
nmf.ffb.split <- function(x, ...) nmf.sem.split(x, ...)

---
title: "Topic Modeling with nmfkc"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Topic Modeling with nmfkc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteDepends{quanteda}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7, 
  fig.height = 5
)
```

## Introduction

Non-negative Matrix Factorization (NMF) is a powerful technique for topic modeling. By decomposing a document-term matrix, we can simultaneously discover latent **Topics** (clusters of words) and their **Trends** (proportions in documents).

This vignette demonstrates how to use the `nmfkc` package to analyze the U.S. presidential inaugural addresses using the `quanteda` package.

We will cover:

1.  **Data Preparation**: Converting text data into a matrix format suitable for `nmfkc`.
2.  **Rank Selection**: Determining the optimal number of topics using robust Cross-Validation.
3.  **Standard NMF**: Extracting basic topics and interpreting key words.
4.  **Kernel NMF**: Modeling the **temporal evolution** of topics using the "Year" as a covariate.

First, let's load the necessary packages.

```{r load-packages}
library(nmfkc)
library(quanteda)
```

## 1\. Data Preparation

We create a Document-Feature Matrix (DFM) where rows represent documents and columns represent words.

```{r data-prep}
# Load the corpus from quanteda
corp <- corpus(data_corpus_inaugural)

# Preprocessing: tokenize, remove stopwords, and punctuation
tok <- tokens(corp, remove_punct = TRUE)
tok <- tokens_remove(tok, pattern = stopwords("en", source = "snowball"))

# Create DFM and filter
df <- dfm(tok)
df <- dfm_select(df, min_nchar = 3) # Remove short words (<= 2 chars)
df <- dfm_trim(df, min_termfreq = 100) # Remove rare words (appearing < 100 times)

# --- CRITICAL STEP ---
# quanteda's DFM is (Documents x Words).
# nmfkc expects (Words x Documents).
# We must transpose the matrix.
d <- as.matrix(df)
# Sort by frequency
index <- order(colSums(d), decreasing=TRUE) 
d <- d[,index]
Y <- t(d)

dim(Y) # Features (Words) x Samples (Documents)
```

## 2\. Rank Selection (Number of Topics)

Before fitting the model, we need to decide the number of topics ($rank$). The `nmfkc.rank()` function helps us choose an appropriate rank.

The default `detail = "full"` performs **Element-wise Cross-Validation (Wold's CV)**. This method randomly holds out individual matrix elements and evaluates how well the model predicts them, providing a robust measure for rank selection (though it takes more computation time).

```{r rank-selection, fig.width=7, fig.height=6}
# Evaluate ranks from 2 to 6
nmfkc.rank(Y, rank = 2:6)
```

Looking at the diagnostics (e.g., the minimum ECV Sigma, the elbow of the R-squared curve, or high Cophenetic Correlation), let's assume **Rank = 3** is a reasonable choice for this overview.

## 3\. Standard NMF

We fit the standard NMF model ($Y \approx XB$) with `rank = 3`.
In the context of topic modeling:

  * **$X$ (Basis Matrix):** Represents **Topics** (distribution of words).
  * **$B$ (Coefficient Matrix):** Represents **Trends** (distribution of topics across documents).

<!-- end list -->

```{r model-fit}
rank <- 3
# Set seed for reproducibility
res_std <- nmfkc(Y, rank = rank, seed = 123, prefix = "Topic")

# Check Goodness of Fit (R-squared)
res_std$r.squared
```

### Interpreting Topics (Keywords)

We can identify the meaning of each topic by looking at the words with the highest weights in the basis matrix `X`.

```{r interpret-topics}
# Extract top 10 words for each topic from X.prob (normalized X)
Xp <- res_std$X.prob
for(q in 1:rank){
  message(paste0("----- Featured words on Topic [", q, "] -----"))
  print(paste0(rownames(Xp), "(", rowSums(Y), ") ", round(100*Xp[,q], 1), "%")[Xp[,q]>=0.5])
}
```

*(Note: Interpretation depends on the result. For example, one topic might contain words like "government, people, states" (Political), while another might have "world, peace, freedom" (International).)*

## 4\. Kernel NMF: Temporal Topic Evolution

One of the unique features of `nmfkc` is **Kernel NMF**.
In standard NMF, the order of documents is ignored; each speech is treated independently. However, inaugural addresses have a strong **temporal component**. By using the "Year" as a covariate, we can smooth the topic proportions over time to see historical shifts.

### Optimizing the Kernel Parameter

We construct a covariate matrix `U` using the year of the address. We then find the optimal kernel bandwidth (`beta`) using Cross-Validation.

```{r kernel-prep}
# Covariate: Year of the address
years <- as.numeric(substring(names(data_corpus_inaugural), 1, 4))
U <- t(as.matrix(years))

# Optimize beta (Gaussian Kernel width)
# We test a specific range of betas to locate the minimum CV error.
beta_candidates <- c(0.2, 0.5, 1, 2, 5) / 10000

# Run CV to find the best beta
# Note: We use the same rank (rank=3) as selected above.
cv_res <- nmfkc.kernel.beta.cv(Y, rank = rank, U = U, beta = beta_candidates, plot = FALSE)
best_beta <- cv_res$beta
print(best_beta)
```

### Fitting Kernel NMF

Now we fit the model using the kernel matrix `A`. This enforces that documents close in time (similar years) should have similar topic distributions.

```{r kernel-fit}
# Create Kernel Matrix
A <- nmfkc.kernel(U, beta = best_beta)

# Fit NMF with Kernel Covariates
res_ker <- nmfkc(Y, A = A, rank = rank, seed = 123, prefix = "Topic")
```

### Visualization: Standard vs Kernel NMF

Let's compare how topic proportions change over time.

  * **Standard NMF (Top):** Shows noisy fluctuations. It captures the specific content of each speech but misses the larger historical context.
  * **Kernel NMF (Bottom):** Reveals smooth historical trends, showing how themes like "Nation Building" vs "Global Affairs" have evolved over centuries.

<!-- end list -->

```{r plot-comparison, fig.width=8, fig.height=8}
oldpar <- par(mfrow = c(2, 1), mar = c(4, 4, 2, 1))

# Prepare Axis Labels (Rounded to integers)
at_points <- seq(1, ncol(Y), length.out = 10)
labels_years <- round(seq(min(years), max(years), length.out = 10))

# 1. Standard NMF (Noisy)
barplot(res_std$B.prob, col = 2:(rank+1), border = NA, xaxt='n',
        main = "Standard NMF: Topic Proportions (Noisy)", ylab = "Probability")
axis(1, at = at_points, labels = labels_years)

# 2. Kernel NMF (Smooth trend)
barplot(res_ker$B.prob, col = 2:(rank+1), border = NA, xaxt='n',
        main = "Kernel NMF: Temporal Topic Evolution (Smooth)", ylab = "Probability")
axis(1, at = at_points, labels = labels_years)

# Legend
legend("topright", legend = paste("Topic", 1:rank), fill = 2:(rank+1), bg="white", cex=0.8)
par(oldpar)
```

---
title: "NMF-SEM with nmfkc"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{NMF-SEM with nmfkc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteDepends{lavaan}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4
)
```

## Overview

This vignette demonstrates how to fit the **NMF-SEM**
(Nonnegative Matrix Factorization–Structural Equation Model)
implemented in the **nmfkc** package.

NMF-SEM extends standard nonnegative matrix factorization by introducing
a **structural equation model in latent space**.

Observed endogenous variables $Y_1$ are approximated as

$$
Y_1 \approx X (\Theta_1 Y_1 + \Theta_2 Y_2).
$$

If the feedback operator $XC_1$ is stable, the model implies the equilibrium mapping

$$
\hat Y_1 \approx (I - X\Theta_1)^{-1} X\Theta_2 Y_2
\equiv M_{\mathrm{model}} Y_2.
$$

---

## 1. Load and preprocess data

```{r}
library(lavaan)
data(HolzingerSwineford1939)

d <- HolzingerSwineford1939
index <- complete.cases(d)
d <- d[index, ]
```

### 1.1 Construct exogenous variables

```{r}
d$age.rev <- -(d$ageyr + d$agemo / 12)
d$sex.2 <- ifelse(d$sex == 2, 1, 0)
d$school.GW <- ifelse(d$school == "Grant-White", 1, 0)
d[, c("id", "sex", "ageyr", "agemo", "school", "grade")] <- NULL
```

### 1.2 Nonnegative normalization

```{r}
library(nmfkc)
d <- nmfkc::nmfkc.normalize(d)
```

## 2. Split into endogenous and exogenous blocks

```{r}
exogenous_vars <- c("age.rev", "sex.2", "school.GW")
endogenous_vars <- setdiff(colnames(d), exogenous_vars)

Y1 <- t(d[, endogenous_vars])
Y2 <- t(d[, exogenous_vars])
```

## 3. Baseline NMF model

```{r}
myepsilon <- 1e-6
Q0 <- 3

res0 <- nmfkc(
  Y = Y1,
  A = Y2,
  rank = Q0,
  epsilon = myepsilon,
  X.L2.ortho = 100
)

M.simple <- res0$X %*% res0$C
```

## 3.5 Hyperparameter Tuning for Sparsity (Optional)

```{r,eval=FALSE}
grid_params <- expand.grid(
    C1.L1     = c(0,1:9/10,1:10),
    C2.L1     = c(0,1:9/10,1:10)
)
n_iter <- nrow(grid_params)
mae.cv <- 0*1:n_iter

for(i in 1:n_iter){
  if (i %% round(n_iter / 10) == 0) {
    message(sprintf("Processing... %d%% (%d/%d)", round(i/n_iter*100), i, n_iter)) 
  }  
  p <- grid_params[i, ]
  res.cv <- nmf.sem.cv(Y1, Y2, rank = Q0,
                       X.init = res0$X,
                       X.L2.ortho = 100,
                       C1.L1 = p$C1.L1,
                       C2.L1 = p$C2.L1,
                       seed = 1, epsilon = myepsilon)
  mae.cv[i] <- res.cv
}

f <- data.frame(grid_params,mae.cv)
f <- f[order(f$mae.cv),]
head(f,5)
#    C2.L2.off C1.L1 C2.L1    mae.cv
#140         0    10   0.6 0.1820841
#160         0    10   0.7 0.1820843
#180         0    10   0.8 0.1820877
#200         0    10   0.9 0.1820907
#120         0    10   0.5 0.1820908
print(p <- f[1,])
```


## 4. NMF-SEM estimation

```{r}
p <- list(C1.L1 = 10, C2.L1 = 0.6)

res <- nmf.sem(
  Y1, Y2,
  rank = Q0,
  X.init = res0$X,
  X.L2.ortho = 100,
  C1.L1 = p$C1.L1,
  C2.L1 = p$C2.L1,
  epsilon = myepsilon
)
```

```{r}
plot(res$objfunc.full, type = "l",
     main = "Objective Function",
     ylab = "Loss", xlab = "Iteration")
```

## 5. Diagnostics

```{r}
SC.map <- cor(as.vector(res$M.model),
              as.vector(M.simple))

cat("Q=     ", Q0, "\n")
cat("RHO=   ", round(res$XC1.radius, 3), "\n")
cat("AR=    ", round(res$amplification, 3), "\n")
cat("SCmap= ", round(SC.map, 3), "\n")
cat("SCcov= ", round(res$SC.cov, 3), "\n")
cat("MAE=   ", round(res$MAE, 3), "\n")
```

## 6. Visualization

```{r}
res.dot <- nmf.sem.DOT(
  res,
  weight_scale = 5,
  rankdir = "TB",
  threshold = 0.01,
  fill = FALSE,
  cluster.box = "none"
)

# plot(res.dot)  # requires DiagrammeR
```

---
title: "Classification with NMF-LAB"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Classification with NMF-LAB}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteDepends{palmerpenguins}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

## Introduction

This vignette demonstrates how to use the `nmfkc` package for **Supervised Classification**, a technique referred to as **NMF-LAB** (Label-based NMF).

### How it works

The NMF-LAB approach treats multi-class classification as a matrix factorization task:
$$Y \approx X C A$$

  * **$Y$ (Target):** One-hot encoded matrix of class labels (Classes $\times$ Samples).
  * **$A$ (Input):** Feature matrix or Kernel matrix constructed from features.
  * **$X$ (Learned Basis):** A mapping matrix between latent factors and class labels.
  * **$C$ (Coefficient/Parameter):** Captures the relationship between input $A$ and classes.

**Key Trade-off:**

1.  **Linear Model ($A = Features$):** Highly interpretable. $C$ tells you exactly which feature contributes to which class.
2.  **Kernel Model ($A = Kernel$):** Highly accurate. Handles non-linear boundaries but is harder to interpret.

We will explore this trade-off using the `iris` dataset.

First, let's load the required packages.

```{r load-packages}
library(nmfkc)
library(palmerpenguins)
```

-----

## Example 1: The Iris Dataset (Linear vs. Kernel)

Our goal is to classify 3 species of iris flowers (*setosa, versicolor, virginica*) based on 4 measurements.

### 1\. Data Preparation

We convert the species labels into a binary class matrix $Y$ and normalize the features $U$.

```{r iris-data-prep}
# 1. Prepare Labels (Y)
label_iris <- iris$Species
Y_iris <- nmfkc.class(label_iris) # One-hot encoding (3 Classes x 150 Samples)
rank_iris <- length(unique(label_iris)) # Number of classes

# 2. Prepare Features (U)
# Normalize features to [0, 1] range and transpose to (Features x Samples)
U_iris <- t(nmfkc.normalize(iris[, -5]))
```

### Step 1: Linear NMF-LAB (Focus on Interpretability)

First, we fit a **Linear Model** by using the features $U$ directly as the input matrix $A$.
This allows us to see **"Which feature drives which class?"** by inspecting the matrix $C$.

```{r iris-linear}
# Use Normalized Features directly as A
A_linear <- U_iris

# Fit Linear NMF-LAB
res_linear <- nmfkc(Y = Y_iris, A = A_linear, rank = rank_iris, seed = 123, prefix = "Class")

# --- Interpretability Check ---
# The matrix C (Q x R) shows the weight of each Feature (columns) for each Class (rows).
# Let's look at the estimated weights:
round(res_linear$C, 2)
```

**Interpretation:**
Looking at the matrix $C$ above:

  * **Class1 (Setosa):** Has high weights on `Sepal.Width` (2nd col).
  * **Class3 (Virginica):** Has high weights on `Petal.Length` (3rd col) and `Petal.Width` (4th col).
    This "White-box" transparency is the main advantage of the linear model.

**Accuracy Check:**
We evaluate the model using a confusion matrix.
*(Note: Linear models often struggle with overlapping classes like versicolor and virginica.)*

```{r iris-linear-acc}
pred_linear <- predict(res_linear, type = "class")
(f_linear <- table(fitted.label = pred_linear, label = label_iris))

# Calculate Accuracy (assuming diagonal correspondence)
acc_linear <- sum(diag(f_linear)) / sum(f_linear)
cat(paste0("Linear Model Accuracy: ", round(acc_linear * 100, 2), "%\n"))
```

### Step 2: Kernel NMF-LAB (Focus on Performance)

To improve accuracy, we switch to the **Kernel Model**. We map the features into a high-dimensional space using a Gaussian kernel.

```{r iris-kernel}
# 1. Optimize Kernel Width (beta)
# Heuristic estimation of beta
res_beta <- nmfkc.kernel.beta.nearest.med(U_iris)

# Cross-validation for fine-tuning (using generated candidates)
cv_res <- nmfkc.kernel.beta.cv(Y_iris, rank = rank_iris, U = U_iris, 
                               beta = res_beta$beta_candidates, plot = FALSE)
best_beta <- cv_res$beta

# 2. Fit Kernel NMF-LAB
A_kernel <- nmfkc.kernel(U_iris, beta = best_beta)
res_kernel <- nmfkc(Y = Y_iris, A = A_kernel, rank = rank_iris, seed = 123, prefix = "Class")

# 3. Prediction and Evaluation
fitted_label <- predict(res_kernel, type = "class")
(f_kernel <- table(fitted.label = fitted_label, label = label_iris))

# Calculate Accuracy
acc_kernel <- sum(diag(f_kernel)) / sum(f_kernel)
cat(paste0("Kernel Model Accuracy: ", round(acc_kernel * 100, 2), "%\n"))
```

**Result:** The accuracy jumps significantly (often \>96%). The kernel successfully separates the complex boundaries.

### Visualization: Class Prototypes (Basis X)

Finally, let's visualize the Basis Matrix $X$ of the successful kernel model. Ideally, it should look like a diagonal matrix, mapping each Latent Factor to a specific Species.

```{r iris-plot-X, fig.height=5}
image(t(res_kernel$X)[, nrow(res_kernel$X):1], 
      main = "Basis Matrix X (Kernel Model)\nMapping Factors to Species",
      axes = FALSE, col = hcl.colors(12, "YlOrRd", rev = TRUE))

axis(1, at = seq(0, 1, length.out = rank_iris), labels = colnames(res_kernel$X))
axis(2, at = seq(0, 1, length.out = rank_iris), labels = rev(rownames(res_kernel$X)), las = 2)
box()
```

-----

## Example 2: The Palmer Penguins Dataset

Let's apply the **Kernel NMF-LAB** workflow to classify penguin species (*Adelie, Chinstrap, Gentoo*), focusing on the **Probabilistic (Soft)** nature of NMF classification.

### 1\. Data Preparation

We must remove rows with missing values (`NA`) as the kernel matrix $A$ cannot handle missing entries in the input features.

```{r penguins-data-prep}
# Load and clean data (remove rows with NAs)
d_penguins <- na.omit(palmerpenguins::penguins)

# Prepare Y (Labels)
label_penguins <- d_penguins$species
Y_penguins <- nmfkc.class(label_penguins)

# Prepare U (Features)
U_penguins <- t(nmfkc.normalize(d_penguins[, 3:6]))
```

### 2\. Model Fitting

We use the heuristic $\beta$ directly for a quick demonstration.

```{r penguins-model-fit}
rank_penguins <- length(unique(label_penguins))

# 1. Heuristic beta estimation
best_beta_penguins <- nmfkc.kernel.beta.nearest.med(U_penguins)$beta

# 2. Optimization
A_penguins <- nmfkc.kernel(U_penguins, beta = best_beta_penguins)
res_penguins <- nmfkc(Y = Y_penguins, A = A_penguins, rank = rank_penguins, 
                      seed = 123, prefix = "Class")
```

### 3\. Visualizing "Soft" Classification

Unlike many classifiers that only output a final label, NMF provides a **probability distribution** over classes.
The plot below shows the predicted probability for each penguin.

  * **Solid Blocks of Color:** Indicate high confidence predictions.
  * **Mixed Colors:** Indicate samples where the model is uncertain (often on the boundary between species).

<!-- end list -->

```{r penguins-soft-vis, fig.width=8, fig.height=4}
# Get probabilistic predictions
probs <- predict(res_penguins, type = "prob")

# Visualize
barplot(probs, col = c("#FF8C00", "#9932CC", "#008B8B"), border = NA,
        main = "Soft Classification Probabilities (Penguins)",
        xlab = "Sample Index", ylab = "Probability")

legend("topright", legend = levels(label_penguins), 
       fill = c("#FF8C00", "#9932CC", "#008B8B"), bg = "white", cex = 0.8)
```

### 4\. Evaluation

Finally, we calculate the accuracy.

```{r penguins-acc}
fitted_label_p <- predict(res_penguins, type = "class")
(f_penguins <- table(Predicted = fitted_label_p, Actual = label_penguins))

acc_p <- sum(diag(f_penguins)) / sum(f_penguins)
cat(paste0("Penguins Accuracy: ", round(acc_p * 100, 2), "%\n"))
```

---
title: "Introduction to nmfkc"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to nmfkc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

## Introduction

Welcome to the `nmfkc` package\!
This vignette provides a beginner-friendly introduction to the core function, `nmfkc()`.

**Non-negative Matrix Factorization (NMF)** is a technique that decomposes a large data matrix $Y$ into two smaller matrices, $X$ and $B$:
$$Y \approx X B$$
The key feature of NMF is that all elements must be **non-negative** ($\ge 0$). This makes the results intuitive, as the original data can be understood as an additive combination of parts.

In this guide, we will cover:

1.  **Basic NMF**: Extracting latent topics using a Movie Ratings example.
2.  **Interpretation**: Understanding what the decomposed matrices represent.
3.  **Missing Values**: How to handle and predict missing data (e.g., for recommendations).

-----

## 1\. Basic Usage: Analyzing Movie Ratings

To understand NMF, let's imagine a scenario with **5 Users** rating **4 Movies** on a scale of 1 to 5.

First, load the package.

```{r load-package}
library(nmfkc)
```

### Creating the Data

We create a rating matrix `Y`.
The dataset contains two hidden genres: "Action" (Movies 1 & 2) and "Romance" (Movies 3 & 4).

```{r create-data}
# Rows: Users (U1-U5), Cols: Movies (M1-M4)
# U1, U2, U3 prefer Action movies.
# U4, U5 prefer Romance movies.
Y <- matrix(
  c(5, 4, 1, 1,
    4, 5, 1, 2,
    5, 5, 2, 2,
    1, 2, 5, 4,
    1, 1, 4, 5),
  nrow = 5, byrow = TRUE
)

# Assign names for better interpretation
rownames(Y) <- paste0("User", 1:5)
colnames(Y) <- c("Action1", "Action2", "Romance1", "Romance2")

# Check the data
print(Y)
```

### Running NMF

We use the `nmfkc()` function to decompose this matrix.
Since we assume there are 2 genres (Action and Romance), we set **rank = 2**.

```{r run-nmfkc}
# Run NMF with rank = 2
res <- nmfkc(Y, rank = 2, seed = 123)
```

### Interpretation

NMF decomposes $Y$ into $X$ (Basis) and $B$ (Coefficient).
*(Note: The order of bases may vary depending on the random seed. In this example with seed=123, Basis 1 corresponds to Action and Basis 2 to Romance.)*

#### 1\. Basis Matrix X: User Preferences

The matrix $X$ represents **"How much each User likes each Genre (Basis)."**

```{r interpret-X}
# Each column represents a latent factor (Basis)
res$X
```

  * **Basis1**: High values for **User1, User2, and User3** (Action fans).
  * **Basis2**: High values for **User4 and User5** (Romance fans).

#### 2\. Coefficient Matrix B: Movie Genres

The matrix $B$ represents **"Which Genre each Movie belongs to."**

```{r interpret-B}
# Each row represents a latent factor
res$B
```

  * **Basis1**: High weights on **Action1 and Action2**.
  * **Basis2**: High weights on **Romance1 and Romance2**.

As you can see, NMF automatically discovered the hidden structures ("Action" vs "Romance") and user preferences without being explicitly told.

-----

## 2\. Visualization

`nmfkc` provides tools to visually diagnose your model.

### Convergence Plot

Use the `plot()` function to check if the error minimized properly during iterations.

```{r plot-convergence}
plot(res, main = "Convergence Plot")
```

### Visualizing the Reconstruction

The `nmfkc.residual.plot()` function allows you to compare the **Original Matrix ($Y$)**, the **Fitted Matrix ($XB$)**, and the **Residuals ($E$)** side-by-side.

```{r plot-residual, fig.width=9, fig.height=4}
# Visualize Original vs Fitted vs Residuals
nmfkc.residual.plot(Y, res)
```

The middle plot (Fitted Matrix) successfully captures the block structure of the original data.

-----

## 3\. Handling Missing Values (Imputation)

A powerful feature of `nmfkc` is its robustness to **Missing Values (`NA`)**. This is useful for tasks like **Recommendation Systems**, where you want to predict how a user would rate a movie they haven't seen yet.

### Creating Data with Missing Values

Let's assume **User1** has not seen **Action1** yet. We set this value to `NA`.

```{r create-na}
Y_missing <- Y
Y_missing["User1", "Action1"] <- NA # Introduce missing value
print(Y_missing)
```

### Running NMF with NAs

Simply pass the matrix with `NA`s to `nmfkc()`. The algorithm automatically handles them by ignoring the missing entries during optimization.

```{r run-na}
res_na <- nmfkc(Y_missing, rank = 2, seed = 123)
```

### Predicting the Unknown Rating

The fitted model ($X \times B$) provides an estimate for the missing entry.

```{r impute-na}
# Extract the predicted value from the fitted matrix XB
predicted_rating <- res_na$XB["User1", "Action1"]
actual_rating <- Y["User1", "Action1"] # The original hidden value (5)

cat(paste0("Actual Rating:    ", actual_rating, "\n"))
cat(paste0("Predicted Rating: ", round(predicted_rating, 2), "\n"))
```

Because User1 liked other Action movies, the model predicted a **reasonably high rating (3.62)** for the missing Action movie, closer to the actual rating (5) than to a low rating.

## Summary

With the `nmfkc` package, you can easily:

1.  **Decompose** complex data into interpretable parts ($X$ and $B$).
2.  **Handle missing values** robustly for imputation and prediction.
3.  **Visualize** the results to verify the fit.

For more advanced topics, such as Time Series Analysis or Covariate-assisted NMF, please refer to the other vignettes (`Topic Modeling` and `Time Series Analysis`).

---
title: "Time Series Analysis with NMF-VAR"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Time Series Analysis with NMF-VAR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteDepends{vars}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

## Introduction

This vignette demonstrates how to apply the `nmfkc` package to time series data using the **NMF-VAR (Non-negative Matrix Factorization - Vector Autoregression)** framework. This approach models the coefficient matrix of NMF as a VAR process ($B \approx C A_{lag}$), allowing for both dimensionality reduction and temporal modeling.

Key functions in `nmfkc` such as `nmfkc.ar` and `nmfkc.ar.degree.cv` directly support R's native **`ts` objects**, preserving time series properties automatically.

We will cover two scenarios:

1.  **Univariate AR:** Forecasting airline passengers (`AirPassengers`).
2.  **Multivariate VAR:** Analyzing macroeconomic indicators (`Canada`).

First, let's load the necessary packages.

```{r load-packages}
library(nmfkc)
library(vars) # For Canada dataset
```

-----

## Example 1: Univariate Autoregression with AirPassengers

The `AirPassengers` dataset contains monthly international airline passenger numbers. We will model this univariate time series using an autoregressive (AR) model.

### 1\. Data Preparation

We simply log-transform the `ts` object to stabilize variance. We do **not** need to manually convert it to a matrix or create time vectors; `nmfkc` functions handle `ts` objects directly.

```{r air-data-prep}
# Load and transform the ts object
d_air <- AirPassengers
d_air_log <- log10(d_air) # Still a ts object
```

### 2\. Model Selection (Lag Order)

We use **Cross-Validation** to find the optimal degree (lag order). By passing the `ts` object `d_air_log`, the function automatically handles the time dimension.

```{r air-degree-cv}
# Evaluate lag orders from 1 to 14
# Note: ts objects are automatically transposed to (Variables x Time) internally
cv_res <- nmfkc.ar.degree.cv(d_air_log, rank = 1, degree = 1:14, epsilon=1e-6, maxit=500000)

# Check the optimal degree
cv_res$degree

# For this example, we will proceed with D=12 (capturing monthly seasonality)
D <- 12
```

### 3\. Model Fitting

We construct the observation matrix `Y` and covariate matrix `A` (lagged Y) using `nmfkc.ar()`. The returned matrices `Y` and `A` will have time-based column names derived from the `ts` object.

```{r air-model-fit}
# Create matrices for the AR(12) model
a_air <- nmfkc.ar(d_air_log, degree = D, intercept = TRUE)

# Fit the NMF-AR model (Rank=1 for univariate)
res_air <- nmfkc(Y = a_air$Y, A = a_air$A, rank = 1, epsilon = 1e-6, maxit=500000)

# Check goodness of fit
res_air$r.squared

# Check for stationarity (spectral radius < 1)
nmfkc.ar.stationarity(res_air)
```

### 4\. Forecasting

We can forecast future values using the fitted model. `nmfkc.ar.predict` uses the time properties stored in the model to generate the correct future time sequence.

```{r air-forecast}
# Forecast next 2 years (24 months)
h <- 24
pred_res <- nmfkc.ar.predict(x = res_air, Y = a_air$Y, n.ahead = h)

# Convert predictions back to original scale
pred_val <- 10^as.vector(pred_res$pred)
pred_time <- pred_res$time # Future time points generated by the function

# --- Plotting ---
# Setup plot range
xlim_range <- range(c(time(d_air), pred_time))
ylim_range <- range(c(d_air, pred_val))

# 1. Observed data (Black)
plot(d_air, type = "l", col = "black", 
     xlim = xlim_range, ylim = ylim_range, lwd = 1,
     xlab = "Year", ylab = "Air Passengers", main = "NMF-VAR Forecast (h=24)")

# 2. Fitted values during training (Red)
# a_air$Y has column names as time strings; we parse them for plotting
fitted_time <- as.numeric(colnames(res_air$XB))
lines(fitted_time, 10^as.vector(res_air$XB), col = "red", lwd = 2)

# 3. Forecast (Blue)
# Connect the last observed point to the first forecast for a continuous line
last_t <- tail(as.numeric(time(d_air)), 1)
last_y <- tail(as.vector(d_air), 1)
lines(c(last_t, pred_time), c(last_y, pred_val), col = "blue", lwd = 2, lty = 2)

# Add legend
legend("topleft", legend = c("Observed", "Fitted", "Forecast"),
       col = c("black", "red", "blue"), lty = c(1, 1, 2), lwd = 2)
```

-----

## Example 2: Vector Autoregression with Canada Dataset

The `Canada` dataset contains four quarterly macroeconomic variables. We'll use a multivariate NMF-VAR model to analyze the relationship between these variables.

### 1\. Data Preparation

We take the first difference to achieve stationarity, normalize the data to $[0, 1]$, and transpose it to the `Variables x Time` format required by NMF.

```{r canada-data-prep}
# Load, difference, and normalize
d0_canada <- Canada
dd_canada <- apply(d0_canada, 2, diff) # Returns a matrix (Time x Vars)
dn_canada <- nmfkc.normalize(dd_canada)

# Transpose to (Variables x Time) for NMF
Y0_canada <- t(dn_canada)

# Create matrices for VAR(1)
a_canada <- nmfkc.ar(Y0_canada, degree = 1, intercept = TRUE)
```

### 2\. Model Fitting

We fit a model with `rank = 2` to identify two latent economic conditions driving the four variables.

```{r canada-model-fit}
# Fit the NMF-VAR model
res_canada <- nmfkc(Y = a_canada$Y, A = a_canada$A, rank = 2, prefix = "Condition", epsilon = 1e-6)

# R-squared and Stationarity
res_canada$r.squared
nmfkc.ar.stationarity(res_canada)
```

### 3\. Latent Structure & Causal Graph

We can visualize how the two latent conditions change over time.

```{r canada-soft-cluster}
# Visualize soft clustering of time trends
barplot(res_canada$B.prob, col = c(2, 3), border = NA,
        main = "Soft Clustering of Economic Conditions",
        xlab = "Time", ylab = "Probability",
        names.arg = colnames(a_canada$Y), las=2, cex.names = 0.5)
legend("topright", legend = colnames(res_canada$X), fill = c(2, 3), bg = "white")
```

Finally, we can generate a **DOT script** to visualize the Granger causality (relationships) between variables inferred by the model.

```{r canada-dot}
# Generate DOT script for graph visualization
dot_script <- nmfkc.ar.DOT(res_canada, intercept = TRUE, threshold=0.01)

# plot(dot_script)  # requires DiagrammeR
```

---
title: "NMF-RE: Mixed-Effects Modeling with nmfkc"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{NMF-RE: Mixed-Effects Modeling with nmfkc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteDepends{nlme}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

## Introduction

This vignette demonstrates **NMF-RE** (Non-negative Matrix Factorization with Random Effects), a mixed-effects extension of NMF implemented in the `nmfkc` package.

### Why Mixed Effects?

Standard NMF with covariates models the data as:
$$Y \approx X \Theta A$$

where $\Theta A$ captures systematic (fixed) effects of covariates $A$ on latent scores. However, in many applications --- such as longitudinal studies, panel data, or clustered observations --- individuals exhibit **unit-specific deviations** that cannot be explained by covariates alone.

NMF-RE addresses this by adding **random effects** $U$:
$$Y = X(\Theta A + U) + \mathcal{E}$$

- $Y$ ($P \times N$): Non-negative observation matrix (e.g., repeated measurements over $P$ time points for $N$ subjects).
- $X$ ($P \times Q$): Non-negative basis matrix learned from data.
- $\Theta$ ($Q \times K$): Coefficient matrix for covariate effects.
- $A$ ($K \times N$): Covariate matrix (e.g., intercept, sex, treatment).
- $U$ ($Q \times N$): Random effects matrix capturing individual deviations.
- $\mathcal{E}$: Residual noise with variance $\sigma^2$.

The random effects $U$ follow a ridge-type prior controlled by variance $\tau^2$, and the penalty $\lambda = \sigma^2 / \tau^2$ determines the degree of shrinkage.

### Key Features

- **Automatic regularization**: `df.rate` controls the effective degrees of freedom consumed by $U$, preventing overfitting.
- **Wild bootstrap inference**: Standard errors, z-values, p-values, and confidence intervals for $\Theta$ without repeated constrained re-optimization.
- **Trace-based ICC**: Quantifies the proportion of variance attributable to random effects.

---

## 1. Data: Orthodont (Longitudinal Growth)

We use the `Orthodont` dataset from the `nlme` package: orthodontic distance measurements for 27 children at ages 8, 10, 12, and 14. The covariate of interest is sex (Male/Female).

```{r load-data}
library(nmfkc)
library(nlme)

data(Orthodont)
head(Orthodont)
```

### 1.1 Prepare the Observation Matrix Y

Each column of $Y$ is a subject, each row is a time point (age). Since there are 4 ages and 27 subjects, $Y$ is $4 \times 27$.

```{r prepare-Y}
Y <- matrix(Orthodont$distance, nrow = 4, ncol = 27)
colnames(Y) <- paste0("S", 1:27)
rownames(Y) <- paste("Age", c(8, 10, 12, 14))
Y[, 1:6]
```

### 1.2 Prepare the Covariate Matrix A

We create a $2 \times 27$ covariate matrix with an intercept row and a binary `male` indicator.

```{r prepare-A}
male <- ifelse(Orthodont$Sex[seq(1, 108, 4)] == "Male", 1, 0)
A <- rbind(intercept = 1, male = male)
A[, 1:6]
```

---

## 2. Selecting dfU Cap Rate

Before fitting the model, we use `nmfre.dfU.scan()` to scan different regularization levels. The **dfU cap rate** limits how many effective degrees of freedom the random effects $U$ can consume, preventing overfitting.

The scan evaluates rates from 0.1 to 1.0 (where 1.0 means no cap). Key columns in the output:

- **dfU**: Effective degrees of freedom actually used.
- **safeguard**: Whether `dfU` stayed below the cap (`TRUE` = safe).
- **hit**: Whether $\tau^2$ was updated freely (`TRUE` = unconstrained).
- **ICC**: Trace-based Intraclass Correlation Coefficient.

```{r dfU-scan}
scan_result <- nmfre.dfU.scan(1:10/10, Y, A, rank = 1)
scan_result
```

### Interpreting the Scan

A good `df.rate` satisfies two criteria:

1. **safeguard = TRUE**: The cap is not overly binding.
2. **hit = TRUE**: The model converges to a natural solution without the cap forcing an artificial penalty.

Choose the smallest rate where both conditions hold. This balances model flexibility against overfitting.

---

## 3. Fitting the NMF-RE Model

Based on the scan, we select `df.rate = 0.2` and fit the model with `rank = 1` (a single latent trend). The `prefix` argument labels the basis.

```{r fit-nmfre}
res <- nmfre(Y, A, rank = 1, df.rate = 0.2, prefix = "Trend")
```

### 3.1 Model Summary

`summary()` displays variance components, fit statistics, and the coefficient table with significance stars.

```{r summary}
summary(res)
```

### Understanding the Output

**Variance components:**

- $\sigma^2$: Residual variance (measurement noise).
- $\tau^2$: Random effect variance (individual heterogeneity).
- **ICC**: Proportion of variance from random effects. Higher ICC means stronger individual differences relative to noise.

**Coefficient table ($\Theta$):**

- **Estimate**: Effect of each covariate on the latent trend.
- **Std. Error**: From wild bootstrap.
- **Pr(>|z|)**: One-sided p-value (H0: $\Theta_{qk} = 0$ vs H1: $\Theta_{qk} > 0$), appropriate because $\Theta$ is non-negative.

**R-squared:**

- $R^2$ (XB+blup): Fit including both fixed and random effects.
- $R^2$ (XB): Fit from fixed effects only --- the gap shows how much random effects improve the fit.

---

## 4. Visualization

We plot the growth curves to see how the model captures both population-level trends (fixed effects) and individual deviations (random effects).

```{r plot-growth, fig.height=6}
age <- c(8, 10, 12, 14)

plot(age, res$XB[, 1], type = "n", ylim = range(Y),
     xlab = "Age (years)", ylab = "Distance (mm)",
     main = "Orthodont: NMF-RE Growth Curves")

# Plot observed data points
for (j in 1:27) {
  pch_j <- ifelse(male[j] == 1, 4, 1)
  points(age, Y[, j], pch = pch_j, col = "gray60")
}

# Plot individual predictions (fixed + random effects)
for (j in 1:27) {
  lines(age, res$XB.blup[, j], col = "steelblue", lty = 3, lwd = 0.8)
}

# Plot population-level fixed effects (two lines: male and female)
for (j in 1:27) {
  lines(age, res$XB[, j], col = "red", lwd = 2)
}

legend("topleft",
  legend = c("Fixed effect (male/female)", "Fixed + Random (BLUP)",
             "Male (observed)", "Female (observed)"),
  lwd = c(2, 1, NA, NA), lty = c(1, 3, NA, NA),
  pch = c(NA, NA, 4, 1),
  col = c("red", "steelblue", "gray60", "gray60"),
  cex = 0.85)
```

### Interpreting the Plot

- **Red solid lines**: Population-level fitted values from fixed effects ($X \Theta A$). There are two distinct curves --- one for males (higher) and one for females (lower) --- reflecting the sex effect in $\Theta$.
- **Blue dashed lines**: Individual-level predictions ($X(\Theta A + U)$). Each line is shifted from the population curve by the subject's random effect $U$, capturing personal growth trajectories.
- **Gray points**: Observed measurements. Crosses ($\times$) for males, circles ($\circ$) for females.

---

## 5. Examining Learned Components

### 5.1 Basis Matrix X

The basis $X$ represents the latent temporal pattern shared across all subjects.

```{r basis-X}
cat("Basis X (temporal pattern):\n")
print(round(res$X, 4))
```

Since `rank = 1`, there is one basis vector. Its shape shows how the latent factor manifests across the four ages. The column is normalized to sum to 1.

### 5.2 Coefficient Matrix $\Theta$

$\Theta$ maps covariates to latent scores:

```{r coef-C}
cat("Coefficient matrix (Theta):\n")
print(round(res$C, 4))
```

- `intercept`: Baseline level of the latent trend for females.
- `male`: Additional contribution for males. A significant positive value indicates that males have higher orthodontic distances on average.

### 5.3 Random Effects U

$U$ captures individual deviations. Subjects with positive $U$ values grow faster than the population average; negative values indicate slower growth.

```{r random-effects, fig.height=4}
barplot(res$U[1, ], names.arg = colnames(Y),
        las = 2, cex.names = 0.7,
        col = ifelse(male == 1, "steelblue", "salmon"),
        main = "Random Effects (U) by Subject",
        ylab = "Random effect value")
legend("topright", fill = c("steelblue", "salmon"),
       legend = c("Male", "Female"), cex = 0.85)
```

---

## 6. Model Diagnostics

### 6.1 Convergence

Check that the algorithm converged properly:

```{r convergence}
cat("Converged:", res$converged, "\n")
cat("Iterations:", res$iter, "\n")
cat("Stop reason:", res$stop.reason, "\n")
```

```{r plot-convergence, fig.height=4}
plot(res$objfunc.iter, type = "l",
     xlab = "Iteration", ylab = "Objective function",
     main = "Convergence Trace")
```

### 6.2 Residual Analysis

Compare the fitted values against the original data:

```{r residual-analysis, fig.height=4}
residuals <- Y - res$XB.blup
cat("Mean absolute residual (BLUP):", round(mean(abs(residuals)), 4), "\n")
cat("Mean absolute residual (fixed):", round(mean(abs(Y - res$XB)), 4), "\n")

# Fitted vs Observed
plot(as.vector(Y), as.vector(res$XB.blup),
     xlab = "Observed", ylab = "Fitted (BLUP)",
     main = "Observed vs Fitted", pch = 16, col = "steelblue")
abline(0, 1, col = "red", lwd = 2)
```

---

## 7. Comparison: With and Without Random Effects

To appreciate the value of random effects, compare NMF-RE with a standard NMF covariate model (no random effects).

```{r compare-models}
# Standard NMF with covariates (no random effects)
res_fixed <- nmfkc(Y, A = A, rank = 1)

cat("=== Standard NMF (fixed effects only) ===\n")
cat("R-squared:", round(1 - sum((Y - res_fixed$XB)^2) / sum((Y - mean(Y))^2), 4), "\n\n")

cat("=== NMF-RE (fixed + random effects) ===\n")
cat("R-squared (XB):      ", round(res$r.squared.fixed, 4), "\n")
cat("R-squared (XB+blup): ", round(res$r.squared, 4), "\n")
cat("ICC:                 ", round(res$ICC, 4), "\n")
```

The improvement from fixed-only $R^2$ to BLUP $R^2$ quantifies the contribution of individual random effects.

---

## Separated Inference with `nmfre.inference()`

When fitting NMF-RE, you can separate optimization from statistical inference. First fit with `wild.bootstrap = FALSE` (faster), then run inference separately:

```{r separated-inference}
# Fit without bootstrap (fast)
res_fast <- nmfre(Y, A, rank = 1, df.rate = 0.2, prefix = "Trend",
                  wild.bootstrap = FALSE)

# Run inference separately
res_inf <- nmfre.inference(res_fast, Y, A, wild.B = 500)
res_inf$coefficients[, c("Basis", "Covariate", "Estimate", "SE", "p_value")]
```

### Visualization with `nmfkc.DOT()`

The inference result is compatible with `nmfkc.DOT()` for graph visualization. Edges are filtered by significance level and decorated with stars:

```{r dot-visualization, eval=requireNamespace("DiagrammeR", quietly=TRUE)}
dot <- nmfkc.DOT(res_inf, type = "YXA", sig.level = 0.05)
plot(dot)
```

---

## Summary

NMF-RE provides a principled way to model **individual heterogeneity** in NMF:

| Step | Function | Purpose |
|:---|:---|:---|
| 1 | `nmfre.dfU.scan()` | Select regularization level |
| 2 | `nmfre()` | Fit the mixed-effects model |
| 3 | `summary()` | Examine variance components and inference |
| 4 | `nmfre.inference()` | Separated inference (optional) |
| 5 | `nmfkc.DOT()` | Visualize with significance stars |

**When to use NMF-RE:**

- Longitudinal or repeated-measures data where subjects differ systematically.
- Panel data with unit-level heterogeneity.
- Any NMF application where fixed covariates alone leave structured residual variation.

For more details on the underlying methodology, see:

- Satoh, K. (2026). Wild Bootstrap Inference for Non-Negative Matrix Factorization with Random Effects. arXiv:2603.01468. <https://arxiv.org/abs/2603.01468>

---
title: "Topic Modeling with nmfkc"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Topic Modeling with nmfkc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteDepends{quanteda}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7, 
  fig.height = 5
)
```

## Introduction

Non-negative Matrix Factorization (NMF) is a powerful technique for topic modeling. By decomposing a document-term matrix, we can simultaneously discover latent **Topics** (clusters of words) and their **Trends** (proportions in documents).

This vignette demonstrates how to use the `nmfkc` package to analyze the U.S. presidential inaugural addresses using the `quanteda` package.

We will cover:

1.  **Data Preparation**: Converting text data into a matrix format suitable for `nmfkc`.
2.  **Rank Selection**: Determining the optimal number of topics using robust Cross-Validation.
3.  **Standard NMF**: Extracting basic topics and interpreting key words.
4.  **Kernel NMF**: Modeling the **temporal evolution** of topics using the "Year" as a covariate.

First, let's load the necessary packages.

```{r load-packages}
library(nmfkc)
library(quanteda)
```

## 1\. Data Preparation

We create a Document-Feature Matrix (DFM) where rows represent documents and columns represent words.

```{r data-prep}
# Load the corpus from quanteda
corp <- corpus(data_corpus_inaugural)

# Preprocessing: tokenize, remove stopwords, and punctuation
tok <- tokens(corp, remove_punct = TRUE)
tok <- tokens_remove(tok, pattern = stopwords("en", source = "snowball"))

# Create DFM and filter
df <- dfm(tok)
df <- dfm_select(df, min_nchar = 3) # Remove short words (<= 2 chars)
df <- dfm_trim(df, min_termfreq = 100) # Remove rare words (appearing < 100 times)

# --- CRITICAL STEP ---
# quanteda's DFM is (Documents x Words).
# nmfkc expects (Words x Documents).
# We must transpose the matrix.
d <- as.matrix(df)
# Sort by frequency
index <- order(colSums(d), decreasing=TRUE) 
d <- d[,index]
Y <- t(d)

dim(Y) # Features (Words) x Samples (Documents)
```

## 2\. Rank Selection (Number of Topics)

Before fitting the model, we need to decide the number of topics ($rank$). The `nmfkc.rank()` function helps us choose an appropriate rank.

The default `detail = "full"` performs **Element-wise Cross-Validation (Wold's CV)**. This method randomly holds out individual matrix elements and evaluates how well the model predicts them, providing a robust measure for rank selection (though it takes more computation time).

```{r rank-selection, fig.width=7, fig.height=6}
# Evaluate ranks from 2 to 6
nmfkc.rank(Y, rank = 2:6)
```

Looking at the diagnostics (the minimum ECV Sigma and the elbow of the R-squared curve; the green `eff.rank` line is shown for context), let's assume **Rank = 3** is a reasonable choice for this overview.

## 3\. Standard NMF

We fit the standard NMF model ($Y \approx XB$) with `rank = 3`.
In the context of topic modeling:

  * **$X$ (Basis Matrix):** Represents **Topics** (distribution of words).
  * **$B$ (Coefficient Matrix):** Represents **Trends** (distribution of topics across documents).

<!-- end list -->

```{r model-fit}
rank <- 3
# Set seed for reproducibility
res_std <- nmfkc(Y, rank = rank, seed = 123, prefix = "Topic")

# Check Goodness of Fit (R-squared)
res_std$r.squared
```

### Interpreting Topics (Keywords)

We can identify the meaning of each topic by looking at the words with the highest weights in the basis matrix `X`.

```{r interpret-topics}
# Extract top 10 words for each topic from X.prob (normalized X)
Xp <- res_std$X.prob
for(q in 1:rank){
  message(paste0("----- Featured words on Topic [", q, "] -----"))
  print(paste0(rownames(Xp), "(", rowSums(Y), ") ", round(100*Xp[,q], 1), "%")[Xp[,q]>=0.5])
}
```

*(Note: Interpretation depends on the result. For example, one topic might contain words like "government, people, states" (Political), while another might have "world, peace, freedom" (International).)*

## 4\. Kernel NMF: Temporal Topic Evolution

One of the unique features of `nmfkc` is **Kernel NMF**.
In standard NMF, the order of documents is ignored; each speech is treated independently. However, inaugural addresses have a strong **temporal component**. By using the "Year" as a covariate, we can smooth the topic proportions over time to see historical shifts.

### Optimizing the Kernel Parameter

We construct a covariate matrix `U` using the year of the address. We then find the optimal kernel bandwidth (`beta`) using Cross-Validation.

```{r kernel-prep}
# Covariate: Year of the address
years <- as.numeric(substring(names(data_corpus_inaugural), 1, 4))
U <- t(as.matrix(years))

# Optimize beta (Gaussian Kernel width)
# We test a specific range of betas to locate the minimum CV error.
beta_candidates <- c(0.2, 0.5, 1, 2, 5) / 10000

# Run CV to find the best beta
# Note: We use the same rank (rank=3) as selected above.
cv_res <- nmfkc.kernel.beta.cv(Y, rank = rank, U = U, beta = beta_candidates, plot = FALSE)
best_beta <- cv_res$beta
print(best_beta)
```

### Fitting Kernel NMF

Now we fit the model using the kernel matrix `A`. This enforces that documents close in time (similar years) should have similar topic distributions.

```{r kernel-fit}
# Create Kernel Matrix
A <- nmfkc.kernel(U, beta = best_beta)

# Fit NMF with Kernel Covariates
res_ker <- nmfkc(Y, A = A, rank = rank, seed = 123, prefix = "Topic")
```

### Visualization: Standard vs Kernel NMF

Let's compare how topic proportions change over time.

  * **Standard NMF (Top):** Shows noisy fluctuations. It captures the specific content of each speech but misses the larger historical context.
  * **Kernel NMF (Bottom):** Reveals smooth historical trends, showing how themes like "Nation Building" vs "Global Affairs" have evolved over centuries.

<!-- end list -->

```{r plot-comparison, fig.width=8, fig.height=8}
oldpar <- par(mfrow = c(2, 1), mar = c(4, 4, 2, 1))

# Prepare Axis Labels (Rounded to integers)
at_points <- seq(1, ncol(Y), length.out = 10)
labels_years <- round(seq(min(years), max(years), length.out = 10))

# 1. Standard NMF (Noisy)
barplot(res_std$B.prob, col = 2:(rank+1), border = NA, xaxt='n',
        main = "Standard NMF: Topic Proportions (Noisy)", ylab = "Probability")
axis(1, at = at_points, labels = labels_years)

# 2. Kernel NMF (Smooth trend)
barplot(res_ker$B.prob, col = 2:(rank+1), border = NA, xaxt='n',
        main = "Kernel NMF: Temporal Topic Evolution (Smooth)", ylab = "Probability")
axis(1, at = at_points, labels = labels_years)

# Legend
legend("topright", legend = paste("Topic", 1:rank), fill = 2:(rank+1), bg="white", cex=0.8)
par(oldpar)
```

---
title: "NMF-FFB with nmfkc"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{NMF-FFB with nmfkc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteDepends{lavaan}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4
)
```

## Overview

This vignette demonstrates how to fit the **NMF-FFB**
(Nonnegative Matrix Factorization with Feed-Forward + Feedback;
formerly NMF-SEM) implemented in the **nmfkc** package.

NMF-FFB extends standard nonnegative matrix factorization by introducing
**feed-forward exogenous loading and feedback in latent space**.

Observed endogenous variables $Y_1$ are approximated as

$$
Y_1 \approx X (\Theta_1 Y_1 + \Theta_2 Y_2).
$$

If the feedback operator $XC_1$ is stable, the model implies the equilibrium mapping

$$
\hat Y_1 \approx (I - X\Theta_1)^{-1} X\Theta_2 Y_2
\equiv M_{\mathrm{model}} Y_2.
$$

---

## 1. Load and preprocess data

```{r}
library(lavaan)
data(HolzingerSwineford1939)

d <- HolzingerSwineford1939
index <- complete.cases(d)
d <- d[index, ]
```

### 1.1 Construct exogenous variables

```{r}
d$age.rev <- -(d$ageyr + d$agemo / 12)
d$sex.2 <- ifelse(d$sex == 2, 1, 0)
d$school.GW <- ifelse(d$school == "Grant-White", 1, 0)
d[, c("id", "sex", "ageyr", "agemo", "school", "grade")] <- NULL
```

### 1.2 Nonnegative normalization

```{r}
library(nmfkc)
d <- nmfkc::nmfkc.normalize(d)
```

## 2. Split into endogenous and exogenous blocks

```{r}
exogenous_vars <- c("age.rev", "sex.2", "school.GW")
endogenous_vars <- setdiff(colnames(d), exogenous_vars)

Y1 <- t(d[, endogenous_vars])
Y2 <- t(d[, exogenous_vars])
```

## 3. Baseline NMF model

```{r}
myepsilon <- 1e-6
Q0 <- 3

res0 <- nmfkc(
  Y = Y1,
  A = Y2,
  rank = Q0,
  epsilon = myepsilon,
  X.L2.ortho = 100
)

M.simple <- res0$X %*% res0$C
```

## 3.5 Hyperparameter Tuning for Sparsity (Optional)

```{r,eval=FALSE}
grid_params <- expand.grid(
    C1.L1     = c(0,1:9/10,1:10),
    C2.L1     = c(0,1:9/10,1:10)
)
n_iter <- nrow(grid_params)
mae.cv <- 0*1:n_iter

for(i in 1:n_iter){
  if (i %% round(n_iter / 10) == 0) {
    message(sprintf("Processing... %d%% (%d/%d)", round(i/n_iter*100), i, n_iter)) 
  }  
  p <- grid_params[i, ]
  res.cv <- nmf.ffb.cv(Y1, Y2, rank = Q0,
                       X.init = res0$X,
                       X.L2.ortho = 100,
                       C1.L1 = p$C1.L1,
                       C2.L1 = p$C2.L1,
                       seed = 1, epsilon = myepsilon)
  mae.cv[i] <- res.cv
}

f <- data.frame(grid_params,mae.cv)
f <- f[order(f$mae.cv),]
head(f,5)
#    C2.L2.off C1.L1 C2.L1    mae.cv
#140         0    10   0.6 0.1820841
#160         0    10   0.7 0.1820843
#180         0    10   0.8 0.1820877
#200         0    10   0.9 0.1820907
#120         0    10   0.5 0.1820908
print(p <- f[1,])
```


## 4. NMF-FFB estimation

```{r}
p <- list(C1.L1 = 10, C2.L1 = 0.6)

res <- nmf.ffb(
  Y1, Y2,
  rank = Q0,
  X.init = res0$X,
  X.L2.ortho = 100,
  C1.L1 = p$C1.L1,
  C2.L1 = p$C2.L1,
  epsilon = myepsilon
)
```

```{r}
plot(res$objfunc.full, type = "l",
     main = "Objective Function",
     ylab = "Loss", xlab = "Iteration")
```

## 5. Diagnostics

```{r}
SC.map <- cor(as.vector(res$M.model),
              as.vector(M.simple))

cat("Q=     ", Q0, "\n")
cat("RHO=   ", round(res$XC1.radius, 3), "\n")
cat("AR=    ", round(res$amplification, 3), "\n")
cat("SCmap= ", round(SC.map, 3), "\n")
cat("SCcov= ", round(res$SC.cov, 3), "\n")
cat("MAE=   ", round(res$MAE, 3), "\n")
```

## 6. Visualization

```{r}
res.dot <- nmf.ffb.DOT(
  res,
  weight_scale = 5,
  rankdir = "TB",
  threshold = 0.01,
  fill = FALSE,
  cluster.box = "none"
)

# plot(res.dot)  # requires DiagrammeR
```

---
title: "Soft community detection in networks with nmfkc.net"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Soft community detection in networks with nmfkc.net}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

## Introduction

`nmfkc.net()` is a **symmetric** NMF for network (adjacency) matrices.  With
`type = "bi"` it factorises a symmetric \eqn{N \times N} matrix as
\eqn{Y \approx X X^\top} with \eqn{X \ge 0}.  Each row of \eqn{X} is a node's
**graded membership** across communities, so this is *soft* community detection:
unlike a hard method (e.g.\ Louvain), boundary nodes are revealed by a mixed
membership rather than being forced into one group.

This vignette uses a **self-contained synthetic network with a known community
structure**, so we can check that `nmfkc.net()` recovers it and that
`nmfkc.net.rank()` finds the right number of communities.

```{r lib}
library(nmfkc)
```

## A synthetic network (stochastic block model, 3 communities)

We draw an undirected graph on `60` nodes from a **stochastic block model**:
three communities of 20 nodes, with a high within-community edge probability and
a low between-community one.  We then deliberately add **three "bridge" nodes**
(one per community), each also wired to a *second* community, so that the network
contains a few genuinely mixed nodes.  The true community of each node is known.

```{r data, fig.width = 6, fig.height = 5.2}
set.seed(1)
K <- 3; n_each <- 20; N <- K * n_each       # 60 nodes, 3 communities
block <- rep(1:K, each = n_each)            # true community labels

p_in <- 0.6; p_out <- 0.05
Prob <- matrix(p_out, N, N)
for (k in 1:K) Prob[block == k, block == k] <- p_in
Y <- matrix(rbinom(N * N, 1, Prob), N, N)   # 0/1 adjacency
Y[lower.tri(Y)] <- t(Y)[lower.tri(Y)]        # make symmetric
diag(Y) <- 0

# add three bridge nodes, each also linked to a second community
bridge <- c(20, 40, 60); into <- c(2, 3, 1)
for (i in seq_along(bridge)) {
  b <- bridge[i]; tgt <- which(block == into[i])
  e <- rbinom(length(tgt), 1, 0.45)
  Y[b, tgt] <- pmax(Y[b, tgt], e); Y[tgt, b] <- Y[b, tgt]
}

isSymmetric(Y); sum(Y) / 2                   # symmetric, number of edges

# adjacency matrix in the original (random) node order -- structure is hidden
image(1:N, 1:N, Y[, N:1], col = c("white", "steelblue"),
      xlab = "node", ylab = "node", main = "Adjacency (original order)")
```

## Soft community detection with `nmfkc.net`

```{r fit}
res <- nmfkc.net(Y, rank = 3, type = "bi", nstart = 10, maxit = 500)
res$r.squared                 # fit
head(round(res$X.prob, 2))    # soft membership: rows = nodes, columns = communities
head(res$X.cluster)           # hard label = argmax membership
```

## Did it recover the known communities?

```{r compare}
table(true = block, estimated = res$X.cluster)

# the three bridge nodes have genuinely split memberships
round(res$X.prob[bridge, ] * 100, 1)
```

The cross-tabulation is essentially block-diagonal: apart from the bridge nodes,
every node is assigned to its true community.  The bridge nodes (rows above) are
roughly **50/50 between two communities**, so their hard label simply tips to
whichever side is marginally stronger --- exactly the situation where a *soft*
membership is more informative than a forced assignment.

## Visualising the result

Re-ordering the nodes by the **recovered** community turns the adjacency matrix
into a clean block-diagonal, and the membership matrix shows three crisp blocks:

```{r viz, fig.width = 8, fig.height = 4}
ord <- order(res$X.cluster)
op <- par(mfrow = c(1, 2), mar = c(4, 4, 3, 1))
image(1:N, 1:N, Y[ord, rev(ord)], col = c("white", "steelblue"),
      xlab = "node (reordered)", ylab = "", main = "Adjacency by community")
image(1:N, 1:K, res$X.prob[ord, , drop = FALSE],
      col = hcl.colors(20, "Blues", rev = TRUE),
      xlab = "node (reordered)", ylab = "community", axes = FALSE,
      main = "Soft membership")
axis(1); axis(2, at = 1:K)
par(op)
```

## Network diagram (Graphviz DOT): soft vs. hard

`nmfkc.net.DOT()` turns the fit into a Graphviz **DOT** graph: each network node
is linked to the community (basis) hubs by edges whose thickness encodes the
membership strength.  We raise `threshold` to `0.25` so that only substantial
memberships are drawn --- each pure node then links to a *single* hub, which lets
the `neato` layout pull the three communities apart, while the bridge nodes keep
*both* links and settle in between.  Rendered with the suggested package
**DiagrammeR**, it embeds directly in the HTML output.  The only difference
between the two views below is the **node colour**:

* `Y.cluster = "soft"` (default) blends the community colours in proportion to
  the membership, so the three bridge nodes appear in a **mixed colour**.
* `Y.cluster = "hard"` paints each node in the **single colour** of its dominant
  community.

```{r dot-check, include = FALSE}
has_dg <- requireNamespace("DiagrammeR", quietly = TRUE)
```

**Soft** -- the three bridge nodes added earlier (`bridge <- c(20, 40, 60)`,
each wired to a second community) show a blended colour:

```{r dot-soft, eval = has_dg, fig.width = 7, fig.height = 7}
dot_soft <- nmfkc.net.DOT(res,
  Y.label = as.character(1:N), X.label = paste("Community", 1:K),
  Y.title = "Network nodes",   X.title = "Communities",
  layout  = "neato", threshold = 0.25, Y.cluster = "soft")
DiagrammeR::grViz(as.character(dot_soft))
```

**Hard** -- those same bridge nodes (20, 40, 60) are forced into one community
colour:

```{r dot-hard, eval = has_dg, fig.width = 7, fig.height = 7}
dot_hard <- nmfkc.net.DOT(res,
  Y.label = as.character(1:N), X.label = paste("Community", 1:K),
  Y.title = "Network nodes",   X.title = "Communities",
  layout  = "neato", threshold = 0.25, Y.cluster = "hard")
DiagrammeR::grViz(as.character(dot_hard))
```

(If DiagrammeR is not installed, both diagrams are skipped.)

## How many communities? `nmfkc.net.rank`

The same rank-selection diagnostics as `nmfkc.rank` are available for the
symmetric model.  The recommended rank is the cross-validation (ECV) minimum;
the effective rank and R-squared elbow corroborate it.

```{r rank, fig.width = 7, fig.height = 6}
rk <- nmfkc.net.rank(Y, rank = 1:6, type = "bi", nstart = 5)
rk$rank.best
round(rk$criteria, 3)
```

## Remarks

`nmfkc.net()` recovers the three planted communities for the core nodes, and
`nmfkc.net.rank()` selects three.  Because the membership is *soft*
(`res$X.prob`), the three bridge nodes that sit between communities show a split
membership instead of being forced into one group --- the practical advantage of
NMF over hard community detection, made visible by the soft-vs-hard DOT graphs
above.

`type = "bi"` (\eqn{Y \approx X X^\top}) is the usual choice for an undirected,
non-negative network; `type = "tri"` (\eqn{Y \approx X C X^\top}) adds a
community-interaction matrix \eqn{C}, and `type = "signed"` handles networks with
negative weights.  For choosing the number of communities, the same advice as in
the *rank-selection* vignette applies: use the ECV minimum as the primary
criterion and corroborate it with the effective rank.

See `?nmfkc.net`, `?nmfkc.net.ecv`, `?nmfkc.net.rank`, and the companion vignette
*"Choosing the NMF rank on data with a known true rank"*.

---
title: "Choosing the NMF rank on data with a known true rank"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Choosing the NMF rank on data with a known true rank}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

## Introduction

Choosing the number of basis vectors (the *rank* `Q`) is the central tuning
problem in NMF.  `nmfkc` offers several rank-selection tools, each based on a
different principle:

| function | principle | a good rank... |
|---|---|---|
| `nmfkc.rank` | integrated diagnostics: R-squared, effective rank, element-wise CV | minimises CV error / R-squared elbow |
| `nmfkc.ecv` | element-wise cross-validation (Wold) | minimises held-out error |
| `nmfkc.bicv` | block bi-cross-validation (Owen & Perry) | minimises held-out error |
| `nmfkc.consensus` | clustering stability (Brunet) | maximises stability (or minimises PAC) |
| `nmfkc.ard` | Bayesian automatic relevance determination (Tan & Fevotte) | surviving components after pruning |

Since there is no single "best" criterion, this vignette applies all of them
to a **synthetic dataset whose true rank is known**, so we can see which ones
recover it.

```{r lib}
library(nmfkc)
```

## A synthetic dataset (true rank = 3)

We build a non-negative `24 x 60` matrix from **three** feature blocks and
**three** sample clusters (20 samples each), then add noise.  By construction
the true rank is `3`.

```{r data, fig.width = 6, fig.height = 4}
set.seed(123)
r_true <- 3
P <- 24                       # features
n_each <- 20                  # samples per cluster
N <- r_true * n_each          # 60 samples

# basis: each latent factor is a distinct block of features
X <- matrix(0, P, r_true)
X[1:8, 1] <- 1; X[9:16, 2] <- 1; X[17:24, 3] <- 1
X <- X + matrix(runif(P * r_true, 0, 0.1), P, r_true)   # small background

# coefficients: each sample loads mainly on its own cluster's factor
cl_true <- rep(1:r_true, each = n_each)
B <- matrix(runif(r_true * N, 0, 0.2), r_true, N)
for (j in seq_len(N)) B[cl_true[j], j] <- runif(1, 1, 2)

Y <- X %*% B
Y <- Y + matrix(rnorm(P * N, 0, 0.15 * mean(X %*% B)), P, N)  # noise
Y[Y < 0] <- 0

dim(Y)
image(1:N, 1:P, t(Y), xlab = "sample", ylab = "feature",
      main = "Synthetic data (true rank = 3)")
```

## 1. Integrated diagnostics: `nmfkc.rank`

`nmfkc.rank()` fits each rank and draws three curves on one plot:
**R-squared** (red, with a "Best (Elbow)" marker), the broken-stick-corrected
**effective rank** index (green, shown for context), and the element-wise
**ECV error** (blue, with a "Best (Min)" marker).  The recommended rank is the
ECV minimum (or the R-squared elbow).

```{r rank, fig.width = 7, fig.height = 6}
rk <- nmfkc.rank(Y, rank = 1:6)
rk$rank.best                       # recommended rank (ECV minimum)
round(rk$criteria, 3)
```

## 2. Cross-validation: `nmfkc.ecv` and `nmfkc.bicv`

Both are predictive engines that return the held-out error per rank; the
recommended rank minimises it.  `nmfkc.ecv` holds out scattered *entries*
(Wold's CV); `nmfkc.bicv` holds out a *block* of rows and columns at once
(Owen & Perry).

```{r cv}
ev <- nmfkc.ecv (Y, rank = 1:6)
bv <- nmfkc.bicv(Y, rank = 1:6)   # nfolds = 2 (Owen & Perry) by default
data.frame(rank = 1:6,
           sigma.ecv  = round(ev$sigma, 3),
           sigma.bicv = round(bv$sigma, 3))
cat("ECV  picks rank", which.min(ev$sigma),
    "| bi-CV picks rank", bv$rank[which.min(bv$sigma)], "\n")
```

## 3. Clustering stability: `nmfkc.consensus`

For each rank, NMF is run many times from random initialisations and the
reproducibility of the sample clustering is summarised by `cophenetic`
(Brunet), `dispersion` (Kim & Park; higher = crisper) and `pac` (Senbabaoglu;
lower = less ambiguous).

```{r consensus, fig.width = 7, fig.height = 5}
cs <- nmfkc.consensus(Y, rank = 2:6, nrun = 20, keep.consensus = TRUE)
cs
plot(cs)                               # stability curves
```

Each panel is the consensus matrix at one rank, reordered by hierarchical
clustering (blue = 0 / different cluster, red = 1 / same cluster):

```{r consensus-heatmap, fig.width = 8, fig.height = 6}
plot(cs, type = "heatmap")   # all ranks
```

A striking point: the consensus stays essentially **three crisp blocks even at
ranks 4--6**.  The surplus basis vectors do **not** create new *reproducible*
clusters --- at ranks 4--5 they stay near-empty (the hard clustering still uses
only three factors), and at rank 6 a cluster splits only *erratically* across
restarts, which averages out (just a few intermediate, pink entries, and a small
drop in `dispersion` / rise in `PAC`).  So consensus reveals the genuine stable
structure (`3`) robustly, **even when the rank is over-specified** --- a useful
property.  (Use `plot(cs, type = "heatmap", rank = 3)` to enlarge a single rank
with sample labels.)

## 4. Bayesian ARD: `nmfkc.ard`

ARD fits NMF **once** at an over-complete rank and prunes unneeded components
automatically; the relevance bar plot is read like a scree plot (look for the
knee).  Because it is a sensitive point estimate, we aggregate over several
random restarts (`nrun`).

```{r ard, fig.width = 6.5, fig.height = 4.5}
ar <- nmfkc.ard(Y, rank = 8, nrun = 20)   # >=10-20 restarts for a stable mode
ar                                        # see "rank over runs" for confidence
plot(ar)
```

## Summary: did they recover the true rank?

```{r summary}
data.frame(
  method = c("nmfkc.rank (ECV)", "nmfkc.ecv", "nmfkc.bicv",
             "nmfkc.consensus (dispersion)", "nmfkc.consensus (PAC)",
             "nmfkc.ard"),
  estimate = c(rk$rank.best,
               which.min(ev$sigma),
               bv$rank[which.min(bv$sigma)],
               cs$rank[which.max(cs$dispersion)],
               cs$rank[which.min(cs$pac)],
               ar$rank),
  true = r_true
)
```

## How the clustering evolves with rank

A complementary view: `nmf.cluster.flow()` draws a Sankey / alluvial diagram of
how the hard sample clustering changes as the rank grows.  We fit ranks `1:6`
(so the list index equals the rank) and colour the flows by the clustering at
the **true rank**, `reference = 3`.  The adjacent-rank ARI (agreement) is printed
along the top.

```{r flow, fig.width = 8, fig.height = 6}
fits <- lapply(1:6, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
fl <- nmf.cluster.flow(fits, reference = 3)
head(fl$clusters)      # rows = individuals, columns = rank, entries = cluster id
```

At `rank = 1` all 60 samples form a single group; at `rank = 3` they split
cleanly into the three true clusters (20 each); beyond that the groups
over-split (the ARI to the next rank starts to fall) --- visual confirmation
that `3` is the right number of basis vectors here.

## Remarks

On this clean dataset with a clear rank-3 structure, **almost every criterion
recovers the true rank**: the predictive cross-validations (`nmfkc.ecv`,
`nmfkc.bicv`), the integrated `nmfkc.rank`, the consensus `dispersion` (with a
crisp three-block consensus matrix) and ARD (every `nrun` restart agrees on
`3`, relevance dropping to zero after the third component) all give `3`.  Only
`PAC` leans to the coarser rank `2` --- a small hint of the low-rank bias of
stability measures.

On **real** data the criteria often *disagree* much more: stability/consensus
tends to favour low (coarse) ranks, and ARD's count can track the starting rank
when the energy spectrum decays smoothly (no clear knee).  A robust practical
recipe is to **triangulate**:
use ECV for the predictive upper bound, the effective-rank / relevance *knee*
for the lower bound, and the consensus heatmap to confirm an interpretable block
structure --- then choose one rank by purpose and domain knowledge.

See `?nmfkc.rank`, `?nmfkc.ecv`, `?nmfkc.bicv`, `?nmfkc.consensus` and
`?nmfkc.ard` (all cross-referenced) for details.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{plot.nmf.cluster.flow}
\alias{plot.nmf.cluster.flow}
\title{Plot a cluster-flow (alluvial) diagram}
\usage{
\method{plot}{nmf.cluster.flow}(
  x,
  col = NULL,
  lwd = 1,
  xlab = "rank (Q)",
  ylab = "individuals",
  main = "Cluster flow across rank",
  ...
)
}
\arguments{
\item{x}{An object of class \code{"nmf.cluster.flow"}.}

\item{col}{Optional colour vector indexed by \strong{reference}
cluster id (\code{col[k]} colours every individual whose reference
cluster is \code{k}); recycled if shorter than the number of
reference clusters.  Defaults to the object's palette.}

\item{lwd}{Line width of the flow segments.}

\item{xlab, ylab}{Axis labels.}

\item{main}{Plot title.}

\item{...}{Further arguments passed to the initial
\code{\link[graphics]{plot}} call.}
}
\value{
\code{x}, invisibly.
}
\description{
Draws the alluvial / Sankey-style cluster-flow diagram for an object
created by \code{\link{nmf.cluster.flow}}.
}
\seealso{
\code{\link{nmf.cluster.flow}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{print.summary.nmfkc.net}
\alias{print.summary.nmfkc.net}
\title{Print method for summary.nmfkc.net objects}
\usage{
\method{print}{summary.nmfkc.net}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{An object of class \code{"summary.nmfkc.net"}.}

\item{digits}{Minimum number of significant digits.}

\item{...}{Unused.}
}
\value{
Invisible \code{x}.
}
\description{
Print method for summary.nmfkc.net objects
}
\seealso{
\code{\link{summary.nmfkc.net}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{print.summary.nmfae.inference}
\alias{print.summary.nmfae.inference}
\title{Print method for summary.nmfae.inference objects}
\usage{
\method{print}{summary.nmfae.inference}(x, digits = max(3L, getOption("digits") - 3L), max.coef = 20, ...)
}
\arguments{
\item{x}{An object of class \code{"summary.nmfae.inference"}.}

\item{digits}{Minimum number of significant digits.}

\item{max.coef}{Maximum coefficient rows to display. Default is 20.}

\item{...}{Additional arguments (currently unused).}
}
\value{
Called for its side effect (printing). Returns \code{x} invisibly.
}
\description{
Prints a formatted summary including the coefficients table.
}
\seealso{
\code{\link{summary.nmfae.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.ard.R
\name{plot.nmfkc.ard}
\alias{plot.nmfkc.ard}
\title{Plot method for nmfkc.ard objects}
\usage{
\method{plot}{nmfkc.ard}(x, main = NULL, ...)
}
\arguments{
\item{x}{An object of class \code{"nmfkc.ard"}.}

\item{main}{Plot title.}

\item{...}{Passed to \code{\link[graphics]{barplot}}.}
}
\value{
\code{x}, invisibly.
}
\description{
Bar plot of the per-component relevance (descending), with a
line at the pruning threshold; bars above it (the estimated rank) are
highlighted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{nmfkc.net.inference}
\alias{nmfkc.net.inference}
\title{Statistical Inference for Symmetric NMF Parameters}
\usage{
nmfkc.net.inference(object, Y, wild.bootstrap = TRUE, ...)
}
\arguments{
\item{object}{A fitted model returned by \code{nmfkc()} with
\code{Y.symmetric = "bi"} or \code{"tri"}, or the newer
\code{\link{nmfkc.net}} with \code{type = "bi"} or \code{"tri"}.}

\item{Y}{The original symmetric matrix (P x P).}

\item{wild.bootstrap}{Logical; if TRUE (default), perform wild bootstrap
inference for the C matrix.}

\item{...}{Additional arguments passed to \code{nmfkc.inference()}
(e.g., \code{wild.B}, \code{wild.seed}, \code{C.p.side}).}
}
\value{
The object augmented with inference fields
(same as \code{nmfkc.inference}), with class
\code{c("nmfkc.net.inference", "nmfkc.inference", "nmf.inference", "nmfkc", "nmf")}.
}
\description{
Performs statistical inference on the parameter matrix \eqn{C} of a
symmetric NMF model (\eqn{Y \approx XCX^\top}).

This is a wrapper around \code{\link{nmfkc.inference}} that automatically
sets the covariate matrix to \eqn{A = X^\top}, which is the defining
property of symmetric NMF.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\examples{
\donttest{
library(nmfkc)
Y <- matrix(c(0,1,1,0,0,0,
              1,0,1,0,0,0,
              1,1,0,1,0,0,
              0,0,1,0,1,1,
              0,0,0,1,0,1,
              0,0,0,1,1,0), 6, 6)
res <- nmfkc.net(Y, rank = 2, type = "tri", nstart = 20)
res_inf <- nmfkc.net.inference(res, Y)
summary(res_inf)
}

}
\seealso{
\code{\link{nmfkc.inference}}, \code{\link{nmfkc.net.DOT}},
\code{\link{summary.nmfkc.net.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{nmfae.signed.ecv}
\alias{nmfae.signed.ecv}
\title{Element-wise Cross-Validation for Signed-Bottleneck NMF-AE}
\usage{
nmfae.signed.ecv(Y1, Y2 = Y1, rank = 1:2, rank.encoder = NULL, ...)
}
\arguments{
\item{Y1}{Output matrix (P1 x N).}

\item{Y2}{Input matrix (P2 x N).  Default \code{Y1}.}

\item{rank}{Integer vector of candidate Q values.  Default \code{1:2}.}

\item{rank.encoder}{Integer vector of candidate R values, or \code{NULL}
(default: pair R = Q, diagonal grid).}

\item{...}{Additional arguments:
\describe{
\item{\code{nfolds} / \code{div}}{Number of folds.  Default 5.}
\item{\code{seed}}{RNG seed for fold assignment.  Default 123.}
\item{\code{nstart}}{Number of random restarts per fit.  Default 1.
Signed models have more local minima (the bottleneck can carry
both signs), so \code{nstart >= 10} is recommended for
reproducible rank selection.}
\item{Other args}{\code{epsilon}, \code{maxit}, \code{warm.start},
etc.\ are passed to \code{\link{nmfae.signed}}.}
}}
}
\value{
An object of class \code{c("nmfae.signed.ecv", "nmfae.ecv")} with
\code{objfunc} (MSE per pair), \code{sigma} (RMSE), \code{objfunc.fold}
(per-fold MSE), \code{folds}, \code{QR}, \code{paired}.
}
\description{
Element-wise k-fold cross-validation for \code{\link{nmfae.signed}} to
select the decoder / encoder ranks \eqn{(Q, R)}.  Mirrors
\code{\link{nmfae.ecv}} but uses the \strong{weighted} Signed-Bottleneck NMF-AE fit
path (\code{Y1.weights}): test-fold elements are zero-weighted during
fitting, and held-out MSE is computed on those elements.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae.signed}}, \code{\link{nmfae.ecv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.signed.rff.R
\name{nmfkc.signed.rff}
\alias{nmfkc.signed.rff}
\title{Random Fourier Features for nmfkc.signed()}
\usage{
nmfkc.signed.rff(U, beta = NULL, D = ceiling(ncol(U)/2), seed = NULL, ...)
}
\arguments{
\item{U}{A \eqn{p \times N} numeric matrix; columns are data points.}

\item{beta}{Positive scalar. Gaussian kernel bandwidth parameter.
Can be obtained via \code{\link{nmfkc.kernel.beta.nearest.med}}.
May be \code{NULL} only when \code{pars} is supplied through
\code{...}.}

\item{D}{Integer. Number of random features. Defaults to
\code{ceiling(ncol(U) / 2)}.  This default is intended for the
\strong{training-time fresh generation} only; for test data, always
supply \code{pars} via \code{...} to inherit the training \eqn{D}
together with \eqn{\omega, b}.  For very large \eqn{N} the default
may be excessive (direct MU cost is \eqn{O(QD^2)}); choose a
smaller \eqn{D} manually.  For very small \eqn{N}, RFF is not
recommended; use a full kernel matrix with \code{\link{nmfkc}}
instead.}

\item{seed}{Optional integer passed to \code{set.seed()} before
generating \eqn{\omega, b}, for reproducibility.  Ignored when
\code{pars} is supplied.}

\item{...}{Hidden option \code{pars}: a list
\code{list(omega, b, D, beta)} obtained from a previous call
(\code{Ztrain$pars}).  When supplied, \eqn{\omega, b} are reused
and the \code{beta}, \code{D}, \code{seed} arguments are ignored.
Use this to apply the same random map to test data.}
}
\value{
A list with two elements:
\describe{
\item{\code{Z}}{A \eqn{D \times N} sign-unrestricted numeric
matrix.  Pass this to \code{\link{nmfkc.signed}} as its
\code{A} argument.}
\item{\code{pars}}{A list \code{list(omega, b, D, beta)}.  Pass
this to \code{\link{nmfkc.signed}} via its \code{pars}
argument (for summary display) and to subsequent
\code{nmfkc.signed.rff()} calls (to reuse the same random map
on new data).}
}
}
\description{
Generates RFF random parameters
\eqn{\omega_d \sim \mathcal{N}(0, 2\beta I_p)},
\eqn{b_d \sim \mathrm{Uniform}(0, 2\pi)} (Rahimi & Recht, 2007) and
applies the RFF transform
\deqn{z_d(u) = \sqrt{2/D}\, \cos(\omega_d^\top u + b_d)}
to each column of \code{U}, yielding a sign-unrestricted
\eqn{D \times N} feature matrix \eqn{Z} such that
\eqn{Z^\top Z \approx K}, the Gaussian kernel matrix with bandwidth
\code{beta}.

The return value is a list with the feature matrix \code{Z} and
the generating parameters \code{pars = list(omega, b, D, beta)} so
that the same random map can be re-applied to new data (by passing
\code{pars} back) and \code{\link{nmfkc.signed}} can record the
parameters for downstream summary.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\examples{
\donttest{
## Iris 3-class classification with RFF + direct MU (Ding-Li-Jordan)
data(iris)
set.seed(1)
idx <- sample(nrow(iris), 100)    # 100 training, 50 test

## Scale features using TRAINING mean/sd; transpose to p x N layout
mn      <- colMeans(iris[idx, 1:4])
sc      <- apply(iris[idx, 1:4], 2, sd)
U.train <- t(scale(iris[idx,  1:4], center = mn, scale = sc))   # 4 x 100
U.test  <- t(scale(iris[-idx, 1:4], center = mn, scale = sc))   # 4 x  50

## One-hot encode training labels as a Q_obs x N target matrix
levs      <- levels(iris$Species)
Y.train   <- sapply(iris$Species[idx], function(s) as.integer(levs == s))
rownames(Y.train) <- levs                                       # 3 x 100
lab.train <- iris$Species[idx]
lab.test  <- iris$Species[-idx]

## Beta candidates from nearest-neighbour median heuristic
beta_info <- nmfkc.kernel.beta.nearest.med(U.train)
betas     <- beta_info$beta_candidates

## CV over beta candidates: for each beta, generate RFF, fit, and
## evaluate column-wise CV-MSE on training data
cv_mse <- numeric(length(betas))
for (i in seq_along(betas)) {
  rff_i   <- nmfkc.signed.rff(U.train, beta = betas[i], D = 50, seed = 1)
  cv_i    <- nmfkc.signed.cv(Y.train, A = rff_i$Z, rank = 3, seed = 123)
  cv_mse[i] <- cv_i$objfunc
}
beta_best <- betas[which.min(cv_mse)]

## Generate signed RFF features with the best beta
rff.train <- nmfkc.signed.rff(U.train, beta = beta_best, D = 50, seed = 1)
rff.test  <- nmfkc.signed.rff(U.test, pars = rff.train$pars)

## Fit on training data only
res <- nmfkc.signed(Y.train, A = rff.train$Z, rank = 3,
                    pars = rff.train$pars, verbose = FALSE)

## Predict on training and test separately
pred.train <- predict(res, newA = rff.train$Z, type = "class")
pred.test  <- predict(res, newA = rff.test$Z,  type = "class")
mean(pred.train == as.character(lab.train))
mean(pred.test  == as.character(lab.test))
}

}
\seealso{
\code{\link{nmfkc.signed}},
\code{\link{nmfkc.kernel.beta.nearest.med}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{nmfae.kernel.beta.cv}
\alias{nmfae.kernel.beta.cv}
\title{Optimize kernel beta for nmfae by cross-validation}
\usage{
nmfae.kernel.beta.cv(
  Y1,
  rank = 2,
  rank.encoder = rank,
  U,
  V = NULL,
  beta = NULL,
  plot = TRUE,
  ...
)
}
\arguments{
\item{Y1}{Output matrix \eqn{Y_1} (P1 x N). Non-negative.}

\item{rank}{Integer. Rank of the decoder basis. Default is 2.}

\item{rank.encoder}{Integer. Rank of the encoder basis. Default is \code{rank}.}

\item{U}{Covariate matrix \eqn{U} (K x M). Rows are features, columns are samples
(or knot points for non-symmetric kernels).}

\item{V}{Covariate matrix \eqn{V} (K x N). If \code{NULL} (default), \code{V = U}
and a symmetric kernel is used.}

\item{beta}{Numeric vector of candidate beta values. If \code{NULL}, automatically
determined via \code{\link{nmfkc.kernel.beta.nearest.med}}.}

\item{plot}{Logical. If \code{TRUE} (default), plots the objective function curve.}

\item{...}{Additional arguments. Kernel-specific args (\code{kernel}, \code{degree})
are passed to \code{\link{nmfkc.kernel}}; all others
(\code{div}, \code{seed}, \code{shuffle}, \code{epsilon}, \code{maxit}, etc.)
are passed to \code{\link{nmfae.cv}}.
For backward compatibility, \code{Q} and \code{R} are accepted as aliases for
\code{rank} and \code{rank.encoder}.}
}
\value{
A list with components:
\item{beta}{The beta value that minimizes the cross-validation objective.}
\item{objfunc}{Named numeric vector of objective function values for each candidate beta.}
}
\description{
\code{nmfae.kernel.beta.cv} selects the optimal \code{beta} parameter of the
kernel function by evaluating \code{\link{nmfae.cv}} for each candidate value.
The kernel matrix \eqn{A = K(U, V; \beta)} replaces \eqn{Y_2} in the three-layer
NMF model.

When \code{beta = NULL}, candidate values are automatically generated via
\code{\link{nmfkc.kernel.beta.nearest.med}}.
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
U <- matrix(cars$speed, nrow = 1)
res <- nmfae.kernel.beta.cv(Y, rank = 1, rank.encoder = 1, U = U,
                             beta = c(0.01, 0.02, 0.05), nfolds = 5)
res$beta

}
\seealso{
\code{\link{nmfae.cv}}, \code{\link{nmfkc.kernel}},
\code{\link{nmfkc.kernel.beta.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmf.cluster.criteria}
\alias{nmf.cluster.criteria}
\title{Sample-clustering quality across ranks}
\usage{
nmf.cluster.criteria(fits, Y, Y2 = NULL, names = NULL, plot = TRUE, ...)
}
\arguments{
\item{fits}{A list of fitted models, one per rank, all over the same
\eqn{N} individuals (a single fitted model is also accepted and
wrapped automatically).  Supported families: \code{\link{nmfkc}},
\code{\link{nmfkc.signed}}, \code{\link{nmfae}}, \code{nmfae.signed},
\code{\link{nmfkc.net}}, \code{\link{nmfre}}, and
\code{\link{nmf.sem}} / \code{nmf.ffb}.}

\item{Y}{The original data matrix used to fit the models (\eqn{Y_1}
for \code{nmf.ffb}); required for the data-space distances.}

\item{Y2}{Exogenous block, required only for \code{nmf.ffb} /
\code{nmf.sem}.}

\item{names}{Optional character vector (length \code{length(fits)}) of
x-axis tick labels.  Defaults to each result's \code{$rank}.}

\item{plot}{Logical; draw the diagnostics plot immediately
(default \code{TRUE}); see \code{\link{plot.nmf.cluster.criteria}}.}

\item{...}{When \code{plot = TRUE}, graphical arguments forwarded to
\code{\link{plot.nmf.cluster.criteria}}.}
}
\value{
An object of class \code{"nmf.cluster.criteria"} (returned
invisibly): a list with \code{criteria} (a data frame with one row
per result and columns \code{rank}, \code{silhouette}, \code{CPCC},
\code{dist.cor}, and \code{hard}) and \code{labels} (the x-axis
labels).  Results are kept in the given order (not sorted).
}
\description{
Computes the clustering-quality criteria \code{silhouette},
\code{CPCC}, and \code{dist.cor} for a list of models fitted at
different ranks (or a single fit), returning one row per rank.  These
are \strong{clustering-stability} diagnostics (how decisively and
faithfully the samples cluster), conceptually separate from the
rank-selection \code{*.rank} functions (which use r.squared, effective
rank, and ECV) and complementary to \code{\link{nmf.cluster.flow}}
(which shows how the hard clustering itself changes across ranks).

Hard sample clustering requires a non-negative coefficient/score
matrix (so the columns form a membership simplex); when a model's
coefficient is signed (e.g.\ \code{nmfkc.signed}, \code{nmfae.signed},
\code{nmfre} fits whose coefficient has negative entries) the
hard-label \code{silhouette} is \code{NA} while the distance-based
\code{CPCC} and \code{dist.cor} are still computed.
}
\examples{
\donttest{
Y <- t(as.matrix(iris[, 1:4]))
fits <- lapply(2:6, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
cc <- nmf.cluster.criteria(fits, Y, plot = FALSE)
cc$criteria
plot(cc)
}
}
\seealso{
\code{\link{nmf.cluster.flow}}, \code{\link{nmfkc.rank}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{fitted.nmf}
\alias{fitted.nmf}
\alias{fitted.nmfae}
\alias{fitted.nmf.sem}
\title{Extract fitted values from NMF models}
\usage{
\method{fitted}{nmf}(object, ...)

\method{fitted}{nmfae}(object, ...)

\method{fitted}{nmf.sem}(object, ...)
}
\arguments{
\item{object}{A fitted model object of class \code{"nmf"}, \code{"nmfkc"},
\code{"nmfae"}, \code{"nmfre"}, or \code{"nmf.sem"}.}

\item{...}{For \code{nmf.sem}: optionally \code{Y1} and \code{Y2}.}
}
\value{
The fitted matrix \eqn{X B}.
}
\description{
Returns the reconstructed matrix \eqn{\hat{Y} = X B} from a fitted
NMF model.

For \code{nmf.sem} objects, returns the equilibrium prediction
\eqn{\hat{Y}_1 = M_{model} Y_2} if available. Supply \code{Y1} and
\code{Y2} to get the direct reconstruction
\eqn{X (C_1 Y_1 + C_2 Y_2)} instead.
}
\examples{
result <- nmfkc(matrix(runif(50), 5, 10), rank = 2)
fitted(result)

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfae}}, \code{\link{nmfre}},
\code{\link{nmf.sem}}, \code{\link{residuals.nmf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.ard.R
\name{nmfkc.ard}
\alias{nmfkc.ard}
\title{Automatic relevance determination for NMF rank (experimental)}
\usage{
nmfkc.ard(Y, rank = NULL, nrun = 10, plot = FALSE, ...)
}
\arguments{
\item{Y}{Observation matrix (\eqn{F \times N}), non-negative.}

\item{rank}{Over-complete starting rank \eqn{K} (must exceed the true
rank).  \code{NULL} (default) uses \code{min(F, N, 20)}.}

\item{nrun}{Number of random-initialization restarts (default \code{10}).
ARD is a sensitive point estimate, so several restarts are advisable; the
reported rank is the \strong{mode} of the per-run estimates
(\code{rank.runs}), with a representative modal fit kept for
\code{plot}/\code{W}/\code{H}.}

\item{plot}{Logical; draw the relevance bar plot.}

\item{...}{Advanced options, rarely needed (defaults in parentheses):
\code{prior} (\code{"L2"}: half-normal / squared-energy group
shrinkage, \code{"L1"}: exponential / sparser); \code{seed} (\code{123},
random-initialization seed); \code{a} (\code{1}) and
\code{b} (\code{(F + N)/K * mean(Y)}), the inverse-gamma prior (a smaller
\code{b} over-prunes, a larger one prunes nothing); \code{maxit}
(\code{3000}) and \code{epsilon} (\code{1e-6}) for optimisation control;
\code{tol} (\code{1e-3}), the relevance threshold below which a component is
counted as pruned.}
}
\value{
An object of class \code{"nmfkc.ard"}: a list with
\code{rank} (estimated = mode over restarts), \code{rank.runs} (the
per-run estimates), \code{relevance} (representative run, descending),
\code{lambda}, \code{W}, \code{H} (ordered by relevance),
\code{rank.init}, \code{prior}, \code{nrun} and \code{objfunc}.
}
\description{
\strong{Prototype} of Tan & Fevotte's (2013) ARD-NMF (Euclidean /
\eqn{\beta = 2}).  Unlike the cross-validation (\code{\link{nmfkc.ecv}},
\code{\link{nmfkc.bicv}}) and stability (\code{\link{nmfkc.consensus}})
engines that \emph{scan} a range of ranks, ARD fits NMF \strong{once} at
an over-complete rank and prunes automatically: each component \eqn{k}
carries a relevance weight \eqn{\lambda_k} with an inverse-gamma prior;
the multiplicative updates gain a penalty \eqn{+ w_{fk}/\lambda_k} (L2 /
half-normal prior) or \eqn{+ 1/\lambda_k} (L1 / exponential), which
drives unsupported components to zero.  The number of surviving
components is the estimated rank.  Covariates are ignored (plain NMF).

This is a model-based point estimate: the result depends on the prior, the
starting rank and the random initialization, and can vary run to run.  Use
it as a complement to the CV / consensus engines, not a sole criterion.
}
\details{
\strong{Relation to Tan & Fevotte (2013).}  The update equations reproduce
the paper's \eqn{\ell_2} / \eqn{\ell_1} ARD-NMF for the \strong{Euclidean
(\eqn{\beta = 2})} case exactly: the multiplicative penalties, the
closed-form \eqn{\lambda_k} update \eqn{(f(w_k) + f(h_k) + b)/c} and the
constant \eqn{c} (their Eq. 33).  Two deliberate simplifications keep it
practical: only \eqn{\beta = 2} is implemented (the paper covers the general
\eqn{\beta}-divergence), and the default \code{b} is an \emph{empirical}
per-component energy scale \eqn{(F + N)/K \cdot \bar{Y}} that reliably avoids
the winner-take-all collapse, rather than the paper's method-of-moments value
(their Eq. 38).
}
\examples{
\donttest{
set.seed(1)
X <- matrix(abs(rnorm(40 * 3)), 40, 3)
B <- matrix(abs(rnorm(3 * 60)), 3, 60)
ar <- nmfkc.ard(X \%*\% B, rank = 10)   # over-complete start
ar$rank                                # ~ 3 surviving components
plot(ar)
}
}
\references{
V. Y. F. Tan and C. Fevotte (2013).  Automatic relevance
determination in nonnegative matrix factorization with the
beta-divergence.  \emph{IEEE TPAMI} 35(7):1592--1605.
\doi{10.1109/TPAMI.2012.240}.
}
\seealso{
\code{\link{nmfkc.ecv}}, \code{\link{nmfkc.bicv}},
\code{\link{nmfkc.consensus}}, \code{\link{nmfkc.rank}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmf.ffb.R, R/nmf.sem.R
\name{nmf.ffb.split}
\alias{nmf.ffb.split}
\alias{nmf.sem.split}
\title{Heuristic Variable Splitting for NMF-FFB}
\usage{
nmf.ffb.split(x, ...)

nmf.sem.split(
  x,
  n.exogenous = NULL,
  threshold = 0.1,
  auto.flipped = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{x}{A numeric matrix or data frame with
\strong{rows = samples} and \strong{columns = observed variables}.}

\item{...}{Reserved for future use; currently unused (also accepted
by the \code{\link{nmf.ffb.split}} alias for argument forwarding).}

\item{n.exogenous}{Optional integer specifying the number of exogenous variables
(\eqn{Y_2}). If \code{NULL}, the number is inferred automatically by the
coefficient cut-off rule.}

\item{threshold}{Standardized regression-coefficient threshold used in the
automatic exogenous–endogenous split. A variable is treated as endogenous
once its maximum standardized parent coefficient exceeds this value.
(Default: \code{0.1})}

\item{auto.flipped}{Logical; if \code{TRUE}, applies PC1-based automatic
sign flipping after standardization to ensure consistent orientation.
(Default: \code{TRUE})}

\item{verbose}{Logical; if \code{TRUE}, prints progress messages and the
resulting variable split. (Default: \code{FALSE})}
}
\value{
A list with:
\item{endogenous.variables}{
Character vector of variables selected as endogenous (\eqn{Y_1}).}
\item{exogenous.variables}{
Character vector of variables selected as exogenous (\eqn{Y_2}).}
\item{ordered.variables}{
Variables in inferred causal order (from exogenous to endogenous).}
\item{is.flipped}{
Logical vector indicating which variables were sign-flipped during processing.}
\item{n.exogenous}{
Integer giving the number of exogenous variables.}
}
\description{
Infers a heuristic partition of observed variables into exogenous (\eqn{Y_2})
and endogenous (\eqn{Y_1}) blocks for use in NMF-FFB.
The method is based on positive-SEM logic, causal ordering, and optional
sign alignment using the first principal component (PC1).

The procedure:
\itemize{
\item internally standardizes variables (mean 0, sd 1),
\item optionally flips signs so that most variables align positively with PC1,
\item infers a causal ordering by repeatedly regressing each variable on the
remaining ones and selecting the variable with the largest minimum
standardized coefficient,
\item determines an exogenous block by scanning the ordering from upstream
and stopping at the first variable whose strongest parent coefficient
exceeds \code{threshold}.
}

If \code{n.exogenous} is supplied, it overrides the automatic threshold rule.
}
\examples{
# Infer exogenous/endogenous split from iris
sp <- nmf.sem.split(iris[, -5], n.exogenous = 2)
sp$endogenous.variables
sp$exogenous.variables

}
\seealso{
\code{\link{nmf.sem}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.kernel.gaussian}
\alias{nmfkc.kernel.gaussian}
\title{Create a Gaussian kernel matrix from covariates}
\source{
Mesquita, D., Gomes, J. P., & Rodrigues, L. R. (2019).
Gaussian kernels for incomplete data. \emph{Applied Soft Computing}, 77, 356--365.
}
\usage{
nmfkc.kernel.gaussian(
  U,
  V = NULL,
  beta = 0.5,
  na.method = c("pds", "egk"),
  ...
)
}
\arguments{
\item{U}{Covariate matrix \eqn{U(K,N) = (u_1, \dots, u_N)}. Each row may be normalized in advance.}

\item{V}{Covariate matrix \eqn{V(K,M) = (v_1, \dots, v_M)}, typically used for prediction. If \code{NULL}, the default is \code{U}. May contain \code{NA} values.}

\item{beta}{Bandwidth parameter for the Gaussian kernel. Default is \code{0.5}.}

\item{na.method}{Method for handling \code{NA} values in \code{V}. Either \code{"pds"} or \code{"egk"}. Ignored if \code{V} has no \code{NA}.}

\item{...}{Additional arguments for EGK method:
\describe{
\item{\code{gmm.G}}{Number of GMM components for EGK. Default is \code{3} (Mesquita et al., 2019).}
}}
}
\value{
Kernel matrix \eqn{A(N,M)}.
}
\description{
\code{nmfkc.kernel.gaussian} constructs a Gaussian (RBF) kernel matrix from covariate matrices.
The kernel is defined as \eqn{K(u,v) = \exp(-\beta \|u - v\|^2)}.
When \code{V} contains \code{NA} values, two methods are available via \code{na.method}:
\describe{
\item{\code{"pds"}}{Partial Distance Strategy. Computes the kernel using only observed (non-NA) rows,
with beta adjusted by \eqn{\beta_{adj} = \beta \times K / K_{obs}} where \eqn{K} is the total number of rows
and \eqn{K_{obs}} is the number of observed rows.}
\item{\code{"egk"}}{Expected Gaussian Kernel (Mesquita et al., 2019). Uses a Gaussian Mixture Model (GMM)
to estimate the conditional distribution of missing values given observed values,
then computes the expected kernel value via a Gamma approximation.
Requires \code{gmm.means}, \code{gmm.sigmas}, and \code{gmm.weights} passed through \code{...}.}
}
}
\examples{
U <- matrix(c(5,10,15,20,25),nrow=1)
V <- matrix(1:25,nrow=1)
A <- nmfkc.kernel.gaussian(U,V,beta=28/1000)
dim(A)

# PDS example: V with NA in first row
U2 <- matrix(rnorm(20), nrow=2)
V2 <- matrix(rnorm(10), nrow=2)
V2[1, c(2,4)] <- NA
A2 <- nmfkc.kernel.gaussian(U2, V2, beta=0.5, na.method="pds")
}
\seealso{
\code{\link{nmfkc.kernel}}, \code{\link{nmfkc.kernel.beta.cv}}, \code{\link{nmfkc.kernel.beta.nearest.med}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.ar.R
\name{nmfkc.ar.DOT}
\alias{nmfkc.ar.DOT}
\title{Generate a Graphviz DOT Diagram for NMF-AR / NMF-VAR Models}
\usage{
nmfkc.ar.DOT(
  result,
  degree = 1,
  intercept = any(colnames(result$C) == "(Intercept)"),
  threshold = 0.1,
  rankdir = "RL",
  fill = TRUE,
  weight_scale_xy = 5,
  weight_scale_lag = 5,
  weight_scale_int = 3,
  hide.isolated = TRUE
)
}
\arguments{
\item{result}{A fitted \code{nmfkc} object representing the AR model.
Must contain matrices \code{X} and \code{C}.}

\item{degree}{Maximum AR lag to visualize.}

\item{intercept}{Logical; if \code{TRUE}, draws intercept nodes for
columns named "(Intercept)" in matrix \code{C}.
Default is \code{TRUE} when an intercept column is detected in \code{C},
\code{FALSE} otherwise (auto-detected).}

\item{threshold}{Minimum coefficient magnitude required to draw an edge.}

\item{rankdir}{Graphviz rank direction (e.g., \code{"RL"}, \code{"LR"}, \code{"TB"}).}

\item{fill}{Logical; whether nodes are filled with color.}

\item{weight_scale_xy}{Scaling factor for edges \eqn{X \rightarrow T}.}

\item{weight_scale_lag}{Scaling factor for lagged edges \eqn{T-k \rightarrow X}.}

\item{weight_scale_int}{Scaling factor for intercept edges.}

\item{hide.isolated}{Logical. If \code{TRUE} (default), Y nodes that have no
edges at or above \code{threshold} are excluded from the graph.}
}
\value{
A character string representing a Graphviz DOT file.
}
\description{
Produces a Graphviz DOT script for visualizing autoregressive
NMF-with-covariates models constructed via \code{nmfkc.ar} + \code{nmfkc}.

The diagram displays three types of directed relationships:
\itemize{
\item Lagged predictors: \eqn{T_{t-k} \rightarrow X},
\item Current latent factors: \eqn{X \rightarrow T_t},
\item Optional intercept effects: \code{Const -> X}.
}

Importantly, \emph{no direct edges from lagged variables to current outputs}
(\eqn{T_{t-k} \rightarrow T_t}) are drawn, in accordance with the NMF-AR
formulation.

Each block of lagged variables is displayed in its own DOT subgraph
(e.g., “T-1”, “T-2”, ...), while latent factor nodes and current-time
outputs are arranged in separate clusters.
}
\examples{
d <- AirPassengers
ar_data <- nmfkc.ar(d, degree = 2)
result <- nmfkc(ar_data$Y, ar_data$A, rank = 1)
dot <- nmfkc.ar.DOT(result, degree = 2)
cat(dot)

}
\seealso{
\code{\link{nmfkc.ar}}, \code{\link{nmfkc}}, \code{\link{plot.nmfkc.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfre.R
\name{nmfre.inference}
\alias{nmfre.inference}
\title{Statistical inference for the coefficient matrix C from NMF-RE}
\usage{
nmfre.inference(object, Y, A = NULL, wild.bootstrap = TRUE, ...)
}
\arguments{
\item{object}{An object of class \code{"nmfre"} returned by
\code{\link{nmfre}}.}

\item{Y}{Observation matrix (P x N). Must match the data used in
\code{nmfre()}.}

\item{A}{Covariate matrix (K x N). Default is \code{NULL} (intercept only).}

\item{wild.bootstrap}{Logical. If \code{TRUE} (default), performs wild
bootstrap for confidence intervals and bootstrap standard errors.}

\item{...}{Additional arguments:
\describe{
\item{\code{wild.B}}{Number of bootstrap replicates. Default is 500.}
\item{\code{wild.seed}}{Seed for bootstrap. Default is 123.}
\item{\code{wild.level}}{Confidence level for bootstrap CI. Default is 0.95.}
\item{\code{C.p.side}}{P-value type: \code{"one.sided"} (default) or \code{"two.sided"}.}
\item{\code{cov.ridge}}{Ridge stabilization. Default is 1e-8.}
\item{\code{print.trace}}{Logical. Default is \code{FALSE}.}
}}
}
\value{
The input \code{object} with additional inference components:
\item{sigma2.used}{Estimated \eqn{\sigma^2} used for inference.}
\item{C.vec.cov}{Full covariance matrix for \eqn{vec(C)}.}
\item{C.se}{Sandwich standard errors for \eqn{C}.}
\item{C.se.boot}{Bootstrap standard errors for \eqn{C}.}
\item{C.ci.lower}{Lower CI bounds for \eqn{C}.}
\item{C.ci.upper}{Upper CI bounds for \eqn{C}.}
\item{coefficients}{Data frame with Basis, Covariate, Estimate, SE, BSE,
z_value, p_value, CI_low, CI_high.}
\item{C.p.side}{P-value type used.}
}
\description{
\code{nmfre.inference} performs statistical inference on the coefficient
matrix \eqn{C} (\eqn{\Theta}) from a fitted \code{nmfre} model,
conditional on the estimated basis matrix \eqn{\hat{X}} and random
effects \eqn{\hat{U}}.

Under the working model \eqn{Y^* = Y - X\hat{U} \approx X C A + \varepsilon},
inference is conducted via sandwich covariance estimation and
one-step wild bootstrap with non-negative projection.

The result is compatible with \code{\link{nmfkc.DOT}} for visualization
(pass the result directly as \code{x} with \code{type = "YXA"}).
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(intercept = 1, speed = cars$speed)
res <- nmfre(Y, A, rank = 1, wild.bootstrap = FALSE)
res2 <- nmfre.inference(res, Y, A)
res2$coefficients

}
\references{
Satoh, K. (2026). Wild Bootstrap Inference for Non-Negative Matrix
Factorization with Random Effects. arXiv:2603.01468.
\url{https://arxiv.org/abs/2603.01468}
}
\seealso{
\code{\link{nmfre}}, \code{\link{nmfkc.DOT}},
\code{\link{summary.nmfre}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{nmfae.signed.heatmap}
\alias{nmfae.signed.heatmap}
\title{Heatmap visualization of nmfae.signed factor matrices}
\usage{
nmfae.signed.heatmap(
  x,
  Y1.label = NULL,
  X1.label = NULL,
  X2.label = NULL,
  Y2.label = NULL,
  palette.pos = NULL,
  palette.signed = NULL,
  show.C = TRUE,
  ...
)
}
\arguments{
\item{x}{An object of class \code{"nmfae.signed"}.}

\item{Y1.label}{Character vector for rows of \eqn{X_1}.}

\item{X1.label}{Decoder basis labels.}

\item{X2.label}{Encoder basis labels.}

\item{Y2.label}{Input variable labels.}

\item{palette.pos}{Palette for non-negative blocks. Default white-orange-red.}

\item{palette.signed}{Palette for signed \eqn{C}. Default blue-white-red.}

\item{show.C}{Logical. If \code{TRUE} (default), shows the combined
signed \eqn{C = C_{+} - C_{-}} as a separate panel.}

\item{...}{Not used.}
}
\value{
Invisible \code{NULL}. Called for its side effect (plot).
}
\description{
Displays the factor blocks of a \code{\link{nmfae.signed}} fit as
side-by-side heatmaps.  Non-negative blocks (\eqn{X_1, C_{+}, C_{-},
X_2}) use the white-orange-red palette; the signed combined
bottleneck \eqn{C = C_{+} - C_{-}} is rendered with a diverging
blue-white-red palette so positive and negative weights are visually
distinguishable.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\examples{
\donttest{
set.seed(1)
Y1 <- matrix(abs(rnorm(12)), 3, 4)
Y2 <- matrix(abs(rnorm(20)), 5, 4)
res <- nmfae.signed(Y1, Y2, rank = 2, rank.encoder = 2, maxit = 200)
nmfae.signed.heatmap(res)
}
}
\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae.signed}}, \code{\link{nmfae.heatmap}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.signed.R
\name{nmfkc.signed.rank}
\alias{nmfkc.signed.rank}
\title{Rank selection for nmfkc.signed (concise diagnostics)}
\usage{
nmfkc.signed.rank(
  Y,
  A,
  rank = 1:5,
  detail = c("full", "fast"),
  plot = TRUE,
  ...
)
}
\arguments{
\item{Y}{Observation matrix (may contain negative entries).}

\item{A}{Covariate matrix (may be signed).}

\item{rank}{Integer vector of ranks to evaluate.}

\item{detail}{\code{"full"} (default) also runs element-wise CV
(\code{sigma.ecv}); \code{"fast"} skips it (plots r.squared and
eff.rank only, and recommends the R-squared elbow).}

\item{plot}{Logical; draw the diagnostics plot (default \code{TRUE}).}

\item{...}{Passed on to \code{\link{nmfkc.signed}} and
\code{nmfkc.signed.ecv} (e.g.\ \code{maxit}, \code{nfolds},
\code{seed}).}
}
\value{
A list with \code{rank.best} and \code{criteria}
(\code{rank}, \code{effective.rank}, \code{effective.rank.ratio},
\code{r.squared}, \code{sigma.ecv}).
}
\description{
Fits \code{\link{nmfkc.signed}} across a range of ranks and reports
\code{r.squared}, the effective rank, and the element-wise CV error
\code{sigma.ecv}, with the same concise plot as
\code{\link{nmfkc.rank}}.
}
\references{
Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
effective dimensionality.  \emph{Proc. EUSIPCO}, 606--610.
(\code{effective.rank})
Wold, S. (1978).  Cross-validatory estimation of the number of
components in factor and principal components models.
\emph{Technometrics}, 20(4), 397--405. (\code{sigma.ecv})
}
\seealso{
\code{\link{nmfkc.signed}}, \code{\link{nmfkc.rank}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.kernel.beta.nearest.med}
\alias{nmfkc.kernel.beta.nearest.med}
\title{Estimate Gaussian/RBF kernel parameter beta from covariates (supports landmarks)}
\usage{
nmfkc.kernel.beta.nearest.med(
  U,
  Uk = NULL,
  block.size = 1000,
  block.size.Uk = 2000,
  sample.size = NULL,
  ...
)
}
\arguments{
\item{U}{A numeric matrix of covariates (\code{K x N}); columns are samples.}

\item{Uk}{An optional numeric matrix of landmarks (\code{K x M}); columns are landmark points.
If provided, distances are computed from samples in \code{U} to landmarks in \code{Uk}.}

\item{block.size}{Integer. Number of columns of \code{U} processed per block when computing
distances (controls memory usage). If \code{N <= 1000}, it is automatically set to \code{N}.}

\item{block.size.Uk}{Integer. Number of columns of \code{Uk} processed per block when \code{Uk}
is not \code{NULL} (controls memory usage). If \code{M <= 2000}, it is automatically set to \code{M}.}

\item{sample.size}{Integer or \code{NULL}. If not \code{NULL}, randomly subsamples this many columns
of \code{U} (without replacement) before computing distances, to reduce computational cost.}

\item{...}{Additional arguments.  Hidden option \code{candidates} controls
the candidate grid: one of \code{"7points"} (default), \code{"4points"},
or a numeric vector of \eqn{t} values.  See Details.}
}
\value{
A list with elements:
\itemize{
\item \code{beta}: Estimated kernel parameter \eqn{\beta_0 = 1/(2\sigma_0^2)}.
\item \code{beta_candidates}: Numeric vector of candidate \eqn{\beta} values (logarithmic grid)
intended for cross-validation.
\item \code{dist_median}: The estimated distance scale \eqn{\sigma_0} (median of nearest-neighbor
or nearest-landmark distances).
\item \code{block.size.used}: The effective block size(s) used. Either a scalar (no \code{Uk}) or
a named vector \code{c(U=..., Uk=...)} when \code{Uk} is provided.
\item \code{sample.size.used}: The number of columns of \code{U} actually used (after subsampling).
\item \code{uk_is_u}: Logical flag indicating whether \code{Uk} was detected as identical to \code{U}
(only returned when \code{Uk} is provided).
}
}
\description{
Computes a data-driven reference scale for the Gaussian/RBF kernel from covariates
using a robust "median nearest-neighbor (or nearest-landmark) distance" heuristic,
and returns the corresponding kernel parameter \eqn{\beta}.

The Gaussian/RBF kernel is assumed to be written in the form
\deqn{k(u,v) = \exp\{-\beta \|u-v\|^2\} = \exp\{-\|u-v\|^2/(2\sigma^2)\},}
hence \eqn{\beta = 1/(2\sigma^2)}. This function first estimates a typical distance
scale \eqn{\sigma_0} by the median of distances, then sets \eqn{\beta_0 = 1/(2\sigma_0^2)}.

If \code{Uk} is \code{NULL}, \eqn{\sigma_0} is estimated as the median of
nearest-neighbor distances within \code{U} (excluding self-distance).
If \code{Uk} is provided, \eqn{\sigma_0} is estimated as the median of
nearest-landmark distances from each sample in \code{U} to its closest landmark in \code{Uk}.

To control memory usage for large \code{N} (and \code{M}), distances are computed in blocks.
Optionally, columns of \code{U} can be randomly subsampled via \code{sample.size} to reduce cost.
}
\details{
\strong{Candidate grid:}
Along with \code{beta}, the function returns \code{beta_candidates}, a
logarithmic grid suitable for cross-validation.  The grid is symmetric on
the bandwidth scale \eqn{\sigma} around \eqn{\sigma_0}:
\deqn{\sigma = \sigma_0 \times 10^{t},}
and since \eqn{\beta = 1/(2\sigma^2)}, this corresponds to
\eqn{\beta = \beta_0 \times 10^{-2t}}.

The grid of \eqn{t} values can be customized through the hidden argument
\code{candidates} (passed via \code{...}):
\itemize{
\item \code{"7points"} (default): \eqn{t \in \{-1,-2/3,-1/3,0,1/3,2/3,1\}}
(7 candidates spanning one decade, matches the grid used in the
RFF-NMF research memo).
\item \code{"4points"}: \eqn{t \in \{-1/2, 0, 1/2, 1\}} yielding
\eqn{\beta_0 \times 10^{(1,0,-1,-2)}} (the legacy short grid).
\item A numeric vector: user-specified \eqn{t} values.  The grid returned
is \eqn{\beta_0 \times 10^{-2t}}.
}

Prior to version 0.6.8, the grid depended on whether \code{Uk} was
supplied (4 candidates for \code{Uk = NULL}, 7 for supplied \code{Uk}).
The current implementation unifies both branches via \code{candidates}.

\strong{Notes:}
\itemize{
\item When \code{Uk} is identical to \code{U}, the function detects this case and excludes
self-distances (distance 0) to avoid \eqn{\sigma_0=0}.
\item \code{sample.size} performs random subsampling without setting a seed. For reproducible
results, set \code{set.seed()} before calling this function.
}
}
\examples{
# Basic (nearest-neighbor within U)
U <- matrix(runif(20), nrow = 2)
beta_info <- nmfkc.kernel.beta.nearest.med(U)
beta0 <- beta_info$beta
betas <- beta_info$beta_candidates

# With landmarks (nearest-landmark distances)
Uk <- matrix(runif(10), nrow = 2)
\donttest{
beta_info2 <- nmfkc.kernel.beta.nearest.med(U, Uk)
}

}
\seealso{
\code{\link{nmfkc.kernel.gaussian}}, \code{\link{nmfkc.kernel.beta.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.rank}
\alias{nmfkc.rank}
\title{Rank selection diagnostics with graphical output}
\usage{
nmfkc.rank(Y, A = NULL, rank = 1:2, detail = "full", plot = TRUE, data, ...)
}
\arguments{
\item{Y}{Observation matrix, or a formula (see \code{\link{nmfkc}} for Formula Mode).}

\item{A}{Covariate matrix. If \code{NULL}, the identity matrix is used.
Ignored when \code{Y} is a formula.}

\item{rank}{A vector of candidate ranks to be evaluated.}

\item{detail}{\code{"full"} (default) also runs the element-wise CV
(\code{sigma.ecv}); \code{"fast"} skips it (the plot then shows only
r.squared and eff.rank, and the recommended rank falls back to the
R-squared elbow).}

\item{plot}{Logical. If \code{TRUE} (default), draws a plot of the diagnostic criteria.}

\item{data}{A data frame (required when \code{Y} is a formula with column names).}

\item{...}{Additional arguments passed to \code{\link{nmfkc}} and \code{\link{nmfkc.ecv}}.
\itemize{
\item \code{Q}: (Deprecated) Alias for \code{rank}.
\item \code{save.time}: (Deprecated) \code{TRUE} maps to \code{detail = "fast"}.
}}
}
\value{
A list containing:
\item{rank.best}{The estimated optimal rank. Prioritizes ECV minimum if available, otherwise R-squared Elbow.}
\item{criteria}{A data frame containing diagnostic metrics for each rank.
The \code{effective.rank} column gives the effective rank
(\eqn{\exp} of the Shannon entropy of the explained-variance
distribution \eqn{p_k = \mathrm{var}(B_{k\cdot}) / \sum_j \mathrm{var}(B_{j\cdot})},
in \eqn{[1, Q]}); when it plateaus well below the nominal
\code{rank}, the extra factors are not carrying additional
coefficient variance, which suggests an over-specified rank.
The \code{effective.rank.ratio} column is \code{effective.rank /
  rank} in \eqn{[0, 1]} (the utilization fraction plotted as
\code{eff.rank} when \code{plot = TRUE}); a peak marks the rank
at which the latent factors carry the most evenly distributed
variance.}
}
\description{
\code{nmfkc.rank} provides diagnostic criteria for selecting the rank (\eqn{Q})
in NMF with kernel covariates. Three rank-selection measures are computed
(R-squared, the effective rank, and the element-wise CV error), and results
can be visualized in a plot. Sample-clustering quality (silhouette / CPCC /
dist.cor) is no longer part of rank selection; use \code{\link{nmf.cluster.criteria}}
on a fitted model for those.

By default (\code{save.time = FALSE}), this function also computes the
Element-wise Cross-Validation error (Wold's CV Sigma) using \code{\link{nmfkc.ecv}}.

The plot explicitly marks the "BEST" rank based on two criteria:
\enumerate{
\item \strong{Elbow Method (Red)}: Based on the curvature of the R-squared values (always computed if Q > 2).
\item \strong{Min RMSE (Blue)}: Based on the minimum Element-wise CV Sigma (only if \code{detail="full"}).
}
}
\examples{
# Example.
Y <- t(iris[,-5])
# Full run (default)
nmfkc.rank(Y, rank=1:4)
# Fast run (skip ECV)
nmfkc.rank(Y, rank=1:4, detail="fast")
}
\references{
Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
effective dimensionality.  \emph{Proc. 15th European Signal Processing
Conf. (EUSIPCO)}, 606--610.  (\code{effective.rank})
Wold, S. (1978).  Cross-validatory estimation of the number of
components in factor and principal components models.
\emph{Technometrics}, 20(4), 397--405.
\doi{10.1080/00401706.1978.10489693}  (\code{sigma.ecv})
}
\seealso{
\code{\link{nmfkc}},
\code{\link{nmfkc.ecv}} (element-wise CV, used internally),
\code{\link{nmfkc.bicv}} (block bi-cross-validation),
\code{\link{nmfkc.consensus}} (stability) and
\code{\link{nmfkc.ard}} (Bayesian ARD) for alternative rank criteria.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.consensus.R
\name{plot.nmfkc.consensus}
\alias{plot.nmfkc.consensus}
\title{Plot a consensus rank-selection (nmfkc.consensus) object}
\usage{
\method{plot}{nmfkc.consensus}(
  x,
  type = c("criteria", "heatmap"),
  rank = NULL,
  mfrow = NULL,
  mar = NULL,
  col = grDevices::hcl.colors(50, "Blue-Red"),
  main = NULL,
  ...
)
}
\arguments{
\item{x}{An object of class \code{"nmfkc.consensus"}.}

\item{type}{\code{"criteria"} or \code{"heatmap"}.}

\item{rank}{For \code{type = "heatmap"}, the rank(s) to display.
\code{NULL} (default) shows every rank stored in \code{x}.}

\item{mfrow}{Panel layout for multiple heatmaps, as
\code{c(nrow, ncol)}.  \code{NULL} (default) uses
\code{\link[grDevices]{n2mfrow}} for a near-square grid.}

\item{mar}{Per-panel margins \code{c(b, l, t, r)} for the heatmap
view.  \code{NULL} (default) uses tight margins for a grid and wider
margins (for sample labels) when a single rank is shown.}

\item{col}{Heatmap colour palette (length-50 blue-to-red by default).}

\item{main}{Plot title for the \code{"criteria"} view (heatmap panels
are titled per rank).}

\item{...}{Further arguments passed to the underlying
\code{\link[graphics]{plot}} / \code{\link[graphics]{image}}.}
}
\value{
\code{x}, invisibly.
}
\description{
Two views of a \code{\link{nmfkc.consensus}} result:
\itemize{
\item \code{type = "criteria"} (default): the stability curves
\code{cophenetic} (blue) and \code{dispersion} (red) against rank.
No "best" marker is drawn -- a stable rank tends to be a coarse one,
so the curves are shown for inspection rather than as an optimum.
\item \code{type = "heatmap"}: the consensus matrix of each requested
\code{rank}, reordered by average-linkage hierarchical clustering of
\eqn{1 - \bar C} (blue = 0 / different cluster, red = 1 / same
cluster).  Requires \code{keep.consensus = TRUE} at compute time.
}
}
\seealso{
\code{\link{nmfkc.consensus}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{print.summary.nmfkc.net.inference}
\alias{print.summary.nmfkc.net.inference}
\title{Print method for summary.nmfkc.net.inference objects}
\usage{
\method{print}{summary.nmfkc.net.inference}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{An object of class \code{"summary.nmfkc.net.inference"}.}

\item{digits}{Minimum number of significant digits.}

\item{...}{Additional arguments (currently unused).}
}
\value{
Called for its side effect (printing). Returns \code{x} invisibly.
}
\description{
Prints a formatted summary of a symmetric NMF model with inference results.
The coefficients table uses \code{Basis.row} and \code{Basis.col} labels
reflecting the symmetric structure \eqn{C_{qr}}.
}
\seealso{
\code{\link{summary.nmfkc.net.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{nmfae.cv}
\alias{nmfae.cv}
\title{Sample-wise k-fold Cross-Validation for nmfae}
\usage{
nmfae.cv(Y1, Y2 = Y1, rank = 2, rank.encoder = rank, ...)
}
\arguments{
\item{Y1}{Output matrix \eqn{Y_1} (P1 x N). Non-negative.}

\item{Y2}{Input matrix \eqn{Y_2} (P2 x N), or a kernel matrix (N x N).
Default is \code{Y1} (autoencoder).}

\item{rank}{Integer. Rank of the decoder basis. Default is 2.}

\item{rank.encoder}{Integer. Rank of the encoder basis. Default is \code{rank}.}

\item{...}{Additional arguments passed to \code{\link{nmfae}}
(e.g., \code{epsilon}, \code{maxit}, \code{Y1.weights}).
Also accepts: \code{nfolds} (number of folds, default 5; \code{div} also accepted),
\code{seed} (integer seed, default 123), \code{shuffle} (logical, default \code{TRUE}).
For backward compatibility, \code{Q}, \code{R} are accepted as aliases for
\code{rank}, \code{rank.encoder}.}
}
\value{
A list with components:
\item{objfunc}{Mean squared error per valid element over all folds.}
\item{sigma}{Residual standard error (RMSE), same scale as \eqn{Y_1}.}
\item{objfunc.block}{Per-fold squared error totals.}
\item{block}{Integer vector of fold assignments (1, ..., \code{div}) for each column.}
}
\description{
\code{nmfae.cv} performs k-fold cross-validation by splitting columns (samples)
of \eqn{Y_1} and \eqn{Y_2} into \code{div} folds. For each fold, the model
\eqn{Y_1 \approx X_1 \Theta X_2 Y_2} is fitted on the training samples and
predictive performance is evaluated on the held-out samples.

When \code{Y2} is a kernel matrix created by \code{\link{nmfkc.kernel}}
(detected via attributes), the symmetric kernel splitting convention is used:
\code{Y2[train, train]} for training and \code{Y2[train, test]} for prediction.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\examples{
Y <- t(iris[1:30, 1:4])
res <- nmfae.cv(Y, rank = 2, rank.encoder = 2, nfolds = 5, maxit = 500)
res$sigma

}
\seealso{
\code{\link{nmfae}}, \code{\link{nmfae.ecv}}, \code{\link{nmfae.kernel.beta.cv}},
\code{\link{nmfkc.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{summary.nmfkc.net.inference}
\alias{summary.nmfkc.net.inference}
\title{Summary method for nmfkc.net.inference objects}
\usage{
\method{summary}{nmfkc.net.inference}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"nmfkc.net.inference"}.}

\item{...}{Additional arguments passed to \code{summary.nmfkc}.}
}
\value{
An object of class \code{"summary.nmfkc.net.inference"}.
}
\description{
Produces a summary of a symmetric NMF model with inference results,
including the coefficients table for \eqn{C}.
}
\seealso{
\code{\link{nmfkc.net.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.kernel}
\alias{nmfkc.kernel}
\title{Create a kernel matrix from covariates}
\source{
Satoh, K. (2024). Applying Non-negative Matrix Factorization with Covariates to the Longitudinal Data as Growth Curve Model.
arXiv preprint arXiv:2403.05359. \url{https://arxiv.org/abs/2403.05359}
}
\usage{
nmfkc.kernel(
  U,
  V = NULL,
  kernel = c("Gaussian", "Exponential", "Periodic", "Linear", "NormalizedLinear",
    "Polynomial"),
  ...
)
}
\arguments{
\item{U}{Covariate matrix \eqn{U(K,N) = (u_1, \dots, u_N)}. Each row may be normalized in advance.}

\item{V}{Covariate matrix \eqn{V(K,M) = (v_1, \dots, v_M)}, typically used for prediction. If \code{NULL}, the default is \code{U}.}

\item{kernel}{Kernel function to use. Default is \code{"Gaussian"}. Options are \code{"Gaussian"}, \code{"Exponential"}, \code{"Periodic"}, \code{"Linear"}, \code{"NormalizedLinear"}, and \code{"Polynomial"}.}

\item{...}{Additional arguments passed to the specific kernel function (e.g., \code{beta}, \code{degree}).}
}
\value{
Kernel matrix \eqn{A(N,M)}.
}
\description{
\code{nmfkc.kernel} constructs a kernel matrix from covariate matrices.
It supports Gaussian, Exponential, Periodic, Linear, Normalized Linear, and Polynomial kernels.
}
\examples{
# Example.
Y <- matrix(cars$dist,nrow=1)
U <- matrix(c(5,10,15,20,25),nrow=1)
V <- matrix(cars$speed,nrow=1)
A <- nmfkc.kernel(U,V,beta=28/1000)
dim(A)
result <- nmfkc(Y,A,rank=1)
plot(as.vector(V),as.vector(Y))
lines(as.vector(V),as.vector(result$XB),col=2,lwd=2)
}
\seealso{
\code{\link{nmfkc.kernel.gaussian}}, \code{\link{nmfkc.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{plot.nmfae.signed}
\alias{plot.nmfae.signed}
\title{Plot method for nmfae.signed (convergence)}
\usage{
\method{plot}{nmfae.signed}(x, ...)
}
\arguments{
\item{x}{An \code{nmfae.signed} object.}

\item{...}{Additional graphical parameters.}
}
\value{
Invisible \code{NULL}.
}
\description{
Displays the convergence trajectory of the objective function.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae.signed}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.ar.R
\name{nmfkc.ar.degree.cv}
\alias{nmfkc.ar.degree.cv}
\title{Optimize lag order for the autoregressive model}
\usage{
nmfkc.ar.degree.cv(
  Y,
  rank = 1,
  degree = 1:2,
  intercept = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{Y}{Observation matrix \eqn{Y(P,N)} or a \code{ts} object.}

\item{rank}{Rank of the basis matrix. For backward compatibility,
\code{Q} is accepted via \code{...}.}

\item{degree}{A vector of candidate lag orders to be evaluated.}

\item{intercept}{Logical. If TRUE (default), an intercept is added to the covariate matrix.}

\item{plot}{Logical. If TRUE (default), a plot of the objective function values is drawn.}

\item{...}{Additional arguments passed to \code{nmfkc.cv}.}
}
\value{
A list with components:
\item{degree}{The lag order that minimizes the cross-validation objective function.}
\item{degree.max}{Maximum recommended lag order, computed as \eqn{10 \log_{10}(N)}
following the \code{ar} function in the \pkg{stats} package.}
\item{objfunc}{Objective function values for each candidate lag order.}
}
\description{
\code{nmfkc.ar.degree.cv} selects the optimal lag order for an autoregressive model
by applying cross-validation over candidate degrees.

This function accepts both standard matrices (Variables x Time) and \code{ts} objects
(Time x Variables). \code{ts} objects are automatically transposed internally.
}
\examples{
# Example using ts object directly
d <- AirPassengers

# Selection of degree (using ts object)
# Note: Y is automatically transposed if it is a ts object
nmfkc.ar.degree.cv(Y=d, rank=1, degree=11:14)
}
\seealso{
\code{\link{nmfkc.ar}}, \code{\link{nmfkc.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{predict.nmfae}
\alias{predict.nmfae}
\title{Predict method for nmfae objects}
\usage{
\method{predict}{nmfae}(object, newY2 = NULL, Y1 = NULL, type = c("response", "class"), ...)
}
\arguments{
\item{object}{An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.}

\item{newY2}{Optional new input matrix (P2 x M) for prediction.
If \code{NULL}, returns in-sample fitted values.}

\item{Y1}{Optional actual output matrix for comparison plotting.}

\item{type}{Character. \code{"response"} (default) returns the predicted matrix.
\code{"class"} returns a factor of predicted class labels (row with max value).}

\item{...}{Not used.}
}
\value{
For \code{type = "response"}: a matrix of class \code{"predict.nmfae"}.
For \code{type = "class"}: a factor of class \code{"predict.nmfae"} with
predicted class labels. If \code{Y1} was provided, actual classes are stored
in \code{attr(result, "actual")}.
}
\description{
\code{predict.nmfae} computes fitted or predicted values from a three-layer NMF model.
Without \code{newY2}, returns the in-sample fitted values \eqn{X_1 \Theta X_2 Y_2}.
With \code{newY2}, computes out-of-sample predictions \eqn{X_1 \Theta X_2 \cdot \mathrm{newY2}}.

When \code{type = "class"}, each column is classified to the row with the
maximum predicted value (useful when \eqn{Y_1} is a one-hot class matrix
from \code{\link{nmfkc.class}}).

If \code{Y1} (actual values) is provided, it is stored as an attribute so that
\code{plot.predict.nmfae} can produce an observed-vs-predicted scatter plot
(for \code{type = "response"}) or a confusion matrix heatmap
(for \code{type = "class"}).
}
\examples{
\donttest{
set.seed(1)
Y <- matrix(runif(20), nrow = 4)
res <- nmfae(Y, rank = 2)
pred <- predict(res)
}
}
\seealso{
\code{\link{nmfae}}, \code{\link{plot.predict.nmfae}},
\code{\link{nmfkc.class}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{nmfae.inference}
\alias{nmfae.inference}
\title{Statistical Inference for NMF-AE Parameter Matrix}
\usage{
nmfae.inference(object, Y1, Y2 = Y1, wild.bootstrap = TRUE, ...)
}
\arguments{
\item{object}{An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.}

\item{Y1}{Output matrix \eqn{Y_1} (P1 x N). Must match the data used in \code{nmfae()}.}

\item{Y2}{Input matrix \eqn{Y_2} (P2 x N). Default is \code{Y1} (autoencoder).}

\item{wild.bootstrap}{Logical. If \code{TRUE} (default), performs wild bootstrap
for bootstrap SE and confidence intervals. If \code{FALSE}, only sandwich SE
and z-test p-values are computed (faster).}

\item{...}{Additional arguments:
\describe{
\item{\code{wild.B}}{Number of bootstrap replicates. Default is 1000.}
\item{\code{wild.seed}}{Seed for bootstrap. Default is 42.}
\item{\code{wild.level}}{Confidence level for bootstrap CI. Default is 0.95.}
\item{\code{sandwich}}{Logical. Use sandwich covariance. Default is \code{TRUE}.}
\item{\code{C.p.side}}{P-value type: \code{"one.sided"} (default) or \code{"two.sided"}.}
\item{\code{cov.ridge}}{Ridge stabilization for information matrix inversion. Default is 1e-8.}
\item{\code{print.trace}}{Logical. If \code{TRUE}, prints progress. Default is \code{FALSE}.}
}}
}
\value{
An object of class \code{c("nmfae.inference", "nmfae")}, inheriting all
components from the input \code{object}, with additional inference components:
\item{sigma2.used}{Estimated \eqn{\sigma^2} used for inference.}
\item{C.se}{Sandwich standard errors for \eqn{\Theta} (Q x R matrix).}
\item{C.se.boot}{Bootstrap standard errors for \eqn{\Theta} (Q x R matrix).}
\item{C.ci.lower}{Lower CI bounds for \eqn{\Theta} (Q x R matrix).}
\item{C.ci.upper}{Upper CI bounds for \eqn{\Theta} (Q x R matrix).}
\item{coefficients}{Data frame with Estimate, SE, BSE, z, p-value for each element of \eqn{\Theta}.}
\item{C.p.side}{P-value type used.}
}
\description{
Performs post-estimation inference for \eqn{\Theta} in the three-layer NMF model
\eqn{Y_1 \approx X_1 \Theta X_2 Y_2}, conditional on \eqn{(\hat{X}_1, \hat{X}_2)}.
Uses sandwich covariance estimation and one-step wild bootstrap with
non-negative projection.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\examples{
Y <- matrix(c(1,0,1,0, 0,1,0,1, 1,1,0,0), nrow=3, byrow=TRUE)
res <- nmfae(Y, rank=2, rank.encoder=2)
res2 <- nmfae.inference(res, Y)
summary(res2)

}
\seealso{
\code{\link{nmfae}}, \code{\link{summary.nmfae.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmf.ffb.R, R/nmf.sem.R
\name{nmf.ffb.inference}
\alias{nmf.ffb.inference}
\alias{nmf.sem.inference}
\title{Statistical inference for NMF-FFB via X-fixed full pair bootstrap}
\usage{
nmf.ffb.inference(
  object,
  Y1,
  Y2,
  B = 1000L,
  threshold = 0.01,
  ci.level = 0.95,
  C1.L1 = 1,
  C2.L1 = 0.1,
  seed = 123L,
  ...
)

nmf.sem.inference(
  object,
  Y1,
  Y2,
  B = 1000L,
  threshold = 0.01,
  ci.level = 0.95,
  C1.L1 = 1,
  C2.L1 = 0.1,
  seed = 123L,
  ...
)
}
\arguments{
\item{object}{A fitted object returned by \code{\link{nmf.sem}}.  Must
contain \code{X}, \code{C1}, \code{C2}.}

\item{Y1}{Endogenous variable matrix (P1 x N).  Must match the data
used in \code{nmf.sem()}.}

\item{Y2}{Exogenous variable matrix (P2 x N).  Same.}

\item{B}{Number of bootstrap replicates.  Default \code{1000}; required
for the \code{***} threshold (sup > 0.999).  Reduce to 500 for
exploratory speed (only \code{*} / \code{**} stay reliable).}

\item{threshold}{Display threshold \eqn{\delta} for the support rate
\eqn{\Pr_{\mathrm{boot}}(\hat c^{(b)} > \delta)}.  Default
\code{0.01}; entries below this magnitude are treated as effectively
zero in the path diagram.}

\item{ci.level}{Confidence level for the percentile bootstrap CI.
Default \code{0.95}.}

\item{C1.L1, C2.L1}{L1 sparsity penalties used by the original
\code{\link{nmf.sem}} fit.  These must match the fit's hyperparameters
for the bootstrap to estimate the correct model.  Defaults
(\code{1.0}, \code{0.1}) match \code{nmf.sem}'s defaults but you
should pass the actual values used.}

\item{seed}{Base RNG seed for the bootstrap.  Each replicate uses
\code{seed + b} (resampling) and \code{seed + 1000 + b}
(\eqn{C_1, C_2} initialization).  Default \code{123}.}

\item{...}{Hidden options:
\describe{
\item{\code{epsilon}}{Convergence tolerance for the inner fixed-X MU
loop.  Default \code{1e-6}.}
\item{\code{maxit}}{Maximum iterations for the inner MU loop.
Default \code{5000}.}
\item{\code{ncores}}{Number of parallel workers.  Default \code{1}
(serial).  Cross-platform: uses \code{parallel::mclapply} on
Linux/macOS and \code{parallel::parLapply} (PSOCK cluster) on
Windows.}
\item{\code{print.trace}}{Logical, print progress.  Default
\code{FALSE}.}
}}
}
\value{
The input \code{object} with additional bootstrap inference
components:
\item{coefficients}{Data frame with rows for every entry of \eqn{C_1}
and \eqn{C_2} and columns \code{Type} ("C1" / "C2"), \code{Basis},
\code{Covariate}, \code{Estimate}, \code{CI_low}, \code{CI_high},
\code{support_rate}, \code{p_value} (\eqn{= 1 - \mathrm{support\_rate}},
for compatibility with downstream consumers such as
\code{\link{nmf.sem.DOT}}), and \code{sig}.}
\item{C1.support.rate, C2.support.rate}{Per-element support rates
(Q x P1 and Q x P2 matrices).}
\item{C1.ci.lower, C1.ci.upper, C2.ci.lower, C2.ci.upper}{Per-element
percentile CI bounds.}
\item{C1.array, C2.array}{Bootstrap distributions: 3D arrays of shape
B x Q x P1 (and B x Q x P2).  Invalid replicates contain \code{NA}.}
\item{rho.boot, AR.boot, iter.boot}{Per-replicate spectral radius,
amplification ratio, and inner-loop iteration count.}
\item{bootstrap.B, bootstrap.threshold, bootstrap.ci.level}{Inputs
recorded for reproducibility.}
\item{bootstrap.n.valid, bootstrap.n.invalid}{Validity counts.}
}
\description{
\code{nmf.sem.inference} performs statistical inference on the structural
coefficient matrices \eqn{C_1} (latent feedback, \eqn{\Theta_1}) and
\eqn{C_2} (exogenous loading, \eqn{\Theta_2}) from a fitted
\code{\link{nmf.sem}} model.

The procedure is a \strong{full pair bootstrap} that holds the basis
matrix \eqn{\hat X} from the original fit fixed across all replicates
(which avoids label switching and gives a clean conditional
interpretation: ``uncertainty of the structural coefficients given the
measurement model''):
\enumerate{
\item For each replicate \eqn{b = 1, \dots, B}, resample column indices
\eqn{(i_1, \dots, i_N)} with replacement from \eqn{\{1, \dots, N\}}
and form \eqn{Y_1^{(b)} = Y_1[, i]}, \eqn{Y_2^{(b)} = Y_2[, i]}.
\item Re-estimate \eqn{(C_1^{(b)}, C_2^{(b)})} by running the
\code{\link{nmf.sem}} multiplicative updates \emph{with \eqn{X = \hat X}
held fixed} (no \eqn{X} update; no centroid sort), using the same
\code{C1.L1}, \code{C2.L1} as the original fit.
\item Discard replicates that violate stationarity
(\eqn{\rho(X C_1^{(b)}) \ge 1}) or have an amplification ratio
exceeding the geometric-series bound by more than 1\\%.
}
Because \eqn{C_1, C_2 \ge 0} are non-negative by construction, exact
zeros are essentially never observed in the bootstrap distribution.
Significance is assessed via a \strong{support rate} at a small display
threshold \eqn{\delta} (default \code{0.01}):
\deqn{
  \mathrm{sup}(c) \;=\; \frac{1}{|\mathrm{valid}|}
    \sum_{b \in \mathrm{valid}} \mathbf{1}\!\left( \hat c^{(b)} > \delta \right).
}
This is a one-sided counterpart of the classical \eqn{p}-value:
large \code{support_rate} indicates strong evidence that the entry is
meaningfully positive.  Significance markers follow the lavaan
convention with the natural correspondence \eqn{p = 1 - \mathrm{sup}}:
\code{*} (sup > 0.95), \code{**} (sup > 0.99), \code{***}
(sup > 0.999).  Cutoffs use strict greater-than so the rule
mirrors the standard R convention for p-values (p < 0.05 / 0.01 /
0.001 → \emph{/\strong{/}}), translated to support_rate via
\eqn{\mathrm{sup} = 1 - p}.
}
\section{Lifecycle}{

This function's interface changed at v0.6.8: the legacy 1-step Newton
wild bootstrap (with sandwich SE) has been replaced by the full pair
bootstrap described above, following the paper revision.  The fields
\code{sigma2.used}, \code{C2.se}, \code{C2.se.boot}, \code{C2.p.side}
that the previous implementation produced are no longer present.
}

\examples{
\donttest{
Y <- t(iris[, -5])
Y1 <- Y[1:2, ]; Y2 <- Y[3:4, ]
res  <- nmf.sem(Y1, Y2, rank = 2)
res2 <- nmf.sem.inference(res, Y1, Y2, B = 200)  # quick demo
head(res2$coefficients)
}
}
\references{
Satoh, K. (2025). Applying non-negative matrix factorization with covariates
to structural equation modeling for blind input-output analysis.
arXiv:2512.18250. \url{https://arxiv.org/abs/2512.18250}
}
\seealso{
\code{\link{nmf.sem}}, \code{\link{nmf.sem.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{nmfae.signed.inference}
\alias{nmfae.signed.inference}
\title{Statistical Inference for Signed-Bottleneck NMF-AE Signed Bottleneck}
\usage{
nmfae.signed.inference(object, Y1, Y2 = Y1, wild.bootstrap = TRUE, ...)
}
\arguments{
\item{object}{A fitted \code{"nmfae.signed"} object.}

\item{Y1}{Output matrix used during fitting.}

\item{Y2}{Input matrix used during fitting.  Default \code{Y1}.}

\item{wild.bootstrap}{Logical.  Default \code{TRUE}.}

\item{...}{Additional arguments:
\describe{
\item{\code{wild.B}}{Bootstrap replicates.  Default 500.}
\item{\code{wild.seed}}{RNG seed.  Default 123.}
\item{\code{wild.level}}{CI confidence level.  Default 0.95.}
\item{\code{sandwich}}{Use sandwich covariance.  Default \code{TRUE}.}
\item{\code{C.p.side}}{P-value type: \code{"two.sided"} (default for
Signed-Bottleneck NMF-AE) or \code{"one.sided"}.}
\item{\code{cov.ridge}}{Ridge stabilization.  Default \code{1e-8}.}
\item{\code{print.trace}}{Logical.  Default \code{FALSE}.}
}}
}
\value{
An object of class \code{c("nmfae.signed.inference",
  "nmfae.inference", "nmfae.signed", "nmfae", "nmf")} with added fields:
\item{sigma2.used}{Estimated \eqn{\sigma^2}.}
\item{C.se, C.se.boot}{Sandwich / bootstrap SEs for \eqn{\Theta} (Q x R).}
\item{C.ci.lower, C.ci.upper}{Bootstrap CIs.}
\item{coefficients}{Data frame with Estimate, SE, BSE, z, p-value, CI.}
\item{C.p.side}{P-value side used.}
}
\description{
Post-estimation inference for the \strong{signed} bottleneck
\eqn{\Theta = C_{+} - C_{-}} in the Signed-Bottleneck NMF-AE model
\eqn{Y_1 \approx X_1 \Theta X_2 Y_2}, conditional on
\eqn{(\hat X_1, \hat X_2)}.  Uses sandwich covariance and wild bootstrap
\strong{without} the non-negativity projection that \code{\link{nmfae.inference}}
applies (because \eqn{\Theta} is unconstrained in sign here).
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae.signed}}, \code{\link{nmfae.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmf.sem.R
\name{nmfkc.DOT}
\alias{nmfkc.DOT}
\title{Generate Graphviz DOT Scripts for NMF or NMF-with-Covariates Models}
\usage{
nmfkc.DOT(
  result,
  type = c("YX", "YA", "YXA"),
  threshold = 0.01,
  sig.level = 0.1,
  rankdir = "LR",
  fill = TRUE,
  weight_scale = 5,
  weight_scale_ax = weight_scale,
  weight_scale_xy = weight_scale,
  weight_scale_ay = weight_scale,
  Y.label = NULL,
  X.label = NULL,
  A.label = NULL,
  Y.title = "Observation (Y)",
  X.title = "Basis (X)",
  A.title = "Covariates (A)",
  hide.isolated = TRUE
)
}
\arguments{
\item{result}{The return value from \code{nmfkc}, containing matrices
\code{X}, \code{B}, and optionally \code{C}.}

\item{type}{Character string specifying the visualization style:
one of \code{"YX"}, \code{"YA"}, \code{"YXA"}.}

\item{threshold}{Minimum coefficient magnitude to display an edge.}

\item{sig.level}{Significance level for filtering C edges when inference
results are available (i.e., \code{x} is of class \code{"nmfkc.inference"}).
Only edges with p-value below \code{sig.level} are shown, decorated with
significance stars (\code{*}, \code{**}, \code{***}). Set to \code{NULL}
to disable filtering and show all edges above \code{threshold}. Default is 0.1.}

\item{rankdir}{Graphviz rank direction (e.g., \code{"LR"}, \code{"TB"}).}

\item{fill}{Logical; whether nodes should be drawn with filled shapes.}

\item{weight_scale}{Base scaling factor for edge widths.}

\item{weight_scale_ax}{Scaling factor for edges \eqn{A \rightarrow X} (type \code{"YXA"}).}

\item{weight_scale_xy}{Scaling factor for edges \eqn{X \rightarrow Y}.}

\item{weight_scale_ay}{Scaling factor for edges \eqn{A \rightarrow Y} (type \code{"YA"}).}

\item{Y.label}{Optional character vector for labels of Y nodes.}

\item{X.label}{Optional character vector for labels of X (latent factor) nodes.}

\item{A.label}{Optional character vector for labels of A (covariate) nodes.}

\item{Y.title}{Cluster title for Y nodes.}

\item{X.title}{Cluster title for X nodes.}

\item{A.title}{Cluster title for A nodes.}

\item{hide.isolated}{Logical. If \code{TRUE} (default), Y and A nodes that have no
edges at or above \code{threshold} are excluded from the graph.}
}
\value{
A character string representing a Graphviz DOT script.
}
\description{
Produces a Graphviz DOT script visualizing the structure of an NMF model
(\eqn{Y \approx X C A}) or its simplified forms.

Supported visualization types:
\itemize{
\item \code{"YX"} — Standard NMF view: latent factors \eqn{X} map to observations \eqn{Y}.
\item \code{"YA"} — Direct regression view: covariates \eqn{A} map directly to \eqn{Y}
using the combined coefficient matrix \eqn{X C}.
\item \code{"YXA"} — Full tri-factorization: \eqn{A \rightarrow C \rightarrow X \rightarrow Y}.
}

Edge widths are scaled by coefficient magnitude, and nodes with no edges
above the threshold are omitted from the visualization.
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(1, cars$speed)
result <- nmfkc(Y, A, rank = 1)
dot <- nmfkc.DOT(result)
cat(dot)

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfae.DOT}}, \code{\link{nmf.sem.DOT}},
\code{\link{nmfkc.ar.DOT}}, \code{\link{plot.nmfkc.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{nmfae.rename}
\alias{nmfae.rename}
\title{Rename decoder and encoder bases}
\usage{
nmfae.rename(x, X1.colnames = NULL, X2.rownames = NULL)
}
\arguments{
\item{x}{An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.}

\item{X1.colnames}{Character vector of length \eqn{Q} for decoder bases
(columns of \eqn{X_1} / rows of \eqn{\Theta}).  If \code{NULL}
(default), the decoder names are left unchanged.}

\item{X2.rownames}{Character vector of length \eqn{R} for encoder bases
(rows of \eqn{X_2} / columns of \eqn{\Theta}).  If \code{NULL}
(default), the encoder names are left unchanged.}
}
\value{
A modified copy of \code{x} with updated names.
}
\description{
Assigns user-specified names to the decoder (X1 columns) and encoder
(X2 rows) bases of an \code{nmfae} object.  The names propagate to
\eqn{\Theta}, the coefficients table, and all downstream displays
such as \code{summary}, \code{nmfae.DOT}, and \code{nmfae.heatmap}.
}
\examples{
\donttest{
set.seed(1)
Y <- matrix(runif(15), nrow = 3)
res <- nmfae(Y, rank = 2, rank.encoder = 2)
res <- nmfae.rename(res,
  X1.colnames = c("Basis1", "Basis2"),
  X2.rownames = c("Enc1", "Enc2"))
summary(res)
}
}
\seealso{
\code{\link{nmfae}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{plot.nmf.cluster.criteria}
\alias{plot.nmf.cluster.criteria}
\title{Plot clustering-quality criteria across a sequence of fits}
\usage{
\method{plot}{nmf.cluster.criteria}(
  x,
  main = "Clustering quality across rank",
  xlab = "rank (Q)",
  ylab = "criterion",
  lwd = 2,
  ...
)
}
\arguments{
\item{x}{An object of class \code{"nmf.cluster.criteria"}.}

\item{main}{Plot title.}

\item{xlab, ylab}{Axis labels.}

\item{lwd}{Line width.}

\item{...}{Further arguments passed to the initial
\code{\link[graphics]{plot}}.}
}
\value{
\code{x}, invisibly.
}
\description{
Line plot of \code{silhouette}, \code{CPCC}, and \code{dist.cor}
against the result index, for an object from
\code{\link{nmf.cluster.criteria}}.  X-axis ticks default to each
result's \code{$rank} (overridable via the \code{names} argument of
\code{\link{nmf.cluster.criteria}}).
}
\seealso{
\code{\link{nmf.cluster.criteria}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{nmfae.signed.rename}
\alias{nmfae.signed.rename}
\title{Rename Dec/Enc labels on nmfae.signed objects}
\usage{
nmfae.signed.rename(x, X1.colnames = NULL, X2.rownames = NULL)
}
\arguments{
\item{x}{An \code{"nmfae.signed"} object.}

\item{X1.colnames}{Character vector of length Q for decoder labels.}

\item{X2.rownames}{Character vector of length R for encoder labels.}
}
\value{
The renamed object (same class).
}
\description{
Replaces the default \code{"Dec1", "Dec2", ...} (decoder / X1 columns
and Cp/Cn/C rows) and \code{"Enc1", "Enc2", ...} (encoder / X2 rows and
Cp/Cn/C columns) with user-supplied labels.  Propagates to coefficients
tables if present (e.g., from \code{nmfae.inference}).
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae.signed}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{plot.nmfkc}
\alias{plot.nmfkc}
\title{Plot method for objects of class \code{nmfkc}}
\usage{
\method{plot}{nmfkc}(x, ...)
}
\arguments{
\item{x}{An object of class \code{nmfkc}, i.e., the return value of \code{nmfkc}.}

\item{...}{Additional arguments passed to the base \code{\link{plot}} function.}
}
\value{
Called for its side effect (a plot). Returns \code{NULL} invisibly.
}
\description{
\code{plot.nmfkc} produces a diagnostic plot for the return value of
\code{nmfkc}, showing the objective function across iterations.
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(1, cars$speed)
result <- nmfkc(Y, A, rank = 1)
plot(result)

}
\seealso{
\code{\link{nmfkc}}, \code{\link{summary.nmfkc}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{nmfae.ecv}
\alias{nmfae.ecv}
\title{Element-wise Cross-Validation for nmfae (Wold's CV)}
\usage{
nmfae.ecv(Y1, Y2 = Y1, rank = 1:2, rank.encoder = NULL, ...)
}
\arguments{
\item{Y1}{Output matrix \eqn{Y_1} (P1 x N).}

\item{Y2}{Input matrix \eqn{Y_2} (P2 x N). Default is \code{Y1}.}

\item{rank}{Integer vector of decoder ranks to evaluate. Default is \code{1:2}.}

\item{rank.encoder}{Integer vector of encoder ranks to evaluate. Default is \code{NULL},
which sets \code{rank.encoder = rank} and evaluates element-wise pairs.
When explicitly specified, all combinations with \code{rank} are evaluated.}

\item{...}{Additional arguments passed to \code{\link{nmfae}} (e.g., \code{epsilon}, \code{maxit}).
Also accepts: \code{nfolds} (number of folds, default 5; \code{div} also accepted),
\code{seed} (integer seed, default 123).
For backward compatibility, \code{Q} and \code{R} are accepted as aliases for
\code{rank} and \code{rank.encoder}.}
}
\value{
A list with components:
\item{objfunc}{Named numeric vector of mean MSE for each (Q, R) pair.}
\item{sigma}{Named numeric vector of RMSE (square root of MSE) for each pair.}
\item{objfunc.fold}{Named list of per-fold MSE vectors for each pair.}
\item{folds}{List of length \code{div} containing the held-out element indices for each fold.}
\item{QR}{Data frame with columns \code{Q} and \code{R} listing the evaluated pairs.}
}
\description{
\code{nmfae.ecv} performs k-fold element-wise cross-validation by randomly
holding out individual elements of \eqn{Y_1}, assigning them a weight of 0
via \code{Y1.weights}, and evaluating the reconstruction error on those
held-out elements.

This method (also known as Wold's CV) is suitable for determining the optimal
rank pair \eqn{(Q, R)} in three-layer NMF. Both \code{rank} and \code{rank.encoder} accept
vector inputs. When \code{rank.encoder = NULL} (default), \code{rank.encoder} is set equal to \code{rank}
and pairs are evaluated element-wise (i.e., \eqn{(Q_1, R_1), (Q_2, R_2), \dots}).
When \code{rank.encoder} is explicitly specified, all combinations of \code{rank} and \code{rank.encoder}
are evaluated via \code{expand.grid}.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\examples{
Y <- t(iris[1:30, 1:4])
# Default: rank.encoder=NULL -> paired rank=rank.encoder
res <- nmfae.ecv(Y, rank = 1:3, nfolds = 3, maxit = 500)
res$sigma
# Explicit rank.encoder: full grid
res2 <- nmfae.ecv(Y, rank = 1:3, rank.encoder = 1:3, nfolds = 3, maxit = 500)
res2$sigma

}
\seealso{
\code{\link{nmfae}}, \code{\link{nmfkc.ecv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{predict.nmfkc}
\alias{predict.nmfkc}
\title{Prediction method for objects of class \code{nmfkc}}
\usage{
\method{predict}{nmfkc}(object, newA = NULL, newdata = NULL, type = "response", ...)
}
\arguments{
\item{object}{An object of class \code{nmfkc}, i.e., the return value of \code{nmfkc}.}

\item{newA}{Optional. A new covariate matrix to be used for prediction.}

\item{newdata}{Optional data frame. Only available when the model was fitted
using a formula. Covariate columns are extracted automatically using the
stored formula metadata. If both \code{newdata} and \code{newA} are
supplied, \code{newdata} takes precedence (with a warning).}

\item{type}{Type of prediction to return. Options are "response" (fitted values matrix),
"prob" (soft-clustering probabilities), or "class" (hard-clustering labels based on row names of X).}

\item{...}{Further arguments passed to or from other methods.}
}
\value{
Depending on \code{type}: a numeric matrix (\code{"response"} or \code{"prob"})
or a character vector of class labels (\code{"class"}).
}
\description{
\code{predict.nmfkc} generates predictions from an object of class \code{nmfkc},
either using the fitted covariates or a new covariate matrix.

When the model was fitted using a formula (Formula Mode), a \code{newdata}
data frame can be supplied instead of \code{newA}; the covariate matrix is
then constructed automatically from the stored formula metadata.
}
\examples{
# Prediction with newA
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(1, cars$speed)
result <- nmfkc(Y, A, rank = 1)
newA <- rbind(1, c(10, 20, 30))
predict(result, newA = newA)

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfkc.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.denormalize}
\alias{nmfkc.denormalize}
\title{Denormalize a matrix from \eqn{[0,1]} back to its original scale}
\usage{
nmfkc.denormalize(x, ref = x)
}
\arguments{
\item{x}{A numeric matrix (or vector) with values in \eqn{[0,1]} to be denormalized.}

\item{ref}{A reference matrix used to obtain the original column-wise minima
and maxima. Must have the same number of columns as \code{x}.}
}
\value{
A numeric matrix with values transformed back to the original scale.
}
\description{
\code{nmfkc.denormalize} rescales a matrix with values in \eqn{[0,1]} back to its
original scale using the column-wise minima and maxima of a reference matrix.
}
\examples{
x <- nmfkc.normalize(iris[, -5])
x_recovered <- nmfkc.denormalize(x, iris[, -5])
apply(x_recovered - iris[, -5], 2, max)
}
\seealso{
\code{\link{nmfkc.normalize}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{summary.nmfae.signed.inference}
\alias{summary.nmfae.signed.inference}
\title{Summary method for nmfae.signed.inference objects}
\usage{
\method{summary}{nmfae.signed.inference}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"nmfae.signed.inference"}.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An object of class \code{"summary.nmfae.signed.inference"}.
}
\description{
Produces a summary of a fitted Signed-Bottleneck NMF-AE model with
inference results.  Extends \code{\link{summary.nmfae.signed}} by
attaching the \code{coefficients} table and p-value side from
\code{\link{nmfae.signed.inference}}.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae.signed.inference}}, \code{\link{summary.nmfae.signed}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{nmfae.signed.rank}
\alias{nmfae.signed.rank}
\title{Rank selection for nmfae.signed (paired rank, concise diagnostics)}
\usage{
nmfae.signed.rank(
  Y1,
  Y2 = Y1,
  rank = 1:5,
  detail = c("full", "fast"),
  plot = TRUE,
  ...
)
}
\arguments{
\item{Y1}{Endogenous matrix (\eqn{P_1 \times N}); may be signed.}

\item{Y2}{Exogenous matrix; defaults to \code{Y1}.}

\item{rank}{Integer vector of (paired) ranks to evaluate.}

\item{detail}{\code{"full"} (default) also runs element-wise CV
(\code{sigma.ecv}); \code{"fast"} skips it (plots r.squared and
eff.rank only, and recommends the R-squared elbow).}

\item{plot}{Logical; draw the diagnostics plot (default \code{TRUE}).}

\item{...}{Passed on to \code{\link{nmfae.signed}} and
\code{\link{nmfae.signed.ecv}}.}
}
\value{
A list with \code{rank.best} and \code{criteria}
(\code{rank}, \code{effective.rank}, \code{effective.rank.ratio},
\code{r.squared}, \code{sigma.ecv}).
}
\description{
Fits \code{\link{nmfae.signed}} with a \strong{paired} decoder/encoder
rank (\eqn{Q = R}) across a range of ranks and reports
\code{r.squared}, the effective rank (of the latent encoding \eqn{H}),
and the element-wise CV error \code{sigma.ecv}, with the same concise
plot as \code{\link{nmfkc.rank}}.  For a full \eqn{(Q, R)} grid use
\code{\link{nmfae.signed.ecv}}.
}
\references{
Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
effective dimensionality.  \emph{Proc. EUSIPCO}, 606--610.
(\code{effective.rank})
Wold, S. (1978).  Cross-validatory estimation of the number of
components in factor and principal components models.
\emph{Technometrics}, 20(4), 397--405. (\code{sigma.ecv})
}
\seealso{
\code{\link{nmfae.signed}}, \code{\link{nmfae.signed.ecv}},
\code{\link{nmfkc.rank}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.criterion}
\alias{nmfkc.criterion}
\title{Compute model selection criteria for a fitted nmfkc model}
\usage{
nmfkc.criterion(object, Y, detail = c("full", "fast", "minimal"), ...)
}
\arguments{
\item{object}{An object of class \code{"nmfkc"} returned by
\code{\link{nmfkc}}.}

\item{Y}{The original observation matrix (P x N) used for fitting.}

\item{detail}{Character string controlling the level of computation:
\code{"full"} (default) computes all criteria including silhouette,
CPCC and dist.cor;
\code{"fast"} skips the expensive distance-based criteria;
\code{"minimal"} skips the fit and clustering statistics.}

\item{...}{Additional arguments: \code{Y.weights} (non-negative
weight matrix; \code{lm()}-style loss \eqn{\sum W \, r^2}; default:
all ones).  See \code{\link{nmfkc}} for full details.}
}
\value{
A list with components:
\describe{
\item{r.squared}{R-squared between Y and XB.}
\item{sigma}{Residual standard deviation.}
\item{mae}{Mean absolute error.}
\item{B.prob}{Column-normalized coefficient matrix (soft-clustering probabilities).}
\item{B.cluster}{Hard clustering labels (argmax of B.prob per column).}
\item{X.prob}{Row-normalized basis matrix.}
\item{X.cluster}{Hard clustering labels per row of X.}
\item{criterion}{Named list: B.prob.max.mean, effective.rank,
silhouette, CPCC, dist.cor.}
}
}
\description{
\code{nmfkc.criterion} computes the effective rank, clustering-quality
measures (silhouette, CPCC, dist.cor), and the clustering-crispness
statistic (\code{B.prob.max.mean}) from a fitted \code{nmfkc} model.

This function can be called on a model that was fitted with
\code{detail = "fast"} or \code{detail = "minimal"} to compute the
full set of criteria afterwards.
}
\examples{
Y <- t(iris[, -5])
res <- nmfkc(Y, rank = 3, detail = "fast")
crit <- nmfkc.criterion(res, Y)
crit$criterion$silhouette

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfkc.rank}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.consensus.R
\name{nmfkc.consensus}
\alias{nmfkc.consensus}
\title{Consensus-clustering rank selection for NMF (Brunet 2004)}
\usage{
nmfkc.consensus(
  Y,
  A = NULL,
  rank = 2:4,
  nrun = 30,
  keep.consensus = FALSE,
  ...
)
}
\arguments{
\item{Y}{Observation matrix (\eqn{P \times N}), non-negative.}

\item{A}{Optional covariate matrix passed to \code{\link{nmfkc}}.}

\item{rank}{Integer vector of ranks to evaluate (\eqn{\ge 2}).}

\item{nrun}{Number of random-initialization runs per rank (default 30).}

\item{keep.consensus}{Logical; if \code{TRUE} also return the list of
consensus matrices (one \eqn{N \times N} matrix per rank).}

\item{...}{Advanced options, rarely needed (defaults in parentheses):
\code{seed} (\code{123}, base seed; run \eqn{r} of rank index \eqn{i} uses
\code{seed + 1000 * i + r}) and \code{pac.range} (\code{c(0.1, 0.9)}, the
ambiguous interval \eqn{(u_1, u_2)} for the PAC measure).  Any other
arguments are passed to \code{\link{nmfkc}} (e.g.\ \code{maxit});
\code{X.init} is forced to \code{"runif"}.}
}
\value{
An object of class \code{"nmfkc.consensus"} (a list) with:
\item{cophenetic}{Cophenetic correlation coefficient for each rank.}
\item{dispersion}{Dispersion coefficient (\eqn{[0, 1]}) for each rank.}
\item{pac}{Proportion of Ambiguous Clustering (\eqn{[0, 1]}, lower is
better) for each rank.}
\item{rank}{The evaluated rank vector.}
\item{nrun}{Number of runs per rank.}
\item{consensus}{List of consensus matrices, or \code{NULL}.}
It has \code{\link{print.nmfkc.consensus}} and
\code{\link{plot.nmfkc.consensus}} (\code{type = "criteria"} /
\code{"heatmap"}) methods.
}
\description{
The bioinformatics-standard stability approach to choosing the NMF rank.
A lightweight engine like \code{\link{nmfkc.ecv}} / \code{\link{nmfkc.bicv}}:
it returns one stability score per rank and nothing more.

For each rank, NMF is run \code{nrun} times from different random
initializations (\code{X.init = "runif"}).  Each run gives a hard
clustering of the samples (the column \eqn{\arg\max} of the coefficient
matrix).  Averaging the \eqn{N \times N} connectivity matrices (1 if two
samples share a cluster) over the runs yields the \strong{consensus
matrix}; its crispness measures how reproducible the clustering is.  Two
summaries are reported per rank:
\itemize{
\item \code{cophenetic}: the cophenetic correlation coefficient (CPCC)
of the consensus matrix (Brunet et al. 2004).  Close to 1 = stable.
\item \code{dispersion}: the Kim & Park (2007) dispersion
\eqn{\frac{1}{N^2}\sum_{ij} 4 (C_{ij} - 1/2)^2 \in [0, 1]}; 1 when
every consensus entry is exactly 0 or 1 (perfectly crisp).
\item \code{pac}: the Proportion of Ambiguous Clustering
(Senbabaoglu et al. 2014) -- the fraction of off-diagonal consensus
entries falling in the ambiguous interval \code{pac.range}
(default \eqn{(0.1, 0.9)}).  \strong{Lower is better} (less
ambiguity); a more sensitive readout than \code{cophenetic}, which
tends to saturate.
}
Unlike the cross-validation engines (where the rank \emph{minimizes} the
error), here a good rank \strong{maximizes} stability, or is the largest
rank before it drops.
}
\examples{
\donttest{
Y <- t(as.matrix(iris[, 1:4]))
cs <- nmfkc.consensus(Y, rank = 2:5, nrun = 20, keep.consensus = TRUE)
cs                       # stability table per rank
plot(cs)                 # type = "criteria": stability curves
plot(cs, type = "heatmap")            # all ranks, n2mfrow grid
plot(cs, type = "heatmap", rank = 3)  # one rank, with labels
}
}
\references{
Brunet, J.-P., Tamayo, P., Golub, T. R., Mesirov, J. P. (2004).
Metagenes and molecular pattern discovery using matrix factorization.
\emph{PNAS} 101(12):4164--4169. \doi{10.1073/pnas.0308531101}.
Kim, H., Park, H. (2007).  Sparse non-negative matrix factorizations
\dots \emph{Bioinformatics} 23(12):1495--1502.
Senbabaoglu, Y., Michailidis, G., Li, J. Z. (2014).  Critical
limitations of consensus clustering in class discovery.
\emph{Sci. Rep.} 4:6207. \doi{10.1038/srep06207}.
}
\seealso{
\code{\link{nmfkc.rank}}, \code{\link{nmfkc.ecv}},
\code{\link{nmfkc.bicv}}, \code{\link{nmfkc.ard}},
\code{\link{nmf.cluster.criteria}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{nmfae}
\alias{nmfae}
\title{Three-Layer Non-negative Matrix Factorization (NMF-AE)}
\source{
Satoh, K. (2025). Applying Non-negative Matrix Factorization with Covariates
to Multivariate Time Series. \emph{Japanese Journal of Statistics and Data Science}.
}
\usage{
nmfae(
  Y1,
  Y2 = Y1,
  rank = 2,
  rank.encoder = rank,
  epsilon = 1e-04,
  maxit = 5000,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{Y1}{Output matrix \eqn{Y_1} (P1 x N). Non-negative. May contain \code{NA}s
(handled via \code{Y1.weights}).}

\item{Y2}{Input matrix \eqn{Y_2} (P2 x N). Non-negative. Default is \code{Y1} (autoencoder).}

\item{rank}{Integer. Rank of the decoder basis \eqn{X_1} (P1 x Q). Default is 2.
For backward compatibility, \code{Q} is accepted via \code{...}.}

\item{rank.encoder}{Integer. Rank of the encoder basis \eqn{X_2} (R x P2).
Default is \code{rank}. For backward compatibility, \code{R} is accepted via \code{...}.}

\item{epsilon}{Positive convergence tolerance. Default is \code{1e-4}.}

\item{maxit}{Maximum number of multiplicative update iterations. Default is 5000.}

\item{verbose}{Logical. If \code{TRUE}, prints progress messages during fitting. Default is \code{FALSE}.}

\item{...}{Additional arguments:
\describe{
\item{\code{Y1.weights}}{Optional non-negative weight matrix
(P1 x N) or vector for \eqn{Y_1}, analogous to the
\code{weights} argument of \code{\link[stats]{lm}}.  Loss becomes
\eqn{\sum W_{ij} \, (Y_{1,ij} - \hat Y_{1,ij})^2}
(\code{lm()}-style, \strong{linear} in \eqn{W}).  Logical
matrices (\code{TRUE} / \code{FALSE}) are also accepted.
Typical ECV / CV usage passes a binary mask
\eqn{W \in \{0,1\}} for held-out elements; real-valued weights
for importance weighting are also supported.  Default: if
\code{Y1} has \code{NA}, a binary mask is auto-generated
(0 for \code{NA}, 1 elsewhere).}
\item{\code{C.L1}}{L1 regularization parameter for \eqn{C}. Default is 0.}
\item{\code{X1.L2.ortho}}{L2 orthogonality regularization for \eqn{X_1} columns. Default is 0.}
\item{\code{X2.L2.ortho}}{L2 orthogonality regularization for \eqn{X_2} rows. Default is 0.}
\item{\code{seed}}{Integer seed for reproducibility. Default is 123.}
\item{\code{print.trace}}{Logical. If \code{TRUE}, prints progress. Default is \code{FALSE}.}
}}
}
\value{
An object of class \code{"nmfae"}, a list with components:
\item{X1}{Decoder basis matrix (P1 x Q), column sum 1.}
\item{C}{Parameter matrix (Q x R).}
\item{X2}{Encoder basis matrix (R x P2), row sum 1.}
\item{Y1hat}{Fitted values \eqn{X_1 \Theta X_2 Y_2} (P1 x N).}
\item{rank}{Named integer vector \code{c(Q, R)}.}
\item{objfunc}{Final objective value.}
\item{objfunc.iter}{Objective values by iteration.}
\item{r.squared}{\eqn{\mathrm{cor}(Y, \widehat Y)^2} (Pearson; in \eqn{[0,1]}).}
\item{r.squared.uncentered}{Uncentered \eqn{R^2 = 1 - \|Y - \widehat Y\|_F^2 / \|Y\|_F^2} (baseline = zero matrix).}
\item{r.squared.centered}{Row-mean centered \eqn{1 - \|Y - \widehat Y\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2}.}
\item{niter}{Number of iterations performed.}
\item{runtime}{Elapsed time as a \code{difftime} object.}
\item{n.missing}{Number of missing (or zero-weighted) elements in \eqn{Y_1}.}
\item{n.total}{Total number of elements in \eqn{Y_1} (P1 x N).}
}
\description{
\code{nmfae} fits a three-layer nonnegative matrix factorization model
\eqn{Y_1 \approx X_1 \Theta X_2 Y_2}, where \eqn{X_1} is a decoder basis
(column sum 1), \eqn{\Theta} is a bottleneck parameter matrix,
\eqn{X_2} is an encoder basis (row sum 1), and \eqn{Y_2} is the input matrix.

When \code{Y2 = Y1}, the model acts as a non-negative autoencoder.
When \code{Y1 != Y2}, it acts as a heteroencoder.

Initialization uses a three-step NMF procedure via \code{\link{nmfkc}}:
(1) \code{nmfkc(Y1, rank=Q)} to obtain \eqn{X_1},
(2) \code{nmfkc(Y1, A=Y2, rank=Q)} with fixed \eqn{X_1} to obtain \eqn{C = \Theta X_2},
(3) \code{nmfkc(Y2, rank=R)} to factor \eqn{C} into \eqn{\Theta} and \eqn{X_2}.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in future versions.
}

\examples{
# Autoencoder example
Y <- matrix(c(1,0,1,0, 0,1,0,1, 1,1,0,0), nrow=3, byrow=TRUE)
res <- nmfae(Y, rank=2, rank.encoder=2)
res$r.squared

# Heteroencoder example
Y1 <- matrix(c(1,0,0,1), nrow=2)
Y2 <- matrix(runif(8), nrow=4)
res2 <- nmfae(Y1, Y2, rank=2, rank.encoder=2)

}
\references{
Lee, D. D. and Seung, H. S. (2001). Algorithms for Non-negative Matrix Factorization.
\emph{Advances in Neural Information Processing Systems}, 13.

Saha, S. et al. (2022). Hierarchical Deep Learning Neural Network (HiDeNN):
An Artificial Intelligence (AI) Framework for Computational Science and Engineering.
\emph{Computer Methods in Applied Mechanics and Engineering}, 399.
}
\seealso{
\code{\link{nmfae.inference}}, \code{\link{predict.nmfae}}, \code{\link{nmfae.ecv}}, \code{\link{nmfae.DOT}}, \code{\link{nmfkc}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.signed.R
\name{plot.nmfkc.signed}
\alias{plot.nmfkc.signed}
\title{Plot method for nmfkc.signed (convergence)}
\usage{
\method{plot}{nmfkc.signed}(x, ...)
}
\arguments{
\item{x}{An \code{nmfkc.signed} object.}

\item{...}{Passed to \code{plot()}.}
}
\value{
Invisible \code{x}.
}
\description{
Plot method for nmfkc.signed (convergence)
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{print.summary.nmf.sem}
\alias{print.summary.nmf.sem}
\title{Print method for summary.nmf.sem objects}
\usage{
\method{print}{summary.nmf.sem}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"summary.nmf.sem"} returned by
\code{\link{summary.nmf.sem}}.}

\item{...}{Not used.}
}
\value{
Invisible \code{x}.
}
\description{
Prints the NMF-FFB model summary (dimensions, convergence, stability
diagnostics, fit statistics, and inference results if available).
}
\seealso{
\code{\link{summary.nmf.sem}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfre.R
\name{nmfre.dfU.scan}
\alias{nmfre.dfU.scan}
\title{Scan dfU cap rates for NMF-RE}
\usage{
nmfre.dfU.scan(
  rates = (1:10)/10,
  Y,
  A,
  rank = NULL,
  X.init = NULL,
  C.init = NULL,
  U.init = NULL,
  print.trace = FALSE,
  ...
)
}
\arguments{
\item{rates}{Numeric vector of cap rates to scan (default \code{(1:10)/10}).}

\item{Y}{Observation matrix (P x N).}

\item{A}{Covariate matrix (K x N).}

\item{rank}{Integer. Rank of the basis matrix.
For backward compatibility, \code{Q} is accepted via \code{...}.}

\item{X.init}{Initial basis matrix, or \code{NULL}.}

\item{C.init}{Initial coefficient matrix, or \code{NULL}.}

\item{U.init}{Initial random effects matrix, or \code{NULL}.}

\item{print.trace}{Logical. Print progress for each fit (default \code{FALSE}).}

\item{...}{Additional arguments passed to \code{\link{nmfre}}.}
}
\value{
An object of class \code{"nmfre.dfU.scan"} with two components:
\describe{
\item{\code{table}}{A data frame with the following columns:
\describe{
\item{\code{rate}}{Cap rate used. The dfU cap is \code{rate * N * Q}.}
\item{\code{dfU.cap}}{The dfU cap value (upper bound on effective degrees of freedom).}
\item{\code{dfU}}{Final effective degrees of freedom for \eqn{U} at convergence.}
\item{\code{safeguard}}{Logical. \code{TRUE} if the dfU cap is functioning
as a safeguard (\code{dfU / dfU.cap < 0.99}): the cap prevents
random-effects saturation without over-constraining \eqn{U}.
\code{FALSE} if dfU is at or near the cap, indicating the cap is
binding and the rate may be too small.}
\item{\code{hit}}{Logical. \code{TRUE} if the cap was reached at least once
during iteration, even if dfU later decreased below the cap.}
\item{\code{converged}}{Logical. \code{TRUE} if the algorithm converged
within the maximum number of iterations.}
\item{\code{tau2}}{Final random effect variance \eqn{\hat{\tau}^2}.}
\item{\code{sigma2}}{Final residual variance \eqn{\hat{\sigma}^2}.}
\item{\code{ICC}}{Trace-based Intraclass Correlation Coefficient
\eqn{\tau^2 \, \mathrm{tr}(X^\top X) /
      (\tau^2 \, \mathrm{tr}(X^\top X) + \sigma^2 P)}.
See \code{\link{nmfre}} for details.}
}}
\item{\code{cap.rate}}{Optimal cap rate selected automatically.
If rows with \code{safeguard = TRUE} and \code{hit = TRUE} exist,
the maximum rate among them is chosen (safeguard activated but
giving \eqn{U} the most freedom). Otherwise, the minimum rate
with \code{safeguard = TRUE} and \code{hit = FALSE} is chosen.
\code{NA} if no suitable rate is found.}
}

When printed, only the \code{table} is displayed. Access \code{cap.rate}
directly from the returned object.
}
\description{
Fits the NMF-RE model across a range of \code{dfU.cap.rate} values and
returns a diagnostic table showing the resulting effective degrees of freedom,
variance components, and convergence diagnostics for each rate.

The dfU cap limits the effective degrees of freedom consumed by the random
effects \eqn{U}. The cap is computed as \code{rate * N * Q}, where \eqn{N}
is the number of observations and \eqn{Q} is the rank. A suitable rate is
one where the final \eqn{\mathrm{df}_U} is below the cap
(\code{safeguard = TRUE}) and the model has converged
(\code{converged = TRUE}).

When called automatically by \code{\link{nmfre}} (i.e., \code{dfU.cap.rate = NULL}),
the minimum rate satisfying both \code{safeguard = TRUE} and
\code{converged = TRUE} is selected.
}
\examples{
# Example 1. cars data (small maxit for speed)
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(intercept = 1, speed = cars$speed)
tab <- nmfre.dfU.scan(rates = c(0.1, 0.2), Y = Y, A = A, rank = 1, maxit = 1000)
print(tab)

\donttest{
# Example 2. Orthodont data (nlme)
if (requireNamespace("nlme", quietly = TRUE)) {
  Y <- matrix(nlme::Orthodont$distance, 4, 27)
  male <- ifelse(nlme::Orthodont$Sex[seq(1, 108, 4)] == "Male", 1, 0)
  A <- rbind(intercept = 1, male = male)
  nmfre.dfU.scan(1:10/10, Y, A, rank = 1)
}}

}
\seealso{
\code{\link{nmfre}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{print.summary.nmfkc}
\alias{print.summary.nmfkc}
\title{Print method for \code{summary.nmfkc} objects}
\usage{
\method{print}{summary.nmfkc}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{An object of class \code{summary.nmfkc}.}

\item{digits}{Minimum number of significant digits to be used.}

\item{...}{Additional arguments (currently unused).}
}
\value{
Called for its side effect (printing). Returns \code{x} invisibly.
}
\description{
Prints a formatted summary of an \code{nmfkc} model fit.
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(1, cars$speed)
result <- nmfkc(Y, A, rank = 1)
print(summary(result))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmf.ffb.R, R/nmf.sem.R
\name{nmf.ffb.DOT}
\alias{nmf.ffb.DOT}
\alias{nmf.sem.DOT}
\title{Generate a Graphviz DOT Diagram for an NMF-FFB Model}
\usage{
nmf.ffb.DOT(result, ...)

nmf.sem.DOT(
  result,
  weight_scale = 5,
  weight_scale_c2 = weight_scale,
  weight_scale_x1 = weight_scale,
  weight_scale_feedback = weight_scale,
  threshold = 0.01,
  sig.level = 0.1,
  rankdir = "LR",
  fill = TRUE,
  cluster.box = c("normal", "faint", "invisible", "none"),
  cluster.labels = NULL,
  hide.isolated = TRUE,
  ...
)
}
\arguments{
\item{result}{A list returned by \code{nmf.sem}, containing matrices
\code{X}, \code{C1}, and \code{C2}.}

\item{...}{For backward compatibility: accepts deprecated names
\code{weight_scale_y2f} (use \code{weight_scale_c2}) and
\code{weight_scale_fy1} (use \code{weight_scale_x1}).}

\item{weight_scale}{Base scaling factor for edge widths.}

\item{weight_scale_c2}{Scaling factor for edges
\eqn{Y_2 \rightarrow F_q} (C2 matrix). Defaults to \code{weight_scale}.}

\item{weight_scale_x1}{Scaling factor for edges
\eqn{F_q \rightarrow Y_1} (X matrix). Defaults to \code{weight_scale}.}

\item{weight_scale_feedback}{Scaling factor for feedback edges
\eqn{Y_1 \rightarrow F_q} (C1 matrix). Defaults to \code{weight_scale}.}

\item{threshold}{Minimum coefficient value needed for an edge to be drawn.}

\item{sig.level}{Significance level for filtering structural edges
(\eqn{C_1} feedback and \eqn{C_2} exogenous loadings) when
inference results are present.  If \code{result} contains a
\code{coefficients} data frame from \code{\link{nmf.sem.inference}},
only edges with \code{p_value < sig.level} are drawn, with
significance stars (\code{*} \code{**} \code{***}) appended to
the edge label.  The \eqn{X} (factor-to-\eqn{Y_1}) edges are
never starred since the basis is not the inference target.
Set to \code{NULL} to disable significance filtering and fall
back to the \code{threshold} magnitude filter for both \eqn{C_1}
and \eqn{C_2}.  Default is \code{0.1}.}

\item{rankdir}{Graphviz rank direction (e.g., \code{"LR"}, \code{"TB"}).}

\item{fill}{Logical; whether to use filled node shapes.}

\item{cluster.box}{Character string controlling the visibility and style
of cluster frames around Y2, factors, and Y1 blocks.
One of \code{"normal"}, \code{"faint"}, \code{"invisible"}, \code{"none"}.}

\item{cluster.labels}{Optional character vector of length 3 giving custom
labels for the Y2, factor, and Y1 clusters.}

\item{hide.isolated}{Logical. If \code{TRUE} (default), Y1 and Y2 nodes
that have no edges at or above \code{threshold} are excluded from the graph.}
}
\value{
A character string representing a valid Graphviz DOT script.
}
\description{
Creates a Graphviz DOT script that visualizes the structural network
estimated by \code{nmf.sem}.
The resulting diagram displays:
\itemize{
\item endogenous observed variables (\eqn{Y_1}),
\item exogenous observed variables (\eqn{Y_2}),
\item latent factors (\eqn{F_1}, \dots, \eqn{F_Q}),
}
together with the non-negative path coefficients whose magnitudes
exceed a user-specified threshold.

Directed edges represent estimated relationships:
\itemize{
\item \eqn{Y_2 \rightarrow F_q}: entries of \code{C2} (exogenous loadings),
\item \eqn{F_q \rightarrow Y_1}: rows of \code{X} (factor-to-endogenous mappings),
\item \eqn{Y_1 \rightarrow F_q}: entries of \code{C1} (feedback paths).
}

Edge widths are scaled by coefficient magnitude, and nodes are placed
in optional visual clusters. Only variables participating in
edges above the threshold are displayed, while latent factors are always shown.
}
\examples{
Y <- t(iris[, -5])
Y1 <- Y[1:2, ]
Y2 <- Y[3:4, ]
result <- nmf.sem(Y1, Y2, rank = 2, maxit = 500)
dot <- nmf.sem.DOT(result)
cat(dot)

}
\seealso{
\code{\link{nmf.sem}}, \code{\link{nmf.sem.inference}},
\code{\link{plot.nmfkc.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{residuals.nmf}
\alias{residuals.nmf}
\alias{residuals.nmfae}
\alias{residuals.nmfre}
\alias{residuals.nmf.sem}
\title{Extract residuals from NMF models}
\usage{
\method{residuals}{nmf}(object, Y, ...)

\method{residuals}{nmfae}(object, Y, ...)

\method{residuals}{nmfre}(object, Y, type = c("blup", "fixed"), ...)

\method{residuals}{nmf.sem}(object, Y, ...)
}
\arguments{
\item{object}{A fitted model object.}

\item{Y}{The original observation matrix used for fitting.}

\item{...}{Not used.}

\item{type}{For \code{nmfre} objects: \code{"blup"} (default) or
\code{"fixed"}.}
}
\value{
The residual matrix.
}
\description{
Returns the residual matrix \eqn{Y - \hat{Y}} from a fitted NMF model.
Requires the original observation matrix \code{Y} to be supplied.

For \code{nmfre} objects, residuals are computed from the BLUP
reconstruction (\eqn{Y - X(B_{blup})}) by default. Set
\code{type = "fixed"} to use fixed-effects only.
}
\examples{
Y <- matrix(runif(50), 5, 10)
result <- nmfkc(Y, rank = 2)
residuals(result, Y)

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfae}}, \code{\link{nmfre}},
\code{\link{nmf.sem}}, \code{\link{fitted.nmf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.bicv.R
\name{nmfkc.bicv}
\alias{nmfkc.bicv}
\title{Bi-cross-validation for NMF rank selection}
\usage{
nmfkc.bicv(Y, rank = 1:3, ...)
}
\arguments{
\item{Y}{Observation matrix (\eqn{P \times N}), non-negative.}

\item{rank}{Integer vector of ranks to evaluate.}

\item{...}{Advanced options, rarely needed (defaults in parentheses):
\code{nfolds} (\code{2}), the number of row and column folds (the grid is
\code{nfolds x nfolds}; \code{2} leaves out half the rows / columns, Owen
& Perry's recommendation); \code{seed} (\code{123}, fold-assignment seed);
and \code{nnls.maxit} (\code{100}, multiplicative-update iterations for the
fold-in non-negative regressions).  Any other arguments are passed to
\code{\link{nmfkc}} for the per-block fits (e.g.\ \code{maxit}).}
}
\value{
A list (cf.\ \code{\link{nmfkc.ecv}}) with:
\item{objfunc}{Held-out mean squared error for each rank.}
\item{sigma}{Its square root (RMSE) for each rank.}
\item{rank}{The evaluated rank vector.}
\item{nfolds}{The number of folds used.}
}
\description{
Owen & Perry's (2009) bi-cross-validation (BCV) for choosing the NMF
rank.  A lightweight CV engine in the spirit of \code{\link{nmfkc.ecv}}:
it returns the held-out error per rank and nothing more (no plot, no
table) -- pass the result to \code{which.min(sigma)}, or build your own
diagnostics.

Unlike the element-wise CV of \code{\link{nmfkc.ecv}} (which holds out
scattered \emph{entries} and refits with weights), BCV holds out a whole
\strong{row-block and column-block} simultaneously: the model is fit only
on the retained block \eqn{D}, and the held-out block \eqn{A} is predicted
by folding the held-out rows/columns onto the fixed \eqn{D}-factors via
non-negative regression (\eqn{\hat A = L_I R_J}).  Because the held-out
rows and columns never enter the fit, there is no information leakage.
Covariates are ignored (plain NMF).  The recommended setting is to leave
out roughly half the rows and half the columns (\code{nfolds = 2}).
}
\details{
Each fold keeps about \eqn{(1 - 1/\text{nfolds})} of the rows
and columns, so the retained block \eqn{D} must have more than
\code{rank} rows \strong{and} columns.  The largest testable rank is
therefore about \eqn{(1 - 1/\text{nfolds})\min(P, N) - 1}; with
\code{nfolds = 2} this is roughly \eqn{\min(P, N)/2 - 1}.  Ranks above
this return \code{NA} and trigger a \code{\link[base]{warning}} that
names the limit and the \code{nfolds} (or \code{\link{nmfkc.ecv}})
that would reach the requested ranks.  Raising \code{nfolds} lifts the
limit at the cost of a smaller hold-out and more compute
(\eqn{(\text{nfolds} - 1)^2} full fits per rank).
}
\examples{
\donttest{
## rank-3 non-negative data; bi-CV needs enough kept rows/cols per
## fold (> rank), so use a matrix with ample dimensions.
set.seed(1)
X <- matrix(abs(rnorm(30 * 3)), 30, 3)
B <- matrix(abs(rnorm(3 * 40)), 3, 40)
bv <- nmfkc.bicv(X \%*\% B, rank = 1:6)   # nfolds = 2 (Owen & Perry) by default
bv$sigma                  # held-out RMSE per rank
bv$rank[which.min(bv$sigma)]
}
}
\references{
A. B. Owen and P. O. Perry (2009).  Bi-cross-validation of
the SVD and the nonnegative matrix factorization.
\emph{Ann. Appl. Stat.} 3(2):564--594. \doi{10.1214/08-AOAS227}.
}
\seealso{
\code{\link{nmfkc.rank}}, \code{\link{nmfkc.ecv}},
\code{\link{nmfkc.consensus}}, \code{\link{nmfkc.ard}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{nmfkc.net}
\alias{nmfkc.net}
\title{Symmetric NMF for networks (tri / bi / signed)}
\usage{
nmfkc.net(
  Y,
  rank = 2,
  type = c("tri", "bi", "signed"),
  epsilon = 1e-04,
  maxit = 5000,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{Y}{Symmetric (N x N) \strong{non-negative} adjacency matrix.
\code{NA} entries are automatically treated as masked edges
(equivalent to supplying \code{Y.weights} with 0 at those positions);
see the note on \code{Y.weights} below.}

\item{rank}{Integer Q.}

\item{type}{\code{"tri"} (default), \code{"bi"}, or \code{"signed"}.}

\item{epsilon, maxit, verbose}{Standard.}

\item{...}{Hidden options: \code{nstart} (default 1; see note below),
\code{seed} (default 123), \code{X.restriction}, \code{X.init},
\code{C.init} (tri only) or \code{Cp.init}/\code{Cn.init} (signed only),
\code{Y.weights}, \code{C.L1} (tri only), \code{X.L2.ortho}, \code{prefix}.

\strong{\code{Y.weights}} is an optional non-negative N x N weight
matrix (symmetric, same shape as \code{Y}).  When supplied, the loss
becomes \eqn{\sum W_{ij} \, (Y_{ij} - \hat Y_{ij})^2}
(\code{\link[stats]{lm}}()-style, \strong{linear} in \eqn{W}).
Logical matrices (\code{TRUE} / \code{FALSE}) are also accepted.
Typical usage by \code{\link{nmfkc.net.ecv}} is a binary mask
(\eqn{W \in \{0,1\}}) holding out test edges on the upper triangle;
real-valued weights for edge-level importance weighting are also
supported.  If \code{Y.weights} is \code{NULL} (default) and \code{Y}
contains \code{NA}, a binary mask is auto-generated (0 at \code{NA}
positions, 1 elsewhere), and the \code{NA} entries in \code{Y} are
replaced by 0 so the multiplicative updates can proceed.

\strong{\code{X.init}} controls the initialization of the N x Q basis
matrix \eqn{X}.  Accepted values:
\itemize{
\item \code{"kmeans"} (default): k-means on the rows of \eqn{Y}
(equivalently columns, since \eqn{Y} is symmetric); the Q
cluster centers become the columns of \eqn{X}.  Each node is
treated as an N-dimensional connectivity profile, so clusters
correspond to nodes with similar neighborhood structure --
essentially a fast proxy for spectral clustering (Kuang, Yun &
Park 2015, \emph{SymNMF}).  Scales well and is the recommended
default for network data.
\item \code{"kmeansar"}: \code{"kmeans"} followed by filling zero
entries of \eqn{X} with \eqn{\mathrm{Uniform}(0, \bar Y / 100)}
to escape trivial stationary points.
\item \code{"nndsvd"}: Non-negative Double SVD with additive
randomness (NNDSVDar).  Requires a full SVD of \eqn{Y}, so for
very large networks (N > a few thousand) \code{"kmeans"} is
preferable.
\item \code{"runif"}: Uniform random entries in \eqn{[0, 1]}.
\item \code{"random"}: Legacy default (pre-v0.6.8), equivalent to
\code{abs(rnorm(N * Q)) * 0.1}.  Kept for backward compatibility.
\item A numeric N x Q matrix supplied by the user (used as-is).
}
When \code{nstart > 1}, each restart uses a distinct seed so that
k-means / runif / NNDSVDar produce different candidate initial
values across the multi-start loop.

\strong{Multi-start recommendation.} For \code{type = "signed"} the
\eqn{C = C_{+} - C_{-}} bottleneck can take both positive and negative
values, so the objective has more local minima than for \code{"tri"}
or \code{"bi"}.  A larger \code{nstart} (e.g., 10-50) is recommended
during exploration to reduce the chance of being trapped at a
suboptimal stationary point.  The default 1 is intended for fast
development; raise for publication-grade runs.}
}
\value{
Object of class \code{c("nmfkc.net.<type>", "nmfkc.net", "nmfkc")}.
For \code{type = "signed"} the return also carries \code{$Cp, $Cn}.
}
\description{
Single entry point for symmetric NMF of network data with correct
multiplicative updates.  Three model types are supported via \code{type}:
\itemize{
\item \strong{tri} (\code{type="tri"}, default):
\eqn{Y \approx X C X^\top} with \eqn{X, C \ge 0} (both non-negative;
\eqn{C} symmetric by design).  Uses Frobenius-full bilateral gradient.
\item \strong{bi} (\code{type="bi"}):
\eqn{Y \approx X X^\top} (\eqn{C} fixed to \eqn{I_Q}),
cube-root damping (He et al. 2011).
\item \strong{signed} (\code{type="signed"}):
\eqn{Y \approx X (C_{+} - C_{-}) X^\top} with \eqn{X \ge 0} and signed
\eqn{C = C_{+} - C_{-}}.  Preserves the soft-clustering interpretation
of \eqn{X} while allowing negative off-diagonals of \eqn{C}
(inter-cluster \emph{repulsion}).
}

\strong{Non-negative adjacency matrix assumption.} All three types
assume \eqn{Y \ge 0} (a non-negative adjacency/affinity matrix).
The qualifier \dQuote{signed} in \code{type = "signed"} refers to the
middle coefficient \eqn{C}, not to \eqn{Y} itself.  The underlying
Ding, Li & Jordan (2010) sign-splitting updates require \eqn{Y \ge 0}
to guarantee monotone descent; supplying a signed \eqn{Y} triggers an
error.  For a signed data matrix, see \code{\link{nmfkc.signed}}.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\seealso{
\code{\link{nmfkc.net.ecv}}, \code{\link{nmfkc.net.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.ard.R
\name{print.nmfkc.ard}
\alias{print.nmfkc.ard}
\title{Print method for nmfkc.ard objects}
\usage{
\method{print}{nmfkc.ard}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmfkc.ard"}.}

\item{...}{Unused.}
}
\value{
\code{x}, invisibly.
}
\description{
Print method for nmfkc.ard objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{plot.nmfae.ecv}
\alias{plot.nmfae.ecv}
\title{Plot method for nmfae.ecv objects}
\usage{
\method{plot}{nmfae.ecv}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmfae.ecv"} returned by \code{\link{nmfae.ecv}}.}

\item{...}{Additional graphical parameters (currently unused).}
}
\value{
Invisible \code{NULL}. Called for its side effect of producing a plot.
}
\description{
Visualizes element-wise cross-validation results.
When \code{rank.encoder} was \code{NULL} (paired), a line plot of sigma vs rank is drawn.
When \code{rank.encoder} was explicitly specified (grid), a heatmap of sigma over the (rank, rank.encoder) grid is drawn.
}
\seealso{
\code{\link{nmfae.ecv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{summary.nmfae.signed}
\alias{summary.nmfae.signed}
\title{Summary method for nmfae.signed}
\usage{
\method{summary}{nmfae.signed}(object, ...)
}
\arguments{
\item{object}{An \code{nmfae.signed} object.}

\item{...}{Unused.}
}
\value{
An object of class \code{"summary.nmfae.signed"}.
}
\description{
Produces a summary with dimensions, convergence, fit statistics,
and structure diagnostics (sparsity and negative-mass ratio).
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae.signed}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.ecv}
\alias{nmfkc.ecv}
\title{Perform Element-wise Cross-Validation (Wold's CV)}
\usage{
nmfkc.ecv(Y, A = NULL, rank = 1:3, data, ...)
}
\arguments{
\item{Y}{Observation matrix, or a formula (see \code{\link{nmfkc}} for Formula Mode).}

\item{A}{Covariate matrix. Ignored when \code{Y} is a formula.}

\item{rank}{Vector of ranks to evaluate (e.g., 1:5). For backward compatibility,
\code{Q} is accepted via \code{...}.}

\item{data}{A data frame (required when \code{Y} is a formula with column names).}

\item{...}{Additional arguments passed to \code{\link{nmfkc}} (e.g., \code{method="EU"}).
Also accepts: \code{nfolds} (number of folds, default 5; \code{div} also accepted),
\code{seed} (integer seed, default 123).}
}
\value{
A list with components:
\item{objfunc}{Numeric vector containing the Mean Squared Error (MSE) for each Q.}
\item{sigma}{Numeric vector containing the Residual Standard Error (RMSE) for each Q. Only available if method="EU".}
\item{objfunc.fold}{List of length equal to Q vector. Each element contains the MSE values for the k folds.}
\item{folds}{A list of length \code{div}, containing the linear indices of held-out elements for each fold (shared across all Q).}
}
\description{
\code{nmfkc.ecv} performs k-fold cross-validation by randomly holding out
individual elements of the data matrix (element-wise), assigning them a
weight of 0 via \code{Y.weights}, and evaluating the reconstruction error on
those held-out elements.

This method (also known as Wold's CV) is theoretically robust for determining
the optimal rank (Q) in NMF. This function supports vector input for \code{Q},
allowing simultaneous evaluation of multiple ranks on the same folds.

For symmetric (network) data use \code{\link{nmfkc.net.ecv}}, which
creates upper-triangle folds to prevent information leakage through
the symmetric entries \eqn{Y_{ij} = Y_{ji}}.  Passing the old
\code{Y.symmetric} argument here is no longer supported and stops
with a redirect message.
}
\examples{
# Element-wise CV to select rank
Y <- t(iris[1:30, 1:4])
res <- nmfkc.ecv(Y, rank = 1:2, nfolds = 3)
res$objfunc

}
\references{
Wold, S. (1978).  Cross-validatory estimation of the number of
components in factor and principal components models.
\emph{Technometrics}, 20(4), 397--405.
\doi{10.1080/00401706.1978.10489693}
Owen, A. B., & Perry, P. O. (2009).  Bi-cross-validation of the SVD
and the nonnegative matrix factorization.  \emph{Ann. Appl. Stat.}
3(2), 564--594. \doi{10.1214/08-AOAS227}  (cross-validation of the
NMF rank; see also \code{\link{nmfkc.bicv}}).
}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfkc.cv}}; other rank-selection
criteria: \code{\link{nmfkc.rank}}, \code{\link{nmfkc.bicv}},
\code{\link{nmfkc.consensus}}, \code{\link{nmfkc.ard}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{nmfkc.net.ecv}
\alias{nmfkc.net.ecv}
\title{Element-wise cross-validation for nmfkc.net (upper-triangle folds)}
\usage{
nmfkc.net.ecv(Y, rank = 1:3, type = c("tri", "bi", "signed"), ...)
}
\arguments{
\item{Y}{Symmetric N x N non-negative matrix.}

\item{rank}{Integer vector of ranks to evaluate. Default \code{1:3}.}

\item{type}{Model type: \code{"tri"} (default), \code{"bi"}, or \code{"signed"}.}

\item{...}{Passed to the underlying fitter; also accepts \code{nfolds}
(default 5; \code{div} alias), \code{seed} (default 123).}
}
\value{
A list with \code{objfunc}, \code{sigma}, \code{objfunc.fold},
\code{folds}, \code{Q.grid}, \code{type}.
}
\description{
k-fold CV with folds taken over the upper triangle of the
symmetric \eqn{Y} (mirrored to the lower triangle) to prevent information
leakage through symmetry.  A single entry point covers all three symmetric
NMF variants; \code{type} selects the fitting function:
\itemize{
\item \code{"tri"} (default): \code{\link{nmfkc.net}} with \eqn{C \ge 0}
\item \code{"bi"}: \code{\link{nmfkc.net}} with \eqn{C = I_Q}
\item \code{"signed"}: \code{\link{nmfkc.net}} with signed \eqn{C = C_{+} - C_{-}}
}
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\seealso{
\code{\link{nmfkc.net}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.cv}
\alias{nmfkc.cv}
\title{Perform k-fold cross-validation for NMF with kernel covariates}
\usage{
nmfkc.cv(Y, A = NULL, rank = 2, data, ...)
}
\arguments{
\item{Y}{Observation matrix, or a formula (see \code{\link{nmfkc}} for Formula Mode).}

\item{A}{Covariate matrix. If \code{NULL}, the identity matrix is used.
Ignored when \code{Y} is a formula.}

\item{rank}{Rank of the basis matrix \eqn{X}. Default is 2.}

\item{data}{A data frame (required when \code{Y} is a formula with column names).}

\item{...}{Additional arguments controlling CV and the internal \code{\link{nmfkc}} call:
\describe{
\item{\code{Y.weights}}{Non-negative weight matrix or vector
(\code{lm()}-style: loss \eqn{\sum W \, r^2}).  Binary
\code{{0,1}} masks (\code{TRUE} / \code{FALSE} also accepted)
are the typical ECV usage -- 0/\code{FALSE} excludes an element.
See \code{\link{nmfkc}} for full details.}
\item{\code{div}}{Number of folds (\eqn{k}); default: \code{5}.}
\item{\code{seed}}{Integer seed for reproducible partitioning; default: \code{123}.}
\item{\code{shuffle}}{Logical. If \code{TRUE} (default), randomly shuffles samples (standard CV);
if \code{FALSE}, splits sequentially (block CV; recommended for time series).}
\item{\code{Q}}{(Deprecated) Alias for \code{rank}.}
\item{\emph{Arguments passed to} \code{\link{nmfkc}}}{e.g., \code{gamma} (\code{B.L1}), \code{epsilon},
\code{maxit}, \code{method} (\code{"EU"} or \code{"KL"}), \code{X.restriction}, \code{X.init}, etc.}
}}
}
\value{
A list with components:
\describe{
\item{\code{objfunc}}{Mean loss per valid entry over all folds (MSE for \code{method="EU"}).}
\item{\code{sigma}}{Residual standard error (RMSE). Available only if \code{method="EU"}; on the same scale as \code{Y}.}
\item{\code{objfunc.block}}{Loss for each fold.}
\item{\code{block}}{Vector of fold indices (1, …, \code{div}) assigned to each column of \eqn{Y}.}
}
}
\description{
\code{nmfkc.cv} performs k-fold cross-validation for the tri-factorization model
\eqn{Y \approx X C A = X B}, where
\itemize{
\item \eqn{Y(P,N)} is the observation matrix,
\item \eqn{A(R,N)} is the covariate (or kernel) matrix,
\item \eqn{X(P,Q)} is the basis matrix,
\item \eqn{C(Q,R)} is the parameter matrix, and
\item \eqn{B(Q,N)} is the coefficient matrix (\eqn{B = C A}).
}
Given \eqn{Y} (and optionally \eqn{A}), \eqn{X} and \eqn{C} are fitted on each
training split and predictive performance is evaluated on the held-out split.
}
\examples{
# Example 1 (with explicit covariates):
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(1, cars$speed)
res <- nmfkc.cv(Y, A, rank = 1)
res$objfunc

# Example 2 (kernel A and beta sweep):
Y <- matrix(cars$dist, nrow = 1)
U <- matrix(c(5, 10, 15, 20, 25), nrow = 1)
V <- matrix(cars$speed, nrow = 1)
betas <- 25:35/1000
obj <- numeric(length(betas))
for (i in seq_along(betas)) {
  A <- nmfkc.kernel(U, V, beta = betas[i])
  obj[i] <- nmfkc.cv(Y, A, rank = 1, nfolds = 10)$objfunc
}
betas[which.min(obj)]

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfkc.kernel.beta.cv}}, \code{\link{nmfkc.ar.degree.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{nmfkc.net.DOT}
\alias{nmfkc.net.DOT}
\title{Generate a Graphviz DOT Diagram for a Symmetric NMF Network}
\usage{
nmfkc.net.DOT(
  result,
  threshold = 0.01,
  sig.level = 0.1,
  weight_scale = 5,
  weight_scale_xy = 1,
  weight_scale_xx = weight_scale,
  rankdir = "TB",
  fill = TRUE,
  hide.isolated = TRUE,
  Y.label = NULL,
  X.label = NULL,
  Y.title = "Nodes (Y)",
  X.title = "Basis (X)",
  show.theta = NULL,
  signed = inherits(result, "nmfkc.net.signed"),
  cluster.box = c("none", "normal", "faint", "invisible"),
  layout = c("neato", "fdp", "twopi", "circo", "dot"),
  X.color = NULL,
  Y.cluster = c("soft", "hard")
)
}
\arguments{
\item{result}{A list returned by \code{nmfkc()} with
\code{Y.symmetric = "bi"} or \code{"tri"}, or the newer
\code{\link{nmfkc.net}} with \code{type = "bi"} or \code{"tri"}.
If inference results are present (from \code{\link{nmfkc.net.inference}}),
C edges are decorated with significance stars.}

\item{threshold}{Minimum coefficient value to display an edge.}

\item{sig.level}{Significance level for filtering C edges (if inference
results are present). Set to \code{NULL} to show all edges above threshold.}

\item{weight_scale}{Base scaling factor for edge widths.}

\item{weight_scale_xy}{Scaling factor for X edges (basis -> node).}

\item{weight_scale_xx}{Scaling factor for C edges (basis <-> basis).}

\item{rankdir}{Graphviz rank direction (\code{"TB"} or \code{"LR"}).}

\item{fill}{Logical; whether nodes are filled with color.}

\item{hide.isolated}{Logical; if TRUE, omit outer nodes with no X edge
above threshold.}

\item{Y.label}{Character vector of labels for outer nodes.}

\item{X.label}{Character vector of labels for basis nodes.}

\item{Y.title}{Cluster title for outer nodes.}

\item{X.title}{Cluster title for basis nodes.}

\item{show.theta}{Logical or NULL. Whether to draw C edges
between basis nodes. NULL = auto-detect (TRUE for tri, FALSE for bi).}

\item{signed}{Logical. If \code{TRUE}, \eqn{C} is treated as a signed
matrix: positive entries rendered as solid edges and negative as
dashed, with edge visibility threshold on \eqn{|C|}.  Default is
\code{inherits(result, "nmfkc.net.signed")} so that results from
\code{\link{nmfkc.net}} are auto-detected.}

\item{cluster.box}{Style of cluster box: \code{"none"}, \code{"normal"},
\code{"faint"}, \code{"invisible"}.}

\item{layout}{Graphviz layout engine, in recommended order:
\code{"neato"} (default; spring model, clearest for small/medium community
graphs), \code{"fdp"} (force-directed, scales to larger graphs),
\code{"twopi"} (radial), \code{"circo"} (circular), \code{"dot"}
(hierarchical).  For community networks, \code{"neato"} or \code{"fdp"} with
a raised \code{threshold} (e.g.\ 0.2--0.3) separate the groups best.}

\item{X.color}{Color palette for basis nodes (length Q).}

\item{Y.cluster}{Coloring mode for outer nodes: \code{"soft"} (weighted mix)
or \code{"hard"} (most probable basis color).}
}
\value{
A character string of class \code{c("nmfkc.net.DOT", "nmfkc.DOT")}
representing a valid Graphviz DOT script.
Use \code{plot()} to render (requires DiagrammeR).
}
\description{
Creates a Graphviz DOT script that visualizes the two-layer structure
of a symmetric NMF model (\eqn{Y \approx X C X^\top}).

The resulting diagram displays:
\itemize{
\item outer nodes: the original network nodes (rows/columns of Y),
\item inner nodes: the latent basis/group nodes (columns of X),
\item directed edges from basis to node: membership weights from X,
\item undirected edges between basis nodes: inter-group interactions
from C matrix (for tri-symmetric models only).
}
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\examples{
\donttest{
library(nmfkc)
Y <- matrix(c(0,1,1,0,0,0,
              1,0,1,0,0,0,
              1,1,0,1,0,0,
              0,0,1,0,1,1,
              0,0,0,1,0,1,
              0,0,0,1,1,0), 6, 6)
res <- nmfkc.net(Y, rank = 2, type = "tri", nstart = 20)
dot <- nmfkc.net.DOT(res)
plot(dot)
}

}
\seealso{
\code{\link{nmfkc.net.inference}}, \code{\link{nmfkc.DOT}},
\code{\link{plot.nmfkc.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.signed.R
\name{predict.nmfkc.signed}
\alias{predict.nmfkc.signed}
\title{Predict method for nmfkc.signed}
\usage{
\method{predict}{nmfkc.signed}(object, newA = NULL, type = c("response", "prob", "class"), ...)
}
\arguments{
\item{object}{A fitted \code{"nmfkc.signed"} object.}

\item{newA}{Real-valued \eqn{D \times N_{\mathrm{new}}} covariate matrix.}

\item{type}{Output: \code{"response"} (raw signed), \code{"prob"},
or \code{"class"}.}

\item{...}{Unused.}
}
\value{
A numeric matrix (\code{"response"} or \code{"prob"}) or a
character vector (\code{"class"}).
}
\description{
Computes \eqn{\widehat Y = X \, C \, A_{\mathrm{new}}}
(\eqn{= X (C_{+} - C_{-})(A_{+}^{\mathrm{new}} - A_{-}^{\mathrm{new}})}).
For \code{type = "response"} the raw prediction is returned
(possibly signed).  For \code{type = "prob"} and \code{"class"},
negative entries of \eqn{\widehat Y} are clipped to zero before
column normalization, since probabilities must be non-negative.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{plot.nmfae.cv}
\alias{plot.nmfae.cv}
\title{Plot method for nmfae.cv objects}
\usage{
\method{plot}{nmfae.cv}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmfae.cv"} returned by \code{\link{nmfae.cv}}.}

\item{...}{Additional graphical parameters passed to \code{barplot}.}
}
\value{
Invisible \code{NULL}. Called for its side effect (plot).
}
\description{
Displays a bar chart of per-fold cross-validation errors from
\code{\link{nmfae.cv}}. The overall RMSE (sigma) is shown in the title.
}
\seealso{
\code{\link{nmfae.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.signed.R
\name{nmfkc.signed.cv}
\alias{nmfkc.signed.cv}
\title{Column-wise k-fold cross-validation for nmfkc.signed}
\usage{
nmfkc.signed.cv(Y, A, rank = 2, ...)
}
\arguments{
\item{Y}{Real-valued \eqn{Q_{\mathrm{obs}} \times N} response matrix
(signed entries allowed).}

\item{A}{Real-valued \eqn{D \times N} covariate matrix (signed).}

\item{rank}{Integer \eqn{Q}.}

\item{...}{Passed to \code{\link{nmfkc.signed}}; also accepts
\code{nfolds} (default 5; \code{div} alias), \code{seed}
(default 123), \code{shuffle} (default \code{TRUE}).}
}
\value{
A list with \code{objfunc} (mean squared prediction error),
\code{sigma} (RMSE), \code{objfunc.block} (per-fold MSE vector),
\code{block} (integer fold assignment of length \eqn{N}).  Field
names match \code{\link{nmfkc.cv}}.
}
\description{
Column-wise k-fold CV by held-out samples: for each
fold, the model is fit on the training columns and evaluated on the
held-out columns by solving for new-sample coefficients via a
weighted refit with \eqn{X} fixed.
}
\section{Lifecycle}{

This function is \strong{experimental}.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfkc.signed}}, \code{\link{nmfkc.signed.ecv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfre.R
\name{nmfre}
\alias{nmfre}
\title{Non-negative Matrix Factorization with Random Effects}
\usage{
nmfre(
  Y,
  A = NULL,
  rank = 2,
  df.rate = NULL,
  wild.bootstrap = TRUE,
  epsilon = 1e-05,
  maxit = 5000,
  ...
)
}
\arguments{
\item{Y}{Observation matrix (P x N), non-negative.}

\item{A}{Covariate matrix (K x N). Default is a row of ones (intercept only).}

\item{rank}{Integer. Rank of the basis matrix \eqn{X}. Default is 2.
For backward compatibility, \code{Q} is accepted via \code{...}.}

\item{df.rate}{Rate for computing the dfU cap (\code{cap = rate * N * Q}).
For backward compatibility, \code{dfU.cap.rate} is accepted via \code{...}.
If \code{NULL} (default), runs \code{\link{nmfre.dfU.scan}} internally and
selects the minimum rate where the cap is not binding.
Use \code{\link{nmfre.dfU.scan}} beforehand to examine dfU behavior across rates
and choose an appropriate value.}

\item{wild.bootstrap}{Logical. If \code{TRUE} (default), perform wild bootstrap inference
on \eqn{\Theta}.}

\item{epsilon}{Convergence tolerance for relative change in objective (default 1e-5).}

\item{maxit}{Maximum number of iterations.  Default \code{5000}
(matches \code{\link{nmfkc}} and the other MU functions in the
package).  When the cap is hit without meeting the relative-
tolerance criterion, a \code{"maximum iterations (...) reached..."}
warning is emitted so users notice unconverged fits.}

\item{...}{Additional arguments for initialization, variance control, dfU control,
optimization, and inference settings.
\itemize{
\item \code{X.init}: Initial basis matrix (P x Q), or \code{NULL}.
When \code{NULL}, \code{\link{nmfkc}} is called internally to generate initial values.
\item \code{C.init}: Initial coefficient matrix (Q x K), or \code{NULL}.
When \code{NULL}, \code{\link{nmfkc}} is called internally to generate initial values.
\item \code{U.init}: Initial random effects matrix (Q x N), or \code{NULL} (all zeros).
\item \code{prefix}: Prefix for basis names (default \code{"Basis"}).
\item \code{sigma2}: Initial residual variance (default 1).
\item \code{sigma2.update}: Logical. Update \eqn{\sigma^2} during iterations (default \code{TRUE}).
\item \code{tau2}: Initial random effect variance (default 1).
\item \code{tau2.update}: Logical. Update \eqn{\tau^2} by moment matching (default \code{TRUE}).
Disabled when dfU cap is active.
\item \code{dfU.control}: Either \code{"cap"} (default) to enforce a cap on dfU,
or \code{"off"} for no cap.
\item \code{print.trace}: Logical. If \code{TRUE}, print progress every 100 iterations (default \code{FALSE}).
\item \code{seed}: Integer seed for reproducibility (default 1).
\item \code{C.p.side}: P-value sidedness: \code{"one.sided"} (default, for boundary null
H0: C=0 vs H1: C>0) or \code{"two.sided"}.
\item \code{wild.B}: Number of wild bootstrap replicates (default 500).
\item \code{wild.seed}: Seed for wild bootstrap (default 123).
}}
}
\value{
A list of class \code{"nmfre"} with components.
The model is \eqn{Y = X(\Theta A + U) + \mathcal{E}}.

\strong{Core matrices}
\describe{
\item{\code{X}}{Basis matrix \eqn{X} (\eqn{P \times Q}), columns normalized to sum to 1.}
\item{\code{C}}{Coefficient matrix \eqn{\Theta} (\eqn{Q \times K}).}
\item{\code{U}}{Random effects matrix \eqn{U} (\eqn{Q \times N}).}
}

\strong{Variance components}
\describe{
\item{\code{sigma2}}{Residual variance \eqn{\hat{\sigma}^2}.}
\item{\code{tau2}}{Random effect variance \eqn{\hat{\tau}^2}.}
\item{\code{lambda}}{Ridge penalty \eqn{\lambda = \sigma^2 / \tau^2}.}
}

\strong{Convergence diagnostics}
\describe{
\item{\code{converged}}{Logical. Whether the algorithm converged.}
\item{\code{stop.reason}}{Character string describing why iteration stopped.}
\item{\code{iter}}{Number of iterations performed.}
\item{\code{maxit}}{Maximum iterations setting used.}
\item{\code{epsilon}}{Convergence tolerance used.}
\item{\code{objfunc}}{Final objective function value
\eqn{\|Y - X(\Theta A + U)\|^2 + \lambda \|U\|^2}.}
\item{\code{rel.change.final}}{Final relative change in objective.}
\item{\code{objfunc.iter}}{Numeric vector of objective values per iteration.}
\item{\code{rss.trace}}{Numeric vector of \eqn{\|Y - X(\Theta A + U)\|^2} per iteration.}
}

\strong{Effective degrees of freedom (dfU) diagnostics}
\describe{
\item{\code{dfU}}{Final effective degrees of freedom
\eqn{\mathrm{df}_U = N \sum_q d_q / (d_q + \lambda)},
where \eqn{d_q} are eigenvalues of \eqn{X'X}.}
\item{\code{dfU.cap}}{Upper bound imposed on \eqn{\mathrm{df}_U}.}
\item{\code{dfU.cap.rate}}{Rate used to compute the cap.}
\item{\code{dfU.cap.scan}}{Result of \code{\link{nmfre.dfU.scan}}, or \code{NULL}.}
\item{\code{lambda.enforced}}{Final \eqn{\lambda} enforced to satisfy the cap.}
\item{\code{dfU.hit.cap}}{Logical. Whether the cap was binding.}
\item{\code{dfU.hit.iter}}{Iteration at which the cap first bound.}
\item{\code{dfU.frac}}{\eqn{\mathrm{df}_U / (NQ)}, fraction of maximum df.}
\item{\code{dfU.cap.frac}}{\eqn{\mathrm{df}_U^{\mathrm{cap}} / (NQ)}.}
}

\strong{Fitted matrices}
\describe{
\item{\code{B}}{Fixed-effect scores \eqn{\Theta A} (\eqn{Q \times N}).}
\item{\code{B.prob}}{Column-normalized probabilities from
\eqn{\max(\Theta A, 0)}.}
\item{\code{B.blup}}{BLUP scores \eqn{\Theta A + U} (\eqn{Q \times N}).}
\item{\code{B.blup.pos}}{Non-negative BLUP scores
\eqn{\max(\Theta A + U, 0)} (\eqn{Q \times N}).}
\item{\code{B.blup.prob}}{Column-normalized probabilities from
\eqn{\max(\Theta A + U, 0)}.}
\item{\code{XB}}{Fitted values from fixed effects
\eqn{X \Theta A} (\eqn{P \times N}).}
\item{\code{XB.blup}}{Fitted values including random effects
\eqn{X(\Theta A + U)} (\eqn{P \times N}).}
}

\strong{Fit statistics}
\describe{
\item{\code{r.squared}}{Pearson \eqn{\mathrm{cor}(Y, X(\Theta A + U))^2}
(BLUP prediction).}
\item{\code{r.squared.uncentered}}{Uncentered
\eqn{1 - \|Y - X(\Theta A + U)\|_F^2 / \|Y\|_F^2} (BLUP;
baseline = zero matrix).}
\item{\code{r.squared.centered}}{Row-mean centered
\eqn{1 - \|Y - X(\Theta A + U)\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2}
(BLUP; baseline = per-row mean).}
\item{\code{r.squared.fixed}}{Pearson \eqn{\mathrm{cor}(Y, X\Theta A)^2}
(fixed-only prediction).}
\item{\code{r.squared.fixed.uncentered}, \code{r.squared.fixed.centered}}{Uncentered
and centered \eqn{R^2} for the fixed-only prediction.}
\item{\code{ICC}}{Trace-based Intraclass Correlation Coefficient.
In the NMF-RE model, the conditional covariance of the \eqn{n}-th
observation column is
\eqn{\mathrm{Var}(Y_n) = \tau^2 X X^\top + \sigma^2 I_P},
a \eqn{P \times P} matrix. Unlike a standard random intercept model
where the design matrix \eqn{Z} is a simple indicator (so the ICC
reduces to \eqn{\tau^2 / (\sigma^2 + \tau^2)}), the basis matrix
\eqn{X} plays the role of \eqn{Z} in a random slopes model,
making the variance contribution of \eqn{U} depend on \eqn{X}.
To obtain a scalar summary, we take the trace of each component:
\deqn{\mathrm{ICC} = \frac{\tau^2 \, \mathrm{tr}(X^\top X)}
      {\tau^2 \, \mathrm{tr}(X^\top X) + \sigma^2 P}.}
This equals the average (over \eqn{P} dimensions) proportion of
per-column variance attributable to the random effects.}
}

\strong{Inference on \eqn{\Theta} (wild bootstrap)}
\describe{
\item{\code{sigma2.used}}{\eqn{\hat{\sigma}^2} used for inference.}
\item{\code{C.vec.cov}}{Variance-covariance matrix for
\eqn{\mathrm{vec}(\Theta)} (\eqn{QK \times QK}).}
\item{\code{C.se}}{Standard error matrix for \eqn{\Theta} (\eqn{Q \times K}).}
\item{\code{C.se.hess}}{Sandwich (Hessian-based) SE matrix for \eqn{\Theta}.}
\item{\code{C.se.boot}}{Bootstrap SE matrix for \eqn{\Theta}.}
\item{\code{coefficients}}{Data frame with columns Estimate, Std. Error, z value,
Pr(>|z|), and confidence interval bounds for each element of \eqn{\Theta}.}
\item{\code{C.ci.lower}}{Lower confidence interval matrix for \eqn{\Theta} (\eqn{Q \times K}).}
\item{\code{C.ci.upper}}{Upper confidence interval matrix for \eqn{\Theta} (\eqn{Q \times K}).}
\item{\code{C.boot.sd}}{Bootstrap standard deviation matrix for \eqn{\Theta} (\eqn{Q \times K}).}
\item{\code{C.p.side}}{P-value sidedness used: \code{"one.sided"} or \code{"two.sided"}.}
}
}
\description{
Estimates the NMF-RE model
\deqn{Y = X(\Theta A + U) + \mathcal{E}}
where \eqn{Y} (\eqn{P \times N}) is a non-negative observation matrix,
\eqn{X} (\eqn{P \times Q}) is a non-negative basis matrix learned from the data,
\eqn{\Theta} (\eqn{Q \times K}) is a non-negative coefficient matrix capturing
systematic covariate effects on latent scores,
\eqn{A} (\eqn{K \times N}) is a covariate matrix, and
\eqn{U} (\eqn{Q \times N}) is a random effects matrix capturing
unit-specific deviations in the latent score space.

NMF-RE can be viewed as a mixed-effects latent-variable model defined on a
reconstruction (mean) structure. The non-negativity constraint on \eqn{X}
induces sparse, parts-based loadings, achieving measurement-side variable
selection without an explicit sparsity penalty. Inference on \eqn{\Theta}
provides covariate-side variable selection by identifying which covariates
significantly affect which components.

Estimation alternates ridge-type BLUP-like closed-form updates for \eqn{U}
with multiplicative non-negative updates for \eqn{X} and \eqn{\Theta}.
The effective degrees of freedom consumed by \eqn{U} are monitored and a
df-based cap can be enforced to prevent near-saturated fits.

When \code{wild.bootstrap = TRUE}, inference on \eqn{\Theta} is performed
conditional on \eqn{(\hat{X}, \hat{U})} via asymptotic linearization,
a one-step Newton update, and a multiplier (wild) bootstrap, yielding
standard errors, z-values, p-values, and confidence intervals without
repeated constrained re-optimization.
}
\examples{
# Example 1. cars data
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(intercept = 1, speed = cars$speed)
res <- nmfre(Y, A, rank = 1, maxit = 5000)
summary(res)

\donttest{
# Example 2. Orthodont data (nlme)
if (requireNamespace("nlme", quietly = TRUE)) {
  Y <- matrix(nlme::Orthodont$distance, 4, 27)
  male <- ifelse(nlme::Orthodont$Sex[seq(1, 108, 4)] == "Male", 1, 0)
  A <- rbind(intercept = 1, male = male)

  # Scan dfU cap rates to choose an appropriate value
  nmfre.dfU.scan(1:10/10, Y, A, rank = 1)

  # Fit with chosen rate
  res <- nmfre(Y, A, rank = 1, df.rate = 0.2)
  summary(res)
}}

}
\references{
Satoh, K. (2026). Wild Bootstrap Inference for Non-Negative Matrix
Factorization with Random Effects. arXiv:2603.01468.
\url{https://arxiv.org/abs/2603.01468}
}
\seealso{
\code{\link{nmfre.inference}}, \code{\link{nmfre.dfU.scan}},
\code{\link{nmfkc.DOT}}, \code{\link{summary.nmfre}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.class}
\alias{nmfkc.class}
\title{Create a class (one-hot) matrix from a categorical vector}
\usage{
nmfkc.class(x)
}
\arguments{
\item{x}{A categorical vector or a factor.}
}
\value{
A binary matrix with one row per unique category and one column per observation. Each column has exactly one entry equal to 1, indicating the category of the observation.
}
\description{
\code{nmfkc.class} converts a categorical or factor vector into a class matrix
(one-hot encoded representation), where each row corresponds to a category
and each column corresponds to an observation.
}
\examples{
# Example.
Y <- nmfkc.class(iris$Species)
Y[,1:6]
}
\seealso{
\code{\link{nmfkc}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.residual.plot}
\alias{nmfkc.residual.plot}
\title{Plot Diagnostics: Original, Fitted, and Residual Matrices as Heatmaps}
\usage{
nmfkc.residual.plot(
  Y,
  result,
  fitted.palette = (grDevices::colorRampPalette(c("white", "orange", "red")))(256),
  residual.palette = (grDevices::colorRampPalette(c("blue", "white", "red")))(256),
  ...
)
}
\arguments{
\item{Y}{The original observation matrix (P x N).}

\item{result}{The result object returned by the nmfkc function.}

\item{fitted.palette}{A vector of colors for Y and XB heatmaps. Defaults to white-orange-red.
For backward compatibility, \code{Y_XB_palette} is accepted via \code{...}.}

\item{residual.palette}{A vector of colors for the residuals heatmap. Defaults to blue-white-red.
For backward compatibility, \code{E_palette} is accepted via \code{...}.}

\item{...}{Additional graphical parameters passed to the internal image calls.}
}
\value{
NULL. The function generates a plot.
}
\description{
This function generates a side-by-side plot of three heatmaps: the original
observation matrix Y, the fitted matrix XB (from NMF), and the residual matrix E (Y - XB).
This visualization aids in diagnosing whether the chosen rank Q is adequate
by assessing if the residual matrix E appears to be random noise.

The axis labels (X-axis: Samples, Y-axis: Features) are integrated into the main title of each plot
to maximize the plot area, reflecting the compact layout settings.
}
\examples{
Y <- t(iris[1:30, 1:4])
result <- nmfkc(Y, rank = 2)
nmfkc.residual.plot(Y, result)

}
\seealso{
\code{\link{nmfkc}}, \code{\link{residuals.nmf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{print.summary.nmfae.signed.inference}
\alias{print.summary.nmfae.signed.inference}
\title{Print method for summary.nmfae.signed.inference objects}
\usage{
\method{print}{summary.nmfae.signed.inference}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{An object of class \code{"summary.nmfae.signed.inference"}.}

\item{digits}{Minimum number of significant digits.}

\item{...}{Additional arguments (currently unused).}
}
\value{
Called for its side effect (printing). Returns \code{x} invisibly.
}
\description{
Prints the Signed-Bottleneck NMF-AE summary followed by the
coefficients table of Theta.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{summary.nmfae.signed.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.signed.R
\name{print.summary.nmfkc.signed}
\alias{print.summary.nmfkc.signed}
\title{Print method for summary.nmfkc.signed}
\usage{
\method{print}{summary.nmfkc.signed}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{Object of class \code{"summary.nmfkc.signed"}.}

\item{digits}{Number of significant digits.}

\item{...}{Unused.}
}
\value{
Invisible \code{x}.
}
\description{
Print method for summary.nmfkc.signed
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{print.nmf.inference}
\alias{print.nmf.inference}
\title{Print method for NMF inference objects}
\usage{
\method{print}{nmf.inference}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmf.inference"}.}

\item{...}{Additional arguments passed to the corresponding
\code{print.summary.*} method.}
}
\value{
Called for its side effect (printing). Returns \code{x} invisibly.
}
\description{
Prints a summary of any NMF inference result object
(\code{"nmfkc.inference"} or \code{"nmfae.inference"}).
}
\seealso{
\code{\link{nmfkc.inference}}, \code{\link{nmfae.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{nmfkc.net.rank}
\alias{nmfkc.net.rank}
\title{Rank selection for nmfkc.net (concise diagnostics)}
\usage{
nmfkc.net.rank(
  Y,
  rank = 1:5,
  type = c("tri", "bi", "signed"),
  detail = c("full", "fast"),
  plot = TRUE,
  ...
)
}
\arguments{
\item{Y}{Symmetric (network) observation matrix.}

\item{rank}{Integer vector of ranks to evaluate.}

\item{type}{One of \code{"tri"}, \code{"bi"}, \code{"signed"}.}

\item{detail}{\code{"full"} (default) also runs element-wise CV
(\code{sigma.ecv}); \code{"fast"} skips it (plots r.squared and
eff.rank only, and recommends the R-squared elbow).}

\item{plot}{Logical; draw the diagnostics plot (default \code{TRUE}).}

\item{...}{Passed on to \code{\link{nmfkc.net}} and
\code{\link{nmfkc.net.ecv}} (e.g.\ \code{nstart}, \code{maxit},
\code{nfolds}, \code{seed}).}
}
\value{
A list with \code{rank.best} (ECV minimum, or the R-squared
elbow under \code{detail = "fast"}) and \code{criteria} (data
frame: \code{rank}, \code{effective.rank}, \code{effective.rank.ratio},
\code{r.squared}, \code{sigma.ecv}).
}
\description{
Fits \code{\link{nmfkc.net}} across a range of ranks and reports the
three rank-selection criteria -- \code{r.squared}, the effective rank
(utilization), and the element-wise CV error \code{sigma.ecv} -- with
the same concise diagnostics plot as \code{\link{nmfkc.rank}}.
}
\examples{
\donttest{
Y <- matrix(c(0,1,1,0,0,0, 1,0,1,0,0,0, 1,1,0,1,0,0,
              0,0,1,0,1,1, 0,0,0,1,0,1, 0,0,0,1,1,0), 6, 6)
nmfkc.net.rank(Y, rank = 1:3, type = "tri", nstart = 5, nfolds = 3)
}
}
\references{
Roy, O., & Vetterli, M. (2007).  The effective rank: A measure of
effective dimensionality.  \emph{Proc. EUSIPCO}, 606--610.
(\code{effective.rank})
Wold, S. (1978).  Cross-validatory estimation of the number of
components in factor and principal components models.
\emph{Technometrics}, 20(4), 397--405. (\code{sigma.ecv})
}
\seealso{
\code{\link{nmfkc.net}}, \code{\link{nmfkc.net.ecv}},
\code{\link{nmfkc.rank}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{print.summary.nmfkc.net.signed}
\alias{print.summary.nmfkc.net.signed}
\title{Print method for summary.nmfkc.net.signed objects}
\usage{
\method{print}{summary.nmfkc.net.signed}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{An object of class \code{"summary.nmfkc.net.signed"}.}

\item{digits}{Minimum number of significant digits.}

\item{...}{Unused.}
}
\value{
Invisible \code{x}.
}
\description{
Print method for summary.nmfkc.net.signed objects
}
\seealso{
\code{\link{summary.nmfkc.net.signed}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.signed.R
\name{nmfkc.signed}
\alias{nmfkc.signed}
\title{NMF-KC with signed covariate matrix}
\usage{
nmfkc.signed(
  Y,
  A,
  rank = NULL,
  epsilon = 1e-04,
  maxit = 5000,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{Y}{Real-valued \eqn{Q_{\mathrm{obs}} \times N} response matrix.
Unlike \code{\link{nmfkc}}, negative entries are allowed.}

\item{A}{Real-valued \eqn{D \times N} covariate matrix (signed).
A single matrix is passed; its positive and negative parts
\eqn{A_{+} = \max(A, 0)} and \eqn{A_{-} = \max(-A, 0)} are computed
internally.  When using Random Fourier Features (Rahimi & Recht
2007) as \eqn{A}, supply the RFF parameters via the hidden
\code{pars} argument so that \code{predict()} can regenerate
features for new data (see \code{pars} entry in \code{\dots} below).}

\item{rank}{Integer.  Number of latent components \eqn{Q} in \eqn{X}.}

\item{epsilon}{Relative convergence tolerance on the objective
(default \code{1e-4}).}

\item{maxit}{Maximum number of iterations (default \code{5000}).}

\item{verbose}{Logical.  Print dimensions at start (default \code{TRUE}).}

\item{...}{Additional arguments:
\itemize{
\item \code{Q}: alias for \code{rank}.
\item \code{X.restriction}: constraint applied to columns of
\eqn{X} after every update, with the scale absorbed into
\eqn{C_{+}, C_{-}}.  One of \code{"colSums"} (default,
\eqn{\mathrm{colSums}(X) = 1}), \code{"colSqSums"},
\code{"totalSum"}, \code{"none"}, \code{"fixed"}.
\item \code{X.init}: initialization strategy for the basis matrix
\eqn{X} (\eqn{Q_{\mathrm{obs}} \times Q}).  Accepts the same
menu as \code{\link{nmfkc}}: \code{"kmeans"} (default),
\code{"kmeansar"}, \code{"nndsvd"}, \code{"runif"}, or a
user-supplied \eqn{Q_{\mathrm{obs}} \times Q} non-negative
numeric matrix.  String methods delegate to the shared
internal helper \code{.init_X_method()} (see \code{\link{nmfkc}}
for the definitions of each method).  For signed \eqn{Y},
\code{"kmeans"} cluster centers may contain negative entries;
they are clipped to zero to satisfy \eqn{X \ge 0}, and any
column that collapses to all-zeros is re-filled with small
\eqn{\mathrm{Uniform}(0, 0.1)} noise.
\item \code{C.init}: explicit initial \eqn{Q \times D} coefficient
matrix \eqn{\Theta} (signed).  Split internally.
\item \code{warm.start}: logical (default \code{TRUE}).  If
\code{TRUE} \strong{and \eqn{Y \ge 0}}, runs
\code{nmfkc(Y, A = rbind(A_+, A_-), rank = Q)} internally to
seed \eqn{X, C_{+}, C_{-}}.  The user's \code{X.init},
\code{seed}, \code{nstart}, and \code{X.restriction} are
forwarded to the internal \code{\link{nmfkc}} call so that
initialization choices propagate consistently between the
warm-start and the signed MU loop.  Ignored when \eqn{Y} has
negative entries (warm-start is disabled; \code{X.init} is
used directly by the signed branch instead).
\item \code{seed}: RNG seed for random initialization (default 123).
\item \code{prefix}: name prefix for rows of \eqn{C} and columns
of \eqn{X} (default \code{"Basis"}).
\item \code{pars}: optional list \code{list(omega, b, D, beta)}
of Random Fourier Feature parameters (Rahimi & Recht 2007;
\code{omega}: frequency matrix, \code{b}: phase offset,
\code{D}: feature dimension, \code{beta}: bandwidth).  When
supplied, it is stored in the returned object so that
\code{summary()} can report \eqn{\beta} and downstream
\code{predict()} calls can regenerate RFF features for new
data.  If \code{A} is not RFF features, leave this \code{NULL}.
\item \code{Y.weights}: Optional non-negative weight matrix
(\eqn{Q_{\mathrm{obs}} \times N}) or vector (length \eqn{N}),
analogous to the \code{weights} argument of
\code{\link[stats]{lm}}.  Loss becomes
\eqn{\sum W_{ij} \, (Y_{ij} - (XCA)_{ij})^2}
(\code{lm()}-style, \strong{linear} in \eqn{W}).  Logical
matrices (\code{TRUE} / \code{FALSE}) are also accepted.
Typical usage by \code{\link{nmfkc.signed.cv}} /
\code{\link{nmfkc.signed.ecv}} passes a binary mask
\eqn{W \in \{0,1\}} to hold out test elements; real-valued
weights for observation-level importance weighting are also
supported.  Default \code{NULL}: if \code{Y} has \code{NA},
a binary mask is auto-constructed (0 for \code{NA}, 1
elsewhere); otherwise no weighting.
\item \code{nstart}: number of random restarts.  \strong{Signed
models have more local minima than non-negative ones} because
\eqn{\Theta = C_{+} - C_{-}} can take both positive and negative
values.  Since \code{nmfkc.signed()} itself does not loop over
restarts (callers control it), set the outer-loop size via e.g.
running the function several times with different \code{seed}
and keeping the fit with the smallest \code{$objfunc}.  A
restart budget of 10-50 is recommended for publication-grade
runs on signed data.
}}
}
\value{
An object of class \code{c("nmfkc.signed", "nmfkc")} with
\itemize{
\item \code{X}: \eqn{Q_{\mathrm{obs}} \times Q} basis matrix (non-negative,
column-normalized according to \code{X.restriction}).
\item \code{Cp}, \code{Cn}: \eqn{Q \times D} non-negative parts of
\eqn{\Theta}, so that \eqn{\Theta = C_{+} - C_{-}}.
\item \code{C}: \eqn{C_{+} - C_{-}} (= \eqn{\Theta}), signed.
\item \code{B}: \eqn{C \, A}, \eqn{Q \times N} (signed).
\item \code{objfunc.iter}: objective values per iteration.
\item \code{objfunc}: final objective.
\item \code{r.squared}: \eqn{\mathrm{cor}(Y, \widehat Y)^2} (Pearson; in \eqn{[0,1]}).
\item \code{r.squared.uncentered}: uncentered \eqn{R^2 = 1 - \|Y - \widehat Y\|_F^2 / \|Y\|_F^2} (baseline = zero matrix).
\item \code{r.squared.centered}: row-mean centered \eqn{1 - \|Y - \widehat Y\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2}.
\item \code{mae}: mean absolute error.
\item \code{iter}: number of iterations performed.
\item \code{runtime}: elapsed seconds.
\item \code{Y.signed}: logical; whether \eqn{Y} contained negative
entries during fitting.
\item \code{pars}: RFF generating parameters, if supplied.
\item \code{call}: the matched call.
}
}
\description{
Solves
\deqn{Y \approx X\,\Theta\,A,\qquad X \ge 0,\;\Theta\in\R^{Q\times D},
  \;A\in\R^{D\times N},}
where the covariate matrix \eqn{A} and the coefficient matrix
\eqn{\Theta} may be \strong{signed}.  Internally \eqn{A = A_{+} - A_{-}}
and \eqn{\Theta = C_{+} - C_{-}} with \eqn{A_{\pm}, C_{\pm} \ge 0}
(sign-splitting trick, Ding et al. 2010), and the problem is solved
by a Direct Multiplicative Update algorithm whose iteration cost is
\eqn{O(Q D^2)}, independent of \eqn{N}.

Only \eqn{X} is structurally constrained to be non-negative (Semi-NMF
sense of Ding, Li, & Jordan 2010).  In particular, \strong{\eqn{Y} may
contain negative entries}, in which case the response is fit in the
least-squares sense without any non-negativity requirement on \eqn{Y}.

When \eqn{A \ge 0} (so \eqn{A_{-} = 0}), the result reduces to
\code{\link{nmfkc}}(Y, A, rank) with Euclidean loss, up to reordering.
}
\section{Lifecycle}{

This function is \strong{experimental}.  The interface may change in
future versions.
}

\examples{
\donttest{
set.seed(1)
## Example 1: signed A (e.g., hand-built RFF features), non-negative Y
## Build simple signed features Z = sqrt(2/D) * cos(omega^T U + b):
U     <- matrix(stats::rnorm(5 * 40), 5, 40)           # raw input
D     <- 20                                            # feature dim
omega <- matrix(stats::rnorm(5 * D), 5, D)             # random freqs
b     <- stats::runif(D, 0, 2 * pi)                    # phase
Z     <- sqrt(2 / D) *
           cos(t(omega) \%*\% U + matrix(b, D, 40))      # D x 40, signed
Y     <- matrix(abs(stats::rnorm(8 * 40)), 8, 40)
res1  <- nmfkc.signed(Y, A = Z, rank = 3, maxit = 200)

## Example 2: signed Y (regression)
Y2    <- matrix(stats::rnorm(8 * 40), 8, 40)           # signed response
res2  <- nmfkc.signed(Y2, A = Z, rank = 3, maxit = 200,
                       warm.start = FALSE)
}

}
\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010).  Convex and
semi-nonnegative matrix factorizations.  \emph{IEEE TPAMI}, 32(1), 45–55.

Rahimi, A., & Recht, B. (2007).  Random features for large-scale
kernel machines.  \emph{Advances in NIPS}, 20.
}
\seealso{
\code{\link{nmfkc}}, \code{\link{predict.nmfkc.signed}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmf.ffb.R, R/nmf.sem.R
\name{nmf.ffb.cv}
\alias{nmf.ffb.cv}
\alias{nmf.sem.cv}
\title{Cross-Validation for NMF-FFB}
\usage{
nmf.ffb.cv(...)

nmf.sem.cv(
  Y1,
  Y2,
  rank = NULL,
  X.init = "nndsvd",
  X.L2.ortho = 100,
  C1.L1 = 1,
  C2.L1 = 0.1,
  epsilon = 1e-06,
  maxit = 5000,
  ...
)
}
\arguments{
\item{...}{Additional arguments passed to \code{nmf.sem} (except for
\code{rank}, \code{seed}, \code{div}, \code{shuffle}, which are handled here).
Also accepts: \code{nfolds} (number of folds, default 5; \code{div} also accepted),
\code{seed} (master random seed, default \code{NULL}),
\code{shuffle} (logical, default \code{TRUE}).}

\item{Y1}{A non-negative numeric matrix of endogenous variables with
\strong{rows = variables (P1), columns = samples (N)}.}

\item{Y2}{A non-negative numeric matrix of exogenous variables with
\strong{rows = variables (P2), columns = samples (N)}.
Must satisfy \code{ncol(Y1) == ncol(Y2)}.}

\item{rank}{Integer; rank (number of latent factors) passed to \code{nmf.sem}.
If \code{NULL}, \code{nmf.sem} decides the effective rank (via \code{...} or \code{nrow(Y2)}).}

\item{X.init}{Initialization strategy for \code{X}, forwarded to
\code{\link{nmf.sem}}.  One of \code{"nndsvd"} (default),
\code{"kmeans"}, \code{"kmeansar"}, \code{"runif"}, a numeric
\eqn{P_1 \times Q} matrix, or \code{NULL} (alias for
\code{"nndsvd"}).  See \code{\link{nmf.sem}} for details.}

\item{X.L2.ortho}{L2 orthogonality penalty for \code{X}.}

\item{C1.L1}{L1 sparsity penalty for \code{C1} (\eqn{\Theta_1}).}

\item{C2.L1}{L1 sparsity penalty for \code{C2} (\eqn{\Theta_2}).}

\item{epsilon}{Convergence threshold for \code{nmf.sem}.}

\item{maxit}{Maximum number of iterations for \code{nmf.sem}.}
}
\value{
A numeric scalar: mean MAE across CV folds.
}
\description{
Performs K-fold cross-validation to evaluate the equilibrium mapping of
the NMF-FFB model.

For each fold, \code{nmf.sem} is fitted on the training samples,
yielding an equilibrium mapping \eqn{\hat Y_1 = M_{\mathrm{model}} Y_2}.
The held-out endogenous variables \eqn{Y_1} are then predicted from \eqn{Y_2}
using this mapping, and the mean absolute error (MAE) over all entries in the
test block is computed. The returned value is the average MAE across folds.

This implements the hyperparameter selection strategy described in the paper:
hyperparameters are chosen by predictive cross-validation rather than direct
inspection of the internal structural matrices.
}
\examples{
Y <- t(iris[, -5])
Y1 <- Y[1:2, ]
Y2 <- Y[3:4, ]
mae <- nmf.sem.cv(Y1, Y2, rank = 2, maxit = 500, nfolds = 3)
mae

}
\seealso{
\code{\link{nmf.sem}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{print.nmf.rank}
\alias{print.nmf.rank}
\title{Print method for rank-selection (nmf.rank) objects}
\usage{
\method{print}{nmf.rank}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmf.rank"}.}

\item{...}{Passed to the criteria table's \code{print}.}
}
\value{
\code{x}, invisibly.
}
\description{
Print method for rank-selection (nmf.rank) objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{coef.nmf}
\alias{coef.nmf}
\alias{coef.nmf.sem}
\title{Extract coefficients from NMF models}
\usage{
\method{coef}{nmf}(object, ...)

\method{coef}{nmf.sem}(object, ...)
}
\arguments{
\item{object}{A fitted model object of class \code{"nmf"}, \code{"nmfkc"},
\code{"nmfae"}, \code{"nmfre"}, or \code{"nmf.sem"}.}

\item{...}{Not used.}
}
\value{
A data frame of coefficients (if inference was performed),
or the parameter matrix \eqn{C}.
}
\description{
Returns the \code{coefficients} data frame from a fitted NMF model
that has been passed through an inference function
(\code{\link{nmfkc.inference}}, \code{\link{nmfae.inference}},
\code{\link{nmfre.inference}}).

If inference has not been run, returns the parameter matrix \eqn{C}
(\eqn{\Theta}) directly.

For \code{nmf.sem} objects, returns \eqn{C_2} (exogenous block) as fallback.
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(1, cars$speed)
result <- nmfkc(Y, A, rank = 1)
coef(result)  # returns C matrix

result2 <- nmfkc.inference(result, Y, A)
coef(result2)  # returns coefficients data frame

}
\seealso{
\code{\link{nmfkc.inference}}, \code{\link{nmfae.inference}},
\code{\link{nmfre.inference}}, \code{\link{nmf.sem.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{summary.nmfkc}
\alias{summary.nmfkc}
\title{Summary method for objects of class \code{nmfkc}}
\usage{
\method{summary}{nmfkc}(object, ...)
}
\arguments{
\item{object}{An object of class \code{nmfkc}, i.e., the return value of \code{nmfkc}.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An object of class \code{summary.nmfkc}, containing summary statistics.
}
\description{
Produces a summary of an \code{nmfkc} object, including matrix dimensions,
runtime, fit statistics, and diagnostics.
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(1, cars$speed)
result <- nmfkc(Y, A, rank = 1)
summary(result)

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfkc.inference}}, \code{\link{plot.nmfkc}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{plot.predict.nmfae}
\alias{plot.predict.nmfae}
\title{Plot method for predict.nmfae objects}
\usage{
\method{plot}{predict.nmfae}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"predict.nmfae"} returned by \code{\link{predict.nmfae}}.}

\item{...}{Additional graphical parameters passed to \code{plot} or \code{image}.}
}
\value{
Invisible \code{NULL}. Called for its side effect (plot).
}
\description{
For \code{type = "response"}: if actual values \eqn{Y_1} were stored,
displays an observed-vs-predicted scatter plot with \eqn{R^2} in the title.
Otherwise, displays the predicted matrix as a heatmap.

For \code{type = "class"}: if actual classes were stored, displays a
confusion matrix heatmap with accuracy (ACC) in the title.
}
\examples{
\donttest{
set.seed(1)
Y <- matrix(runif(20), nrow = 4)
res <- nmfae(Y, rank = 2)
pred <- predict(res)
plot(pred)
}
}
\seealso{
\code{\link{predict.nmfae}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{print.nmf.cluster.flow}
\alias{print.nmf.cluster.flow}
\title{Print method for nmf.cluster.flow objects}
\usage{
\method{print}{nmf.cluster.flow}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmf.cluster.flow"}.}

\item{...}{Passed to the table's \code{print}.}
}
\value{
\code{x}, invisibly.
}
\description{
Prints a one-line header, the adjacent-result ARI, and the
\eqn{N \times R} cluster table (rows = individuals, columns =
results, entries = cluster number).  Use \code{\link{plot}} for the
diagram.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{plot.nmfre}
\alias{plot.nmfre}
\alias{plot.nmf.sem}
\title{Plot convergence diagnostics for NMF models}
\usage{
\method{plot}{nmfre}(x, ...)

\method{plot}{nmf.sem}(x, ..., which = c("full", "reconstruction", "both"))
}
\arguments{
\item{x}{A fitted model object.}

\item{...}{Additional graphical arguments passed to \code{\link{plot}}.}

\item{which}{For \code{plot.nmf.sem}: which objective to plot.
One of \code{"full"} (default; \code{loss + penalties}, the actual
monotonically-decreasing quantity that the multiplicative updates
minimize), \code{"reconstruction"} (Frobenius distance only,
\eqn{\| Y_1 - X B \|_F^2}), or \code{"both"} (overlay both with a
legend).  \code{"both"} is useful for diagnosing whether
regularization is actively shaping the solution: if the two curves
diverge, the penalties are pulling the optimizer away from the
pure least-squares minimum.}
}
\value{
Invisible \code{NULL}.
}
\description{
Plots the objective function value over iterations for \code{nmfre} and
\code{nmf.sem} objects. (For \code{nmfkc} and \code{nmfae}, plot methods
are defined in their respective source files.)
}
\examples{
\donttest{
set.seed(1)
Y <- matrix(runif(20), nrow = 4)
A <- diag(5)
res <- nmfre(Y, A, rank = 2, wild.bootstrap = FALSE)
plot(res)
}

}
\seealso{
\code{\link{nmfre}}, \code{\link{nmf.sem}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{plot.nmf.rank}
\alias{plot.nmf.rank}
\title{Plot a rank-selection (nmf.rank) object}
\usage{
\method{plot}{nmf.rank}(
  x,
  main = NULL,
  xlab = "Rank (Q)",
  ylab = "R-squared / eff.rank (0-1)",
  lwd = 3,
  ...
)
}
\arguments{
\item{x}{An object of class \code{"nmf.rank"}.}

\item{main}{Plot title (defaults to the title stored in \code{x}).}

\item{xlab, ylab}{Axis labels.}

\item{lwd}{Line width of the criterion curves.}

\item{...}{Further arguments passed to the initial
\code{\link[graphics]{plot}}.}
}
\value{
\code{x}, invisibly.
}
\description{
Draws the concise three-criterion rank-selection plot for an object
returned by \code{\link{nmfkc.rank}} (or \code{nmfkc.net.rank},
\code{nmfkc.signed.rank}, \code{nmfae.rank}, \code{nmfae.signed.rank}):
\code{r.squared} (red) and the effective-rank utilization
\code{eff.rank} (green) on the left \eqn{[0, 1]} axis, and the
cross-validation error \code{sigma.ecv} (blue) on the right axis, each
with points, rank-number labels and a highlighted best marker.
}
\seealso{
\code{\link{nmfkc.rank}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.ar.R
\name{nmfkc.ar.predict}
\alias{nmfkc.ar.predict}
\title{Forecast future values for NMF-VAR model}
\usage{
nmfkc.ar.predict(x, Y, degree = NULL, n.ahead = 1)
}
\arguments{
\item{x}{An object of class \code{nmfkc} (the fitted model).}

\item{Y}{The historical observation matrix used for fitting (or at least the last \code{degree} columns).}

\item{degree}{Optional integer. Lag order (D). If \code{NULL} (default), it is inferred
from \code{x$A.attr} (when available) or from the dimensions of \code{x$C}.}

\item{n.ahead}{Integer (>=1). Number of steps ahead to forecast.}
}
\value{
A list with components:
\item{pred}{A \eqn{P \times n.ahead} matrix of predicted values. Column names are future time points if time information is available.}
\item{time}{A numeric vector of future time points corresponding to the columns of \code{pred}.}
}
\description{
\code{nmfkc.ar.predict} computes multi-step-ahead forecasts for a fitted NMF-VAR model
using recursive forecasting.

If the fitted model contains time series property information (from \code{nmfkc.ar}),
the forecasted values will have appropriate time-based column names.
}
\examples{
# Forecast AirPassengers
d <- AirPassengers
ar_data <- nmfkc.ar(d, degree = 2)
result <- nmfkc(ar_data$Y, ar_data$A, rank = 1)
pred <- nmfkc.ar.predict(result, Y = matrix(d, nrow = 1), degree = 2, n.ahead = 3)
pred$pred

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfkc.ar}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{summary.nmf.sem}
\alias{summary.nmf.sem}
\title{Summary method for nmf.sem objects}
\usage{
\method{summary}{nmf.sem}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"nmf.ffb"} (or legacy
\code{"nmf.sem"}) returned by \code{\link{nmf.ffb}} /
\code{\link{nmf.sem}}.}

\item{...}{Not used.}
}
\value{
An object of class \code{"summary.nmf.sem"} (the fitted model
tagged for printing); printed by \code{\link{print.summary.nmf.sem}}.
}
\description{
Produces a formatted summary of a fitted NMF-FFB model, including
matrix dimensions, convergence, stability diagnostics, fit statistics,
and inference results (if available).
}
\examples{
Y <- t(iris[, -5])
Y1 <- Y[1:2, ]; Y2 <- Y[3:4, ]
result <- nmf.ffb(Y1, Y2, rank = 2, maxit = 500)
summary(result)

}
\seealso{
\code{\link{nmf.ffb}}, \code{\link{nmf.ffb.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{nmfae.DOT}
\alias{nmfae.DOT}
\title{DOT graph visualization for nmfae objects}
\usage{
nmfae.DOT(
  result,
  type = c("XCX", "YXCXY"),
  threshold = 0.01,
  sig.level = 0.1,
  rankdir = "LR",
  fill = TRUE,
  weight_scale = 5,
  weight_scale_x1 = weight_scale,
  weight_scale_theta = weight_scale,
  weight_scale_x2 = weight_scale,
  Y1.label = NULL,
  X1.label = NULL,
  X2.label = NULL,
  Y2.label = NULL,
  Y1.title = "Output (Y1)",
  X1.title = "Decoder (X1)",
  X2.title = "Encoder (X2)",
  Y2.title = "Input (Y2)",
  hide.isolated = TRUE
)
}
\arguments{
\item{result}{An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.}

\item{type}{Character. Graph type: \code{"XCX"} (default) or \code{"YXCXY"}.}

\item{threshold}{Numeric. Edges with values below this are omitted. Default is 0.01.}

\item{sig.level}{Numeric or \code{NULL}. Significance level for filtering C edges
when inference results are available. Only edges with p-value below \code{sig.level}
are shown, with significance stars. Set to \code{NULL} to disable. Default is 0.1.}

\item{rankdir}{Character. Graph direction for DOT layout. Default is \code{"LR"} (left to right).}

\item{fill}{Logical. If \code{TRUE}, nodes are filled with color. Default is \code{TRUE}.}

\item{weight_scale}{Numeric. Base scale factor for edge widths. Default is 5.}

\item{weight_scale_x1}{Numeric. Scale factor for \eqn{X_1} edges.}

\item{weight_scale_theta}{Numeric. Scale factor for \eqn{\Theta} edges.}

\item{weight_scale_x2}{Numeric. Scale factor for \eqn{X_2} edges.}

\item{Y1.label}{Character vector of output variable labels.}

\item{X1.label}{Character vector of decoder basis labels.}

\item{X2.label}{Character vector of encoder basis labels.}

\item{Y2.label}{Character vector of input variable labels.}

\item{Y1.title}{Character. Title for output node group. Default is \code{"Output (Y1)"}.}

\item{X1.title}{Character. Title for decoder node group. Default is \code{"Decoder (X1)"}.}

\item{X2.title}{Character. Title for encoder node group. Default is \code{"Encoder (X2)"}.}

\item{Y2.title}{Character. Title for input node group. Default is \code{"Input (Y2)"}.}

\item{hide.isolated}{Logical. If \code{TRUE} (default), Y1 and Y2 nodes that have no
edges at or above \code{threshold} are excluded from the graph. Only
applies when \code{type = "YXCXY"}.}
}
\value{
A character string containing the DOT graph specification.
}
\description{
\code{nmfae.DOT} generates a DOT language string for visualizing the structure
of a three-layer NMF model. Two graph types are supported:
\code{"XCX"} shows encoder factors, \eqn{\Theta}, and decoder factors;
\code{"YXCXY"} shows the full structure from \eqn{Y_2} through \eqn{X_2}, \eqn{\Theta},
\eqn{X_1} to \eqn{Y_1}.

Edge widths are proportional to matrix element values, and edges below
\code{threshold} are omitted for clarity.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\examples{
\donttest{
set.seed(1)
Y <- matrix(runif(20), nrow = 4)
res <- nmfae(Y, rank = 2)
dot <- nmfae.DOT(res)
}
}
\seealso{
\code{\link{nmfae}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.inference}
\alias{nmfkc.inference}
\title{Statistical inference for the parameter matrix C (Theta)}
\usage{
nmfkc.inference(object, Y, A = NULL, wild.bootstrap = TRUE, ...)
}
\arguments{
\item{object}{An object of class \code{"nmfkc"} returned by \code{\link{nmfkc}}.}

\item{Y}{Observation matrix (P x N). Must match the data used in \code{nmfkc()}.}

\item{A}{Covariate matrix (K x N). Default is \code{NULL} (same as identity;
in this case \eqn{B = C} and inference is on \eqn{B} directly).}

\item{wild.bootstrap}{Logical. If \code{TRUE} (default), performs wild bootstrap
for confidence intervals and bootstrap standard errors. Set to \code{FALSE}
to skip bootstrap (faster, only sandwich SE is computed).}

\item{...}{Additional arguments:
\describe{
\item{\code{wild.B}}{Number of bootstrap replicates. Default is 1000.}
\item{\code{wild.seed}}{Seed for bootstrap. Default is 42.}
\item{\code{wild.level}}{Confidence level for bootstrap CI. Default is 0.95.}
\item{\code{sandwich}}{Logical. Use sandwich covariance. Default is \code{TRUE}.}
\item{\code{C.p.side}}{P-value type: \code{"one.sided"} (default) or \code{"two.sided"}.}
\item{\code{cov.ridge}}{Ridge stabilization for information matrix inversion. Default is 1e-8.}
\item{\code{print.trace}}{Logical. If \code{TRUE}, prints progress. Default is \code{FALSE}.}
}}
}
\value{
An object of class \code{c("nmfkc.inference", "nmfkc")}, inheriting all
components from the input \code{object}, with additional inference components:
\item{sigma2.used}{Estimated \eqn{\sigma^2} used for inference.}
\item{C.se}{Sandwich standard errors for \eqn{C} (Q x K matrix).}
\item{C.se.boot}{Bootstrap standard errors for \eqn{C} (Q x K matrix).}
\item{C.ci.lower}{Lower CI bounds for \eqn{C} (Q x K matrix).}
\item{C.ci.upper}{Upper CI bounds for \eqn{C} (Q x K matrix).}
\item{coefficients}{Data frame with Estimate, SE, BSE, z, p-value for each element of \eqn{C}.}
\item{C.p.side}{P-value type used.}
}
\description{
\code{nmfkc.inference} performs statistical inference on the parameter matrix
\eqn{C} (\eqn{\Theta}) from a fitted \code{nmfkc} model, conditional on
the estimated basis matrix \eqn{\hat{X}}.

Under the working model \eqn{Y = X C A + \varepsilon} where
\eqn{\varepsilon_{pn} \stackrel{iid}{\sim} N(0, \sigma^2)},
inference is conducted via sandwich covariance estimation and
one-step wild bootstrap with non-negative projection.
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(intercept = 1, speed = cars$speed)
result <- nmfkc(Y, A, rank = 1)
result2 <- nmfkc.inference(result, Y, A)
summary(result2)

}
\references{
Satoh, K. (2026). Wild Bootstrap Inference for Non-Negative Matrix
Factorization with Random Effects. arXiv:2603.01468.
\url{https://arxiv.org/abs/2603.01468}
}
\seealso{
\code{\link{nmfkc}}, \code{\link{summary.nmfkc.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{summary.nmfae}
\alias{summary.nmfae}
\title{Summary method for nmfae objects}
\usage{
\method{summary}{nmfae}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An object of class \code{"summary.nmfae"}, a list with components:
\item{call}{The matched call.}
\item{dims}{Named vector \code{c(P1, P2, N)}.}
\item{Q}{Decoder rank.}
\item{R}{Encoder rank.}
\item{n.params}{Total number of parameters (P1\emph{Q + Q}R + R*P2).}
\item{autoencoder}{Logical; TRUE if P1 == P2 and Y1 was used as Y2.}
\item{niter}{Number of iterations.}
\item{runtime}{Elapsed time.}
\item{objfunc}{Final objective value.}
\item{r.squared}{R-squared.}
\item{sigma}{Residual standard error (RMSE).}
\item{mae}{Mean absolute error.}
\item{n.missing}{Number of missing elements.}
\item{prop.missing}{Percentage of missing elements.}
\item{X1.sparsity}{Proportion of near-zero elements in X1.}
\item{C.sparsity}{Proportion of near-zero elements in C.}
\item{X2.sparsity}{Proportion of near-zero elements in X2.}
}
\description{
\code{summary.nmfae} produces a summary of a fitted NMF-AE model,
including dimensions, convergence status, goodness-of-fit statistics,
and structure diagnostics (sparsity of factor matrices).
}
\seealso{
\code{\link{nmfae}}, \code{\link{print.summary.nmfae}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmf.ffb.R, R/nmf.sem.R
\name{nmf.ffb}
\alias{nmf.ffb}
\alias{nmf.sem}
\title{NMF-FFB Main Estimation Algorithm (formerly NMF-SEM)}
\usage{
nmf.ffb(
  Y1,
  Y2,
  rank = NULL,
  X.init = "nndsvd",
  X.L2.ortho = 100,
  C1.L1 = 1,
  C2.L1 = 0.1,
  epsilon = 1e-06,
  maxit = 5000,
  seed = 123,
  ...
)

nmf.sem(
  Y1,
  Y2,
  rank = NULL,
  X.init = "nndsvd",
  X.L2.ortho = 100,
  C1.L1 = 1,
  C2.L1 = 0.1,
  epsilon = 1e-06,
  maxit = 5000,
  seed = 123,
  ...
)
}
\arguments{
\item{Y1}{A non-negative numeric matrix of endogenous variables with
\strong{rows = variables (P1), columns = samples (N)}.}

\item{Y2}{A non-negative numeric matrix of exogenous variables with
\strong{rows = variables (P2), columns = samples (N)}.
Must satisfy \code{ncol(Y1) == ncol(Y2)}.}

\item{rank}{Integer; number of latent factors \eqn{Q}. If \code{NULL},
\eqn{Q} is taken from a hidden argument in \code{...} or defaults to
\code{nrow(Y2)}.}

\item{X.init}{Initialization strategy for the basis matrix
\code{X} (\eqn{P_1 \times Q}).  One of:
\itemize{
\item \code{"nndsvd"} (default): Non-negative Double SVD with
additive randomness (NNDSVDar; Boutsidis & Gallopoulos 2008),
computed internally via \code{.nndsvdar(Y1, Q)}.  Requires
\eqn{Q \le \min(P_1, N)} (over-rank case falls back to
\code{"runif"}).  Uses a full SVD of \eqn{Y_1}, so for very
large \eqn{Y_1} consider switching to \code{"kmeans"} to
avoid SVD memory / compute cost.
\item \code{"kmeans"}: k-means on the columns of \eqn{Y_1}
(samples clustered into \eqn{Q} groups); the transposed
cluster centers become \eqn{X}.  Scales well for large
\eqn{Y_1}; this is the default of \code{\link{nmfkc}}.
\item \code{"kmeansar"}: \code{"kmeans"} followed by filling
zero entries of \eqn{X} with \eqn{\mathrm{Uniform}(0,
      \bar Y_1 / 100)} (NNDSVDar-style additive randomness to
escape trivial stationary points).
\item \code{"runif"}: Uniform random entries in \eqn{[0, 1]}.
\item A numeric \eqn{P_1 \times Q} matrix supplied by the user;
negative entries are projected to 0.
\item \code{NULL}: backward-compatible alias for \code{"nndsvd"}.
}
In all cases the result is column-normalized to \code{colSums(X) = 1}
before iteration.  The menu mirrors \code{\link{nmfkc}}'s
\code{X.init} option for consistency across the package.}

\item{X.L2.ortho}{L2 orthogonality penalty for \code{X}. This controls
the penalty term \eqn{\lambda_X \lVert X^\top X - \mathrm{diag}(X^\top X)
  \rVert_F^2}. Default: \code{100}.}

\item{C1.L1}{L1 sparsity penalty for \code{C1} (i.e., \eqn{\Theta_1}).
Default: \code{1.0}.}

\item{C2.L1}{L1 sparsity penalty for \code{C2} (i.e., \eqn{\Theta_2}).
Default: \code{0.1}.}

\item{epsilon}{Relative convergence threshold for the objective function.
Iterations stop when the relative change in reconstruction loss falls
below this value. Default: \code{1e-6}.}

\item{maxit}{Maximum number of iterations for the multiplicative updates.
Default: \code{5000} (matches \code{\link{nmfkc}} and other MU
functions in the package).}

\item{seed}{Random seed used to initialize \code{X}, \code{C1}, and \code{C2}.
Default: \code{123}.}

\item{...}{Additional hidden arguments controlling the optional
feedforward baseline (used both as an \eqn{X} warm-start and as
the reference for \code{SC.map}, the input-output structural
fidelity defined in Satoh (2025) §4.SC.map):
\describe{
\item{\code{nmfkc.baseline}}{Controls whether a feedforward
\code{\link{nmfkc}}(Y1, A = Y2) fit is used as baseline.
Possible values:
\itemize{
\item Default (not given) — \code{nmf.sem} runs
\code{\link{nmfkc}} \strong{internally} when \code{X.init}
is a string method (\code{"nndsvd"}, \code{"kmeans"},
\dots) or \code{NULL}, forwarding \code{X.init},
\code{X.L2.ortho}, \code{epsilon}, \code{maxit},
\code{seed}.  The fitted \eqn{X} of the baseline is then
used as warm-start for the nmf.sem MU iterations, and
\code{SC.map} is computed.  This means
\code{nmf.sem(Y1, Y2, rank = Q)} runs end-to-end without
a prior \code{nmfkc} call.
\item \code{TRUE} — same as above, but force the internal
\code{\link{nmfkc}} call even when \code{X.init} is a
user-supplied matrix (the matrix is overridden).
\item \code{FALSE} — opt out; no internal call,
\code{SC.map = NA} (pre-v0.6.8 behavior).
\item An \code{\link{nmfkc}} result (list with \code{$X}
and \code{$C}) — use as the baseline directly (no
internal call); also adopted as \code{X.init} when the
latter is a string / NULL.
}}
\item{\code{M.simple}}{Optional \eqn{P_1 \times P_2} pre-computed
baseline mapping.  Takes precedence over \code{nmfkc.baseline}
for the SC.map calculation but does not affect warm-start.}
\item{\code{Q}}{Backward-compat alias for \code{rank}.}
}}
}
\value{
A list with components:
\item{X}{Estimated basis matrix (\eqn{P_1 \times Q}).}
\item{C1}{Estimated latent feedback matrix (\eqn{\Theta_1}, \eqn{Q \times P_1}).}
\item{C2}{Estimated exogenous loading matrix (\eqn{\Theta_2}, \eqn{Q \times P_2}).}
\item{XC1}{Feedback matrix \eqn{X \Theta_1}.}
\item{XC2}{Direct-effect matrix \eqn{X \Theta_2}.}
\item{XC1.radius}{Spectral radius \eqn{\rho(X \Theta_1)}.}
\item{XC1.norm1}{Induced 1-norm \eqn{\lVert X \Theta_1 \rVert_{1,\mathrm{op}}}.}
\item{Leontief.inv}{Leontief-type inverse \eqn{(I - X \Theta_1)^{-1}.}}
\item{M.model}{Equilibrium mapping
\eqn{M_{\mathrm{model}} = (I - X \Theta_1)^{-1} X \Theta_2}.}
\item{amplification}{Latent amplification factor
\eqn{\lVert M_{\mathrm{model}} \rVert_{1,\mathrm{op}} /
         \bigl\lVert X \Theta_2 \bigr\rVert_{1,\mathrm{op}}}.}
\item{amplification.bound}{Geometric-series upper bound
\eqn{1 / (1 - \lVert X \Theta_1 \rVert_{1,\mathrm{op}})} if
\eqn{\lVert X \Theta_1 \rVert_{1,\mathrm{op}} < 1}, otherwise \code{Inf}.}
\item{Q}{Effective latent dimension used in the fit.}
\item{SC.cov}{Correlation between sample and model-implied covariance
(flattened) of \eqn{Y_1}.  See \emph{second-moment fidelity} in
Satoh (2025).}
\item{SC.map}{Correlation between the equilibrium operator
\eqn{M_{\mathrm{model}}} and a feedforward baseline mapping
\eqn{M_{\mathrm{simple}} = X_0 \Theta_0}, computed only when the
baseline is supplied via \code{M.simple} or \code{nmfkc.baseline}
in \code{...}; otherwise \code{NA}.  See \emph{input-output
structural fidelity} in Satoh (2025).}
\item{MAE}{Mean absolute error between \eqn{Y_1} and its equilibrium
prediction \eqn{\hat Y_1 = M_{\mathrm{model}} Y_2}.}
\item{objfunc}{Vector of reconstruction losses per iteration.}
\item{objfunc.full}{Vector of penalized objective values per iteration.}
\item{iter}{Number of iterations actually performed.}
}
\description{
Fits the NMF-FFB model
\deqn{
  Y_1 \approx X \bigl( \Theta_1 Y_1 + \Theta_2 Y_2 \bigr)
}
under non-negativity constraints with orthogonality and sparsity regularization.
The function returns the estimated latent factors, structural coefficient matrices,
and the implied equilibrium (input–output) mapping.

At equilibrium, the model can be written as
\deqn{
  Y_1 \approx (I - X \Theta_1)^{-1} X \Theta_2 Y_2
  \equiv M_{\mathrm{model}} Y_2,
}
where \eqn{M_{\mathrm{model}} = (I - X \Theta_1)^{-1} X \Theta_2} is a
Leontief-type cumulative-effect operator in latent space.

Internally, the latent feedback and exogenous loading matrices are stored as
\code{C1} and \code{C2}, corresponding to \eqn{\Theta_1} and \eqn{\Theta_2},
respectively.
}
\examples{
# Simple NMF-FFB with iris data (non-negative)
Y <- t(iris[, -5])
Y1 <- Y[1:2, ]  # Sepal
Y2 <- Y[3:4, ]  # Petal
result <- nmf.sem(Y1, Y2, rank = 2, maxit = 500)
result$MAE

}
\references{
Satoh, K. (2025). Applying non-negative matrix factorization with covariates
to structural equation modeling for blind input-output analysis.
arXiv:2512.18250. \url{https://arxiv.org/abs/2512.18250}
}
\seealso{
\code{\link{nmf.sem.inference}}, \code{\link{nmf.sem.cv}},
\code{\link{nmf.sem.split}}, \code{\link{nmf.sem.DOT}},
\code{\link{summary.nmf.sem}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.net.R
\name{summary.nmfkc.net.signed}
\alias{summary.nmfkc.net.signed}
\title{Summary method for nmfkc.net.signed objects}
\usage{
\method{summary}{nmfkc.net.signed}(object, ...)
}
\arguments{
\item{object}{An \code{nmfkc.net.signed} object.}

\item{...}{Unused.}
}
\value{
An object of class \code{"summary.nmfkc.net.signed"}.
}
\description{
Summary method for nmfkc.net.signed objects
}
\seealso{
\code{\link{nmfkc.net}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.normalize}
\alias{nmfkc.normalize}
\title{Normalize a matrix to the range \eqn{[0,1]}}
\usage{
nmfkc.normalize(x, ref = x)
}
\arguments{
\item{x}{A numeric matrix (or vector) to be normalized.}

\item{ref}{A reference matrix from which the column-wise minima and maxima are taken.
Default is \code{x}.}
}
\value{
A matrix of the same dimensions as \code{x}, with each column rescaled to the \eqn{[0,1]} range.
}
\description{
\code{nmfkc.normalize} rescales the values of a matrix to lie between 0 and 1
using the column-wise minimum and maximum values of a reference matrix.
}
\examples{
# Example.
x <- nmfkc.normalize(iris[,-5])
apply(x,2,range)
}
\seealso{
\code{\link{nmfkc.denormalize}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmf.sem.R
\name{plot.nmfkc.DOT}
\alias{plot.nmfkc.DOT}
\title{Plot method for nmfkc.DOT objects}
\usage{
\method{plot}{nmfkc.DOT}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmfkc.DOT"} (or a subclass thereof).}

\item{...}{Not used.}
}
\value{
Called for its side effect (rendering). Returns \code{x} invisibly.
}
\description{
Renders a DOT graph string using \code{DiagrammeR::grViz}.
If the \pkg{DiagrammeR} package is not installed, prints the DOT source
to the console instead.

This method handles all DOT objects produced by the nmfkc package:
\code{\link{nmfkc.DOT}}, \code{\link{nmfae.DOT}}, \code{\link{nmf.sem.DOT}},
and \code{\link{nmfkc.ar.DOT}}.
}
\seealso{
\code{\link{nmfkc.DOT}}, \code{\link{nmfae.DOT}},
\code{\link{nmf.sem.DOT}}, \code{\link{nmfkc.ar.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfre.R
\name{summary.nmfre}
\alias{summary.nmfre}
\title{Summary method for objects of class \code{nmfre}}
\usage{
\method{summary}{nmfre}(object, show_ci = FALSE, ...)
}
\arguments{
\item{object}{An object of class \code{nmfre}, returned by \code{\link{nmfre}}.}

\item{show_ci}{Logical. If \code{TRUE}, show confidence interval columns (default \code{FALSE}).}

\item{...}{Additional arguments (currently unused).}
}
\value{
The input object, invisibly.
}
\description{
Displays a concise summary of an NMF-RE model fit, including dimensions,
convergence, variance components, and a coefficient table following
standard R regression output conventions.
}
\examples{
Y <- matrix(cars$dist, nrow = 1)
A <- rbind(intercept = 1, speed = cars$speed)
res <- nmfre(Y, A, rank = 1, maxit = 5000)
summary(res)

}
\seealso{
\code{\link{nmfre}}, \code{\link{nmfre.inference}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmf.cluster.flow}
\alias{nmf.cluster.flow}
\title{Cluster-flow (alluvial) diagram across a sequence of fits}
\usage{
nmf.cluster.flow(fits, reference = NULL, names = NULL, plot = TRUE, ...)
}
\arguments{
\item{fits}{A list (length \eqn{\ge 2}) of fitted models, all over the
\strong{same} \eqn{N} individuals.  The results are taken in the
given order (\strong{not} sorted), so they may be different ranks or
different models at the same rank.}

\item{reference}{The \strong{index} (1-based position in \code{fits})
of the result whose clustering defines the line colours.  Defaults
to the central result, \code{floor(length(fits) / 2) + 1} (e.g.\ the
2nd of 2 or 3 results).}

\item{names}{Optional character vector (length \code{length(fits)}) of
x-axis tick labels.  Defaults to each result's \code{$rank}.}

\item{plot}{Logical; draw the diagram immediately by calling
\code{\link{plot.nmf.cluster.flow}} (default \code{TRUE}).  Set
\code{FALSE} to only build the object and plot it later.}

\item{...}{When \code{plot = TRUE}, graphical arguments forwarded to
\code{\link{plot.nmf.cluster.flow}} (e.g.\ \code{col}, \code{lwd},
\code{xlab}, \code{ylab}, \code{main}).}
}
\value{
An object of class \code{"nmf.cluster.flow"} (returned
invisibly): a list with \code{clusters} (the \eqn{N \times R} table:
rows = individuals, columns = results, entries = cluster number = the
dominant-factor index of each fit, so it matches the factor/basis
numbering of \code{fits}; a factor that never dominates leaves an
empty, unused cluster number),
\code{ypos} (the layout positions), \code{ranks} (each result's
rank), \code{labels} (the x-axis labels), \code{reference} (the
reference index), \code{ref.cluster} (the reference hard labels),
\code{ARI} (adjusted Rand index between each pair of adjacent
results, length \eqn{R - 1}), and \code{colors}
(the default per-individual reference colour).  Call
\code{\link{plot}} on it to (re)draw the diagram.
}
\description{
Visualizes how the hard sample clustering changes across a sequence of
fitted models -- typically the same model at increasing ranks, but
also different models at the same rank.  Each individual is a line
flowing left-to-right across the results (x-axis); its vertical
position at each result is determined by its cluster, and clusters are
reordered (barycenter method) to reduce crossings.  Lines are coloured
by the individual's cluster in the \code{reference} result, so one can
see how the reference clusters split or merge.  The adjusted Rand
index (ARI) between each pair of adjacent results is printed along the
top of the figure.  X-axis ticks default to each result's \code{$rank}
and can be overridden with \code{names}.

Works for any non-negative multiplicative-update family
(\code{nmfkc}, \code{nmfae}, \code{nmfkc.net}, \code{nmfre},
and the signed variants); the hard label is the argmax of the
coefficient/score matrix.
}
\examples{
\donttest{
Y <- t(as.matrix(iris[, 1:4]))
fits <- lapply(2:6, function(q) nmfkc(Y, Q = q, print.dims = FALSE))
fl <- nmf.cluster.flow(fits, reference = 2, plot = FALSE)  # 2nd result
head(fl$clusters)
plot(fl, lwd = 2, main = "iris cluster flow")
}
}
\seealso{
\code{\link{plot.nmf.cluster.flow}},
\code{\link{nmf.cluster.criteria}}, \code{\link{nmfkc.rank}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{plot.nmfae.kernel.beta.cv}
\alias{plot.nmfae.kernel.beta.cv}
\title{Plot method for nmfae.kernel.beta.cv objects}
\usage{
\method{plot}{nmfae.kernel.beta.cv}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmfae.kernel.beta.cv"} returned by
\code{\link{nmfae.kernel.beta.cv}}.}

\item{...}{Additional graphical parameters passed to \code{plot}.}
}
\value{
Invisible \code{NULL}. Called for its side effect (plot).
}
\description{
Displays the cross-validation objective function across candidate
\code{beta} values (log scale). The optimal beta is highlighted in red.
}
\seealso{
\code{\link{nmfae.kernel.beta.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{nmfae.signed}
\alias{nmfae.signed}
\title{Signed-Bottleneck NMF-AE: Three-Layer NMF-AE with Signed Bottleneck}
\usage{
nmfae.signed(
  Y1,
  Y2 = Y1,
  rank = 2,
  rank.encoder = rank,
  epsilon = 1e-04,
  maxit = 5000,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{Y1}{Output matrix \eqn{Y_1} (P1 x N).  Negative entries allowed
(\code{Y.signed = TRUE} is auto-detected).}

\item{Y2}{Input matrix \eqn{Y_2} (P2 x N).  Must be non-negative.
Default \code{Y1} (autoencoder).}

\item{rank}{Integer.  Decoder rank Q.  Default 2.
Alias \code{Q} accepted via \code{...}.}

\item{rank.encoder}{Integer.  Encoder rank R.  Default \code{rank}.
Alias \code{R} accepted via \code{...}.}

\item{epsilon}{Relative convergence tolerance on the objective.
Default \code{1e-4}.}

\item{maxit}{Maximum iterations.  Default 5000.}

\item{verbose}{Logical.  Print progress.  Default \code{FALSE}.}

\item{...}{Additional arguments:
\describe{
\item{\code{warm.start}}{One of \code{TRUE} (default, \strong{hybrid}:
warm-start \eqn{X_1, X_2} from \code{\link{nmfae}} but initialize
\eqn{C_{+}, C_{-}} randomly), \code{"full"} (warm-start everything
including \eqn{C_{+} = C_{\mathrm{tri}}}, \eqn{C_{-} = \delta}),
or \code{FALSE} (random for all blocks).  The hybrid default avoids
the \eqn{C_{-} = 0} local-minimum trap inherited from tri-NMF-AE
while still benefiting from good \eqn{X_1, X_2} initialization.
Ignored when \eqn{Y_1} has negative entries.}
\item{\code{nstart}}{Integer, default 1 (cf. \code{\link{kmeans}}).
Number of random restarts for \eqn{C_{+}, C_{-}}.  Each restart
uses seed \code{seed + 7919 * (s-1)}.  Returns the best run by
final objective.  \strong{Signed models have more local minima
than non-negative ones} because the bottleneck
\eqn{\Theta = C_{+} - C_{-}} can take both positive and negative
values; during exploration a larger \code{nstart} (e.g., 10-50)
reduces the chance of being trapped at an inferior stationary
point (particularly the \eqn{C_{-} = 0} trap from warm-start
from non-negative tri-NMF-AE).  Use the default 1 for fast
development and raise for publication-grade runs.}
\item{\code{Y1.weights}}{Optional non-negative weight matrix
(P1 x N) or vector (length N) for \eqn{Y_1}, analogous to the
\code{weights} argument of \code{\link[stats]{lm}}.  Loss
becomes \eqn{\sum W_{ij} \, (Y_{1,ij} - \hat Y_{1,ij})^2}
(\code{lm()}-style, \strong{linear} in \eqn{W}).  Logical
matrices (\code{TRUE} / \code{FALSE}) are also accepted.
Used by \code{\link{nmfae.signed.ecv}} to hold out test
elements via a binary mask \eqn{W \in \{0,1\}}; real-valued
weights for importance weighting are also supported.  Default:
if \code{Y1} has \code{NA}, a binary mask is auto-generated
(0 for \code{NA}, 1 elsewhere).}
\item{\code{Cp.init}, \code{Cn.init}}{Explicit \eqn{Q \times R}
non-negative matrices for initialization.  Overrides warm.start.}
\item{\code{C.init}}{Explicit signed \eqn{Q \times R} matrix,
internally split into \eqn{(C_{+}, C_{-})}.}
\item{\code{X1.init}, \code{X2.init}}{Explicit basis matrices.}
\item{\code{seed}}{RNG seed.  Default 123.}
\item{\code{print.trace}}{Logical.  Print iteration trace.
Default \code{FALSE}.}
\item{\code{prefix.dec}, \code{prefix.enc}}{Label prefixes for
decoder/encoder factors.  Default \code{"Dec"}, \code{"Enc"}.}
}}
}
\value{
An object of class \code{c("nmfae.signed", "nmfae", "nmf")} with:
\item{X1}{Decoder basis (P1 x Q), column sum 1.}
\item{Cp, Cn}{Non-negative parts of \eqn{\Theta} (each Q x R).}
\item{C}{Signed bottleneck \eqn{\Theta = C_{+} - C_{-}} (Q x R).}
\item{X2}{Encoder basis (R x P2), row sum 1.}
\item{Y1hat}{Fitted values \eqn{X_1 (C_{+} - C_{-}) X_2 Y_2}.}
\item{H}{Encoding \eqn{(C_{+} - C_{-}) X_2 Y_2} (Q x N, signed).}
\item{rank}{\code{c(Q = Q, R = R)}.}
\item{dims}{\code{c(P1, P2, N)}.}
\item{objfunc, objfunc.iter}{Final and per-iteration objective values.}
\item{r.squared}{\eqn{\mathrm{cor}(Y_1, \widehat Y_1)^2} (Pearson; in \eqn{[0,1]}).}
\item{r.squared.uncentered}{Uncentered \eqn{R^2 = 1 - \|Y_1 - \widehat Y_1\|_F^2 / \|Y_1\|_F^2} (baseline = zero matrix).}
\item{r.squared.centered}{Row-mean centered \eqn{1 - \|Y_1 - \widehat Y_1\|_F^2 / \|Y_1 - \bar Y_{p\cdot}\|_F^2}.}
\item{sigma, mae}{Residual SE and mean absolute error.}
\item{niter, runtime}{Iterations and elapsed seconds.}
\item{Y.signed}{Logical; whether \eqn{Y_1} contained negative entries.}
\item{call}{Matched call.}
}
\description{
\code{nmfae.signed} fits a three-layer non-negative matrix factorization
autoencoder with a \strong{signed} bottleneck, solving
\deqn{Y_1 \approx X_1 (C_{+} - C_{-}) X_2 Y_2,
  \quad X_1 \ge 0,\; C_{+} \ge 0,\; C_{-} \ge 0,\; X_2 \ge 0,}
where \eqn{\Theta = C_{+} - C_{-}} is the signed bottleneck.  The basis
matrices \eqn{X_1} (columns sum to 1) and \eqn{X_2} (rows sum to 1) retain
their non-negative "parts-based" interpretability, while \eqn{\Theta} can
express anti-correlations (e.g., refractive index up vs. Abbe number down).

The algorithm uses \strong{Direct Multiplicative Updates} derived from
Ding et al. (2010) sign-splitting technique, applied block-wise to the
four non-negative blocks \eqn{(C_{+}, C_{-}, X_1, X_2)}.  Each block update
monotonically decreases the true objective
\eqn{\|Y_1 - X_1(C_{+} - C_{-})X_2 Y_2\|_F^2} (Lee-Seung auxiliary
function method).

\strong{Relation to \code{\link{nmfae}}:} When \eqn{\Theta \ge 0} suffices
(the \code{nmfae} case), \code{nmfae.signed} reduces to \code{nmfae} up to
the \eqn{C_{+} - C_{-}} parameterization.  Use \code{nmfae.signed} when the
data exhibit negative cross-property correlations that tri-NMF-AE cannot
express (e.g., high refractive index <-> low Abbe number trade-off).
}
\section{Lifecycle}{

This function is \strong{experimental}; interface may change.
}

\examples{
\donttest{
set.seed(1)
Y1 <- matrix(abs(rnorm(12)), 3, 4)
Y2 <- matrix(abs(rnorm(20)), 5, 4)
res <- nmfae.signed(Y1, Y2, rank = 2, rank.encoder = 2, maxit = 500)
summary(res)
}
}
\references{
Ding, C.H.Q., Li, T., and Jordan, M.I. (2010).  Convex and
Semi-Nonnegative Matrix Factorizations.
\emph{IEEE TPAMI}, 32(1), 45-55.

Satoh, K. (2026).  Signed-Bottleneck NMF-AE: Signed-Bottleneck 3-Layer NMF
(research memo, 2026-04-18).

Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae}}, \code{\link{predict.nmfae.signed}},
\code{\link{summary.nmfae.signed}}, \code{\link{nmfae.signed.rename}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.ar.R
\name{nmfkc.ar.stationarity}
\alias{nmfkc.ar.stationarity}
\title{Check stationarity of an NMF-VAR model}
\usage{
nmfkc.ar.stationarity(x)
}
\arguments{
\item{x}{The return value of \code{nmfkc} for a VAR model.}
}
\value{
A list with components:
\item{spectral.radius}{Numeric. The spectral radius of the companion matrix. A value less than 1 indicates stationarity.}
\item{stationary}{Logical. \code{TRUE} if the spectral radius is less than 1 (i.e., the system is stationary), \code{FALSE} otherwise.}
}
\description{
\code{nmfkc.ar.stationarity} assesses the dynamic stability of a VAR model
by computing the spectral radius of its companion matrix.
It returns both the spectral radius and a logical indicator of stationarity.
}
\examples{
# Check stationarity of fitted AR model
d <- AirPassengers
ar_data <- nmfkc.ar(d, degree = 2)
result <- nmfkc(ar_data$Y, ar_data$A, rank = 1)
nmfkc.ar.stationarity(result)

}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfkc.ar}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.signed.R
\name{nmfkc.signed.ecv}
\alias{nmfkc.signed.ecv}
\title{Element-wise cross-validation for nmfkc.signed}
\usage{
nmfkc.signed.ecv(Y, A, rank = 1:3, ...)
}
\arguments{
\item{Y}{Real-valued \eqn{Q_{\mathrm{obs}} \times N} response matrix
(signed entries allowed).}

\item{A}{Real-valued \eqn{D \times N} covariate matrix (signed).}

\item{rank}{Integer vector of candidate ranks (default \code{1:3}).}

\item{...}{Passed to \code{\link{nmfkc.signed}}; also accepts
\code{nfolds} (default 5; \code{div} alias), \code{seed}
(default 123).}
}
\value{
A list with \code{objfunc} (MSE per rank), \code{sigma}
(RMSE), \code{objfunc.fold} (per-fold per-rank), \code{folds},
\code{Q.grid}.
}
\description{
Element-wise k-fold CV (Wold's CV): held-out elements
are masked via \code{Y.weights = 0} during fitting, and the RMSE on
those elements is reported.  Loops over candidate \code{rank} values.
}
\section{Lifecycle}{

This function is \strong{experimental}.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfkc.signed}}, \code{\link{nmfkc.signed.cv}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{nmfae.heatmap}
\alias{nmfae.heatmap}
\title{Heatmap visualization of nmfae factor matrices}
\usage{
nmfae.heatmap(
  x,
  Y1.label = NULL,
  X1.label = NULL,
  X2.label = NULL,
  Y2.label = NULL,
  palette = NULL,
  ...
)
}
\arguments{
\item{x}{An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.}

\item{Y1.label}{Character vector of output variable names (rows of \eqn{X_1}).}

\item{X1.label}{Character vector of decoder basis labels (columns of \eqn{X_1}).}

\item{X2.label}{Character vector of encoder basis labels (rows of \eqn{X_2}).}

\item{Y2.label}{Character vector of input variable names (columns of \eqn{X_2}).}

\item{palette}{Color palette vector. Default is white-orange-red (64 colors).}

\item{...}{Not used.}
}
\value{
Invisible \code{NULL}. Called for its side effect (plot).
}
\description{
\code{nmfae.heatmap} displays the three factor matrices \eqn{X_1}, \eqn{\Theta},
and \eqn{X_2} as side-by-side heatmaps. This provides an alternative to DOT graph
visualization, especially when \eqn{Y_2} has many variables (e.g., kernel matrix).
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions; details are to be described in an upcoming paper.
}

\examples{
\donttest{
set.seed(1)
Y <- matrix(runif(20), nrow = 4)
res <- nmfae(Y, rank = 2)
nmfae.heatmap(res)
}
}
\seealso{
\code{\link{nmfae}}, \code{\link{plot.nmfae}}, \code{\link{nmfae.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.ar.R
\name{nmfkc.ar}
\alias{nmfkc.ar}
\title{Construct observation and covariate matrices for a vector autoregressive model}
\usage{
nmfkc.ar(Y, degree = 1, intercept = TRUE)
}
\arguments{
\item{Y}{An observation matrix (P x N) or a \code{ts} object.
If \code{Y} is a \code{ts} object (typically N x P), it is automatically transposed to match the (P x N) format.}

\item{degree}{The lag order of the autoregressive model. The default is 1.}

\item{intercept}{Logical. If TRUE (default), an intercept term is added to the covariate matrix.}
}
\value{
A list containing:
\item{Y}{Observation matrix (P x N_A) used for NMF. Includes adjusted \code{"tsp_info"} attribute and time-based column names.}
\item{A}{Covariate matrix (R x N_A) constructed according to the specified lag order. Includes adjusted \code{"tsp_info"} attribute and time-based column names.}
\item{A.columns}{Index matrix used to generate \code{A}.}
\item{degree.max}{Maximum lag order.}
}
\description{
\code{nmfkc.ar} generates the observation matrix and covariate matrix
corresponding to a specified autoregressive lag order.

If the input \code{Y} is a \code{ts} object, its time properties are preserved
in the \code{"tsp_info"} attribute, adjusted for the lag.
Additionally, the column names of \code{Y} and \code{A} are set to the corresponding time points.
}
\examples{
# Example using AirPassengers (ts object)
d <- AirPassengers
ar_data <- nmfkc.ar(d, degree = 2)
dim(ar_data$Y)
dim(ar_data$A)

# Example using matrix input
Y <- matrix(1:20, nrow = 2)
ar_data <- nmfkc.ar(Y, degree = 1)
ar_data$degree.max

}
\references{
Satoh, K. (2025). Applying non-negative matrix factorization with covariates
to multivariate time series data as a vector autoregression model.
\emph{Japanese Journal of Statistics and Data Science}. arXiv:2501.17446.
\doi{10.1007/s42081-025-00314-0}
}
\seealso{
\code{\link{nmfkc}}, \code{\link{nmfkc.ar.degree.cv}}, \code{\link{nmfkc.ar.stationarity}}, \code{\link{nmfkc.ar.DOT}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.signed.R
\name{predict.nmfae.signed}
\alias{predict.nmfae.signed}
\title{Predict method for nmfae.signed}
\usage{
\method{predict}{nmfae.signed}(object, newY2 = NULL, Y1 = NULL, type = c("response", "class"), ...)
}
\arguments{
\item{object}{A fitted \code{"nmfae.signed"} object.}

\item{newY2}{New input matrix (P2 x N_new).  If \code{NULL}, returns
the training fitted values.}

\item{Y1}{Optional reference Y1 for scatter / confusion plot.}

\item{type}{Output: \code{"response"} (raw signed) or \code{"class"}.}

\item{...}{Unused.}
}
\value{
A numeric matrix (\code{"response"}) or factor (\code{"class"}).
}
\description{
Computes \eqn{\hat Y_1 = X_1 (C_{+} - C_{-}) X_2 Y_2^{\mathrm{new}}}.
Since \eqn{\Theta = C_{+} - C_{-}} is signed, predictions may contain
negative entries even when \eqn{Y_1 \ge 0} in training.
}
\section{Lifecycle}{

This function is \strong{experimental}. The interface may change in
future versions.
}

\references{
Ding, C. H. Q., Li, T., & Jordan, M. I. (2010). Convex and
semi-nonnegative matrix factorizations. \emph{IEEE Transactions on
Pattern Analysis and Machine Intelligence}, 32(1), 45--55.
}
\seealso{
\code{\link{nmfae.signed}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc}
\alias{nmfkc}
\title{Optimize NMF with kernel covariates (Full Support for Missing Values)}
\usage{
nmfkc(
  Y,
  A = NULL,
  rank = NULL,
  data,
  epsilon = 1e-04,
  maxit = 5000,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{Y}{Observation matrix (P x N), OR a formula object for Formula Mode.
In Formula Mode, use \code{Y1 + Y2 ~ A1 + A2} with \code{data}, or
\code{Y_matrix ~ A_matrix} for direct matrix evaluation.
Supports dot notation (\code{. ~ A1 + A2}) when \code{data} is supplied.}

\item{A}{Covariate matrix. Default is \code{NULL} (no covariates).
Ignored when \code{Y} is a formula.}

\item{rank}{Integer. The rank of the basis matrix \eqn{X} (Q). Preferred over \code{Q}.}

\item{data}{Optional. A data frame from which variables in the formula should be taken.}

\item{epsilon}{Positive convergence tolerance.}

\item{maxit}{Maximum number of iterations.}

\item{verbose}{Logical. If \code{TRUE} (default), prints matrix dimensions and elapsed time.}

\item{...}{Additional arguments passed for fine-tuning regularization, initialization, constraints,
and output control. This includes the backward-compatible arguments \code{Q} and \code{method}.
\itemize{
\item \code{Y.weights}: Optional weight matrix (P x N) or vector
(length N) with non-negative entries, analogous to the \code{weights}
argument of \code{\link[stats]{lm}}.  When supplied, the objective
becomes \eqn{\sum W_{ij} \, (Y_{ij} - (XB)_{ij})^2}
(i.e.\ \strong{linear} in \eqn{W}; \code{lm()}-style weighted
least squares).  Logical matrices (\code{TRUE} / \code{FALSE})
are also accepted and coerced to 1 / 0.  The primary use case is
missing-value masking for ECV / CV, where \eqn{W_{ij} \in \{0, 1\}}
(\code{FALSE} / \code{TRUE}) indicates held-out vs.\ used elements;
real-valued weights for observation-level importance weighting are
also supported.  Default \code{NULL}: if \code{Y} contains \code{NA}
a binary mask is auto-constructed (0 for \code{NA}, 1 elsewhere);
otherwise no weighting.
\item \code{X.L2.ortho}: Nonnegative penalty parameter for the orthogonality of \eqn{X} (default: 0).
It minimizes the off-diagonal elements of the Gram matrix \eqn{X^\top X}, reducing the correlation
between basis vectors (conceptually minimizing \eqn{\| X^\top X - \mathrm{diag}(X^\top X) \|_F^2}).
(Formerly \code{lambda.ortho}).
\item \code{B.L1}: Nonnegative penalty parameter for L1 regularization on \eqn{B = C A} (default: 0).
Promotes \strong{sparsity in the coefficients}. (Formerly \code{gamma}).
\item \code{C.L1}: Nonnegative penalty parameter for L1 regularization on \eqn{C} (default: 0).
Promotes \strong{sparsity in the parameter matrix}. (Formerly \code{lambda}).
\item \code{Q}: Backward-compatible name for the rank of the basis matrix (Q).
\item \code{method}: Objective function: Euclidean distance \code{"EU"} (default) or Kullback–Leibler divergence \code{"KL"}.
\item \code{X.restriction}: Constraint for columns of \eqn{X}. Options: \code{"colSums"} (default), \code{"colSqSums"}, \code{"totalSum"}, \code{"none"}, or \code{"fixed"}.
\code{"none"} applies no normalization to \eqn{X} after each update, allowing it to absorb the scale freely.
\item \code{X.init}: Method for initializing the basis matrix \eqn{X}. Options: \code{"kmeans"} (default), \code{"kmeansar"}, \code{"runif"}, \code{"nndsvd"}, or a user-specified matrix. \code{"kmeansar"} applies \eqn{k}-means initialization and then fills zero entries with \code{Uniform(0, mean(Y)/100)}, analogous to NNDSVDar.
\item \code{nstart}: Number of random starts for initialization of \eqn{X} (default: 1).
Used by \code{kmeans} (when \code{X.init = "kmeans"} or \code{"kmeansar"}) and by the
multi-start evaluation (when \code{X.init = "runif"}).
\item \code{seed}: Integer seed for reproducibility (default: 123).
\item \code{C.init}: Optional numeric matrix giving the initial value of the parameter matrix \eqn{C}
(i.e., \eqn{\Theta}). If \code{A} is \code{NULL}, \code{C} has dimension \eqn{Q \times N} (equivalently \eqn{B});
otherwise, \code{C} has dimension \eqn{Q \times K} where \eqn{K = nrow(A)}. Default initializes all entries to 1.
\item \code{Y.symmetric}: \strong{Removed.} Symmetric NMF
(\eqn{Y \approx X X^\top} or \eqn{X C X^\top}) has moved to the
dedicated \code{\link{nmfkc.net}} function (types \code{"tri"},
\code{"bi"}, \code{"signed"}), which uses the correct
Frobenius bilateral-gradient updates.  Passing \code{Y.symmetric}
to \code{nmfkc()} now stops with a message pointing to
\code{nmfkc.net()}.
\item \code{prefix}: Prefix for column names of \eqn{X} and row names of \eqn{B} (default: "Basis").
\item \code{print.trace}: Logical. If \code{TRUE}, prints progress every 10 iterations (default: \code{FALSE}).
\item \code{print.dims}: Deprecated. Use \code{verbose} instead.
\item \code{detail}: Level of post-fit criterion computation.
\code{"full"} computes all criteria including silhouette, CPCC, dist.cor;
\code{"fast"} skips expensive distance-based criteria;
\code{"minimal"} returns only information criteria.
Default is \code{"full"}. For backward compatibility,
\code{save.time = TRUE} maps to \code{"fast"} and
\code{save.memory = TRUE} maps to \code{"minimal"}.
}}
}
\value{
A list with components:
\item{call}{The matched call, as captured by \code{match.call()}.}
\item{dims}{A character string summarizing the matrix dimensions of the model.}
\item{runtime}{A character string summarizing the computation time.}
\item{X}{Basis matrix. Column normalization depends on \code{X.restriction}.}
\item{B}{Coefficient matrix \eqn{B = C A}.}
\item{XB}{Fitted values for \eqn{Y}.}
\item{C}{Parameter matrix.}
\item{B.prob}{Soft-clustering probabilities derived from columns of \eqn{B}.}
\item{B.cluster}{Hard-clustering labels (argmax over \eqn{B.prob} for each column).}
\item{X.prob}{Row-wise soft-clustering probabilities derived from \eqn{X}.}
\item{X.cluster}{Hard-clustering labels (argmax over \eqn{X.prob} for each row).}
\item{A.attr}{List of attributes of the input covariate matrix \code{A}, containing metadata like lag order and intercept status if created by \code{nmfkc.ar} or \code{nmfkc.kernel}.}
\item{formula.meta}{If fitted via Formula Mode, a list with \code{formula}, \code{Y_cols}, and \code{A_cols}; otherwise \code{NULL}.}
\item{objfunc}{Final objective value.}
\item{objfunc.iter}{Objective values by iteration.}
\item{r.squared}{\eqn{R^2 = \mathrm{cor}(Y, XB)^2} (Pearson; scale-invariant; \eqn{[0,1]}).}
\item{r.squared.uncentered}{Uncentered \eqn{R^2 = 1 - \|Y - XB\|_F^2 / \|Y\|_F^2} (baseline = zero matrix; natural for non-negative factorizations without an intercept).}
\item{r.squared.centered}{Row-mean centered \eqn{R^2 = 1 - \|Y - XB\|_F^2 / \|Y - \bar Y_{p\cdot}\|_F^2}, the multivariate regression \eqn{R^2}.}
\item{method}{Character string indicating the optimization method used (\code{"EU"} or \code{"KL"}).}
\item{n.missing}{Number of missing (or zero-weighted) elements in \eqn{Y}.}
\item{n.total}{Total number of elements in \eqn{Y}.}
\item{rank}{The rank \eqn{Q} used in the factorization.}
\item{sigma}{The residual standard error, representing the typical deviation of the observed values \eqn{Y} from the fitted values \eqn{X B}.}
\item{mae}{Mean Absolute Error between \eqn{Y} and \eqn{X B}.}
\item{criterion}{A list of selection criteria: \code{silhouette} (mean silhouette width of the hard clustering, computed in the original data space \code{dist(t(Y))} with the per-sample labels), \code{CPCC} (cophenetic correlation of a hierarchical clustering of the coefficient distances \code{dist(t(B))}), \code{dist.cor} (correlation between original-data and coefficient distances), \code{B.prob.max.mean} (clustering crispness: mean dominant-cluster membership, in \eqn{[1/Q, 1]}; meaningful at fixed \eqn{Q} as a confidence check before using \code{B.cluster} as hard labels), and \code{effective.rank}.  The last is the \strong{effective rank}: \eqn{\exp} of the Shannon entropy of the explained-variance distribution \eqn{p_k = \mathrm{var}(B_{k\cdot}) / \sum_j \mathrm{var}(B_{j\cdot})}.  By the trace identity \eqn{\sum_k \mathrm{var}(B_{k\cdot}) = \mathrm{tr}(\mathrm{Cov}(B))}, \eqn{p_k} is the exact fraction of the total coefficient variance carried by factor \eqn{k}, so the entropy measures how that variance is spread across factors.  It ranges in \eqn{[1, Q]} (1 when one factor carries all the variance, \eqn{Q} when all contribute equally) and counts the number of latent factors that actively shape across-sample variation.  This is the PCA-style explained-variance / effective-dimensionality measure and reuses the \eqn{\exp(\mathrm{entropy})} functional form of Roy & Vetterli (2007).}
}
\description{
\code{nmfkc} fits a nonnegative matrix factorization with kernel covariates
under the tri-factorization model \eqn{Y \approx X C A = X B}.

This function supports two major input modes:
\enumerate{
\item \strong{Matrix Mode (Existing)}: \code{nmfkc(Y=matrix, A=matrix, ...)}
\item \strong{Formula Mode (New)}: \code{nmfkc(formula=Y_vars ~ A_vars, data=df, rank=Q, ...)}
}

The rank of the basis matrix can be specified using either the \code{rank} argument
(preferred for formula mode) or the hidden \code{Q} argument (for backward compatibility).
}
\examples{
# Example 1. Matrix Mode (Existing)
X <- cbind(c(1,0,1),c(0,1,0))
B <- cbind(c(1,0),c(0,1),c(1,1))
Y <- X \%*\% B
rownames(Y) <- paste0("P",1:nrow(Y))
colnames(Y) <- paste0("N",1:ncol(Y))
print(X); print(B); print(Y)
res <- nmfkc(Y,rank=2,epsilon=1e-6)
res$X
res$B

# Example 2. Formula Mode
set.seed(1)
dummy_data <- data.frame(Y1=rpois(10,5), Y2=rpois(10,10),
                         A1=abs(rnorm(10,5)), A2=abs(rnorm(10,3)))
res_f <- nmfkc(Y1 + Y2 ~ A1 + A2, data=dummy_data, rank=2)

# For symmetric NMF (Y approximated by X X^T or X C X^T),
# use \code{\link{nmfkc.net}()} instead.

}
\references{
Satoh, K. (2024). Applying Non-negative Matrix Factorization with Covariates
to the Longitudinal Data as Growth Curve Model. arXiv:2403.05359.
\url{https://arxiv.org/abs/2403.05359}

Satoh, K. (2025). Applying non-negative matrix factorization with covariates
to multivariate time series data as a vector autoregression model.
\emph{Japanese Journal of Statistics and Data Science}. arXiv:2501.17446.
\doi{10.1007/s42081-025-00314-0}

Satoh, K. (2025). Applying non-negative matrix factorization with covariates
to label matrix for classification. arXiv:2510.10375.
\url{https://arxiv.org/abs/2510.10375}

Ding, C., Li, T., Peng, W., & Park, H. (2006). Orthogonal Nonnegative Matrix
Tri-Factorizations for Clustering.
In \emph{Proc. 12th ACM SIGKDD} (pp. 126--135).
\doi{10.1145/1150402.1150420}

Roy, O., & Vetterli, M. (2007). The effective rank: A measure of
effective dimensionality. In \emph{15th European Signal Processing
Conference (EUSIPCO)} (pp. 606--610).
}
\seealso{
\code{\link{nmfkc.cv}}, \code{\link{nmfkc.rank}}, \code{\link{nmfkc.kernel}}, \code{\link{nmfkc.ar}}, \code{\link{predict.nmfkc}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfkc.R
\name{nmfkc.kernel.beta.cv}
\alias{nmfkc.kernel.beta.cv}
\title{Optimize beta of the Gaussian kernel function by cross-validation}
\usage{
nmfkc.kernel.beta.cv(Y, rank = 2, U, V = NULL, beta = NULL, plot = TRUE, ...)
}
\arguments{
\item{Y}{Observation matrix \eqn{Y(P,N)}.}

\item{rank}{Rank of the basis matrix.}

\item{U}{Covariate matrix \eqn{U(K,N) = (u_1, \dots, u_N)}. Each row may be normalized in advance.}

\item{V}{Covariate matrix \eqn{V(K,M) = (v_1, \dots, v_M)}, typically used for prediction. If \code{NULL}, the default is \code{U}.}

\item{beta}{A numeric vector of candidate kernel parameters to evaluate via cross-validation.}

\item{plot}{Logical. If TRUE (default), plots the objective function values for each candidate \code{beta}.}

\item{...}{Additional arguments passed to \code{nmfkc.cv}.}
}
\value{
A list with components:
\item{beta}{The beta value that minimizes the cross-validation objective function.}
\item{objfunc}{Objective function values for each candidate \code{beta}.}
}
\description{
\code{nmfkc.kernel.beta.cv} selects the optimal beta parameter of the kernel function by applying cross-validation over a set of candidate values.
}
\examples{
# Example.
Y <- matrix(cars$dist,nrow=1)
U <- matrix(c(5,10,15,20,25),nrow=1)
V <- matrix(cars$speed,nrow=1)
nmfkc.kernel.beta.cv(Y,rank=1,U,V,beta=25:30/1000)
A <- nmfkc.kernel(U,V,beta=28/1000)
result <- nmfkc(Y,A,rank=1)
plot(as.vector(V),as.vector(Y))
lines(as.vector(V),as.vector(result$XB),col=2,lwd=2)
}
\seealso{
\code{\link{nmfkc.kernel.gaussian}}, \code{\link{nmfkc.kernel.beta.nearest.med}},
\code{\link{nmfkc.kernel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmfae.R
\name{plot.nmfae}
\alias{plot.nmfae}
\title{\code{plot.nmfae} displays the convergence trajectory of the objective function
across iterations. The title shows the achieved \eqn{R^2}.}
\usage{
\method{plot}{nmfae}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nmfae"} returned by \code{\link{nmfae}}.}

\item{...}{Additional graphical parameters passed to \code{plot}.}
}
\value{
Invisible \code{NULL}. Called for its side effect (plot).
}
\description{
\code{plot.nmfae} displays the convergence trajectory of the objective function
across iterations. The title shows the achieved \eqn{R^2}.
}
\examples{
\donttest{
set.seed(1)
Y <- matrix(runif(20), nrow = 4)
res <- nmfae(Y, rank = 2)
plot(res)
}
}
\seealso{
\code{\link{nmfae}}, \code{\link{nmfae.heatmap}}
}
